library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use work.gpciepkg.all;

entity ifpga is
  generic (
    DEVICE               : string := "Stratix II GX";
    NLANE                : integer := 1
  );
  port (
    clk32                : in  std_logic;
    clk100_ext           : in  std_logic;
    mperst               : in  std_logic;

    -- PCI-Express interface
    rx_in                : in  std_logic_vector(NLANE-1 downto 0);
    tx_out               : out std_logic_vector(NLANE-1 downto 0);
    wake                 : out std_logic;

    o_led: out std_logic_vector(7 downto 0)
  );

end ifpga;

architecture default of ifpga is

  component hib
    generic (
      DEVICE              : string := DEVICE;
      USE_CLK32           : integer := 1;  -- use clock input pin 'clk32'.
      PIOWBUF_DEPTH       : integer := 10; -- 1k words = 2kB.
      TXBUF_DEPTH         : integer := 12; -- 4k words = 8kB.
      NLANE               : integer := NLANE
    );
    port (
      phy_linkup           : out std_logic;
      dl_linkup            : out std_logic;
      clk32                : in  std_logic;

      -- PCIe signals
      clk100_ext           : in  std_logic;
      mperst               : in  std_logic;
      rx_in                : in  std_logic_vector(NLANE-1 downto 0);
      tx_out               : out std_logic_vector(NLANE-1 downto 0);
      wake                 : out std_logic;

      -- global output lines
      clk_out              : out std_logic;

      -- interface to the backend 
      hib_we               : out std_logic;
      hib_data             : out std_logic_vector(NLANE*16-1 downto 0);
      backend_we           : in  std_logic;
      backend_data         : in  std_logic_vector(NLANE*16-1 downto 0);
      reset_backend        : out std_logic;  -- active high

      -- board information register initial value
      board_info           : in  std_logic_vector(31 downto 0)
    );
  end component;

  component backend
    generic (
      WIDTH               : integer := NLANE*16
    );
    port (
      -- DMA signals
      hib_we             : in  std_logic;
      hib_data           : in  std_logic_vector(WIDTH-1 downto 0);
      backend_we         : out std_logic;
      backend_data       : out std_logic_vector(WIDTH-1 downto 0);
      board_info         : out std_logic_vector(31 downto 0);
      hib_clk            : in  std_logic;
      backend_clk0       : in  std_logic;
      rst                : in  std_logic  -- negative logic
    );
  end component;

  signal hib_clk        : std_logic;
  signal hib_we         : std_logic;
  signal hib_data       : std_logic_vector(NLANE*16-1 downto 0);
  signal backend_we     : std_logic;
  signal backend_data   : std_logic_vector(NLANE*16-1 downto 0);
  signal reset_backend  : std_logic;
  signal board_info     : std_logic_vector(31 downto 0);

begin

  o_led(0) <= '1';
  
  hib_instance : hib
    port map (
      dl_linkup      => o_led(2),
      phy_linkup     => o_led(1),
      clk32          => clk32,
      -- PCIe signals
      clk100_ext     => clk100_ext,
      mperst         => mperst,
      rx_in          => rx_in,
      tx_out         => tx_out,
      wake           => wake,
      
      -- interface to the backend 
      clk_out        => hib_clk,
      hib_we         => hib_we,
      hib_data       => hib_data,
      backend_we     => backend_we,
      backend_data   => backend_data,
      reset_backend  => reset_backend,
      board_info     => board_info
    );

  backend_instance : backend
    port map (
      -- DMA signals
      hib_we         => hib_we,
      hib_data       => hib_data,
      backend_we     => backend_we,
      backend_data   => backend_data,
      board_info     => board_info,
      hib_clk        => hib_clk,
      backend_clk0   => hib_clk,
      rst            => not reset_backend
    );

end default;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

entity backend is
  generic (
    WIDTH              : integer;
    -- board info register initial value
    KFCR_PRODUCT_ID    : integer := 0;
    G7_MODEL_NO        : integer := 0;
    G7_BACKEND         : integer := 0;
    P3MCUTOFF          : integer := 0;
    EPS2_IN            : integer := 0;
    J_MEM_SIZE         : integer := 0;
    NPIPES             : integer := 8
  );
  port (
    -- DMA signals
    hib_we             : in  std_logic;
    hib_data           : in  std_logic_vector(WIDTH-1 downto 0);
    backend_we         : out std_logic;
    backend_data       : out std_logic_vector(WIDTH-1 downto 0);
    -- board info register initial value
    board_info         : out std_logic_vector(31 downto 0);

    hib_clk            : in  std_logic;
    backend_clk0       : in std_logic;
    rst                : in  std_logic  -- negative logic
  );
end backend;

architecture default of backend is

begin

  -- the simplest loop back logic
  backend_we   <= hib_we;
  backend_data <= hib_data;

  -- let the interface logic know the value of board info register
  -- below is just an example. you can hardwire it to all 0,
  -- if the register is not necessary.
  board_info <=
      conv_std_logic_vector(KFCR_PRODUCT_ID, 4) &
      conv_std_logic_vector(G7_MODEL_NO, 4) &
      conv_std_logic_vector(G7_BACKEND, 4) &
      conv_std_logic_vector(0, 4) &
      conv_std_logic_vector(P3MCUTOFF, 1) &
      conv_std_logic_vector(EPS2_IN, 1) &
      conv_std_logic_vector(0, 2) &
      conv_std_logic_vector(0, 2) &
      conv_std_logic_vector(J_MEM_SIZE, 2) &
      conv_std_logic_vector(NPIPES, 8);

end default;
