
--
-- Type0 Configuration Register
--
-- !! accepts DW (32-bit) access only.
--
-------------------------------------------------------------
-- addr    31:24      23:16      15:08      07:00
-- byte DW
-------------------------------------------------------------
-- 000  00 device_id             vendor_id
-- 004  01 status                command
-- 008  02 class_code                       revision_id
-- 00c  03 BIST       hdr_type   lat_timer  cache_line_size
-- 010  04 bar0
-- 014  05 bar1
-- 018  06 bar2
-- 01c  07 bar3
-- 020  08 bar4
-- 024  09 bar5
-- 028  0a CardBus CIS pointer
-- 02c  0b sub_device_id         sub_vendor_id
-- 030  0c bar_rom
-- 034  0d                                  cap_ptr
-- 038  0e 
-- 03c  0f max_gnt    min_gnt    int_pin    int_line
-------------------------------------------------------------
--
--
-- PCI Express Capability Register
-------------------------------------------------------------
-- addr    31:24      23:16      15:08      07:00
-- byte DW
-------------------------------------------------------------
-- 080  00 pcie_cap              next_ptr   cap_id
-- 084  01 device_cap
-- 088  02 device_status         device_ctl
-- 08c  03 link_cap
-- 090  04 link_status           link_ctl
-- 094  05 not used.
-- 098  06 not used.
-- 09c  07 not used.
-- 0a0  08 not used.
-------------------------------------------------------------
--
-- PCI Express Extended Capability Register
-------------------------------------------------------------
-- addr    31:24      23:16      15:08      07:00
-- byte 
-- 100     00000000 (which indicates not available).
-------------------------------------------------------------
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity cfgreg is
  generic (
    CFG_VENDOR_ID_INIT           : std_logic_vector(15 downto 0);
    CFG_DEVICE_ID_INIT           : std_logic_vector(15 downto 0);
    CFG_COMMAND_INIT             : std_logic_vector(15 downto 0);

    CFG_REVISION_ID_INIT         : std_logic_vector( 7 downto 0);
    CFG_CLASS_CODE_INIT          : std_logic_vector(23 downto 0);
    CFG_HEADER_TYPE_INIT         : std_logic_vector( 7 downto 0);
    CFG_BAR0_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT                : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT             : std_logic_vector(31 downto 0);
    CFG_SUB_VENDOR_ID_INIT       : std_logic_vector(15 downto 0);
    CFG_SUB_DEVICE_ID_INIT       : std_logic_vector(15 downto 0);
    CFG_INT_PIN_INIT             : std_logic_vector( 7 downto 0);

    PCIE_CAP_INIT                : std_logic_vector(15 downto 0);
    PCIE_DEVICE_CAP_INIT         : std_logic_vector(31 downto 0);
    PCIE_DEVICE_CTL_INIT         : std_logic_vector(15 downto 0);
    PCIE_DEVICE_STATUS_INIT      : std_logic_vector(15 downto 0);
    PCIE_LINK_CAP_INIT           : std_logic_vector(31 downto 0);
    PCIE_LINK_CTL_INIT           : std_logic_vector(15 downto 0);
    PCIE_LINK_STATUS_INIT        : std_logic_vector(15 downto 0);

    NLANE               : integer;
    MAX_LINK_SPEED      : std_logic_vector(3 downto 0);
    MAX_READ_REQ_SIZE   : integer;
    MAX_PAYLOAD         : integer;
    SIMULATION          : integer
  );
  port (
    clk                 : in  std_logic;
    rst                 : in  std_logic;

    mpls                : out std_logic_vector(11 downto 0);    -- max payload size
    mrrs                : out std_logic_vector(11+1 downto 0);  -- max read request size
    rcb                 : out std_logic;                        -- read completion boundary

    cfg_bar0_out        : out std_logic_vector(31 downto 0);
    cfg_bar1_out        : out std_logic_vector(31 downto 0);
    cfg_bar2_out        : out std_logic_vector(31 downto 0);
    cfg_bar3_out        : out std_logic_vector(31 downto 0);
    cfg_bar4_out        : out std_logic_vector(31 downto 0);
    cfg_bar5_out        : out std_logic_vector(31 downto 0);
    cfg_bar_rom_out     : out std_logic_vector(31 downto 0);

    rd                  : in  std_logic;
    wr                  : in  std_logic;
    addr                : in  std_logic_vector(11 downto 0);
    be                  : in  std_logic_vector(3 downto 0);
    datain              : in  std_logic_vector(31 downto 0);
    dataout             : out std_logic_vector(31 downto 0)
  );
end cfgreg;

architecture default of cfgreg is

  -- Type0 Configuration Registers
  signal cfg_vendor_id        : std_logic_vector(15 downto 0);
  signal cfg_device_id        : std_logic_vector(15 downto 0);
  signal cfg_command          : std_logic_vector(15 downto 0);
  signal cfg_status           : std_logic_vector(15 downto 0);
  signal cfg_revision_id      : std_logic_vector( 7 downto 0);
  signal cfg_class_code       : std_logic_vector(23 downto 0);
  signal cfg_header_type      : std_logic_vector(7 downto 0);
  signal cfg_cache_line_size  : std_logic_vector(7 downto 0);
  signal cfg_latency_timer    : std_logic_vector(7 downto 0);
  signal cfg_bist             : std_logic_vector(7 downto 0);
  signal cfg_bar0             : std_logic_vector(31 downto 0);
  signal cfg_bar1             : std_logic_vector(31 downto 0);
  signal cfg_bar2             : std_logic_vector(31 downto 0);
  signal cfg_bar3             : std_logic_vector(31 downto 0);
  signal cfg_bar4             : std_logic_vector(31 downto 0);
  signal cfg_bar5             : std_logic_vector(31 downto 0);
  signal cfg_cardbus_cis_ptr  : std_logic_vector(31 downto 0);
  signal cfg_sub_vendor_id    : std_logic_vector(15 downto 0);
  signal cfg_sub_device_id    : std_logic_vector(15 downto 0);
  signal cfg_bar_rom          : std_logic_vector(31 downto 0);
  signal cfg_cap_ptr          : std_logic_vector(7 downto 0);
  signal cfg_int_line         : std_logic_vector(7 downto 0);
  signal cfg_int_pin          : std_logic_vector(7 downto 0);
  signal cfg_min_gnt          : std_logic_vector(7 downto 0);
  signal cfg_max_gnt          : std_logic_vector(7 downto 0);

  -- PCI Express Capability Registers
  signal pcie_cap_id          : std_logic_vector(7 downto 0);
  signal pcie_cap_ptr         : std_logic_vector(7 downto 0);
  signal pcie_cap             : std_logic_vector(15 downto 0);
  signal pcie_device_cap      : std_logic_vector(31 downto 0);
  signal pcie_device_ctl      : std_logic_vector(15 downto 0);
  signal pcie_device_status   : std_logic_vector(15 downto 0);
  signal pcie_link_cap        : std_logic_vector(31 downto 0);
  signal pcie_link_ctl        : std_logic_vector(15 downto 0);
  signal pcie_link_status     : std_logic_vector(15 downto 0);

  -- for internal use
  signal addr0                : std_logic_vector(11 downto 0);

begin

  cfg_bar0_out        <= cfg_bar0;
  cfg_bar1_out        <= cfg_bar1;
  cfg_bar2_out        <= cfg_bar2;
  cfg_bar3_out        <= cfg_bar3;
  cfg_bar4_out        <= cfg_bar4;
  cfg_bar5_out        <= cfg_bar5;
  cfg_bar_rom_out     <= cfg_bar_rom;

  addr0 <= addr(11 downto 2) & "00";

  -- configuration read
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (rd = '1') then
        case addr0 is

          when x"000" =>
            dataout <= cfg_device_id & cfg_vendor_id;

          when x"004" =>
            dataout <= cfg_status & cfg_command;

          when x"008" =>
            dataout <= cfg_class_code & cfg_revision_id;

          when x"00c" =>
            dataout <= cfg_bist & cfg_header_type & cfg_latency_timer & cfg_cache_line_size;

          when x"010" =>
            dataout <= cfg_bar0;

          when x"014" =>
            dataout <= cfg_bar1;

          when x"018" =>
            dataout <= cfg_bar2;

          when x"01c" =>
            dataout <= cfg_bar3;

          when x"020" =>
            dataout <= cfg_bar4;

          when x"024" =>
            dataout <= cfg_bar5;

          when x"028" =>
            dataout <= cfg_cardbus_cis_ptr;

          when x"02c" =>
            dataout <= cfg_sub_device_id & cfg_sub_vendor_id;

          when x"030" =>
            dataout <= cfg_bar_rom;

          when x"034" =>
            dataout <= conv_std_logic_vector(0, 24) & cfg_cap_ptr;

          when x"038" =>
            dataout <= conv_std_logic_vector(0, 32);

          when x"03c" =>
            dataout <= cfg_max_gnt & cfg_min_gnt & cfg_int_pin & cfg_int_line;

          when x"080" =>
            dataout <= pcie_cap & pcie_cap_ptr & pcie_cap_id;

          when x"084" =>
            dataout <= pcie_device_cap;

          when x"088" =>
            dataout <= pcie_device_status & pcie_device_ctl;

          when x"08c" =>
            dataout <= pcie_link_cap;

          when x"090" =>
            dataout <= pcie_link_status & pcie_link_ctl;

          when x"100" =>
            dataout <= (others => '0');

          when others =>
            dataout <= (others => '1');

        end case;
      end if;
    end if;
  end process;

  -- configuration write
  process (clk, rst)

    variable mask16 : std_logic_vector(15 downto 0);
    variable mask32 : std_logic_vector(31 downto 0);

  begin
    if (rst = '1') then

      cfg_vendor_id       <= CFG_VENDOR_ID_INIT;
      cfg_device_id       <= CFG_DEVICE_ID_INIT;
      cfg_command         <= CFG_COMMAND_INIT;
      cfg_status          <= x"0010";
      -- bit 04        cap_list present bit. shoud be '1'.
      -- bit 15:08     error status bits. all set to zero for now.
      -- other bits    should always be zero.

      cfg_revision_id     <= CFG_REVISION_ID_INIT;
      cfg_class_code      <= CFG_CLASS_CODE_INIT;
      cfg_header_type     <= CFG_HEADER_TYPE_INIT;
      cfg_cache_line_size <= x"00";       -- not used for PCIe, but should be writable.
      cfg_latency_timer   <= x"00";       -- not used for PCIe. must be tied to 0.
      cfg_bist            <= x"00";       -- do not support buit-in self-test.
      cfg_bar0            <= CFG_BAR0_INIT;
      cfg_bar1            <= CFG_BAR1_INIT;
      cfg_bar2            <= CFG_BAR2_INIT;
      cfg_bar3            <= CFG_BAR3_INIT;
      cfg_bar4            <= CFG_BAR4_INIT;
      cfg_bar5            <= CFG_BAR5_INIT;
      cfg_cardbus_cis_ptr <= x"00000000"; -- not used.
      cfg_sub_vendor_id   <= CFG_SUB_VENDOR_ID_INIT;
      cfg_sub_device_id   <= CFG_SUB_DEVICE_ID_INIT;
      cfg_bar_rom         <= CFG_BAR_ROM_INIT;
      cfg_cap_ptr         <= x"80";       -- points to top address of PCIe capability register.
      cfg_int_line        <= x"00";       -- set by RC. 00:IRQ0 01:IRQ1 ... 0f:IRQ15
      cfg_int_pin         <= CFG_INT_PIN_INIT;
      cfg_min_gnt         <= x"00";       -- not used for PCIe.
      cfg_max_gnt         <= x"00";       -- not used for PCIe.

      pcie_cap_id         <= x"10";       -- capability ID.  10:PCIe capability register
      pcie_cap_ptr        <= x"00";       -- end of linked list.
      pcie_cap            <= PCIE_CAP_INIT;
      pcie_device_cap     <= PCIE_DEVICE_CAP_INIT(31 downto 3) &
                             conv_std_logic_vector(mylog2(MAX_PAYLOAD) - 7, 3);
                             -- 000:128B 001:256B 010:512B 011:1kB 100:2kB 101:4kB

      pcie_device_ctl     <= PCIE_DEVICE_CTL_INIT(15) &
                             conv_std_logic_vector(mylog2(MAX_READ_REQ_SIZE) - 7, 3) &
                             PCIE_DEVICE_CTL_INIT(11 downto 8) &
                             conv_std_logic_vector(0, 3) &  -- initial value of max payload size should be set to 128B.
                             PCIE_DEVICE_CTL_INIT(4 downto 0);

      pcie_device_status  <= PCIE_DEVICE_STATUS_INIT;

      pcie_link_cap       <= PCIE_LINK_CAP_INIT(31 downto 10) &
                             conv_std_logic_vector(NLANE, 6) &  -- max link width
                             MAX_LINK_SPEED;

      pcie_link_ctl       <= PCIE_LINK_CTL_INIT;
      pcie_link_status    <= PCIE_LINK_STATUS_INIT(15 downto 10) &
                             conv_std_logic_vector(NLANE, 6) &  -- max link width
                             MAX_LINK_SPEED;

    elsif (rising_edge(clk)) then

      if (wr = '1') then
        case addr0 is

          when x"004" =>
            mask16 := be_to_mask16(be(1 downto 0)) and x"0547";  -- writable bits: 10,8,6,2,1,0
            cfg_command <= (datain(15 downto 0) and mask16) or (cfg_command and not mask16);

          when x"00c" =>
            if (be(0) = '1') then
              cfg_cache_line_size <= datain(7 downto 0);
            end if;

          when x"010" =>
            mask32 := be_to_mask32(be) and (CFG_BAR0_INIT(31 downto 7) & "0000000");
            cfg_bar0 <= (datain(31 downto 0) and mask32) or (cfg_bar0 and not mask32);

          when x"014" =>
            if (CFG_BAR0_INIT(2 downto 1) = "10") then  -- 64-bit address
              cfg_bar1 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR1_INIT(31 downto 7) & "0000000");
              cfg_bar1 <= (datain(31 downto 0) and mask32) or (cfg_bar1 and not mask32);
            end if;

          when x"018" =>
            mask32 := be_to_mask32(be) and (CFG_BAR2_INIT(31 downto 7) & "0000000");
            cfg_bar2 <= (datain(31 downto 0) and mask32) or (cfg_bar2 and not mask32);

          when x"01c" =>
            if (CFG_BAR2_INIT(2 downto 1) = "10") then  -- 64-bit address
              cfg_bar3 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR3_INIT(31 downto 7) & "0000000");
              cfg_bar3 <= (datain(31 downto 0) and mask32) or (cfg_bar3 and not mask32);
            end if;

          when x"020" =>
            mask32 := be_to_mask32(be) and (CFG_BAR4_INIT(31 downto 7) & "0000000");
            cfg_bar4 <= (datain(31 downto 0) and mask32) or (cfg_bar4 and not mask32);

          when x"024" =>
            if (CFG_BAR4_INIT(2 downto 1) = "10") then  -- 64-bit address
              cfg_bar5 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR5_INIT(31 downto 7) & "0000000");
              cfg_bar5 <= (datain(31 downto 0) and mask32) or (cfg_bar5 and not mask32);
            end if;

          when x"030" =>
            mask32 := be_to_mask32(be) and (CFG_BAR_ROM_INIT(31 downto 11) & "00000000001");
            cfg_bar_rom <= (datain(31 downto 0) and mask32) or (cfg_bar_rom and not mask32);

          when x"03c" =>
            if (be(0) = '1') then
              cfg_int_line <= datain(7 downto 0);
            end if;

          when x"088" =>
              mask16 := be_to_mask16(be(1 downto 0)) and x"70e0";
              pcie_device_ctl <= (datain(15 downto 0) and mask16) or (pcie_device_ctl and not mask16);

          when x"090" =>
              mask16 := be_to_mask16(be(1 downto 0)) and x"70e0";
              pcie_link_ctl <= (datain(15 downto 0) and mask16) or (pcie_link_ctl and not mask16);

          when others =>
            -- written data is discarded.

        end case;
      end if;
    end if;
  end process;

  -- max payload size  000:128B ... 101:4kB
  process (clk)
  begin
    if (rising_edge(clk)) then
      case pcie_device_ctl(7 downto 5) is
        when "000" => 
          mpls <= conv_std_logic_vector(128, mpls'length);
        when "001" => 
          mpls <= conv_std_logic_vector(256, mpls'length);
        when "010" => 
          mpls <= conv_std_logic_vector(512, mpls'length);
        when "011" =>
          mpls <= conv_std_logic_vector(1024, mpls'length);
        when "100" => 
          mpls <= conv_std_logic_vector(2048, mpls'length);
        when "101" => 
          mpls <= conv_std_logic_vector(4096, mpls'length);
        when others => 
          mpls <= conv_std_logic_vector(128, mpls'length);
      end case;
    end if;
  end process;

  -- max read request size  000:128B ... 101:4kB
  process (clk)
  begin
    if (rising_edge(clk)) then
      case pcie_device_ctl(14 downto 12) is
        when "000" => 
          mrrs <= conv_std_logic_vector(128, mrrs'length);
        when "001" => 
          mrrs <= conv_std_logic_vector(256, mrrs'length);
        when "010" => 
          mrrs <= conv_std_logic_vector(512, mrrs'length);
        when "011" =>
          mrrs <= conv_std_logic_vector(1024, mrrs'length);
        when "100" => 
          mrrs <= conv_std_logic_vector(2048, mrrs'length);
        when "101" => 
          mrrs <= conv_std_logic_vector(4096, mrrs'length);
        when others => 
          mrrs <= conv_std_logic_vector(128, mrrs'length);
      end case;
    end if;
  end process;

  -- read completion boundary  0:64B 1:128B
  rcb  <= pcie_link_ctl(3);

end default;

--
-- DMA Controller (for each channel)
--
-- dma_control[0]    : we for dma_paddrlow_in.
--            [1]    : we for dma_paddrhigh_in.
--            [2]    : we for dma_laddr_in.
--            [3]    : we for dma_size_in.
--            [4]    : we for dma_param.
--            [5]    : kickoff a new transfer.
--            [6]    : abort currently running transfer.
-- 
-- dma_param[0]      : 0:fifo mode  1:ram mode
--          [1]      : not used.
--          [3:2]    : read latency. 00:1clk  01:2clk  10:3clk  11:4clk
--          [7:4]    : be for DW transfer. not used for burst transfer.
--          [8]      : direction. 0 for MemRd burst. 1 for MemWr burst. 
--          [10:9]   : not used.
--          [13:11]  : TC. tied to "000".
--          [15:14]  : ATTR.
-- 
-- dma_status[3]     : set to 1 if transfer is done.
--           [2:0]   : verbose diagnostics of a transfer being done.
-- 
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity dmactl is
  generic (
    NLANE             : integer;
    NDMACH            : integer
  );
  port (
    clk               : in  std_logic;
    rst               : in  std_logic;

    mpls              : in  std_logic_vector(11 downto 0);  -- max payload size
    mrrs              : in  std_logic_vector(11+1 downto 0);  -- max read request size

    -- interface to user-side logic:
    dma_control       : in  std_logic_vector(6 downto 0);
    dma_param         : in  std_logic_vector(15 downto 0);
    dma_status        : out std_logic_vector(3 downto 0);
    dma_fifocnt       : in  std_logic_vector(12 downto 0);

    dma_paddrlow_in   : in  std_logic_vector(31 downto 0);
    dma_paddrhigh_in  : in  std_logic_vector(31 downto 0);
    dma_laddr_in      : in  std_logic_vector(31 downto 0);
    dma_size_in       : in  std_logic_vector(31 downto 0);

    dma_paddrlow_out  : out std_logic_vector(31 downto 0);
    dma_paddrhigh_out : out std_logic_vector(31 downto 0);
    dma_laddr_out     : out std_logic_vector(31 downto 0);
    dma_size_out      : out std_logic_vector(31 downto 0);

    -- interface to master controller:
    dma_fifocnt_off   : in  std_logic_vector(4 downto 0);  -- width >= log2(NLANE*2)
    wrreq             : out std_logic;
    wrack             : in  std_logic;
    rdreq             : out std_logic;
    rdack             : in  std_logic;

    length            : out std_logic_vector(PAYLOAD_WIDTH-1 downto 0);
    paddr             : out std_logic_vector(63 downto 0);
    laddr             : out std_logic_vector(31 downto 0);

    -- interface to slave controller:
    rdcpl             : in  std_logic;
    rdcpl_length      : in  std_logic_vector(9 downto 0);
    rdcpl_bcnt        : in  std_logic_vector(11 downto 0);
    rdcpl_addr        : out std_logic_vector(31 downto 0)
  );
end dmactl;

architecture default of dmactl is

  type DMA_STATE is (
    IDLE,

    WAIT_Wr,
    READY_Wr,
    BURST_Wr,
    CPL_Wr,
    
    WAIT_Rd,
    READY_Rd,
    BURST_Rd,
    CPL_Rd
  );

  signal dmasm          : DMA_STATE;
  signal kickoff        : std_logic;
  signal abort          : std_logic;
  signal done           : std_logic;
  signal bcnt           : std_logic_vector(mpls'left + 1 downto 0);

  signal paddr_reg      : std_logic_vector(63 downto 0);
  signal laddr_reg      : std_logic_vector(31 downto 0);
  signal size_reg       : std_logic_vector(31 downto 0);
  signal dir_reg        : std_logic;

  signal paddr_reg_next : std_logic_vector(63 downto 0);
  signal laddr_reg_next : std_logic_vector(31 downto 0);
  signal size_reg_next  : std_logic_vector(31 downto 0);
  signal dma_fifocnt_off0 : std_logic_vector(dma_fifocnt_off'range);

  signal rdcpl_addr0    : std_logic_vector(31 downto 0);
  signal wcnt           : std_logic_vector(1 downto 0);

begin

  --
  -- DMA register read
  --
  dma_paddrhigh_out <= paddr_reg(63 downto 32);
  dma_paddrlow_out  <= paddr_reg(31 downto 0);
  dma_laddr_out     <= laddr_reg;
  dma_size_out      <= size_reg;
  dma_status        <= done & "000";
  paddr             <= paddr_reg;
  laddr             <= laddr_reg;

  process(clk, rst)
      variable bcnttmp  : std_logic_vector(bcnt'range);
      variable fcnttmp  : std_logic_vector(dma_fifocnt'range);
      variable blentmp  : std_logic_vector(length'left + 3 downto 0);
  begin
    if (rst = '1') then
      abort <= '0';
      done  <= '1';
      wrreq <= '0';
      rdreq <= '0';
      dmasm <= IDLE;
      dma_fifocnt_off0 <= (others => '0');
      wcnt             <= conv_std_logic_vector(0, wcnt'length);
    elsif (rising_edge(clk)) then

      kickoff <= '0';

      case dmasm is
        when IDLE =>
          --
          -- DMA register write
          --
          if (dma_control(0) = '1') then
            paddr_reg(31 downto 0) <= dma_paddrlow_in;
            paddr_reg_next(31 downto 0) <= dma_paddrlow_in;
          end if;

          if (dma_control(1) = '1') then
            paddr_reg(63 downto 32) <= dma_paddrhigh_in;
            paddr_reg_next(63 downto 32) <= dma_paddrhigh_in;
          else
            paddr_reg_next <= paddr_reg;
          end if;

          if (dma_control(2) = '1') then
            laddr_reg <= dma_laddr_in;
            laddr_reg_next <= dma_laddr_in;
          else
            laddr_reg_next <= laddr_reg;
          end if;

          if (dma_control(3) = '1') then
            size_reg <= dma_size_in;
            size_reg_next <= dma_size_in;
          else
            size_reg_next <= size_reg;
          end if;

          if (dma_control(4) = '1') then
            dir_reg <= dma_param(8);
          end if;

          if (dma_control(5) = '1') then
            kickoff <= '1';
          end if;

          if (dma_control(6) = '1') then
            abort <= '1';
          end if;

          if (kickoff = '1') then
            done <= '0';
            if (dir_reg = '1') then
              dmasm <= WAIT_Wr;
            else
              dmasm <= WAIT_Rd;
            end if;
          end if;

          dma_fifocnt_off0 <= (others => '0');

        when WAIT_Wr =>

          -- determine length of next burst.
          if (unsigned(size_reg_next) < unsigned(mpls)) then
            bcnttmp := size_reg_next(bcnt'range);
          else
            bcnttmp := '0' & mpls;
          end if;
          bcnt <= bcnttmp;

          -- issue request for burst write when data is ready.
          fcnttmp := unsigned(dma_fifocnt) - unsigned(dma_fifocnt_off0);
          if (unsigned(bcnttmp) <= unsigned(fcnttmp)) then
            wrreq  <= '1';
            length <= bcnttmp(bcnttmp'left - 1 downto 2);  -- size in DW unit. note that MSB is discatded,
                                                           -- i.e., length=0 denotes 4096B.
            dma_fifocnt_off0 <= (others => '0');
            dmasm  <= READY_Wr;
          end if;

          if (dma_fifocnt_off0 /= conv_std_logic_vector(0, dma_fifocnt_off0'length)) then
            dma_fifocnt_off0 <= unsigned(dma_fifocnt_off0) - NLANE * 2;
          end if;
          
          size_reg  <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;

        when READY_Wr =>
          if (wrack = '1') then
            wrreq <= '0';
            dmasm <= BURST_Wr;
          end if;

        when BURST_Wr =>
          if (wrack = '0') then
            dma_fifocnt_off0 <= dma_fifocnt_off;

            if (unsigned(size_reg) > unsigned(bcnt)) then
              dmasm <= WAIT_Wr;
            else
              done     <= '1';
              dmasm <= CPL_Wr;
            end if;
          end if;

          size_reg_next  <= unsigned(size_reg)  - unsigned(bcnt);
          paddr_reg_next <= unsigned(paddr_reg) + unsigned(bcnt);
          laddr_reg_next <= unsigned(laddr_reg) + unsigned(bcnt);

        when CPL_Wr =>
          size_reg  <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          dmasm <= IDLE;

        when WAIT_Rd =>

          -- determine length of next burst.
          if (unsigned(size_reg_next) < unsigned(mrrs)) then
            bcnttmp := size_reg_next(bcnt'range);
          else
            bcnttmp := mrrs(bcnt'range);
          end if;
          bcnt <= bcnttmp;

          -- issue a MemRd TLP when rx fifo has enough room.
          if (unsigned(bcnttmp) <= unsigned(dma_fifocnt)) then
            rdreq  <= '1';
            length <= bcnttmp(bcnttmp'left - 1 downto 2);  -- size in DW unit. note that MSB is discatded,
                                                           -- i.e., length=0 denotes 4096B.
            dmasm  <= READY_Rd;
          end if;
          
          size_reg  <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          rdcpl_addr0 <= laddr_reg_next;
          
        when READY_Rd =>
          if (rdack = '1') then
            rdreq <= '0';
            dmasm <= BURST_Rd;
          end if;

        when BURST_Rd =>
          -- received a completion TLP for current request.
          if (rdcpl = '1') then
            blentmp := '0' & rdcpl_length & "00";
            if (rdcpl_length = conv_std_logic_vector(0, rdcpl_length'length)) then
              blentmp := (blentmp'left => '1', others => '0'); -- length=0 denotes 4096B.
            end if;
            rdcpl_addr0 <= unsigned(rdcpl_addr0) + unsigned(blentmp);

            -- received all completion TLPs for current request.
            bcnttmp := '0' & rdcpl_bcnt;
            if (rdcpl_bcnt = conv_std_logic_vector(0, rdcpl_bcnt'length)) then
              bcnttmp := (bcnttmp'left => '1', others => '0'); -- bcnt=0 denotes 4096B.
            end if;
            if (unsigned(bcnttmp) <= unsigned(blentmp)) then
              if (unsigned(size_reg) > unsigned(bcnt)) then
                dmasm <= WAIT_Rd;
              else
                done     <= '1';
                dmasm <= CPL_Rd;
              end if;
            end if;
          end if;

          size_reg_next  <= unsigned(size_reg)  - unsigned(bcnt);
          paddr_reg_next <= unsigned(paddr_reg) + unsigned(bcnt);
          laddr_reg_next <= unsigned(laddr_reg) + unsigned(bcnt);

        when CPL_Rd =>
          size_reg  <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          dmasm <= IDLE;

        when others =>

      end case;

      -- abort DMA after currently running burst is completed.
      if (abort = '1') then
        if (dmasm = IDLE or dmasm = WAIT_Wr or dmasm = WAIT_Rd) then
          abort <= '0';
          done  <= '1';
          wrreq <= '0';
          rdreq <= '0';
          dmasm <= IDLE;
        end if;
      end if;
      
    end if; -- clk
  end process;

end default;
