--
-- Data-Link Layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;

entity datalink is

  generic (
    -- Flow-control buffer size
    MAX_PAYLOAD       : integer;
    CA_PH_VC0_INIT    : integer;
    CA_NPH_VC0_INIT   : integer;
    CA_CH_VC0_INIT    : integer;
    CA_PD_VC0_INIT    : integer;
    CA_NPD_VC0_INIT   : integer;
    CA_CD_VC0_INIT    : integer;
    TXDPBUF_DEPTH     : integer := 6;
    NLANE             : integer;
    SIMULATION        : integer
  );
  port (
    st2_trigger       : out std_logic;
    txdpbuffull       : out std_logic;

    clk               : in  std_logic;
    rst               : in  std_logic;   -- active high

    -- DLL-MAC interface
    rxwedp            : in  std_logic;
    rxdp              : in  std_logic_vector(47 downto 0);

    txemptydp         : out std_logic;
    txrdreqdp         : in  std_logic;
    txdp              : out std_logic_vector(47 downto 0);

    rxwetp            : in  std_logic;
    rxbetp            : in  std_logic_vector(NLANE*2-1 downto 0);
    rxtptop           : in  std_logic;
    rxtp              : in  std_logic_vector(NLANE*16-1 downto 0);

    txtoptp           : out std_logic;
    txwetp            : out std_logic;
    txbetp            : out std_logic_vector(NLANE*2-1 downto 0);
    txtp              : out std_logic_vector(NLANE*16-1 downto 0);

    -- TL-DLL interface
    rxtop             : out std_logic;
    rxwe              : out std_logic;
    rx                : out std_logic_vector(NLANE*16-1 downto 0);

    txtop             : in  std_logic;
    txwe              : in  std_logic;
    txbe              : in  std_logic_vector(NLANE*2-1 downto 0);
    tx                : in  std_logic_vector(NLANE*16-1 downto 0);
    tpsize            : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);

    -- CA notified by transaction layer
    ca_nph            : in  std_logic_vector(7 downto 0);
    ca_ph             : in  std_logic_vector(7 downto 0);
    ca_ch             : in  std_logic_vector(7 downto 0);
    ca_npd            : in  std_logic_vector(11 downto 0);
    ca_pd             : in  std_logic_vector(11 downto 0);
    ca_cd             : in  std_logic_vector(11 downto 0);
    ca_p_flush        : in  std_logic;
    ca_np_flush       : in  std_logic;
    ca_c_flush        : in  std_logic;
    
    -- notify CL to transaction layer
    cl_nph            : out std_logic_vector( 7 downto 0);
    cl_ph             : out std_logic_vector( 7 downto 0);
    cl_ch             : out std_logic_vector( 7 downto 0);
    cl_npd            : out std_logic_vector(11 downto 0);
    cl_pd             : out std_logic_vector(11 downto 0);
    cl_cd             : out std_logic_vector(11 downto 0);

    inf_cl_ph  : out std_logic;
    inf_cl_pd  : out std_logic;
    inf_cl_nph : out std_logic;
    inf_cl_npd : out std_logic;
    inf_cl_ch  : out std_logic;
    inf_cl_cd  : out std_logic;
    
    -- core status
    phy_linkup        : in  std_logic;
    dl_linkup         : out std_logic
  );
end datalink;

architecture default of datalink is
  component acknakmgr
    generic (
      NLANE       : integer := NLANE;
      MAX_PAYLOAD : integer := MAX_PAYLOAD
    );
    port (
      clk        : in  std_logic;
      rst        : in  std_logic;

      we         : in  std_logic;
      seqno      : in  std_logic_vector(11 downto 0);
      lcrcerr    : in  std_logic;

      next_seqno : out std_logic_vector(11 downto 0);
      req        : out std_logic;
      antype     : out std_logic_vector(7 downto 0);
      data       : out std_logic_vector(23 downto 0);
      ack        : in  std_logic
    );
  end component;

  component updatefcmgr
    generic (
      NLANE       : integer := NLANE;
      MAX_PAYLOAD : integer := MAX_PAYLOAD
    );
    port (
      clk        : in  std_logic;
      rst        : in  std_logic;

      -- CA notified by transaction layer
      ca_ph      : in  std_logic_vector( 7 downto 0);
      ca_nph     : in  std_logic_vector( 7 downto 0);
      ca_ch      : in  std_logic_vector( 7 downto 0);
      ca_pd      : in  std_logic_vector(11 downto 0);
      ca_npd     : in  std_logic_vector(11 downto 0);
      ca_cd      : in  std_logic_vector(11 downto 0);

      -- notify CL to transaction layer
      cl_ph      : out std_logic_vector( 7 downto 0);
      cl_nph     : out std_logic_vector( 7 downto 0);
      cl_ch      : out std_logic_vector( 7 downto 0);
      cl_pd      : out std_logic_vector(11 downto 0);
      cl_npd     : out std_logic_vector(11 downto 0);
      cl_cd      : out std_logic_vector(11 downto 0);

      inf_cl_ph  : out std_logic;
      inf_cl_pd  : out std_logic;
      inf_cl_nph : out std_logic;
      inf_cl_npd : out std_logic;
      inf_cl_ch  : out std_logic;
      inf_cl_cd  : out std_logic;
      
      -- Rx InitFC/UpdateFC DLLP
      we         : in  std_logic;
      fctype_in  : in  std_logic_vector(7 downto 0);
      hsize_in   : in  std_logic_vector(7 downto 0);
      dsize_in   : in  std_logic_vector(11 downto 0);

      -- Tx UpdateFC DLLP
      flush_p    : in  std_logic;
      flush_np   : in  std_logic;
      flush_c    : in  std_logic;
      req        : out std_logic;
      fctype_out : out std_logic_vector(7 downto 0);
      hsize_out  : out std_logic_vector(7 downto 0);
      dsize_out  : out std_logic_vector(11 downto 0);
      ack        : in  std_logic
    );
  end component;

  component txdpctl
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;

      dptype    : in  std_logic_vector(7 downto 0);
      data      : in  std_logic_vector(23 downto 0);
      kickoff   : in  std_logic;

      dp        : out std_logic_vector(47 downto 0);
      done      : out std_logic
    );
  end component;

  component rxdpctl
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;

      we        : in  std_logic;
      dp        : in  std_logic_vector(47 downto 0);

      dptype    : out std_logic_vector(7 downto 0);
      dpbody    : out std_logic_vector(23 downto 0);
      crcerr    : out std_logic;
      done      : out std_logic
    );
  end component;

  component txtpctl
    generic (
      NLANE      : integer := NLANE
    );
    port (
      topin      : in  std_logic;
      wein       : in  std_logic;
      bein       : in  std_logic_vector(NLANE*2-1 downto 0);
      tpbody     : in  std_logic_vector(NLANE*16-1 downto 0);
      tpbodysize : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);

      topout     : out std_logic;
      weout      : out std_logic;
      beout      : out std_logic_vector(NLANE*2-1 downto 0);
      tp         : out std_logic_vector(NLANE*16-1 downto 0);

      clk        : in  std_logic;
      rst        : in  std_logic
      );
  end component;

  component rxtpctl
    generic (
      NLANE      : integer := NLANE
    );
    port (
      clk        : in  std_logic;
      rst        : in  std_logic;

      wein       : in  std_logic;
      be         : in  std_logic_vector(NLANE*2-1 downto 0);
      tptop      : in  std_logic;
      tp         : in  std_logic_vector(NLANE*16-1 downto 0);
      next_seqno : in  std_logic_vector(11 downto 0);

      topout     : out std_logic;
      weout      : out std_logic;
      tpbody     : out std_logic_vector(NLANE*16-1 downto 0);
      seqno      : out std_logic_vector(11 downto 0);
      crcerr     : out std_logic;
      done       : out std_logic
    );
  end component;

  component scfifo
    generic (
      ADD_RAM_OUTPUT_REGISTER : string := "ON";
      INTENDED_DEVICE_FAMILY  : string := "Stratix II GX";
      LPM_NUMWORDS            : natural;
      LPM_SHOWAHEAD           : string := "ON";
      LPM_TYPE                : string := "scfifo";
      LPM_WIDTH               : natural;
      OVERFLOW_CHECKING       : string := "OFF";
      UNDERFLOW_CHECKING      : string := "OFF";
      USE_EAB                 : string := "ON"
    );
    port (
      rdreq    : in  std_logic;
      empty    : out std_logic;
      clock    : in  std_logic;
      sclr     : in  std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
      wrreq    : in  std_logic;
      data     : in  std_logic_vector(LPM_WIDTH-1 downto 0);
      full     : out std_logic 
    );
  end component;

  -- DLCMSM States
  type DLCMSM_STATE is (
    DL_INACTIVE,
    DL_INIT1,
    DL_INIT2,
    DL_ACTIVE,
    INVALID_STATE
  );

--  attribute enum_encoding: string;
--  attribute enum_encoding of DLCMSM_STATE : type is "000 001 010 011 100 101 110 111";
  
  -- internal use
  signal dlcmsm           : DLCMSM_STATE;
  signal prev_dlcmsm      : DLCMSM_STATE;

  signal txdp_type        : std_logic_vector(7 downto 0);
  signal txdp_data        : std_logic_vector(23 downto 0);
  signal txdp_kickoff     : std_logic;
  signal txdp_done_pulse  : std_logic;
  signal txdp_done_p      : std_logic;
  signal txdp_done_np     : std_logic;
  signal txdp_done_c      : std_logic;
  signal txdp_dp          : std_logic_vector(47 downto 0);
  signal txdp_clear       : std_logic;

  signal rxdp_type        : std_logic_vector(7 downto 0);
  signal rxdp_data        : std_logic_vector(23 downto 0);
  signal rxdp_dcrcerr     : std_logic;
  signal rxdp_done_pulse  : std_logic;
  signal rxdp_done_p      : std_logic;
  signal rxdp_done_np     : std_logic;
  signal rxdp_done_c      : std_logic;

  signal rxtp_seqno         : std_logic_vector(11 downto 0);
  signal rxtp_lcrcerr       : std_logic;
  signal rxtp_done_pulse    : std_logic;

  signal an_req             : std_logic;
  signal an_ack             : std_logic;
  signal an_type            : std_logic_vector(7 downto 0);
  signal an_data            : std_logic_vector(23 downto 0);
  signal next_seqno         : std_logic_vector(11 downto 0);

  signal fc_we              : std_logic;
  signal fc_type_in         : std_logic_vector(7 downto 0);
  signal fc_hsize_in        : std_logic_vector(7 downto 0);
  signal fc_dsize_in        : std_logic_vector(11 downto 0);

  signal fc_req             : std_logic;
  signal fc_ack             : std_logic;
  signal fc_type_out        : std_logic_vector(7 downto 0);
  signal fc_hsize_out       : std_logic_vector(7 downto 0);
  signal fc_dsize_out       : std_logic_vector(11 downto 0);

  signal dl_linkup0         : std_logic;
  signal txdpcleared        : std_logic;

-- synthesis translate_off
  attribute keep : boolean;
  attribute keep of txdp_done_pulse : signal is true;
  attribute keep of txdpbuffull     : signal is true;
  attribute keep of rxtp_seqno      : signal is true;
  attribute keep of rxtp_lcrcerr    : signal is true;
  attribute keep of rxtp_done_pulse : signal is true;
-- synthesis translate_on
    
begin

  dl_linkup     <= dl_linkup0;
  
  process (clk, rst)
  begin
    if (rst = '1') then

      dl_linkup0    <= '0';
      txdp_kickoff  <= '0';
      txdp_done_p   <= '0';
      txdp_done_np  <= '0';
      txdp_done_c   <= '0';
      rxdp_done_p   <= '0';
      rxdp_done_np  <= '0';
      rxdp_done_c   <= '0';

      if (SIMULATION = 0) then
        dlcmsm <= DL_INACTIVE;
      else
        dlcmsm <= DL_ACTIVE; -- skip training sequence for simulation.
      end if;

    elsif (rising_edge(clk)) then

      dl_linkup0 <= '0';
      prev_dlcmsm <= dlcmsm;

      case dlcmsm is

        when DL_INACTIVE =>
          txdp_kickoff  <= '0';
          txdp_done_p   <= '0';
          txdp_done_np  <= '0';
          txdp_done_c   <= '0';
          rxdp_done_p   <= '0';
          rxdp_done_np  <= '0';
          rxdp_done_c   <= '0';
          txdpcleared   <= '0';

          if (phy_linkup = '1') then
            dlcmsm <= DL_INIT1;
          end if;

        when DL_INIT1 =>
          -- repeatedly send InitFC1-P,Np,Cpl w/Credit Allocated.
          if (dlcmsm /= prev_dlcmsm) then
            txdp_type     <= DPTYPE_INITFC1_P_VC0;

#if 1 // remember to modify DL_INIT2 when you activate this part.
            txdp_data     <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                             "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
#else // infinite CA
            txdp_data     <= "00" & conv_std_logic_vector(0, 8) &
                             "00" & conv_std_logic_vector(0, 12);
#endif
            txdp_kickoff  <= '1';
          elsif (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC1_C_VC0) then
              txdp_type    <= DPTYPE_INITFC1_P_VC0;
#if 1
              txdp_data    <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
#else // infinite CA
              txdp_data     <= "00" & conv_std_logic_vector(0, 8) &
                               "00" & conv_std_logic_vector(0, 12);
#endif
              txdp_kickoff <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_P_VC0) then
              txdp_type    <= DPTYPE_INITFC1_NP_VC0;
              txdp_data    <= "00" & conv_std_logic_vector(CA_NPH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_NPD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            else
              txdp_type    <= DPTYPE_INITFC1_C_VC0;
              txdp_data    <= "00" & conv_std_logic_vector(CA_CH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_CD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            end if;
          else
            txdp_kickoff <= '0';
          end if;

          -- sent at least one InitFC1-P,Np,Cpl.
          if (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC1_P_VC0) then
              txdp_done_p   <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_NP_VC0) then
              txdp_done_np  <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_C_VC0) then
              txdp_done_c <= '1';
            end if;
          end if;

          -- receive InitFC1-P,Np,Cpl and set Credit Limit.
          fc_we <= '0';
          if (rxdp_done_pulse = '1') then
            if ((rxdp_type = DPTYPE_INITFC1_P_VC0) or
               (rxdp_type = DPTYPE_INITFC1_NP_VC0) or
               (rxdp_type = DPTYPE_INITFC1_C_VC0)) then
              fc_we       <= '1';
              fc_type_in  <= rxdp_type;
              fc_hsize_in <= rxdp_data(21 downto 14);
              fc_dsize_in <= rxdp_data(11 downto  0);
              if (rxdp_type = DPTYPE_INITFC1_P_VC0) then
                rxdp_done_p   <= '1';
              elsif (rxdp_type = DPTYPE_INITFC1_NP_VC0) then
                rxdp_done_np  <= '1';
              elsif (rxdp_type = DPTYPE_INITFC1_C_VC0) then
                rxdp_done_c <= '1';
              end if;
            end if;
          end if;

          -- sent and received at least one InitFC1-P,Np,Cpl.
          if (txdp_done_p = '1' and txdp_done_np = '1' and txdp_done_c = '1' and
              rxdp_done_p = '1' and rxdp_done_np = '1' and rxdp_done_c = '1') then 
            txdp_kickoff  <= '0';
            txdp_done_p   <= '0';
            txdp_done_np  <= '0';
            txdp_done_c   <= '0';
            rxdp_done_p   <= '0';
            rxdp_done_np  <= '0';
            rxdp_done_c   <= '0';
            fc_we <= '0';
            dlcmsm <= DL_INIT2;
          end if;

          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;
  
        when DL_INIT2 =>
--          dl_linkup0 <= '1';

          -- repeatedly send InitFC2-P,Np,Cpl w/Credit Allocated.
          if (dlcmsm /= prev_dlcmsm) then
            txdp_type     <= DPTYPE_INITFC2_P_VC0;
#if 1
            txdp_data     <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                             "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
#else // infinite CA
            txdp_data     <= "00" & conv_std_logic_vector(0, 8) &
                             "00" & conv_std_logic_vector(0, 12);
#endif
            txdp_kickoff  <= '1';
          elsif (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC2_C_VC0) then
              txdp_type    <= DPTYPE_INITFC2_P_VC0;
#if 1
              txdp_data    <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
#else // infinite CA
              txdp_data     <= "00" & conv_std_logic_vector(0, 8) &
                               "00" & conv_std_logic_vector(0, 12);
#endif
              txdp_kickoff <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_P_VC0) then
              txdp_type    <= DPTYPE_INITFC2_NP_VC0;
              txdp_data    <= "00" & conv_std_logic_vector(CA_NPH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_NPD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            else
              txdp_type    <= DPTYPE_INITFC2_C_VC0;
              txdp_data    <= "00" & conv_std_logic_vector(CA_CH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_CD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            end if;
          else
            txdp_kickoff <= '0';
          end if;

          -- sent at least one InitFC2-P,Np,Cpl.
          if (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC2_P_VC0) then
              txdp_done_p   <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_NP_VC0) then
              txdp_done_np  <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_C_VC0) then
              txdp_done_c <= '1';
            end if;
          end if;

          -- receive InitFC2-P,Np,Cpl and set Credit Limit.
          if (rxdp_done_pulse = '1') then
            if (rxdp_type = DPTYPE_INITFC2_P_VC0) then
              rxdp_done_p   <= '1';
            elsif (rxdp_type = DPTYPE_INITFC2_NP_VC0) then
              rxdp_done_np  <= '1';
            elsif (rxdp_type = DPTYPE_INITFC2_C_VC0) then
              rxdp_done_c <= '1';
            end if;
          end if;

          -- sent and received at least one InitFC2-P,Np,Cpl.
          if (txdp_done_p = '1' and txdp_done_np = '1' and txdp_done_c = '1' and
              rxdp_done_p = '1' and rxdp_done_np = '1' and rxdp_done_c = '1') then 
            txdp_kickoff  <= '0';
            txdp_done_p   <= '0';
            txdp_done_np  <= '0';
            txdp_done_c   <= '0';
            rxdp_done_p   <= '0';
            rxdp_done_np  <= '0';
            rxdp_done_c   <= '0';
            dlcmsm <= DL_ACTIVE;
          end if;

          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;

        when DL_ACTIVE =>

          st2_trigger <= '0';

#if 1 // for debugging
          if (rxdp_done_pulse = '1') then
            -- Ack/Nak
            if (rxdp_type = DPTYPE_NAK) then
              st2_trigger <= '1';
            end if;
          end if;
#endif

          dl_linkup0   <= '1';
          txdp_clear   <= '0';
          an_ack       <= '0';
          fc_ack       <= '0';
          txdp_kickoff <= '0';
          -- transmit Ack/Nak DLLP.
          if (an_req = '1' and an_ack = '0') then
            txdp_type     <= an_type;
            txdp_data     <= an_data;
            txdp_kickoff  <= '1';
            an_ack        <= '1';

            -- the 1st TLP from Root Complex arrived.
            -- clear unnecessary InitFC2s in TxDPbuf.
            if (an_data = conv_std_logic_vector(0, an_data'length) and txdpcleared = '0') then
              txdp_clear <= '1';
              txdpcleared <= '1';
            end if;

          -- transmit UpdateFC DLLP.
          elsif (fc_req = '1' and fc_ack = '0') then
            txdp_type    <= fc_type_out;
            txdp_data    <= "00" & fc_hsize_out & "00" & fc_dsize_out;
            txdp_kickoff <= '1';
            fc_ack       <= '1';
          end if;

          -- receive UpdateFC-P,Np,Cpl and set Credit Limit.
          fc_we <= '0';
          if (rxdp_done_pulse = '1') then
            if ((rxdp_type = DPTYPE_UPDATEFC_P_VC0) or
               (rxdp_type = DPTYPE_UPDATEFC_NP_VC0) or
               (rxdp_type = DPTYPE_UPDATEFC_C_VC0)) then
              fc_we       <= '1';
              fc_type_in  <= rxdp_type;
              fc_hsize_in <= rxdp_data(21 downto 14);
              fc_dsize_in <= rxdp_data(11 downto  0);
            end if;
          end if;

          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;
  
        when others =>
          dlcmsm <= DL_INACTIVE;

      end case;
    end if;
  end process;

  acknakmgr0 : acknakmgr
    port map (
      clk         => clk,
      rst         => not dl_linkup0,

      we          => rxtp_done_pulse,
      seqno       => rxtp_seqno,
      lcrcerr     => rxtp_lcrcerr,
      
      next_seqno  => next_seqno,
      req         => an_req,
      antype      => an_type,
      data        => an_data,
      ack         => an_ack
    );
  
  updatefcmgr0 : updatefcmgr
    port map (
      clk         => clk,
      rst         => rst,

      ca_nph      => ca_nph,
      ca_ph       => ca_ph,
      ca_ch       => ca_ch,
      ca_npd      => ca_npd,
      ca_pd       => ca_pd,
      ca_cd       => ca_cd,

      cl_nph      => cl_nph,
      cl_ph       => cl_ph,
      cl_ch       => cl_ch,
      cl_npd      => cl_npd,
      cl_pd       => cl_pd,
      cl_cd       => cl_cd,

      inf_cl_ph   => inf_cl_ph,
      inf_cl_pd   => inf_cl_pd,
      inf_cl_nph  => inf_cl_nph,
      inf_cl_npd  => inf_cl_npd,
      inf_cl_ch   => inf_cl_ch,
      inf_cl_cd   => inf_cl_cd,
      
      we          => fc_we,
      fctype_in   => fc_type_in,
      hsize_in    => fc_hsize_in,
      dsize_in    => fc_dsize_in,

      flush_p     => ca_p_flush,
      flush_np    => ca_np_flush,
      flush_c     => ca_c_flush,
      req         => fc_req,
      fctype_out  => fc_type_out,
      hsize_out   => fc_hsize_out,
      dsize_out   => fc_dsize_out,
      ack         => fc_ack
    );

  txdpctl0 : txdpctl
    port map (
      clk      => clk,
      rst      => rst,

      dptype   => txdp_type,
      data     => txdp_data,
      kickoff  => txdp_kickoff,

      dp       => txdp_dp,
      done     => txdp_done_pulse
    );

  rxdpctl0 : rxdpctl
    port map (
      clk      => clk,
      rst      => rst,

      we       => rxwedp,
      dp       => rxdp,

      dptype   => rxdp_type,
      dpbody   => rxdp_data,
      crcerr   => rxdp_dcrcerr,
      done     => rxdp_done_pulse
    );

  txdpbuf0 : scfifo
    generic map (
      LPM_NUMWORDS => 2**TXDPBUF_DEPTH,
      LPM_WIDTH    => 48
    )
    port map (
      clock => clk,
      sclr  => rst or txdp_clear,

      wrreq => txdp_done_pulse,
      data  => txdp_dp,

      rdreq => txrdreqdp,
      empty => txemptydp,
--      full  => open,
      full  => txdpbuffull,
      q     => txdp
    );

  rxtpctl0 : rxtpctl
    port map (
      clk        => clk,
      rst      => not dl_linkup0,

      wein       => rxwetp,
      be         => rxbetp,
      tptop      => rxtptop,
      tp         => rxtp,
      next_seqno => next_seqno,

      topout     => rxtop,
      weout      => rxwe,
      tpbody     => rx,
      seqno      => rxtp_seqno,
      crcerr     => rxtp_lcrcerr,
      done       => rxtp_done_pulse
    );
  
  txtpctl0 : txtpctl
    port map (
      clk        => clk,
      rst        => rst,

      topin      => txtop,
      wein       => txwe,
      bein       => txbe,
      tpbody     => tx,
      tpbodysize => tpsize,

      topout     => txtoptp,
      weout      => txwetp,
      beout      => txbetp,
      tp         => txtp
    );
  
end default;

--
-- Ack/Nak Manager
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity acknakmgr is
  generic (
    NLANE       : integer;
    MAX_PAYLOAD : integer
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;

    we         : in  std_logic;
    seqno      : in  std_logic_vector(11 downto 0);
    lcrcerr    : in  std_logic;

    next_seqno : out std_logic_vector(11 downto 0);
    req        : out std_logic;
    antype     : out std_logic_vector(7 downto 0);
    data       : out std_logic_vector(23 downto 0);
    ack        : in  std_logic
  );
end acknakmgr;

architecture default of acknakmgr is

  signal latency          : std_logic_vector(15 downto 0);
  signal timer            : std_logic_vector(15 downto 0);
  signal next_rcv_seq     : std_logic_vector(11 downto 0);
  signal ack_scheduled    : std_logic;
  signal nak_scheduled    : std_logic;
  signal last_good_seq    : std_logic_vector(11 downto 0);

begin

  data <= conv_std_logic_vector(0, 12) & last_good_seq;
  next_seqno <= next_rcv_seq;

  process (clk, rst)
    variable reqtmp  : std_logic;
  begin
    if (rst = '1') then
      last_good_seq <= (others => '1'); -- no good-TLP received yet.
      next_rcv_seq  <= (others => '0');
      ack_scheduled <= '0';
      nak_scheduled <= '0';
      req <= '0';
      timer         <= (others => '0');
    elsif (rising_edge(clk)) then

      reqtmp := '0';
      if (we = '1') then

        if (lcrcerr = '0' and seqno = next_rcv_seq) then  -- a new TLP coming in.

          -- do not schedule any Ack/Nak here.
          -- just update seq# and type.
          last_good_seq <= seqno;
          antype        <= DPTYPE_ACK;
          next_rcv_seq  <= unsigned(next_rcv_seq) + 1;
          ack_scheduled <= '1';
          nak_scheduled <= '0';

          reqtmp := '1'; -- transmit Ack immediately.  !!!

        elsif (lcrcerr = '0' and valid_seqno(seqno, next_rcv_seq) = '1') then -- a replayed TLP coming in.

          antype        <= DPTYPE_ACK;
          ack_scheduled <= '1';
          nak_scheduled <= '0';
          reqtmp := '1'; -- transmit Ack immediately.

        elsif (nak_scheduled = '0') then  -- an invalid TLP coming in.
          -- schedule a Nak.
          -- only one Nak is scheduled for a sequence of bad-TLPs.
          nak_scheduled <= '1';
          antype <= DPTYPE_NAK;
          reqtmp := '1';

        end if;

      end if;

      -- flush an Ack due to latency-timer expiration.
      if (timer > latency) then
        reqtmp := '1';
      elsif (ack_scheduled = '1') then
        timer <= unsigned(timer) + 1;
      end if;

      if (reqtmp = '1') then
        req           <= '1';
      elsif (ack = '1') then  -- scheduled request is acknowledged by TxDPctl and transmitted.
        req           <= '0';
        ack_scheduled <= '0';
        nak_scheduled <= '0';
        timer         <= (others => '0');
      end if;

      -- 
      -- Ack/Nak latency timer load value (in 125MHz UI, i.e., 8 ns unit):
      -- 
      case NLANE is
        when 1 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(237/2 - (20+128)/1/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(416/2 - (20+256)/1/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when 4 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(73/2 - (20+128)/4/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(118/2 - (20+256)/4/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when 8 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(67/2 - (20+128)/8/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(107/2 - (20+256)/8/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when others =>
          latency <= conv_std_logic_vector(100/2, latency'length);

      end case;

    end if;
  end process;

end default;


--
-- UpdateFC Manager
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity updatefcmgr is
  generic (
    NLANE       : integer;
    MAX_PAYLOAD : integer
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;

    -- CA notified by transaction layer
    ca_ph      : in  std_logic_vector( 7 downto 0);
    ca_nph     : in  std_logic_vector( 7 downto 0);
    ca_ch      : in  std_logic_vector( 7 downto 0);
    ca_pd      : in  std_logic_vector(11 downto 0);
    ca_npd     : in  std_logic_vector(11 downto 0);
    ca_cd      : in  std_logic_vector(11 downto 0);

    -- notify CL to transaction layer
    cl_ph      : out std_logic_vector( 7 downto 0);
    cl_nph     : out std_logic_vector( 7 downto 0);
    cl_ch      : out std_logic_vector( 7 downto 0);
    cl_pd      : out std_logic_vector(11 downto 0);
    cl_npd     : out std_logic_vector(11 downto 0);
    cl_cd      : out std_logic_vector(11 downto 0);

    inf_cl_ph  : out std_logic;
    inf_cl_pd  : out std_logic;
    inf_cl_nph : out std_logic;
    inf_cl_npd : out std_logic;
    inf_cl_ch  : out std_logic;
    inf_cl_cd  : out std_logic;
    
    -- Rx InitFC/UpdateFC DLLP
    we         : in  std_logic;
    fctype_in  : in  std_logic_vector(7 downto 0);
    hsize_in   : in  std_logic_vector(7 downto 0);
    dsize_in   : in  std_logic_vector(11 downto 0);

    -- Tx UpdateFC DLLP
    flush_p    : in  std_logic;
    flush_np   : in  std_logic;
    flush_c    : in  std_logic;
    req        : out std_logic;
    fctype_out : out std_logic_vector(7 downto 0);
    hsize_out  : out std_logic_vector(7 downto 0);
    dsize_out  : out std_logic_vector(11 downto 0);
    ack        : in  std_logic
  );
end updatefcmgr;

architecture default of updatefcmgr is

  signal latency       : std_logic_vector(15 downto 0);
  signal timer         : std_logic_vector(15 downto 0);
  signal expired       : std_logic;

 constant TIME30us    : std_logic_vector(timer'left downto 0) := x"0ea6";
--  constant TIME30us    : std_logic_vector(timer'left downto 0) := x"0da6";  -- !! 30us - 4kB/16 clk margin.
--  constant TIME120us   : std_logic_vector(timer'left downto 0) := x"3a98";

  type FCOUT_STATE is (
    IDLE,
    REQ_NP,
    REQ_P,
    REQ_C
  );

--  attribute enum_encoding: string;
--  attribute enum_encoding of FCOUT_STATE : type is "00 01 10 11";

  signal fcoutsm       : FCOUT_STATE;
  signal flush_p0      : std_logic;
  signal flush_np0     : std_logic;
  signal flush_c0      : std_logic;

begin

  --
  -- Reception of Credit Limit:
  -- 1) take in 'hsize_in' and 'dsize_in' with 'we'.
  -- 2) update CL of type specified by 'fctype_in'.
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      inf_cl_ph  <= '0';
      inf_cl_pd  <= '0';
      inf_cl_nph <= '0';
      inf_cl_npd <= '0';
      inf_cl_ch  <= '0';
      inf_cl_cd  <= '0';
    elsif (rising_edge(clk)) then
      if (we = '1') then
        case fctype_in is

          when DPTYPE_INITFC1_P_VC0 =>
            cl_ph <= hsize_in;
            cl_pd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_ph <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_pd <= '1';
            end if;
            
          when DPTYPE_INITFC1_NP_VC0 =>
            cl_nph <= hsize_in;
            cl_npd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_nph <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_npd <= '1';
            end if;

          when DPTYPE_INITFC1_C_VC0 =>
            cl_ch <= hsize_in;
            cl_cd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_ch <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_cd <= '1';
            end if;

          when DPTYPE_UPDATEFC_P_VC0 =>
            cl_ph <= hsize_in;
            cl_pd <= dsize_in;

          when DPTYPE_UPDATEFC_NP_VC0 =>
            cl_nph <= hsize_in;
            cl_npd <= dsize_in;

          when DPTYPE_UPDATEFC_C_VC0 =>
            cl_ch <= hsize_in;
            cl_cd <= dsize_in;

          when others =>

        end case;
      end if;

      -- 
      -- UpdateFC latency timer load value (in 125MHz UI, i.e., 8 ns unit):
      -- 
      case NLANE is
        when 1 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(237/2 - (20+128)/1/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(416/2 - (20+256)/1/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when 4 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(73/2 - (20+128)/4/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(118/2 - (20+256)/4/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when 8 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(67/2 - (20+128)/8/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(107/2 - (20+256)/8/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;

        when others =>
          latency <= conv_std_logic_vector(100/2, latency'length);

      end case;

    end if;
  end process;

  --
  -- Transmission of Credit Allocated:
  -- 1) assert 'req' pulse with 'fctype_out' and 'h/dsize_out',
  --    when UpdateFC should be transmitted,
  --    i.e., 'timer' expired or 'flush' asserted.
  -- 2) deassert 'req' and reset UpdateFC timer
  --    when 'ack' detect a pulse.
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      timer     <= (others => '0');
      expired   <= '0';
      flush_p0  <= '0';
      flush_np0 <= '0';
      flush_c0  <= '0';
      req       <= '0';
      fcoutsm   <= IDLE;
    elsif (rising_edge(clk)) then

      -- save a flush request.
      if (flush_p = '1') then
        flush_p0 <= '1';
      end if;
      if (flush_np = '1') then
        flush_np0 <= '1';
      end if;
      if (flush_c = '1') then
        flush_c0 <= '1';
      end if;

      case fcoutsm is
        when IDLE =>

          if (flush_p0 = '1') then
            flush_p0   <= '0';          -- update due to flush request.
            req        <= '1';          -- send UpdateFC NP only.
            fctype_out <= DPTYPE_UPDATEFC_P_VC0;
            hsize_out  <= ca_ph;
            dsize_out  <= ca_pd;
            fcoutsm    <= REQ_P;
          elsif (flush_np0 = '1') then
            flush_np0  <= '0';
            req        <= '1';
            fctype_out <= DPTYPE_UPDATEFC_NP_VC0;
            hsize_out  <= ca_nph;
            dsize_out  <= ca_npd;
            fcoutsm    <= REQ_NP;
          elsif (flush_c0 = '1') then
            flush_c0   <= '0';
            req        <= '1';
            fctype_out <= DPTYPE_UPDATEFC_C_VC0;
            hsize_out  <= ca_ch;
            dsize_out  <= ca_cd;
            fcoutsm    <= REQ_C;
--          elsif (timer > latency) then
          elsif (timer > TIME30us) then  -- !!
            expired    <= '1';           -- update due to timer expiration.
            req        <= '1';           -- send all three UpdateFCs (P/NP/C).
            fctype_out <= DPTYPE_UPDATEFC_NP_VC0;
            hsize_out  <= ca_nph;
            dsize_out  <= ca_npd;
            fcoutsm    <= REQ_NP;
          else
            timer <= unsigned(timer) + 1;
          end if;

        when REQ_NP =>
          if (ack = '1') then
            if (expired = '1') then
              req        <= '1';
              fctype_out <= DPTYPE_UPDATEFC_P_VC0;
              hsize_out  <= ca_ph;
              dsize_out  <= ca_pd;
              fcoutsm    <= REQ_P;
            else
              req     <= '0';
              fcoutsm <= IDLE;
            end if;
          end if;

        when REQ_P =>
          if (ack = '1') then
            if (expired = '1') then
              req        <= '1';
              fctype_out <= DPTYPE_UPDATEFC_C_VC0;
              hsize_out  <= ca_ch;
              dsize_out  <= ca_cd;
              fcoutsm <= REQ_C;
            else
              req     <= '0';
              fcoutsm <= IDLE;
            end if;
          end if;

        when REQ_C =>
          if (ack = '1') then
            req     <= '0';
            timer   <= (others => '0');
            expired <= '0';
            fcoutsm <= IDLE;
          end if;

        when others =>
          fcoutsm <= IDLE;

      end case;

    end if; --  rising(clk)

  end process;

end default;

--
-- DLLP transmitter
--   1) set 'dptype' and 'data' when 'kickoff'.
--   2) calc CRC of 'dptype' & 'data'.
--   3) output DLLP to 'dp' when 'done'.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity txdpctl is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;

    dptype    : in  std_logic_vector(7 downto 0);
    data      : in  std_logic_vector(23 downto 0);
    kickoff   : in  std_logic;

    dp        : out std_logic_vector(47 downto 0);
    done      : out std_logic
  );
end txdpctl;

architecture default of txdpctl is

  component calc_crc
    generic (
      DIVISOR        : integer := 16#100B#;
      DATA_WIDTH     : integer := 32;
      CRC_WIDTH      : integer := 16
    );
    port (
      clr            : in  std_logic;
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

#if (NLANE==1)
  constant TXDP_WAIT   : integer := 3;
#else
  constant TXDP_WAIT   : integer := 0;
#endif
  constant CRCDELAY    : integer := 2;

  signal dpbody        : std_logic_vector(31 downto 0);
  signal crcout        : std_logic_vector(15 downto 0);

  type arrayNWAITxDPW  is array (TXDP_WAIT+CRCDELAY downto 0) of std_logic_vector(dpbody'left downto 0);
  type arrayNWAITxCRCW is array (TXDP_WAIT+CRCDELAY downto 0) of std_logic_vector(crcout'left downto 0);
  signal pkickoff      : std_logic_vector(TXDP_WAIT+CRCDELAY downto 0);
  signal pdp           : arrayNWAITxDPW;
  signal pcrcout       : arrayNWAITxCRCW;

begin

  -- alias for input data.
  dpbody <= data(7 downto 0) & data(15 downto 8) & data(23 downto 16) & dptype;

  pkickoff(0) <= kickoff;
  process (clk)
  begin
    if (rising_edge(clk)) then
      pkickoff(pkickoff'left downto 1) <= pkickoff(pkickoff'left-1 downto 0);
    end if;
  end process;
  done <= pkickoff(TXDP_WAIT+CRCDELAY);

  pdp(0)(31 downto 0) <= dpbody;
  process (clk)
  begin
    if (rising_edge(clk)) then
      pdp(pdp'left downto 1) <= pdp(pdp'left-1 downto 0);
    end if;
  end process;

  pcrcout(CRCDELAY) <= crcout;
  pcrcoutdelay0 : if (TXDP_WAIT > 0) generate
    process (clk)
    begin
      if (rising_edge(clk)) then
        pcrcout(pcrcout'left downto CRCDELAY+1) <= pcrcout(pcrcout'left-1 downto CRCDELAY);
      end if;
    end process;
  end generate pcrcoutdelay0;

  dp(39 downto 32) <= pcrcout(TXDP_WAIT+CRCDELAY)(15 downto 8);
  dp(47 downto 40) <= pcrcout(TXDP_WAIT+CRCDELAY)( 7 downto 0);
  dp(31 downto 0)  <= pdp(TXDP_WAIT+CRCDELAY);

  txdpcrc0 : calc_crc
    port map (
      clr  => '1',
      we   => kickoff,
      be   => (others => '1'),
      data => dpbody,
      crc  => crcout,
      rst  => rst,
      clk  => clk
    );

end default;


--
-- DLLP receiver
--   1) DLLP comes in. check CRC of it.
--   2) output its 'dptype', 'dpbody' and 'crcerr' with 'done' pulse.
--
--   input to output delay : 3clk
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity rxdpctl is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;

    we        : in  std_logic;
    dp        : in  std_logic_vector(47 downto 0);

    dptype    : out std_logic_vector(7 downto 0);
    dpbody    : out std_logic_vector(23 downto 0);
    crcerr    : out std_logic;
    done      : out std_logic
  );
end rxdpctl;

architecture default of rxdpctl is

  component calc_crc
    generic (
      DIVISOR        : integer := 16#100B#;
      DATA_WIDTH     : integer := 32;
      CRC_WIDTH      : integer := 16
    );
    port (
      clr            : in  std_logic;
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

  signal we1       : std_logic;
  signal we2       : std_logic;
  signal dp1       : std_logic_vector(47 downto 0);
  signal dp2       : std_logic_vector(47 downto 0);
  signal crc_recvd : std_logic_vector(15 downto 0);
  signal crc_calcd : std_logic_vector(15 downto 0);
  
begin

  process (clk, rst)
  begin
    if (rst = '1') then
      done <= '0';
    elsif (rising_edge(clk)) then

      dp1 <= dp;
      dp2 <= dp1;

      -- dptype & dpbody
      dptype <= dp2( 7 downto 0);
      dpbody <= dp2(15 downto 8) & dp2(23 downto 16) & dp2(31 downto 24);

      -- crcerr
      crc_recvd  <= dp1(39 downto 32) & dp1(47 downto 40);
      if (we2 = '1') then
        if (crc_recvd = crc_calcd) then
          crcerr <= '0';
        else
          crcerr <= '1';
        end if;
      end if;

      -- done pulse
      we1  <= we;
      we2  <= we1;
      done <= we2;

    end if;
  end process;

  rxdpcrc0 : calc_crc
    port map (
      clr  => '1',
      we   => we,
      be   => (others => '1'),
      data => dp(31 downto 0),
      crc  => crc_calcd,
      rst  => rst,
      clk  => clk
    );

end default;


--
-- TLP receiver
--  TLP comes in from 'tp' port with 'wein' asserted.
--  TLP contents are decoded and sent out to 'tpdata' port
--  with 'tptype', 'weout', 'crcerr'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;

entity rxtpctl is
  generic (
    NLANE      : integer
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;

    wein       : in  std_logic;
    be         : in  std_logic_vector(NLANE*2-1 downto 0);
    tptop      : in  std_logic;
    tp         : in  std_logic_vector(NLANE*16-1 downto 0);
    next_seqno : in  std_logic_vector(11 downto 0);

    topout     : out std_logic;
    weout      : out std_logic;         -- not used for x4 or wider.
    tpbody     : out std_logic_vector(NLANE*16-1 downto 0);
    seqno      : out std_logic_vector(11 downto 0);
    crcerr     : out std_logic;
    done       : out std_logic
  );
end rxtpctl;

architecture default of rxtpctl is

  component calc_crc
    generic (
      DIVISOR        : integer := 16#04C11DB7#;
      DATA_WIDTH     : integer := NLANE*16;
      CRC_WIDTH      : integer := 32
    );
    port (
      clr            : in  std_logic;
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

  component calc_lcrc_walma
    generic (
      DIVISOR        : integer := 16#04C11DB7#;
      DATA_WIDTH     : integer := NLANE*16;
      NPIPE          : integer := NLANE/2;
      CRC_WIDTH      : integer := 32
    );
    port (
      top            : in  std_logic;
      packetsize     : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

begin

  x1 : if (NLANE = 1) generate

    x1block : block

      constant CRCDELAY : integer := NSTAGE+1;  -- num of clks from the falling edge of wecrc to crcout.
      type arrayNDELAYxSEQNOW is array (CRCDELAY+1 downto 0) of std_logic_vector(11 downto 0);
      signal seqno1      : std_logic_vector(11 downto 0);
      signal tpend1      : std_logic;
      signal wecrc       : std_logic;
      signal crcout      : std_logic_vector(31 downto 0);

      signal we1, we2, we3             : std_logic;
      signal tptop1, tptop2, tptop3    : std_logic;
      signal next_rcv_seq              : std_logic_vector(11 downto 0); -- !! should be removed later.
      signal tp1, tp2                  : std_logic_vector(NLANE*16-1 downto 0);

    begin

      wecrc  <= wein;
      topout <= tptop3;
      weout  <= wein and we1 and we2 and we3;
      tpbody <= tp2;
      
      process (clk, rst)
        variable seqnotmp : std_logic_vector(seqno'range);
      begin
        if (rst = '1') then
          next_rcv_seq  <= (others => '0'); -- !! should be modified later.
          crcerr      <= '0';
        elsif (rising_edge(clk)) then

          -- delayed inputs.
          we1 <= wein;
          we2 <= we1;
          we3 <= we2;

          tptop2 <= tptop1;
          tptop3 <= tptop2;
          tp1 <= tp;
          tp2 <= tp1;

          -- detect end of packet.
          if (wein = '0' and we1 = '1') then
            tpend1 <= '1';
          else
            tpend1 <= '0';
          end if;

          -- sequence number of incoming TLP.
          if (tptop = '1') then
            seqno1  <= tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
          end if;

          -- check if sequence number of incoming TLP is the expected one.
          -- tptop should be passed to transaction layer if and only if
          -- incoming TLP is a new (i.e. non-replayed) one.
          -- !! this version skips LCRC check. should be modified later. !!
          tptop1 <= '0';
          if (tptop = '1') then
            seqnotmp := tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
            if (seqnotmp = next_rcv_seq) then  -- a new TLP is coming in.
              next_rcv_seq  <= unsigned(next_rcv_seq) + 1;
              tptop1 <= '1';
            end if;
          end if;

          -- check CRC. it should be all-one.
          if (tpend1 = '1') then
            if (crcout = x"ffffffff") then
              crcerr <= '0';
            else
              crcerr <= '1';
            end if;
          end if;

        end if;
      end process;

      seqno <= seqno1;
      done <= tpend1;
      
-- synthesis translate_off
      rxlcrc0 : calc_crc
        port map (
          clr  => '0',
          we   => wecrc,
          be   => (others => '1'),
          data => tp,
          crc  => crcout,
          rst  => rst,
          clk  => clk
        );
-- synthesis translate_on

      crcout <= (others => '1');  -- just for debugging
    end block x1block;

  end generate x1;

  x4 : if (NLANE >= 4) generate

    x4block : block

      constant CRCDELAY : integer := NSTAGE+1;  -- num of clks from the falling edge of wecrc to crcout.
      type arrayNDELAYxSEQNOW is array (CRCDELAY+1 downto 0) of std_logic_vector(11 downto 0);
      signal pseqno      : arrayNDELAYxSEQNOW;
      signal ptpend      : std_logic_vector(CRCDELAY+1 downto 0);
      signal we1         : std_logic;
      signal be1         : std_logic_vector(NLANE*2-1 downto 0);
      signal tp1         : std_logic_vector(NLANE*16-1 downto 0);
      signal crctop      : std_logic;
      signal wecrc       : std_logic;
      signal becrc       : std_logic_vector(NLANE*2-1 downto 0);
      signal crcin       : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout      : std_logic_vector(31 downto 0);
      signal beall1      : std_logic_vector(NLANE*2-1 downto 0);
      signal tptop1     : std_logic;
      signal next_rcv_seq : std_logic_vector(11 downto 0); -- !! should be removed later.

    begin

      beall1 <= (others => '1');
      topout <= tptop1;
      weout <= '0';  -- not used.

      process (clk, rst)
        variable seqnotmp : std_logic_vector(seqno'range);
      begin
        if (rst = '1') then
          next_rcv_seq  <= (others => '0'); -- !! should be modified later.
          crcerr      <= '0';
        elsif (rising_edge(clk)) then

          -- delayed inputs.
          we1 <= wein;
          be1 <= be;
          tp1 <= tp;

          -- detect end of packet.
          if (wein = '1' and be /= beall1) then
            ptpend(0) <= '1';
          else
            ptpend(0) <= '0';
          end if;
          ptpend(ptpend'left downto 1) <= ptpend(ptpend'left-1 downto 0);

          -- inputs to crc calculator.
          crctop <= tptop;
          wecrc  <= wein;
          becrc  <= be;

          -- sequence number of incoming TLP.
          if (tptop = '1') then
            pseqno(0)  <= tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
          end if;
          pseqno(pseqno'left downto 1) <= pseqno(pseqno'left-1 downto 0);

          -- check if sequence number of incoming TLP is the expected one.
          -- tptop should be passed to transaction layer if and only if
          -- incoming TLP is a new (i.e. non-replayed) one.
          -- !! this version skips LCRC check. should be modified later. !!
          tptop1 <= '0';
          if (tptop = '1') then
            seqnotmp := tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
            if (seqnotmp = next_rcv_seq) then  -- a new TLP is coming in.
              next_rcv_seq  <= unsigned(next_rcv_seq) + 1;
              tptop1 <= '1';
            end if;
          end if;

          -- check CRC. it should be all-one.
          if (ptpend(CRCDELAY) = '1') then
            if (crcout = x"ffffffff") then
              crcerr <= '0';
            else
              crcerr <= '1';
            end if;
          end if;

        end if;
      end process;

      seqno <= pseqno(CRCDELAY+1);
      done <= ptpend(CRCDELAY+1);
--      tpbody <= tp(1*8+7 downto 0*8+0) & tp1(7*8+7 downto 2*8+0);
      tpbody <= tp(1*8+7 downto 0*8+0) & tp1(NLANE*16-1 downto 2*8+0);
      
      -- pass incoming TLP into crc calculator.
      process (tp1, wein, be, we1, be1)
      begin
        if (NLANE = 4) then
          if (wein = '1' and be = x"03") then  -- CRC at packet tail is negated.
            crcin(7*8+7 downto 6*8+0) <= not tp1(7*8+7 downto 6*8+0);
            crcin(5*8+7 downto 0*8+0) <=     tp1(5*8+7 downto 0*8+0);
          elsif (we1 = '1' and be1 = x"03") then
            crcin(1*8+7 downto 0*8+0) <= not tp1(1*8+7 downto 0*8+0);
          elsif (we1 = '1' and be1 = x"3f") then
            crcin(5*8+7 downto 2*8+0) <= not tp1(5*8+7 downto 2*8+0);
            crcin(1*8+7 downto 0*8+0) <= tp1(1*8+7 downto 0*8+0);
          else
            crcin <= tp1;
          end if;
        end if;

        if (NLANE = 8) then
          -- pass incoming TLP to CRC calculator.
          crcin <= tp1;
          -- the last 4-byte of TLP (LCRC) is negated.
          if (wein = '1' and be = x"0003") then
            crcin(15*8+7 downto 14*8+0) <= not tp1(15*8+7 downto 14*8+0);
          elsif (we1 = '1' and be1 = x"0003") then
            crcin( 1*8+7 downto  0*8+0) <= not tp1( 1*8+7 downto  0*8+0);
          elsif (we1 = '1' and be1 = x"003f") then
            crcin( 5*8+7 downto  2*8+0) <= not tp1( 5*8+7 downto  2*8+0);
          elsif (we1 = '1' and be1 = x"03ff") then
            crcin( 9*8+7 downto  6*8+0) <= not tp1( 9*8+7 downto  6*8+0);
          elsif (we1 = '1' and be1 = x"3fff") then
            crcin(13*8+7 downto 10*8+0) <= not tp1(13*8+7 downto 10*8+0);
          end if;
        end if;

      end process;

-- synthesis translate_off
      rxlcrc0 : calc_lcrc_walma
        port map (
          top  => crctop,
          packetsize => packetsize,
          we   => wecrc,
          be   => becrc,
          data => crcin,
          crc  => crcout,
          rst  => rst,
          clk  => clk
        );
-- synthesis translate_on

      crcout <= (others => '1');  -- just for debugging

    end block x4block;

  end generate x4;

end default;




--
-- TLP transmitter
--  bare TLP-body comes in from 'tpbody' port with 'wein' asserted.
--  the body is wrapped with sequencial number and LCRC,
--  and then sent out to 'tp' port.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;

entity txtpctl is
  generic (
    NLANE      : integer
  );
  port (
    topin      : in  std_logic;
    wein       : in  std_logic;
    bein       : in  std_logic_vector(NLANE*2-1 downto 0);
    tpbody     : in  std_logic_vector(NLANE*16-1 downto 0);
    tpbodysize : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);

    topout     : out std_logic;
    weout      : out std_logic;
    beout      : out std_logic_vector(NLANE*2-1 downto 0);
    tp         : out std_logic_vector(NLANE*16-1 downto 0);

    clk        : in  std_logic;
    rst        : in  std_logic
  );
end txtpctl;

architecture default of txtpctl is

  component calc_crc
    generic (
      DIVISOR        : integer := 16#04C11DB7#;
      DATA_WIDTH     : integer := NLANE*16;
      CRC_WIDTH      : integer := 32
    );
    port (
      clr            : in  std_logic;
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

  component calc_lcrc_walma
    generic (
      DIVISOR        : integer := 16#04C11DB7#;
      DATA_WIDTH     : integer := NLANE*16;
      NPIPE          : integer := NLANE/2;
      CRC_WIDTH      : integer := 32
    );
    port (
      top            : in  std_logic;
      packetsize     : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
      we             : in  std_logic;
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
      );
  end component;

  constant CRCDELAY : integer := NSTAGE+1;  -- num of clks from the falling edge of wecrc to crcout.
  
begin

  x1 : if (NLANE = 1) generate

    x1block : block

      type TXTP_STATE is (
        IDLE,
        SEND_TLP,
        WAIT_LCRC,
        SEND_LCRC0,
        SEND_LCRC1,
        INVALID_STATE
      );

      signal txtpsm    : TXTP_STATE;
      signal seqno     : std_logic_vector(11 downto 0);
      signal seqno16   : std_logic_vector(15 downto 0);

      signal we        : std_logic_vector(4 downto 0);
      signal wecrc     : std_logic;
      signal tpbody1   : std_logic_vector(NLANE*16-1 downto 0);
      signal tpbody2   : std_logic_vector(NLANE*16-1 downto 0);
      signal crcin     : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout      : std_logic_vector(31 downto 0);
      
    begin

      process (clk, rst)
      begin
        if (rst = '1') then
          seqno <= (others => '0');
          txtpsm <= IDLE;
        elsif (rising_edge(clk)) then
          
          case txtpsm is
            when IDLE =>
              if (wein = '1') then
                txtpsm <= SEND_TLP;
              end if;

            when SEND_TLP =>
              if (wein = '0') then
                txtpsm <= WAIT_LCRC;
              end if;

            when WAIT_LCRC => 
              seqno <= unsigned(seqno) + 1;
              txtpsm <= SEND_LCRC0;

            when SEND_LCRC0 => 
              txtpsm <= SEND_LCRC1;

            when SEND_LCRC1 => 
              txtpsm <= IDLE;
              
            when others =>
              txtpsm <= INVALID_STATE;

          end case;
        end if;
      end process;

      beout <= (others => '1');

      process (clk)
      begin
        if (rising_edge(clk)) then
          topout <= topin;
        end if;
      end process;

      -- calc_crc has 2-clk delay from 'data' input to 'crc' output.
      weout <= we(1) or we(2) or we(3) or we(4);
      wecrc <= we(0) or we(1);

      we(0) <= wein;
      process (clk)
      begin
        if (rising_edge(clk)) then
          we(4 downto 1)  <= we(3 downto 0);
          tpbody1 <= tpbody;
          tpbody2 <= tpbody1;
        end if;
      end process;

      -- extend seqno to 16-bit width and then swap the higher and lower bytes.
      seqno16 <= seqno(7 downto 0) &
                 conv_std_logic_vector(0, 16 - seqno'length) & seqno(11 downto 8);
      crcin <= tpbody1 when we(1) = '1' else seqno16;

      process (we, txtpsm, tpbody2, crcout, seqno, seqno16)
      begin
        if (we(2) = '1') then
          tp <= tpbody2;
        elsif (txtpsm = SEND_LCRC0) then
          tp <= crcout(23 downto 16) & crcout(31 downto 24);
        elsif (txtpsm = SEND_LCRC1) then
          tp <= crcout(7 downto 0) & crcout(15 downto 8);
        else
          tp <= seqno16;
        end if;
      end process;

      txlcrc0 : calc_crc
        port map (
          clr  => '0',
          we   => wecrc,
          be   => (others => '1'),
          data => crcin,
          crc  => crcout,
          rst  => rst,
          clk  => clk
        );
    end block x1block;

  end generate x1;

  x4 : if (NLANE >= 4) generate

    x4block : block

      signal seqno     : std_logic_vector(11 downto 0);
      signal top       : std_logic;
      signal wecrc     : std_logic;
      signal becrc     : std_logic_vector(NLANE*2-1 downto 0);
      signal crcin     : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout    : std_logic_vector(31 downto 0);
      signal crcout1   : std_logic_vector(31 downto 0);

      type arrayNDELAYxDATAW     is array (CRCDELAY+2 downto 0) of std_logic_vector(NLANE*16-1 downto 0);
      type arrayNDELAYxHALFNLANE is array (CRCDELAY+2 downto 0) of std_logic_vector(NLANE/2-1 downto 0);
      signal ptp            : arrayNDELAYxDATAW;
      signal ptop           : std_logic_vector(CRCDELAY downto 0);
      signal ptail          : std_logic_vector(CRCDELAY downto 0);
      signal pwe            : std_logic_vector(CRCDELAY+2 downto 0);
      signal be0            : std_logic_vector(NLANE*2-1 downto 0);
      signal be1            : std_logic_vector(NLANE*2-1 downto 0);
      signal tpbody1        : std_logic_vector(NLANE*16-1 downto 0);
      signal endat          : arrayNDELAYxHALFNLANE;
      signal packetsize     : std_logic_vector(OFFSET_WIDTH-1 downto 0);

    begin
      -- set initial values of pipeline registers.
      pwe(0) <= wein;

      -- shift tbody by 2 bytes in order to insert a sequence number.
      ptp(0)(NLANE*16-1 downto 2*8+0) <= tpbody(NLANE*16-1-2*8 downto 0*8+0);
      ptp(0)(1*8+7 downto 0*8+0)      <= seqno(7 downto 0) & "0000" & seqno(11 downto 8) when ptop(0) = '1'
                                         else tpbody1(NLANE*16-1    downto NLANE*16-2*8);

      -- check byte-enable at packet tail and
      -- set endat(0) to 1, 2, 4, 8, if packet ends at lane 3L, 7L, 3H, 7H,
      -- for x8, for example.
      process(pwe, be0)
        variable endattmp : std_logic;
      begin
        endat(0) <= (others => '0');
        if (pwe(0) = '1') then
          endattmp := '0';
          for i in NLANE/2-1 downto 0 loop
            if (be0(i*4) = '1' and endattmp = '0') then
              endat(0)(i) <= '1';
              endattmp    := '1';
            end if;
          end loop;
        end if;
      end process;
      ptop(0) <= topin;  -- top of TLP input.
      topout <= ptop(CRCDELAY);  -- top of TLP output.

      -- tail of TLP input.
      process (pwe, topin)
      begin
        ptail(0) <= '0';
        if (pwe(1) = '1' and (pwe(0) = '0' or topin = '1')) then
          ptail(0) <= '1';
        end if;
      end process;

      packetsize <= tpbodysize + conv_std_logic_vector(2, packetsize'length);  -- add 2 bytes for sequence number.
      crcin <= ptp(0);
      be0 <= bein;

      process(clk, rst)
      begin
        if (rst = '1') then
          seqno <= (others => '0');
        elsif (rising_edge(clk)) then

          -- delayed inputs:
          ptop(ptop'left downto 1) <= ptop(ptop'left-1 downto 0);
          ptail(ptail'left downto 1) <= ptail(ptail'left-1 downto 0);
          pwe(pwe'left downto 1) <= pwe(pwe'left-1 downto 0);
          ptp(ptp'left downto 1) <= ptp(ptp'left-1 downto 0);
          endat(endat'left downto 1) <= endat(endat'left-1 downto 0);

          -- count up sequence number at the rising edge of wein.
          if (ptop(0) = '1') then
            seqno <= unsigned(seqno) + 1;
          end if;

          crcout1 <= crcout;
          be1 <= be0;
          tpbody1 <= tpbody;

        end if;
      end process;
      
      -- write enable for CRC calculation.
      process(endat, pwe)
      begin
        if (NLANE = 4) then
          wecrc <= '0';
          if (pwe(0) = '1') then
            -- wecrc is the same as wein, in most cases.
            wecrc <= '1';
          end if;
          if (endat(1)(1) = '1' and ptail(0) = '1') then
            -- assert wecrc 1clk longer if TLP end at lane 3H.
            wecrc <= '1';
          end if;
        end if;

        if (NLANE = 8) then
          wecrc <= '0';
          if (pwe(0) = '1') then
            -- wecrc is the same as wein, in most cases.
            wecrc <= '1';
          end if;
          if (endat(1)(3) = '1' and ptail(0) = '1') then
            -- assert wecrc 1clk longer if TLP end at lane 7H.
            wecrc <= '1';
          end if;
        end if;
      end process;

      -- byte enable for CRC calculation.
      process(pwe, be0, be1)
      begin
        if (NLANE = 4) then
          becrc <= (others => '1');
          if (ptail(0) = '1' and be1(4) = '1') then
            becrc <= x"03";
          elsif (pwe(0) = '1' and be0(4) = '0') then
            becrc <= x"3f";
          end if;
        end if;

        if (NLANE = 8) then
          becrc <= (others => '1');
          if (ptail(0) = '1' and be1(12) = '1') then
            becrc <= x"0003";
          elsif (pwe(0) = '1') then
            if (be0(4) = '0') then
              becrc <= x"003f";
            elsif (be0(8) = '0') then
              becrc <= x"03ff";
            elsif (be0(12) = '0') then
              becrc <= x"3fff";
            end if;
          end if;
        end if;

      end process;

      -- write enable for final output.
      process(clk)
      begin
        if (rising_edge(clk)) then
          if (NLANE = 4) then
            if (pwe(CRCDELAY-1) = '1') then
              weout <= '1';
            elsif (pwe(CRCDELAY) = '1') then
              weout <= '1';
            else
              weout <= '0';
            end if;
          end if;

          if (NLANE = 8) then
            -- weout is the same as wein delayed by CRCDELAY-clk, in most cases.
            weout <= pwe(CRCDELAY-1);
            -- assert weout 1clk longer if TLP end at lane 3H or 7H.
            if (pwe(CRCDELAY) = '1') then
              if (endat(CRCDELAY)(3) = '1' or endat(CRCDELAY)(2) = '1') then
                weout <= '1';
              end if;
            else
            end if;
          end if;

        end if;
      end process;

      -- byte enable for final output.
      process(clk)
      begin
        if (rising_edge(clk)) then
          if (NLANE = 4) then
            beout <= (others => '1');
            if (ptail(CRCDELAY-1) = '1') then
              if (endat(CRCDELAY)(1) = '1') then
                beout <= x"3f";
              else
                beout <= x"03";
              end if;
            end if;
          end if;

          if (NLANE = 8) then
            beout <= (others => '1');
            if (endat(CRCDELAY)(3) = '1' and ptail(CRCDELAY-1) = '1') then
              beout <= x"003f";
            end if;
            if (endat(CRCDELAY)(2) = '1' and ptail(CRCDELAY-1) = '1') then
              beout <= x"0003";
            end if;
            if (endat(CRCDELAY-1)(1) = '1' and ptail(CRCDELAY-2) = '1') then
              beout <= x"3fff";
            end if;
            if (endat(CRCDELAY-1)(0) = '1' and ptail(CRCDELAY-2) = '1') then
              beout <= x"03ff";
            end if;
          end if;
        end if;
      end process;

      -- final data output.
      process(endat, pwe, ptp, crcout, crcout1)
      begin
        tp <= ptp(CRCDELAY);            -- default output other than packet tail.

        if (NLANE = 4) then
          if (endat(CRCDELAY+1)(1) = '1' and ptail(CRCDELAY) = '1') then
            tp(5*8+7 downto 2*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;

          if (endat(CRCDELAY+1)(0) = '1' and ptail(CRCDELAY) = '1') then
            tp(1*8+7 downto 0*8+0) <= crcout1(0*8+7 downto 0*8+0) & crcout1(1*8+7 downto 1*8+0);
          end if;
          if (endat(CRCDELAY)(0) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(7*8+7 downto 6*8+0) <= crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;

        end if;

        if (NLANE = 8) then
          if (endat(CRCDELAY+1)(3) = '1' and ptail(CRCDELAY) = '1') then
            tp(5*8+7 downto 2*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;

          if (endat(CRCDELAY+1)(2) = '1' and ptail(CRCDELAY) = '1') then
            tp(1*8+7 downto 0*8+0) <= crcout1(0*8+7 downto 0*8+0) & crcout1(1*8+7 downto 1*8+0);
          end if;
          if (endat(CRCDELAY)(2) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(15*8+7 downto 14*8+0) <= crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;

          if (endat(CRCDELAY)(1) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(13*8+7 downto 10*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);

          end if;

          if (endat(CRCDELAY)(0) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(9*8+7 downto 6*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;

        end if;
      end process;

      txlcrc0 : calc_lcrc_walma
        port map (
          top        => ptop(0),
          packetsize => packetsize,
          we         => wecrc,
          be         => becrc,
          data       => crcin,
          crc        => crcout,
          rst        => rst,
          clk        => clk
        );

    end block x4block;
  end generate x4;

end default;


--
-- calculate CRC_WIDTH-bit CRC of input data stream of DATA_WIDTH-bit width.
-- input data to output crc delay : 2clk
-- output remains unchanged while we is deasserted.
--
-- this version calculates CRC using the straight forward, simplest algorithm.
-- it is handy to calculate 16-bit CRC for DLLP,
-- but is not fast enough to calculate 32-bit LCRC for lanes wider than x2.
-- for x4 (i.e. 64-bit input per clock) and wider lanes, use calc_crc_walma.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity calc_crc is
  generic (
    DIVISOR        : integer;
    DATA_WIDTH     : integer;
    CRC_WIDTH      : integer
  );
  port (
    clr            : in  std_logic;
    we             : in  std_logic;
    be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
    data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
    crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
    rst            : in  std_logic;     -- active high reset
    clk            : in  std_logic
  );
end calc_crc;

architecture default of calc_crc is

  signal we1         : std_logic;
  signal crc_reg     : std_logic_vector(CRC_WIDTH-1 downto 0);
  signal ncrc1       : std_logic_vector(CRC_WIDTH-1 downto 0);
  signal crc2        : std_logic_vector(CRC_WIDTH-1 downto 0);

begin

  process (clk)
    variable crc_var : std_logic_vector(CRC_WIDTH-1 downto 0);
  begin
    if (rising_edge(clk)) then
      we1 <= we;

      if (we = '1') then
        if (we1 = '0' or clr = '1') then -- init crc_reg at the rising edge of we.
          crc_var := (others => '1');
        else
          crc_var := crc_reg;
        end if;
        -- shift CRC 1-bit to the left.
        eachbit : for i in 0 to DATA_WIDTH-1 loop
          -- CRC is XORed with the divisor if its MSB is 1.
          if (be(i/8) = '1') then
            if ((data(i) xor crc_var(CRC_WIDTH-1)) = '1') then
              crc_var := (crc_var(CRC_WIDTH-2 downto 0) & '0') xor
                         conv_std_logic_vector(DIVISOR, CRC_WIDTH);
            else
              crc_var := (crc_var(CRC_WIDTH-2 downto 0) & '0');
            end if;
          end if;
        end loop;
        crc_reg <= crc_var;
      end if;  -- we
    end if;  -- clk
  end process;

  -- negate.
  ncrc1 <= not crc_reg;

  -- reverse bit order of each byte.
  process (clk)
  begin
    if (rising_edge(clk)) then
      for j in 0 to CRC_WIDTH / 8 - 1 loop
        for i in 0 to 7 loop
          crc2(j * 8 + i) <= ncrc1(j * 8 + 7 - i);
        end loop;
      end loop;
    end if;
  end process;

  crc <= crc2;
  
end default;

--
-- CRC calculater, pipelined & parallelized with
-- algorithm proposed by Walma (2007).
--
-- calculate 32-bit CRC of input data stream of 64-bit width.
--   + input data to output crc delay : ??? clk
--   + support DIVISOR : 16#04C11DB7# so far.
--   + top must be asserted 1clk at the top of data stream.
--   + be is handled correctly only at the tail of data stream.
--     be='0' in the middle of data-stream causes an incorrect result.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;

entity calc_lcrc_walma is
  generic (
    DIVISOR        : integer;
    DATA_WIDTH     : integer;
    NPIPE          : integer;
    CRC_WIDTH      : integer
  );
  port (
    top            : in  std_logic;
    packetsize     : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
    we             : in  std_logic;
    be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
    data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
    crc            : out std_logic_vector(CRC_WIDTH-1 downto 0);
    rst            : in  std_logic;     -- active high reset
    clk            : in  std_logic
  );
end calc_lcrc_walma;

architecture default of calc_lcrc_walma is
  component crcpipe
    generic (
      DIVISOR        : integer;
      DATA_WIDTH     : integer;
      CRC_WIDTH      : integer
    );
    port (
      sop            : in  std_logic;
      data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      offset         : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
      eop            : in  std_logic;
      offadj         : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
      lfsr           : out std_logic_vector(CRC_WIDTH-1 downto 0);
      done           : out std_logic;
      rst            : in  std_logic;     -- active high reset
      clk            : in  std_logic
    );
  end component;

  type arrayNPIPExDATAW is array (NPIPE-1 downto 0) of std_logic_vector(DATA_WIDTH/NPIPE-1 downto 0);
  type arrayNPIPExBEW   is array (NPIPE-1 downto 0) of std_logic_vector(DATA_WIDTH/NPIPE/8-1 downto 0);

  -- input to each pipe.
  signal sop     : std_logic;
  signal eop     : std_logic;
  signal pdata   : arrayNPIPExDATAW;
  signal pbe     : arrayNPIPExBEW;
  signal poffset : arrayNPIPExOFFW;
  signal offadj  : std_logic_vector(OFFSET_WIDTH-1 downto 0);

  -- output from each pipe.
  signal plfsr   : arrayNPIPExCRCW;
  signal pdone   : std_logic_vector(NPIPE-1 downto 0);

  -- LFSR reduced from all pipes.
  signal lfsr    : std_logic_vector(CRC_WIDTH-1 downto 0);

  -- for internal use.
  signal we1     : std_logic;

begin

  process(clk, rst)

    variable lfsrtmp  : std_logic_vector(CRC_WIDTH-1 downto 0);
    variable nbytetmp : integer;

  begin

    if (rst = '1') then
      for i in 0 to NPIPE-1 loop
        poffset(i) <= (others => '0');  -- just to remove warning message :-)
      end loop;
    elsif (rising_edge(clk)) then

      sop <= top;
      we1 <= we;
      --
      -- at start of a new data stream...
      --
      for i in 0 to NPIPE-1 loop

#if 1  // for use with calc_lfsr_wbe_serial
        -- negate the first CRC_WIDTH-bit of the input.
        if (top = '1' and DATA_WIDTH/NPIPE*i < CRC_WIDTH) then
          pdata(i) <= not data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
        else
          pdata(i) <= data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
        end if;

        pbe(i)   <=   be((i+1)*DATA_WIDTH/NPIPE/8-1 downto i*DATA_WIDTH/NPIPE/8);

#else  // for use with calc_lfsr_serial
        -- negate the first CRC_WIDTH-bit of the input.
        if (top = '1' and DATA_WIDTH/NPIPE*i < CRC_WIDTH) then
          pdata(i) <= not data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
        else
         -- all data fed to one pipeline need to have a same be.
          if (be(i*DATA_WIDTH/NPIPE/8) = '1') then
            pdata(i) <= data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
          else
            pdata(i) <= (others => '0');
          end if;
        end if;
#endif

        -- set offset, the number of bytes from the end of max-possible TLP.
        if (top = '1') then
          if (USE_G = 1) then
            poffset(i) <= conv_std_logic_vector(MAX_TLP_SIZE-DATA_WIDTH/NPIPE/8*(i+1), OFFSET_WIDTH); -- !!
            -- eg) for x4, the 1st data to be processed by the 1st pipe (crcpipe0) has (8k-8/NPIPE)-byte offset.
          else
            if (unsigned(packetsize) > DATA_WIDTH/NPIPE/8*(i+1)) then -- care for packet tail.
              poffset(i) <= packetsize - conv_std_logic_vector(DATA_WIDTH/NPIPE/8*(i+1), OFFSET_WIDTH);
            else
              poffset(i) <= (others => '0');
            end if;
          end if;
        else
          if (unsigned(poffset(i)) > DATA_WIDTH/8) then  -- care for packet tail.
            poffset(i) <= unsigned(poffset(i)) - DATA_WIDTH/8;
          else
            poffset(i) <= (others => '0');
          end if;
        end if;

      end loop;

      --
      -- while pipelines processing a data stream...
      --
      if (top = '1') then
        offadj <= conv_std_logic_vector(MAX_TLP_SIZE-DATA_WIDTH/8, OFFSET_WIDTH);
      elsif (we = '1') then
        -- count number of bytes taken in.
        nbytetmp := 0;
        for i in 0 to DATA_WIDTH/8-1 loop
          if (be(i) = '1') then
            nbytetmp := nbytetmp + 1;
          end if;
        end loop;

        -- as data-stream grows bigger, offadj (i.e., difference between
        -- data-stream size and MAX_TLP_SIZE) becomes shorter. the
        -- final value of offadj can be obtained when 'we' goes down,
        -- and it can be read out by pipelines when 'eop' is asserted.
        offadj <= unsigned(offadj) - nbytetmp;

      end if;

      --
      -- at end of a data stream...
      --
      lfsrtmp := (others => '0');
      for i in 0 to NPIPE-1 loop
        -- reduce LFSR output from each pipe.
        if (pdone(0) = '1') then
          lfsrtmp := lfsrtmp xor plfsr(i);
        end if;
        
      end loop;
      lfsr <= lfsrtmp;

    end if;
  end process;

  -- assert 'eop' at
  --   . the falling edge of 'we'.
  --   . 1-clk before 'top' of the next packet.
  eop <= '1' when (we = '0' and we1 = '1') or (top = '1' and we1 = '1') else '0';

    u0 : crcpipe
      generic map (
        DIVISOR    => DIVISOR,
        DATA_WIDTH => DATA_WIDTH/NPIPE,
        CRC_WIDTH  => CRC_WIDTH
      )
      port map (
        sop    => sop,
        eop    => eop,

        -- taken in every clock cycle.
        data   => pdata(0),
        be     => pbe(0),
        offset => poffset(0),
        
        -- taken in at the end of a packet.
        offadj => offadj,

        -- output becomes valid several clocks after eop pulse.
        -- the delay is fixed and thus pdone pulse is not really necessary.
        -- just for debugging and easy maintenance.
        done   => pdone(0),
        lfsr   => plfsr(0),

        rst    => rst,
        clk    => clk
      );

  pipe : for i in 1 to NPIPE-1 generate
    u0 : crcpipe
      generic map (
        DIVISOR    => DIVISOR,
        DATA_WIDTH => DATA_WIDTH/NPIPE,
        CRC_WIDTH  => CRC_WIDTH
      )
      port map (
        sop    => sop,
        eop    => eop,

        -- taken in every clock cycle.
        data   => pdata(i),
        be     => pbe(i),
        offset => poffset(i),
        
        -- taken in at the end of a packet.
        offadj => offadj,

        -- output becomes valid several clocks after eop pulse.
        -- the delay is fixed and thus pdone pulse is not really necessary.
        -- just for debugging and easy maintenance.
        done   => pdone(i),
        lfsr   => plfsr(i),

        rst    => rst,
        clk    => clk
      );
  end generate pipe;

  -- convert LFSR to CRC.
  process (lfsr)
    variable nlfsrtmp : std_logic_vector(CRC_WIDTH-1 downto 0);
  begin
    -- negate.
    nlfsrtmp := not lfsr;

    -- reverse bit order of each byte.
    for j in 0 to CRC_WIDTH / 8 - 1 loop
      for i in 0 to 7 loop
        crc(j * 8 + i) <= nlfsrtmp(j * 8 + 7 - i);
      end loop;
    end loop;

  end process;

end default;

--
-- A pipeline to calculate a CRC fragment.
--
--  + sop pulse indicates start of a new packet.
--    it clears an internal LFSR register.
--
--  + data & offset are taken in every clock cycle, and pushed
--    into the pipeline. the pipeline consists of several stages:
--
--    - prestage: calculate LFSR for data taken in.
--
--    - H stages: multiply H-matrices to LFSR for offset-times.
--
--    - accumulation stage: accumulate LFSRs.
--
--    - G stages: multiply G-matrices to the accumulated LFSR
--      for offadj-times. this part is not pipelined. each of
--      G stages is activated only once per packet.
--
--  + offadj is taken in at end of a packet with eop pulse.
--    it indicates the number of over-processed 'zero's.
--    contribution to LFSR values from these zeros are canceled
--    by multiplying G-matrix with LFSR offadj times.
--
--  + output becomes valid several clocks after eop pulse, when the
--    above mentioned cancelation has done. the delay is fixed, and
--    thus done pulse is not really necessary. it exists just for
--    debugging and easy maintenance.
-- 
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;

entity crcpipe is
  generic (
    DIVISOR        : integer := 16#04C11DB7#;
    DATA_WIDTH     : integer;
    CRC_WIDTH      : integer
  );
  port (
    sop            : in  std_logic;
    data           : in  std_logic_vector(DATA_WIDTH-1 downto 0);
    be             : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
    offset         : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
    eop            : in  std_logic;
    offadj         : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);
    lfsr           : out std_logic_vector(CRC_WIDTH-1 downto 0);
    done           : out std_logic;
    rst            : in  std_logic;     -- active high reset
    clk            : in  std_logic
  );
end crcpipe;

architecture default of crcpipe is

  component lpm_rom
    generic (
      LPM_WIDTH    : natural;
      LPM_WIDTHAD  : natural;
      LPM_NUMWORDS : natural;
      LPM_FILE     : string
    );
    port (
      address : in  std_logic_vector(LPM_WIDTHAD-1 downto 0);
      q       : out std_logic_vector(LPM_WIDTH-1 downto 0)
    );
  end component;
  
  -- pipeline registers.
  signal psop     : std_logic_vector(FINAL_STAGE downto 0);
  signal peop     : std_logic_vector(FINAL_STAGE downto 0);
  signal plfsr    : arrayNSTGxCRCW;
  signal paccum   : arrayNSTGxCRCW;
  signal poffset  : arrayNSTGxOFFW;
  signal poffadj  : arrayNSTGxOFFW;

begin
  process(clk, rst)

    variable  plfsrtmp  : std_logic_vector(CRC_WIDTH-1 downto 0);
    variable  paccumtmp : std_logic_vector(CRC_WIDTH-1 downto 0);

  begin

    if (rst = '1') then

    elsif (rising_edge(clk)) then

      --
      -- input to pipeline registers
      --
      psop(0) <= sop;
      psop(FINAL_STAGE downto 1) <= psop(FINAL_STAGE-1 downto 0);

      peop(0) <= eop;
      peop(FINAL_STAGE downto 1) <= peop(FINAL_STAGE-1 downto 0);

      poffset(0) <= offset;
      poffset(FINAL_STAGE downto 1) <= poffset(FINAL_STAGE-1 downto 0);

      if (eop = '1') then
        poffadj(0) <= offadj;
      end if;
      poffadj(FINAL_STAGE downto 1) <= poffadj(FINAL_STAGE-1 downto 0);


      --
      -- prestage:
      -- calculate LFSR for a single-byte input.
      --
--      plfsr(0) <= calc_lfsr_serial(data, DIVISOR, CRC_WIDTH);
      plfsr(0) <= calc_lfsr_wbe_serial(data, be, DIVISOR, CRC_WIDTH);

      --
      -- H stages:
      -- multiply H-matrices to LFSR for offset-times.
      --
      if (USE_H11 = 0) then
        case MULT_PER_STAGE is
          when 1 =>
            for i in 0 to ACCUM_STAGE-2 loop
              if (poffset(i)(i) = '1') then
                plfsr(i+1) <= mult_m_v(H_MATRIX(i), plfsr(i));
              else
                plfsr(i+1) <= plfsr(i);
              end if;
            end loop;

          when others =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp  := plfsr(i);
              for k in 0 to MULT_PER_STAGE-1 loop
                if (poffset(i)(i*MULT_PER_STAGE+k) = '1') then
                  plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE+k), plfsrtmp);
                end if;
              end loop;
              plfsr(i+1)   <= plfsrtmp;
            end loop;

        end case;

      else  -- USE_H11

        case MULT_PER_STAGE is
          when 1 =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp  := plfsr(i);
              case poffset(i)(i*2+1 downto i*2+0) is
                when "00" =>
                  plfsrtmp := plfsrtmp;
                when "01" =>
                  plfsrtmp := mult_m_v(H_MATRIX(i*2+0), plfsrtmp);
                when "10" =>
                  plfsrtmp := mult_m_v(H_MATRIX(i*2+1), plfsrtmp);
                when "11" =>
                  plfsrtmp := mult_m_v(H11_MATRIX(i*2+0), plfsrtmp);
                when others =>
              end case;
              plfsr(i+1) <= plfsrtmp;
            end loop;

          when others =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp  := plfsr(i);
              for k in 0 to MULT_PER_STAGE-1 loop
                case poffset(i)(i*MULT_PER_STAGE*2+k*2+1 downto i*MULT_PER_STAGE*2+k*2+0) is
                  when "00" =>
                    plfsrtmp := plfsrtmp;
                  when "01" =>
                    plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE*2+k*2+0), plfsrtmp);
                  when "10" =>
                    plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE*2+k*2+1), plfsrtmp);
                  when "11" =>
                    plfsrtmp := mult_m_v(H11_MATRIX(i*MULT_PER_STAGE*2+k*2+0), plfsrtmp);
                  when others =>
                end case;
              end loop;
              plfsr(i+1) <= plfsrtmp;
            end loop;

        end case;

      end if;

      --
      -- accumulation stage
      --
      if (psop(ACCUM_STAGE-1) = '1') then
        -- paccum is set to plfsr at start of a packet.
        if (poffset(ACCUM_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(ACCUM_STAGE) <= mult_m_v(H_MATRIX(OFFSET_WIDTH-1), plfsr(ACCUM_STAGE-1));
        else
          paccum(ACCUM_STAGE) <= plfsr(ACCUM_STAGE-1);
        end if;
      else
        -- plfsr is accumulated to paccum.
        if (poffset(ACCUM_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(ACCUM_STAGE) <= paccum(ACCUM_STAGE) xor mult_m_v(H_MATRIX(OFFSET_WIDTH-1), plfsr(ACCUM_STAGE-1));
        else
          paccum(ACCUM_STAGE) <= paccum(ACCUM_STAGE) xor plfsr(ACCUM_STAGE-1);
        end if;
      end if;

      --
      -- G stages:
      -- multiply G-matrices to the accumulated LFSR for offadj-times.
      --
      if (USE_G = 1) then
        if (USE_H11 = 0) then

          case MULT_PER_STAGE is
            when 1 =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                if (poffadj(ACCUM_STAGE+i)(i) = '1') then
                  paccum(ACCUM_STAGE+i+1) <= mult_m_v(G_MATRIX(i), paccum(ACCUM_STAGE+i));
                else
                  paccum(ACCUM_STAGE+i+1) <= paccum(ACCUM_STAGE+i);
                end if;
              end loop;

            when others =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                for k in 0 to MULT_PER_STAGE-1 loop
                  if (poffadj(ACCUM_STAGE+i)(i*MULT_PER_STAGE+k) = '1') then
                    paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE+k), paccumtmp);
                  end if;
                end loop;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;

          end case;

        else

          case MULT_PER_STAGE is
            when 1 =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                case poffadj(ACCUM_STAGE+i)(i*2+1 downto i*2+0) is
                  when "00" =>
                    paccumtmp := paccumtmp;
                  when "01" =>
                    paccumtmp := mult_m_v(G_MATRIX(i*2+0), paccumtmp);
                  when "10" =>
                    paccumtmp := mult_m_v(G_MATRIX(i*2+1), paccumtmp);
                  when "11" =>
                    paccumtmp := mult_m_v(G11_MATRIX(i*2+0), paccumtmp);
                  when others =>
                end case;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;

            when others =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                for k in 0 to MULT_PER_STAGE-1 loop
                  case poffadj(ACCUM_STAGE+i)(i*MULT_PER_STAGE*2+k*2+1 downto i*MULT_PER_STAGE*2+k*2+0) is
                    when "00" =>
                      paccumtmp := paccumtmp;
                    when "01" =>
                      paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE*2+k*2+0), paccumtmp);
                    when "10" =>
                      paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE*2+k*2+1), paccumtmp);
                    when "11" =>
                      paccumtmp := mult_m_v(G11_MATRIX(i*MULT_PER_STAGE*2+k*2+0), paccumtmp);
                    when others =>
                  end case;
                end loop;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;

          end case;

        end if; -- USE_H11

        if (poffadj(FINAL_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(FINAL_STAGE) <= mult_m_v(G_MATRIX(OFFSET_WIDTH-1), paccum(FINAL_STAGE-1));
        else
          paccum(FINAL_STAGE) <= paccum(FINAL_STAGE-1);
        end if;
      end if; -- USE_G

    end if; -- clk
  end process;

  done <= peop(FINAL_STAGE);
  lfsr <= paccum(FINAL_STAGE);

end default;
