library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity gpcie is
  generic (
    NLANE             : integer range 1 to 8 := 8;        -- number of lanes. should be one of 1, 4, 8.
    NDMACH            : integer range 0 to 7 := 2;        -- number of DMA channels.

    NVC               : integer := 1;
    N_FTS             : integer := 6;
    DATARATE          : std_logic_vector(7 downto 0) := "00000010";
    TSCTRL            : std_logic_vector(7 downto 0) := "00000000";  -- bit0:hotreset 1:disable_link 2:loopback 3:disable_scrambling

    MAX_LINK_SPEED    : std_logic_vector(3 downto 0) := "0001";  -- 2.5Gb/s
    MAX_READ_REQ_SIZE : integer := 4096;  -- 4096 gives the best performance on DMA read,
                                          -- at the expense of hogging the link.
    MAX_PAYLOAD       : integer := 256;

    CA_PH_VC0_INIT    : integer := 16;  -- PIO write
    CA_PD_VC0_INIT    : integer := 64;

--    CA_NPH_VC0_INIT   : integer := 4;   -- PIO read
    CA_NPH_VC0_INIT   : integer := 32;   -- PIO read
    CA_NPD_VC0_INIT   : integer := 2;   -- data buf not used.

    CA_CH_VC0_INIT    : integer := 16;  -- completion for DMA read
    CA_CD_VC0_INIT    : integer := 64;

    CL_PH_VC0_INIT    : integer := 16;  -- DMA write
    CL_PD_VC0_INIT    : integer := 64;

    CL_NPH_VC0_INIT   : integer := 2;   -- DMA read
    CL_NPD_VC0_INIT   : integer := 2;   -- data buf not used.

--    CL_CH_VC0_INIT    : integer := 4;   -- completion for PIO read
--    CL_CD_VC0_INIT    : integer := 4;

    CL_CH_VC0_INIT    : integer := 16;   -- completion for PIO read
    CL_CD_VC0_INIT    : integer := 16;

    CFG_VENDOR_ID_INIT           : std_logic_vector(15 downto 0) := x"1b1a";    -- KFCR
    CFG_DEVICE_ID_INIT           : std_logic_vector(15 downto 0) := x"0e70";    -- PCI device ID.
    CFG_COMMAND_INIT             : std_logic_vector(15 downto 0) := "00000" &   -- 15:11  reserved.
                                                                    "1" &       -- 10     interrupt disable. set to 1 to disable interrupt.
                                                                    "0000000" &  --       some are reserved by the specification. others are not used for now.
                                                                    '1' &       -- 2      bus master enable. set to 1 to make DMA engine to work.
                                                                    '1' &       -- 1      memory address space decoder enable. set to 1 to accept memory transaction.
                                                                    '0';        -- 0      io address space decode enable. set to 0 for now. does not accept IO transaction.

    CFG_REVISION_ID_INIT         : std_logic_vector( 7 downto 0) := x"01";
    CFG_CLASS_CODE_INIT          : std_logic_vector(23 downto 0) := x"ff0000";    -- device does not fit any of the defined class codes.
    CFG_HEADER_TYPE_INIT         : std_logic_vector( 7 downto 0) := x"00";        -- single function, non-bridge.
    CFG_BAR0_INIT                : std_logic_vector(31 downto 0) := x"ffff8008";  -- 32kB, prefetchable, 32-bit address, memory space.
    CFG_BAR1_INIT                : std_logic_vector(31 downto 0) := x"fffff008";  --  4kB, prefetchable, 32-bit address, memory space.
    CFG_BAR2_INIT                : std_logic_vector(31 downto 0) := x"ffff8008";  -- 32kB, prefetchable, 32-bit address, memory space.
    CFG_BAR3_INIT                : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR4_INIT                : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR5_INIT                : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR_ROM_INIT             : std_logic_vector(31 downto 0) := x"00000000";
    CFG_SUB_VENDOR_ID_INIT       : std_logic_vector(15 downto 0) := x"1b1a";
    CFG_SUB_DEVICE_ID_INIT       : std_logic_vector(15 downto 0) := x"0e70";
    CFG_INT_PIN_INIT             : std_logic_vector( 7 downto 0) := x"00";        -- 00:generate no interrupts. 01:INTA 02:INTB 03:INTC 04:INTD

    PCIE_CAP_INIT                : std_logic_vector(15 downto 0) := "0000000" &   -- interrupt message number. not used for now.
                                                                    "0" &      -- slot implemented. not used for endpoint.
                                                                    "0001" &   -- device type.  0001:legacy PCIe endpoint.
                                                                    "0001";    -- capability version.

    PCIE_DEVICE_CAP_INIT         : std_logic_vector(31 downto 0) := "0000" &  -- reserved.
                                                                    "00" &    -- captured slot power limit scale. not used for now.
                                                                    x"00" &   -- captured slot power limit value. not used for now.
                                                                    "000" &   -- reserved.
                                                                    '0' &     -- no power indicator present.
                                                                    '0' &     -- no attention indicator present.
                                                                    '0' &     -- no attention button present.
                                                                    "000" &   -- L1 latency is less than 1us.
                                                                    "000" &   -- L0s latency is less than 64ns.
                                                                    "000" &   -- do not support phantom functions.
                                                                    "000";    -- supported max payload size is automatically set to MAX_PAYLOAD.

    PCIE_DEVICE_CTL_INIT         : std_logic_vector(15 downto 0) := '0' &     -- reserved.
                                                                    "000" &   -- max reqd request value is automatically set to MAX_READ_REQ_SIZE.
                                                                    "0000" &  -- no snoop, aux power pm, phantom functions, extended tag are not supported.
                                                                    "000" &   -- initial value of max payload size should be set to 128B.
                                                                    "00000";  -- relaxed ordering, advanced error reportings are not supported.

    PCIE_DEVICE_STATUS_INIT      : std_logic_vector(15 downto 0) := x"0000";

    PCIE_LINK_CAP_INIT           : std_logic_vector(31 downto 0) := x"00" &    -- port number. designer should assign different number for different port.
                                                                    "00" &     -- reserved.
                                                                    "0" &      -- link bandwidth notification capability.
                                                                    "0" &      -- data link layer link active reporting capable.
                                                                    "0" &      -- surprise down error reporting capable.
                                                                    "0" &      -- clock power management.
                                                                    "000" &    -- L1 latency is less than 1us.
                                                                    "000" &    -- L0s latency is less than 64ns.
                                                                    "01" &     -- active state power management support. 01:L0s  11:L0s & L1
                                                                    "000000" & -- max link width value is automatically set to MAX_LINK_WIDTH.
                                                                    "0001";    -- max link speed value is automatically set to MAX_LINK_SPEED.

    PCIE_LINK_CTL_INIT           : std_logic_vector(15 downto 0) := x"0000";
    PCIE_LINK_STATUS_INIT        : std_logic_vector(15 downto 0) := x"0000";   -- negotiated link speed/width are fixed to MAX_LINK_SPEED/WIDTH so far.

    DEFRAMER_TLPBUF_DEPTH        : integer := 8;  -- buffer to store head of a TLP/DLLP which is received
    DEFRAMER_DLLPBUF_DEPTH       : integer := 8;  -- with the tail of previous packet in a single clk.
                                                  -- 8 bit (256 word) is sufficient for nForce 780i.
                                                  -- but it should be tested on various chipsets.
    TXDPBUF_DEPTH                : integer := 6;  -- 6 (i.e. 64 words) should be enough.
    SIMULATION                   : integer          := 0
  );

  port (

    st2_trigger     : out std_logic;
    st2_ltssm       : out LTSSM_STATE;
    st2_txd         : out each16b(NLANE-1 downto 0);
    st2_txk         : out each2b(NLANE-1 downto 0);
    st2_rxd         : out each16b(NLANE-1 downto 0);
    st2_rxk         : out each2b(NLANE-1 downto 0);

    phy_linkup      : out std_logic;
    dl_linkup       : out std_logic;
    
    clk             : in  std_logic;
    rstn            : in  std_logic;

    -- PIPE interface
    phystatus       : in  std_logic;
    powerdown       : out std_logic_vector(1 downto 0);
    txdetectrx      : out std_logic;
    txdata          : out std_logic_vector(NLANE*16-1 downto 0);
    txdatak         : out std_logic_vector(NLANE*2-1 downto 0);
    txelecidle      : out std_logic_vector(NLANE-1 downto 0);
    txcompl         : out std_logic_vector(NLANE-1 downto 0);
    rxpolarity      : out std_logic_vector(NLANE-1 downto 0);
    rxdata          : in  std_logic_vector(NLANE*16-1 downto 0);
    rxdatak         : in  std_logic_vector(NLANE*2-1 downto 0);
    rxvalid         : in  std_logic_vector(NLANE-1 downto 0);
    rxelecidle      : in  std_logic_vector(NLANE-1 downto 0);
    rxstatus        : in  std_logic_vector(NLANE*3-1 downto 0);

    --
    -- interface to user-side logic:
    --

    -- interface as a target (slave) device:
    slv_read        : out std_logic;
    slv_write       : out std_logic;
    slv_bar         : out std_logic_vector(6 downto 0);
    slv_addr        : out std_logic_vector(63 downto 0);
    slv_bytevalid   : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout     : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain      : in  std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq     : out std_logic;
    slv_writereq    : out std_logic;
    slv_accept      : in  std_logic;
    slv_bytecount   : out std_logic_vector(11 downto 0);

    -- interface as an initiator (master) device:
    ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_write          : out std_logic;
    ms_wraddr         : out std_logic_vector(31 downto 0);  -- 32-bit local address.
    ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);
    ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_read           : out std_logic;
    ms_rdaddr         : out std_logic_vector(31 downto 0);  -- 32-bit local address.
    ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

    -- interface to DMA controllers
    dma_control       : in  each7b(NDMACH-1 downto 0);
    dma_param         : in  each16b(NDMACH-1 downto 0);
    dma_status        : out each4b(NDMACH-1 downto 0);
    dma_fifocnt       : in  each13b(NDMACH-1 downto 0);

    dma_paddrlow_in   : in  each32b(NDMACH-1 downto 0);
    dma_paddrhigh_in  : in  each32b(NDMACH-1 downto 0);
    dma_laddr_in      : in  each32b(NDMACH-1 downto 0);
    dma_size_in       : in  each32b(NDMACH-1 downto 0);

    dma_paddrlow_out  : out each32b(NDMACH-1 downto 0);
    dma_paddrhigh_out : out each32b(NDMACH-1 downto 0);
    dma_laddr_out     : out each32b(NDMACH-1 downto 0);
    dma_size_out      : out each32b(NDMACH-1 downto 0)
  );

end gpcie;

architecture default of gpcie is

  component phymac
    generic (
      NLANE                  : integer := NLANE;
      N_FTS                  : integer := N_FTS;
      DATARATE               : std_logic_vector(7 downto 0) := DATARATE;
      TSCTRL                 : std_logic_vector(7 downto 0) := TSCTRL;
      DEFRAMER_TLPBUF_DEPTH  : integer := DEFRAMER_TLPBUF_DEPTH;
      DEFRAMER_DLLPBUF_DEPTH : integer := DEFRAMER_DLLPBUF_DEPTH;
      SIMULATION             : integer := SIMULATION
    );
    port (
      st2_txd           : out each16b(NLANE-1 downto 0);
      st2_txk           : out each2b(NLANE-1 downto 0);
      st2_rxd           : out each16b(NLANE-1 downto 0);
      st2_rxk           : out each2b(NLANE-1 downto 0);

      clk               : in  std_logic;
      rst               : in  std_logic;   -- active high

      -- PIPE interface
      txelecidleL       : out each1b(NLANE-1 downto 0);
      txdatakL          : out each2b(NLANE-1 downto 0);
      txdataL           : out each16b(NLANE-1 downto 0);
      rxelecidleL       : in  each1b(NLANE-1 downto 0);
      rxdatakL          : in  each2b(NLANE-1 downto 0);
      rxdataL           : in  each16b(NLANE-1 downto 0);
      rxpolarityL       : out each1b(NLANE-1 downto 0);

      -- DLL-MAC interface
      rxwedp            : out std_logic;
      rxdp              : out std_logic_vector(47 downto 0);
      rxwetp            : out std_logic;
      rxbetp            : out std_logic_vector(NLANE*2-1 downto 0);
      rxtptop           : out std_logic;
      rxtp              : out std_logic_vector(NLANE*16-1 downto 0);

      txdp              : in  std_logic_vector(47 downto 0);
      txemptydp         : in  std_logic;
      txrdreqdp         : out std_logic;
      txtp              : in  std_logic_vector(NLANE*16-1 downto 0);
      txtoptp           : in  std_logic;
      txwetp            : in  std_logic;
      txbetp            : in  std_logic_vector(NLANE*2-1 downto 0);

      -- TL-PHYMAC interface
      txreq             : in  std_logic;
      txack             : out std_logic;

      dmltssm           : out LTSSM_STATE;
      dmlinknumL        : out each8b(NLANE-1 downto 0);
      dmlanenumL        : out each8b(NLANE-1 downto 0);
      dmdatarateL       : out each8b(NLANE-1 downto 0);
      phy_linkup        : out std_logic;
      dl_linkup         : in  std_logic
    );
  end component;

  component datalink
    generic (
      MAX_PAYLOAD       : integer := MAX_PAYLOAD;
      CA_PH_VC0_INIT    : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT   : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT    : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT    : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT   : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT    : integer := CA_CD_VC0_INIT;
      TXDPBUF_DEPTH     : integer := TXDPBUF_DEPTH;
      NLANE             : integer := NLANE;
      SIMULATION        : integer := SIMULATION
    );
    port (
      st2_trigger       : out std_logic;

      clk               : in  std_logic;
      rst               : in  std_logic;   -- active high

      -- DLL-MAC interface
      rxwedp            : in  std_logic;
      rxdp              : in  std_logic_vector(47 downto 0);
      rxwetp            : in  std_logic;
      rxbetp            : in  std_logic_vector(NLANE*2-1 downto 0);
      rxtptop           : in  std_logic;
      rxtp              : in  std_logic_vector(NLANE*16-1 downto 0);

      txemptydp         : out std_logic;
      txrdreqdp         : in  std_logic;
      txdp              : out std_logic_vector(47 downto 0);
      txtoptp           : out std_logic;
      txwetp            : out std_logic;
      txbetp            : out std_logic_vector(NLANE*2-1 downto 0);
      txtp              : out std_logic_vector(NLANE*16-1 downto 0);

      -- TL-DLL interface
      rxtop             : out std_logic;
      rxwe              : out std_logic;
      rx                : out std_logic_vector(NLANE*16-1 downto 0);

      txtop             : in  std_logic;
      txwe              : in  std_logic;
      txbe              : in  std_logic_vector(NLANE*2-1 downto 0);
      tx                : in  std_logic_vector(NLANE*16-1 downto 0);
      tpsize            : in  std_logic_vector(OFFSET_WIDTH-1 downto 0);

      -- CA notified by transaction layer
      ca_nph            : in  std_logic_vector(7 downto 0);
      ca_ph             : in  std_logic_vector(7 downto 0);
      ca_ch             : in  std_logic_vector(7 downto 0);
      ca_npd            : in  std_logic_vector(11 downto 0);
      ca_pd             : in  std_logic_vector(11 downto 0);
      ca_cd             : in  std_logic_vector(11 downto 0);
      ca_p_flush        : in  std_logic;
      ca_np_flush       : in  std_logic;
      ca_c_flush        : in  std_logic;

      -- notify CL to transaction layer
      cl_nph            : out std_logic_vector( 7 downto 0);
      cl_ph             : out std_logic_vector( 7 downto 0);
      cl_ch             : out std_logic_vector( 7 downto 0);
      cl_npd            : out std_logic_vector(11 downto 0);
      cl_pd             : out std_logic_vector(11 downto 0);
      cl_cd             : out std_logic_vector(11 downto 0);

      inf_cl_ph         : out std_logic;
      inf_cl_pd         : out std_logic;
      inf_cl_nph        : out std_logic;
      inf_cl_npd        : out std_logic;
      inf_cl_ch         : out std_logic;
      inf_cl_cd         : out std_logic;

      -- core status
      phy_linkup        : in  std_logic;
      dl_linkup         : out std_logic
    );
  end component;

  component transaction
    generic (
      NLANE             : integer := NLANE;
      NVC               : integer := NVC;
      MAX_PAYLOAD       : integer := MAX_PAYLOAD;

      CA_PH_VC0_INIT    : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT   : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT    : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT    : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT   : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT    : integer := CA_CD_VC0_INIT;

      CL_PH_VC0_INIT    : integer := CL_PH_VC0_INIT;
      CL_NPH_VC0_INIT   : integer := CL_NPH_VC0_INIT;
      CL_CH_VC0_INIT    : integer := CL_CH_VC0_INIT;
      CL_PD_VC0_INIT    : integer := CL_PD_VC0_INIT;
      CL_NPD_VC0_INIT   : integer := CL_NPD_VC0_INIT;
      CL_CD_VC0_INIT    : integer := CL_CD_VC0_INIT;

      CFG_BAR0_INIT     : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT     : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT     : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT     : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT     : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT     : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT  : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;

      NDMACH            : integer := NDMACH;
      SIMULATION        : integer := SIMULATION
    );
    port (
      clk               : in  std_logic;
      rst               : in  std_logic;   -- active high

      -- TL-PHYMAC interface
      txreq             : out std_logic;
      txack             : in  std_logic;

      -- TL-DLL interface
      rxtop             : in  std_logic;
      rxwr              : in  std_logic;
      rx                : in  std_logic_vector(NLANE*16-1 downto 0);

      txtop             : out std_logic;
      txwr              : out std_logic;
      txbe              : out std_logic_vector(NLANE*2-1 downto 0);
      tx                : out std_logic_vector(NLANE*16-1 downto 0);
      tpsize            : out  std_logic_vector(OFFSET_WIDTH-1 downto 0);

      -- notify CA to data-link layer
      ca_nph            : out std_logic_vector(7 downto 0);
      ca_ph             : out std_logic_vector(7 downto 0);
      ca_ch             : out std_logic_vector(7 downto 0);
      ca_npd            : out std_logic_vector(11 downto 0);
      ca_pd             : out std_logic_vector(11 downto 0);
      ca_cd             : out std_logic_vector(11 downto 0);
      ca_p_flush        : out std_logic;
      ca_np_flush       : out std_logic;
      ca_c_flush        : out std_logic;

      -- CL notified by data-link layer
      cl_nph            : in  std_logic_vector( 7 downto 0);
      cl_ph             : in  std_logic_vector( 7 downto 0);
      cl_ch             : in  std_logic_vector( 7 downto 0);
      cl_npd            : in  std_logic_vector(11 downto 0);
      cl_pd             : in  std_logic_vector(11 downto 0);
      cl_cd             : in  std_logic_vector(11 downto 0);

      inf_cl_ph         : in  std_logic;
      inf_cl_pd         : in  std_logic;
      inf_cl_nph        : in  std_logic;
      inf_cl_npd        : in  std_logic;
      inf_cl_ch         : in  std_logic;
      inf_cl_cd         : in  std_logic;

      -- AL-TL interface
      mpls              : in  std_logic_vector(11 downto 0);

      cfg_bar0          : in  std_logic_vector(31 downto 0);
      cfg_bar1          : in  std_logic_vector(31 downto 0);
      cfg_bar2          : in  std_logic_vector(31 downto 0);
      cfg_bar3          : in  std_logic_vector(31 downto 0);
      cfg_bar4          : in  std_logic_vector(31 downto 0);
      cfg_bar5          : in  std_logic_vector(31 downto 0);
      cfg_bar_rom       : in  std_logic_vector(31 downto 0);

      cfg_read          : out std_logic;
      cfg_write         : out std_logic;
      cfg_addr          : out std_logic_vector(11 downto 0);
      cfg_be            : out std_logic_vector(3 downto 0);
      cfg_dataout       : out std_logic_vector(31 downto 0);
      cfg_datain        : in  std_logic_vector(31 downto 0);

      -- core interface
      slv_read          : out std_logic;
      slv_write         : out std_logic;
      slv_bar           : out std_logic_vector(6 downto 0);
      slv_addr          : out std_logic_vector(63 downto 0);
      slv_bytevalid     : out std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout       : out std_logic_vector(NLANE*16-1 downto 0);
      slv_datain        : in  std_logic_vector(NLANE*16-1 downto 0);

      slv_readreq       : out std_logic;
      slv_writereq      : out std_logic;
      slv_accept        : in  std_logic;
      slv_bytecount     : out std_logic_vector(11 downto 0);

      dma_fifocnt_off   : out std_logic_vector(4 downto 0);
      ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_write          : out std_logic;
      ms_wraddr         : out std_logic_vector(31 downto 0);
      ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);
      ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_read           : out std_logic;
      ms_rdaddr         : out std_logic_vector(31 downto 0);
      ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

      dmach_wrreq       : in  std_logic_vector(NDMACH-1 downto 0);
      dmach_wrack       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdreq       : in  std_logic_vector(NDMACH-1 downto 0);
      dmach_rdack       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_length      : in  each10b(NDMACH-1 downto 0);
      dmach_paddr       : in  each64b(NDMACH-1 downto 0);
      dmach_laddr       : in  each32b(NDMACH-1 downto 0);

      dmach_rdcpl       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdcpllength : out std_logic_vector(9 downto 0);
      dmach_rdcplbcnt   : out std_logic_vector(11 downto 0);
      dmach_rdcpladdr   : in  each32b(NDMACH-1 downto 0);

      phy_linkup        : in  std_logic;
      dl_linkup         : in  std_logic
    );
  end component;

  component cfgreg
    generic (
      CFG_VENDOR_ID_INIT           : std_logic_vector(15 downto 0) := CFG_VENDOR_ID_INIT;
      CFG_DEVICE_ID_INIT           : std_logic_vector(15 downto 0) := CFG_DEVICE_ID_INIT;
      CFG_COMMAND_INIT             : std_logic_vector(15 downto 0) := CFG_COMMAND_INIT;
      CFG_REVISION_ID_INIT         : std_logic_vector( 7 downto 0) := CFG_REVISION_ID_INIT;
      CFG_CLASS_CODE_INIT          : std_logic_vector(23 downto 0) := CFG_CLASS_CODE_INIT;
      CFG_HEADER_TYPE_INIT         : std_logic_vector( 7 downto 0) := CFG_HEADER_TYPE_INIT;
      CFG_BAR0_INIT                : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT                : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT                : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT                : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT                : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT                : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT             : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;
      CFG_SUB_VENDOR_ID_INIT       : std_logic_vector(15 downto 0) := CFG_SUB_VENDOR_ID_INIT;
      CFG_SUB_DEVICE_ID_INIT       : std_logic_vector(15 downto 0) := CFG_SUB_DEVICE_ID_INIT;
      CFG_INT_PIN_INIT             : std_logic_vector( 7 downto 0) := CFG_INT_PIN_INIT;

      PCIE_CAP_INIT                : std_logic_vector(15 downto 0) := PCIE_CAP_INIT;
      PCIE_DEVICE_CAP_INIT         : std_logic_vector(31 downto 0) := PCIE_DEVICE_CAP_INIT;
      PCIE_DEVICE_CTL_INIT         : std_logic_vector(15 downto 0) := PCIE_DEVICE_CTL_INIT;
      PCIE_DEVICE_STATUS_INIT      : std_logic_vector(15 downto 0) := PCIE_DEVICE_STATUS_INIT;
      PCIE_LINK_CAP_INIT           : std_logic_vector(31 downto 0) := PCIE_LINK_CAP_INIT;
      PCIE_LINK_CTL_INIT           : std_logic_vector(15 downto 0) := PCIE_LINK_CTL_INIT;
      PCIE_LINK_STATUS_INIT        : std_logic_vector(15 downto 0) := PCIE_LINK_STATUS_INIT;

      NLANE                        : integer := NLANE;
      MAX_LINK_SPEED               : std_logic_vector( 3 downto 0) := MAX_LINK_SPEED;
      MAX_READ_REQ_SIZE            : integer := MAX_READ_REQ_SIZE;
      MAX_PAYLOAD                  : integer := MAX_PAYLOAD;
      SIMULATION                   : integer := SIMULATION
    );
    port (
      clk                 : in  std_logic;
      rst                 : in  std_logic;

      mpls                : out std_logic_vector(11 downto 0);    -- max payload size
      mrrs                : out std_logic_vector(11+1 downto 0);  -- max read request size
      rcb                 : out std_logic;                        -- read completion boundary

      cfg_bar0_out        : out std_logic_vector(31 downto 0);
      cfg_bar1_out        : out std_logic_vector(31 downto 0);
      cfg_bar2_out        : out std_logic_vector(31 downto 0);
      cfg_bar3_out        : out std_logic_vector(31 downto 0);
      cfg_bar4_out        : out std_logic_vector(31 downto 0);
      cfg_bar5_out        : out std_logic_vector(31 downto 0);
      cfg_bar_rom_out     : out std_logic_vector(31 downto 0);

      rd                  : in  std_logic;
      wr                  : in  std_logic;
      addr                : in  std_logic_vector(11 downto 0);
      be                  : in  std_logic_vector(3 downto 0);
      datain              : in  std_logic_vector(31 downto 0);
      dataout             : out std_logic_vector(31 downto 0)
    );
  end component;

  component dmactl
    generic (
      NLANE             : integer := NLANE;
      NDMACH            : integer := NDMACH
    );
    port (
      clk               : in  std_logic;
      rst               : in  std_logic;

      mpls              : in  std_logic_vector(11 downto 0);    -- max payload size
      mrrs              : in  std_logic_vector(11+1 downto 0);  -- max read request size

      -- interface to user-side logic:
      dma_control       : in  std_logic_vector(6 downto 0);
      dma_param         : in  std_logic_vector(15 downto 0);
      dma_status        : out std_logic_vector(3 downto 0);
      dma_fifocnt       : in  std_logic_vector(12 downto 0);

      dma_paddrlow_in   : in  std_logic_vector(31 downto 0);
      dma_paddrhigh_in  : in  std_logic_vector(31 downto 0);
      dma_laddr_in      : in  std_logic_vector(31 downto 0);
      dma_size_in       : in  std_logic_vector(31 downto 0);

      dma_paddrlow_out  : out std_logic_vector(31 downto 0);
      dma_paddrhigh_out : out std_logic_vector(31 downto 0);
      dma_laddr_out     : out std_logic_vector(31 downto 0);
      dma_size_out      : out std_logic_vector(31 downto 0);

      -- interface to master controller:
      dma_fifocnt_off   : in  std_logic_vector(4 downto 0);
      wrreq             : out std_logic;
      wrack             : in  std_logic;
      rdreq             : out std_logic;
      rdack             : in  std_logic;

      length            : out std_logic_vector(PAYLOAD_WIDTH-1 downto 0);
      paddr             : out std_logic_vector(63 downto 0);
      laddr             : out std_logic_vector(31 downto 0);

      -- interface to slave controller:
      rdcpl             : in  std_logic;
      rdcpl_length      : in  std_logic_vector(9 downto 0);
      rdcpl_bcnt        : in  std_logic_vector(11 downto 0);
      rdcpl_addr        : out std_logic_vector(31 downto 0)
    );
  end component;

  -- PIPE interface
  signal txelecidleL       : each1b(NLANE-1 downto 0);
  signal txdatakL          : each2b(NLANE-1 downto 0);
  signal txdataL           : each16b(NLANE-1 downto 0);
  signal rxelecidleL       : each1b(NLANE-1 downto 0);
  signal rxdatakL          : each2b(NLANE-1 downto 0);
  signal rxdataL           : each16b(NLANE-1 downto 0);
  signal rxpolarityL       : each1b(NLANE-1 downto 0);

  -- signals commonly used inside the core
  signal linknumL          : each8b(NLANE-1 downto 0);
  signal lanenumL          : each8b(NLANE-1 downto 0);
  signal datarateL         : each8b(NLANE-1 downto 0);
  signal phy_linkup0       : std_logic;
  signal dl_linkup0        : std_logic;

  -- DLL-MAC interface
  signal rxdp              : std_logic_vector(47 downto 0);
  signal rxwedp            : std_logic;
  signal rxtp              : std_logic_vector(NLANE*16-1 downto 0);
  signal rxwetp            : std_logic;
  signal rxtptop           : std_logic;
  signal rxbetp            : std_logic_vector(NLANE*2-1 downto 0);

  signal txdp              : std_logic_vector(47 downto 0);
  signal txemptydp         : std_logic;
  signal txrdreqdp         : std_logic;
  signal txtp              : std_logic_vector(NLANE*16-1 downto 0);
  signal txtoptp           : std_logic;
  signal txwetp            : std_logic;
  signal txbetp            : std_logic_vector(NLANE*2-1 downto 0);

  -- TL-DLL interface
  signal rxtop             : std_logic;
  signal rxwe              : std_logic;
  signal rx                : std_logic_vector(NLANE*16-1 downto 0);
  signal txtop             : std_logic;
  signal txwe              : std_logic;
  signal txbe              : std_logic_vector(NLANE*2-1 downto 0);
  signal tx                : std_logic_vector(NLANE*16-1 downto 0);
  signal tpsize            : std_logic_vector(OFFSET_WIDTH-1 downto 0);

  signal ca_nph            : std_logic_vector(7 downto 0);
  signal ca_ph             : std_logic_vector(7 downto 0);
  signal ca_ch             : std_logic_vector(7 downto 0);
  signal ca_npd            : std_logic_vector(11 downto 0);
  signal ca_pd             : std_logic_vector(11 downto 0);
  signal ca_cd             : std_logic_vector(11 downto 0);
  signal ca_p_flush        : std_logic;
  signal ca_np_flush       : std_logic;
  signal ca_c_flush        : std_logic;

  signal cl_nph            : std_logic_vector(7 downto 0);
  signal cl_ph             : std_logic_vector(7 downto 0);
  signal cl_ch             : std_logic_vector(7 downto 0);
  signal cl_npd            : std_logic_vector(11 downto 0);
  signal cl_pd             : std_logic_vector(11 downto 0);
  signal cl_cd             : std_logic_vector(11 downto 0);

  signal inf_cl_nph        : std_logic;
  signal inf_cl_ph         : std_logic;
  signal inf_cl_ch         : std_logic;
  signal inf_cl_npd        : std_logic;
  signal inf_cl_pd         : std_logic;
  signal inf_cl_cd         : std_logic;

  -- TL-PHYMAC interface
  signal txreq             : std_logic;
  signal txack             : std_logic;

  -- AL-TL interface
  signal pcie_device_ctl   : std_logic_vector(15 downto 0);
  signal pcie_link_ctl     : std_logic_vector(15 downto 0);
  signal cfg_bar0          : std_logic_vector(31 downto 0);
  signal cfg_bar1          : std_logic_vector(31 downto 0);
  signal cfg_bar2          : std_logic_vector(31 downto 0);
  signal cfg_bar3          : std_logic_vector(31 downto 0);
  signal cfg_bar4          : std_logic_vector(31 downto 0);
  signal cfg_bar5          : std_logic_vector(31 downto 0);
  signal cfg_bar_rom       : std_logic_vector(31 downto 0);

  signal cfg_read          : std_logic;
  signal cfg_write         : std_logic;
  signal cfg_addr          : std_logic_vector(11 downto 0);
  signal cfg_be            : std_logic_vector(3 downto 0);
  signal cfg_dataout       : std_logic_vector(31 downto 0);
  signal cfg_datain        : std_logic_vector(31 downto 0);

  signal mpls              : std_logic_vector(11 downto 0);
  signal mrrs              : std_logic_vector(11+1 downto 0);
  signal rcb               : std_logic;

  signal dma_fifocnt_off   : std_logic_vector(4 downto 0);
  signal dmach_wrreq       : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_wrack       : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdreq       : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdack       : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_length      : each10b(NDMACH-1 downto 0);
  signal dmach_paddr       : each64b(NDMACH-1 downto 0);
  signal dmach_laddr       : each32b(NDMACH-1 downto 0);

  signal dmach_rdcpl       : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdcpllength : std_logic_vector(9 downto 0);
  signal dmach_rdcplbcnt   : std_logic_vector(11 downto 0);
  signal dmach_rdcpladdr   : each32b(NDMACH-1 downto 0);

  -- for debugging purpose
  signal trigger           : std_logic;
  signal dmltssm           : LTSSM_STATE;
  signal dmltssm1          : LTSSM_STATE;
  signal txd               : each16b(NLANE-1 downto 0);
  signal txk               : each2b(NLANE-1 downto 0);
  signal rxd               : each16b(NLANE-1 downto 0);
  signal rxk               : each2b(NLANE-1 downto 0);
  
  attribute keep : boolean;
  attribute keep of txd         : signal is true;
  attribute keep of txk         : signal is true;

-- synthesis translate_off

  attribute keep of trigger     : signal is true;
  attribute keep of dmltssm     : signal is true;
  attribute keep of clk         : signal is true;

  attribute keep of rxdp        : signal is true;
  attribute keep of rxwedp      : signal is true;
  attribute keep of txd         : signal is true;
  attribute keep of txk         : signal is true;

  attribute keep of dmltssm     : signal is true;
  attribute keep of rxd         : signal is true;
  attribute keep of rxk         : signal is true;
  attribute keep of rxelecidleL : signal is true;
  attribute keep of txelecidleL : signal is true;
  attribute keep of txdataL     : signal is true;
  attribute keep of txdatakL    : signal is true;

  attribute keep of rxdp        : signal is true;
  attribute keep of rxwedp      : signal is true;
  attribute keep of rxtp        : signal is true;
  attribute keep of rxwetp      : signal is true;

  attribute keep of txdp        : signal is true;
  attribute keep of txemptydp   : signal is true;
  attribute keep of txrdreqdp   : signal is true;
  attribute keep of txtp        : signal is true;
  attribute keep of txemptytp   : signal is true;
  attribute keep of txrdreqtp   : signal is true;

-- synthesis translate_on

  signal rxk1 : each2b(NLANE-1 downto 0);
  signal rxd1 : each16b(NLANE-1 downto 0);

begin  
  
  -- for debugging
  process (clk)
  begin
    if (rising_edge(clk)) then
      st2_trigger <= trigger;
      st2_ltssm   <= dmltssm;
      st2_txd     <= txd;
      st2_txk     <= txk;
      st2_rxd     <= rxd;
      st2_rxk     <= rxk;
      dmltssm1    <= dmltssm;

      rxk1 <= rxk;
      rxd1 <= rxd;

    end if;
  end process;

  phy_linkup <= phy_linkup0;
  dl_linkup <= dl_linkup0;

  phymac0 : phymac
    port map (
      st2_txd         => txd,
      st2_txk         => txk,
      st2_rxd         => rxd,
      st2_rxk         => rxk,

      clk             => clk,
      rst             => not rstn,

      txelecidleL     => txelecidleL,
      txdatakL        => txdatakL,
      txdataL         => txdataL,
      rxelecidleL     => rxelecidleL,
      rxdatakL        => rxdatakL,
      rxdataL         => rxdataL,
      rxpolarityL     => rxpolarityL,

      rxwedp          => rxwedp,
      rxdp            => rxdp,
      rxwetp          => rxwetp,
      rxtptop         => rxtptop,
      rxtp            => rxtp,
      rxbetp          => rxbetp,

      txdp            => txdp,
      txemptydp       => txemptydp,
      txrdreqdp       => txrdreqdp,

      txreq           => txreq,
      txack           => txack,
      txtoptp         => txtoptp,
      txwetp          => txwetp,
      txbetp          => txbetp,
      txtp            => txtp,

      dmltssm         => dmltssm,
      dmlinknumL      => linknumL,
      dmlanenumL      => lanenumL,
      dmdatarateL     => datarateL,

      phy_linkup      => phy_linkup0,
      dl_linkup       => dl_linkup0
    );

  sig_lane :  for i in 0 to NLANE-1 generate
    -- inputs
    rxdatakL(i)       <= rxdatak(2*(i+1)-1 downto 2*i);
    rxdataL(i)        <= rxdata(16*(i+1)-1 downto 16*i);
    rxelecidleL(i)    <= rxelecidle(i);
    -- outputs
    txdatak(2*(i+1)-1 downto 2*i)  <= txdatakL(i);
    txdata(16*(i+1)-1 downto 16*i) <= txdataL(i);
    txelecidle(i)                  <= txelecidleL(i);
    rxpolarity(i)                  <= rxpolarityL(i);
    txcompl(i)                     <= '0';
  end generate sig_lane;
  
  powerdown    <= "00";
  txdetectrx   <= '0';

  dll0 : datalink
    port map (
      st2_trigger     => trigger,

      clk             => clk,
      rst             => not (rstn and phy_linkup0),

      -- DLL-MAC interface
      rxdp            => rxdp,
      rxwedp          => rxwedp,

      txdp            => txdp,
      txemptydp       => txemptydp,
      txrdreqdp       => txrdreqdp,

      rxwetp          => rxwetp,
      rxtptop         => rxtptop,
      rxtp            => rxtp,
      txtoptp         => txtoptp,
      txwetp          => txwetp,
      txbetp          => txbetp,
      txtp            => txtp,
      rxbetp          => rxbetp,

      -- TL-DLL interface
      rxtop           => rxtop,
      rxwe            => rxwe,
      rx              => rx,
      txtop           => txtop,
      txwe            => txwe,
      txbe            => txbe,
      tx              => tx,
      tpsize          => tpsize,

      ca_nph          => ca_nph,
      ca_ph           => ca_ph,
      ca_ch           => ca_ch,
      ca_npd          => ca_npd,
      ca_pd           => ca_pd,
      ca_cd           => ca_cd,
      ca_p_flush      => ca_p_flush,
      ca_np_flush     => ca_np_flush,
      ca_c_flush      => ca_c_flush,

      cl_nph          => cl_nph,
      cl_ph           => cl_ph,
      cl_ch           => cl_ch,
      cl_npd          => cl_npd,
      cl_pd           => cl_pd,
      cl_cd           => cl_cd,

      inf_cl_nph      => inf_cl_nph,
      inf_cl_ph       => inf_cl_ph,
      inf_cl_ch       => inf_cl_ch,
      inf_cl_npd      => inf_cl_npd,
      inf_cl_pd       => inf_cl_pd,
      inf_cl_cd       => inf_cl_cd,

      phy_linkup      => phy_linkup0,
      dl_linkup       => dl_linkup0
    );

  tl0 : transaction
    port map (
      clk             => clk,
      rst             => not (rstn and phy_linkup0),
      rx              => rx,
      rxtop           => rxtop,
      rxwr            => rxwe,
      txreq           => txreq,
      txack           => txack,
      tx              => tx,
      tpsize          => tpsize,
      txtop           => txtop,
      txwr            => txwe,
      txbe            => txbe,

      ca_nph          => ca_nph,
      ca_ph           => ca_ph,
      ca_ch           => ca_ch,
      ca_npd          => ca_npd,
      ca_pd           => ca_pd,
      ca_cd           => ca_cd,
      ca_p_flush      => ca_p_flush,
      ca_np_flush     => ca_np_flush,
      ca_c_flush      => ca_c_flush,

      cl_nph          => cl_nph,
      cl_ph           => cl_ph,
      cl_ch           => cl_ch,
      cl_npd          => cl_npd,
      cl_pd           => cl_pd,
      cl_cd           => cl_cd,

      inf_cl_nph      => inf_cl_nph,
      inf_cl_ph       => inf_cl_ph,
      inf_cl_ch       => inf_cl_ch,
      inf_cl_npd      => inf_cl_npd,
      inf_cl_pd       => inf_cl_pd,
      inf_cl_cd       => inf_cl_cd,

      mpls            => mpls,

      cfg_bar0        => cfg_bar0,
      cfg_bar1        => cfg_bar1,
      cfg_bar2        => cfg_bar2,
      cfg_bar3        => cfg_bar3,
      cfg_bar4        => cfg_bar4,
      cfg_bar5        => cfg_bar5,
      cfg_bar_rom     => cfg_bar_rom,

      cfg_read        => cfg_read,
      cfg_write       => cfg_write,
      cfg_addr        => cfg_addr,
      cfg_be          => cfg_be,
      cfg_dataout     => cfg_dataout,
      cfg_datain      => cfg_datain,

      slv_read        => slv_read,
      slv_write       => slv_write,
      slv_bar         => slv_bar,
      slv_addr        => slv_addr,
      slv_bytevalid   => slv_bytevalid,
      slv_dataout     => slv_dataout,
      slv_datain      => slv_datain,

      slv_readreq     => slv_readreq,
      slv_writereq    => slv_writereq,
      slv_accept      => slv_accept,
      slv_bytecount   => slv_bytecount,

      ms_wrchannel    => ms_wrchannel,
      ms_write        => ms_write,
      ms_wraddr       => ms_wraddr,
      ms_wrdata       => ms_wrdata,

      ms_rdchannel    => ms_rdchannel,
      ms_read         => ms_read,
      ms_rdaddr       => ms_rdaddr,
      ms_rddata       => ms_rddata,

      dma_fifocnt_off => dma_fifocnt_off,
      dmach_wrreq     => dmach_wrreq,
      dmach_wrack     => dmach_wrack,
      dmach_rdreq     => dmach_rdreq,
      dmach_rdack     => dmach_rdack,
      dmach_length    => dmach_length,
      dmach_paddr     => dmach_paddr,
      dmach_laddr     => dmach_laddr,

      dmach_rdcpl       => dmach_rdcpl,
      dmach_rdcpllength => dmach_rdcpllength,
      dmach_rdcplbcnt   => dmach_rdcplbcnt,
      dmach_rdcpladdr   => dmach_rdcpladdr,

      phy_linkup      => phy_linkup0,
      dl_linkup       => dl_linkup0
    );

  cfgreg0 : cfgreg
    port map (
      clk                 => clk,
--      rst                 => not (rstn and phy_linkup0),
      rst                 => not rstn,  -- !!!

      mpls                => mpls,
      mrrs                => mrrs,
      rcb                 => rcb,

      cfg_bar0_out        => cfg_bar0,
      cfg_bar1_out        => cfg_bar1,
      cfg_bar2_out        => cfg_bar2,
      cfg_bar3_out        => cfg_bar3,
      cfg_bar4_out        => cfg_bar4,
      cfg_bar5_out        => cfg_bar5,
      cfg_bar_rom_out     => cfg_bar_rom,

      rd                  => cfg_read,
      wr                  => cfg_write,
      be                  => cfg_be,
      addr                => cfg_addr,
      datain              => cfg_dataout,
      dataout             => cfg_datain
    );

  dma_ch : for i in 0 to NDMACH-1 generate
    u0 : dmactl
      port map (
        clk               => clk,
--        rst               => not (rstn and phy_linkup0),
        rst                 => not rstn,  -- !!!

        dma_control       => dma_control(i),
        dma_param         => dma_param(i),
        dma_status        => dma_status(i),
        dma_fifocnt       => dma_fifocnt(i),

        dma_paddrlow_in   => dma_paddrlow_in(i),
        dma_paddrhigh_in  => dma_paddrhigh_in(i),
        dma_laddr_in      => dma_laddr_in(i),
        dma_size_in       => dma_size_in(i),

        dma_paddrlow_out  => dma_paddrlow_out(i),
        dma_paddrhigh_out => dma_paddrhigh_out(i),
        dma_laddr_out     => dma_laddr_out(i),
        dma_size_out      => dma_size_out(i),

        dma_fifocnt_off   => dma_fifocnt_off,

        mpls              => mpls,
        mrrs              => mrrs,

        wrreq             => dmach_wrreq(i),
        wrack             => dmach_wrack(i),
        rdreq             => dmach_rdreq(i),
        rdack             => dmach_rdack(i),
        length            => dmach_length(i),
        paddr             => dmach_paddr(i),
        laddr             => dmach_laddr(i),

        rdcpl             => dmach_rdcpl(i),
        rdcpl_length      => dmach_rdcpllength,
        rdcpl_bcnt        => dmach_rdcplbcnt,
        rdcpl_addr        => dmach_rdcpladdr(i)
      );

  end generate dma_ch;

end default;
