library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

package gpciepkg is

  --
  -- PCI Express-specific constants
  --

  -- flow control credit
  constant MAX_CREDIT_H : integer := 128;  -- max credit for header 
  constant MAX_CREDIT_D : integer := 2048; -- max credit for data (in 16byte unit)
  constant PAYLOAD_WIDTH : integer := 10;  -- payload size width in DW unit (10 bit denotes max possible payload is 4kB).

  -- payload size
  constant MAX_TLP_SIZE : integer := 8192; -- max-possible TLP size in byte unit. rounded up to powers of 2.
  constant OFFSET_WIDTH : integer := 13;   -- ceil(log2(MAX_TLP_SIZE))

  -- 10-bit character symbols (K-code)
  constant K_COM        : std_logic_vector := "10111100"; -- comma       0xbc
  constant K_STP        : std_logic_vector := "11111011"; -- start TLP   0xfb
  constant K_SDP        : std_logic_vector := "01011100"; -- start DLLP  0x5c
  constant K_END        : std_logic_vector := "11111101"; -- end         0xfd
  constant K_EDB        : std_logic_vector := "11111110"; -- end bad     0xfe
  constant K_PAD        : std_logic_vector := "11110111"; -- pad         0xf7
  constant K_SKP        : std_logic_vector := "00011100"; -- skip        0x1c
  constant K_FTS        : std_logic_vector := "00111100"; -- fast training sequence 0x3c
  constant K_IDL        : std_logic_vector := "01111100"; -- idle        0x7c
  constant LOGICAL_IDLE : std_logic_vector := "00000000"; -- send 0x00 if no data to send.
  constant TS1ID        : std_logic_vector := "01001010"; -- D10.2  0x4a
  constant TS1ID_N      : std_logic_vector := "10110101"; -- D21.5  0xb5, TS1ID inverted.
  constant TS2ID        : std_logic_vector := "01000101"; -- D5.2   0x45
  constant TS2ID_N      : std_logic_vector := "10111010"; -- D26.5  0xba, TS2ID inverted.

  -- DLLP type
  constant DPTYPE_ACK              : std_logic_vector := "00000000"; -- acknowledged              0x00
  constant DPTYPE_NAK              : std_logic_vector := "00010000"; -- not acknowledged          0x10
  constant DPTYPE_PM_ENTER_L1      : std_logic_vector := "00100000"; -- PM enter L1               0x20
  constant DPTYPE_PM_ENTER_L23     : std_logic_vector := "00100001"; -- PM enter L2 or L3         0x21
  constant DPTYPE_PM_ASR1          : std_logic_vector := "00100011"; -- PM active state request1  0x23
  constant DPTYPE_PM_REQACK        : std_logic_vector := "00100100"; -- PM request acknowledge    0x24
  constant DPTYPE_VENDOR           : std_logic_vector := "00110000"; -- vendor specific           0x30
  constant DPTYPE_INITFC1_P_VC0    : std_logic_vector := "01000000"; -- init   FC1, posted        0x40
  constant DPTYPE_INITFC1_NP_VC0   : std_logic_vector := "01010000"; -- init   FC1, non-posted    0x50
  constant DPTYPE_INITFC1_C_VC0    : std_logic_vector := "01100000"; -- init   FC1, completion    0x60
  constant DPTYPE_INITFC2_P_VC0    : std_logic_vector := "11000000"; -- init   FC2, posted        0xc0
  constant DPTYPE_INITFC2_NP_VC0   : std_logic_vector := "11010000"; -- init   FC2, non-posted    0xd0
  constant DPTYPE_INITFC2_C_VC0    : std_logic_vector := "11100000"; -- init   FC2, completion    0xe0
  constant DPTYPE_UPDATEFC_P_VC0   : std_logic_vector := "10000000"; -- update FC,  posted        0x80
  constant DPTYPE_UPDATEFC_NP_VC0  : std_logic_vector := "10010000"; -- update FC,  non-posted    0x90
  constant DPTYPE_UPDATEFC_C_VC0   : std_logic_vector := "10100000"; -- update FC,  completion    0xa0

  -- TLP format
  constant TPFMT_12BH_NODATA       : std_logic_vector := "00";  -- 12-byte header w/o data.
  constant TPFMT_16BH_NODATA       : std_logic_vector := "01";  -- 16-byte header w/o data.
  constant TPFMT_12BH_DATA         : std_logic_vector := "10";  -- 12-byte header w/data.
  constant TPFMT_16BH_DATA         : std_logic_vector := "11";  -- 16-byte header w/data.

  -- Flow Control Buffer type
  constant TPBUFTYPE_P             : std_logic_vector := "00";  -- posted.
  constant TPBUFTYPE_NP            : std_logic_vector := "01";  -- non posted.
  constant TPBUFTYPE_C             : std_logic_vector := "10";  -- completion.

  -- TLP type
  constant TPTYPE_MRd32            : std_logic_vector := "0000000";  -- 0x00
  constant TPTYPE_MRd64            : std_logic_vector := "0100000";  -- 0x20
  constant TPTYPE_MRdLk32          : std_logic_vector := "0000001";
  constant TPTYPE_MRdLk64          : std_logic_vector := "0100001";
  constant TPTYPE_MWr32            : std_logic_vector := "1000000";  -- 0x40
  constant TPTYPE_MWr64            : std_logic_vector := "1100000";  -- 0x60
  constant TPTYPE_IORd             : std_logic_vector := "0000010";
  constant TPTYPE_IOWr             : std_logic_vector := "1000010";
  constant TPTYPE_CfgRd0           : std_logic_vector := "0000100";  -- 0x04
  constant TPTYPE_CfgWr0           : std_logic_vector := "1000100";  -- 0x44
  constant TPTYPE_CfgRd1           : std_logic_vector := "0000101";
  constant TPTYPE_CfgWr1           : std_logic_vector := "1000101";
  constant TPTYPE_Msg              : std_logic_vector := "0110XXX";
  constant TPTYPE_MsgD             : std_logic_vector := "1110XXX";  -- 0x7X
  constant TPTYPE_Cpl              : std_logic_vector := "0001010";  -- 0x0a
  constant TPTYPE_CplD             : std_logic_vector := "1001010";  -- 0x4a
  constant TPTYPE_CplLk            : std_logic_vector := "0001011";
  constant TPTYPE_CplDLk           : std_logic_vector := "1001011";

  -- completion status
  constant CPLSTAT_SC              : std_logic_vector := "000";  -- successful completion
  constant CPLSTAT_UR              : std_logic_vector := "001";  -- unsupported request
  constant CPLSTAT_CRS             : std_logic_vector := "010";  -- configuration request retry status
  constant CPLSTAT_CA              : std_logic_vector := "100";  -- completer abort

  --
  -- type definitions
  --

  -- LTSSM States
  type LTSSM_STATE is (
    DETECT_QUIET, DETECT_ACTIVE,
    POLLING_ACTIVE, POLLING_COMPLIANCE, POLLING_CONFIGWAIT, POLLING_CONFIGACC, POLLING_SPEED,
    CONFIG_LINKSTART, CONFIG_LINKACC, CONFIG_LANEWAIT, CONFIG_LANEACC, CONFIG_COMPLETE, CONFIG_COMPLETE2, CONFIG_IDLE, CONFIG_IDLE2,
    L0,
    RECOVERY_LOCK, RECOVERY_CONFIG, RECOVERY_CONFIG2, RECOVERY_IDLE, RECOVERY_IDLE2, RECOVERY_SPEED,
    INVALID_STATE
  );
--  attribute enum_encoding: string;
--  attribute enum_encoding of LTSSM_STATE : type is "00000 00001 00010 00011 00100 00101 00110 00111 01000 01001 01010 01011 01100 01101 01110 01111 10000 10001 10010 10011 10100 10101 10110 10111 11000 11001 11010 11011 11100 11101 11110 11111 ";

  -- define arrays of std_logic_vector.
  type each1b  is array (natural range <>) of std_logic;
  type each2b  is array (natural range <>) of std_logic_vector(1 downto 0);
  type each4b  is array (natural range <>) of std_logic_vector(3 downto 0);
  type each7b  is array (natural range <>) of std_logic_vector(6 downto 0);
  type each8b  is array (natural range <>) of std_logic_vector(7 downto 0);
  type each10b is array (natural range <>) of std_logic_vector(9 downto 0);
  type each12b is array (natural range <>) of std_logic_vector(11 downto 0);
  type each13b is array (natural range <>) of std_logic_vector(12 downto 0);
  type each16b is array (natural range <>) of std_logic_vector(15 downto 0);
  type each32b is array (natural range <>) of std_logic_vector(31 downto 0);
  type each64b is array (natural range <>) of std_logic_vector(63 downto 0);

  --
  -- utility functions
  --

  -- returns log2(ceil(n)), where n is a natural number.
  function mylog2(
    n : integer
  ) return integer;

  -- convert 2-bit byte enable bits to 16-bit mask.
  function be_to_mask16(
    be    : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;

  -- convert 4-bit byte enable bits to 32-bit mask.
  function be_to_mask32(
    be    : in std_logic_vector(3 downto 0)
  )
  return std_logic_vector;

  function length_to_credit(
    len    : in std_logic_vector(9 downto 0)
  )
  return std_logic_vector;

  function have_enough_credit(
    limit   : in std_logic_vector;
    request : in std_logic_vector
  )
  return std_logic;

  function valid_seqno(
    received : in std_logic_vector;
    expected : in std_logic_vector
  )
  return std_logic;

end package;

package body gpciepkg is

  --
  -- a function to return ceil(log2(n)).
  -- n must be a natural number.
  --
  function mylog2(n : in integer) return integer is
    variable i : integer := 0;
  begin
    while (2**i < n) loop
      i := i + 1;
    end loop;
    return i;
  end mylog2;

  --
  -- convert 2-bit byte enable bits to 16-bit mask.
  --
  function be_to_mask16(
    be    : in std_logic_vector(1 downto 0)  -- byte enable
  )
  return std_logic_vector is

    variable mask  : std_logic_vector(15 downto 0);

  begin

    case be is
      when "11" =>
        mask := x"ffff";
      when "10" =>
        mask := x"ff00";
      when "01" =>
        mask := x"00ff";
      when others =>
          mask := (others => '0');
    end case;

    return mask;

  end be_to_mask16;

  --
  -- convert 4-bit byte enable bits to 32-bit mask.
  --
  function be_to_mask32(
    be    : in std_logic_vector(3 downto 0)  -- byte enable
  )
  return std_logic_vector is

    variable mask  : std_logic_vector(31 downto 0);

  begin

    case be is
      when "1111" =>
        mask := x"ffffffff";
      when "1110" =>
        mask := x"ffffff00";
      when "1101" =>
        mask := x"ffff00ff";
      when "1100" =>
        mask := x"ffff0000";

      when "1011" =>
        mask := x"ff00ffff";
      when "1010" =>
        mask := x"ff00ff00";
      when "1001" =>
        mask := x"ff0000ff";
      when "1000" =>
        mask := x"ff000000";

      when "0111" =>
        mask := x"00ffffff";
      when "0110" =>
        mask := x"00ffff00";
      when "0101" =>
        mask := x"00ff00ff";
      when "0100" =>
        mask := x"00ff0000";

      when "0011" =>
        mask := x"0000ffff";
      when "0010" =>
        mask := x"0000ff00";
      when "0001" =>
        mask := x"000000ff";
      when others =>
        mask := (others => '0');
    end case;

    return mask;

  end be_to_mask32;

  --
  -- convert length in DW unit into credit in 16-byte unit.
  -- the result is rounded up to 16-byte boundary.
  -- i.e., credit = ceil(length/4) = (length+3)>>2
  --
  function length_to_credit(
    len    : in std_logic_vector(9 downto 0)
  )
  return std_logic_vector is

    variable lenplus3 : std_logic_vector(9 downto 0);
    variable credit   : std_logic_vector(11 downto 0);

  begin

    lenplus3 := unsigned(len) + 3;
    credit := conv_std_logic_vector(0, credit'length - lenplus3'length + 2) &
              lenplus3(lenplus3'left downto 2);

    return credit;

  end length_to_credit;

  function have_enough_credit(
    limit   : in std_logic_vector;
    request : in std_logic_vector
  )
  return std_logic is

    variable diff   : std_logic_vector(limit'range);
    variable enough : std_logic;

  begin

    diff := unsigned(limit) - unsigned(request);
    enough := not diff(diff'left);

    return enough;

  end have_enough_credit;

  function valid_seqno(
    received : in std_logic_vector;
    expected : in std_logic_vector
  )
  return std_logic is

    variable diff                 : std_logic_vector(expected'range);
    variable larger_than_expected : std_logic;

  begin

    diff := unsigned(expected) - unsigned(received);
    larger_than_expected := diff(diff'left);

    -- received TLP is valid if its seqno is equal to or smaller than
    -- expected seqno.
    return not larger_than_expected;

  end valid_seqno;

end gpciepkg;
