--
-- A GPCIe instantiation in the User-side Logic
-- including DMA controller.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity hibctl is
  generic (
    DEVICE              : string;
    USE_ASMI            : integer;
    PIOWBUF_DEPTH       : integer :=  8; -- 256 words (x1:0.5kB   x4:2kB  x8: 4kB  x16: 8kB) x 2bufs
                                         -- should not exceed 14, 12, 11 for x1, x4, x8, respectively.
    TXBUF_DEPTH         : integer := 10; -- 1k  words (x1:2kB     x4:8kB  x8:16kB  x16:32kB)
    NLANE               : integer;
    NDMACH              : integer;
    SIMULATION          : integer
  );
  port (
    -- for debugging.
    txunderflow         : out std_logic;
    txoverflow          : out std_logic;
    dupr                : out std_logic;
    dupw                : out std_logic;
    
    clk                 : in  std_logic;
    pll_locked          : in  std_logic;
    mperst              : in  std_logic;

    -- interdace to GPCIe
    rstn_out            : out std_logic;

    slv_read            : in  std_logic;
    slv_write           : in  std_logic;
    slv_bar             : in  std_logic_vector(6 downto 0);
    slv_addr            : in  std_logic_vector(63 downto 0);
    slv_bytevalid       : in  std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout         : in  std_logic_vector(NLANE*16-1 downto 0);
    slv_datain          : out std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq         : in  std_logic;
    slv_writereq        : in  std_logic;
    slv_accept          : out std_logic;
    slv_bytecount       : in  std_logic_vector(11 downto 0);

    ms_wrchannel        : in  std_logic_vector(NDMACH-1 downto 0);
    ms_write            : in  std_logic;
    ms_wraddr           : in  std_logic_vector(31 downto 0);
    ms_wrdata           : out std_logic_vector(NLANE*16-1 downto 0);
    ms_rdchannel        : in  std_logic_vector(NDMACH-1 downto 0);
    ms_read             : in  std_logic;
    ms_rdaddr           : in  std_logic_vector(31 downto 0);
    ms_rddata           : in  std_logic_vector(NLANE*16-1 downto 0);

    dma_control         : out each7b(NDMACH-1 downto 0);
    dma_param           : out each16b(NDMACH-1 downto 0);
    dma_status          : in  each4b(NDMACH-1 downto 0);
    dma_fifocnt         : out each13b(NDMACH-1 downto 0);

    dma_paddrlow_in     : out each32b(NDMACH-1 downto 0);
    dma_paddrhigh_in    : out each32b(NDMACH-1 downto 0);
    dma_laddr_in        : out each32b(NDMACH-1 downto 0);
    dma_size_in         : out each32b(NDMACH-1 downto 0);

    dma_paddrlow_out    : in  each32b(NDMACH-1 downto 0);
    dma_paddrhigh_out   : in  each32b(NDMACH-1 downto 0);
    dma_laddr_out       : in  each32b(NDMACH-1 downto 0);
    dma_size_out        : in  each32b(NDMACH-1 downto 0);

    -- interface to the backend 
    hib_we              : out std_logic;
    hib_data            : out std_logic_vector(NLANE*16-1 downto 0);
    backend_we          : in std_logic;
    backend_data        : in std_logic_vector(NLANE*16-1 downto 0);
    reset_backend       : out std_logic; -- active high

    -- board information register initial value
    board_info          : in std_logic_vector(31 downto 0)
  );
end hibctl;

architecture default of hibctl is

  component scfifo
    generic (
      LPM_NUMWORDS            : natural := 2**TXBUF_DEPTH;
      LPM_WIDTHU              : natural := TXBUF_DEPTH;
      LPM_WIDTH               : natural := NLANE*16;
      LPM_SHOWAHEAD           : string  := "OFF";
      ADD_RAM_OUTPUT_REGISTER : string  := "ON";
      OVERFLOW_CHECKING       : string  := "OFF";
      UNDERFLOW_CHECKING      : string  := "OFF";
      USE_EAB                 : string  := "ON";
      INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
      LPM_TYPE                : string  := "scfifo"
      );
    port (
      clock    : in  std_logic;
      sclr     : in  std_logic;

      wrreq    : in  std_logic;
      data     : in  std_logic_vector(LPM_WIDTH-1 downto 0);

      rdreq    : in  std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
      full     : out std_logic;
      empty    : out std_logic;
      usedw    : out std_logic_vector(LPM_WIDTHU-1 downto 0)
    );
  end component;

  component piowbuf
    generic (
      DEPTH   : integer := PIOWBUF_DEPTH;
      WIDTH   : integer := NLANE*16
    );
    port (
      clk        : in  std_logic;
      rst        : in  std_logic;

      swap       : in  std_logic;
      wrcnt      : in  std_logic_vector(DEPTH-1 downto 0);

      wrin       : in  std_logic;
      bein       : in  std_logic_vector(WIDTH/8-1 downto 0);
      addrin     : in  std_logic_vector(DEPTH-1 downto 0);
      datain     : in  std_logic_vector(WIDTH-1 downto 0);

      wrlock0out : out std_logic;
      wrout      : out std_logic;
      dataout    : out std_logic_vector(WIDTH-1 downto 0)
    );
  end component;

  component asmictl
    generic (
      DEVICE   : string := DEVICE;
      USE_ASMI : integer := USE_ASMI
    );
    port  (
      clk      : in  std_logic;
      rst      : in  std_logic;

      wr       : in  std_logic;
      cmd      : in  std_logic_vector(23 downto 0);
      addr     : in  std_logic_vector(23 downto 0);
      datain   : in  std_logic_vector(7 downto 0);
      dataout  : out std_logic_vector(31 downto 0)
    );
  end component;

  signal arst            : std_logic;
  signal rstn            : std_logic;
  signal rst             : std_logic;
  signal rstn_cnt        : std_logic_vector(19 downto 0);
  signal reset_backend0  : std_logic;
  signal reset_dma       : std_logic;

  type array_slv_datain is array (5 downto 0) of std_logic_vector(slv_datain'range);
  signal slv_datain_bar  : array_slv_datain;

  signal test_mode       : std_logic_vector(1 downto 0);
  signal board_info_reg  : std_logic_vector(31 downto 0);
  signal dma0_fifocnt    : std_logic_vector(12 downto 0);
  signal dma1_fifocnt    : std_logic_vector(12 downto 0);  -- 0 to 4kB-1 or infinite (= 1 0000 0000 0000 = 4kB)

  signal hib_we0         : std_logic;
  signal hib_data0       : std_logic_vector(NLANE*16-1 downto 0);

  signal pio_swap        : std_logic;
  signal pio_wrcnt       : std_logic_vector(PIOWBUF_DEPTH-1 downto 0);
  signal pio_wrin        : std_logic;
  signal pio_addr        : std_logic_vector(PIOWBUF_DEPTH-1 downto 0);
  signal pio_wrlock0     : std_logic;
  signal pio_write       : std_logic;
  signal pio_wrdata      : std_logic_vector(NLANE*16-1 downto 0);
  signal pio_be          : std_logic_vector(NLANE*2-1 downto 0);
  signal pio_datain      : std_logic_vector(NLANE*16-1 downto 0);

  signal txwr            : std_logic;
  signal txdata          : std_logic_vector(NLANE*16-1 downto 0);
  signal txrd            : std_logic;
  signal txrst           : std_logic;
  signal txusedw         : std_logic_vector(TXBUF_DEPTH-1 downto 0);
  signal txfull          : std_logic;
  signal txempty         : std_logic;

  signal asmi_write    : std_logic;
  signal asmi_cmd      : std_logic_vector(23 downto 0);
  signal asmi_addr     : std_logic_vector(23 downto 0);
  signal asmi_datain   : std_logic_vector(7 downto 0);
  signal asmi_dataout  : std_logic_vector(31 downto 0);

  -- temporary buf for lower 16-bit of DMA registers.
  -- used only when NLANE=1.
  signal dma_laddr_0low       : std_logic_vector(15 downto 0);
  signal dma_paddrhigh_0low   : std_logic_vector(15 downto 0);
  signal dma_paddrlow_0low    : std_logic_vector(15 downto 0);
  signal dma_size_0low        : std_logic_vector(15 downto 0);
  signal dma_laddr_1low       : std_logic_vector(15 downto 0);
  signal dma_paddrhigh_1low   : std_logic_vector(15 downto 0);
  signal dma_paddrlow_1low    : std_logic_vector(15 downto 0);
  signal dma_size_1low        : std_logic_vector(15 downto 0);

  signal ms_wrdata0           : std_logic_vector(NLANE*16-1 downto 0);
  signal ms_wrdata1           : std_logic_vector(NLANE*16-1 downto 0);
  signal ms_wrdata2           : std_logic_vector(NLANE*16-1 downto 0);
  signal txdata1              : std_logic_vector(NLANE*16-1 downto 0);
  signal txdata2              : std_logic_vector(NLANE*16-1 downto 0);
  signal txrd1                : std_logic;
  
  attribute keep : boolean;
  attribute keep of txusedw      : signal is true;
  attribute keep of txunderflow  : signal is true;
  attribute keep of txoverflow   : signal is true;
  attribute keep of txrd         : signal is true;
  attribute keep of txwr         : signal is true;
  attribute keep of txfull       : signal is true;
  attribute keep of txempty      : signal is true;
  attribute keep of dupw, dupr   : signal is true;

begin

  -- generate synchronous reset signals, rstn & rst.
  rstn_out <= rstn;
  rst      <= not rstn;
  arst     <= pll_locked and mperst;
  process (clk, arst)
  begin
    if arst = '0' then
      rstn_cnt <= (others => '0');
      rstn <= '0';
    elsif rising_edge(clk) then
      if rstn_cnt(3) = '1'  and SIMULATION = 1 then       -- 8clk
        rstn <= '1';
      elsif rstn_cnt(19) = '1' and SIMULATION = 0 then    -- 4ms
        rstn <= '1';
      else
        rstn_cnt <= unsigned(rstn_cnt) + '1';
      end if;
    end if;
  end process;

#if 1 // severe timing margin. need no delay for response to write request.

  process (slv_bar(0), slv_writereq, slv_addr(8 downto 2), pio_write)
  begin
    -- accept read/write request by default.
    slv_accept <= '1';

    -- block write to pio_swap register
    -- until read from piowbuf is completed.
    if (NLANE = 1) then
      if (slv_bar(0) = '1' and slv_writereq = '1' and
          slv_addr(8 downto 1) = "10010000" and
          slv_bytecount >= conv_std_logic_vector(4, slv_bytecount'length) and
          pio_write = '1') then
        slv_accept <= '0';
      end if;
    else
      if (slv_bar(0) = '1' and slv_writereq = '1' and
          slv_addr(8 downto 2) = "1001000" and pio_write = '1') then
        slv_accept <= '0';
      end if;
    end if;
  end process;      

#else // larger timing margin. need 1clk delay for response to write request.

 process (clk)
  begin
    if (rising_edge(clk)) then
      -- accept read/write request by default.
      slv_accept <= '1';

      -- block write to pio_swap register
      -- until read from piowbuf is completed.
      if (slv_bar(0) = '1' and slv_writereq = '1' and
          slv_addr(8 downto 2) = "1001000" and pio_write = '1') then
        slv_accept <= '0';
      end if;
    end if;
  end process;      

#endif

  --
  --  Slave (Target) I/O handling.
  --
  --
  --   BAR0 register map:
  --
  --     000h   DMA0 local address (r/w)
  --     004h   DMA0 PCI address MSB (r/w)
  --     008h   DMA0 PCI address LSB (r/w)
  --     00ch   DMA0 data size (r/w), with side effect to kick off a new transfer.
  --     010h   DMA0 command/param (wo)
  --                 1       : abort current transfer.
  --     014h   DMA0 status (ro)
  --                 13      : dma done.
  --                 12..0   : dma fifo count.
  --
  --     040h   DMA1 local address (r/w)
  --     044h   DMA1 PCI address MSB (r/w)
  --     048h   DMA1 PCI address LSB (r/w)
  --     04ch   DMA1 data size (r/w), with side effect to kick off a new transfer.
  --     050h   DMA1 command/param (wo)
  --                 1       : abort current transfer.
  --     054h   DMA1 status (ro)
  --                 13      : dma done.
  --                 12..0   : dma fifo count.
  --
  --     120h    command
  --                 31      : reset backend & ASMI controller. (wo)
  --                 25      : PIOW double buffer0 write lock. (ro)
  --                 24      : swap PIOW double buffer. (wo)
  --                 ..0     : PIOW double buffer write count (in byte unit). (r/w)
  --
  --     124h    testmode (for debugging and performance measurement)
  --                 1:0    : 00: normal mode. dma_wrdata is passed to the backend.
  --                          01: dma_wrdata is looped back to internal FIFO.
  --                          10: internal FIFO is bypassed (received data is abandoned).
  --     130h    KFCR board info register
  --     138h    ASMI controll register
  --     140h    PLLCONF controll register
  --
  --
  reset_backend <= reset_backend0;

#if DMACNT_NOWAIT
  dma_fifocnt(0) <= dma0_fifocnt;
  dma_fifocnt(1) <= dma1_fifocnt;
#else
  process (clk)
  begin
    if (rising_edge(clk)) then
      dma_fifocnt(0) <= dma0_fifocnt;
      dma_fifocnt(1) <= dma1_fifocnt;
    end if;
  end process;
#endif
  ms_wrdata <= ms_wrdata0;

  gendmax1 : if (NLANE = 1) generate
    process(clk, rst, board_info)
    begin
      if (rst = '1') then

        asmi_write <= '0';
        pio_wrcnt <= (others => '0');
        test_mode <= (others => '0');
        board_info_reg(31 downto 14) <= board_info(31 downto 14);
        board_info_reg(13 downto 12) <= conv_std_logic_vector(PIOWBUF_DEPTH-7, 2);
        board_info_reg(11 downto 10) <= conv_std_logic_vector(TXBUF_DEPTH-8, 2);
        board_info_reg(9 downto 0) <= board_info(9 downto 0);
        dma_laddr_0low       <= (others => '0');
        dma_paddrhigh_0low   <= (others => '0');
        dma_paddrlow_0low    <= (others => '0');
        dma_size_0low        <= (others => '0');
        dma_laddr_1low       <= (others => '0');
        dma_paddrhigh_1low   <= (others => '0');
        dma_paddrlow_1low    <= (others => '0');
        dma_size_1low        <= (others => '0');

      elsif (rising_edge(clk)) then

        -- 1-clk pulse
        reset_backend0 <= '0';
        reset_dma <= '0';
        pio_swap <= '0';
        asmi_write <= '0';

        --
        -- BAR0 write
        --
        dma_control(0) <= (others => '0');
        dma_control(1) <= (others => '0');

        if (slv_bar(0) = '1' and slv_write = '1') then
          case slv_addr(8 downto 1) is

            --
            -- DMA channel 0
            --
            when "00000000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 00h  DMA0 local address lower 16-bit.
                dma_laddr_0low <= slv_dataout;
              end if;

            when "00000001" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 02h  DMA0 local address higher 16-bit.
                dma_control(0)(2) <= '1';
                dma_laddr_in(0) <= slv_dataout & dma_laddr_0low;
              end if;

            when "00000010" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 04h  DMA0 PCI address bit 47:32.
                dma_paddrhigh_0low <= slv_dataout;
              end if;

            when "00000011" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 06h  DMA0 PCI address bit 63:48.
                dma_control(0)(1) <= '1';
                dma_paddrhigh_in(0) <= slv_dataout & dma_paddrhigh_0low;
              end if;

            when "00000100" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 08h  DMA0 PCI address bit 15:0.
                dma_paddrlow_0low <= slv_dataout;
              end if;

            when "00000101" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 0ah  DMA0 PCI address bit 31:16.
                dma_control(0)(0) <= '1';
                dma_paddrlow_in(0) <= slv_dataout & dma_paddrlow_0low;
              end if;

            when "00000110" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 0ch  DMA0 data size lower 16-bit.
                dma_size_0low <= slv_dataout;
              end if;

            when "00000111" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 0eh  DMA0 data size higher 16-bit.
                dma_control(0)(3) <= '1';
                dma_size_in(0) <= slv_dataout & dma_size_0low;
                dma_control(0)(5) <= '1';  -- kickoff a new transfer.
              end if;

            when "00001000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 10h  DMA0 command/param
                -- set DMA parameter.
                dma_control(0)(4) <= '1';
                dma_param(0) <= x"0000";      -- DMA read, fifo mode.

                -- abort current transfer.
                if (slv_dataout(1) = '1') then
                  dma_control(0)(6) <= '1';
                  reset_dma <= '1';
                end if;

              end if;

            --
            -- DMA channel 1
            --
            when "00100000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 40h  DMA1 local address lower 16-bit.
                dma_laddr_1low <= slv_dataout;
              end if;

            when "00100001" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 42h  DMA1 local address higher 16-bit.
                dma_control(1)(2) <= '1';
                dma_laddr_in(1) <= slv_dataout & dma_laddr_1low;
              end if;

            when "00100010" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 44h  DMA1 PCI address bit 47:32.
                dma_paddrhigh_1low <= slv_dataout;
              end if;

            when "00100011" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 46h  DMA1 PCI address bit 63:48.
                dma_control(1)(1) <= '1';
                dma_paddrhigh_in(1) <= slv_dataout & dma_paddrhigh_1low;
              end if;

            when "00100100" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 48h  DMA1 PCI address bit 15:0.
                dma_paddrlow_1low <= slv_dataout;
              end if;

            when "00100101" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 4ah  DMA1 PCI address bit 31:16.
                dma_control(1)(0) <= '1';
                dma_paddrlow_in(1) <= slv_dataout & dma_paddrlow_1low;
              end if;

            when "00100110" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 4ch  DMA1 data size lower 16-bit.
                dma_size_1low <= slv_dataout;
              end if;

            when "00100111" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 4eh  DMA1 data size higher 16-bit.
                dma_control(1)(3) <= '1';
                dma_size_in(1) <= slv_dataout & dma_size_1low;
                dma_control(1)(5) <= '1';  -- kickoff a new transfer.
              end if;


            when "00101000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 50h  DMA1 command/param
                -- set DMA parameter.
                dma_control(1)(4) <= '1';
                dma_param(1) <= x"0100";      -- DMA write, fifo mode.

                -- abort current transfer.
                if (slv_dataout(1) = '1') then
                  dma_control(1)(6) <= '1';
                  reset_dma <= '1';
                end if;
              end if;

              
            when "10010000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 120h command
                pio_wrcnt <= slv_dataout(pio_wrcnt'left+1 downto 1);
              end if;

            when "10010001" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 122h command
                pio_swap  <= slv_dataout(8);
                if (slv_dataout(15) = '1') then  -- writing a '1' to bit 15 resets backend pipeline.
                  reset_backend0 <= '1';
                end if;
              end if;

            when "10010100" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 128h test mode.
                test_mode <= slv_dataout(1 downto 0);
              end if;

            when "10011000" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 130h KFCR board info lower 16-bit.
                board_info_reg(15 downto 0) <= slv_dataout(15 downto 0);
              end if;

            when "10011001" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 132h KFCR board info higher 16-bit.
                board_info_reg(31 downto 16) <= slv_dataout(15 downto 0);
              end if;

            when "10011100" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 138h ASMI controll lower 16-bit.
                -- to be implemented.
              end if;

            when "10011101" =>
              if (slv_bytevalid(1 downto 0) = "11") then      -- 13ah ASMI controll higher 16-bit.
                -- to be implemented.
              end if;

            when "10100000" =>

            when others =>

          end case;
        end if;

        --
        -- BAR0 read
        --
        slv_datain_bar(0) <= (others => '0');

        if (slv_bar(0) = '1' and slv_read = '1') then
          case slv_addr(8 downto 1) is
            --
            -- DMA channel 0
            --
            when "00000000" =>
              slv_datain_bar(0)(15 downto 0) <= dma_laddr_out(0)(15 downto 0);      -- 00h  DMA0 local address lower 16-bit.
            when "00000001" =>
              slv_datain_bar(0)(15 downto 0) <= dma_laddr_out(0)(31 downto 16);     -- 02h  DMA0 local address higher 16-bit.

            when "00000010" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrhigh_out(0)(15 downto 0);  -- 04h  DMA0 PCI address bit 47:32.
            when "00000011" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrhigh_out(0)(31 downto 16); -- 06h  DMA0 PCI address bit 63:48.

            when "00000100" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrlow_out(0)(15 downto 0);   -- 08h  DMA0 PCI address bit 15:0.
            when "00000101" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrlow_out(0)(31 downto 16);  -- 0ah  DMA0 PCI address bit 31:16.

            when "00000110" =>
              slv_datain_bar(0)(15 downto 0) <= dma_size_out(0)(15 downto 0);       -- 0ch  DMA0 data size lower 16-bit.
            when "00000111" =>
              slv_datain_bar(0)(15 downto 0) <= dma_size_out(0)(31 downto 16);      -- 0eh  DMA0 data size higher 16-bit.

            when "00001010" =>
              slv_datain_bar(0)(13) <= dma_status(0)(3);                            -- 14h  DMA0 status
              slv_datain_bar(0)(12 downto 0) <= dma0_fifocnt;

            --
            -- DMA channel 1
            --
            when "00100000" =>
              slv_datain_bar(0)(15 downto 0) <= dma_laddr_out(1)(15 downto 0);      -- 40h  DMA1 local address lower 16-bit.
            when "00100001" =>
              slv_datain_bar(0)(15 downto 0) <= dma_laddr_out(1)(31 downto 16);     -- 42h  DMA1 local address higher 16-bit.

            when "00100010" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrhigh_out(1)(15 downto 0);  -- 44h  DMA1 PCI address bit 47:32.
            when "00100011" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrhigh_out(1)(31 downto 16); -- 46h  DMA1 PCI address bit 63:48.

            when "00100100" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrlow_out(1)(15 downto 0);   -- 48h  DMA1 PCI address bit 15:0.
            when "00100101" =>
              slv_datain_bar(0)(15 downto 0) <= dma_paddrlow_out(1)(31 downto 16);  -- 4ah  DMA1 PCI address bit 31:16.

            when "00100110" =>
              slv_datain_bar(0)(15 downto 0) <= dma_size_out(1)(15 downto 0);       -- 4ch  DMA1 data size lower 16-bit.
            when "00100111" =>
              slv_datain_bar(0)(15 downto 0) <= dma_size_out(1)(31 downto 16);      -- 4eh  DMA1 data size higher 16-bit.

            when "00101010" =>
              slv_datain_bar(0)(13) <= dma_status(1)(3);                            -- 54h  DMA1 status
              slv_datain_bar(0)(12 downto 0) <= dma1_fifocnt;

            when "10010000" =>
              slv_datain_bar(0)(pio_wrcnt'left downto 0) <= pio_wrcnt;              -- 120h command lower 16-bit.

            when "10010001" =>
              slv_datain_bar(0)(9) <= pio_wrlock0;                                  -- 122h command higher 16-bit.
              slv_datain_bar(0)(10) <= not pio_wrlock0;

            when "10010100" =>
              slv_datain_bar(0)(1 downto 0) <= test_mode;                           -- 128h test mode.

            when "10011000" =>
              slv_datain_bar(0)(15 downto 0) <= board_info_reg(15 downto 0);        -- 130h KFCR board info lower 16-bit.
            when "10011001" =>
              slv_datain_bar(0)(15 downto 0) <= board_info_reg(31 downto 16);       -- 132h KFCR board info higher 16-bit.

            when "10011100" =>
              slv_datain_bar(0)(15 downto 0) <= asmi_dataout(15 downto 0);          -- 138h ASMI control lower 16-bit.
            when "10011101" =>
              slv_datain_bar(0)(15 downto 0) <= asmi_dataout(31 downto 16);         -- 13ah ASMI control higher 16-bit.

            when "10100000" =>
              slv_datain_bar(0)(15 downto 0) <= x"ef98";                            -- 140h PLL configuration lower 16-bit.
            when "10100001" =>
              slv_datain_bar(0)(15 downto 0) <= x"abcd";                            -- 142h PLL configuration higher 16-bit.

            when others =>

          end case;
        end if;

      end if; -- clk
    end process;
  end generate gendmax1;

  gendmax4 : if (NLANE >= 4) generate
    process(clk, rst, board_info)
      variable pio_wrcnttmp       : std_logic_vector(PIOWBUF_DEPTH-1+1 downto 0);
    begin
      if (rst = '1') then

        asmi_write <= '0';
        pio_wrcnt <= (others => '0');
        test_mode <= (others => '0');
        board_info_reg(31 downto 14) <= board_info(31 downto 14);
        board_info_reg(13 downto 12) <= conv_std_logic_vector(PIOWBUF_DEPTH-7, 2);
        board_info_reg(11 downto 10) <= conv_std_logic_vector(TXBUF_DEPTH-8, 2);
        board_info_reg(9 downto 0) <= board_info(9 downto 0);

    elsif (rising_edge(clk)) then

      -- 1-clk pulse
      reset_backend0 <= '0';
      reset_dma <= '0';
      pio_swap <= '0';
      asmi_write <= '0';

      --
      -- BAR0 write
      --
      dma_control(0) <= (others => '0');
      dma_control(1) <= (others => '0');

      if (slv_bar(0) = '1' and slv_write = '1') then
        case slv_addr(8 downto 2) is

          --
          -- DMA channel 0
          --
          when "0000000" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 00h  DMA0 local address
              dma_control(0)(2) <= '1';
              dma_laddr_in(0) <= slv_dataout(31 downto 0);
            end if;

            if (slv_bytevalid(7 downto 4) = x"f") then      -- 04h  DMA0 PCI address higher 32-bit.
              dma_control(0)(1) <= '1';
              dma_paddrhigh_in(0) <= slv_dataout(63 downto 32);
            end if;

          when "0000001" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 04h  DMA0 PCI address higher 32-bit.
              dma_control(0)(1) <= '1';
              dma_paddrhigh_in(0) <= slv_dataout(31 downto 0);
            end if;

          when "0000010" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 08h  DMA0 PCI address lower 32-bit.
              dma_control(0)(0) <= '1';
              dma_paddrlow_in(0) <= slv_dataout(31 downto 0);
            end if;

            if (slv_bytevalid(7 downto 4) = x"f") then      -- 0ch  DMA0 data size
              dma_control(0)(3) <= '1';
              dma_size_in(0) <= slv_dataout(63 downto 32);
              dma_control(0)(5) <= '1';  -- kickoff a new transfer.
            end if;

          when "0000011" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 0ch  DMA0 data size
              dma_control(0)(3) <= '1';
              dma_size_in(0) <= slv_dataout(31 downto 0);
              dma_control(0)(5) <= '1';  -- kickoff a new transfer.
            end if;

          when "0000100" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 10h  DMA0 command/param
              -- set DMA parameter.
              dma_control(0)(4) <= '1';
              dma_param(0) <= x"0000";      -- DMA read, fifo mode.

              -- abort current transfer.
              if (slv_dataout(1) = '1') then
                dma_control(0)(6) <= '1';
                reset_dma <= '1';
              end if;

            end if;

          --
          -- DMA channel 1
          --
          when "0010000" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 40h  DMA1 local address
              dma_control(1)(2) <= '1';
              dma_laddr_in(1) <= slv_dataout(31 downto 0);
            end if;

            if (slv_bytevalid(7 downto 4) = x"f") then      -- 44h  DMA1 PCI address higher 32-bit.
              dma_control(1)(1) <= '1';
              dma_paddrhigh_in(1) <= slv_dataout(63 downto 32);
            end if;

          when "0010001" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 44h  DMA1 PCI address higher 32-bit.
              dma_control(1)(1) <= '1';
              dma_paddrhigh_in(1) <= slv_dataout(31 downto 0);
            end if;

          when "0010010" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 48h  DMA1 PCI address lower 32-bit.
              dma_control(1)(0) <= '1';
              dma_paddrlow_in(1) <= slv_dataout(31 downto 0);
            end if;

            if (slv_bytevalid(7 downto 4) = x"f") then      -- 4ch  DMA1 data size
              dma_control(1)(3) <= '1';
              dma_size_in(1) <= slv_dataout(63 downto 32);
              dma_control(1)(5) <= '1';  -- kickoff a new transfer.
            end if;


          when "0010011" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 4ch  DMA1 data size
              dma_control(1)(3) <= '1';
              dma_size_in(1) <= slv_dataout(31 downto 0);
              dma_control(1)(5) <= '1';  -- kickoff a new transfer.
            end if;


          when "0010100" =>
            if (slv_bytevalid(3 downto 0) = x"f") then      -- 50h  DMA1 command/param
              -- set DMA parameter.
              dma_control(1)(4) <= '1';
              dma_param(1) <= x"0100";      -- DMA write, fifo mode.

              -- abort current transfer.
              if (slv_dataout(1) = '1') then
                dma_control(1)(6) <= '1';
                reset_dma <= '1';
              end if;

            end if;

          when "1001000" =>
            if (slv_bytevalid(3 downto 0) = x"f") then  -- 120h command
              case NLANE is
                when 4 => 
                  pio_wrcnt <= slv_dataout(pio_wrcnt'left+3 downto 3);

                when 8 => 
                  -- round up to a multiple of 16byte.
                  pio_wrcnttmp := slv_dataout(pio_wrcnt'left+4 downto 3);
                  pio_wrcnttmp := unsigned(pio_wrcnttmp) + 1;
                  pio_wrcnt <= pio_wrcnttmp(pio_wrcnt'left+1 downto 1);

                when others =>

              end case;
              pio_swap  <= slv_dataout(24);

              if (slv_dataout(31) = '1') then  -- writing a '1' to bit 31 resets backend pipeline.
                reset_backend0 <= '1';
              end if;
            end if;

          when "1001010" =>
            if (slv_bytevalid(3 downto 0) = x"f") then       -- 128h test mode
              test_mode <= slv_dataout(1 downto 0);
            end if;

          when "1001100" =>
            if (slv_bytevalid(3 downto 0) = x"f") then       -- 130h KFCR board info
              board_info_reg <= slv_dataout(31 downto 0);
            end if;

          when "1001110" =>
            if (slv_bytevalid(3 downto 0) = x"f") then       -- 138h ASMI controll
              asmi_write <= '1';
              asmi_cmd <= slv_dataout(23+8 downto 8);
              asmi_addr <= slv_dataout(23+8 downto 8);
              asmi_datain <= slv_dataout(7 downto 0);
            end if;

          when "1010000" =>
--            if (slv_bytevalid(3 downto 0) = x"f") then       -- 140h PLL configuration
--            end if;

          when others =>

        end case;
      end if;

      --
      -- BAR0 read
      --
      slv_datain_bar(0) <= (others => '0');

      if (slv_bar(0) = '1' and slv_read = '1') then
        case slv_addr(8 downto 2) is
          --
          -- DMA channel 0
          --
          when "0000000" =>
            slv_datain_bar(0)(31 downto 0) <= dma_laddr_out(0);        -- 00h  DMA0 local address
            slv_datain_bar(0)( 63 downto 32) <= dma_paddrhigh_out(0);  -- 04h  DMA0 PCI address higher 32-bit.

          when "0000001" =>
            slv_datain_bar(0)( 31 downto  0) <= dma_paddrhigh_out(0);  -- 04h  DMA0 PCI address higher 32-bit.

          when "0000010" =>
            slv_datain_bar(0)( 31 downto  0) <= dma_paddrlow_out(0);   -- 08h  DMA0 PCI address lower 32-bit.
            slv_datain_bar(0)(63 downto 32) <= dma_size_out(0);        -- 0ch  DMA0 data size

          when "0000011" =>
            slv_datain_bar(0)(31 downto 0) <= dma_size_out(0);         -- 0ch  DMA0 data size

          when "0000100" =>
            slv_datain_bar(0)(32+13) <= dma_status(0)(3);              -- 14h  DMA0 status
            slv_datain_bar(0)(32+12 downto 32+0) <= dma0_fifocnt;

          when "0000101" =>
            slv_datain_bar(0)(13) <= dma_status(0)(3);                 -- 14h  DMA0 status
            slv_datain_bar(0)(12 downto 0) <= dma0_fifocnt;

          --
          -- DMA channel 1
          --
          when "0010000" =>
            slv_datain_bar(0)( 31 downto  0) <= dma_laddr_out(1);      -- 40h  DMA1 local address
            slv_datain_bar(0)( 63 downto 32) <= dma_paddrhigh_out(1);  -- 44h  DMA1 PCI address higher 32-bit.

          when "0010001" =>
            slv_datain_bar(0)( 31 downto  0) <= dma_paddrhigh_out(1);  -- 44h  DMA1 PCI address higher 32-bit.

          when "0010010" =>
            slv_datain_bar(0)(31 downto  0) <= dma_paddrlow_out(1);    -- 48h  DMA1 PCI address lower 32-bit.
            slv_datain_bar(0)(63 downto 32) <= dma_size_out(1);        -- 4ch  DMA1 data size

          when "0010011" =>
            slv_datain_bar(0)(31 downto 0) <= dma_size_out(1);         -- 4ch  DMA1 data size

          when "0010100" =>
            slv_datain_bar(0)(32+13) <= dma_status(1)(3);              -- 54h  DMA1 status
            slv_datain_bar(0)(32+12 downto 32+0) <= dma1_fifocnt;

          when "0010101" =>
            slv_datain_bar(0)(13) <= dma_status(1)(3);                 -- 54h  DMA1 status
            slv_datain_bar(0)(12 downto 0) <= dma1_fifocnt;

          when "1001000" =>
            slv_datain_bar(0)(25) <= pio_wrlock0;                      -- 120h command
            slv_datain_bar(0)(26) <= not pio_wrlock0;
            slv_datain_bar(0)(pio_wrcnt'left downto 0) <= pio_wrcnt;

          when "1001010" =>
            slv_datain_bar(0)(1 downto 0) <= test_mode;                -- 128h test mode

          when "1001100" =>
            slv_datain_bar(0)(31 downto 0) <= board_info_reg;          -- 130h KFCR board info

          when "1001110" =>
            slv_datain_bar(0)(31 downto 0) <= asmi_dataout;            -- 138h ASMI control

          when "1010000" =>
            slv_datain_bar(0)(31 downto 0) <= x"abcdef98";             -- 140h PLL configuration

          when others =>

        end case;
      end if;

    end if; -- clk
  end process;

  end generate gendmax4;


  --
  -- slv_datain mux
  --
  process(slv_bar, slv_datain_bar)
  begin
    if (slv_bar(0) = '1') then
      slv_datain <= slv_datain_bar(0);
    else
      slv_datain <= slv_datain_bar(1);
    end if;
  end process;

  --
  -- write request from hib to backend (DMA read or PIO write).
  --
  hib_we <= hib_we0;
  process(test_mode, ms_read, ms_rdchannel, pio_write)
  begin
    hib_we0 <= '0';
    if (test_mode = "00" or test_mode = "01") then
      if (ms_read = '1' and ms_rdchannel(0) = '1') then
        hib_we0 <= '1';
      elsif (pio_write = '1') then
        hib_we0 <= '1';
      end if;
    end if;
  end process;

  hib_data <= hib_data0;
  process (pio_write, pio_wrdata, ms_rddata)
  begin
    if (pio_write = '1') then
      hib_data0 <= pio_wrdata;
    else
      hib_data0 <= ms_rddata;
    end if;
  end process;

  -- DMA read always accept read completion.
  dma0_fifocnt <= (dma0_fifocnt'left => '1', others => '0');

  --
  -- write request from backend (or loopback request from hib in test mode)
  -- to hib (DMA write).
  --
  process(test_mode, backend_we, hib_we0)
  begin
    txwr <= '0';
    if (test_mode = "00") then
      txwr <= backend_we;
    elsif (test_mode = "01") then
      txwr <= hib_we0;
    end if;
  end process;

  process(test_mode, backend_data, hib_data0)
  begin
    if (test_mode = "00") then
      txdata <= backend_data;
    else
      txdata <= hib_data0;
    end if;
  end process;

  txrst <= rst or reset_dma;
  txrd  <= '1' when ms_write = '1' and ms_wrchannel(1) = '1' and test_mode /= "10" else '0'; -- !!

  process(test_mode, txusedw, txfull)
    variable usedbyte : std_logic_vector(txusedw'left+6 downto 0);
    variable infinite : std_logic_vector(dma1_fifocnt'range);
  begin

    infinite := (dma1_fifocnt'left => '1', others => '0');
--    infinite := '1' & conv_std_logic_vector(0, dma1_fifocnt'length - 1);

    if (test_mode = "10") then
      -- DMAW does not block if test_mode = "10" (measurement of DMA write performance).
      dma1_fifocnt <= infinite;
    else
      case NLANE is
        when 1 =>
          usedbyte := "0000" & txfull & txusedw & "0";  -- 2byte per word
        when 4 =>
          usedbyte := "00" & txfull & txusedw & "000";  -- 8byte per word
        when 8 =>
          usedbyte := '0' & txfull & txusedw & "0000";  -- 16byte per word
        when others =>
      end case;

      if (unsigned(usedbyte) > unsigned(infinite)) then
        dma1_fifocnt <= infinite;
      else
        dma1_fifocnt <= usedbyte(dma1_fifocnt'range);
      end if;

    end if;
  end process;

  txbuf0 : scfifo
    port map (
      clock    => clk,
      sclr     => txrst,

      wrreq    => txwr,
      data     => txdata,

      rdreq    => txrd,
      q        => ms_wrdata0,
      full     => txfull,
      empty    => txempty,
      usedw    => txusedw
    );

  -- for debugging purpose.
  -- detect duplicated read.
  process (clk)
  begin
    if (rising_edge(clk)) then
      txrd1 <= txrd;
      if (txrd1 = '1') then
        ms_wrdata1 <= ms_wrdata0;
        ms_wrdata2 <= unsigned(ms_wrdata1)+1;
      end if;

      dupr <= '0';
      if (ms_wrdata1 /= ms_wrdata2) then
        dupr <= '1';
      end if;
    end if;
  end process;

  -- detect duplicated write.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (txwr = '1') then
        txdata1 <= txdata;
        txdata2 <= unsigned(txdata1)+1;
      end if;
      dupw <= '0';
      if (txdata1 /= txdata2) then
        dupw <= '1';
      end if;
    end if;
  end process;

  
  process (txempty, txrd)
  begin
    txunderflow <= '0';
    if (txempty = '1' and txrd = '1') then
      txunderflow <= '1';
    end if;
  end process;

  process (txfull, txwr)
  begin
    txoverflow <= '0';
    if (txfull = '1' and txwr = '1') then
      txoverflow <= '1';
    end if;
  end process;

  process(slv_bar, slv_write)
  begin
    pio_wrin <= '0';
    if (slv_bar(2) = '1' and slv_write = '1') then
      pio_wrin <= '1';
    end if;
  end process;

  process(slv_addr, slv_bytevalid, slv_dataout)
  begin
    case NLANE is
      when 1 =>
        pio_addr <= slv_addr(PIOWBUF_DEPTH-1+1 downto 0+1);
        pio_be <= slv_bytevalid;
        pio_datain <= slv_dataout;

      when 4 =>
        pio_addr <= slv_addr(PIOWBUF_DEPTH-1+3 downto 0+3);
        pio_be <= slv_bytevalid;
        pio_datain <= slv_dataout;

      when 8 =>
        pio_addr <= slv_addr(PIOWBUF_DEPTH-1+4 downto 0+4);
        if (slv_addr(3) = '0') then
          -- lower 64-bit or full 128-bit.
          pio_be <= slv_bytevalid;
          pio_datain <= slv_dataout;
        else
          -- higher 64-bit.
          pio_be <= slv_bytevalid(NLANE-1 downto 0) & conv_std_logic_vector(0, NLANE);
          pio_datain <= slv_dataout(NLANE*8-1 downto 0) & conv_std_logic_vector(0, NLANE*8);
        end if;

      when others =>

    end case;
  end process;

  piowbuf0 : piowbuf
    port map (
      clk        => clk,
      rst        => rst,

      swap       => pio_swap,
      wrcnt      => pio_wrcnt,
      wrin       => pio_wrin,
      bein       => pio_be,
      addrin     => pio_addr,
      datain     => pio_datain,

      wrlock0out => pio_wrlock0,
      wrout      => pio_write,
      dataout    => pio_wrdata
    );

  asmictl0 : asmictl
    port map (
      clk     => clk,
      rst     => rst or reset_backend0,
      wr      => asmi_write,
      cmd     => asmi_cmd,
      addr    => asmi_addr,
      datain  => asmi_datain,
      dataout => asmi_dataout
    );
  
end default;

-- 
-- Double buffer for PIO write
-- 
-- Receive data written to the host's memory space mapped to BAR2,
-- and pass them to the backend logic.
--
-- Note that data may not arrive in address-ascending order, if CPU's
-- MTRR is set to write-combined mode for BAR2. Therefore, a buffer
-- should not be read out during a write operation is in progress.
--
-- In order to perform read and write operation simultaneously, this
-- module internally uses two SRAM buffers, buf0 and buf1. Data can be
-- read from one buffer, while another is being written (see figure
-- below. transfer A and B can be perfromed in parallel).  These two
-- buffers are exclusively write locked. The size of written data is
-- noticed by the host via pio_wrcnt register.
--
--                   A
-- host       buf0 ----> backend
--      ----> buf1
--        B
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;

entity piowbuf is
  generic (
    DEPTH   : integer;
    WIDTH   : integer
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;

    swap       : in  std_logic;
    wrcnt      : in  std_logic_vector(DEPTH-1 downto 0);

    wrin       : in  std_logic;
    bein       : in  std_logic_vector(WIDTH/8-1 downto 0);
    addrin     : in  std_logic_vector(DEPTH-1 downto 0);  -- WIDTH-bit word address
    datain     : in  std_logic_vector(WIDTH-1 downto 0);

    wrlock0out : out std_logic;
    wrout      : out std_logic;
    dataout    : out std_logic_vector(WIDTH-1 downto 0)
  );
end piowbuf;

architecture default of piowbuf is

  component altsyncram
    generic (
      RAM_BLOCK_TYPE  : string  := "AUTO";
      OPERATION_MODE  : string  := "DUAL_PORT";
      WIDTH_A         : natural := WIDTH;
      WIDTHAD_A       : natural := DEPTH;
      NUMWORDS_A      : natural := 2**DEPTH;
      WIDTH_B         : natural := WIDTH;
      WIDTHAD_B       : natural := DEPTH;
      NUMWORDS_B      : natural := 2**DEPTH;
      LPM_TYPE        : string  := "ALTSYNCRAM";
      WIDTH_BYTEENA_A : natural := WIDTH/8;
      BYTE_SIZE       : natural := 8;
      ADDRESS_REG_B   : string  := "CLOCK1";
      OUTDATA_REG_B   : string := "UNREGISTERED"
      );
    port (
      clock0    : in std_logic;
      wren_a    : in std_logic;
      byteena_a : in std_logic_vector(WIDTH_BYTEENA_A-1 downto 0);
      address_a : in std_logic_vector(WIDTHAD_A-1 downto 0);
      data_a    : in std_logic_vector(WIDTH_B-1 downto 0);

      clock1    : in std_logic;
      address_b : in std_logic_vector(WIDTHAD_B-1 downto 0);
      q_b       : out std_logic_vector(WIDTH_A-1 downto 0)
      );
  end component;

  signal wrlock0  : std_logic;
  signal wrlock1  : std_logic;
  signal wrin0    : std_logic;
  signal wrin1    : std_logic;
  signal rcnt0    : std_logic_vector(DEPTH-1 downto 0);
  signal rcnt1    : std_logic_vector(DEPTH-1 downto 0);
  signal rdaddr0  : std_logic_vector(DEPTH-1 downto 0);
  signal rdaddr1  : std_logic_vector(DEPTH-1 downto 0);
  signal rddata0  : std_logic_vector(WIDTH-1 downto 0);
  signal rddata1  : std_logic_vector(WIDTH-1 downto 0);

begin

  wrlock0out <= wrlock0;
  wrlock1    <= not wrlock0;  -- just for readability.

  process(clk, rst)
  begin
    if (rst = '1') then
      wrlock0 <= '0';
      rcnt0 <= (others => '0');
      rcnt1 <= (others => '0');
    elsif (rising_edge(clk)) then
      -- toggle write lock.
      if (swap = '1') then
        wrlock0 <= not wrlock0;
      end if;

      -- handle read address counters.
      wrout <= '0';
      if (swap = '1' and wrlock0 = '0') then
        rcnt0 <= wrcnt;
        rdaddr0 <= (others => '0');
      elsif (rcnt0 /= conv_std_logic_vector(0, rcnt0'length)) then
        wrout <= '1';
        rcnt0 <= unsigned(rcnt0) - 1;
        rdaddr0 <= unsigned(rdaddr0) + 1;
      end if;

      if (swap = '1' and wrlock1 = '0') then
        rcnt1 <= wrcnt;
        rdaddr1 <= (others => '0');
      elsif (rcnt1 /= conv_std_logic_vector(0, rcnt1'length)) then
        wrout <= '1';
        rcnt1 <= unsigned(rcnt1) - 1;
        rdaddr1 <= unsigned(rdaddr1) + 1;
      end if;

    end if;
  end process;

  process(wrlock0, rddata0, rddata1)
  begin
    if (wrlock0 = '1') then
      dataout <= rddata0;
    else
      dataout <= rddata1;
    end if;
  end process;

  wrin0 <= wrin and not wrlock0;
  wrin1 <= wrin and not wrlock1;

  buf0 : altsyncram
    port map (
      clock0    => clk,
      wren_a    => wrin0,
      byteena_a => bein,
      address_a => addrin,
      data_a    => datain,

      clock1    => clk,
      address_b => rdaddr0,
      q_b       => rddata0
    );

  buf1 : altsyncram
    port map (
      clock0    => clk,
      wren_a    => wrin1,
      byteena_a => bein,
      address_a => addrin,
      data_a    => datain,

      clock1    => clk,
      address_b => rdaddr1,
      q_b       => rddata1
    );

end default;
  
-- 
-- ASMI : parallel interface for access to EPCS.
-- 
-- ASMI Control Register (138h)
-- 
-- (31:8)  command or address
--    command
--      fffffd  read silicon id
--      fffffe  write bytes
--      ffffff  erase bulk
--    address
--      '0' & addr(22:0)  start writing to addr in serial ROM device memory space
-- 
-- (7:0)   data to be written
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity asmictl is
  generic (
    DEVICE   : string;
    USE_ASMI : integer
  );
  port  (
    clk      : in  std_logic;
    rst      : in  std_logic;

    wr       : in  std_logic;
    cmd      : in  std_logic_vector(23 downto 0);
    addr     : in  std_logic_vector(23 downto 0);
    datain   : in  std_logic_vector(7 downto 0);
    dataout  : out std_logic_vector(31 downto 0)
  );
end asmictl;

architecture default of asmictl is

  component asmi_agx
    port ( 
      clkin         : in   std_logic;

      -- for write bytes
      wren          : in   std_logic := '0';
      write         : in   std_logic := '0';
      shift_bytes   : in   std_logic := '0';
      addr          : in   std_logic_vector (23 downto 0);
      datain        : in   std_logic_vector (7 downto 0);

      -- for erase bulk
      bulk_erase    : in   std_logic := '0';

      -- for read silicon id
      read_sid      : in   std_logic := '0';
      epcs_id       : out  std_logic_vector (7 downto 0);

      -- status outputs
      busy          : out  std_logic;
      illegal_erase : out  std_logic;
      illegal_write : out  std_logic;

      -- unused signals
      data_valid    : out  std_logic;
      dataout       : out  std_logic_vector (7 downto 0);
      rden          : in   std_logic;
      read          : in   std_logic
      ); 
  end component;

  component asmi_s2gx
    port ( 
      clkin         : in   std_logic;

      -- for write bytes
      wren          : in   std_logic := '0';
      write         : in   std_logic := '0';
      shift_bytes   : in   std_logic := '0';
      addr          : in   std_logic_vector (23 downto 0);
      datain        : in   std_logic_vector (7 downto 0);

      -- for erase bulk
      bulk_erase    : in   std_logic := '0';

      -- for read silicon id
      read_sid      : in   std_logic := '0';
      epcs_id       : out  std_logic_vector (7 downto 0);

      -- status outputs
      busy          : out  std_logic;
      illegal_erase : out  std_logic;
      illegal_write : out  std_logic;

      -- unused signals
      data_valid    : out  std_logic;
      dataout       : out  std_logic_vector (7 downto 0);
      rden          : in   std_logic;
      read          : in   std_logic
      ); 
  end component;

  component lpm_counter
    generic (
      LPM_WIDTH  : INTEGER := 2
      );
    port (
      clock    : in std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0)
      );
  end component;

  component dcfifo
    generic (
      LPM_NUMWORDS            : natural := 256;
      LPM_WIDTHU              : natural := 8;
      LPM_WIDTH               : natural := 8;
      LPM_SHOWAHEAD           : string  := "OFF";
      ADD_RAM_OUTPUT_REGISTER : string  := "ON";
      OVERFLOW_CHECKING       : string  := "OFF";
      UNDERFLOW_CHECKING      : string  := "ON";
      USE_EAB                 : string  := "ON";
      RDSYNC_DELAYPIPE        : natural := 4;
      WRSYNC_DELAYPIPE        : natural := 4;
      LPM_TYPE                : string  := "dcfifo";
      INTENDED_DEVICE_FAMILY  : string  := "Arria GX"
    );
    port (
      aclr     : in std_logic;
      data     : in std_logic_vector(LPM_WIDTH-1 downto 0);
      rdclk    : in std_logic;
      rdreq    : in std_logic;
      wrclk    : in std_logic;
      wrreq    : in std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
      rdempty  : out std_logic 
    );
  end component;

  constant NCLK_DIV      : integer := 3;
  constant ASMI_CMD_RSID : std_logic_vector(23 downto 0) := x"fffffd";
  constant ASMI_CMD_WP   : std_logic_vector(23 downto 0) := x"fffffe";
  constant ASMI_CMD_EB   : std_logic_vector(23 downto 0) := x"ffffff";

  -- ASMI interface
  signal nclk          : std_logic;
  signal counter_q     : std_logic_vector(NCLK_DIV-1 downto 0);
  signal reset_asmi    : std_logic;
  signal reset_cnt     : std_logic_vector(4 downto 0);
  signal asmi_clk      : std_logic;
  signal asmi_w        : std_logic;
  signal asmi_we       : std_logic;
  signal asmi_wew1     : std_logic;
  signal asmi_wew2     : std_logic;
  signal asmi_wee      : std_logic;
  signal asmi_sb       : std_logic;
  signal asmi_addr     : std_logic_vector(23 downto 0);
  signal asmi_data     : std_logic_vector(7 downto 0);
  signal asmi_erase    : std_logic;
  signal asmi_rsid     : std_logic;
  signal asmi_epcsid   : std_logic_vector(7 downto 0);
  signal asmi_busy     : std_logic;
  signal asmi_busy_r   : std_logic;
  signal asmi_errw     : std_logic;
  signal asmi_errw_r   : std_logic;
  signal asmi_erre     : std_logic;
  signal asmi_erre_r   : std_logic;
  signal af_wrreq      : std_logic;
  signal af_data       : std_logic_vector(7 downto 0);
  signal af_rdempty    : std_logic;
  signal af_rdreq      : std_logic;

begin

  -- for debugging
--  led <= asmi_errw_r & asmi_erre_r & asmi_busy;
--  usr_led_r <= "011111" & not (asmi_errw_r or asmi_erre_r) & not asmi_busy;  -- negative logic

  -- ASMI control register read access:
  dataout <= conv_std_logic_vector(0, 16)
             & asmi_epcsid
             & conv_std_logic_vector(0, 6) & (asmi_erre_r or asmi_errw_r) & asmi_busy_r;
  
  -- interface to EPCS.
  genagx : if (DEVICE = "Arria GX" and USE_ASMI = 1) generate
    asmi_controller : asmi_agx
      port map ( 
        clkin         => asmi_clk,

        -- for write bytes
        write         => asmi_w,
        wren          => asmi_we,
        shift_bytes   => asmi_sb,
        addr          => asmi_addr,
        datain        => asmi_data,

        -- for erase bulk
        bulk_erase    => asmi_erase,

        -- for read silicon id
        read_sid      => asmi_rsid,
        epcs_id       => asmi_epcsid,

        -- status outputs
        busy          => asmi_busy,
        illegal_write => asmi_errw,
        illegal_erase => asmi_erre,

        -- unused signals
        data_valid    => open,
        dataout       => open,
        rden          => '0',
        read          => '0'
      ); 
  end generate genagx;

  gens2gx : if (DEVICE = "Stratix II GX" and USE_ASMI = 1) generate
    asmi_controller : asmi_s2gx
      port map ( 
        clkin         => asmi_clk,

        -- for write bytes
        write         => asmi_w,
        wren          => asmi_we,
        shift_bytes   => asmi_sb,
        addr          => asmi_addr,
        datain        => asmi_data,

        -- for erase bulk
        bulk_erase    => asmi_erase,

        -- for read silicon id
        read_sid      => asmi_rsid,
        epcs_id       => asmi_epcsid,

        -- status outputs
        busy          => asmi_busy,
        illegal_write => asmi_errw,
        illegal_erase => asmi_erre,

        -- unused signals
        data_valid    => open,
        dataout       => open,
        rden          => '0',
        read          => '0'
      ); 
  end generate gens2gx;

  -- shift phase by 180deg to avoid changing inputs
  -- at the rising edge of asmi_clk.
  nclk <= not clk;

  clk_counter : lpm_counter
    generic map (
      LPM_WIDTH => NCLK_DIV
    )
    port map (
      clock    => nclk,
      q        => counter_q
    );
  asmi_clk <= counter_q(NCLK_DIV-1);   -- asmi_clk = clk/8 = 125/8 MHz = 16MHz < 20MHz

  -- generate long enough reset signal.
  process (clk, rst)
  begin
    if rst = '1' then
      reset_asmi <= '1';
      reset_cnt <= (others => '1');
    elsif (rising_edge(clk)) then
      if (reset_cnt /= conv_std_logic_vector(0, reset_cnt'length)) then
        reset_asmi <= '1';
        reset_cnt <= unsigned(reset_cnt) - 1;
      else
        reset_asmi <= '0';
      end if;
    end if;
  end process; 

  asmi_we <= asmi_wee or asmi_wew1 or asmi_wew2;
  asmi_busy_r <= asmi_busy or asmi_we;

  --
  -- read silicon id
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      asmi_rsid <= '0';
    elsif (rising_edge(clk)) then
      if (wr = '1' and cmd = ASMI_CMD_RSID) then
        asmi_rsid <= '1';
      elsif asmi_busy = '1' then
        asmi_rsid <= '0';
      elsif reset_asmi = '1' then
        asmi_rsid <= '0';
      end if;
    end if;
  end process; 

  --
  -- erase bulk
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      asmi_erase <= '0';
      asmi_wee <= '0';
    elsif (rising_edge(clk)) then
      if (wr = '1' and cmd = ASMI_CMD_EB) then
        asmi_erase <= '1';
        asmi_wee <= '1';
      elsif asmi_busy = '1' then
        asmi_erase <= '0';
        asmi_wee <= '0';
      elsif reset_asmi = '1' then
        asmi_erase <= '0';
        asmi_wee <= '0';
      end if;
    end if;
  end process; 

  --
  -- write a byte to the megafunction internal buffer (write page stage1)
  --

  process (clk, rst)
  begin
    if (rst = '1') then
      af_wrreq <= '0';
      af_data <= conv_std_logic_vector(0, 8);
    elsif (rising_edge(clk)) then
      if (wr = '1' and cmd = ASMI_CMD_WP) then
        af_wrreq <= '1';
        af_data(7) <= datain(0);  -- reverse bit order
        af_data(6) <= datain(1);
        af_data(5) <= datain(2);
        af_data(4) <= datain(3);
        af_data(3) <= datain(4);
        af_data(2) <= datain(5);
        af_data(1) <= datain(6);
        af_data(0) <= datain(7);
      else
        af_wrreq <= '0';
      end if;
    end if;
  end process; 

  --
  -- dual clock fifo
  -- write in from PCIe, read out to ASMI.
  --
  -- asmi_clk     ~_~_~_~_~_~_~_~_~_~_~_
  -- asmi_fifo    --<><><><>-------
  -- rdempty        ~~________~~~~~~~~
  -- rdreq          __~~~~~~~~________
  -- asmi_sb          __~~~~~~~~______
  -- asmi_q           --<><><><>--------
  --
  
  dcfifo0 : dcfifo
    port map (
      aclr      => reset_asmi,
      rdempty   => af_rdempty,

      wrclk     => clk,
      wrreq     => af_wrreq,
      data      => af_data,

      rdclk     => asmi_clk,
      rdreq     => af_rdreq,
      q         => asmi_data
    );

  af_rdreq <= not af_rdempty;

  process (asmi_clk)
  begin
    if (rising_edge(asmi_clk)) then  
      asmi_sb   <= af_rdreq;
      asmi_wew1 <= af_rdreq;
    end if;
  end process;
  
  --
  -- flush megafunction internal buffer contents to the serial rom (write page stage2)
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      asmi_w <= '0';
      asmi_wew2 <= '0';
    elsif (rising_edge(clk)) then
      if (wr = '1' and cmd(cmd'left) = '0') then
        asmi_w <= '1';
        asmi_wew2 <= '1';
        asmi_addr <= addr;
      elsif asmi_busy = '1' then
        asmi_w <= '0';
        asmi_wew2 <= '0';
      elsif reset_asmi = '1' then
        asmi_w <= '0';
        asmi_wew2 <= '0';
      end if;
    end if;
  end process; 

  --
  -- diagnostics
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      asmi_erre_r <= '0';
      asmi_errw_r <= '0';
    elsif (rising_edge(clk)) then
      if reset_asmi = '1' then
        asmi_erre_r <= '0';
        asmi_errw_r <= '0';
      else
        if asmi_erre = '1' then
          asmi_erre_r <= '1';
        end if;
        if asmi_errw = '1' then
          asmi_errw_r <= '1';
        end if;
      end if;
    end if;
  end process; 

end default;
