library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library lpm;
use lpm.lpm_components.all;
use work.gpciepkg.all;

entity hib is
  generic (
    DEVICE              : string  := "Arria GX";
    NLANE               : integer;
    PIOWBUF_DEPTH       : integer := 8;  -- 256-word x2 double buffer for PIO write.
    TXBUF_DEPTH         : integer := 10; -- 1k-word buffer for DMA write.
    USE_ASMI            : integer := 1;  -- implement an interface to Active Serial Configuration ROM.
    USE_CLK32           : integer := 1;  -- use clock input pin 'clk32'.
    NDMACH              : integer := 2;  -- should be 2. DMA0 for DMA read, DMA1 for DMA write.
    SIMULATION          : integer := 0
  );
  port (
    phy_linkup           : out std_logic;
    dl_linkup            : out std_logic;
    clk32                : in  std_logic;

    -- PCIe signals
    clk100_ext           : in  std_logic;
    mperst               : in  std_logic;
    rx_in                : in  std_logic_vector(NLANE-1 downto 0);
    tx_out               : out std_logic_vector(NLANE-1 downto 0);
    wake                 : out std_logic;

    -- global lines
    clk_out              : out std_logic;

    -- interface to the backend 
    hib_we               : out std_logic;
    hib_data             : out std_logic_vector(NLANE*16-1 downto 0);
    backend_we           : in  std_logic;
    backend_data         : in  std_logic_vector(NLANE*16-1 downto 0);
    reset_backend        : out std_logic;

    -- board information register initial value
    board_info           : in  std_logic_vector(31 downto 0)
  );
end hib;

architecture default of hib is

  component phy125
    generic (
      DEVICE       : string  := DEVICE;
      USE_CLK32    : integer := USE_CLK32;
      NLANE        : integer := NLANE;
      SIMULATION   : integer := SIMULATION
    );
    port (
      cal_blk_clk             : in    std_logic;
      clk125plllock           : out   std_logic;
      clk32                   : in    std_logic;
      clk100                  : in    std_logic;
      clk125out               : out   std_logic;
      rstn                    : in    std_logic;

      rx_in                   : in    std_logic_vector(NLANE-1 downto 0);
      tx_out                  : out   std_logic_vector(NLANE-1 downto 0);

      wake                    : out   std_logic;
      phystatus               : out   std_logic;
      powerdown               : in    std_logic_vector(1 downto 0);
      txdetectrx              : in    std_logic;

      txdata                  : in    std_logic_vector(NLANE*16-1 downto 0);
      txdatak                 : in    std_logic_vector(NLANE*2-1 downto 0);
      txelecidle              : in    std_logic_vector(NLANE-1 downto 0);
      txcompl                 : in    std_logic_vector(NLANE-1 downto 0);
      rxpolarity              : in    std_logic_vector(NLANE-1 downto 0);
      rxdata                  : out   std_logic_vector(NLANE*16-1 downto 0);
      rxdatak                 : out   std_logic_vector(NLANE*2-1 downto 0);
      rxvalid                 : out   std_logic_vector(NLANE-1 downto 0);
      rxelecidle              : out   std_logic_vector(NLANE-1 downto 0);
      rxstatus                : out   std_logic_vector(NLANE*3-1 downto 0)
    );
  end component;

  component gpcie
    generic (
      NLANE              : integer := NLANE;
      NDMACH             : integer := NDMACH;
      CFG_DEVICE_ID_INIT : std_logic_vector(15 downto 0) := x"0e70";
      SIMULATION         : integer := SIMULATION
    );
    port(
      phy_linkup      : out std_logic;
      dl_linkup       : out std_logic;

      clk             : in  std_logic;
      rstn            : in  std_logic;

      -- PIPE interface
      phystatus       : in  std_logic;
      powerdown       : out std_logic_vector(1 downto 0);
      txdetectrx      : out std_logic;
      txdata          : out std_logic_vector(NLANE*16-1 downto 0);
      txdatak         : out std_logic_vector(NLANE*2-1 downto 0);
      txelecidle      : out std_logic_vector(NLANE-1 downto 0);
      txcompl         : out std_logic_vector(NLANE-1 downto 0);
      rxpolarity      : out std_logic_vector(NLANE-1 downto 0);
      rxdata          : in  std_logic_vector(NLANE*16-1 downto 0);
      rxdatak         : in  std_logic_vector(NLANE*2-1 downto 0);
      rxvalid         : in  std_logic_vector(NLANE-1 downto 0);
      rxelecidle      : in  std_logic_vector(NLANE-1 downto 0);
      rxstatus        : in  std_logic_vector(NLANE*3-1 downto 0);

      -- core interface
      slv_read        : out std_logic;
      slv_write       : out std_logic;
      slv_bar         : out std_logic_vector(6 downto 0);
      slv_addr        : out std_logic_vector(63 downto 0);
      slv_bytevalid   : out std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout     : out std_logic_vector(NLANE*16-1 downto 0);
      slv_datain      : in  std_logic_vector(NLANE*16-1 downto 0);

      slv_readreq     : out std_logic;
      slv_writereq    : out std_logic;
      slv_accept      : in  std_logic;
      slv_bytecount   : out std_logic_vector(11 downto 0);

      ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_write          : out std_logic;
      ms_wraddr         : out std_logic_vector(31 downto 0);
      ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);
      ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_read           : out std_logic;
      ms_rdaddr         : out std_logic_vector(31 downto 0);
      ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

      dma_control       : in  each7b(NDMACH-1 downto 0);
      dma_param         : in  each16b(NDMACH-1 downto 0);
      dma_status        : out each4b(NDMACH-1 downto 0);
      dma_fifocnt       : in  each13b(NDMACH-1 downto 0);

      dma_paddrlow_in   : in  each32b(NDMACH-1 downto 0);
      dma_paddrhigh_in  : in  each32b(NDMACH-1 downto 0);
      dma_laddr_in      : in  each32b(NDMACH-1 downto 0);
      dma_size_in       : in  each32b(NDMACH-1 downto 0);

      dma_paddrlow_out  : out each32b(NDMACH-1 downto 0);
      dma_paddrhigh_out : out each32b(NDMACH-1 downto 0);
      dma_laddr_out     : out each32b(NDMACH-1 downto 0);
      dma_size_out      : out each32b(NDMACH-1 downto 0)
    );
  end component;

  component hibctl 
    generic (
      DEVICE          : string  := DEVICE;
      USE_ASMI        : integer := USE_ASMI;
      PIOWBUF_DEPTH   : integer := PIOWBUF_DEPTH;
      TXBUF_DEPTH     : integer := TXBUF_DEPTH;
      NLANE           : integer := NLANE;
      NDMACH          : integer := NDMACH;
      SIMULATION      : integer := SIMULATION
    );
    port (
      clk                 : in  std_logic;
      pll_locked          : in  std_logic;
      mperst              : in  std_logic;
      rstn_out            : out std_logic;

      -- interdace to GPCIe
      slv_read            : in  std_logic;
      slv_write           : in  std_logic;
      slv_bar             : in  std_logic_vector(6 downto 0);
      slv_addr            : in  std_logic_vector(63 downto 0);
      slv_bytevalid       : in  std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout         : in  std_logic_vector(NLANE*16-1 downto 0);
      slv_datain          : out std_logic_vector(NLANE*16-1 downto 0);
      slv_readreq         : in  std_logic;
      slv_writereq        : in  std_logic;
      slv_accept          : out std_logic;
      slv_bytecount       : in  std_logic_vector(11 downto 0);

      ms_wrchannel        : in  std_logic_vector(NDMACH-1 downto 0);
      ms_write            : in  std_logic;
      ms_wraddr           : in  std_logic_vector(31 downto 0);
      ms_wrdata           : out std_logic_vector(NLANE*16-1 downto 0);
      ms_rdchannel        : in  std_logic_vector(NDMACH-1 downto 0);
      ms_read             : in  std_logic;
      ms_rdaddr           : in  std_logic_vector(31 downto 0);
      ms_rddata           : in  std_logic_vector(NLANE*16-1 downto 0);

      dma_control         : out each7b(NDMACH-1 downto 0);
      dma_param           : out each16b(NDMACH-1 downto 0);
      dma_status          : in  each4b(NDMACH-1 downto 0);
      dma_fifocnt         : out each13b(NDMACH-1 downto 0);

      dma_paddrlow_in     : out each32b(NDMACH-1 downto 0);
      dma_paddrhigh_in    : out each32b(NDMACH-1 downto 0);
      dma_laddr_in        : out each32b(NDMACH-1 downto 0);
      dma_size_in         : out each32b(NDMACH-1 downto 0);

      dma_paddrlow_out    : in  each32b(NDMACH-1 downto 0);
      dma_paddrhigh_out   : in  each32b(NDMACH-1 downto 0);
      dma_laddr_out       : in  each32b(NDMACH-1 downto 0);
      dma_size_out        : in  each32b(NDMACH-1 downto 0);

      -- interface to the backend 
      hib_we              : out std_logic;
      hib_data            : out std_logic_vector(NLANE*16-1 downto 0);
      backend_we          : in std_logic;
      backend_data        : in std_logic_vector(NLANE*16-1 downto 0);
      reset_backend       : out std_logic;

      -- board information register initial value
      board_info          : in std_logic_vector(31 downto 0)
    );
  end component;

  signal clk125          : std_logic;
  signal clk125plllocked : std_logic;
  signal hib_rstn        : std_logic;

  -- PCIe signals
  signal phystatus       : std_logic;
  signal powerdown       : std_logic_vector(1 downto 0);
  signal txdetectrx      : std_logic;

  signal txdata          : std_logic_vector(NLANE*16-1 downto 0);
  signal txdatak         : std_logic_vector(NLANE*2-1 downto 0);
  signal txelecidle      : std_logic_vector(NLANE-1 downto 0);
  signal txcompl         : std_logic_vector(NLANE-1 downto 0);
  signal rxpolarity      : std_logic_vector(NLANE-1 downto 0);
  signal rxdata          : std_logic_vector(NLANE*16-1 downto 0);
  signal rxdatak         : std_logic_vector(NLANE*2-1 downto 0);
  signal rxvalid         : std_logic_vector(NLANE-1 downto 0);
  signal rxelecidle      : std_logic_vector(NLANE-1 downto 0);
  signal rxstatus        : std_logic_vector(NLANE*3-1 downto 0);

  signal slv_read        : std_logic;
  signal slv_write       : std_logic;
  signal slv_bar         : std_logic_vector(6 downto 0);
  signal slv_addr        : std_logic_vector(63 downto 0);
  signal slv_bytevalid   : std_logic_vector(NLANE*2-1 downto 0);
  signal slv_dataout     : std_logic_vector(NLANE*16-1 downto 0);
  signal slv_datain      : std_logic_vector(NLANE*16-1 downto 0);
  signal slv_readreq     : std_logic;
  signal slv_writereq    : std_logic;
  signal slv_accept      : std_logic;
  signal slv_bytecount   : std_logic_vector(11 downto 0);

  signal ms_write            : std_logic;
  signal ms_wraddr           : std_logic_vector(31 downto 0);
  signal ms_wrdata           : std_logic_vector(NLANE*16-1 downto 0);
  signal ms_read             : std_logic;
  signal ms_rdaddr           : std_logic_vector(31 downto 0);
  signal ms_rddata           : std_logic_vector(NLANE*16-1 downto 0);

  signal ms_wrchannel        : std_logic_vector(NDMACH-1 downto 0);
  signal ms_rdchannel        : std_logic_vector(NDMACH-1 downto 0);
  signal dma_control         : each7b(NDMACH-1 downto 0);
  signal dma_param           : each16b(NDMACH-1 downto 0);
  signal dma_status          : each4b(NDMACH-1 downto 0);
  signal dma_fifocnt         : each13b(NDMACH-1 downto 0);

  signal dma_paddrlow_in     : each32b(NDMACH-1 downto 0);
  signal dma_paddrhigh_in    : each32b(NDMACH-1 downto 0);
  signal dma_laddr_in        : each32b(NDMACH-1 downto 0);
  signal dma_size_in         : each32b(NDMACH-1 downto 0);

  signal dma_paddrlow_out    : each32b(NDMACH-1 downto 0);
  signal dma_paddrhigh_out   : each32b(NDMACH-1 downto 0);
  signal dma_laddr_out       : each32b(NDMACH-1 downto 0);
  signal dma_size_out        : each32b(NDMACH-1 downto 0);

begin
  --
  -- interface to embedded transceiver quad-block(s).
  --
  clk_out <= clk125;

  phy : phy125
    port map (
      cal_blk_clk       => clk32,    -- should be common to all cal_blk_clk in the design.
      clk125plllock     => clk125plllocked, -- clk valid only when this is high.
      clk32             => clk32,
      clk100            => clk100_ext,
      clk125out         => clk125,
      rstn              => mperst,

      rx_in             => rx_in,
      tx_out            => tx_out,

      wake              => wake,

      phystatus         => phystatus,
      powerdown         => powerdown,
      txdetectrx        => txdetectrx,

      txdata            => txdata,
      txdatak           => txdatak,
      txelecidle        => txelecidle,
      txcompl           => txcompl,
      rxpolarity        => rxpolarity,
      rxdata            => rxdata,
      rxdatak           => rxdatak,
      rxvalid           => rxvalid,
      rxelecidle        => rxelecidle,
      rxstatus          => rxstatus
    );

  gpcie_inst : gpcie
    port map (
      dl_linkup       => dl_linkup,
      phy_linkup      => phy_linkup,

      clk             => clk125,
      rstn            => hib_rstn,

      -- PIPE interface
      phystatus       => phystatus,
      powerdown       => powerdown,
      txdetectrx      => txdetectrx,

      txdata          => txdata,
      txdatak         => txdatak,
      txelecidle      => txelecidle,
      txcompl         => txcompl,
      rxpolarity      => rxpolarity,
      rxdata          => rxdata,
      rxdatak         => rxdatak,
      rxvalid         => rxvalid,
      rxelecidle      => rxelecidle,
      rxstatus        => rxstatus,

      -- core interface
      slv_read        => slv_read,
      slv_write       => slv_write,
      slv_bar         => slv_bar,
      slv_addr        => slv_addr,
      slv_bytevalid   => slv_bytevalid,
      slv_dataout     => slv_dataout,
      slv_datain      => slv_datain,

      slv_readreq     => slv_readreq,
      slv_writereq    => slv_writereq,
      slv_accept      => slv_accept,
      slv_bytecount   => slv_bytecount,

      ms_wrchannel      => ms_wrchannel,
      ms_write          => ms_write,
      ms_wraddr         => ms_wraddr,
      ms_wrdata         => ms_wrdata,
      ms_rdchannel      => ms_rdchannel,
      ms_read           => ms_read,
      ms_rdaddr         => ms_rdaddr,
      ms_rddata         => ms_rddata,

      dma_control       => dma_control,
      dma_param         => dma_param,
      dma_status        => dma_status,
      dma_fifocnt       => dma_fifocnt,

      dma_paddrlow_in   => dma_paddrlow_in,
      dma_paddrhigh_in  => dma_paddrhigh_in,
      dma_laddr_in      => dma_laddr_in,
      dma_size_in       => dma_size_in,

      dma_paddrlow_out  => dma_paddrlow_out,
      dma_paddrhigh_out => dma_paddrhigh_out,
      dma_laddr_out     => dma_laddr_out,
      dma_size_out      => dma_size_out
    );

  hibctl_instance : hibctl
    port map (
      clk             => clk125,
      pll_locked      => clk125plllocked,
      mperst          => mperst,             -- asynchronous power on reset input (active low)
      rstn_out        => hib_rstn,

      -- interface between hibctl & PCIe core
      slv_read        => slv_read,
      slv_write       => slv_write,
      slv_bar         => slv_bar,
      slv_addr        => slv_addr,
      slv_bytevalid   => slv_bytevalid,
      slv_dataout     => slv_dataout,
      slv_datain      => slv_datain,

      slv_readreq     => slv_readreq,
      slv_writereq    => slv_writereq,
      slv_accept      => slv_accept,
      slv_bytecount   => slv_bytecount,

      ms_wrchannel      => ms_wrchannel,
      ms_write          => ms_write,
      ms_wraddr         => ms_wraddr,
      ms_wrdata         => ms_wrdata,
      ms_rdchannel      => ms_rdchannel,
      ms_read           => ms_read,
      ms_rdaddr         => ms_rdaddr,
      ms_rddata         => ms_rddata,

      dma_control       => dma_control,
      dma_param         => dma_param,
      dma_status        => dma_status,
      dma_fifocnt       => dma_fifocnt,

      dma_paddrlow_in   => dma_paddrlow_in,
      dma_paddrhigh_in  => dma_paddrhigh_in,
      dma_laddr_in      => dma_laddr_in,
      dma_size_in       => dma_size_in,

      dma_paddrlow_out  => dma_paddrlow_out,
      dma_paddrhigh_out => dma_paddrhigh_out,
      dma_laddr_out     => dma_laddr_out,
      dma_size_out      => dma_size_out,

      -- backend interface
      hib_we          => hib_we,
      hib_data        => hib_data,
      backend_we      => backend_we,
      backend_data    => backend_data,
      reset_backend   => reset_backend,

      -- board information register initial value
      board_info      => board_info
    );

end default;
