
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

package pcrcpkg is

  -- parameters determined by PCIe spec.
  constant CRC_WIDTH    : integer := 32;
  constant NELEM        : integer := CRC_WIDTH;
  constant NMATRICE     : integer := OFFSET_WIDTH;

  -- parameters which may be adjusted to achieve
  -- better trade off between performance & area.
  constant NPIPEMAX       : integer := 4;  -- set to 4 or larger for x8, 2 or larger for x4.
                                           -- both ArriaGX and StratixII GX can handle 32bit/pipe.
  constant USE_G          : integer := 0;  -- 0 or 1
  constant USE_H11        : integer := 1;  -- 0 or 1
  constant MULT_PER_STAGE : integer := 1;  -- fmax on StratixII 1:300MHz 2:180MHz 3:130MHz
  constant ACCUM_STAGE    : integer := 7;
  constant FINAL_STAGE    : integer := ACCUM_STAGE*(USE_G+1);
  constant NSTAGE         : integer := FINAL_STAGE+2;  -- prestage, stage0, 1, ... ,FINAL_STAGE
  --                                                             
  -- USE_G  USE_H11  MULT_PER_STAGE  ACCUM_STAGE  FINAL_STAGE  NSTAGE   DELAY(of calc_crc)
  --  1    0               1           13           26         28       29
  --  1    0               2            7           14         16       17
  --  1    0               3            5           10         12       13
  --  1    0               4            4            8         10       11
  --  1    1               1            7           14         16       17
  --  1    1               2            4            8         10       11
  --  1    1               3            3            6          8        9
  --  0    0               1           13           13         15       16
  --  0    0               2            7            7          9       10
  --  0    0               3            5            5          7        8
  --  0    0               4            4            4          6        7
  --  0    1               1            7            7          9       10 <- ArriaGX default setting.
  --  0    1               2            4            4          6        7 <- StratixII default setting.
  --  0    1               3            3            3          5        6 <- probably OK with StratixII GX.

  type arrayNPIPExOFFW  is array (NPIPEMAX-1 downto 0) of std_logic_vector(OFFSET_WIDTH-1 downto 0);
  type arrayNPIPExCRCW  is array (NPIPEMAX-1 downto 0) of std_logic_vector(CRC_WIDTH-1 downto 0);
  type arrayNSTGxCRCW   is array (NSTAGE-1 downto 0) of std_logic_vector(CRC_WIDTH-1 downto 0);
  type arrayNSTGxOFFW   is array (NSTAGE-1 downto 0) of std_logic_vector(OFFSET_WIDTH-1 downto 0);
  type matrix           is array (0 to NELEM-1, NELEM-1 downto 0) of std_logic;
  type matrice          is array (0 to NMATRICE-1) of matrix;
  --
  --  powers of H-matrix are listed below.
  --  H-matrix is defined as a 32x32 matrix that converts CRC32(data) to CRC32(data & '0').
  --  you can obtain CRC32 of data trailed by n zeros (CRC32(data & conv_std_logic_vector(0, n)))
  --  by multiplying H to CRC32(data) for n times.
  --
  --  H^2^(i+3) is stored to H_MATRIX(i).
  --  for example, H_MATRIX(0)  := H^8,
  --               H_MATRIX(1)  := H^16,
  --               H_MATRIX(2)  := H^32,
  --               ...
  --               H_MATRIX(12) := H^32k, 
  constant H_MATRIX : matrice :=
    (
      -- H8
      (
       x"41000000",
       x"c3000000",
       x"c7000000",
       x"8e000000",
       x"5d000000",
       x"fb000000",
       x"f6000000",
       x"ad000000",
       x"1b000001",
       x"36000002",
       x"2d000004",
       x"1b000008",
       x"77000010",
       x"ee000020",
       x"dc000040",
       x"b8000080",
       x"31000100",
       x"62000200",
       x"c4000400",
       x"88000800",
       x"10001000",
       x"20002000",
       x"01004000",
       x"43008000",
       x"86010000",
       x"0c020000",
       x"59040000",
       x"b2080000",
       x"64100000",
       x"c8200000",
       x"90400000",
       x"20800000"
      ),

      -- H16
      (
       x"16410000",
       x"3ac30000",
       x"63c70000",
       x"c78e0000",
       x"995d0000",
       x"24fb0000",
       x"49f60000",
       x"85ad0000",
       x"1d1b0000",
       x"3a360000",
       x"622d0000",
       x"d21b0000",
       x"b2770000",
       x"64ee0000",
       x"c9dc0000",
       x"93b80000",
       x"31310001",
       x"62620002",
       x"c4c40004",
       x"89880008",
       x"13100010",
       x"26200020",
       x"5a010040",
       x"a2430080",
       x"44860100",
       x"890c0200",
       x"04590400",
       x"08b20800",
       x"11641000",
       x"22c82000",
       x"45904000",
       x"8b208000"
      ),

      -- H32
      (
       x"f7011641",
       x"19033ac3",
       x"c50763c7",
       x"8a0ec78e",
       x"e31c995d",
       x"313824fb",
       x"627049f6",
       x"33e185ad",
       x"90c21d1b",
       x"21843a36",
       x"b409622d",
       x"9f13d21b",
       x"c926b277",
       x"924d64ee",
       x"249ac9dc",
       x"493593b8",
       x"656a3131",
       x"cad46262",
       x"95a8c4c4",
       x"2b518988",
       x"56a31310",
       x"ad462620",
       x"ad8d5a01",
       x"ac1ba243",
       x"58374486",
       x"b06e890c",
       x"97dc0459",
       x"2fb808b2",
       x"5f701164",
       x"bee022c8",
       x"7dc04590",
       x"fb808b20"
      ),

      -- H64
      (
       x"b4e5b025",
       x"dd2ed06e",
       x"0eb810f9",
       x"1d7021f3",
       x"8e05f3c2",
       x"a8ee57a0",
       x"51dcaf40",
       x"175ceea5",
       x"9a5c6d6f",
       x"34b8dadf",
       x"dd94059b",
       x"0fcdbb12",
       x"ab7ec600",
       x"56fd8c01",
       x"adfb1803",
       x"5bf63006",
       x"0309d029",
       x"0613a052",
       x"0c2740a5",
       x"184e814b",
       x"309d0296",
       x"613a052c",
       x"7691ba7c",
       x"59c6c4dc",
       x"b38d89b9",
       x"671b1372",
       x"7ad396c0",
       x"f5a72d81",
       x"eb4e5b02",
       x"d69cb604",
       x"ad396c09",
       x"5a72d812"
      ),

      -- H128
      (
       x"e8f6c5af",
       x"391b4ef0",
       x"9ac0584f",
       x"3580b09e",
       x"83f7a492",
       x"ef198c8a",
       x"de331914",
       x"5490f786",
       x"41d72aa2",
       x"83ae5545",
       x"efaa6f25",
       x"37a21be4",
       x"87b2f267",
       x"0f65e4ce",
       x"1ecbc99d",
       x"3d97933a",
       x"93d9e3da",
       x"27b3c7b4",
       x"4f678f69",
       x"9ecf1ed2",
       x"3d9e3da4",
       x"7b3c7b49",
       x"1e8e333c",
       x"d5eaa3d7",
       x"abd547ae",
       x"57aa8f5c",
       x"47a3db16",
       x"8f47b62d",
       x"1e8f6c5a",
       x"3d1ed8b5",
       x"7a3db16b",
       x"f47b62d7"
      ),

      -- H256
      (
       x"9108245d",
       x"b3186ce7",
       x"f738fd93",
       x"ee71fb26",
       x"4debd211",
       x"0adf807f",
       x"15bf00fe",
       x"ba7625a1",
       x"e5e46f1e",
       x"cbc8de3d",
       x"06999827",
       x"9c3b1412",
       x"a97e0c78",
       x"52fc18f0",
       x"a5f831e1",
       x"4bf063c2",
       x"06e8e3d8",
       x"0dd1c7b1",
       x"1ba38f63",
       x"37471ec7",
       x"6e8e3d8f",
       x"dd1c7b1f",
       x"2b30d262",
       x"c7698099",
       x"8ed30133",
       x"1da60266",
       x"aa442091",
       x"54884122",
       x"a9108245",
       x"5221048b",
       x"a4420917",
       x"4884122e"
      ),

      -- H512
      (
       x"dc70ff4f",
       x"649101d0",
       x"1552fcee",
       x"2aa5f9dc",
       x"893b0cf7",
       x"ce06e6a0",
       x"9c0dcd40",
       x"e46b65ce",
       x"14a634d3",
       x"294c69a7",
       x"8ee82c00",
       x"c1a0a74f",
       x"5f31b1d0",
       x"be6363a0",
       x"7cc6c740",
       x"f98d8e81",
       x"2f6be24c",
       x"5ed7c499",
       x"bdaf8932",
       x"7b5f1264",
       x"f6be24c8",
       x"ed7c4991",
       x"06886c6c",
       x"d1602796",
       x"a2c04f2c",
       x"45809e59",
       x"5771c3fd",
       x"aee387fa",
       x"5dc70ff4",
       x"bb8e1fe9",
       x"771c3fd3",
       x"ee387fa7"
      ),

      -- H1k
      (
       x"2d9b71d5",
       x"76ad927f",
       x"c0c0552a",
       x"8180aa55",
       x"2e9a257e",
       x"70af3b29",
       x"e15e7653",
       x"ef279d73",
       x"f3d44b33",
       x"e7a89666",
       x"e2ca5d19",
       x"e80fcbe7",
       x"fd84e61b",
       x"fb09cc36",
       x"f613986d",
       x"ec2730db",
       x"f5d51063",
       x"ebaa20c7",
       x"d754418f",
       x"aea8831f",
       x"5d51063f",
       x"baa20c7f",
       x"58df692b",
       x"9c25a382",
       x"384b4704",
       x"70968e09",
       x"ccb66dc7",
       x"996cdb8e",
       x"32d9b71d",
       x"65b36e3a",
       x"cb66dc75",
       x"96cdb8ea"
      ),

      -- H2k
      (
       x"b7739aa3",
       x"d994afe5",
       x"045ac569",
       x"08b58ad2",
       x"a6188f07",
       x"fb4284ad",
       x"f685095a",
       x"5a798816",
       x"03808a8e",
       x"0701151d",
       x"b971b098",
       x"c590fb92",
       x"3c526d86",
       x"78a4db0d",
       x"f149b61a",
       x"e2936c34",
       x"725542ca",
       x"e4aa8595",
       x"c9550b2b",
       x"92aa1657",
       x"25542caf",
       x"4aa8595f",
       x"2223281d",
       x"f335ca99",
       x"e66b9532",
       x"ccd72a64",
       x"2eddce6a",
       x"5dbb9cd5",
       x"bb7739aa",
       x"76ee7354",
       x"eddce6a8",
       x"dbb9cd51"
      ),

      -- H4k
      (
       x"1aedbae1",
       x"2f36cf22",
       x"448024a4",
       x"89004948",
       x"08ed2871",
       x"0b37ea03",
       x"166fd407",
       x"363212ee",
       x"76899f3c",
       x"ed133e79",
       x"c0cbc613",
       x"9b7a36c7",
       x"2c19d76f",
       x"5833aedf",
       x"b0675dbf",
       x"60cebb7e",
       x"db70cc1d",
       x"b6e1983a",
       x"6dc33075",
       x"db8660ea",
       x"b70cc1d4",
       x"6e1983a8",
       x"c6debdb0",
       x"9750c181",
       x"2ea18302",
       x"5d430605",
       x"a06bb6eb",
       x"40d76dd7",
       x"81aedbae",
       x"035db75c",
       x"06bb6eb8",
       x"0d76dd70"
      ),

      -- H8k
      (
       x"9d928f1c",
       x"a6b79125",
       x"d0fdad57",
       x"a1fb5aae",
       x"de643a40",
       x"215afb9d",
       x"42b5f73a",
       x"18f96168",
       x"ac604dcc",
       x"58c09b98",
       x"2c13b82d",
       x"c5b5ff46",
       x"16f97190",
       x"2df2e321",
       x"5be5c643",
       x"b7cb8c87",
       x"f2059613",
       x"e40b2c26",
       x"c816584c",
       x"902cb098",
       x"20596130",
       x"40b2c260",
       x"1cf70bdc",
       x"a47c98a4",
       x"48f93148",
       x"91f26290",
       x"be764a3c",
       x"7cec9478",
       x"f9d928f1",
       x"f3b251e3",
       x"e764a3c7",
       x"cec9478e"
      ),

      -- H16k
      (
       x"d8f285c0",
       x"69178e41",
       x"0add9942",
       x"15bb3284",
       x"f384e0c9",
       x"3ffb4453",
       x"7ff688a6",
       x"271f948d",
       x"96cdacda",
       x"2d9b59b5",
       x"83c436aa",
       x"df7ae894",
       x"660754e8",
       x"cc0ea9d1",
       x"981d53a3",
       x"303aa747",
       x"b887cb4f",
       x"710f969e",
       x"e21f2d3d",
       x"c43e5a7b",
       x"887cb4f7",
       x"10f969ee",
       x"f900561d",
       x"2af229fa",
       x"55e453f5",
       x"abc8a7eb",
       x"8f63ca17",
       x"1ec7942e",
       x"3d8f285c",
       x"7b1e50b8",
       x"f63ca170",
       x"ec7942e0"
      ),

      -- H32k
      (
       x"7eaf4a9e",
       x"83f1dfa2",
       x"794cf5da",
       x"f299ebb4",
       x"9b9c9df7",
       x"49967170",
       x"932ce2e0",
       x"58f68f5f",
       x"cf425421",
       x"9e84a843",
       x"43a61a19",
       x"f9e37ead",
       x"8d69b7c5",
       x"1ad36f8b",
       x"35a6df17",
       x"6b4dbe2e",
       x"a83436c2",
       x"50686d85",
       x"a0d0db0a",
       x"41a1b615",
       x"83436c2a",
       x"0686d855",
       x"73a2fa34",
       x"99eabef6",
       x"33d57ded",
       x"67aafbda",
       x"b1fabd2a",
       x"63f57a54",
       x"c7eaf4a9",
       x"8fd5e953",
       x"1fabd2a7",
       x"3f57a54f"
      )
    );

  --
  --  powers of G-matrix are listed below.
  --  G-matrix is an inverse of H-matrix.
  --  you can obtain CRC32(data) from CRC32 of data trailed by n zeros
  --  (CRC32(data & conv_std_logic_vector(0, n))) by multiplying G to
  --  CRC32(data) for n times.
  --
  --  G^2^(i+3) is stored to G_MATRIX(i).
  --  for example, G_MATRIX(0)  := G^8,
  --               G_MATRIX(1)  := G^16,
  --               G_MATRIX(2)  := G^32,
  --               ...
  --               G_MATRIX(12) := G^32k, 
  constant G_MATRIX : matrice :=
    (
      -- G8
      (
       x"000001aa",
       x"000002ff",
       x"00000455",
       x"000008aa",
       x"000010fe",
       x"00002057",
       x"000040ae",
       x"000080f7",
       x"00010044",
       x"00020089",
       x"000400b9",
       x"000800d8",
       x"0010001a",
       x"00200035",
       x"0040006b",
       x"008000d7",
       x"01000005",
       x"0200000b",
       x"04000016",
       x"0800002c",
       x"10000059",
       x"200000b2",
       x"400000cf",
       x"80000035",
       x"0000006b",
       x"000000d6",
       x"00000006",
       x"0000000d",
       x"0000001a",
       x"00000035",
       x"0000006a",
       x"000000d5"
      ),

      -- G16
      (
       x"0001aab1",
       x"0002ffd3",
       x"00045516",
       x"0008aa2d",
       x"0010feea",
       x"00205765",
       x"0040aecb",
       x"0080f727",
       x"010044fe",
       x"020089fc",
       x"0400b948",
       x"0800d821",
       x"10001af2",
       x"200035e4",
       x"40006bc9",
       x"8000d792",
       x"00000594",
       x"00000b29",
       x"00001652",
       x"00002ca5",
       x"0000594a",
       x"0000b294",
       x"0000cf99",
       x"00003583",
       x"00006b06",
       x"0000d60d",
       x"000006aa",
       x"00000d55",
       x"00001aab",
       x"00003556",
       x"00006aac",
       x"0000d558"
      ),

      -- G32
      (
       x"aab14226",
       x"ffd3c66b",
       x"5516cef0",
       x"aa2d9de1",
       x"feea79e5",
       x"5765b1ec",
       x"aecb63d9",
       x"f7278595",
       x"44fe490c",
       x"89fc9218",
       x"b9486617",
       x"d8218e09",
       x"1af25e34",
       x"35e4bc69",
       x"6bc978d2",
       x"d792f1a5",
       x"0594a16d",
       x"0b2942da",
       x"165285b4",
       x"2ca50b68",
       x"594a16d1",
       x"b2942da3",
       x"cf991961",
       x"358370e5",
       x"6b06e1cb",
       x"d60dc397",
       x"06aac508",
       x"0d558a11",
       x"1aab1422",
       x"35562844",
       x"6aac5089",
       x"d558a113"
      ),

      -- G64
      (
       x"ff0a7996",
       x"011e8aba",
       x"fd376ce3",
       x"fa6ed9c6",
       x"0bd7ca1b",
       x"e8a5eda1",
       x"d14bdb43",
       x"5d9dcf10",
       x"4431e7b7",
       x"8863cf6e",
       x"efcde74a",
       x"2091b703",
       x"be291791",
       x"7c522f22",
       x"f8a45e45",
       x"f148bc8b",
       x"1d9b0081",
       x"3b360102",
       x"766c0205",
       x"ecd8040b",
       x"d9b00816",
       x"b360102d",
       x"99ca59cc",
       x"cc9eca0e",
       x"993d941c",
       x"327b2838",
       x"9bfc29e6",
       x"37f853cc",
       x"6ff0a799",
       x"dfe14f32",
       x"bfc29e65",
       x"7f853ccb"
      ),

      -- G128
      (
       x"54e65c12",
       x"fd2ae436",
       x"aeb3947e",
       x"5d6728fd",
       x"ee280de8",
       x"88b647c2",
       x"116c8f85",
       x"763f4318",
       x"b898da23",
       x"7131b446",
       x"b685349f",
       x"39ec352c",
       x"273e364a",
       x"4e7c6c95",
       x"9cf8d92a",
       x"39f1b255",
       x"270538b9",
       x"4e0a7172",
       x"9c14e2e4",
       x"3829c5c9",
       x"70538b93",
       x"e0a71726",
       x"95a8725f",
       x"7fb6b8ac",
       x"ff6d7158",
       x"fedae2b1",
       x"a9539970",
       x"52a732e0",
       x"a54e65c1",
       x"4a9ccb82",
       x"95399704",
       x"2a732e09"
      ),

      -- G256
      (
       x"662e22fb",
       x"aa72670c",
       x"32caece2",
       x"6595d9c4",
       x"ad059173",
       x"3c25001d",
       x"784a003a",
       x"96ba228f",
       x"4b5a67e5",
       x"96b4cfcb",
       x"4b47bd6d",
       x"f0a15821",
       x"876c92b9",
       x"0ed92572",
       x"1db24ae4",
       x"3b6495c9",
       x"10e70968",
       x"21ce12d1",
       x"439c25a2",
       x"87384b44",
       x"0e709688",
       x"1ce12d10",
       x"5fec78da",
       x"d9f6d34e",
       x"b3eda69c",
       x"67db4d38",
       x"a998b88b",
       x"53317117",
       x"a662e22f",
       x"4cc5c45f",
       x"998b88be",
       x"3317117d"
      ),

      -- G512
      (
       x"acf22d24",
       x"f516776d",
       x"46dec3fe",
       x"8dbd87fd",
       x"b78922de",
       x"c3e06898",
       x"87c0d130",
       x"a3738f45",
       x"ea1533ae",
       x"d42a675c",
       x"04a6e39d",
       x"a5bfea1f",
       x"e78df91b",
       x"cf1bf236",
       x"9e37e46c",
       x"3c6fc8d9",
       x"d42dbc97",
       x"a85b792f",
       x"50b6f25e",
       x"a16de4bd",
       x"42dbc97b",
       x"85b792f7",
       x"a79d08cb",
       x"e3c83cb2",
       x"c7907964",
       x"8f20f2c8",
       x"b2b3c8b4",
       x"65679169",
       x"cacf22d2",
       x"959e45a4",
       x"2b3c8b49",
       x"56791692"
      ),

      -- G1k
      (
       x"3e787f4d",
       x"428881d6",
       x"bb697ce1",
       x"76d2f9c3",
       x"d3dd8ccb",
       x"99c366da",
       x"3386cdb5",
       x"5975e426",
       x"8c93b700",
       x"19276e00",
       x"0c36a34d",
       x"261539d7",
       x"72520ce2",
       x"e4a419c5",
       x"c948338a",
       x"92906714",
       x"1b58b164",
       x"36b162c8",
       x"6d62c590",
       x"dac58b20",
       x"b58b1640",
       x"6b162c80",
       x"e854264d",
       x"eed033d6",
       x"dda067ac",
       x"bb40cf58",
       x"48f9e1fd",
       x"91f3c3fa",
       x"23e787f4",
       x"47cf0fe9",
       x"8f9e1fd3",
       x"1f3c3fa6"
      ),

      -- G2k
      (
       x"ca02a002",
       x"5e07e006",
       x"760d600f",
       x"ec1ac01e",
       x"1237203f",
       x"ee6ce07c",
       x"dcd9c0f8",
       x"73b121f3",
       x"2d60e3e5",
       x"5ac1c7cb",
       x"7f812f94",
       x"3500ff2a",
       x"a0035e56",
       x"4006bcad",
       x"800d795a",
       x"001af2b4",
       x"ca37456a",
       x"946e8ad5",
       x"28dd15aa",
       x"51ba2b55",
       x"a37456aa",
       x"46e8ad55",
       x"47d3faa9",
       x"45a55550",
       x"8b4aaaa0",
       x"16955541",
       x"e7280a80",
       x"ce501500",
       x"9ca02a00",
       x"39405400",
       x"7280a800",
       x"e5015001"
      ),

      -- G4k
      (
       x"c5319c8a",
       x"4f52a59e",
       x"5b94d7b6",
       x"b729af6d",
       x"ab62c250",
       x"93f4182a",
       x"27e83055",
       x"8ae1fc20",
       x"d0f264cb",
       x"a1e4c997",
       x"86f80fa5",
       x"c8c183c1",
       x"54b29b09",
       x"a9653612",
       x"52ca6c25",
       x"a594d84b",
       x"8e182c1d",
       x"1c30583b",
       x"3860b076",
       x"70c160ec",
       x"e182c1d8",
       x"c30583b0",
       x"433a9bea",
       x"4344ab5f",
       x"868956be",
       x"0d12ad7c",
       x"df14c672",
       x"be298ce4",
       x"7c5319c8",
       x"f8a63391",
       x"f14c6722",
       x"e298ce45"
      ),

      -- G8k
      (
       x"2b536e2b",
       x"7df5b27d",
       x"d0b80ad1",
       x"a17015a2",
       x"69b3456f",
       x"f835e4f5",
       x"f06bc9ea",
       x"cb84fdfe",
       x"bc5a95d7",
       x"78b52baf",
       x"da393974",
       x"9f211cc3",
       x"151157ac",
       x"2a22af58",
       x"54455eb1",
       x"a88abd63",
       x"7a4614ec",
       x"f48c29d8",
       x"e91853b0",
       x"d230a760",
       x"a4614ec1",
       x"48c29d83",
       x"bad6552c",
       x"5effc472",
       x"bdff88e4",
       x"7bff11c9",
       x"dcad4db8",
       x"b95a9b71",
       x"72b536e2",
       x"e56a6dc5",
       x"cad4db8a",
       x"95a9b715"
      ),

      -- G16k
      (
       x"46649bb0",
       x"caadacd0",
       x"d33fc211",
       x"a67f8423",
       x"0a9b93f6",
       x"5353bc5d",
       x"a6a778bb",
       x"0b2a6ac7",
       x"50304e3e",
       x"a0609c7d",
       x"06a5a34a",
       x"4b2fdd25",
       x"d03b21fa",
       x"a07643f5",
       x"40ec87ea",
       x"81d90fd5",
       x"45d6841b",
       x"8bad0837",
       x"175a106f",
       x"2eb420de",
       x"5d6841bd",
       x"bad0837a",
       x"33c59d45",
       x"21efa13b",
       x"43df4277",
       x"87be84ef",
       x"4919926e",
       x"923324dd",
       x"246649bb",
       x"48cc9376",
       x"919926ec",
       x"23324dd8"
      ),

      -- G32k
      (
       x"790e5a18",
       x"8b12ee28",
       x"6f2b8649",
       x"de570c92",
       x"c5a0433c",
       x"f24edc61",
       x"e49db8c2",
       x"b0352b9c",
       x"19640d21",
       x"32c81a43",
       x"1c9e6e9f",
       x"40328726",
       x"f96b5455",
       x"f2d6a8ab",
       x"e5ad5157",
       x"cb5aa2ae",
       x"efbb1f44",
       x"df763e89",
       x"beec7d12",
       x"7dd8fa24",
       x"fbb1f448",
       x"f763e891",
       x"97c98b3b",
       x"569d4c6e",
       x"ad3a98dc",
       x"5a7531b8",
       x"cde43968",
       x"9bc872d0",
       x"3790e5a1",
       x"6f21cb43",
       x"de439686",
       x"bc872d0c"
      )
    );

  --  H^(2^(i+3)+2^(i+4)) is stored to H11_MATRIX(i).
  --  for example, H11_MATRIX(0)  := H^24,
  --               H11_MATRIX(1)  := H^48,
  --               H11_MATRIX(2)  := H^96,
  --               ...
  --               H11_MATRIX(11) := H^48k, 
  constant H11_MATRIX : matrice :=
    (
      -- H24
      (
       x"01164100",
       x"033ac300",
       x"0763c700",
       x"0ec78e00",
       x"1c995d00",
       x"3824fb00",
       x"7049f600",
       x"e185ad00",
       x"c21d1b00",
       x"843a3600",
       x"09622d00",
       x"13d21b00",
       x"26b27700",
       x"4d64ee00",
       x"9ac9dc00",
       x"3593b800",
       x"6a313100",
       x"d4626200",
       x"a8c4c400",
       x"51898800",
       x"a3131000",
       x"46262000",
       x"8d5a0100",
       x"1ba24300",
       x"37448601",
       x"6e890c02",
       x"dc045904",
       x"b808b208",
       x"70116410",
       x"e022c820",
       x"c0459040",
       x"808b2080"
      ),

      -- H48
      (
       x"b025f701",
       x"d06e1903",
       x"10f9c507",
       x"21f38a0e",
       x"f3c2e31c",
       x"57a03138",
       x"af406270",
       x"eea533e1",
       x"6d6f90c2",
       x"dadf2184",
       x"059bb409",
       x"bb129f13",
       x"c600c926",
       x"8c01924d",
       x"1803249a",
       x"30064935",
       x"d029656a",
       x"a052cad4",
       x"40a595a8",
       x"814b2b51",
       x"029656a3",
       x"052cad46",
       x"ba7cad8d",
       x"c4dcac1b",
       x"89b95837",
       x"1372b06e",
       x"96c097dc",
       x"2d812fb8",
       x"5b025f70",
       x"b604bee0",
       x"6c097dc0",
       x"d812fb80"
      ),

      -- H96
      (
       x"c0be831e",
       x"41c38523",
       x"43398959",
       x"867312b2",
       x"cc58a67a",
       x"580fcfeb",
       x"b01f9fd7",
       x"a081bcb0",
       x"81bdfa7e",
       x"037bf4fd",
       x"c6496ae4",
       x"4c2c56d7",
       x"58e62eb0",
       x"b1cc5d61",
       x"6398bac2",
       x"c7317585",
       x"4edc6814",
       x"9db8d028",
       x"3b71a050",
       x"76e340a0",
       x"edc68140",
       x"db8d0280",
       x"77a4861e",
       x"2ff78f22",
       x"5fef1e44",
       x"bfde3c89",
       x"bf02fa0c",
       x"7e05f418",
       x"fc0be831",
       x"f817d063",
       x"f02fa0c7",
       x"e05f418f"
      ),

      -- H192
      (
       x"d4c01ec6",
       x"7d40234a",
       x"2e405853",
       x"5c80b0a7",
       x"6dc17f88",
       x"0f42e1d6",
       x"1e85c3ad",
       x"e9cb999c",
       x"07572dff",
       x"0eae5bfe",
       x"c99ca93b",
       x"47f94cb1",
       x"5b3287a4",
       x"b6650f48",
       x"6cca1e91",
       x"d9943d23",
       x"67e86481",
       x"cfd0c902",
       x"9fa19204",
       x"3f432409",
       x"7e864813",
       x"fd0c9027",
       x"2ed93e88",
       x"897263d7",
       x"12e4c7af",
       x"25c98f5e",
       x"9f53007b",
       x"3ea600f6",
       x"7d4c01ec",
       x"fa9803d8",
       x"f53007b1",
       x"ea600f63"
      ),

      -- H384
      (
       x"25534ce2",
       x"6ff5d526",
       x"fab8e6ae",
       x"f571cd5d",
       x"cfb0d658",
       x"ba32e053",
       x"7465c0a6",
       x"cd98cdaf",
       x"be62d7bc",
       x"7cc5af78",
       x"dcd81212",
       x"9ce368c7",
       x"1c959d6c",
       x"392b3ad9",
       x"725675b2",
       x"e4aceb64",
       x"ec0a9a2a",
       x"d8153454",
       x"b02a68a8",
       x"6054d151",
       x"c0a9a2a3",
       x"81534547",
       x"27f5c66c",
       x"6ab8c03a",
       x"d5718074",
       x"aae300e8",
       x"70954d33",
       x"e12a9a67",
       x"c25534ce",
       x"84aa699c",
       x"0954d338",
       x"12a9a671"
      ),

      -- H768
      (
       x"ac5a5971",
       x"f4eeeb93",
       x"45878e57",
       x"8b0f1cae",
       x"ba44602c",
       x"d8d29929",
       x"b1a53252",
       x"cf103dd5",
       x"327a22db",
       x"64f445b6",
       x"65b2d21d",
       x"673ffd4b",
       x"6225a3e6",
       x"c44b47cd",
       x"88968f9a",
       x"112d1f35",
       x"8e00671b",
       x"1c00ce36",
       x"38019c6c",
       x"700338d9",
       x"e00671b2",
       x"c00ce364",
       x"2c439fb9",
       x"f4dd6602",
       x"e9bacc05",
       x"d375980a",
       x"0ab16965",
       x"1562d2cb",
       x"2ac5a597",
       x"558b4b2e",
       x"ab16965c",
       x"562d2cb8"
      ),

      -- H1_5k
      (
       x"7caca516",
       x"85f5ef3b",
       x"77477b61",
       x"ee8ef6c2",
       x"a1b14892",
       x"3fce3432",
       x"7f9c6865",
       x"839475dc",
       x"7b844eaf",
       x"f7089d5e",
       x"92bd9fab",
       x"59d79a41",
       x"cf039194",
       x"9e072329",
       x"3c0e4653",
       x"781c8ca6",
       x"8c95bc5b",
       x"192b78b7",
       x"3256f16e",
       x"64ade2dc",
       x"c95bc5b9",
       x"92b78b72",
       x"59c3b3f3",
       x"cf2bc2f0",
       x"9e5785e0",
       x"3caf0bc1",
       x"05f2b294",
       x"0be56528",
       x"17caca51",
       x"2f9594a2",
       x"5f2b2945",
       x"be56528b"
      ),

      -- H3k
      (
       x"3e255879",
       x"426fe88a",
       x"bafa896d",
       x"75f512db",
       x"d5cf7dce",
       x"95bba3e5",
       x"2b7747cb",
       x"68cbd7ef",
       x"efb2f7a6",
       x"df65ef4c",
       x"80ee86e1",
       x"3ff855ba",
       x"41d5f30d",
       x"83abe61b",
       x"0757cc36",
       x"0eaf986d",
       x"237a68a2",
       x"46f4d144",
       x"8de9a289",
       x"1bd34512",
       x"37a68a25",
       x"6f4d144a",
       x"e0bf70ed",
       x"ff5bb9a3",
       x"feb77346",
       x"fd6ee68c",
       x"c4f89561",
       x"89f12ac3",
       x"13e25587",
       x"27c4ab0f",
       x"4f89561e",
       x"9f12ac3c"
      ),

      -- H6k
      (
       x"1c36c004",
       x"245b400c",
       x"5480401c",
       x"a9008038",
       x"4e37c074",
       x"805940ed",
       x"00b281da",
       x"1d53c3b1",
       x"26914766",
       x"4d228ecd",
       x"8673dd9f",
       x"10d17b3a",
       x"3d943670",
       x"7b286ce0",
       x"f650d9c0",
       x"eca1b381",
       x"c575a706",
       x"8aeb4e0d",
       x"15d69c1b",
       x"2bad3836",
       x"575a706d",
       x"aeb4e0db",
       x"415f01b2",
       x"9e88c360",
       x"3d1186c1",
       x"7a230d82",
       x"e870db00",
       x"d0e1b600",
       x"a1c36c00",
       x"4386d800",
       x"870db001",
       x"0e1b6002"
      ),

      -- H12k
      (
       x"b1f069fd",
       x"d210ba06",
       x"15d11df1",
       x"2ba23be3",
       x"e6b41e3b",
       x"7c98558b",
       x"f930ab17",
       x"43913fd3",
       x"36d2165a",
       x"6da42cb4",
       x"6ab83094",
       x"648008d5",
       x"78f07856",
       x"f1e0f0ac",
       x"e3c1e158",
       x"c783c2b1",
       x"3ef7ec9e",
       x"7defd93c",
       x"fbdfb278",
       x"f7bf64f0",
       x"ef7ec9e1",
       x"defd93c3",
       x"0c0b4e7b",
       x"a9e6f50b",
       x"53cdea16",
       x"a79bd42d",
       x"fec7c1a7",
       x"fd8f834f",
       x"fb1f069f",
       x"f63e0d3f",
       x"ec7c1a7f",
       x"d8f834fe"
      ),

      -- H24k
      (
       x"fd07abfa",
       x"0708fc0f",
       x"f31653e5",
       x"e62ca7ca",
       x"315ee46e",
       x"9fba6326",
       x"3f74c64d",
       x"83ee2761",
       x"fadbe538",
       x"f5b7ca71",
       x"16683f18",
       x"d1d7d5ca",
       x"5ea8006e",
       x"bd5000dd",
       x"7aa001bb",
       x"f5400377",
       x"1787ad15",
       x"2f0f5a2a",
       x"5e1eb455",
       x"bc3d68aa",
       x"787ad154",
       x"f0f5a2a9",
       x"1ceceea8",
       x"c4de76aa",
       x"89bced55",
       x"1379daaa",
       x"dbf41eaf",
       x"b7e83d5f",
       x"6fd07abf",
       x"dfa0f57f",
       x"bf41eafe",
       x"7e83d5fd"
      ),

      -- H48k
      (
       x"9775f984",
       x"b99e0a8d",
       x"e449ec9e",
       x"c893d93d",
       x"06524bfe",
       x"9bd16e79",
       x"37a2dcf2",
       x"f8304061",
       x"67157946",
       x"ce2af28c",
       x"0b201c9d",
       x"8135c0be",
       x"951e78f9",
       x"2a3cf1f3",
       x"5479e3e6",
       x"a8f3c7cd",
       x"c692761e",
       x"8d24ec3d",
       x"1a49d87a",
       x"3493b0f4",
       x"692761e8",
       x"d24ec3d0",
       x"33e87e24",
       x"f0a505cc",
       x"e14a0b98",
       x"c2941731",
       x"125dd7e6",
       x"24bbafcc",
       x"49775f98",
       x"92eebf30",
       x"25dd7e61",
       x"4bbafcc2"
      ),

      -- dummy
      (
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000"
      )
    );

  --  G^(2^(i+3)+2^(i+4)) is stored to G11_MATRIX(i).
  --  for example, G11_MATRIX(0)  := G^24,
  --               G11_MATRIX(1)  := G^48,
  --               G11_MATRIX(2)  := G^96,
  --               ...
  --               G11_MATRIX(11) := G^48k, 
  constant G11_MATRIX : matrice :=
    (
      -- G24
      (
       x"01aab142",
       x"02ffd3c6",
       x"045516ce",
       x"08aa2d9d",
       x"10feea79",
       x"205765b1",
       x"40aecb63",
       x"80f72785",
       x"0044fe49",
       x"0089fc92",
       x"00b94866",
       x"00d8218e",
       x"001af25e",
       x"0035e4bc",
       x"006bc978",
       x"00d792f1",
       x"000594a1",
       x"000b2942",
       x"00165285",
       x"002ca50b",
       x"00594a16",
       x"00b2942d",
       x"00cf9919",
       x"00358370",
       x"006b06e1",
       x"00d60dc3",
       x"0006aac5",
       x"000d558a",
       x"001aab14",
       x"00355628",
       x"006aac50",
       x"00d558a1"
      ),

      -- G48
      (
       x"4226ff0a",
       x"c66b011e",
       x"cef0fd37",
       x"9de1fa6e",
       x"79e50bd7",
       x"b1ece8a5",
       x"63d9d14b",
       x"85955d9d",
       x"490c4431",
       x"92188863",
       x"6617efcd",
       x"8e092091",
       x"5e34be29",
       x"bc697c52",
       x"78d2f8a4",
       x"f1a5f148",
       x"a16d1d9b",
       x"42da3b36",
       x"85b4766c",
       x"0b68ecd8",
       x"16d1d9b0",
       x"2da3b360",
       x"196199ca",
       x"70e5cc9e",
       x"e1cb993d",
       x"c397327b",
       x"c5089bfc",
       x"8a1137f8",
       x"14226ff0",
       x"2844dfe1",
       x"5089bfc2",
       x"a1137f85"
      ),

      -- G96
      (
       x"7a79406a",
       x"8e8bc0be",
       x"676ec117",
       x"cedd822f",
       x"e7c24434",
       x"b5fdc803",
       x"6bfb9007",
       x"ad8e6064",
       x"216580a2",
       x"42cb0145",
       x"ffef42e0",
       x"85a7c5ab",
       x"7136cb3c",
       x"e26d9678",
       x"c4db2cf0",
       x"89b659e1",
       x"6915f3a8",
       x"d22be750",
       x"a457cea0",
       x"48af9d41",
       x"915f3a82",
       x"22be7504",
       x"3f05aa63",
       x"047214ad",
       x"08e4295a",
       x"11c852b5",
       x"59e9e501",
       x"b3d3ca03",
       x"67a79406",
       x"cf4f280d",
       x"9e9e501a",
       x"3d3ca035"
      ),

      -- G192
      (
       x"63b8c9ce",
       x"a4c95a52",
       x"2a2a7d6b",
       x"5454fad6",
       x"cb113c62",
       x"f59ab10a",
       x"eb356214",
       x"b5d20de7",
       x"081cd200",
       x"1039a400",
       x"43cb81cf",
       x"e42fca50",
       x"abe75d6f",
       x"57cebade",
       x"af9d75bc",
       x"5f3aeb78",
       x"ddcd1f3e",
       x"bb9a3e7c",
       x"77347cf8",
       x"ee68f9f0",
       x"dcd1f3e1",
       x"b9a3e7c3",
       x"10ff0649",
       x"4246c55d",
       x"848d8aba",
       x"091b1574",
       x"718ee327",
       x"e31dc64e",
       x"c63b8c9c",
       x"8c771939",
       x"18ee3273",
       x"31dc64e7"
      ),

      -- G384
      (
       x"4897a883",
       x"d9b8f984",
       x"fbe65b8a",
       x"f7ccb715",
       x"a70ec6a9",
       x"068a25d1",
       x"0d144ba2",
       x"52bf3fc7",
       x"ede9d70c",
       x"dbd3ae19",
       x"ff30f4b1",
       x"b6f641e1",
       x"257b2b40",
       x"4af65681",
       x"95ecad03",
       x"2bd95a06",
       x"1f251c8e",
       x"3e4a391d",
       x"7c94723a",
       x"f928e474",
       x"f251c8e8",
       x"e4a391d0",
       x"81d08b23",
       x"4b36bec4",
       x"966d7d88",
       x"2cdafb10",
       x"11225ea2",
       x"2244bd44",
       x"44897a88",
       x"8912f510",
       x"1225ea20",
       x"244bd441"
      ),

      -- G768
      (
       x"c0b45547",
       x"41dcffc9",
       x"430daad4",
       x"861b55a9",
       x"cc82fe15",
       x"59b1a96c",
       x"b36352d9",
       x"a672f0f5",
       x"8c51b4ac",
       x"18a36958",
       x"f1f287f6",
       x"23515aab",
       x"8616e011",
       x"0c2dc023",
       x"185b8046",
       x"30b7008c",
       x"a1da545e",
       x"43b4a8bd",
       x"8769517a",
       x"0ed2a2f4",
       x"1da545e9",
       x"3b4a8bd2",
       x"b62142e2",
       x"acf6d082",
       x"59eda104",
       x"b3db4209",
       x"a702d155",
       x"4e05a2aa",
       x"9c0b4554",
       x"38168aa8",
       x"702d1551",
       x"e05a2aa3"
      ),

      -- G1_5k
      (
       x"0fef71bf",
       x"103192c1",
       x"2f8c543d",
       x"5f18a87b",
       x"b1de2149",
       x"6c53332c",
       x"d8a66659",
       x"bea3bd0d",
       x"72a80ba5",
       x"e550174a",
       x"c54f5f2a",
       x"8571cfea",
       x"050cee6a",
       x"0a19dcd5",
       x"1433b9aa",
       x"28677354",
       x"5f219716",
       x"be432e2d",
       x"7c865c5a",
       x"f90cb8b4",
       x"f2197169",
       x"e432e2d3",
       x"c78ab418",
       x"80fa198f",
       x"01f4331e",
       x"03e8663c",
       x"083fbdc6",
       x"107f7b8d",
       x"20fef71b",
       x"41fdee37",
       x"83fbdc6f",
       x"07f7b8df"
      ),

      -- G3k
      (
       x"b81cb118",
       x"c825d328",
       x"28571749",
       x"50ae2e92",
       x"1940ec3d",
       x"8a9d6963",
       x"153ad2c6",
       x"92691494",
       x"9cce9831",
       x"399d3062",
       x"cb26d1dc",
       x"2e5112a0",
       x"e4be9458",
       x"c97d28b1",
       x"92fa5163",
       x"25f4a2c7",
       x"f3f5f496",
       x"e7ebe92d",
       x"cfd7d25a",
       x"9fafa4b5",
       x"3f5f496a",
       x"7ebe92d4",
       x"456194b1",
       x"32df987b",
       x"65bf30f7",
       x"cb7e61ee",
       x"2ee072c4",
       x"5dc0e588",
       x"bb81cb11",
       x"77039623",
       x"ee072c46",
       x"dc0e588c"
      ),

      -- G6k
      (
       x"ff7c9b03",
       x"0185ad04",
       x"fc77c10a",
       x"f8ef8214",
       x"0ea39f2b",
       x"e23ba554",
       x"c4774aa9",
       x"77920e51",
       x"105887a0",
       x"20b10f41",
       x"be1e8581",
       x"83419001",
       x"f9ffbb00",
       x"f3ff7600",
       x"e7feec00",
       x"cffdd801",
       x"60872b00",
       x"c10e5600",
       x"821cac01",
       x"04395802",
       x"0872b005",
       x"10e5600a",
       x"deb65b17",
       x"42102d2d",
       x"84205a5b",
       x"0840b4b7",
       x"effdf26c",
       x"dffbe4d8",
       x"bff7c9b0",
       x"7fef9360",
       x"ffdf26c0",
       x"ffbe4d81"
      ),

      -- G12k
      (
       x"4748fb34",
       x"c9d90d5c",
       x"d4fae18d",
       x"a9f5c31b",
       x"14a37d03",
       x"6e0e0133",
       x"dc1c0266",
       x"ff70fff8",
       x"b9a904c4",
       x"73520989",
       x"a1ece826",
       x"04912b78",
       x"4e6aadc5",
       x"9cd55b8b",
       x"39aab716",
       x"73556e2c",
       x"a1e2276c",
       x"43c44ed9",
       x"87889db2",
       x"0f113b64",
       x"1e2276c8",
       x"3c44ed91",
       x"3fc12017",
       x"38cabb1b",
       x"71957636",
       x"e32aec6c",
       x"811d23ec",
       x"023a47d9",
       x"04748fb3",
       x"08e91f66",
       x"11d23ecd",
       x"23a47d9a"
      ),

      -- G24k
      (
       x"cadd08f4",
       x"5f67191d",
       x"74133acf",
       x"e826759f",
       x"1a91e3ca",
       x"fffecf61",
       x"fffd9ec2",
       x"35263571",
       x"a0916217",
       x"4122c42f",
       x"489880aa",
       x"5bec09a0",
       x"7d051bb5",
       x"fa0a376b",
       x"f4146ed6",
       x"e828ddac",
       x"1a8cb3ad",
       x"3519675a",
       x"6a32ceb4",
       x"d4659d68",
       x"a8cb3ad0",
       x"519675a1",
       x"69f1e3b7",
       x"193ecf9a",
       x"327d9f35",
       x"64fb3e6b",
       x"032b7423",
       x"0656e847",
       x"0cadd08f",
       x"195ba11e",
       x"32b7423d",
       x"656e847a"
      ),

      -- G48k
      (
       x"49e05290",
       x"da20f7b1",
       x"fda1bdf3",
       x"fb437be6",
       x"bf66a55c",
       x"372d1829",
       x"6e5a3053",
       x"95543236",
       x"634836fd",
       x"c6906dfb",
       x"c4c08966",
       x"c061405c",
       x"c922d228",
       x"9245a451",
       x"248b48a2",
       x"49169145",
       x"dbcd701b",
       x"b79ae037",
       x"6f35c06e",
       x"de6b80dc",
       x"bcd701b9",
       x"79ae0372",
       x"babc5475",
       x"3c98fa7b",
       x"7931f4f6",
       x"f263e9ed",
       x"ad27814a",
       x"5a4f0294",
       x"b49e0529",
       x"693c0a52",
       x"d27814a4",
       x"a4f02948"
      ),

      -- dummy
      (
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",

        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000"
      )
    );

  function calc_lfsr_serial(
    data     : in std_logic_vector;
    divisor  : in integer;
    crcwidth : in integer
  )
  return std_logic_vector;

  function calc_lfsr_wbe_serial(
    data     : in std_logic_vector;
    be       : in std_logic_vector;
    divisor  : in integer;
    crcwidth : in integer
  )
  return std_logic_vector;

  function mult_m_v(
    msrc : in matrix;
    vsrc : in std_logic_vector(NELEM-1 downto 0)
  )
  return std_logic_vector;

end package;

package body pcrcpkg is

  --
  -- feed data'length input to LFSR and process it with naive algorithm.
  --
  function calc_lfsr_serial(
    data     : in std_logic_vector;
    divisor  : in integer;
    crcwidth : in integer
  )
  return std_logic_vector is

    variable lfsr : std_logic_vector(crcwidth-1 downto 0);

  begin

      -- initial LFSR value should be all-zero, not all-one.
      lfsr := (others => '0');

      -- proces input bit by bit.
      eachbit : for i in 0 to data'length - 1 loop
        if ((data(i) xor lfsr(lfsr'left)) = '1') then
          -- LFSR is XORed with the divisor
          lfsr := (lfsr(lfsr'left-1 downto 0) & '0') xor
                     conv_std_logic_vector(divisor, lfsr'length);
        else
          -- shift LFSR 1-bit to the left.
          lfsr := (lfsr(lfsr'left-1 downto 0) & '0');
        end if;
      end loop eachbit;

      return lfsr;

  end calc_lfsr_serial;

  --
  -- feed data'length input and its byte enable to LFSR and process it with naive algorithm.
  --
  function calc_lfsr_wbe_serial(
    data     : in std_logic_vector;
    be       : in std_logic_vector;
    divisor  : in integer;
    crcwidth : in integer
  )
  return std_logic_vector is

    variable lfsr : std_logic_vector(crcwidth-1 downto 0);

  begin

      -- initial LFSR value should be all-zero, not all-one.
      lfsr := (others => '0');

      -- proces input bit by bit.
      eachbit : for i in 0 to data'length - 1 loop
        if (be(i/8) = '1') then
          if ((data(i) xor lfsr(lfsr'left)) = '1') then
            -- LFSR is XORed with the divisor
            lfsr := (lfsr(lfsr'left-1 downto 0) & '0') xor
                    conv_std_logic_vector(divisor, lfsr'length);
          else
            -- shift LFSR 1-bit to the left.
            lfsr := (lfsr(lfsr'left-1 downto 0) & '0');
          end if;
        end if;
      end loop eachbit;

      return lfsr;

  end calc_lfsr_wbe_serial;

  function mult_m_v(
    msrc : in matrix;
    vsrc : in std_logic_vector(NELEM-1 downto 0)
  )
  return std_logic_vector is

    variable vdst, vtmpA, vtmpB : std_logic_vector(NELEM-1 downto 0);

  begin

#if 0 // naive
    for r in 0 to NELEM-1 loop
      vdst(r) := '0';
      for c in 0 to NELEM-1 loop
        if (msrc(r, c) = '1' and vsrc(c) = '1') then
          vdst(r) := not vdst(r);
        end if;
      end loop;
    end loop;

#else // tuned
    for r in 0 to NELEM-1 loop
      vtmpA(r) := '0';
      for c in 0 to NELEM/2-1 loop
        if (msrc(r, c) = '1' and vsrc(c) = '1') then
          vtmpA(r) := not vtmpA(r);
        end if;
      end loop;
    end loop;

    for r in 0 to NELEM-1 loop
      vtmpB(r) := '0';
      for c in NELEM/2 to NELEM-1 loop
        if (msrc(r, c) = '1' and vsrc(c) = '1') then
          vtmpB(r) := not vtmpB(r);
        end if;
      end loop;
    end loop;

    for r in 0 to NELEM-1 loop
      vdst(r) := vtmpA(r) xor vtmpB(r);
    end loop;

#endif
    return vdst;

  end mult_m_v;

end pcrcpkg;
