--
-- Transaction Layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
use work.msslvpkg.all;

entity transaction is

  generic (
    NLANE             : integer;
    NVC               : integer;
    NDMACH            : integer;

    MAX_PAYLOAD       : integer;

    CA_PH_VC0_INIT    : integer;
    CA_NPH_VC0_INIT   : integer;
    CA_CH_VC0_INIT    : integer;
    CA_PD_VC0_INIT    : integer;
    CA_NPD_VC0_INIT   : integer;
    CA_CD_VC0_INIT    : integer;

    CL_PH_VC0_INIT    : integer;
    CL_NPH_VC0_INIT   : integer;
    CL_CH_VC0_INIT    : integer;
    CL_PD_VC0_INIT    : integer;
    CL_NPD_VC0_INIT   : integer;
    CL_CD_VC0_INIT    : integer;

    CFG_BAR0_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT  : std_logic_vector(31 downto 0);

    SIMULATION        : integer
  );
  port (

    triggerout        : out std_logic;

#if 1 // for debugging
    txpd_usedw         : out std_logic_vector(8 downto 0);
    txpd_full          : out std_logic;
    txpd_ovflw         : out std_logic;
    txpd_undflw        : out std_logic;
    rxempty_nph        : out std_logic;
    rxempty_ph         : out std_logic;
    rxempty_ch         : out std_logic;
    rxfull_ph          : out std_logic;
    rxfull_pd          : out std_logic;
#endif

    clk               : in  std_logic;
    rst               : in  std_logic;   -- active high

    -- TL-PHYMAC interface
    txreq             : out std_logic;
    txack             : in  std_logic;
    
    -- TL-DLL interface
    rxtop             : in  std_logic;
    rxwr              : in  std_logic;
    rx                : in  std_logic_vector(NLANE*16-1 downto 0);

    txtop             : out std_logic;
    txwr              : out std_logic;
    txbe              : out std_logic_vector(NLANE*2-1 downto 0);
    tx                : out std_logic_vector(NLANE*16-1 downto 0);
    tpsize            : out  std_logic_vector(OFFSET_WIDTH-1 downto 0);

    -- notify CA to data-link layer
    ca_nph            : out std_logic_vector(7 downto 0);
    ca_ph             : out std_logic_vector(7 downto 0);
    ca_ch             : out std_logic_vector(7 downto 0);
    ca_npd            : out std_logic_vector(11 downto 0);
    ca_pd             : out std_logic_vector(11 downto 0);
    ca_cd             : out std_logic_vector(11 downto 0);
    ca_p_flush        : out std_logic;
    ca_np_flush       : out std_logic;
    ca_c_flush        : out std_logic;

    -- CL notified by data-link layer
    cl_nph            : in  std_logic_vector( 7 downto 0);
    cl_ph             : in  std_logic_vector( 7 downto 0);
    cl_ch             : in  std_logic_vector( 7 downto 0);
    cl_npd            : in  std_logic_vector(11 downto 0);
    cl_pd             : in  std_logic_vector(11 downto 0);
    cl_cd             : in  std_logic_vector(11 downto 0);

    inf_cl_ph         : in  std_logic;
    inf_cl_pd         : in  std_logic;
    inf_cl_nph        : in  std_logic;
    inf_cl_npd        : in  std_logic;
    inf_cl_ch         : in  std_logic;
    inf_cl_cd         : in  std_logic;

    -- AL-TL interface
    mpls              : in  std_logic_vector(11 downto 0);

    cfg_bar0          : in  std_logic_vector(31 downto 0);
    cfg_bar1          : in  std_logic_vector(31 downto 0);
    cfg_bar2          : in  std_logic_vector(31 downto 0);
    cfg_bar3          : in  std_logic_vector(31 downto 0);
    cfg_bar4          : in  std_logic_vector(31 downto 0);
    cfg_bar5          : in  std_logic_vector(31 downto 0);
    cfg_bar_rom       : in  std_logic_vector(31 downto 0);
    
    cfg_read          : out std_logic;
    cfg_write         : out std_logic;
    cfg_addr          : out std_logic_vector(11 downto 0);
    cfg_be            : out std_logic_vector(3 downto 0);
    cfg_dataout       : out std_logic_vector(31 downto 0);
    cfg_datain        : in  std_logic_vector(31 downto 0);

    -- core interface
    slv_read          : out std_logic;
    slv_write         : out std_logic;
    slv_bar           : out std_logic_vector(6 downto 0);
    slv_addr          : out std_logic_vector(63 downto 0);
    slv_bytevalid     : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout       : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain        : in  std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq       : out std_logic;
    slv_writereq      : out std_logic;
    slv_accept        : in  std_logic;
    slv_bytecount     : out std_logic_vector(11 downto 0);

    ms_write          : out std_logic;
    ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_wraddr         : out std_logic_vector(31 downto 0);
    ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);

    ms_read           : out std_logic;
    ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_rdaddr         : out std_logic_vector(31 downto 0);
    ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

    dma_fifocnt_off   : out std_logic_vector(4 downto 0);
    dmach_wrreq       : in  std_logic_vector(NDMACH-1 downto 0);
    dmach_wrack       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdreq       : in  std_logic_vector(NDMACH-1 downto 0);
    dmach_rdack       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_length      : in  each10b(NDMACH-1 downto 0);
    dmach_paddr       : in  each64b(NDMACH-1 downto 0);
    dmach_laddr       : in  each32b(NDMACH-1 downto 0);

    dmach_rdcpl       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdcpllength : out std_logic_vector(9 downto 0);
    dmach_rdcplbcnt   : out std_logic_vector(11 downto 0);
    dmach_rdcpladdr   : in  each32b(NDMACH-1 downto 0);

    phy_linkup        : in  std_logic;
    dl_linkup         : in  std_logic
  );
end transaction;

architecture default of transaction is

  component rxctl
    generic (
      NLANE      : integer := NLANE
    );
    port (
      clk         : in  std_logic;
      rst         : in  std_logic;

      rxtop       : in  std_logic;
      rxwr        : in  std_logic;
      rx          : in  std_logic_vector(NLANE*16-1 downto 0);

      vc          : out std_logic_vector(2 downto 0);
      rxwr_bufsel : out std_logic;
      bufsel      : out std_logic_vector(1 downto 0);

      rxwr_header : out std_logic;
      header      : out std_logic_vector(127 downto 0);

      rxwr_data   : out std_logic;
      data        : out std_logic_vector(NLANE*16-1 downto 0)
    );
  end component;

  component txctl
    generic (
      NLANE             : integer := NLANE;
      CL_PH_VC0_INIT    : integer := CL_PH_VC0_INIT;
      CL_NPH_VC0_INIT   : integer := CL_NPH_VC0_INIT;
      CL_CH_VC0_INIT    : integer := CL_CH_VC0_INIT;
      CL_PD_VC0_INIT    : integer := CL_PD_VC0_INIT;
      CL_NPD_VC0_INIT   : integer := CL_NPD_VC0_INIT;
      CL_CD_VC0_INIT    : integer := CL_CD_VC0_INIT
    );
    port (
      clk           : in  std_logic;
      rst           : in  std_logic;

      -- CL notified by data-link layer
      cl_header     : in  std_logic_vector( 7 downto 0);
      cl_data       : in  std_logic_vector(11 downto 0);
      inf_cl_header : in  std_logic;
      inf_cl_data   : in  std_logic;

      -- notify TxFCbuf CA to slvctl
      cat_nph       : out std_logic_vector(7 downto 0);
      cat_ph        : out std_logic_vector(7 downto 0);
      cat_ch        : out std_logic_vector(7 downto 0);
      cat_npd       : out std_logic_vector(11 downto 0);
      cat_pd        : out std_logic_vector(11 downto 0);
      cat_cd        : out std_logic_vector(11 downto 0);

      hbufsel        : out std_logic_vector(1 downto 0);
      dbufsel        : out std_logic_vector(1 downto 0);
      
      empty_bufsel  : in  std_logic;
      empty_header  : in  std_logic;
      empty_data    : in  std_logic;

      rd_bufsel     : out std_logic;
      rd_header     : out std_logic;
      rd_data       : out std_logic;

      q_bufsel      : in  std_logic_vector(1 downto 0);
      q_header      : in  std_logic_vector(127 downto 0);
      q_data        : in  std_logic_vector(NLANE*16-1 downto 0);

      txreq         : out std_logic;
      txack         : in  std_logic;
      txtop         : out std_logic;
      txwr          : out std_logic;
      txbe          : out std_logic_vector(NLANE*2-1 downto 0);
      tpsize        : out  std_logic_vector(OFFSET_WIDTH-1 downto 0);
      tx            : out std_logic_vector(NLANE*16-1 downto 0)
    );
  end component;

  component txarb
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;   -- active high

      c_req     : in  std_logic;
      c_ack     : out std_logic;

      p_req     : in  std_logic;
      p_ack     : out std_logic;

      np_req    : in  std_logic;
      np_ack    : out std_logic;

      srcsel    : out std_logic_vector(1 downto 0)  -- TPBUFTYPE_P, _NP, or _C
    );
  end component;

  component slvctl
    generic (
      CA_PH_VC0_INIT    : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT   : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT    : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT    : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT   : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT    : integer := CA_CD_VC0_INIT;

      CFG_BAR0_INIT     : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT     : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT     : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT     : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT     : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT     : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT  : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;
      NLANE             : integer := NLANE;
      NDMACH            : integer := NDMACH
    );
    port (
      clk               : in  std_logic;
      rst               : in  std_logic;   -- active high

      ca_nph            : out std_logic_vector(7 downto 0);
      ca_ph             : out std_logic_vector(7 downto 0);
      ca_ch             : out std_logic_vector(7 downto 0);
      ca_npd            : out std_logic_vector(11 downto 0);
      ca_pd             : out std_logic_vector(11 downto 0);
      ca_cd             : out std_logic_vector(11 downto 0);
      ca_p_flush        : out std_logic;
      ca_np_flush       : out std_logic;
      ca_c_flush        : out std_logic;

      mpls              : in  std_logic_vector(11 downto 0);  -- max payload size

      cfg_bar0          : in  std_logic_vector(31 downto 0);
      cfg_bar1          : in  std_logic_vector(31 downto 0);
      cfg_bar2          : in  std_logic_vector(31 downto 0);
      cfg_bar3          : in  std_logic_vector(31 downto 0);
      cfg_bar4          : in  std_logic_vector(31 downto 0);
      cfg_bar5          : in  std_logic_vector(31 downto 0);
      cfg_bar_rom       : in  std_logic_vector(31 downto 0);

      rxhbufsel         : out std_logic_vector(1 downto 0);
      rxdbufsel         : out std_logic_vector(1 downto 0);
      rxempty_bufsel    : in  std_logic;

      rxrd_bufsel       : out std_logic;
      rxrd_header       : out std_logic;
      rxrd_data         : out std_logic;

      rxq_bufsel        : in  std_logic_vector(1 downto 0);
      rxq_header        : in  std_logic_vector(127 downto 0);
      rxq_data          : in  std_logic_vector(NLANE*16-1 downto 0);

      txvc              : out std_logic_vector(2 downto 0);

      cat_header        : in  std_logic_vector(7 downto 0);
      cat_data          : in  std_logic_vector(11 downto 0);

      bdf               : out std_logic_vector(15 downto 0);

      txwr_req          : out std_logic;
      txwr_ack          : in  std_logic;
      txwr_bufsel       : out std_logic;
      txwr_header       : out std_logic;
      txwr_data         : out std_logic;

      tx_header         : out std_logic_vector(127 downto 0);
      tx_data           : out std_logic_vector(NLANE*16-1 downto 0);

      slv_read          : out std_logic;
      slv_write         : out std_logic;
      slv_bar           : out std_logic_vector(6 downto 0);
      slv_addr          : out std_logic_vector(63 downto 0);
      slv_bytevalid     : out std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout       : out std_logic_vector(NLANE*16-1 downto 0);
      slv_datain        : in  std_logic_vector(NLANE*16-1 downto 0);

      slv_readreq       : out std_logic;
      slv_writereq      : out std_logic;
      slv_accept        : in  std_logic;
      slv_bytecount     : out std_logic_vector(11 downto 0);

      ms_read           : out std_logic;
      ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_rdaddr         : out std_logic_vector(31 downto 0);
      ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

      dmach_rdcpl       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdcpllength : out std_logic_vector(9 downto 0);
      dmach_rdcplbcnt   : out std_logic_vector(11 downto 0);
      dmach_rdcpladdr   : in  each32b(NDMACH-1 downto 0);

      cfg_read          : out std_logic;
      cfg_write         : out std_logic;
      cfg_addr          : out std_logic_vector(11 downto 0);
      cfg_be            : out std_logic_vector(3 downto 0);
      cfg_dataout       : out std_logic_vector(31 downto 0);
      cfg_datain        : in  std_logic_vector(31 downto 0)
    );
  end component;

  component msctl
    generic (
      NLANE             : integer := NLANE;
      NDMACH            : integer := NDMACH
    );
    port (
      clk               : in  std_logic;
      rst               : in  std_logic;   -- active high

      cat_ph            : in  std_logic_vector(7 downto 0);
      cat_pd            : in  std_logic_vector(11 downto 0);
      cat_nph           : in  std_logic_vector(7 downto 0);
      cat_npd           : in  std_logic_vector(11 downto 0);
      bdf               : in  std_logic_vector(15 downto 0);

      -- interface to TxFCbuf and txarb
      p_txwr_req        : out std_logic;
      p_txwr_ack        : in  std_logic;
      p_txwr_bufsel     : out std_logic;
      txwr_ph           : out std_logic;
      tx_ph             : out std_logic_vector(127 downto 0);
      txwr_pd           : out std_logic;
      tx_pd             : out std_logic_vector(NLANE*16-1 downto 0);

      np_txwr_req       : out std_logic;
      np_txwr_ack       : in  std_logic;
      np_txwr_bufsel    : out std_logic;
      txwr_nph          : out std_logic;
      tx_nph            : out std_logic_vector(127 downto 0);
      txwr_npd          : out std_logic;
      tx_npd            : out std_logic_vector(NLANE*16-1 downto 0);

      -- interface to master controllers in the user-side logic.
      ms_write          : out std_logic;
      ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
      ms_wraddr         : out std_logic_vector(31 downto 0);
      ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);

      -- interface to DMA controllers
      dma_fifocnt_off   : out std_logic_vector(4 downto 0);
      dmach_wrreq       : in  std_logic_vector(NDMACH-1 downto 0);
      dmach_wrack       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdreq       : in  std_logic_vector(NDMACH-1 downto 0);
      dmach_rdack       : out std_logic_vector(NDMACH-1 downto 0);
      dmach_length      : in  each10b(NDMACH-1 downto 0);
      dmach_paddr       : in  each64b(NDMACH-1 downto 0);
      dmach_laddr       : in  each32b(NDMACH-1 downto 0)
    );
  end component;

  component fcbuf
    generic (
      VIRTUAL_CHANNEL : integer;
      DATA_WIDTH      : integer;
      DATA_DEPTH      : integer
    );
    port (
      clk         : in  std_logic;
      rst         : in  std_logic;     -- active high reset

      vc          : in  std_logic_vector(2 downto 0);
      wr          : in  std_logic;
      rd          : in  std_logic;
      data        : in  std_logic_vector(DATA_WIDTH-1 downto 0);

      empty       : out std_logic;
      full        : out std_logic;
      q           : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
  end component;

  component fcbuf2
    generic (
      VIRTUAL_CHANNEL : integer;
      DATA_WIDTH      : integer;
      DATA_DEPTH      : integer
    );
    port (
      clk         : in  std_logic;
      rst         : in  std_logic;     -- active high reset

      vc          : in  std_logic_vector(2 downto 0);
      wr          : in  std_logic;
      rd          : in  std_logic;
      data        : in  std_logic_vector(DATA_WIDTH-1 downto 0);

      empty       : out std_logic;
      full        : out std_logic;
      usedw       : out std_logic_vector(8 downto 0);
      q           : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
  end component;

  signal rxvc            , txvc            : std_logic_vector(2 downto 0);

  --
  -- selection buffer control lines
  --
  signal rxrd_bufsel     , txrd_bufsel     : std_logic;
  signal rxempty_bufsel  , txempty_bufsel  : std_logic;
  signal rx_bufselin     , tx_bufselin     : std_logic_vector(1 downto 0);
  signal                   tx_hbufselout   : std_logic_vector(1 downto 0);
  signal                   tx_dbufselout   : std_logic_vector(1 downto 0);
  signal rx_hbufselout   , rx_dbufselout   : std_logic_vector(1 downto 0);
  signal rxq_bufsel      , txq_bufsel      : std_logic_vector(1 downto 0);

  --
  -- header buffer control lines
  --
  signal rxwr_bufsel                       : std_logic;
  signal rxwr_header    , txwr_header      : std_logic;
  signal rxrd_header    , txrd_header      : std_logic;
  signal                  txempty_header   : std_logic;
  signal rx_header                         : std_logic_vector(127 downto 0);
  signal rxq_header     , txq_header       : std_logic_vector(127 downto 0);
  signal tx_ch                             : std_logic_vector(127 downto 0);
  signal tx_cd                             : std_logic_vector(NLANE*16-1 downto 0);

  -- non-posted
  signal rxwr_nph       , txwr_nph         : std_logic;
  signal rxrd_nph       , txrd_nph         : std_logic;
  signal                  txempty_nph      : std_logic;
  signal rxq_nph        , txq_nph          : std_logic_vector(127 downto 0);
  signal cat_nph                           : std_logic_vector(7 downto 0);

  -- posted
  signal rxwr_ph       , txwr_ph           : std_logic;
  signal rxrd_ph       , txrd_ph           : std_logic;
  signal                 txempty_ph        : std_logic;
  signal rxq_ph        , txq_ph            : std_logic_vector(127 downto 0);
  signal cat_ph                            : std_logic_vector(7 downto 0);

  -- completion
  signal rxwr_ch       , txwr_ch           : std_logic;
  signal rxrd_ch       , txrd_ch           : std_logic;
  signal                 txempty_ch        : std_logic;
  signal rxq_ch        , txq_ch            : std_logic_vector(127 downto 0);
  signal cat_ch                            : std_logic_vector(7 downto 0);

  --
  -- data buffer control lines
  --
  signal rxwr_data      , txwr_data        : std_logic;
  signal rxrd_data      , txrd_data        : std_logic;
  signal                  txempty_data     : std_logic;
  signal rx_data                           : std_logic_vector(NLANE*16-1 downto 0);
  signal rxq_data       , txq_data         : std_logic_vector(NLANE*16-1 downto 0);

  -- non-posted
  signal rxwr_npd       , txwr_npd         : std_logic;
  signal rxrd_npd       , txrd_npd         : std_logic;
  signal                  txempty_npd      : std_logic;
  signal rxq_npd        , txq_npd          : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_npd                           : std_logic_vector(11 downto 0);

  -- posted
  signal rxwr_pd        , txwr_pd          : std_logic;
  signal rxrd_pd        , txrd_pd          : std_logic;
  signal                  txempty_pd       : std_logic;
  signal rxq_pd         , txq_pd           : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_pd                            : std_logic_vector(11 downto 0);

  -- completion
  signal rxwr_cd        , txwr_cd          : std_logic;
  signal rxrd_cd        , txrd_cd          : std_logic;
  signal                  txempty_cd       : std_logic;
  signal rxq_cd         , txq_cd           : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_cd                            : std_logic_vector(11 downto 0);

  signal cl_header       : std_logic_vector( 7 downto 0);
  signal cl_data         : std_logic_vector(11 downto 0);
  signal inf_cl_header   : std_logic;
  signal inf_cl_data     : std_logic;

  signal txwr_bufsel : std_logic;

  signal c_txwr_req      : std_logic;
  signal c_txwr_ack      : std_logic;
  signal c_txwr_bufsel   : std_logic;

  signal p_txwr_req      : std_logic;
  signal p_txwr_ack      : std_logic;
  signal p_txwr_bufsel   : std_logic;
  signal tx_ph           : std_logic_vector(127 downto 0);
  signal tx_pd           : std_logic_vector(NLANE*16-1 downto 0);

  signal np_txwr_req     : std_logic;
  signal np_txwr_ack     : std_logic;
  signal np_txwr_bufsel  : std_logic;
  signal tx_nph          : std_logic_vector(127 downto 0);
  signal tx_npd          : std_logic_vector(NLANE*16-1 downto 0);

  signal tx_srcsel       : std_logic_vector(1 downto 0);
  signal bdf : std_logic_vector(15 downto 0);


  -- for debugging
  signal txpd_full0         : std_logic;

  attribute keep : boolean;
  attribute keep of txpd_full   : signal is true;
  attribute keep of txpd_usedw  : signal is true;

-- synthesis translate_off
  attribute keep of txpd_ovflw  : signal is true;
  attribute keep of txpd_undflw : signal is true;

  attribute keep : boolean;
  attribute keep of txpd_full : signal is true;
  attribute keep of txrd_bufsel     : signal is true;
  attribute keep of txempty_bufsel  : signal is true;
  attribute keep of txrd_ph         : signal is true;
  attribute keep of txempty_ph      : signal is true;
  attribute keep of txrd_ch         : signal is true;
  attribute keep of txempty_ch      : signal is true;
  attribute keep of txwr_bufsel     : signal is true;

  attribute keep of txq_bufsel      : signal is true;
  attribute keep of rxfull_ph     : signal is true;
  attribute keep of rxfull_pd     : signal is true;
-- synthesis translate_on

begin

  -- Rx Flow Control Buffer Input Controller
  rxctl0 : rxctl
    port map (
      clk         => clk,
      rst         => rst,

      rxtop       => rxtop,
      rxwr        => rxwr,
      rx          => rx,

      vc          => rxvc,
      rxwr_bufsel => rxwr_bufsel,
      bufsel      => rx_bufselin,

      rxwr_header => rxwr_header,
      rxwr_data   => rxwr_data,

      header      => rx_header,
      data        => rx_data
    );

  -- Rx Flow Control Buffer Input Selector
  process (rx_bufselin, rxwr_header, rxwr_data)
  begin

    rxwr_nph <= '0';
    rxwr_npd <= '0';
    rxwr_ph  <= '0';
    rxwr_pd  <= '0';
    rxwr_ch  <= '0';
    rxwr_cd  <= '0';

    case rx_bufselin is

      when TPBUFTYPE_NP =>
        rxwr_nph <= rxwr_header;
        rxwr_npd <= rxwr_data;

      when TPBUFTYPE_P =>
        rxwr_ph <= rxwr_header;
        rxwr_pd <= rxwr_data;

      when TPBUFTYPE_C =>
        rxwr_ch <= rxwr_header;
        rxwr_cd <= rxwr_data;

      when others =>

    end case;
  end process;

  -- Slave Controller
  slvctl0 : slvctl
    port map (
      clk             => clk,
      rst             => rst,

      ca_nph          => ca_nph,
      ca_ph           => ca_ph,
      ca_ch           => ca_ch,
      ca_npd          => ca_npd,
      ca_pd           => ca_pd,
      ca_cd           => ca_cd,
      ca_p_flush      => ca_p_flush,
      ca_np_flush     => ca_np_flush,
      ca_c_flush      => ca_c_flush,

      mpls            => mpls,

      cfg_bar0        => cfg_bar0,
      cfg_bar1        => cfg_bar1,
      cfg_bar2        => cfg_bar2,
      cfg_bar3        => cfg_bar3,
      cfg_bar4        => cfg_bar4,
      cfg_bar5        => cfg_bar5,
      cfg_bar_rom     => cfg_bar_rom,

      rxhbufsel       => rx_hbufselout,
      rxdbufsel       => rx_dbufselout,
      rxempty_bufsel  => rxempty_bufsel,

      rxrd_bufsel     => rxrd_bufsel,
      rxrd_header     => rxrd_header,
      rxrd_data       => rxrd_data,

      rxq_bufsel      => rxq_bufsel,
      rxq_header      => rxq_header,
      rxq_data        => rxq_data,

      txvc            => txvc,

      cat_header      => cat_ch,
      cat_data        => cat_cd,

      bdf             => bdf,

      txwr_req        => c_txwr_req,
      txwr_ack        => c_txwr_ack,
      txwr_bufsel     => c_txwr_bufsel,
      txwr_header     => txwr_ch,
      txwr_data       => txwr_cd,

      tx_header       => tx_ch,
      tx_data         => tx_cd,

      slv_read        => slv_read,
      slv_write       => slv_write,
      slv_bar         => slv_bar,
      slv_addr        => slv_addr,
      slv_bytevalid   => slv_bytevalid,
      slv_dataout     => slv_dataout,
      slv_datain      => slv_datain,

      slv_readreq     => slv_readreq,
      slv_writereq    => slv_writereq,
      slv_accept      => slv_accept,
      slv_bytecount   => slv_bytecount,

      ms_read         => ms_read,
      ms_rdchannel    => ms_rdchannel,
      ms_rdaddr       => ms_rdaddr,
      ms_rddata       => ms_rddata,

      dmach_rdcpl        => dmach_rdcpl,
      dmach_rdcpllength  => dmach_rdcpllength,
      dmach_rdcplbcnt    => dmach_rdcplbcnt,
      dmach_rdcpladdr    => dmach_rdcpladdr,

      cfg_read        => cfg_read,
      cfg_write       => cfg_write,
      cfg_addr        => cfg_addr,
      cfg_be          => cfg_be,
      cfg_dataout     => cfg_dataout,
      cfg_datain      => cfg_datain
    );

  -- Rx Header Flow Control Buffer Output Selector
  process (rx_hbufselout, rxrd_header, rxq_nph, rxq_ph, rxq_ch)
  begin

    rxrd_nph <= '0';
    rxrd_ph  <= '0';
    rxrd_ch  <= '0';

    case rx_hbufselout is
      when TPBUFTYPE_NP =>
        rxq_header <= rxq_nph;
        rxrd_nph <= rxrd_header;

      when TPBUFTYPE_P =>
        rxq_header <= rxq_ph;
        rxrd_ph <= rxrd_header;

      when TPBUFTYPE_C =>
        rxq_header <= rxq_ch;
        rxrd_ch <= rxrd_header;

      when others =>
        rxq_header <= (others => '0');

    end case;
  end process;

  -- Rx Data Flow Control Buffer Output Selector
  process (rx_dbufselout, rxrd_data, rxq_npd, rxq_pd, rxq_cd)
  begin

    rxrd_npd <= '0';
    rxrd_pd  <= '0';
    rxrd_cd  <= '0';

    case rx_dbufselout is

      when TPBUFTYPE_NP =>
        rxq_data   <= rxq_npd;
        rxrd_npd <= rxrd_data;

      when TPBUFTYPE_P =>
        rxq_data   <= rxq_pd;
        rxrd_pd <= rxrd_data;

      when TPBUFTYPE_C =>
        rxq_data   <= rxq_cd;
        rxrd_cd <= rxrd_data;

      when others =>
        rxq_data   <= (others => '0');

    end case;
  end process;

  -- Tx Flow Control Buffer Output Controller
  txctl0 : txctl
    port map (
      clk           => clk,
      rst           => rst,

      cl_header     => cl_header,
      cl_data       => cl_data,
      inf_cl_header => inf_cl_header,
      inf_cl_data   => inf_cl_data,

      cat_nph       => cat_nph,
      cat_ph        => cat_ph,
      cat_ch        => cat_ch,
      cat_npd       => cat_npd,
      cat_pd        => cat_pd,
      cat_cd        => cat_cd,

      hbufsel        => tx_hbufselout,
      dbufsel        => tx_dbufselout,

      empty_bufsel  => txempty_bufsel,
      empty_header  => txempty_header,
      empty_data    => txempty_data,

      rd_bufsel     => txrd_bufsel,
      rd_header     => txrd_header,
      rd_data       => txrd_data,

      q_bufsel      => txq_bufsel,
      q_header      => txq_header,
      q_data        => txq_data,

      txreq         => txreq,
      txack         => txack,
      txtop         => txtop,
      txwr          => txwr,
      txbe          => txbe,
      tpsize        => tpsize,
      tx            => tx
    );

  -- Tx Header Flow Control Buffer Output Selector
  process (tx_hbufselout, txrd_header,
           txempty_nph, txq_nph,
           txempty_ph , txq_ph,
           txempty_ch , txq_ch)
  begin

    txrd_nph <= '0';
    txrd_ph  <= '0';
    txrd_ch  <= '0';

    case tx_hbufselout is

      when TPBUFTYPE_NP =>
        txempty_header <= txempty_nph;
        txq_header     <= txq_nph;

        txrd_nph    <= txrd_header;

      when TPBUFTYPE_P =>
        txempty_header <= txempty_ph;
        txq_header     <= txq_ph;

        txrd_ph     <= txrd_header;

      when TPBUFTYPE_C =>
        txempty_header <= txempty_ch;
        txq_header     <= txq_ch;

        txrd_ch     <= txrd_header;

      when others =>
        txempty_header <= '1';
        txq_header     <= (others => '0');

    end case;
  end process;


  -- Tx Data Flow Control Buffer Output Selector
  process (tx_dbufselout, txrd_data,
           txempty_npd, txq_npd,
           txempty_pd, txq_ph,
           txempty_cd, txq_ch,
           cl_nph, cl_npd, inf_cl_nph, inf_cl_npd,
           cl_ph , cl_pd , inf_cl_ph , inf_cl_pd,
           cl_ch , cl_cd , inf_cl_ch , inf_cl_cd)
           
  begin

    txrd_npd <= '0';
    txrd_pd  <= '0';
    txrd_cd  <= '0';

    case tx_dbufselout is

      when TPBUFTYPE_NP =>
        cl_header      <= cl_nph;
        cl_data        <= cl_npd;
        inf_cl_header  <= inf_cl_nph;
        inf_cl_data    <= inf_cl_npd;

        txempty_data   <= txempty_npd;
        txq_data       <= txq_npd;

        txrd_npd    <= txrd_data;

      when TPBUFTYPE_P =>
        cl_header      <= cl_ph;
        cl_data        <= cl_pd;
        inf_cl_header  <= inf_cl_ph;
        inf_cl_data    <= inf_cl_pd;

        txempty_data   <= txempty_pd;
        txq_data       <= txq_pd;

        txrd_pd     <= txrd_data;

      when TPBUFTYPE_C =>
        cl_header      <= cl_ch;
        cl_data        <= cl_cd;
        inf_cl_header  <= inf_cl_ch;
        inf_cl_data    <= inf_cl_cd;

        txempty_data   <= txempty_cd;
        txq_data       <= txq_cd;

        txrd_cd     <= txrd_data;

      when others =>
        cl_header      <= (others => '0');
        cl_data        <= (others => '0');
        inf_cl_header  <= '0';
        inf_cl_data    <= '0';

        txempty_data   <= '1';
        txq_data       <= (others => '0');

    end case;
  end process;

  
  msctl0 : msctl
    port map (
      clk              => clk,
      rst              => rst,

      cat_ph           => cat_ph,
      cat_pd           => cat_pd,
      cat_nph          => cat_nph,
      cat_npd          => cat_npd,

      bdf              => bdf,
      
      p_txwr_req       => p_txwr_req,
      p_txwr_ack       => p_txwr_ack,
      p_txwr_bufsel    => p_txwr_bufsel,
      txwr_ph          => txwr_ph,
      tx_ph            => tx_ph,
      txwr_pd          => txwr_pd,
      tx_pd            => tx_pd,

      np_txwr_req      => np_txwr_req,
      np_txwr_ack      => np_txwr_ack,
      np_txwr_bufsel   => np_txwr_bufsel,
      txwr_nph         => txwr_nph,
      tx_nph           => tx_nph,
      txwr_npd         => txwr_npd,
      tx_npd           => tx_npd,

      ms_wrchannel     => ms_wrchannel,
      ms_write         => ms_write,
      ms_wraddr        => ms_wraddr,
      ms_wrdata        => ms_wrdata,

      dma_fifocnt_off  => dma_fifocnt_off,
      dmach_wrreq      => dmach_wrreq,
      dmach_wrack      => dmach_wrack,
      dmach_rdreq      => dmach_rdreq,
      dmach_rdack      => dmach_rdack,
      dmach_length     => dmach_length,
      dmach_paddr      => dmach_paddr,
      dmach_laddr      => dmach_laddr
    );

  -- exclusively select write requests from slvctl (completion) or
  -- msctl (posted or non-posted).
  txarb0 : txarb
    port map (
      clk     => clk,
      rst     => rst,

      c_req   => c_txwr_req,
      c_ack   => c_txwr_ack,

      p_req   => p_txwr_req,
      p_ack   => p_txwr_ack,

      np_req  => np_txwr_req,
      np_ack  => np_txwr_ack,

      srcsel  => tx_srcsel
    );

  process(tx_srcsel, c_txwr_bufsel, p_txwr_bufsel, np_txwr_bufsel)
  begin
    case (tx_srcsel) is
      when TPBUFTYPE_C =>
        txwr_bufsel <= c_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_C;

      when TPBUFTYPE_P =>
        txwr_bufsel <= p_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_P;

      when TPBUFTYPE_NP =>
        txwr_bufsel <= np_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_NP;

      when others =>

    end case;
  end process;

  --
  -- Flow Control Buffers
  --

  vc : for i in 0 to NVC-1 generate

    -- Rx Buffer Selection buffer
    rxselbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPH_VC0_INIT + CA_PH_VC0_INIT + CA_CH_VC0_INIT,
        DATA_WIDTH => rx_bufselin'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_bufsel,
        data  => rx_bufselin,

        rd    => rxrd_bufsel,
        empty => rxempty_bufsel,
        q     => rxq_bufsel
      );

    -- Rx Non-Posted Header buffer
    rxnphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_nph,
        data  => rx_header,

        rd    => rxrd_nph,
        empty => rxempty_nph,
        q     => rxq_nph
      );

    -- Rx Posted Header buffer
    rxphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_PH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_ph,
        data  => rx_header,

        rd    => rxrd_ph,
        empty => rxempty_ph,
        full  => rxfull_ph,
        q     => rxq_ph
      );

    -- Rx Completion Header buffer
    rxchbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_CH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_ch,
        data  => rx_header,

        rd    => rxrd_ch,
        empty => rxempty_ch,
        q     => rxq_ch
      );

    -- Rx Non-Posted Data buffer
    rxnpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_npd,
        data  => rx_data,

        rd    => rxrd_npd,
        empty => open,
        q     => rxq_npd
      );

    -- Rx Posted Data buffer
    rxpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_PD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_pd,
        data  => rx_data,

        rd    => rxrd_pd,
        empty => open,
        q     => rxq_pd
      );

    -- Rx Completion Data buffer
    rxcdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_CD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => rxvc,
        wr    => rxwr_cd,
        data  => rx_data,

        rd    => rxrd_cd,
        empty => open,
        full  => rxfull_pd,
        q     => rxq_cd
      );

    -- Tx Buffer Selection buffer
    txselbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH  => CL_NPH_VC0_INIT + CL_PH_VC0_INIT + CL_CH_VC0_INIT,
        DATA_WIDTH  => tx_bufselin'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_bufsel,
        data  => tx_bufselin,

        rd    => txrd_bufsel,
        empty => txempty_bufsel,
        q     => txq_bufsel
      );

    -- Tx Non-Posted Header buffer
    txnphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_NPH_VC0_INIT,
        DATA_WIDTH => tx_nph'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_nph,
        data  => tx_nph,

        rd    => txrd_nph,
        empty => txempty_nph,
        q     => txq_nph
      );

    -- Tx Posted Header buffer
    txphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_PH_VC0_INIT,
        DATA_WIDTH => tx_ph'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_ph,
        data  => tx_ph,

        rd    => txrd_ph,
        empty => txempty_ph,
        q     => txq_ph
      );

    -- Tx Completion Header buffer
    txchbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_CH_VC0_INIT,
        DATA_WIDTH => tx_ch'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_ch,
        data  => tx_ch,

        rd    => txrd_ch,
        empty => txempty_ch,
        q     => txq_ch
      );

    -- Tx Non-Posted Data buffer
    txnpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_NPD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_npd'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_npd,
        data  => tx_npd,

        rd    => txrd_npd,
        empty => txempty_npd,
        q     => txq_npd
      );

    -- Tx Posted Data buffer
    txpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_PD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_pd'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_pd,
        data  => tx_pd,

        rd    => txrd_pd,
        empty => txempty_pd,
--        full  => txpd_full0,
--        usedw => txpd_usedw,
        q     => txq_pd
      );

--    txpd_full   <= txpd_full0;
--    txpd_ovflw  <= '1' when txpd_full0 = '1' and txwr_pd = '1' else '0';
--    txpd_undflw <= '1' when txempty_pd = '1' and txrd_pd = '1' else '0';
    
    -- Tx Completion Data buffer
    txcdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_CD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_cd'length
      )
      port map (
        clk   => clk,
        rst   => rst,

        vc    => txvc,
        wr    => txwr_cd,
        data  => tx_cd,

        rd    => txrd_cd,
        empty => txempty_cd,
        q     => txq_cd
      );

  end generate vc;

end default;


--
-- TLP receiver
--   parse incoming TLP and output its
--   header/data to Rx{NP,P,C}{H,D}buf,
--   according to its type.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity rxctl is
  generic (
    NLANE      : integer
  );
  port (
--    cont        : out std_logic;

    clk         : in  std_logic;
    rst         : in  std_logic;
    
    rxtop       : in  std_logic;
    rxwr        : in  std_logic;        -- not used
    rx          : in  std_logic_vector(NLANE*16-1 downto 0);

    vc          : out std_logic_vector(2 downto 0);
    rxwr_bufsel : out std_logic;
    bufsel      : out std_logic_vector(1 downto 0);

    rxwr_header : out std_logic;
    header      : out std_logic_vector(127 downto 0);

    rxwr_data   : out std_logic;
    data        : out std_logic_vector(NLANE*16-1 downto 0)
  );
end rxctl;

architecture default of rxctl is

  type array3x3 is array (7 downto 0) of std_logic_vector(2 downto 0);
  signal tc2vc        : array3x3;

--  attribute keep : boolean;
--  attribute keep of cont    : signal is true;
  
begin

  x1 : if (NLANE = 1) generate
    x1block : block

      type RXCTL_STATE is (
        IDLE,
        RECV_B2_B3,
        RECV_HEADRx1,
        RECV_DATA
      );

--  attribute enum_encoding: string;
--  attribute enum_encoding of RXCTL_STATE : type is "00 01 10 11";

      signal rxsm         : RXCTL_STATE;
      signal have_data    : std_logic;
      signal payload_left : std_logic_vector(11 downto 0);  -- in byte unit
      signal header_left  : std_logic_vector(3 downto 0);
      signal header_count : std_logic_vector(3 downto 0);
      signal short_header : std_logic;
      signal rx1, rx2     : std_logic_vector(NLANE*16-1 downto 0);
      signal header0      : std_logic_vector(127 downto 0);

    begin

      header <= header0;

      process (clk, rst)
      begin
        if (rst = '1') then
          rxwr_bufsel  <= '0';
          rxwr_header  <= '0';
          rxwr_data    <= '0';
          rxsm <= IDLE;
        elsif (rising_edge(clk)) then

          rxwr_bufsel <= '0';
          rxwr_header <= '0';
          rxwr_data   <= '0';
          rx1 <= rx;
          rx2 <= rx1;

          case rxsm is
            when IDLE =>

              -- TLP byte 0 & 1
--              if (rxwr = '1') then
              if (rxtop = '1') then      -- !!! replayed TLPs are filtered here.

                -- TLP has data?
                if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA) or
                    (rx(0*8+6 downto 0*8+5) = TPFMT_16BH_DATA)) then
                  have_data    <= '1';
                else
                  have_data    <= '0';
                end if;

                -- header size
                if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_NODATA) or
                    (rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA)) then
                  short_header <= '1';  -- 12-byte header.
                  if (NLANE = 1) then
                    header_left <= conv_std_logic_vector(12-NLANE*2, header_left'length);
                  end if;
                else
                  short_header <= '0';  -- 16-byte header.
                  header_left <= conv_std_logic_vector(16-NLANE*2, header_left'length);
                end if;

                -- buffer type                                     fmt type   bufsel
                if (rx(0*8+6 downto 0*8+0) = "1000000" or        -- MWr    10  00000  P
                    rx(0*8+6 downto 0*8+0) = "1100000" or        -- MWr    11  00000  P
                    rx(0*8+6 downto 0*8+3) = "0110" or           -- Msg    01  10XXX  P
                    rx(0*8+6 downto 0*8+3) = "1110") then        -- Msg    11  10XXX  P
                  bufsel <= TPBUFTYPE_P;
                elsif (rx(0*8+6 downto 0*8+0) = "0001010" or     -- Cpl    00  01010  C
                       rx(0*8+6 downto 0*8+0) = "1001010" or     -- CplD   10  01010  C
                       rx(0*8+6 downto 0*8+0) = "0001011" or     -- CplLk  00  01011  C
                       rx(0*8+6 downto 0*8+0) = "1001011") then  -- CplDLk 10  01011  C
                  bufsel <= TPBUFTYPE_C;
                else                                     -- others            NP
                  bufsel <= TPBUFTYPE_NP;
                end if;

                -- virtual channel
                for i in 0 to 5 loop
                  if (rx(1*8+6 downto 1*8+4) = conv_std_logic_vector(i, 3)) then
                    vc <= tc2vc(i);
                  end if;
                end loop;           -- virtual channel

                -- output header byte 0 & 1.
                header0(1*8+7 downto 0*8+0) <= rx;
                header_count <= conv_std_logic_vector(NLANE*2, header_count'length);
                rxsm <= RECV_B2_B3;

              end if;

            when RECV_B2_B3 =>
              -- payload in byte unit
              payload_left <= rx(0*8+1 downto 0*8+0) & rx(1*8+7 downto 1*8+0) & "00";

              -- output header byte 2 & 3.
              header0(3*8+7 downto 2*8+0) <= rx;
              header_count <= unsigned(header_count) + NLANE*2;
              header_left  <= unsigned(header_left) - NLANE*2;
              rxsm <= RECV_HEADRx1;

            when RECV_HEADRx1 =>
              if (header_left = conv_std_logic_vector(NLANE*2, header_left'length)) then
                rxwr_bufsel  <= '1';
                rxwr_header  <= '1';
                if (have_data = '1') then
                  rxsm <= RECV_DATA;
                else
                  rxsm <= IDLE;
                end if;
              else
                header_left <= unsigned(header_left) - NLANE*2;
              end if;

              -- output header byte : 4&5, 6&7, 8&9, 10&11 (, 12&13, 14&15).
              -- 2*i runs through   : 4,   6,   8,   10,      12,    14
              for i in 2 to 7 loop
                if (header_count = conv_std_logic_vector(2*i, header_left'length)) then
                  header0(16*(i+1)-1 downto 16*i)   <= rx;
                end if;
              end loop;
              header_count <= unsigned(header_count) + NLANE*2;

            when RECV_DATA =>
              rxwr_data    <= '1';
              data <= rx;

              if (payload_left <= conv_std_logic_vector(NLANE*2, payload_left'length)) then
                rxsm <= IDLE;
              else
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;

            when others =>
              rxsm <= IDLE;

          end case;

        end if;
      end process;

    end block x1block;
  end generate x1;

  x4 : if (NLANE >= 4) generate
    x4block : block
      
      signal have_data    : std_logic;
      signal short_header : std_logic;
      signal bufsel1      : std_logic_vector(1 downto 0);
      signal bufsel2      : std_logic_vector(1 downto 0);
      signal top1, top2   : std_logic;
      signal rx1, rx2     : std_logic_vector(NLANE*16-1 downto 0);
      signal payload_left : std_logic_vector(11 downto 0);  -- in byte unit

    begin

#if 0 // for debugging
      process (clk, rst)
      begin
        if (rst = '1') then
          cont <= '0';
        elsif (rising_edge(clk)) then
          if (rxtop = '1' and top1 = '1') then
            cont <= '1';
          end if;
        end if;
      end process;
#endif

      bufsel <= bufsel2;

      process (clk, rst)
      begin
        if (rst = '1') then
          rxwr_bufsel  <= '0';
          rxwr_header  <= '0';
          rxwr_data    <= '0';
        elsif (rising_edge(clk)) then

          rxwr_bufsel  <= '0';
          rxwr_header  <= '0';
          rxwr_data    <= '0';

          top1 <= rxtop;
          top2 <= top1;
          rx1 <= rx;
          rx2 <= rx1;
          bufsel2 <= bufsel1;

          if (rxtop = '1') then

            -- TLP has data?
            if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA) or
                (rx(0*8+6 downto 0*8+5) = TPFMT_16BH_DATA)) then
              have_data    <= '1';
            else
              have_data    <= '0';
            end if;

            -- header size
            if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_NODATA) or
                (rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA)) then
              short_header <= '1';  -- 12-byte header.
            else
              short_header <= '0';  -- 16-byte header.
            end if;

            -- buffer type                                     fmt type   bufsel
            if (rx(0*8+6 downto 0*8+0) = "1000000" or        -- MWr    10  00000  P
                rx(0*8+6 downto 0*8+0) = "1100000" or        -- MWr    11  00000  P
                rx(0*8+6 downto 0*8+3) = "0110" or           -- Msg    01  10XXX  P
                rx(0*8+6 downto 0*8+3) = "1110") then        -- Msg    11  10XXX  P
              bufsel1 <= TPBUFTYPE_P;
            elsif (rx(0*8+6 downto 0*8+0) = "0001010" or     -- Cpl    00  01010  C
                   rx(0*8+6 downto 0*8+0) = "1001010" or     -- CplD   10  01010  C
                   rx(0*8+6 downto 0*8+0) = "0001011" or     -- CplLk  00  01011  C
                   rx(0*8+6 downto 0*8+0) = "1001011") then  -- CplDLk 10  01011  C
              bufsel1 <= TPBUFTYPE_C;
            else                                     -- others            NP
              bufsel1 <= TPBUFTYPE_NP;
            end if;

            -- virtual channel
            for i in 0 to 5 loop
              if (rx(1*8+6 downto 1*8+4) = conv_std_logic_vector(i, 3)) then
                vc <= tc2vc(i);
              end if;
            end loop;           -- virtual channel

          end if;


          -- header and its write enable
          if (top1 = '1') then
            case NLANE is
              when 4 =>
                rxwr_bufsel <= '1';
                rxwr_header <= '1';
                header <= rx & rx1;

              when 8 =>
                rxwr_bufsel <= '1';
                rxwr_header <= '1';
                header <= rx1;

              when others =>
            end case;
          end if;

          -- payload_left
          case NLANE is
            when 4 => 
              if (rxtop = '1') then
                -- load payload size (in byte unit).
                payload_left <= rx(2*8+1 downto 2*8+0) & rx(3*8+7 downto 3*8+0) & "00";
              elsif (top1 = '1') then
                -- do nothing.
              elsif (top2 = '1') then
                -- do nothing.
              elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;

            when 8 => 
              if (rxtop = '1') then
                -- load payload size (in byte unit).
                payload_left <= rx(2*8+1 downto 2*8+0) & rx(3*8+7 downto 3*8+0) & "00";
              elsif (top1 = '1') then
                -- do nothing.
              elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;

            when others =>

          end case;

          --  data write enable
          if (have_data = '1') then
            case NLANE is
              when 4 => 
                if (top1 = '1') then
                  -- do nothing.
                elsif (top2 = '1') then
                  rxwr_data <= '1';
                elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                  rxwr_data <= '1';
                end if;

              when 8 => 
                if (top1 = '1') then
                  rxwr_data <= '1';
                elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                  rxwr_data <= '1';
                end if;

              when others =>

            end case;
          end if; -- have_data

          -- data
          case NLANE is
            when 4 => 
              if (short_header = '1') then
                data <= rx(3*8+7 downto 0*8+0) & rx1(7*8+7 downto 4*8+0);
              else
                data <= rx;
              end if;

            when 8 => 
              if (short_header = '1') then
                data <= rx(11*8+7 downto 0*8+0) & rx1(15*8+7 downto 12*8+0);
              else
                data <= rx;
              end if;

            when others => 

          end case;

        end if;
      end process;
    end block x4block;

    -- Traffic Class vs Virtual Channel mapping:
    tc2vc(0) <= "000";
    tc2vc(1) <= "000";
    tc2vc(2) <= "000";
    tc2vc(3) <= "000";
    tc2vc(4) <= "000";
    tc2vc(5) <= "000";
    tc2vc(6) <= "000";
    tc2vc(7) <= "000";

  end generate x4;
end default;


--
-- TLP transmitter
--   multiplex outgoing header/data from Tx{NP,P,C}{H,D}buf,
--   into TLPs and pass them to data-link layer.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity txctl is
  generic (
    NLANE             : integer;
    CL_PH_VC0_INIT    : integer;
    CL_NPH_VC0_INIT   : integer;
    CL_CH_VC0_INIT    : integer;
    CL_PD_VC0_INIT    : integer;
    CL_NPD_VC0_INIT   : integer;
    CL_CD_VC0_INIT    : integer
  );
  port (
    clk           : in  std_logic;
    rst           : in  std_logic;

    -- CL notified by data-link layer
    cl_header     : in  std_logic_vector( 7 downto 0);
    cl_data       : in  std_logic_vector(11 downto 0);
    inf_cl_header : in  std_logic;
    inf_cl_data   : in  std_logic;

    -- notify TxFCbuf CA to slvctl or msctl
    cat_nph       : out std_logic_vector(7 downto 0);
    cat_ph        : out std_logic_vector(7 downto 0);
    cat_ch        : out std_logic_vector(7 downto 0);
    cat_npd       : out std_logic_vector(11 downto 0);
    cat_pd        : out std_logic_vector(11 downto 0);
    cat_cd        : out std_logic_vector(11 downto 0);

    hbufsel        : out std_logic_vector(1 downto 0);
    dbufsel        : out std_logic_vector(1 downto 0);
    
    empty_bufsel  : in  std_logic;
    empty_header  : in  std_logic;
    empty_data    : in  std_logic;

    rd_bufsel     : out std_logic;
    rd_header     : out std_logic;
    rd_data       : out std_logic;

    q_bufsel      : in  std_logic_vector(1 downto 0);
    q_header      : in  std_logic_vector(127 downto 0);
    q_data        : in  std_logic_vector(NLANE*16-1 downto 0);

    txreq         : out std_logic;
    txack         : in  std_logic;
    txtop         : out std_logic;
    txwr          : out std_logic;
    txbe          : out std_logic_vector(NLANE*2-1 downto 0);
    tpsize        : out  std_logic_vector(OFFSET_WIDTH-1 downto 0);
    tx            : out std_logic_vector(NLANE*16-1 downto 0)
  );
end txctl;

#ifdef TXCTL_NOWAIT // low latency, severe timing constraint.

architecture default of txctl is

  type TXCTL_STATE is (
    IDLE,
    PARSE_HEADER,
    SEND_HEADERx1,
    SEND_HEADERx4,
    SEND_HEADERx8,
    SEND_DATA,
    WAITx1,
    INVALID_STATE
  );

  signal txsm         : TXCTL_STATE;

  -- credit consumed so far
  signal cc_nph       : std_logic_vector( 7 downto 0);
  signal cc_ph        : std_logic_vector( 7 downto 0);
  signal cc_ch        : std_logic_vector( 7 downto 0);
  signal cc_npd       : std_logic_vector(11 downto 0);
  signal cc_pd        : std_logic_vector(11 downto 0);
  signal cc_cd        : std_logic_vector(11 downto 0);

  signal cc_header    : std_logic_vector( 7 downto 0);
  signal cc_data      : std_logic_vector(11 downto 0);
  signal credit_inc   : std_logic_vector(11 downto 0);

  -- TxFCbuf credit allocated
  signal cat_nph0     : std_logic_vector(7 downto 0);
  signal cat_ph0      : std_logic_vector(7 downto 0);
  signal cat_ch0      : std_logic_vector(7 downto 0);
  signal cat_npd0     : std_logic_vector(11 downto 0);
  signal cat_pd0      : std_logic_vector(11 downto 0);
  signal cat_cd0      : std_logic_vector(11 downto 0);

  signal cr_header    : std_logic_vector(7 downto 0);
  signal cr_data      : std_logic_vector(11 downto 0);

  signal cl_header1     : std_logic_vector(7 downto 0);
  signal cl_data1       : std_logic_vector(11 downto 0);
  signal inf_cl_data1   : std_logic;
  signal inf_cl_header1 : std_logic;

  signal bufsel0      : std_logic_vector(1 downto 0);
  signal header       : std_logic_vector(127 downto 0);
  signal have_data    : std_logic;
  signal payload_left : std_logic_vector(11 downto 0);  -- in byte unit
  signal header_left  : std_logic_vector(3 downto 0);
  signal header_count : std_logic_vector(3 downto 0);
  signal short_header : std_logic;
  signal data1        : std_logic_vector(NLANE*16-1 downto 0);
  signal txacked      : std_logic;
  signal wcnt         : std_logic_vector(1 downto 0);
  
begin

  cat_ph     <= cat_ph0;
  cat_nph    <= cat_nph0;
  cat_ch     <= cat_ch0;
  cat_pd     <= cat_pd0;
  cat_npd    <= cat_npd0;
  cat_cd     <= cat_cd0;

  hbufsel     <= q_bufsel;
  dbufsel     <= bufsel0;

  -- chose credit consumed according to bufsel.
  process (bufsel0,
           cc_nph, cc_npd,
           cc_ph , cc_pd,
           cc_ch , cc_cd)
           
  begin
    case bufsel0 is

      when TPBUFTYPE_NP =>
        cc_header <= cc_nph;
        cc_data   <= cc_npd;

      when TPBUFTYPE_P =>
        cc_header <= cc_ph;
        cc_data   <= cc_pd;

      when TPBUFTYPE_C =>
        cc_header <= cc_ch;
        cc_data   <= cc_cd;

      when others =>
        cc_header <= (others => '0');
        cc_data   <= (others => '0');

    end case;
  end process;

  process (clk, rst)
    variable fmttmp          : std_logic_vector(1 downto 0);
    variable lengthtmp       : std_logic_vector(9 downto 0);
    variable cr_headertmp    : std_logic_vector(7 downto 0);
    variable cr_datatmp      : std_logic_vector(11 downto 0);
    variable header_lefttmp  : std_logic_vector(header_left'range);
    variable short_headertmp : std_logic;
    variable have_datatmp    : std_logic;
    variable payload_lefttmp : std_logic_vector(payload_left'range);

  begin
    if (rst = '1') then
      cc_nph <= conv_std_logic_vector(0, cc_nph'length);
      cc_ph  <= conv_std_logic_vector(0, cc_ph'length);
      cc_ch  <= conv_std_logic_vector(0, cc_ch'length);
      cc_npd <= conv_std_logic_vector(0, cc_npd'length);
      cc_pd  <= conv_std_logic_vector(0, cc_pd'length);
      cc_cd  <= conv_std_logic_vector(0, cc_cd'length);

      -- init TxFCbuf Credit Allocated.
      cat_ph0     <= conv_std_logic_vector(CL_PH_VC0_INIT, cat_ph0'length);
      cat_nph0    <= conv_std_logic_vector(CL_NPH_VC0_INIT, cat_nph0'length);
      cat_ch0     <= conv_std_logic_vector(CL_CH_VC0_INIT, cat_ch0'length);
      cat_pd0     <= conv_std_logic_vector(CL_PD_VC0_INIT, cat_pd0'length);
      cat_npd0    <= conv_std_logic_vector(CL_NPD_VC0_INIT, cat_npd0'length);
      cat_cd0     <= conv_std_logic_vector(CL_CD_VC0_INIT, cat_cd0'length);

      -- initial values:
      rd_bufsel  <= '0';
      rd_header  <= '0';
      rd_data    <= '0';
      txtop      <= '0';
      txwr       <= '0';
      txbe       <= (others => '1');
      wcnt       <= conv_std_logic_vector(0, wcnt'length);

      txsm <= IDLE;
    elsif (rising_edge(clk)) then

      -- default values:
      rd_bufsel  <= '0';
      rd_header  <= '0';
      rd_data    <= '0';
      txtop      <= '0';
      txwr       <= '0';
      txbe       <= (others => '1');

      -- delayed values:
      data1      <= q_data;
      cl_header1     <= cl_header;      -- CLs need to be latched just to satisfy timing constraint.
      inf_cl_header1 <= inf_cl_header;
      cl_data1       <= cl_data;
      inf_cl_data1   <= inf_cl_data;

      case txsm is

        when IDLE =>
          -- header field aliases:
          lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
          
          txreq <= '0';
          txacked <= '0';
          if (empty_bufsel = '0') then  -- something resides in the queue.
            bufsel0 <= q_bufsel;
            if (empty_header = '0') then
              header  <= q_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rd_bufsel  <= '1';
              rd_header  <= '1';

              credit_inc  <= length_to_credit(lengthtmp);
              txreq <= '1';
              txsm <= PARSE_HEADER;
            end if;
          end if;

        when PARSE_HEADER => 
          -- header field aliases:
          fmttmp                    := header( 0*8+6 downto  0*8+5);
          lengthtmp(9 downto 0)     := header(2*8+1 downto 2*8+0) & header(3*8+7 downto 3*8+0);

          -- header size
          if (fmttmp = TPFMT_12BH_NODATA or fmttmp = TPFMT_12BH_DATA) then
            short_headertmp := '1';
            short_header <= '1';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(12, header_left'length);
              header_left <= conv_std_logic_vector(12, header_left'length);
            end if;
          else
            short_headertmp := '0';
            short_header <= '0';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(16, header_left'length);
              header_left <= conv_std_logic_vector(16, header_left'length);
            end if;
          end if;

          -- credit requested for header
          cr_headertmp := unsigned(cc_header) + 1;
          cr_header <= cr_headertmp;
          
          -- credit requested for data
          if (fmttmp = TPFMT_12BH_DATA or fmttmp = TPFMT_16BH_DATA) then
            have_datatmp := '1';
            have_data <= '1';
            payload_lefttmp := lengthtmp & "00"; -- in byte unit.
            payload_left <= lengthtmp & "00"; -- in byte unit.
            cr_datatmp := cc_data + credit_inc;
            cr_data <= cr_datatmp;
          else
            have_datatmp := '0';
            have_data <= '0';
          end if;

          case fmttmp is
            when TPFMT_12BH_NODATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length);
            when TPFMT_16BH_NODATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length);
            when TPFMT_12BH_DATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length) + ("0" & lengthtmp & "00");
            when TPFMT_16BH_DATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length) + ("0" & lengthtmp & "00");
            when others =>
          end case;

          if (txack = '1' or txacked = '1') then
            txacked <= '1';
            txreq <= '0';
          end if;

          if (have_enough_credit(cl_header1, cr_headertmp) = '1' or inf_cl_header1 = '1') then  -- check credit for header.

            if (have_datatmp = '0' or -- header only TLP.
                (empty_data = '0' and  -- data is arrived in Tx buf.
                 (have_enough_credit(cl_data1, cr_datatmp) = '1' or inf_cl_data1 = '1'))) then  -- check credit for data.
              if (txack = '1' or txacked = '1') then   -- start tx when PHYMAC:Framer is ready.

                -- send the first word.
                case NLANE is

                  when 1 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(1*8+7 downto 0*8+0);
                    header_left <= unsigned(header_lefttmp) - NLANE*2;
                    header_count <= conv_std_logic_vector(NLANE*2, header_count'length);
                    txsm <= SEND_HEADERx1;

                  when 4 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(7*8+7 downto 0*8+0);
                    if (have_datatmp = '1' and short_headertmp = '1') then
                      rd_data <= '1';
                    end if;
                    txsm <= SEND_HEADERx4;

                  when 8 =>
                    if (have_datatmp = '0') then   -- TLP w/o data.
                      txtop <= '1';
                      txwr <= '1';
                      tx <= header(15*8+7 downto 0*8+0);
                      if (short_headertmp = '1') then
                        txbe <= x"0fff";
                      end if;
                      txsm <= IDLE;
                    else                        -- TLP w/data.
                      rd_data <= '1';
                      if (short_headertmp = '0') then
                        txtop <= '1';
                        txwr <= '1';
                        tx <= header(15*8+7 downto 0*8+0);
                        txsm <= SEND_DATA;
                      else
                        txwr <= '0';    -- don't transmit header right now.
                        tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 0*8+0);
                        payload_left <= unsigned(payload_lefttmp) - 4;  -- pack header & 4-byte data.
                        txsm <= SEND_HEADERx8;
                      end if;
                    end if;

                    if (have_datatmp = '0' or short_headertmp = '0') then
                      --
                      -- update header-credit consumed.
                      --
                      case bufsel0 is
                        when TPBUFTYPE_NP =>
                          cc_nph <= cr_headertmp;
                          cat_nph0 <= unsigned(cat_nph0) + 1;
                        when TPBUFTYPE_P =>
                          cc_ph <= cr_headertmp;
                          cat_ph0 <= unsigned(cat_ph0) + 1;
                        when TPBUFTYPE_C =>
                          cc_ch <= cr_headertmp;
                          cat_ch0 <= unsigned(cat_ch0) + 1;
                        when others =>
                      end case;
                    end if;

                  when others =>

                end case;

              end if;  -- acktp
            end if;  -- cl_data
          end if;  -- cl_header

        when SEND_HEADERx1 =>
          if (NLANE = 1) then
            txwr <= '1';
            -- output header byte : 2&3, 4&5, 6&7, 8&9, 10&11 (, 12&13, 14&15).
            -- 2*i runs through   : 2,   4,   6,   8,   10,      12,    14
            for i in 1 to 7 loop
              if (header_count = conv_std_logic_vector(2*i, header_left'length)) then
                tx <= header(16*(i+1)-1 downto 16*i);
              end if;
            end loop;
            header_count <= unsigned(header_count) + NLANE*2;

            if (header_left = conv_std_logic_vector(NLANE*2, header_left'length)) then
              if (have_data = '1') then
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                txsm <= WAITx1;
              end if;

              --
              -- update header-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_nph <= cr_header;
                  cat_nph0 <= unsigned(cat_nph0) + 1;
                when TPBUFTYPE_P =>
                  cc_ph <= cr_header;
                  cat_ph0 <= unsigned(cat_ph0) + 1;
                when TPBUFTYPE_C =>
                  cc_ch <= cr_header;
                  cat_ch0 <= unsigned(cat_ch0) + 1;
                when others =>
              end case;

            else
              header_left <= unsigned(header_left) - NLANE*2;
            end if;
          end if;

        when SEND_HEADERx4 =>
          if (NLANE = 4) then
            txwr <= '1';

            if (have_data = '0') then   -- TLP w/o data.
              tx <= header(15*8+7 downto 8*8+0);
              if (short_header = '1') then
                txbe <= x"0f";
              end if;
              txsm <= IDLE;
            else                        -- TLP w/data.
              if (short_header = '0') then
                -- transmit last half (8bytes) of the header, then go to SEND_DATA.
                -- rd_data is not asserted yet. assert it right now.
                tx <= header(15*8+7 downto 8*8+0);
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                -- pack 4-byte header & 4-byte data.
                tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 8*8+0);
                payload_left <= unsigned(payload_left) - 4;

                -- rd_data is already asserted in the previous state (PARSE_HEADER).
                -- 4-byte data is transmitted so far, and 8-byte can be transmitted
                -- in the next clock. rd_data should be deasserted now, if payload <= 8byte (=1word in the queue).

                if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                  -- all data (4bytes) are transmitted. quit.
                  txsm <= IDLE;
                  --
                  -- update data-credit consumed.
                  --
                  case bufsel0 is
                    when TPBUFTYPE_NP =>
                      cc_npd <= cr_data;
                      cat_npd0 <= cat_npd0 + credit_inc;
                    when TPBUFTYPE_P =>
                      cc_pd <= cr_data;
                      cat_pd0 <= cat_pd0 + credit_inc;
                    when TPBUFTYPE_C =>
                      cc_cd <= cr_data;
                      cat_cd0 <= cat_cd0 + credit_inc;
                    when others =>
                  end case;
                else
                  -- data yet resides in the queue. goto SEND_DATA.
                  if (payload_left <= conv_std_logic_vector(8, payload_left'length)) then
                    -- SEND_DATA state will be finished in a sigle clock.
                    -- rd_data should be deasserted now.
                    rd_data <= '0';
                  else
                    -- SEND_DATA state continues two or more clocks.
                    -- rd_data should be kept asserted.
                    rd_data <= '1';
                  end if;
                  txsm <= SEND_DATA;
                end if;
              end if;
            end if;

            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;

          end if;

        when SEND_HEADERx8 =>
          if (NLANE = 8) then
            txtop <= '1';
            txwr <= '1';  -- transmit header which was set in the previous state (PARSE_HEADER).

            if (payload_left = conv_std_logic_vector(0, payload_left'length)) then
              -- all data (4bytes) are transmitted. quit.
              txsm <= IDLE;
              --
              -- update data-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_npd <= cr_data;
                  cat_npd0 <= cat_npd0 + credit_inc;
                when TPBUFTYPE_P =>
                  cc_pd <= cr_data;
                  cat_pd0 <= cat_pd0 + credit_inc;
                when TPBUFTYPE_C =>
                  cc_cd <= cr_data;
                  cat_cd0 <= cat_cd0 + credit_inc;
                when others =>
              end case;
            else
              -- data yet resides in the queue. goto SEND_DATA.
              if (payload_left <= conv_std_logic_vector(12, payload_left'length)) then
                -- rd_data should be asserted only for 1clk since payload size is 16B or smaller.
                rd_data <= '0';
              else
                -- SEND_DATA state continues two or more clocks.
                -- rd_data should be kept asserted.
                rd_data <= '1';
              end if;
              txsm <= SEND_DATA;
            end if;

            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;

          end if;

        when SEND_DATA =>
          txwr <= '1';

          case NLANE is
            when 1 =>
              tx <= q_data; -- always transmit 2-byte data.

              if (payload_left > conv_std_logic_vector(2, payload_left'length)) then
                rd_data <= '1';
              end if;

            when 4 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(7*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;

              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"0f";
              end if;

              if (short_header = '1') then
                if (payload_left > conv_std_logic_vector(12, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                if (payload_left > conv_std_logic_vector(8, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;

            when 8 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(15*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;

              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(12, payload_left'length)) then
                txbe <= x"0fff";
              elsif (payload_left = conv_std_logic_vector(8, payload_left'length)) then
                txbe <= x"00ff";
              elsif (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"000f";
              end if;
              
              if (short_header = '1') then
                -- assert rd_data if payload_left > NLANE*2-4.
                if (payload_left > conv_std_logic_vector(28, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                -- assert rd_data if payload_left > NLANE.
                if (payload_left > conv_std_logic_vector(16, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;

            when others =>

          end case;

          -- exit to IDLE.
          if (payload_left <= conv_std_logic_vector(NLANE*2, payload_left'length)) then
            if (NLANE = 1) then
              txsm <= WAITx1;
            else
              txsm <= IDLE;
            end if;
            --
            -- update data-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_npd <= cr_data;
                cat_npd0 <= cat_npd0 + credit_inc;
              when TPBUFTYPE_P =>
                cc_pd <= cr_data;
                cat_pd0 <= cat_pd0 + credit_inc;
              when TPBUFTYPE_C =>
                cc_cd <= cr_data;
                cat_cd0 <= cat_cd0 + credit_inc;
              when others =>
            end case;
#if 1 // !!!!
            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            -- this is necessary only for x8 32-bit addressing TLP w/payload.
            if (NLANE = 8) then
              fmttmp := q_header( 0*8+6 downto  0*8+5);
              lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
              if (empty_bufsel = '0' and fmttmp = TPFMT_12BH_DATA) then
                txreq <= '0';
                txacked <= '0';
                bufsel0 <= q_bufsel;
                if (empty_header = '0') then
                  header  <= q_header;
                  -- bufsel & header is saved to a register.
                  -- fifo output is no longer necessary.
                  rd_bufsel  <= '1';
                  rd_header  <= '1';
                  credit_inc  <= length_to_credit(lengthtmp);
                  txreq <= '1';
                  txsm <= PARSE_HEADER;
                end if;
              end if;
            end if;
#endif
          else
            payload_left <= unsigned(payload_left) - NLANE*2;
          end if;

        when WAITx1 =>
          -- wait for 3clk.
          if (wcnt = conv_std_logic_vector(3-1, wcnt'length)) then
            wcnt <= conv_std_logic_vector(0, wcnt'length);
            txsm <= IDLE;
          else
            wcnt <= unsigned(wcnt) + 1;
          end if;

        when others =>
          txsm <= INVALID_STATE;

      end case;

    end if;

  end process;

end default;

#else // high latency, easy timing constraint.

architecture default of txctl is

  type TXCTL_STATE is (
    IDLE,
    PARSE_HEADER,
    WAIT_CREDITS,
    SEND_HEADERx1,
    SEND_HEADERx4,
    SEND_HEADERx8,
    SEND_DATA,
    WAITx1,
    INVALID_STATE
  );

  signal txsm         : TXCTL_STATE;

  -- credit consumed so far
  signal cc_nph       : std_logic_vector( 7 downto 0);
  signal cc_ph        : std_logic_vector( 7 downto 0);
  signal cc_ch        : std_logic_vector( 7 downto 0);
  signal cc_npd       : std_logic_vector(11 downto 0);
  signal cc_pd        : std_logic_vector(11 downto 0);
  signal cc_cd        : std_logic_vector(11 downto 0);

  signal cc_header    : std_logic_vector( 7 downto 0);
  signal cc_data      : std_logic_vector(11 downto 0);
  signal credit_inc   : std_logic_vector(11 downto 0);

  -- TxFCbuf credit allocated
  signal cat_nph0     : std_logic_vector(7 downto 0);
  signal cat_ph0      : std_logic_vector(7 downto 0);
  signal cat_ch0      : std_logic_vector(7 downto 0);
  signal cat_npd0     : std_logic_vector(11 downto 0);
  signal cat_pd0      : std_logic_vector(11 downto 0);
  signal cat_cd0      : std_logic_vector(11 downto 0);

  signal cr_header    : std_logic_vector(7 downto 0);
  signal cr_data      : std_logic_vector(11 downto 0);

  signal cl_header1     : std_logic_vector(7 downto 0);
  signal cl_data1       : std_logic_vector(11 downto 0);
  signal inf_cl_data1   : std_logic;
  signal inf_cl_header1 : std_logic;

  signal bufsel0      : std_logic_vector(1 downto 0);
  signal header       : std_logic_vector(127 downto 0);
  signal have_data    : std_logic;
  signal payload_left : std_logic_vector(11 downto 0);  -- in byte unit
  signal header_left  : std_logic_vector(3 downto 0);
  signal header_count : std_logic_vector(3 downto 0);
  signal short_header : std_logic;
  signal data1        : std_logic_vector(NLANE*16-1 downto 0);
  signal txacked      : std_logic;
  signal wcnt         : std_logic_vector(1 downto 0);
  
begin

  cat_ph     <= cat_ph0;
  cat_nph    <= cat_nph0;
  cat_ch     <= cat_ch0;
  cat_pd     <= cat_pd0;
  cat_npd    <= cat_npd0;
  cat_cd     <= cat_cd0;

  hbufsel     <= q_bufsel;
  dbufsel     <= bufsel0;

  -- chose credit consumed according to bufsel.
  process (bufsel0,
           cc_nph, cc_npd,
           cc_ph , cc_pd,
           cc_ch , cc_cd)
           
  begin
    case bufsel0 is

      when TPBUFTYPE_NP =>
        cc_header <= cc_nph;
        cc_data   <= cc_npd;

      when TPBUFTYPE_P =>
        cc_header <= cc_ph;
        cc_data   <= cc_pd;

      when TPBUFTYPE_C =>
        cc_header <= cc_ch;
        cc_data   <= cc_cd;

      when others =>
        cc_header <= (others => '0');
        cc_data   <= (others => '0');

    end case;
  end process;

  process (clk, rst)
    variable fmttmp          : std_logic_vector(1 downto 0);
    variable lengthtmp       : std_logic_vector(9 downto 0);
    variable cr_headertmp    : std_logic_vector(7 downto 0);
    variable cr_datatmp      : std_logic_vector(11 downto 0);
    variable header_lefttmp  : std_logic_vector(header_left'range);
    variable short_headertmp : std_logic;
    variable have_datatmp    : std_logic;
    variable payload_lefttmp : std_logic_vector(payload_left'range);

  begin
    if (rst = '1') then
      cc_nph <= conv_std_logic_vector(0, cc_nph'length);
      cc_ph  <= conv_std_logic_vector(0, cc_ph'length);
      cc_ch  <= conv_std_logic_vector(0, cc_ch'length);
      cc_npd <= conv_std_logic_vector(0, cc_npd'length);
      cc_pd  <= conv_std_logic_vector(0, cc_pd'length);
      cc_cd  <= conv_std_logic_vector(0, cc_cd'length);

      -- init TxFCbuf Credit Allocated.
      cat_ph0     <= conv_std_logic_vector(CL_PH_VC0_INIT, cat_ph0'length);
      cat_nph0    <= conv_std_logic_vector(CL_NPH_VC0_INIT, cat_nph0'length);
      cat_ch0     <= conv_std_logic_vector(CL_CH_VC0_INIT, cat_ch0'length);
      cat_pd0     <= conv_std_logic_vector(CL_PD_VC0_INIT, cat_pd0'length);
      cat_npd0    <= conv_std_logic_vector(CL_NPD_VC0_INIT, cat_npd0'length);
      cat_cd0     <= conv_std_logic_vector(CL_CD_VC0_INIT, cat_cd0'length);

      -- initial values:
      rd_bufsel  <= '0';
      rd_header  <= '0';
      rd_data    <= '0';
      txtop      <= '0';
      txwr       <= '0';
      txbe       <= (others => '1');
      wcnt       <= conv_std_logic_vector(0, wcnt'length);

      txsm <= IDLE;
    elsif (rising_edge(clk)) then

      -- default values:
      rd_bufsel  <= '0';
      rd_header  <= '0';
      rd_data    <= '0';
      txtop      <= '0';
      txwr       <= '0';
      txbe       <= (others => '1');

      -- delayed values:
      data1      <= q_data;
      cl_header1     <= cl_header;      -- CLs need to be latched just to satisfy timing constraint.
      inf_cl_header1 <= inf_cl_header;
      cl_data1       <= cl_data;
      inf_cl_data1   <= inf_cl_data;

      case txsm is

        when IDLE =>
          -- header field aliases:
          lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
          
          txreq <= '0';
          txacked <= '0';
          if (empty_bufsel = '0') then  -- something resides in the queue.
            bufsel0 <= q_bufsel;
            if (empty_header = '0') then
              header  <= q_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rd_bufsel  <= '1';
              rd_header  <= '1';

              credit_inc  <= length_to_credit(lengthtmp);
              txreq <= '1';
              txsm <= PARSE_HEADER;
            end if;
          end if;

        when PARSE_HEADER => 
          -- header field aliases:
          fmttmp                    := header( 0*8+6 downto  0*8+5);
          lengthtmp(9 downto 0)     := header(2*8+1 downto 2*8+0) & header(3*8+7 downto 3*8+0);

          -- header size
          if (fmttmp = TPFMT_12BH_NODATA or fmttmp = TPFMT_12BH_DATA) then
            short_headertmp := '1';
            short_header <= '1';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(12, header_left'length);
              header_left <= conv_std_logic_vector(12, header_left'length);
            end if;
          else
            short_headertmp := '0';
            short_header <= '0';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(16, header_left'length);
              header_left <= conv_std_logic_vector(16, header_left'length);
            end if;
          end if;

          -- credit requested for header
          cr_headertmp := unsigned(cc_header) + 1;
          cr_header <= cr_headertmp;
          
          -- credit requested for data
          if (fmttmp = TPFMT_12BH_DATA or fmttmp = TPFMT_16BH_DATA) then
            have_datatmp := '1';
            have_data <= '1';
            payload_lefttmp := lengthtmp & "00"; -- in byte unit.
            payload_left <= lengthtmp & "00"; -- in byte unit.
            cr_datatmp := cc_data + credit_inc;
            cr_data <= cr_datatmp;
          else
            have_datatmp := '0';
            have_data <= '0';
          end if;

          case fmttmp is
            when TPFMT_12BH_NODATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length);
            when TPFMT_16BH_NODATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length);
            when TPFMT_12BH_DATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length) + ("0" & lengthtmp & "00");
            when TPFMT_16BH_DATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length) + ("0" & lengthtmp & "00");
            when others =>
          end case;

          if (txack = '1' or txacked = '1') then
            txacked <= '1';
            txreq <= '0';
          end if;

          txsm <= WAIT_CREDITS;

        when WAIT_CREDITS =>
          if (txack = '1' or txacked = '1') then
            txacked <= '1';
            txreq <= '0';
          end if;

          if (have_enough_credit(cl_header1, cr_header) = '1' or inf_cl_header1 = '1') then  -- check credit for header.

            if (have_data = '0' or -- header only TLP.
                (empty_data = '0' and  -- data is arrived in Tx buf.
                 (have_enough_credit(cl_data1, cr_data) = '1' or inf_cl_data1 = '1'))) then  -- check credit for data.
              if (txack = '1' or txacked = '1') then   -- start tx when PHYMAC:Framer is ready.

                -- send the first word.
                case NLANE is

                  when 1 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(1*8+7 downto 0*8+0);
                    header_left <= unsigned(header_left) - NLANE*2;
                    header_count <= conv_std_logic_vector(NLANE*2, header_count'length);
                    txsm <= SEND_HEADERx1;

                  when 4 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(7*8+7 downto 0*8+0);
                    if (have_data = '1' and short_header = '1') then
                      rd_data <= '1';
                    end if;
                    txsm <= SEND_HEADERx4;

                  when 8 =>
                    if (have_data = '0') then   -- TLP w/o data.
                      txtop <= '1';
                      txwr <= '1';
                      tx <= header(15*8+7 downto 0*8+0);
                      if (short_header = '1') then
                        txbe <= x"0fff";
                      end if;
                      txsm <= IDLE;
                    else                        -- TLP w/data.
                      rd_data <= '1';
                      if (short_header = '0') then
                        txtop <= '1';
                        txwr <= '1';
                        tx <= header(15*8+7 downto 0*8+0);
                        txsm <= SEND_DATA;
                      else
                        txwr <= '0';    -- don't transmit header right now.
                        tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 0*8+0);
                        payload_left <= unsigned(payload_left) - 4;  -- pack header & 4-byte data.
                        txsm <= SEND_HEADERx8;
                      end if;
                    end if;

                    if (have_data = '0' or short_header = '0') then
                      --
                      -- update header-credit consumed.
                      --
                      case bufsel0 is
                        when TPBUFTYPE_NP =>
                          cc_nph <= cr_header;
                          cat_nph0 <= unsigned(cat_nph0) + 1;
                        when TPBUFTYPE_P =>
                          cc_ph <= cr_header;
                          cat_ph0 <= unsigned(cat_ph0) + 1;
                        when TPBUFTYPE_C =>
                          cc_ch <= cr_header;
                          cat_ch0 <= unsigned(cat_ch0) + 1;
                        when others =>
                      end case;
                    end if;

                  when others =>

                end case;

              end if;  -- acktp
            end if;  -- cl_data
          end if;  -- cl_header

        when SEND_HEADERx1 =>
          if (NLANE = 1) then
            txwr <= '1';
            -- output header byte : 2&3, 4&5, 6&7, 8&9, 10&11 (, 12&13, 14&15).
            -- 2*i runs through   : 2,   4,   6,   8,   10,      12,    14
            for i in 1 to 7 loop
              if (header_count = conv_std_logic_vector(2*i, header_left'length)) then
                tx <= header(16*(i+1)-1 downto 16*i);
              end if;
            end loop;
            header_count <= unsigned(header_count) + NLANE*2;

            if (header_left = conv_std_logic_vector(NLANE*2, header_left'length)) then
              if (have_data = '1') then
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                txsm <= WAITx1;
              end if;

              --
              -- update header-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_nph <= cr_header;
                  cat_nph0 <= unsigned(cat_nph0) + 1;
                when TPBUFTYPE_P =>
                  cc_ph <= cr_header;
                  cat_ph0 <= unsigned(cat_ph0) + 1;
                when TPBUFTYPE_C =>
                  cc_ch <= cr_header;
                  cat_ch0 <= unsigned(cat_ch0) + 1;
                when others =>
              end case;

            else
              header_left <= unsigned(header_left) - NLANE*2;
            end if;
          end if;

        when SEND_HEADERx4 =>
          if (NLANE = 4) then
            txwr <= '1';

            if (have_data = '0') then   -- TLP w/o data.
              tx <= header(15*8+7 downto 8*8+0);
              if (short_header = '1') then
                txbe <= x"0f";
              end if;
              txsm <= IDLE;
            else                        -- TLP w/data.
              if (short_header = '0') then
                -- transmit last half (8bytes) of the header, then go to SEND_DATA.
                -- rd_data is not asserted yet. assert it right now.
                tx <= header(15*8+7 downto 8*8+0);
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                -- pack 4-byte header & 4-byte data.
                tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 8*8+0);
                payload_left <= unsigned(payload_left) - 4;

                -- rd_data is already asserted in the previous state (PARSE_HEADER).
                -- 4-byte data is transmitted so far, and 8-byte can be transmitted
                -- in the next clock. rd_data should be deasserted now, if payload <= 8byte (=1word in the queue).

                if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                  -- all data (4bytes) are transmitted. quit.
                  txsm <= IDLE;
                  --
                  -- update data-credit consumed.
                  --
                  case bufsel0 is
                    when TPBUFTYPE_NP =>
                      cc_npd <= cr_data;
                      cat_npd0 <= cat_npd0 + credit_inc;
                    when TPBUFTYPE_P =>
                      cc_pd <= cr_data;
                      cat_pd0 <= cat_pd0 + credit_inc;
                    when TPBUFTYPE_C =>
                      cc_cd <= cr_data;
                      cat_cd0 <= cat_cd0 + credit_inc;
                    when others =>
                  end case;
                else
                  -- data yet resides in the queue. goto SEND_DATA.
                  if (payload_left <= conv_std_logic_vector(8, payload_left'length)) then
                    -- SEND_DATA state will be finished in a sigle clock.
                    -- rd_data should be deasserted now.
                    rd_data <= '0';
                  else
                    -- SEND_DATA state continues two or more clocks.
                    -- rd_data should be kept asserted.
                    rd_data <= '1';
                  end if;
                  txsm <= SEND_DATA;
                end if;
              end if;
            end if;

            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;

          end if;

        when SEND_HEADERx8 =>
          if (NLANE = 8) then
            txtop <= '1';
            txwr <= '1';  -- transmit header which was set in the previous state (PARSE_HEADER).

            if (payload_left = conv_std_logic_vector(0, payload_left'length)) then
              -- all data (4bytes) are transmitted. quit.
              txsm <= IDLE;
              --
              -- update data-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_npd <= cr_data;
                  cat_npd0 <= cat_npd0 + credit_inc;
                when TPBUFTYPE_P =>
                  cc_pd <= cr_data;
                  cat_pd0 <= cat_pd0 + credit_inc;
                when TPBUFTYPE_C =>
                  cc_cd <= cr_data;
                  cat_cd0 <= cat_cd0 + credit_inc;
                when others =>
              end case;
            else
              -- data yet resides in the queue. goto SEND_DATA.
              if (payload_left <= conv_std_logic_vector(12, payload_left'length)) then
                -- rd_data should be asserted only for 1clk since payload size is 16B or smaller.
                rd_data <= '0';
              else
                -- SEND_DATA state continues two or more clocks.
                -- rd_data should be kept asserted.
                rd_data <= '1';
              end if;
              txsm <= SEND_DATA;
            end if;

            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;

          end if;

        when SEND_DATA =>
          txwr <= '1';

          case NLANE is
            when 1 =>
              tx <= q_data; -- always transmit 2-byte data.

              if (payload_left > conv_std_logic_vector(2, payload_left'length)) then
                rd_data <= '1';
              end if;

            when 4 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(7*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;

              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"0f";
              end if;

              if (short_header = '1') then
                if (payload_left > conv_std_logic_vector(12, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                if (payload_left > conv_std_logic_vector(8, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;

            when 8 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(15*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;

              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(12, payload_left'length)) then
                txbe <= x"0fff";
              elsif (payload_left = conv_std_logic_vector(8, payload_left'length)) then
                txbe <= x"00ff";
              elsif (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"000f";
              end if;
              
              if (short_header = '1') then
                -- assert rd_data if payload_left > NLANE*2-4.
                if (payload_left > conv_std_logic_vector(28, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                -- assert rd_data if payload_left > NLANE.
                if (payload_left > conv_std_logic_vector(16, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;

            when others =>

          end case;

          -- exit to IDLE.
          if (payload_left <= conv_std_logic_vector(NLANE*2, payload_left'length)) then
            if (NLANE = 1) then
              txsm <= WAITx1;
            else
              txsm <= IDLE;
            end if;
            --
            -- update data-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_npd <= cr_data;
                cat_npd0 <= cat_npd0 + credit_inc;
              when TPBUFTYPE_P =>
                cc_pd <= cr_data;
                cat_pd0 <= cat_pd0 + credit_inc;
              when TPBUFTYPE_C =>
                cc_cd <= cr_data;
                cat_cd0 <= cat_cd0 + credit_inc;
              when others =>
            end case;
#if 1 // !!!!
            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            -- this is necessary only for x8 32-bit addressing TLP w/payload.
            if (NLANE = 8) then
              fmttmp := q_header( 0*8+6 downto  0*8+5);
              lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
              if (empty_bufsel = '0' and fmttmp = TPFMT_12BH_DATA) then
                txreq <= '0';
                txacked <= '0';
                bufsel0 <= q_bufsel;
                if (empty_header = '0') then
                  header  <= q_header;
                  -- bufsel & header is saved to a register.
                  -- fifo output is no longer necessary.
                  rd_bufsel  <= '1';
                  rd_header  <= '1';
                  credit_inc  <= length_to_credit(lengthtmp);
                  txreq <= '1';
                  txsm <= PARSE_HEADER;
                end if;
              end if;
            end if;
#endif
          else
            payload_left <= unsigned(payload_left) - NLANE*2;
          end if;

        when WAITx1 =>
          -- wait for 3clk.
          if (wcnt = conv_std_logic_vector(3-1, wcnt'length)) then
            wcnt <= conv_std_logic_vector(0, wcnt'length);
            txsm <= IDLE;
          else
            wcnt <= unsigned(wcnt) + 1;
          end if;

        when others =>
          txsm <= INVALID_STATE;

      end case;

    end if;

  end process;

end default;

#endif

--
-- PCIe Slvae (Target) Controller
--   . receive Posted and Non-posted TLPs from Rx FC bufs.
--   . communicate with application layer as master.
--   . transmit Completion TLPs to Tx FC bufs.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;

entity slvctl is
  generic (
    CA_PH_VC0_INIT    : integer;
    CA_NPH_VC0_INIT   : integer;
    CA_CH_VC0_INIT    : integer;
    CA_PD_VC0_INIT    : integer;
    CA_NPD_VC0_INIT   : integer;
    CA_CD_VC0_INIT    : integer;

    CFG_BAR0_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT     : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT  : std_logic_vector(31 downto 0);
    NLANE             : integer;
    NDMACH            : integer
  );
  port (
    clk               : in  std_logic;
    rst               : in  std_logic;   -- active high

    ca_nph            : out std_logic_vector(7 downto 0);
    ca_ph             : out std_logic_vector(7 downto 0);
    ca_ch             : out std_logic_vector(7 downto 0);
    ca_npd            : out std_logic_vector(11 downto 0);
    ca_pd             : out std_logic_vector(11 downto 0);
    ca_cd             : out std_logic_vector(11 downto 0);
    ca_p_flush        : out std_logic;
    ca_np_flush       : out std_logic;
    ca_c_flush        : out std_logic;

    mpls              : in  std_logic_vector(11 downto 0);  -- max payload size

    cfg_bar0          : in  std_logic_vector(31 downto 0);
    cfg_bar1          : in  std_logic_vector(31 downto 0);
    cfg_bar2          : in  std_logic_vector(31 downto 0);
    cfg_bar3          : in  std_logic_vector(31 downto 0);
    cfg_bar4          : in  std_logic_vector(31 downto 0);
    cfg_bar5          : in  std_logic_vector(31 downto 0);
    cfg_bar_rom       : in  std_logic_vector(31 downto 0);
    
    rxhbufsel         : out std_logic_vector(1 downto 0);
    rxdbufsel         : out std_logic_vector(1 downto 0);
    rxempty_bufsel    : in  std_logic;

    rxrd_bufsel       : out std_logic;
    rxrd_header       : out std_logic;
    rxrd_data         : out std_logic;

    rxq_bufsel        : in  std_logic_vector(1 downto 0);
    rxq_header        : in  std_logic_vector(127 downto 0);
    rxq_data          : in  std_logic_vector(NLANE*16-1 downto 0);

    txvc              : out std_logic_vector(2 downto 0);

    cat_header        : in  std_logic_vector(7 downto 0);
    cat_data          : in  std_logic_vector(11 downto 0);

    bdf               : out std_logic_vector(15 downto 0);  -- completer id

    txwr_req          : out std_logic;
    txwr_ack          : in  std_logic;
    txwr_bufsel       : out std_logic;
    txwr_header       : out std_logic;
    txwr_data         : out std_logic;

    tx_header         : out std_logic_vector(127 downto 0);
    tx_data           : out std_logic_vector(NLANE*16-1 downto 0);

    slv_read          : out std_logic;
    slv_write         : out std_logic;
    slv_bar           : out std_logic_vector(6 downto 0);
    slv_addr          : out std_logic_vector(63 downto 0);
    slv_bytevalid     : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout       : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain        : in  std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq       : out std_logic;
    slv_writereq      : out std_logic;
    slv_accept        : in  std_logic;
    slv_bytecount     : out std_logic_vector(11 downto 0);

    ms_read           : out std_logic;
    ms_rdchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_rdaddr         : out std_logic_vector(31 downto 0);
    ms_rddata         : out std_logic_vector(NLANE*16-1 downto 0);

    dmach_rdcpl       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdcpllength : out std_logic_vector(9 downto 0);
    dmach_rdcplbcnt   : out std_logic_vector(11 downto 0);
    dmach_rdcpladdr   : in  each32b(NDMACH-1 downto 0);

    cfg_read          : out std_logic;
    cfg_write         : out std_logic;
    cfg_addr          : out std_logic_vector(11 downto 0);
    cfg_be            : out std_logic_vector(3 downto 0);
    cfg_dataout       : out std_logic_vector(31 downto 0);
    cfg_datain        : in  std_logic_vector(31 downto 0)
  );

end slvctl;

architecture default of slvctl is

  type SLV_STATE is (
    IDLE,
    PARSE_HEADER,
    -- CfgRd0
    WAIT_CfgRd0,
    CPL_CfgRd0,
    CPL_CfgRd0_2,
    -- CfgWr0
    WAIT_CfgWr0,
    CPL_CfgWr0,
    CPL_CfgWr0_2,
    CPL_CfgWr0_3,
    -- MsgD
    BURST_MsgD,
    -- MemRd
    WAIT_MemRd,
    WAIT_MemRd_2,
    READY_MemRd,
    BURST_MemRd,
    -- MemWr
    WAIT_MemWr,
    BURST_MemWr,
    -- CplD
    WAIT_CplD,
    BURST_CplD,
    -- Non-Supported Request
    INVALID_STATE
  );

  signal slvsm           : SLV_STATE;
  
  -- credit allocated
  signal ca_nph0        : std_logic_vector(7 downto 0);
  signal ca_ph0         : std_logic_vector(7 downto 0);
  signal ca_ch0         : std_logic_vector(7 downto 0);
  signal ca_npd0        : std_logic_vector(11 downto 0);
  signal ca_pd0         : std_logic_vector(11 downto 0);
  signal ca_cd0         : std_logic_vector(11 downto 0);
  signal credit_inc     : std_logic_vector(11 downto 0);

  -- TxFCbuf credit consumed
  signal cct_ch         : std_logic_vector(7 downto 0);
  signal cct_cd         : std_logic_vector(11 downto 0);
  signal crt_header     : std_logic_vector(7 downto 0);
  signal crt_data       : std_logic_vector(11 downto 0);

  signal rxbufsel0      : std_logic_vector(1 downto 0);
  signal header         : std_logic_vector(127 downto 0);
  signal be_first       : std_logic_vector(3 downto 0);
  signal be_last        : std_logic_vector(3 downto 0);
  signal byte_ucnt      : std_logic_vector(11 downto 0);
  signal byte_dcnt      : std_logic_vector(11 downto 0);
  signal barsel         : std_logic_vector(6 downto 0);
  signal slv_addr0      : std_logic_vector(63 downto 0);
  signal bdf0           : std_logic_vector(15 downto 0);
  signal loweraddr      : std_logic_vector(6 downto 0);
  signal length         : std_logic_vector(9 downto 0);
  signal bytecount      : std_logic_vector(11 downto 0);
  signal bytecount_next : std_logic_vector(11 downto 0);
  signal dwcount        : std_logic_vector(9 downto 0);
  signal dwcount_next   : std_logic_vector(9 downto 0);
  signal cpllength      : std_logic_vector(9 downto 0);
  signal cplbytelength  : std_logic_vector(11 downto 0);
  signal data16         : std_logic_vector(15 downto 0);

  signal ms_rdaddr0     : std_logic_vector(31 downto 0);
  signal ms_rdchannel0  : std_logic_vector(NDMACH-1 downto 0);
  
begin

  rxhbufsel     <= rxq_bufsel;
  rxdbufsel     <= rxbufsel0;
  slv_addr      <= slv_addr0;
  slv_dataout   <= rxq_data;
  slv_bytecount <= byte_dcnt;
  ms_rddata     <= rxq_data;
  ms_rdaddr     <= ms_rdaddr0;
  ms_rdchannel  <= ms_rdchannel0;

  -- credit allocated
  ca_ph  <= ca_ph0;
  ca_nph <= ca_nph0;
  ca_ch  <= ca_ch0;
  ca_pd  <= ca_pd0;
  ca_npd <= ca_npd0;
  ca_cd  <= ca_cd0;

  bdf <= bdf0;

  -- Slave Controller State Machine
  process (clk, rst)

    variable typetmp          : std_logic_vector(6 downto 0);
    variable be_firsttmp      : std_logic_vector(3 downto 0);
    variable be_lasttmp       : std_logic_vector(3 downto 0);
    variable lengthtmp        : std_logic_vector(9 downto 0);
    variable cpllengthtmp     : std_logic_vector(9 downto 0);
    variable cplbytelengthtmp : std_logic_vector(11 downto 0);
    variable bdftmp           : std_logic_vector(15 downto 0);
    variable addrtmp32        : std_logic_vector(31 downto 0);
    variable addrtmp64        : std_logic_vector(63 downto 0);
    variable addrtmp          : std_logic_vector(6 downto 0);
    variable loweraddrtmp     : std_logic_vector(6 downto 0);
    variable bytecounttmp     : std_logic_vector(11 downto 0);
    variable crt_datatmp      : std_logic_vector(11 downto 0);
    variable byte_dcnttmp     : std_logic_vector(11 downto 0);
    variable tagtmp           : std_logic_vector(7 downto 0);
    variable bcnttmp          : std_logic_vector(11 downto 0);

  begin

    if (rst = '1') then

      -- init Flow Control Credit Allocated.
      ca_ph0    <= conv_std_logic_vector(CA_PH_VC0_INIT, ca_ph0'length);
      ca_nph0   <= conv_std_logic_vector(CA_NPH_VC0_INIT, ca_nph0'length);
      ca_ch0    <= conv_std_logic_vector(CA_CH_VC0_INIT, ca_ch0'length);
      ca_pd0    <= conv_std_logic_vector(CA_PD_VC0_INIT, ca_pd0'length);
      ca_npd0   <= conv_std_logic_vector(CA_NPD_VC0_INIT, ca_npd0'length);
      ca_cd0    <= conv_std_logic_vector(CA_CD_VC0_INIT, ca_cd0'length);

      -- clear TxFCbuf Credit Consumed.
      cct_ch     <= conv_std_logic_vector(0, cct_ch'length);
      cct_cd     <= conv_std_logic_vector(0, cct_cd'length);

      -- initial values:
      rxrd_bufsel   <= '0';
      rxrd_header   <= '0';
      rxrd_data     <= '0';
      txwr_bufsel   <= '0';
      txwr_header   <= '0';
      txwr_data     <= '0';

      slv_read      <= '0';
      slv_write     <= '0';
      slv_bytevalid <= (others => '0');
      cfg_read      <= '0';
      cfg_write     <= '0';
      ca_p_flush    <= '0';
      ca_np_flush   <= '0';
      ca_c_flush    <= '0';

      txwr_req      <= '0';
      slv_readreq   <= '0';
      slv_writereq  <= '0';

      ms_read       <= '0';

      slvsm <= IDLE;

    elsif (rising_edge(clk)) then

      -- header field aliases:
      typetmp                   := header( 0*8+6 downto  0*8+0);
      be_firsttmp(3 downto 0)   := header( 7*8+3 downto  7*8+0);
      be_lasttmp(3 downto 0)    := header( 7*8+7 downto  7*8+4);
      lengthtmp(9 downto 0)     := header(2*8+1 downto 2*8+0) & header(3*8+7 downto 3*8+0);
      bdftmp                    := header(9*8+7 downto 8*8+0);
      
      addrtmp32 := header( 8*8+7 downto  8*8+0) & header( 9*8+7 downto  9*8+0) &
                   header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+2) & "00";

      addrtmp64 := header( 8*8+7 downto  8*8+0) & header( 9*8+7 downto  9*8+0) &
                   header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+0) &
                   header(12*8+7 downto 12*8+0) & header(13*8+7 downto 13*8+0) &
                   header(14*8+7 downto 14*8+0) & header(15*8+7 downto 15*8+2) & "00";

      tagtmp    := header(10*8+7 downto 10*8+0);
      bcnttmp   := header( 6*8+3 downto  6*8+0) & header( 7*8+7 downto  7*8+0);
      loweraddrtmp := header(11*8+6 downto 11*8+0);

      -- default values:
      rxrd_bufsel   <= '0';
      rxrd_header   <= '0';
      rxrd_data     <= '0';
      txwr_bufsel   <= '0';
      txwr_header   <= '0';
      txwr_data     <= '0';
      slv_read      <= '0';
      slv_write     <= '0';
      cfg_read      <= '0';
      cfg_write     <= '0';
      ca_p_flush    <= '0';
      ca_np_flush   <= '0';
      ca_c_flush    <= '0';
      ms_read       <= '0';
      dmach_rdcpl   <= (others => '0');
      
      case slvsm is

        when IDLE =>

          if (rxempty_bufsel = '0') then  -- something arrived in the queue.
            rxbufsel0 <= rxq_bufsel;
            header    <= rxq_header;
            -- bufsel & header is saved to a register.
            -- fifo output is no longer necessary.
            rxrd_bufsel <= '1';
            rxrd_header <= '1';

            -- header is parsed in the next state, PARSE_HEADER.
            -- it is difficult to parse header here in this state
            -- due to timing restriction.
            slvsm <= PARSE_HEADER;
          end if;

        when PARSE_HEADER =>
          slv_bar   <= barsel;
          slv_addr0 <= mask_addr(barsel, header,
                                 CFG_BAR0_INIT, CFG_BAR1_INIT, CFG_BAR2_INIT,
                                 CFG_BAR3_INIT, CFG_BAR4_INIT, CFG_BAR5_INIT, CFG_BAR_ROM_INIT);
          credit_inc <= length_to_credit(lengthtmp);

          case rxbufsel0 is

            when TPBUFTYPE_NP =>
              ca_nph0    <= unsigned(ca_nph0) + 1;
              crt_header <= unsigned(cct_ch) + 1;

              case typetmp is

                when TPTYPE_MRd32 =>
                  addrtmp := addrtmp32(6 downto 0);

                  -- remaining data size in byte/DW
                  bytecounttmp := memrd_1st_cpl_bytecount(lengthtmp, be_firsttmp, be_lasttmp);
                  bytecount <= bytecounttmp;
                  dwcount <= lengthtmp;

                  -- lower address of the first completion
                  loweraddrtmp := memrd_1st_cpl_loweraddr(addrtmp, be_firsttmp);
                  loweraddr <= loweraddrtmp;

                  -- length of the first completion in byte/DW
                  cplbytelength <= memrd_1st_cpl_bytelength(lengthtmp, mpls, loweraddrtmp, bytecounttmp);
                  cpllength     <= memrd_1st_cpl_length(lengthtmp, mpls, loweraddrtmp);

                  txwr_req   <= '1';
                  txvc <= "000";
                  slv_readreq <= '1';
                  slvsm <= WAIT_MemRd;
                  
                when TPTYPE_MRd64 =>
                  addrtmp := addrtmp64(6 downto 0);

                when TPTYPE_MRdLk32 =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_MRdLk64 =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_IORd =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_IOWr =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_CfgRd0 =>
                  txwr_req   <= '1';
                  cfg_read <= '1';
                  cfg_addr <= "0000" & addrtmp32(7 downto 2) & "00";
                  txvc <= "000";
                  tx_header <= cfgrd0_cpl_header(header);
                  crt_data <= unsigned(cct_cd) + 1;

                  bdf0 <= bdftmp;  -- save my BDF (bus id, device id, function id)
                  -- for future use.
                  -- note that the saved value is byte reversed.

                  slvsm <= WAIT_CfgRd0;

                when TPTYPE_CfgWr0 =>
                  txwr_req   <= '1';
                  txvc <= "000";
                  tx_header <= cfgwr0_cpl_header(header);
                  bdf0 <= bdftmp;
                  slvsm <= WAIT_CfgWr0;

                when TPTYPE_CfgRd1 =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_CfgWr1 =>
                  slvsm <= INVALID_STATE;

                when others =>
                  slvsm <= INVALID_STATE;
              end case;

            when TPBUFTYPE_P =>
              ca_ph0    <= unsigned(ca_ph0) + 1;
              be_first  <= be_firsttmp;
              be_last   <= be_lasttmp;
              byte_dcnttmp := lengthtmp & "00";
              byte_dcnt <= byte_dcnttmp; 
              byte_ucnt <= conv_std_logic_vector(0, byte_ucnt'length);
              slv_bytevalid <= (others => '1');

              if (NLANE = 1) then
                slv_bytevalid <= be_firsttmp(1 downto 0);
              else
                -- byte-valid bits of the lowest 4 bytes are given by be_firsttmp.
                slv_bytevalid(3 downto 0) <= be_firsttmp(3 downto 0);
              end if;

              -- byte-enable of the last NLANE*2 byte.
              if (byte_dcnttmp <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
                case NLANE is
                  when 4 =>
                    case byte_dcnttmp is
                      when conv_std_logic_vector(8, byte_dcnt'length) =>
                        slv_bytevalid <= be_lasttmp & be_firsttmp;
                      when conv_std_logic_vector(4, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & be_firsttmp;
                      when others =>
                    end case;

                  when 8 =>
                    case byte_dcnttmp is
                      when conv_std_logic_vector(16, byte_dcnt'length) =>
                        slv_bytevalid <= be_lasttmp & "1111" & "1111" & be_firsttmp;
                      when conv_std_logic_vector(12, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & be_lasttmp & "1111" & be_firsttmp;
                      when conv_std_logic_vector(8, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & "0000" & be_lasttmp & be_firsttmp;
                      when conv_std_logic_vector(4, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & "0000" & "0000" & be_firsttmp;
                      when others =>
                    end case;

                  when others =>

                end case;
              end if;


              case typetmp is

                when TPTYPE_MWr32 | TPTYPE_MWr64 =>
                  slv_writereq <= '1';
                  slvsm <= WAIT_MemWr;

                when others =>
--                  msg_write <= '1';

                  if (typetmp(6 downto 3) = "0110") then  -- TPTYPE_Msg
                    -- !!! this part should be rewritten.    !!!
                    -- !!! all messages are ignored for now. !!!
                    slvsm <= IDLE;

                  elsif (typetmp(6 downto 3) = "1110") then  -- TPTYPE_MsgD
                    rxrd_data <= '1';
                    slvsm <= BURST_MsgD;
                  else
                    slvsm <= INVALID_STATE;
                  end if;

              end case;

            when TPBUFTYPE_C =>
              ca_ch0 <= unsigned(ca_ch0) + 1;

              case typetmp is

                when TPTYPE_Cpl =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_CplD =>
                  for i in 0 to NDMACH-1 loop
                    if (tagtmp = conv_std_logic_vector(i, tagtmp'length)) then
                      ms_rdaddr0     <= dmach_rdcpladdr(i);
                      ms_rdchannel0  <= (i => '1', others => '0');
                      dmach_rdcpl(i) <= '1';
                    end if;
                  end loop;
                  dmach_rdcpllength <= lengthtmp;
                  dmach_rdcplbcnt   <= bcnttmp;
                  byte_dcnt <= lengthtmp & "00";
                  byte_ucnt <= conv_std_logic_vector(0, byte_ucnt'length);
                  slvsm <= WAIT_CplD;

                when TPTYPE_CplLk =>
                  slvsm <= INVALID_STATE;

                when TPTYPE_CplDLk =>
                  slvsm <= INVALID_STATE;

                when others =>
                  slvsm <= INVALID_STATE;

              end case;

            when others =>
              slvsm <= INVALID_STATE;

          end case; -- end of PARSE_HEADER state


        --
        -- Type0 Configuration Read
        --
        when WAIT_CfgRd0 =>
          -- wait until Tx buffers are ready.
          if (txwr_ack = '1' and                                    -- got write permission to Tx buf.
              have_enough_credit(cat_header, crt_header) = '1' and  -- Tx buf have enough room.
              have_enough_credit(cat_data, crt_data) = '1') then
            cct_ch     <= crt_header;
            cct_cd     <= crt_data;
            txwr_bufsel <= '1';
            txwr_header <= '1';
            slvsm <= CPL_CfgRd0;
          end if;

        when CPL_CfgRd0 =>
          ca_np_flush   <= '1';
          txwr_data   <= '1';
          if (NLANE = 1) then
            tx_data <= cfg_datain(15 downto 0);
            slvsm <= CPL_CfgRd0_2;
          else
            tx_data(NLANE*16-1 downto 32) <= (others => '0');
            tx_data(31 downto 0)          <= cfg_datain;
            txwr_req <= '0';
            slvsm <= IDLE;
          end if;

        -- x1 only
        when CPL_CfgRd0_2 =>
          if (NLANE = 1) then
            txwr_data   <= '1';
            tx_data <= cfg_datain(31 downto 16);
            txwr_req <= '0';
            slvsm <= IDLE;
          end if;

        --
        -- Type0 Configuration Write
        --
        when WAIT_CfgWr0 =>
          -- wait until Tx buffers are ready.
          if (txwr_ack = '1' and
              have_enough_credit(cat_header, crt_header) = '1') then
            cct_ch <= crt_header;
            txwr_bufsel <= '1';
            txwr_header <= '1';
            rxrd_data <= '1';
            ca_npd0 <= ca_npd0 + credit_inc;
            slvsm <= CPL_CfgWr0;
          end if;

        when CPL_CfgWr0 =>
          ca_np_flush   <= '1';
          cfg_addr <= "0000" & header(11*8+7 downto 11*8+2) & "00";
          cfg_be <= header( 7*8+3 downto  7*8+0);
          if (NLANE = 1) then
            rxrd_data <= '1';
            data16 <= rxq_data;
            slvsm <= CPL_CfgWr0_2;
          else
            cfg_write <= '1';
            cfg_dataout <= rxq_data(31 downto 0);
            slvsm <= CPL_CfgWr0_3;
          end if;

        -- x1 only
        when CPL_CfgWr0_2 =>
          if (NLANE = 1) then
            cfg_write <= '1';
            cfg_dataout <= rxq_data & data16;
            slvsm <= CPL_CfgWr0_3;
          end if;

        -- wait for the written configuration register to be updated.
        when CPL_CfgWr0_3 =>
          txwr_req <= '0';
          slvsm <= IDLE;
          
        --
        -- Memory Read
        --
        when WAIT_MemRd =>
          byte_dcnt      <= cplbytelength;
          bytecount_next <= bytecount - cplbytelength;
          dwcount_next   <= dwcount - cpllength;
          tx_header <= memrd_cpl_header(header, bdf0, cpllength, bytecount, loweraddr);

          crt_data <= cct_cd + length_to_credit(cpllength);
          ca_np_flush <= '1';

          slvsm <= WAIT_MemRd_2;

        when WAIT_MemRd_2 =>
          -- wait until Tx buffers are ready.
          if (
              slv_accept = '1' and      -- !!!
              txwr_ack = '1' and
              have_enough_credit(cat_header, crt_header) = '1' and
              have_enough_credit(cat_data, crt_data) = '1') then
            txwr_bufsel <= '1';
            txwr_header <= '1';
            cct_ch     <= crt_header;
            cct_cd     <= crt_data;
            slv_readreq <= '0';
            slv_read  <= '1';
            slvsm <= READY_MemRd;
          end if;

        when READY_MemRd =>
          -- keep slv_read asserted if burst continues longer than 1clk.
          if (byte_dcnt > conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            slv_read  <= '1';
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;
          end if;
          txwr_req <= '0';
          slvsm <= BURST_MemRd;

        when BURST_MemRd =>
          txwr_data <= '1';
          tx_data   <= slv_datain;
          byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;

          -- deassert slv_addr 2-clk earlier than end of burst.
          if (byte_dcnt > conv_std_logic_vector(NLANE*2*2, byte_dcnt'length)) then
            slv_read  <= '1';
          end if;

          if (byte_dcnt > conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then  -- read burst in progress.
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;

          else  -- about to finish current completion.

            -- some more data to be sent. prepare for the next completion.
            if (dwcount_next /= conv_std_logic_vector(0, dwcount_next'length)) then
              bytecount <= bytecount_next;
              dwcount   <= dwcount_next;
              loweraddr <= (others => '0');

              -- next completion is not the final one.
              -- cpllength is set to max payload size.
              if (mpls(11 downto 2) < dwcount_next) then
                cpllengthtmp      := mpls(11 downto 2);
                cpllength         <= cpllengthtmp;
                cplbytelengthtmp  := mpls;
                cplbytelength     <= cplbytelengthtmp;

                dwcount_next   <= dwcount_next - mpls(11 downto 2);
                bytecount_next <= bytecount_next - mpls;

              -- next completion is the final one.
              -- cpllength is set to remaining data size.
              else
                cpllengthtmp      := dwcount_next;
                cpllength         <= cpllengthtmp;
                cplbytelengthtmp  := bytecount_next;
                cplbytelength     <= cplbytelengthtmp;
              end if;

              slv_readreq <= '1';
              slvsm <= WAIT_MemRd;

            else

--              txwr_req <= '0';        -- !!!
              slvsm <= IDLE;

            end if;
          end if;


        --
        -- Memory Write
        --
        when WAIT_MemWr =>
          if (slv_accept = '1') then
            -- flush CA at the beginning of a burst.  flush is
            -- necessary only when RxFCbuf is almost full.  but
            -- for simplicity, flush is asserted every time a
            -- TLP is retrieved from RxFCbuf.
--            ca_p_flush   <= '1';      -- !!!
            rxrd_data    <= '1';
            slv_writereq <= '0';
            slv_write    <= '1';
            slvsm <= BURST_MemWr;
          end if;

        when BURST_MemWr =>
          slv_bytevalid <= (others => '1');
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_pd0       <= ca_pd0 + credit_inc;
            ca_p_flush   <= '1';        -- !!!
            slvsm <= IDLE;

            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            if (rxempty_bufsel = '0') then  -- something arrived in the queue.
              rxbufsel0 <= rxq_bufsel;
              header    <= rxq_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rxrd_bufsel <= '1';
              rxrd_header <= '1';

              -- header is parsed in the next state, PARSE_HEADER.
              -- it is difficult to parse header here in this state
              -- due to timing restriction.
              slvsm <= PARSE_HEADER;
            end if;

          else
            rxrd_data  <= '1';
            slv_write <= '1';
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;
            byte_ucnt <= unsigned(byte_ucnt) + NLANE*2;
            byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          end if;

          -- byte-enable of the first DW for x1.
          if (byte_ucnt = conv_std_logic_vector(0, byte_ucnt'length) and NLANE = 1) then
            slv_bytevalid <= be_first(3 downto 2);
          end if;

          -- byte-enable of the last NLANE*2 byte for x1.
          if (byte_dcnt = conv_std_logic_vector(6, byte_dcnt'length) and NLANE = 1) then
            slv_bytevalid <= be_last(1 downto 0);
          end if;

          if (byte_dcnt <= conv_std_logic_vector((NLANE*2)*2, byte_dcnt'length)) then
            case NLANE is
              when 1 =>
                if (byte_ucnt >= conv_std_logic_vector(4, byte_ucnt'length)) then
                  slv_bytevalid <= be_last(3 downto 2);
                end if;

              when 4 =>
                case byte_dcnt is
                  when conv_std_logic_vector(16, byte_dcnt'length) =>
                    slv_bytevalid <= be_last & "1111";
                  when conv_std_logic_vector(12, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & be_last;
                  when others =>
                end case;

              when 8 =>
                case byte_dcnt is
                  when conv_std_logic_vector(32, byte_dcnt'length) =>
                    slv_bytevalid <= be_last & "1111" & "1111" & "1111";
                  when conv_std_logic_vector(28, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & be_last & "1111" & "1111";
                  when conv_std_logic_vector(24, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & "0000" & be_last & "1111";
                  when conv_std_logic_vector(20, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & "0000" & "0000" & be_last;
                  when others =>
                end case;

              when others =>

            end case;
          end if;

        --
        -- Completion with Data
        --
        when WAIT_CplD =>
--          ca_c_flush   <= '1';        -- !!!
          rxrd_data    <= '1';
          ms_read      <= '1';
          slvsm <= BURST_CplD;

        when BURST_CplD =>
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_cd0       <= ca_cd0 + credit_inc;  -- !!!
            ca_c_flush   <= '1';
            slvsm <= IDLE;

            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            if (rxempty_bufsel = '0') then  -- something arrived in the queue.
              rxbufsel0 <= rxq_bufsel;
              header    <= rxq_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rxrd_bufsel <= '1';
              rxrd_header <= '1';

              -- header is parsed in the next state, PARSE_HEADER.
              -- it is difficult to parse header here in this state
              -- due to timing restriction.
              slvsm <= PARSE_HEADER;
            end if;

          else
            rxrd_data  <= '1';
            ms_read    <= '1';
            ms_rdaddr0 <= unsigned(ms_rdaddr0) + NLANE*2;
            byte_ucnt  <= unsigned(byte_ucnt)  + NLANE*2;
            byte_dcnt  <= unsigned(byte_dcnt)  - NLANE*2;
          end if;

        when BURST_MsgD =>
          -- !!! this part should be rewritten.    !!!
          -- !!! all messages are ignored for now. !!!
          -- just to remove data from rxq_npd.
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_pd0    <= ca_pd0 + credit_inc;
            slvsm <= IDLE;
          else
            rxrd_data  <= '1';
            byte_ucnt <= unsigned(byte_ucnt) + NLANE*2;
            byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          end if;
 
        when others =>
          slvsm <= INVALID_STATE;

      end case;
    end if;
  end process;

  --
  -- Address Decoder
  -- decode address by comparing it with values stored in base address registers.
  --

  process(clk)

    variable addrtmp32  : std_logic_vector(31 downto 0);
    variable addrtmp64  : std_logic_vector(63 downto 0);
    variable mask32     : std_logic_vector(31 downto 0);
    variable mask64     : std_logic_vector(63 downto 0);

  begin

    if (rising_edge(clk)) then

      -- alias for address fields in rxq_header
      addrtmp32 := rxq_header( 8*8+7 downto  8*8+0) & rxq_header( 9*8+7 downto  9*8+0) &
                   rxq_header(10*8+7 downto 10*8+0) & rxq_header(11*8+7 downto 11*8+2) & "00";

      addrtmp64 := rxq_header( 8*8+7 downto  8*8+0) & rxq_header( 9*8+7 downto  9*8+0) &
                   rxq_header(10*8+7 downto 10*8+0) & rxq_header(11*8+7 downto 11*8+0) &
                   rxq_header(12*8+7 downto 12*8+0) & rxq_header(13*8+7 downto 13*8+0) &
                   rxq_header(14*8+7 downto 14*8+0) & rxq_header(15*8+7 downto 15*8+2) & "00";

      barsel         <= (others => '0');

      -- BAR0 & BAR1, 64-bit address
      if (CFG_BAR0_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR1_INIT(31 downto 0) & CFG_BAR0_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar1 & cfg_bar0) and mask64)) then
          barsel(0) <= '1';
        end if;
      else
        -- BAR0, 32-bit address
        if (CFG_BAR0_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then  -- bar0 is in use.
          mask32 := CFG_BAR0_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar0 and mask32)) then
            barsel(0) <= '1';
          end if;
        end if;
        -- BAR1, 32-bit address
        if (CFG_BAR1_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR1_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar1 and mask32)) then
            barsel(1) <= '1';
          end if;
        end if;
      end if;

      -- BAR2 & BAR3, 64-bit address
      if (CFG_BAR2_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR3_INIT(31 downto 0) & CFG_BAR2_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar3 & cfg_bar2) and mask64)) then
          barsel(2) <= '1';
        end if;
      else
        -- BAR2, 32-bit address
        if (CFG_BAR2_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR2_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar2 and mask32)) then
            barsel(2) <= '1';
          end if;
        end if;
        -- BAR3, 32-bit address
        if (CFG_BAR3_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR3_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar3 and mask32)) then
            barsel(3) <= '1';
          end if;
        end if;
      end if;

      -- BAR4 & BAR5, 64-bit address
      if (CFG_BAR4_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR5_INIT(31 downto 0) & CFG_BAR4_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar5 & cfg_bar4) and mask64)) then
          barsel(4) <= '1';
        end if;
      else
        -- BAR4, 32-bit address
        if (CFG_BAR4_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR4_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar4 and mask32)) then
            barsel(4) <= '1';
          end if;
        end if;
        -- BAR5, 32-bit address
        if (CFG_BAR5_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR5_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar5 and mask32)) then
            barsel(5) <= '1';
          end if;
        end if;
      end if;

      -- BAR ROM
      if (CFG_BAR_ROM_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := CFG_BAR_ROM_INIT(31 downto 11) & "00000000001";
        if ((addrtmp32(31 downto 0) and mask32) = (cfg_bar_rom and mask32)) then
          barsel(6) <= '1';
        end if;
      end if;

    end if;  -- end of rising_edge(clk)

  end process;

end default;

--
-- functions used in msctl and slvctl
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

package msslvpkg is

  -- completion header generators:
  function cfgrd0_cpl_header(
    rhdr    : in std_logic_vector  -- header of the requester.
  )
  return std_logic_vector;

  function cfgwr0_cpl_header(
    rhdr    : in std_logic_vector  -- header of the requester.
  )
  return std_logic_vector;

  function memrd_1st_cpl_loweraddr(
    addr        : in std_logic_vector; -- byte address of the request.
    first_be    : in std_logic_vector  -- byte enable of the 1st DW of the request.
  )
  return std_logic_vector;

  function memrd_1st_cpl_bytecount(
    length     : in std_logic_vector; -- requested data size in DW unit.
    first_be   : in std_logic_vector; -- byte enable of the 1st DW of the request.
    last_be    : in std_logic_vector  -- byte enable of the last DW of the request.
  )
  return std_logic_vector;

  function memrd_cpl_header(
    rhdr      : in std_logic_vector(127 downto 0); -- header of the requester.
    bdf       : in std_logic_vector(15 downto 0);  -- completer id (bus id, device id, function id)
    len       : in std_logic_vector( 9 downto 0);
    bcnt      : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector( 6 downto 0)
  )
  return std_logic_vector;

  function memrd_1st_cpl_bytelength(
    len       : in std_logic_vector(9 downto 0);
    mpls      : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0);
    bytecount : in std_logic_vector(11 downto 0)
  )
  return std_logic_vector;

  function memrd_1st_cpl_length(
    len    : in std_logic_vector(9 downto 0);
    mpls   : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector;

  function mask_addr(
    barsel : in std_logic_vector(6 downto 0);
    header : in std_logic_vector(127 downto 0);
    bar0   : in std_logic_vector(31 downto 0);
    bar1   : in std_logic_vector(31 downto 0);
    bar2   : in std_logic_vector(31 downto 0);
    bar3   : in std_logic_vector(31 downto 0);
    bar4   : in std_logic_vector(31 downto 0);
    bar5   : in std_logic_vector(31 downto 0);
    barrom : in std_logic_vector(31 downto 0)
  )
  return std_logic_vector;

  function memwr32_header(
    len       : in std_logic_vector(9 downto 0);
    addr      : in std_logic_vector(31 downto 0);
    tag       : in std_logic_vector(7 downto 0);
    bdf       : in std_logic_vector(15 downto 0);
    tc        : in std_logic_vector(2 downto 0);
    attr      : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;

  function memrd32_header(
    len       : in std_logic_vector(9 downto 0);
    addr      : in std_logic_vector(31 downto 0);
    tag       : in std_logic_vector(7 downto 0);
    bdf       : in std_logic_vector(15 downto 0);
    tc        : in std_logic_vector(2 downto 0);
    attr      : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;

end package;

package body msslvpkg is

  -- CfgRd0 Completion header generator
  function cfgrd0_cpl_header(
    rhdr    : in std_logic_vector(127 downto 0)  -- header of the requester.
  )
  return std_logic_vector is

    variable chdr : std_logic_vector(127 downto 0);  -- header of the completer.
    variable len : std_logic_vector(9 downto 0);
    variable bcnt : std_logic_vector(11 downto 0);

  begin

    len  := conv_std_logic_vector(1, len'length);  -- 1DW length
    bcnt := conv_std_logic_vector(4, bcnt'length);  -- 4bytes to read

    chdr :=
      conv_std_logic_vector(0, 32) &                     -- byte12-15 not used
      x"00" &                                            -- byte11    lower address should be 0 for CfgWr/Rd
      rhdr(6*8+7 downto 6*8+0) &                         -- byte10    tag
      rhdr(5*8+7 downto 4*8+0) &                         -- byte8,9   requester id
      bcnt(7 downto 0) &                                 -- byte7     byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) &             -- byte6     cpl stat, bcm, byte count higher 4bits.
      rhdr(9*8+7 downto 8*8+0) &                         -- byte4,5   completer id
      len(7 downto 0) & len(9 downto 8) &                -- byte3     length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" &           -- byte2     td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" &          -- byte1     tc
      '0' & TPTYPE_CplD;                                 -- byte0     fmt,type

    return chdr;

  end cfgrd0_cpl_header;

  -- CfgWr0 Completion header generator
  function cfgwr0_cpl_header(
    rhdr    : in std_logic_vector(127 downto 0)  -- header of the requester.
  )
  return std_logic_vector is

    variable chdr : std_logic_vector(127 downto 0);  -- header of the completer.
    variable len : std_logic_vector(9 downto 0);
    variable bcnt : std_logic_vector(11 downto 0);

  begin

    len  := conv_std_logic_vector(0, len'length);  -- 0DW length
    bcnt := conv_std_logic_vector(4, bcnt'length);  -- 4bytes to write

    chdr :=
      conv_std_logic_vector(0, 32) &                     -- byte12-15 not used
      x"00" &                                            -- byte11    lower address should be 0 for CfgWr/Rd
      rhdr(6*8+7 downto 6*8+0) &                         -- byte10    tag
      rhdr(5*8+7 downto 4*8+0) &                         -- byte8,9   requester id
      bcnt(7 downto 0) &                                 -- byte7     byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) &             -- byte6     cpl stat, bcm, byte count higher 4bits.
      rhdr(9*8+7 downto 8*8+0) &                         -- byte4,5   completer id
      len(7 downto 0) & len(9 downto 8) &                -- byte3     length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" &           -- byte2     td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" &          -- byte1     tc
      '0' & TPTYPE_Cpl;                                  -- byte0     fmt,type

    return chdr;

  end cfgwr0_cpl_header;

  function memrd_1st_cpl_loweraddr(
    addr        : in std_logic_vector(6 downto 0); -- lower 7 bits of byte address of the request.
    first_be    : in std_logic_vector  -- byte enable of the 1st DW of the request.
  )
  return std_logic_vector is

    variable loweraddr : std_logic_vector(6 downto 0);

  begin

    -- higher 5 bits
    loweraddr(6 downto 2) := addr(6 downto 2);

    -- lower 2 bits
    if (first_be = "0000") then
      loweraddr(1 downto 0) := "00";
    elsif (first_be(0) = '1') then
      loweraddr(1 downto 0) := "00";
    elsif (first_be(1) = '1') then
      loweraddr(1 downto 0) := "01";
    elsif (first_be(2) = '1') then
      loweraddr(1 downto 0) := "10";
    else
      loweraddr(1 downto 0) := "11";
    end if;

    return loweraddr;

  end memrd_1st_cpl_loweraddr;

  function memrd_1st_cpl_bytecount(
    length     : in std_logic_vector; -- requested data size in DW unit.
    first_be   : in std_logic_vector; -- byte enable of the 1st DW of the request.
    last_be    : in std_logic_vector  -- byte enable of the last DW of the request.
  )
  return std_logic_vector is

    variable bytecount : std_logic_vector(11 downto 0);

  begin
    if (last_be = "0000") then          -- 1DW or zero length request
      if (first_be(3) = '1' and first_be(0) = '1') then
        bytecount := conv_std_logic_vector(4, bytecount'length);
      elsif (first_be(3 downto 2) = "01" and first_be(0) = '1') then
        bytecount := conv_std_logic_vector(3, bytecount'length);
      elsif (first_be(3) = '1' and first_be(1 downto 0) = "10") then
        bytecount := conv_std_logic_vector(3, bytecount'length);
      elsif (first_be = "0011") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      elsif (first_be = "0110") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      elsif (first_be = "1100") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      else
        bytecount := conv_std_logic_vector(1, bytecount'length);
      end if;
    else                                -- multi-DW length request
      if (first_be(0) = '1') then
        if (last_be(3) = '1') then
          bytecount := length & "00";
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(1, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        end if;
      elsif (first_be(1) = '1') then
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(1, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        end if;
      elsif (first_be(2) = '1') then
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(5, bytecount'length);
        end if;
      else
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(5, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(6, bytecount'length);
        end if;
      end if;
    end if;

    return bytecount;

  end memrd_1st_cpl_bytecount;

  -- MemRd Completion header generator
  function memrd_cpl_header(
    rhdr      : in std_logic_vector(127 downto 0); -- header of the requester.
    bdf       : in std_logic_vector(15 downto 0);  -- completer id (bus id, device id, function id)
    len       : in std_logic_vector(9 downto 0);
    bcnt      : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector is

    variable chdr : std_logic_vector(127 downto 0);  -- header of the completer.

  begin

    chdr :=
      conv_std_logic_vector(0, 32) &                     -- byte12-15 not used
      '0' & loweraddr(6 downto 0) &                      -- byte11    lower address
      rhdr(6*8+7 downto 6*8+0) &                         -- byte10    tag
      rhdr(5*8+7 downto 4*8+0) &                         -- byte8,9   requester id
      bcnt(7 downto 0) &                                 -- byte7     byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) &             -- byte6     cpl stat, bcm, byte count higher 4bits.
      bdf &                                              -- byte4,5   completer id
      len(7 downto 0) & len(9 downto 8) &                -- byte3     length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" &           -- byte2     td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" &          -- byte1     tc
      '0' & TPTYPE_CplD;                                 -- byte0     fmt,type

    return chdr;

  end memrd_cpl_header;

  function memrd_1st_cpl_bytelength(
    len       : in std_logic_vector(9 downto 0);
    mpls      : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0);
    bytecount : in std_logic_vector(11 downto 0)
  )
  return std_logic_vector is

    variable blen : std_logic_vector(11 downto 0);

  begin
    
    if (len < mpls(11 downto 2)) then
      blen := bytecount;
    else
      blen := mpls - ("00000" & loweraddr);
    end if;

    return blen;

  end memrd_1st_cpl_bytelength;


  function memrd_1st_cpl_length(
    len    : in std_logic_vector(9 downto 0);
    mpls   : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector is

    variable dwlen : std_logic_vector(9 downto 0);

  begin

    if (len < mpls(11 downto 2)) then
      dwlen := len;
    else
      dwlen := mpls(11 downto 2) - ("00000" & loweraddr(6 downto 2));
    end if;

    return dwlen;

  end memrd_1st_cpl_length;

  function mask_addr(
    barsel : in std_logic_vector(6 downto 0);
    header : in std_logic_vector(127 downto 0);
    bar0   : in std_logic_vector(31 downto 0);
    bar1   : in std_logic_vector(31 downto 0);
    bar2   : in std_logic_vector(31 downto 0);
    bar3   : in std_logic_vector(31 downto 0);
    bar4   : in std_logic_vector(31 downto 0);
    bar5   : in std_logic_vector(31 downto 0);
    barrom : in std_logic_vector(31 downto 0)
  )
  return std_logic_vector is

    variable maddr      : std_logic_vector(63 downto 0);
    variable addrtmp32  : std_logic_vector(31 downto 0);
    variable addrtmp64  : std_logic_vector(63 downto 0);
    variable mask32     : std_logic_vector(31 downto 0);
    variable mask64     : std_logic_vector(63 downto 0);

  begin

    -- alias for address fields in rxq_header
    addrtmp32 := header( 8*8+7 downto  8*8+0) & header( 9*8+7 downto  9*8+0) &
                 header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+2) & "00";

    addrtmp64 := header( 8*8+7 downto  8*8+0) & header( 9*8+7 downto  9*8+0) &
                 header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+0) &
                 header(12*8+7 downto 12*8+0) & header(13*8+7 downto 13*8+0) &
                 header(14*8+7 downto 14*8+0) & header(15*8+7 downto 15*8+2) & "00";

    -- BAR0 & BAR1, 64-bit address
    if (bar0(2 downto 1) = "10") then
      mask64 := bar1(31 downto 0) & bar0(31 downto 7) & "0000000";
      if (barsel(0) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR0, 32-bit address
      if (bar0(31 downto 0) /= conv_std_logic_vector(0, 32)) then  -- bar0 is in use.
        mask32 := bar0(31 downto 7) & "0000000";
        if (barsel(0) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR1, 32-bit address
      if (bar1(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar1(31 downto 7) & "0000000";
        if (barsel(1) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;

    -- BAR2 & BAR3, 64-bit address
    if (bar2(2 downto 1) = "10") then
      mask64 := bar3(31 downto 0) & bar2(31 downto 7) & "0000000";
      if (barsel(2) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR2, 32-bit address
      if (bar2(31 downto 0) /= conv_std_logic_vector(0, 32)) then  -- bar0 is in use.
        mask32 := bar2(31 downto 7) & "0000000";
        if (barsel(2) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR3, 32-bit address
      if (bar3(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar3(31 downto 7) & "0000000";
        if (barsel(3) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;

    -- BAR4 & BAR5, 64-bit address
    if (bar4(2 downto 1) = "10") then
      mask64 := bar5(31 downto 0) & bar4(31 downto 7) & "0000000";
      if (barsel(4) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR4, 32-bit address
      if (bar4(31 downto 0) /= conv_std_logic_vector(0, 32)) then  -- bar0 is in use.
        mask32 := bar4(31 downto 7) & "0000000";
        if (barsel(4) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR5, 32-bit address
      if (bar5(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar5(31 downto 7) & "0000000";
        if (barsel(5) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;

    -- BAR ROM, 32-bit address
    if (barrom(31 downto 0) /= conv_std_logic_vector(0, 32)) then
      mask32 := barrom(31 downto 7) & "0000000";
      if (barsel(6) = '1') then
        maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
      end if;
    end if;

    return maddr;

  end mask_addr;

  -- MemWr32 header generator
  function memwr32_header(
    len       : in std_logic_vector(9 downto 0);
    addr      : in std_logic_vector(31 downto 0);
    tag       : in std_logic_vector(7 downto 0);
    bdf       : in std_logic_vector(15 downto 0);  -- completer id (bus id, device id, function id)
    tc        : in std_logic_vector(2 downto 0);
    attr      : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector is

    variable hdr       : std_logic_vector(127 downto 0);  -- header of the completer.
    variable rvsdatmp32 : std_logic_vector(31 downto 0);

  begin

    -- reverse byte order.
    rvsdatmp32 := addr(0*8+7 downto 0*8+0) & addr(1*8+7 downto 1*8+0) &
                  addr(2*8+7 downto 2*8+0) & addr(3*8+7 downto 3*8+0);

    hdr :=
      conv_std_logic_vector(0, 32) &                     -- byte12-15 not used
      rvsdatmp32(31 downto 0) &                          -- byte8-11  address
      x"ff" &                                            -- byte7     last BE, first BE
      tag &                                              -- byte6     tag
      bdf &                                              -- byte4,5   completer id
      len(7 downto 0) &                                  -- byte3     length
      "00" & attr & "00" & len(9 downto 8) &             -- byte2     no td, no ep, attr
      '0' & tc & "0000" &                                -- byte1     tc
      '0' & TPTYPE_MWr32;                                -- byte0     fmt,type

    return hdr;

  end memwr32_header;

  -- MemRd32 header generator
  function memrd32_header(
    len       : in std_logic_vector(9 downto 0);
    addr      : in std_logic_vector(31 downto 0);
    tag       : in std_logic_vector(7 downto 0);
    bdf       : in std_logic_vector(15 downto 0);  -- completer id (bus id, device id, function id)
    tc        : in std_logic_vector(2 downto 0);
    attr      : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector is

    variable hdr       : std_logic_vector(127 downto 0);  -- header of the completer.
    variable rvsdatmp32 : std_logic_vector(31 downto 0);

  begin

    -- reverse byte order.
    rvsdatmp32 := addr(0*8+7 downto 0*8+0) & addr(1*8+7 downto 1*8+0) &
                  addr(2*8+7 downto 2*8+0) & addr(3*8+7 downto 3*8+0);

    hdr :=
      conv_std_logic_vector(0, 32) &                     -- byte12-15 not used
      rvsdatmp32(31 downto 0) &                          -- byte8-11  address
      x"ff" &                                            -- byte7     last BE, first BE
      tag &                                              -- byte6     tag
      bdf &                                              -- byte4,5   completer id
      len(7 downto 0) &                                  -- byte3     length
      "00" & attr & "00" & len(9 downto 8) &             -- byte2     no td, no ep, attr
      '0' & tc & "0000" &                                -- byte1     tc
      '0' & TPTYPE_MRd32;                                -- byte0     fmt,type

    return hdr;

  end memrd32_header;

end msslvpkg;

--
-- Flow Control Buffer
--
--   . Single clock FIFO, show ahead.
--   . Read/write operation is valid only when 'vc' is
--     matched with 'VIRTUAL_CHANNEL'.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;

entity fcbuf is 
  generic (
    VIRTUAL_CHANNEL : integer;
    DATA_WIDTH      : integer;
    DATA_DEPTH      : integer
  );
  port (
    clk         : in  std_logic;
    rst         : in  std_logic;     -- active high reset

    vc          : in  std_logic_vector(2 downto 0);
    wr          : in  std_logic;
    rd          : in  std_logic;
    data        : in  std_logic_vector(DATA_WIDTH-1 downto 0);

    empty       : out std_logic;
    full        : out std_logic;
    usedw       : out std_logic_vector(8 downto 0);
    q           : out std_logic_vector(DATA_WIDTH-1 downto 0)
  );
end fcbuf;

architecture default of fcbuf is

  component scfifo
    generic (
      LPM_WIDTH               : natural := DATA_WIDTH;
      LPM_NUMWORDS            : natural := DATA_DEPTH;
      LPM_SHOWAHEAD           : string  := "ON";
      ADD_RAM_OUTPUT_REGISTER : string  := "ON";
      OVERFLOW_CHECKING       : string  := "OFF";
      UNDERFLOW_CHECKING      : string  := "OFF";
      USE_EAB                 : string  := "ON";
      INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
      LPM_TYPE                : string  := "scfifo"
    );
    port (
      rdreq    : in  std_logic;
      empty    : out std_logic;
      full     : out std_logic;
      clock    : in  std_logic;
      sclr     : in  std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
      wrreq    : in  std_logic;
      data     : in  std_logic_vector(LPM_WIDTH-1 downto 0)
    );
  end component;

  signal we          : std_logic;
  signal full0       : std_logic;
  signal fullout     : std_logic;
  
begin

    buf0 : scfifo
      generic map (
        LPM_NUMWORDS => DATA_DEPTH,
        LPM_WIDTH    => data'length
      )
      port map (
        clock => clk,
        sclr  => rst,

        wrreq => we,
        rdreq => rd,
        data  => data,

        empty => empty,
        full  => full0,
        q     => q
      );

    we <= '1' when wr = '1' and vc = conv_std_logic_vector(VIRTUAL_CHANNEL, vc'length)
              else '0';

#if 0 // for debugging

    full <= fullout;
    process (clk, rst)
    begin
      if (rst = '1') then
        fullout <= '0';
      elsif (rising_edge(clk)) then
        if (full0 = '1') then
          fullout <= '1';
        end if;
      end if;
    end process;

#else

   full <= full0;

#endif

end default;

--
-- Tx FC Buffer Input Arbiter
--   . receives write requests from slvctl:c_req, msctl:np_req, and msctl:p_req.
--   . returns acknowledge to one of them with first come first served policy.
--   . c_req has the highest priority and p_req has the 2nd.
--   
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;

entity txarb is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;   -- active high

    c_req     : in  std_logic;
    c_ack     : out std_logic;

    p_req     : in  std_logic;
    p_ack     : out std_logic;

    np_req    : in  std_logic;
    np_ack    : out std_logic;

    srcsel    : out std_logic_vector(1 downto 0)  --  TPBUFTYPE_P, _NP, or _C
  );
end txarb;


#if 1 // asynchronous version. faster, but requires severe timing constrain.

architecture default of txarb is

  signal c_ack0     : std_logic;
  signal p_ack0     : std_logic;
  signal np_ack0    : std_logic;

begin

  c_ack  <= c_ack0;
  p_ack  <= p_ack0;
  np_ack <= np_ack0;
  
  process(c_req, np_req, p_req)
  begin
    if (c_req = '1') then
      if (c_ack0 = '1') then            -- already locked.
        c_ack0 <= '1';
      elsif (np_ack0 = '0' and p_ack0 = '0') then  -- no one locked.
        c_ack0 <= '1';
      end if;
    else
      c_ack0 <= '0';
    end if;

    if (p_req = '1') then
      if (p_ack0 = '1') then            -- already locked.
        p_ack0 <= '1';
      elsif (c_ack0 = '0' and np_ack0 = '0') then  -- no one locked.
        if (c_req = '0') then           -- no pending request from slvctl:c_req.
          p_ack0 <= '1';
        else
          p_ack0 <= '0';
        end if;
      end if;
    else
      p_ack0 <= '0';
    end if;

    if (np_req = '1') then
      if (np_ack0 = '1') then           -- already locked.
        np_ack0 <= '1';
      elsif (c_ack0 = '0' and p_ack0 = '0') then  -- no one locked.
        if (c_req = '0' and p_req = '0') then  -- no pending request from slvctl:c_req nor msctl:p_req.
          np_ack0 <= '1';
        else
          np_ack0 <= '0';
        end if;
      end if;
    else
      np_ack0 <= '0';
    end if;

  end process;

  process(p_ack0, np_ack0)
  begin
    srcsel <= TPBUFTYPE_C;
    if (p_ack0 = '1') then
      srcsel <= TPBUFTYPE_P;
    elsif (np_ack0 = '1') then
      srcsel <= TPBUFTYPE_NP;
    end if;
  end process;

end default;

#else // synchronous version.

architecture default of txarb is
  type TXARB_STATE is (
    IDLE,
    C_LOCK,
    P_LOCK,
    NP_LOCK
  );

  signal txarbsm : TXARB_STATE;

begin

  process(clk, rst)
  begin
    if (rst = '1') then
      c_ack  <= '0';
      p_ack  <= '0';
      np_ack <= '0';
      txarbsm <= IDLE;
    elsif (rising_edge(clk)) then
      case txarbsm is
        when IDLE =>
          if (c_req = '1') then
            srcsel <= TPBUFTYPE_C;
            c_ack <= '1';
            txarbsm <= C_LOCK;
          elsif (p_req = '1') then
            srcsel <= TPBUFTYPE_P;
            p_ack  <= '1';
            txarbsm <= P_LOCK;
          elsif (np_req = '1') then
            srcsel <= TPBUFTYPE_NP;
            np_ack  <= '1';
            txarbsm <= NP_LOCK;
          end if;

        when C_LOCK =>
          if (c_req = '0') then
            c_ack <= '0';
            txarbsm <= IDLE;
          end if;

        when P_LOCK =>
          if (p_req = '0') then
            p_ack <= '0';
            txarbsm <= IDLE;
          end if;

        when NP_LOCK =>
          if (np_req = '0') then
            np_ack <= '0';
            txarbsm <= IDLE;
          end if;

        when others =>

      end case;
    end if;
  end process;

end default;

#endif

--
-- PCIe Master (Initiator) Controller
--   . transmit Posted and Non-posted TLPs to Tx FC bufs.
--   . communicate with application layer (DMA manager) as slave.
--   . receive Completion TLPs from Rx FC bufs.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;

entity msctl is
  generic (
    NLANE             : integer;
    NDMACH            : integer
  );
  port (
    clk               : in  std_logic;
    rst               : in  std_logic;   -- active high

    cat_ph            : in  std_logic_vector(7 downto 0);
    cat_pd            : in  std_logic_vector(11 downto 0);
    cat_nph           : in  std_logic_vector(7 downto 0);
    cat_npd           : in  std_logic_vector(11 downto 0);
    bdf               : in  std_logic_vector(15 downto 0);  -- completer id

    -- interface to TxFCbuf and txarb
    p_txwr_req        : out std_logic;
    p_txwr_ack        : in  std_logic;
    p_txwr_bufsel     : out std_logic;
    txwr_ph           : out std_logic;
    tx_ph             : out std_logic_vector(127 downto 0);
    txwr_pd           : out std_logic;
    tx_pd             : out std_logic_vector(NLANE*16-1 downto 0);

    np_txwr_req       : out std_logic;
    np_txwr_ack       : in  std_logic;
    np_txwr_bufsel    : out std_logic;
    txwr_nph          : out std_logic;
    tx_nph            : out std_logic_vector(127 downto 0);
    txwr_npd          : out std_logic;
    tx_npd            : out std_logic_vector(NLANE*16-1 downto 0);

    -- interface to master controllers in the user-side logic.
    ms_write          : out std_logic;
    ms_wrchannel      : out std_logic_vector(NDMACH-1 downto 0);
    ms_wraddr         : out std_logic_vector(31 downto 0);
    ms_wrdata         : in  std_logic_vector(NLANE*16-1 downto 0);

    -- interface to DMA controllers
    dma_fifocnt_off   : out std_logic_vector(4 downto 0);  -- width >= log2(NLANE*2)
    dmach_wrreq       : in  std_logic_vector(NDMACH-1 downto 0);
    dmach_wrack       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdreq       : in  std_logic_vector(NDMACH-1 downto 0);
    dmach_rdack       : out std_logic_vector(NDMACH-1 downto 0);
    dmach_length      : in  each10b(NDMACH-1 downto 0);
    dmach_paddr       : in  each64b(NDMACH-1 downto 0);
    dmach_laddr       : in  each32b(NDMACH-1 downto 0)
  );
end msctl;

architecture default of msctl is

  type WR_STATE is (
    IDLE,
    WAIT_MemWr,
    READY_MemWr,
    BURST_MemWr
  );

  type RD_STATE is (
    IDLE,
    WAIT_MemRd
  );

  signal wrsm  : WR_STATE;
  signal rdsm  : RD_STATE;

  -- TxFCbuf credit consumed
  signal cct_nph        : std_logic_vector(7 downto 0);
  signal cct_ph         : std_logic_vector(7 downto 0);
  signal cct_npd        : std_logic_vector(11 downto 0);
  signal cct_pd         : std_logic_vector(11 downto 0);

  signal wr             : std_logic;
  signal wrladdr        : std_logic_vector(31 downto 0);
  signal wrpaddr        : std_logic_vector(63 downto 0);
  signal wrchannel      : std_logic_vector(NDMACH-1 downto 0);
  signal wrack          : std_logic_vector(NDMACH-1 downto 0);
  signal wrcnt          : std_logic_vector(PAYLOAD_WIDTH+2-1 downto 0); -- size in byte.

  signal rd             : std_logic;
  signal rdpaddr        : std_logic_vector(63 downto 0);
  signal rdack          : std_logic_vector(NDMACH-1 downto 0);
  signal rdcnt          : std_logic_vector(PAYLOAD_WIDTH+2-1 downto 0);

  signal wrtag, rdtag   : std_logic_vector(7 downto 0);
  signal crt_ph         : std_logic_vector(7 downto 0);
  signal crt_pd         : std_logic_vector(11 downto 0);

begin

  ms_wrchannel <= wrchannel;
  dmach_wrack  <= wrack;
  dmach_rdack  <= rdack;
  ms_write     <= wr;
  ms_wraddr    <= wrladdr;
  tx_pd        <= ms_wrdata;
  tx_npd       <= (others => '0');      -- not used (reserved for non-posted data).

  --
  -- DMA Write State Machine
  --   transmit MemWr TLP on request from a DMA channel.
  --   Tx Posted Buffer is inter locked by the channel.
  --   No other channel can start a new transfer until the one completes.
  --
  process(clk, rst)
    variable wrreqtmp    : std_logic;
    variable wrcnttmp    : std_logic_vector(wrcnt'range); -- size in byte.
    variable cat_pdtmp   : std_logic_vector(11 downto 0);

  begin

    if (rst = '1') then
      cct_ph     <= conv_std_logic_vector(0, cct_ph'length);
      cct_pd     <= conv_std_logic_vector(0, cct_pd'length);
      p_txwr_req <= '0';
      wrchannel  <= (others => '0');
      wrack      <= (others => '0');
      wrsm       <= IDLE;

    elsif (rising_edge(clk)) then
      wr              <= '0';
      p_txwr_bufsel   <= '0';
      txwr_ph         <= '0';
      txwr_pd         <= '0';
      dma_fifocnt_off <= (others => '0');

      case wrsm is
        when IDLE =>
          -- accept write request from the youngest channel.
          wrchannel <= (others => '0');
          wrack <= (others => '0');
          wrreqtmp := '0';
          for i in 0 to NDMACH-1 loop
            if (dmach_wrreq(i) = '1' and wrreqtmp = '0') then
              wrreqtmp     := '1';
              wrack(i)     <= '1';      -- acknowledge.
              wrchannel(i) <= '1';      -- alloc DMA channel.
              wrladdr      <= dmach_laddr(i);
              wrpaddr      <= dmach_paddr(i);

              wrcnttmp     := dmach_length(i) & "00";  -- size in byte.
              wrcnt        <= wrcnttmp;
              crt_ph       <= unsigned(cct_ph) + 1;
              crt_pd       <= cct_pd + length_to_credit(wrcnttmp(wrcnt'left downto 2));

              wrtag          <= conv_std_logic_vector(i, wrtag'length);
              p_txwr_req   <= '1';      -- write request to TxFCbuf.
              wrsm         <= WAIT_MemWr;
            end if;
          end loop;

        when WAIT_MemWr =>
          if (p_txwr_ack = '1' and
              have_enough_credit(cat_ph, crt_ph) = '1' and
              have_enough_credit(cat_pd, crt_pd) = '1') then

            tx_ph <= memwr32_header(wrcnt(wrcnt'left downto 2),
                                    wrpaddr(31 downto 0),
                                    wrtag, bdf, "000", "00");
            cct_ph        <= crt_ph;
            cct_pd        <= crt_pd;
            wr            <= '1';
            p_txwr_bufsel <= '1';
            txwr_ph       <= '1';
            wrsm          <= READY_MemWr;
          end if;

        when READY_MemWr =>
          -- keep wr asserted if burst continues longer than 1clk.
          if (wrcnt > conv_std_logic_vector(NLANE*2, wrcnt'length)) then
            wr      <= '1';
            wrladdr <= unsigned(wrladdr) + NLANE*2;
          end if;
          txwr_pd    <= '1';
          p_txwr_req <= '0';          -- !!!
          wrsm       <= BURST_MemWr;

        when BURST_MemWr =>
#if DMACNT_NOWAIT
          -- deassert write acknowledge 4 or 3-clk earlier than end of burst
          -- (for better perormance).
          if (NLANE >= 4) then
            if (wrcnt = conv_std_logic_vector(NLANE*2*4, wrcnt'length)) then
              wrack <= (others => '0');
              dma_fifocnt_off <= conv_std_logic_vector(NLANE*2*1, dma_fifocnt_off'length);
            end if;
          end if;

          if (wrcnt <= conv_std_logic_vector(NLANE*2*3, wrcnt'length)) then
            wrack <= (others => '0');
            dma_fifocnt_off <= conv_std_logic_vector(0, dma_fifocnt_off'length);
          end if;
#else
          -- deassert write acknowledge 4 or 3-clk earlier than end of burst
          -- (for better perormance).
--          if (NLANE >= 4) then
--            if (wrcnt = conv_std_logic_vector(NLANE*2*4, wrcnt'length)) then
--              wrack <= (others => '0');
--              dma_fifocnt_off <= conv_std_logic_vector(NLANE*2*2, dma_fifocnt_off'length);
--            end if;
--            if (wrcnt = conv_std_logic_vector(NLANE*2*3, wrcnt'length)) then
--              wrack <= (others => '0');
--              dma_fifocnt_off <= conv_std_logic_vector(NLANE*2*1, dma_fifocnt_off'length);
--            end if;
--          end if;

          if (wrcnt <= conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wrack <= (others => '0');
            dma_fifocnt_off <= conv_std_logic_vector(0, dma_fifocnt_off'length);
          end if;
#endif

          -- deassert wr 2-clk earlier than end of burst.
          -- release write channel at the same clk.
          if (wrcnt > conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wr      <= '1';
            wrladdr <= unsigned(wrladdr) + NLANE*2;
          end if;
          if (wrcnt <= conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wrchannel <= (others => '0');
          end if;

          if (wrcnt > conv_std_logic_vector(NLANE*2, wrcnt'length)) then
            -- proceed burst.
            -- receive wdata from user-side logic and push it into TxFCbuf.
            wrcnt  <= unsigned(wrcnt) - NLANE*2;
            txwr_pd <= '1';
          else
            -- burst going to end.
            wrsm <= IDLE;

            if (NLANE >= 4) then        -- !!!
              -- skip IDLE state and go directly to WAIT_MemWr if possible.
              wrchannel <= (others => '0');
              wrack <= (others => '0');
              wrreqtmp := '0';
              for i in 0 to NDMACH-1 loop
                if (dmach_wrreq(i) = '1' and wrreqtmp = '0') then
                  wrreqtmp     := '1';
                  wrack(i)     <= '1';      -- acknowledge.
                  wrchannel(i) <= '1';      -- alloc DMA channel.
                  wrladdr      <= dmach_laddr(i);
                  wrpaddr      <= dmach_paddr(i);

                  wrcnttmp     := dmach_length(i) & "00";  -- size in byte.
                  wrcnt        <= wrcnttmp;
                  crt_ph       <= unsigned(cct_ph) + 1;
                  crt_pd       <= cct_pd + length_to_credit(wrcnttmp(wrcnt'left downto 2));

                  wrtag          <= conv_std_logic_vector(i, wrtag'length);
                  p_txwr_req   <= '1';      -- write request to TxFCbuf.
                  wrsm         <= WAIT_MemWr;
                end if;
              end loop;
              if (wrreqtmp = '1') then
                wrsm <= WAIT_MemWr;
              end if;
            end if;

          end if;

        when others =>

      end case;

    end if;
  end process;

  --
  -- DMA Read State Machine
  --   transmit MemRd TLP on request from a DMA channel.
  --
  process(clk, rst)
    variable rdreqtmp      : std_logic;
    variable crt_nphtmp : std_logic_vector(7 downto 0);

  begin

    if (rst = '1') then
      cct_nph      <= conv_std_logic_vector(0, cct_nph'length);
      cct_npd      <= conv_std_logic_vector(0, cct_npd'length);
      np_txwr_req  <= '0';
      rdack        <= (others => '0');
      rdsm         <= IDLE;

    elsif (rising_edge(clk)) then
      np_txwr_bufsel <= '0';
      txwr_nph <= '0';
      txwr_npd <= '0';

      case rdsm is
        when IDLE =>
          np_txwr_req    <= '0';        -- !!!
          -- accept read request from the youngest channel.
          rdack <= (others => '0');
          rdreqtmp := '0';
          for i in 0 to NDMACH-1 loop
            if (dmach_rdreq(i) = '1' and rdreqtmp = '0') then
              rdreqtmp     := '1';
              rdack(i)     <= '1';
              rdpaddr      <= dmach_paddr(i);
              rdcnt        <= dmach_length(i) & "00";
              rdtag        <= conv_std_logic_vector(i, rdtag'length);
              np_txwr_req  <= '1';
              rdsm         <= WAIT_MemRd;
            end if;
          end loop;

        when WAIT_MemRd =>
          crt_nphtmp := unsigned(cct_nph) + 1;

          if (np_txwr_ack = '1' and have_enough_credit(cat_nph, crt_nphtmp) = '1') then
            tx_nph  <= memrd32_header(rdcnt(rdcnt'left downto 2),
                                      rdpaddr(31 downto 0),
                                      rdtag, bdf, "000", "00");
            cct_nph        <= crt_nphtmp;
            np_txwr_bufsel <= '1';
            txwr_nph       <= '1';
            rdack          <= (others => '0');
            rdsm           <= IDLE;
          end if;

        when others =>

      end case;

    end if;
  end process;

end default;


#if 0

-- for debugging purpose
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;

entity fcbuf2 is 
  generic (
    VIRTUAL_CHANNEL : integer;
    DATA_WIDTH      : integer;
    DATA_DEPTH      : integer
  );
  port (
    clk         : in  std_logic;
    rst         : in  std_logic;     -- active high reset

    vc          : in  std_logic_vector(2 downto 0);
    wr          : in  std_logic;
    rd          : in  std_logic;
    data        : in  std_logic_vector(DATA_WIDTH-1 downto 0);

    empty       : out std_logic;
    full        : out std_logic;
    usedw       : out std_logic_vector(8 downto 0);
    q           : out std_logic_vector(DATA_WIDTH-1 downto 0)
  );
end fcbuf2;

architecture default of fcbuf2 is

  component scfifo
    generic (
      LPM_WIDTHU              : natural := 9;  -- !!
      LPM_WIDTH               : natural := DATA_WIDTH;
      LPM_NUMWORDS            : natural := DATA_DEPTH;
      LPM_SHOWAHEAD           : string  := "ON";
      ADD_RAM_OUTPUT_REGISTER : string  := "ON";
      OVERFLOW_CHECKING       : string  := "OFF";
      UNDERFLOW_CHECKING      : string  := "OFF";
      USE_EAB                 : string  := "ON";
      INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
      LPM_TYPE                : string  := "scfifo"
    );
    port (
      usedw    : out std_logic_vector(8 downto 0);  -- !!
      rdreq    : in  std_logic;
      empty    : out std_logic;
      full     : out std_logic;
      clock    : in  std_logic;
      sclr     : in  std_logic;
      q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
      wrreq    : in  std_logic;
      data     : in  std_logic_vector(LPM_WIDTH-1 downto 0)
    );
  end component;

  signal we          : std_logic;
  signal full0       : std_logic;
  signal fullout     : std_logic;
  
begin

    buf0 : scfifo
      generic map (
        LPM_NUMWORDS => DATA_DEPTH,
        LPM_WIDTH    => data'length
      )
      port map (
        clock => clk,
        sclr  => rst,

        wrreq => we,
        rdreq => rd,
        data  => data,

        usedw => usedw,
        empty => empty,
        full  => full0,
        q     => q
      );

    we <= '1' when wr = '1' and vc = conv_std_logic_vector(VIRTUAL_CHANNEL, vc'length)
              else '0';

   full <= full0;

end default;

#endif

