

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity gpcie is
  generic (
    NLANE : integer range 1 to 8 := 8; -- number of lanes. should be one of 1, 4, 8.
    NDMACH : integer range 0 to 7 := 2; -- number of DMA channels.

    NVC : integer := 1;
    N_FTS : integer := 6;
    DATARATE : std_logic_vector(7 downto 0) := "00000010";
    TSCTRL : std_logic_vector(7 downto 0) := "00000000"; -- bit0:hotreset 1:disable_link 2:loopback 3:disable_scrambling

    MAX_LINK_SPEED : std_logic_vector(3 downto 0) := "0001"; -- 2.5Gb/s
    MAX_READ_REQ_SIZE : integer := 4096; -- 4096 gives the best performance on DMA read,
                                          -- at the expense of hogging the link.
    MAX_PAYLOAD : integer := 256;

    CA_PH_VC0_INIT : integer := 16; -- PIO write
    CA_PD_VC0_INIT : integer := 64;

-- CA_NPH_VC0_INIT : integer := 4; -- PIO read
    CA_NPH_VC0_INIT : integer := 32; -- PIO read
    CA_NPD_VC0_INIT : integer := 2; -- data buf not used.

    CA_CH_VC0_INIT : integer := 16; -- completion for DMA read
    CA_CD_VC0_INIT : integer := 64;

    CL_PH_VC0_INIT : integer := 16; -- DMA write
    CL_PD_VC0_INIT : integer := 64;

    CL_NPH_VC0_INIT : integer := 2; -- DMA read
    CL_NPD_VC0_INIT : integer := 2; -- data buf not used.

-- CL_CH_VC0_INIT : integer := 4; -- completion for PIO read
-- CL_CD_VC0_INIT : integer := 4;

    CL_CH_VC0_INIT : integer := 16; -- completion for PIO read
    CL_CD_VC0_INIT : integer := 16;

    CFG_VENDOR_ID_INIT : std_logic_vector(15 downto 0) := x"1b1a"; -- KFCR
    CFG_DEVICE_ID_INIT : std_logic_vector(15 downto 0) := x"0e70"; -- PCI device ID.
    CFG_COMMAND_INIT : std_logic_vector(15 downto 0) := "00000" & -- 15:11 reserved.
                                                                    "1" & -- 10 interrupt disable. set to 1 to disable interrupt.
                                                                    "0000000" & -- some are reserved by the specification. others are not used for now.
                                                                    '1' & -- 2 bus master enable. set to 1 to make DMA engine to work.
                                                                    '1' & -- 1 memory address space decoder enable. set to 1 to accept memory transaction.
                                                                    '0'; -- 0 io address space decode enable. set to 0 for now. does not accept IO transaction.

    CFG_REVISION_ID_INIT : std_logic_vector( 7 downto 0) := x"01";
    CFG_CLASS_CODE_INIT : std_logic_vector(23 downto 0) := x"ff0000"; -- device does not fit any of the defined class codes.
    CFG_HEADER_TYPE_INIT : std_logic_vector( 7 downto 0) := x"00"; -- single function, non-bridge.
    CFG_BAR0_INIT : std_logic_vector(31 downto 0) := x"ffff8008"; -- 32kB, prefetchable, 32-bit address, memory space.
    CFG_BAR1_INIT : std_logic_vector(31 downto 0) := x"fffff008"; -- 4kB, prefetchable, 32-bit address, memory space.
    CFG_BAR2_INIT : std_logic_vector(31 downto 0) := x"ffff8008"; -- 32kB, prefetchable, 32-bit address, memory space.
    CFG_BAR3_INIT : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR4_INIT : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR5_INIT : std_logic_vector(31 downto 0) := x"00000000";
    CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0) := x"00000000";
    CFG_SUB_VENDOR_ID_INIT : std_logic_vector(15 downto 0) := x"1b1a";
    CFG_SUB_DEVICE_ID_INIT : std_logic_vector(15 downto 0) := x"0e70";
    CFG_INT_PIN_INIT : std_logic_vector( 7 downto 0) := x"00"; -- 00:generate no interrupts. 01:INTA 02:INTB 03:INTC 04:INTD

    PCIE_CAP_INIT : std_logic_vector(15 downto 0) := "0000000" & -- interrupt message number. not used for now.
                                                                    "0" & -- slot implemented. not used for endpoint.
                                                                    "0001" & -- device type. 0001:legacy PCIe endpoint.
                                                                    "0001"; -- capability version.

    PCIE_DEVICE_CAP_INIT : std_logic_vector(31 downto 0) := "0000" & -- reserved.
                                                                    "00" & -- captured slot power limit scale. not used for now.
                                                                    x"00" & -- captured slot power limit value. not used for now.
                                                                    "000" & -- reserved.
                                                                    '0' & -- no power indicator present.
                                                                    '0' & -- no attention indicator present.
                                                                    '0' & -- no attention button present.
                                                                    "000" & -- L1 latency is less than 1us.
                                                                    "000" & -- L0s latency is less than 64ns.
                                                                    "000" & -- do not support phantom functions.
                                                                    "000"; -- supported max payload size is automatically set to MAX_PAYLOAD.

    PCIE_DEVICE_CTL_INIT : std_logic_vector(15 downto 0) := '0' & -- reserved.
                                                                    "000" & -- max reqd request value is automatically set to MAX_READ_REQ_SIZE.
                                                                    "0000" & -- no snoop, aux power pm, phantom functions, extended tag are not supported.
                                                                    "000" & -- initial value of max payload size should be set to 128B.
                                                                    "00000"; -- relaxed ordering, advanced error reportings are not supported.

    PCIE_DEVICE_STATUS_INIT : std_logic_vector(15 downto 0) := x"0000";

    PCIE_LINK_CAP_INIT : std_logic_vector(31 downto 0) := x"00" & -- port number. designer should assign different number for different port.
                                                                    "00" & -- reserved.
                                                                    "0" & -- link bandwidth notification capability.
                                                                    "0" & -- data link layer link active reporting capable.
                                                                    "0" & -- surprise down error reporting capable.
                                                                    "0" & -- clock power management.
                                                                    "000" & -- L1 latency is less than 1us.
                                                                    "000" & -- L0s latency is less than 64ns.
                                                                    "01" & -- active state power management support. 01:L0s 11:L0s & L1
                                                                    "000000" & -- max link width value is automatically set to MAX_LINK_WIDTH.
                                                                    "0001"; -- max link speed value is automatically set to MAX_LINK_SPEED.

    PCIE_LINK_CTL_INIT : std_logic_vector(15 downto 0) := x"0000";
    PCIE_LINK_STATUS_INIT : std_logic_vector(15 downto 0) := x"0000"; -- negotiated link speed/width are fixed to MAX_LINK_SPEED/WIDTH so far.

    DEFRAMER_TLPBUF_DEPTH : integer := 8; -- buffer to store head of a TLP/DLLP which is received
    DEFRAMER_DLLPBUF_DEPTH : integer := 8; -- with the tail of previous packet in a single clk.
                                                  -- 8 bit (256 word) is sufficient for nForce 780i.
                                                  -- but it should be tested on various chipsets.
    TXDPBUF_DEPTH : integer := 6; -- 6 (i.e. 64 words) should be enough.
    SIMULATION : integer := 0
  );

  port (

    st2_trigger : out std_logic;
    st2_ltssm : out LTSSM_STATE;
    st2_txd : out each16b(NLANE-1 downto 0);
    st2_txk : out each2b(NLANE-1 downto 0);
    st2_rxd : out each16b(NLANE-1 downto 0);
    st2_rxk : out each2b(NLANE-1 downto 0);

    phy_linkup : out std_logic;
    dl_linkup : out std_logic;

    clk : in std_logic;
    rstn : in std_logic;

    -- PIPE interface
    phystatus : in std_logic;
    powerdown : out std_logic_vector(1 downto 0);
    txdetectrx : out std_logic;
    txdata : out std_logic_vector(NLANE*16-1 downto 0);
    txdatak : out std_logic_vector(NLANE*2-1 downto 0);
    txelecidle : out std_logic_vector(NLANE-1 downto 0);
    txcompl : out std_logic_vector(NLANE-1 downto 0);
    rxpolarity : out std_logic_vector(NLANE-1 downto 0);
    rxdata : in std_logic_vector(NLANE*16-1 downto 0);
    rxdatak : in std_logic_vector(NLANE*2-1 downto 0);
    rxvalid : in std_logic_vector(NLANE-1 downto 0);
    rxelecidle : in std_logic_vector(NLANE-1 downto 0);
    rxstatus : in std_logic_vector(NLANE*3-1 downto 0);

    --
    -- interface to user-side logic:
    --

    -- interface as a target (slave) device:
    slv_read : out std_logic;
    slv_write : out std_logic;
    slv_bar : out std_logic_vector(6 downto 0);
    slv_addr : out std_logic_vector(63 downto 0);
    slv_bytevalid : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain : in std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq : out std_logic;
    slv_writereq : out std_logic;
    slv_accept : in std_logic;
    slv_bytecount : out std_logic_vector(11 downto 0);

    -- interface as an initiator (master) device:
    ms_wrchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_write : out std_logic;
    ms_wraddr : out std_logic_vector(31 downto 0); -- 32-bit local address.
    ms_wrdata : in std_logic_vector(NLANE*16-1 downto 0);
    ms_rdchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_read : out std_logic;
    ms_rdaddr : out std_logic_vector(31 downto 0); -- 32-bit local address.
    ms_rddata : out std_logic_vector(NLANE*16-1 downto 0);

    -- interface to DMA controllers
    dma_control : in each7b(NDMACH-1 downto 0);
    dma_param : in each16b(NDMACH-1 downto 0);
    dma_status : out each4b(NDMACH-1 downto 0);
    dma_fifocnt : in each13b(NDMACH-1 downto 0);

    dma_paddrlow_in : in each32b(NDMACH-1 downto 0);
    dma_paddrhigh_in : in each32b(NDMACH-1 downto 0);
    dma_laddr_in : in each32b(NDMACH-1 downto 0);
    dma_size_in : in each32b(NDMACH-1 downto 0);

    dma_paddrlow_out : out each32b(NDMACH-1 downto 0);
    dma_paddrhigh_out : out each32b(NDMACH-1 downto 0);
    dma_laddr_out : out each32b(NDMACH-1 downto 0);
    dma_size_out : out each32b(NDMACH-1 downto 0)
  );

end gpcie;

architecture default of gpcie is

  component phymac
    generic (
      NLANE : integer := NLANE;
      N_FTS : integer := N_FTS;
      DATARATE : std_logic_vector(7 downto 0) := DATARATE;
      TSCTRL : std_logic_vector(7 downto 0) := TSCTRL;
      DEFRAMER_TLPBUF_DEPTH : integer := DEFRAMER_TLPBUF_DEPTH;
      DEFRAMER_DLLPBUF_DEPTH : integer := DEFRAMER_DLLPBUF_DEPTH;
      SIMULATION : integer := SIMULATION
    );
    port (
      st2_txd : out each16b(NLANE-1 downto 0);
      st2_txk : out each2b(NLANE-1 downto 0);
      st2_rxd : out each16b(NLANE-1 downto 0);
      st2_rxk : out each2b(NLANE-1 downto 0);

      clk : in std_logic;
      rst : in std_logic; -- active high

      -- PIPE interface
      txelecidleL : out each1b(NLANE-1 downto 0);
      txdatakL : out each2b(NLANE-1 downto 0);
      txdataL : out each16b(NLANE-1 downto 0);
      rxelecidleL : in each1b(NLANE-1 downto 0);
      rxdatakL : in each2b(NLANE-1 downto 0);
      rxdataL : in each16b(NLANE-1 downto 0);
      rxpolarityL : out each1b(NLANE-1 downto 0);

      -- DLL-MAC interface
      rxwedp : out std_logic;
      rxdp : out std_logic_vector(47 downto 0);
      rxwetp : out std_logic;
      rxbetp : out std_logic_vector(NLANE*2-1 downto 0);
      rxtptop : out std_logic;
      rxtp : out std_logic_vector(NLANE*16-1 downto 0);

      txdp : in std_logic_vector(47 downto 0);
      txemptydp : in std_logic;
      txrdreqdp : out std_logic;
      txtp : in std_logic_vector(NLANE*16-1 downto 0);
      txtoptp : in std_logic;
      txwetp : in std_logic;
      txbetp : in std_logic_vector(NLANE*2-1 downto 0);

      -- TL-PHYMAC interface
      txreq : in std_logic;
      txack : out std_logic;

      dmltssm : out LTSSM_STATE;
      dmlinknumL : out each8b(NLANE-1 downto 0);
      dmlanenumL : out each8b(NLANE-1 downto 0);
      dmdatarateL : out each8b(NLANE-1 downto 0);
      phy_linkup : out std_logic;
      dl_linkup : in std_logic
    );
  end component;

  component datalink
    generic (
      MAX_PAYLOAD : integer := MAX_PAYLOAD;
      CA_PH_VC0_INIT : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT : integer := CA_CD_VC0_INIT;
      TXDPBUF_DEPTH : integer := TXDPBUF_DEPTH;
      NLANE : integer := NLANE;
      SIMULATION : integer := SIMULATION
    );
    port (
      st2_trigger : out std_logic;

      clk : in std_logic;
      rst : in std_logic; -- active high

      -- DLL-MAC interface
      rxwedp : in std_logic;
      rxdp : in std_logic_vector(47 downto 0);
      rxwetp : in std_logic;
      rxbetp : in std_logic_vector(NLANE*2-1 downto 0);
      rxtptop : in std_logic;
      rxtp : in std_logic_vector(NLANE*16-1 downto 0);

      txemptydp : out std_logic;
      txrdreqdp : in std_logic;
      txdp : out std_logic_vector(47 downto 0);
      txtoptp : out std_logic;
      txwetp : out std_logic;
      txbetp : out std_logic_vector(NLANE*2-1 downto 0);
      txtp : out std_logic_vector(NLANE*16-1 downto 0);

      -- TL-DLL interface
      rxtop : out std_logic;
      rxwe : out std_logic;
      rx : out std_logic_vector(NLANE*16-1 downto 0);

      txtop : in std_logic;
      txwe : in std_logic;
      txbe : in std_logic_vector(NLANE*2-1 downto 0);
      tx : in std_logic_vector(NLANE*16-1 downto 0);
      tpsize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);

      -- CA notified by transaction layer
      ca_nph : in std_logic_vector(7 downto 0);
      ca_ph : in std_logic_vector(7 downto 0);
      ca_ch : in std_logic_vector(7 downto 0);
      ca_npd : in std_logic_vector(11 downto 0);
      ca_pd : in std_logic_vector(11 downto 0);
      ca_cd : in std_logic_vector(11 downto 0);
      ca_p_flush : in std_logic;
      ca_np_flush : in std_logic;
      ca_c_flush : in std_logic;

      -- notify CL to transaction layer
      cl_nph : out std_logic_vector( 7 downto 0);
      cl_ph : out std_logic_vector( 7 downto 0);
      cl_ch : out std_logic_vector( 7 downto 0);
      cl_npd : out std_logic_vector(11 downto 0);
      cl_pd : out std_logic_vector(11 downto 0);
      cl_cd : out std_logic_vector(11 downto 0);

      inf_cl_ph : out std_logic;
      inf_cl_pd : out std_logic;
      inf_cl_nph : out std_logic;
      inf_cl_npd : out std_logic;
      inf_cl_ch : out std_logic;
      inf_cl_cd : out std_logic;

      -- core status
      phy_linkup : in std_logic;
      dl_linkup : out std_logic
    );
  end component;

  component transaction
    generic (
      NLANE : integer := NLANE;
      NVC : integer := NVC;
      MAX_PAYLOAD : integer := MAX_PAYLOAD;

      CA_PH_VC0_INIT : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT : integer := CA_CD_VC0_INIT;

      CL_PH_VC0_INIT : integer := CL_PH_VC0_INIT;
      CL_NPH_VC0_INIT : integer := CL_NPH_VC0_INIT;
      CL_CH_VC0_INIT : integer := CL_CH_VC0_INIT;
      CL_PD_VC0_INIT : integer := CL_PD_VC0_INIT;
      CL_NPD_VC0_INIT : integer := CL_NPD_VC0_INIT;
      CL_CD_VC0_INIT : integer := CL_CD_VC0_INIT;

      CFG_BAR0_INIT : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;

      NDMACH : integer := NDMACH;
      SIMULATION : integer := SIMULATION
    );
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high

      -- TL-PHYMAC interface
      txreq : out std_logic;
      txack : in std_logic;

      -- TL-DLL interface
      rxtop : in std_logic;
      rxwr : in std_logic;
      rx : in std_logic_vector(NLANE*16-1 downto 0);

      txtop : out std_logic;
      txwr : out std_logic;
      txbe : out std_logic_vector(NLANE*2-1 downto 0);
      tx : out std_logic_vector(NLANE*16-1 downto 0);
      tpsize : out std_logic_vector(OFFSET_WIDTH-1 downto 0);

      -- notify CA to data-link layer
      ca_nph : out std_logic_vector(7 downto 0);
      ca_ph : out std_logic_vector(7 downto 0);
      ca_ch : out std_logic_vector(7 downto 0);
      ca_npd : out std_logic_vector(11 downto 0);
      ca_pd : out std_logic_vector(11 downto 0);
      ca_cd : out std_logic_vector(11 downto 0);
      ca_p_flush : out std_logic;
      ca_np_flush : out std_logic;
      ca_c_flush : out std_logic;

      -- CL notified by data-link layer
      cl_nph : in std_logic_vector( 7 downto 0);
      cl_ph : in std_logic_vector( 7 downto 0);
      cl_ch : in std_logic_vector( 7 downto 0);
      cl_npd : in std_logic_vector(11 downto 0);
      cl_pd : in std_logic_vector(11 downto 0);
      cl_cd : in std_logic_vector(11 downto 0);

      inf_cl_ph : in std_logic;
      inf_cl_pd : in std_logic;
      inf_cl_nph : in std_logic;
      inf_cl_npd : in std_logic;
      inf_cl_ch : in std_logic;
      inf_cl_cd : in std_logic;

      -- AL-TL interface
      mpls : in std_logic_vector(11 downto 0);

      cfg_bar0 : in std_logic_vector(31 downto 0);
      cfg_bar1 : in std_logic_vector(31 downto 0);
      cfg_bar2 : in std_logic_vector(31 downto 0);
      cfg_bar3 : in std_logic_vector(31 downto 0);
      cfg_bar4 : in std_logic_vector(31 downto 0);
      cfg_bar5 : in std_logic_vector(31 downto 0);
      cfg_bar_rom : in std_logic_vector(31 downto 0);

      cfg_read : out std_logic;
      cfg_write : out std_logic;
      cfg_addr : out std_logic_vector(11 downto 0);
      cfg_be : out std_logic_vector(3 downto 0);
      cfg_dataout : out std_logic_vector(31 downto 0);
      cfg_datain : in std_logic_vector(31 downto 0);

      -- core interface
      slv_read : out std_logic;
      slv_write : out std_logic;
      slv_bar : out std_logic_vector(6 downto 0);
      slv_addr : out std_logic_vector(63 downto 0);
      slv_bytevalid : out std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout : out std_logic_vector(NLANE*16-1 downto 0);
      slv_datain : in std_logic_vector(NLANE*16-1 downto 0);

      slv_readreq : out std_logic;
      slv_writereq : out std_logic;
      slv_accept : in std_logic;
      slv_bytecount : out std_logic_vector(11 downto 0);

      dma_fifocnt_off : out std_logic_vector(4 downto 0);
      ms_wrchannel : out std_logic_vector(NDMACH-1 downto 0);
      ms_write : out std_logic;
      ms_wraddr : out std_logic_vector(31 downto 0);
      ms_wrdata : in std_logic_vector(NLANE*16-1 downto 0);
      ms_rdchannel : out std_logic_vector(NDMACH-1 downto 0);
      ms_read : out std_logic;
      ms_rdaddr : out std_logic_vector(31 downto 0);
      ms_rddata : out std_logic_vector(NLANE*16-1 downto 0);

      dmach_wrreq : in std_logic_vector(NDMACH-1 downto 0);
      dmach_wrack : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdreq : in std_logic_vector(NDMACH-1 downto 0);
      dmach_rdack : out std_logic_vector(NDMACH-1 downto 0);
      dmach_length : in each10b(NDMACH-1 downto 0);
      dmach_paddr : in each64b(NDMACH-1 downto 0);
      dmach_laddr : in each32b(NDMACH-1 downto 0);

      dmach_rdcpl : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdcpllength : out std_logic_vector(9 downto 0);
      dmach_rdcplbcnt : out std_logic_vector(11 downto 0);
      dmach_rdcpladdr : in each32b(NDMACH-1 downto 0);

      phy_linkup : in std_logic;
      dl_linkup : in std_logic
    );
  end component;

  component cfgreg
    generic (
      CFG_VENDOR_ID_INIT : std_logic_vector(15 downto 0) := CFG_VENDOR_ID_INIT;
      CFG_DEVICE_ID_INIT : std_logic_vector(15 downto 0) := CFG_DEVICE_ID_INIT;
      CFG_COMMAND_INIT : std_logic_vector(15 downto 0) := CFG_COMMAND_INIT;
      CFG_REVISION_ID_INIT : std_logic_vector( 7 downto 0) := CFG_REVISION_ID_INIT;
      CFG_CLASS_CODE_INIT : std_logic_vector(23 downto 0) := CFG_CLASS_CODE_INIT;
      CFG_HEADER_TYPE_INIT : std_logic_vector( 7 downto 0) := CFG_HEADER_TYPE_INIT;
      CFG_BAR0_INIT : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;
      CFG_SUB_VENDOR_ID_INIT : std_logic_vector(15 downto 0) := CFG_SUB_VENDOR_ID_INIT;
      CFG_SUB_DEVICE_ID_INIT : std_logic_vector(15 downto 0) := CFG_SUB_DEVICE_ID_INIT;
      CFG_INT_PIN_INIT : std_logic_vector( 7 downto 0) := CFG_INT_PIN_INIT;

      PCIE_CAP_INIT : std_logic_vector(15 downto 0) := PCIE_CAP_INIT;
      PCIE_DEVICE_CAP_INIT : std_logic_vector(31 downto 0) := PCIE_DEVICE_CAP_INIT;
      PCIE_DEVICE_CTL_INIT : std_logic_vector(15 downto 0) := PCIE_DEVICE_CTL_INIT;
      PCIE_DEVICE_STATUS_INIT : std_logic_vector(15 downto 0) := PCIE_DEVICE_STATUS_INIT;
      PCIE_LINK_CAP_INIT : std_logic_vector(31 downto 0) := PCIE_LINK_CAP_INIT;
      PCIE_LINK_CTL_INIT : std_logic_vector(15 downto 0) := PCIE_LINK_CTL_INIT;
      PCIE_LINK_STATUS_INIT : std_logic_vector(15 downto 0) := PCIE_LINK_STATUS_INIT;

      NLANE : integer := NLANE;
      MAX_LINK_SPEED : std_logic_vector( 3 downto 0) := MAX_LINK_SPEED;
      MAX_READ_REQ_SIZE : integer := MAX_READ_REQ_SIZE;
      MAX_PAYLOAD : integer := MAX_PAYLOAD;
      SIMULATION : integer := SIMULATION
    );
    port (
      clk : in std_logic;
      rst : in std_logic;

      mpls : out std_logic_vector(11 downto 0); -- max payload size
      mrrs : out std_logic_vector(11+1 downto 0); -- max read request size
      rcb : out std_logic; -- read completion boundary

      cfg_bar0_out : out std_logic_vector(31 downto 0);
      cfg_bar1_out : out std_logic_vector(31 downto 0);
      cfg_bar2_out : out std_logic_vector(31 downto 0);
      cfg_bar3_out : out std_logic_vector(31 downto 0);
      cfg_bar4_out : out std_logic_vector(31 downto 0);
      cfg_bar5_out : out std_logic_vector(31 downto 0);
      cfg_bar_rom_out : out std_logic_vector(31 downto 0);

      rd : in std_logic;
      wr : in std_logic;
      addr : in std_logic_vector(11 downto 0);
      be : in std_logic_vector(3 downto 0);
      datain : in std_logic_vector(31 downto 0);
      dataout : out std_logic_vector(31 downto 0)
    );
  end component;

  component dmactl
    generic (
      NLANE : integer := NLANE;
      NDMACH : integer := NDMACH
    );
    port (
      clk : in std_logic;
      rst : in std_logic;

      mpls : in std_logic_vector(11 downto 0); -- max payload size
      mrrs : in std_logic_vector(11+1 downto 0); -- max read request size

      -- interface to user-side logic:
      dma_control : in std_logic_vector(6 downto 0);
      dma_param : in std_logic_vector(15 downto 0);
      dma_status : out std_logic_vector(3 downto 0);
      dma_fifocnt : in std_logic_vector(12 downto 0);

      dma_paddrlow_in : in std_logic_vector(31 downto 0);
      dma_paddrhigh_in : in std_logic_vector(31 downto 0);
      dma_laddr_in : in std_logic_vector(31 downto 0);
      dma_size_in : in std_logic_vector(31 downto 0);

      dma_paddrlow_out : out std_logic_vector(31 downto 0);
      dma_paddrhigh_out : out std_logic_vector(31 downto 0);
      dma_laddr_out : out std_logic_vector(31 downto 0);
      dma_size_out : out std_logic_vector(31 downto 0);

      -- interface to master controller:
      dma_fifocnt_off : in std_logic_vector(4 downto 0);
      wrreq : out std_logic;
      wrack : in std_logic;
      rdreq : out std_logic;
      rdack : in std_logic;

      length : out std_logic_vector(PAYLOAD_WIDTH-1 downto 0);
      paddr : out std_logic_vector(63 downto 0);
      laddr : out std_logic_vector(31 downto 0);

      -- interface to slave controller:
      rdcpl : in std_logic;
      rdcpl_length : in std_logic_vector(9 downto 0);
      rdcpl_bcnt : in std_logic_vector(11 downto 0);
      rdcpl_addr : out std_logic_vector(31 downto 0)
    );
  end component;

  -- PIPE interface
  signal txelecidleL : each1b(NLANE-1 downto 0);
  signal txdatakL : each2b(NLANE-1 downto 0);
  signal txdataL : each16b(NLANE-1 downto 0);
  signal rxelecidleL : each1b(NLANE-1 downto 0);
  signal rxdatakL : each2b(NLANE-1 downto 0);
  signal rxdataL : each16b(NLANE-1 downto 0);
  signal rxpolarityL : each1b(NLANE-1 downto 0);

  -- signals commonly used inside the core
  signal linknumL : each8b(NLANE-1 downto 0);
  signal lanenumL : each8b(NLANE-1 downto 0);
  signal datarateL : each8b(NLANE-1 downto 0);
  signal phy_linkup0 : std_logic;
  signal dl_linkup0 : std_logic;

  -- DLL-MAC interface
  signal rxdp : std_logic_vector(47 downto 0);
  signal rxwedp : std_logic;
  signal rxtp : std_logic_vector(NLANE*16-1 downto 0);
  signal rxwetp : std_logic;
  signal rxtptop : std_logic;
  signal rxbetp : std_logic_vector(NLANE*2-1 downto 0);

  signal txdp : std_logic_vector(47 downto 0);
  signal txemptydp : std_logic;
  signal txrdreqdp : std_logic;
  signal txtp : std_logic_vector(NLANE*16-1 downto 0);
  signal txtoptp : std_logic;
  signal txwetp : std_logic;
  signal txbetp : std_logic_vector(NLANE*2-1 downto 0);

  -- TL-DLL interface
  signal rxtop : std_logic;
  signal rxwe : std_logic;
  signal rx : std_logic_vector(NLANE*16-1 downto 0);
  signal txtop : std_logic;
  signal txwe : std_logic;
  signal txbe : std_logic_vector(NLANE*2-1 downto 0);
  signal tx : std_logic_vector(NLANE*16-1 downto 0);
  signal tpsize : std_logic_vector(OFFSET_WIDTH-1 downto 0);

  signal ca_nph : std_logic_vector(7 downto 0);
  signal ca_ph : std_logic_vector(7 downto 0);
  signal ca_ch : std_logic_vector(7 downto 0);
  signal ca_npd : std_logic_vector(11 downto 0);
  signal ca_pd : std_logic_vector(11 downto 0);
  signal ca_cd : std_logic_vector(11 downto 0);
  signal ca_p_flush : std_logic;
  signal ca_np_flush : std_logic;
  signal ca_c_flush : std_logic;

  signal cl_nph : std_logic_vector(7 downto 0);
  signal cl_ph : std_logic_vector(7 downto 0);
  signal cl_ch : std_logic_vector(7 downto 0);
  signal cl_npd : std_logic_vector(11 downto 0);
  signal cl_pd : std_logic_vector(11 downto 0);
  signal cl_cd : std_logic_vector(11 downto 0);

  signal inf_cl_nph : std_logic;
  signal inf_cl_ph : std_logic;
  signal inf_cl_ch : std_logic;
  signal inf_cl_npd : std_logic;
  signal inf_cl_pd : std_logic;
  signal inf_cl_cd : std_logic;

  -- TL-PHYMAC interface
  signal txreq : std_logic;
  signal txack : std_logic;

  -- AL-TL interface
  signal pcie_device_ctl : std_logic_vector(15 downto 0);
  signal pcie_link_ctl : std_logic_vector(15 downto 0);
  signal cfg_bar0 : std_logic_vector(31 downto 0);
  signal cfg_bar1 : std_logic_vector(31 downto 0);
  signal cfg_bar2 : std_logic_vector(31 downto 0);
  signal cfg_bar3 : std_logic_vector(31 downto 0);
  signal cfg_bar4 : std_logic_vector(31 downto 0);
  signal cfg_bar5 : std_logic_vector(31 downto 0);
  signal cfg_bar_rom : std_logic_vector(31 downto 0);

  signal cfg_read : std_logic;
  signal cfg_write : std_logic;
  signal cfg_addr : std_logic_vector(11 downto 0);
  signal cfg_be : std_logic_vector(3 downto 0);
  signal cfg_dataout : std_logic_vector(31 downto 0);
  signal cfg_datain : std_logic_vector(31 downto 0);

  signal mpls : std_logic_vector(11 downto 0);
  signal mrrs : std_logic_vector(11+1 downto 0);
  signal rcb : std_logic;

  signal dma_fifocnt_off : std_logic_vector(4 downto 0);
  signal dmach_wrreq : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_wrack : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdreq : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdack : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_length : each10b(NDMACH-1 downto 0);
  signal dmach_paddr : each64b(NDMACH-1 downto 0);
  signal dmach_laddr : each32b(NDMACH-1 downto 0);

  signal dmach_rdcpl : std_logic_vector(NDMACH-1 downto 0);
  signal dmach_rdcpllength : std_logic_vector(9 downto 0);
  signal dmach_rdcplbcnt : std_logic_vector(11 downto 0);
  signal dmach_rdcpladdr : each32b(NDMACH-1 downto 0);

  -- for debugging purpose
  signal trigger : std_logic;
  signal dmltssm : LTSSM_STATE;
  signal dmltssm1 : LTSSM_STATE;
  signal txd : each16b(NLANE-1 downto 0);
  signal txk : each2b(NLANE-1 downto 0);
  signal rxd : each16b(NLANE-1 downto 0);
  signal rxk : each2b(NLANE-1 downto 0);

  attribute keep : boolean;
  attribute keep of txd : signal is true;
  attribute keep of txk : signal is true;

-- synthesis translate_off

  attribute keep of trigger : signal is true;
  attribute keep of dmltssm : signal is true;
  attribute keep of clk : signal is true;

  attribute keep of rxdp : signal is true;
  attribute keep of rxwedp : signal is true;
  attribute keep of txd : signal is true;
  attribute keep of txk : signal is true;

  attribute keep of dmltssm : signal is true;
  attribute keep of rxd : signal is true;
  attribute keep of rxk : signal is true;
  attribute keep of rxelecidleL : signal is true;
  attribute keep of txelecidleL : signal is true;
  attribute keep of txdataL : signal is true;
  attribute keep of txdatakL : signal is true;

  attribute keep of rxdp : signal is true;
  attribute keep of rxwedp : signal is true;
  attribute keep of rxtp : signal is true;
  attribute keep of rxwetp : signal is true;

  attribute keep of txdp : signal is true;
  attribute keep of txemptydp : signal is true;
  attribute keep of txrdreqdp : signal is true;
  attribute keep of txtp : signal is true;
  attribute keep of txemptytp : signal is true;
  attribute keep of txrdreqtp : signal is true;

-- synthesis translate_on

  signal rxk1 : each2b(NLANE-1 downto 0);
  signal rxd1 : each16b(NLANE-1 downto 0);

begin

  -- for debugging
  process (clk)
  begin
    if (rising_edge(clk)) then
      st2_trigger <= trigger;
      st2_ltssm <= dmltssm;
      st2_txd <= txd;
      st2_txk <= txk;
      st2_rxd <= rxd;
      st2_rxk <= rxk;
      dmltssm1 <= dmltssm;

      rxk1 <= rxk;
      rxd1 <= rxd;

    end if;
  end process;

  phy_linkup <= phy_linkup0;
  dl_linkup <= dl_linkup0;

  phymac0 : phymac
    port map (
      st2_txd => txd,
      st2_txk => txk,
      st2_rxd => rxd,
      st2_rxk => rxk,

      clk => clk,
      rst => not rstn,

      txelecidleL => txelecidleL,
      txdatakL => txdatakL,
      txdataL => txdataL,
      rxelecidleL => rxelecidleL,
      rxdatakL => rxdatakL,
      rxdataL => rxdataL,
      rxpolarityL => rxpolarityL,

      rxwedp => rxwedp,
      rxdp => rxdp,
      rxwetp => rxwetp,
      rxtptop => rxtptop,
      rxtp => rxtp,
      rxbetp => rxbetp,

      txdp => txdp,
      txemptydp => txemptydp,
      txrdreqdp => txrdreqdp,

      txreq => txreq,
      txack => txack,
      txtoptp => txtoptp,
      txwetp => txwetp,
      txbetp => txbetp,
      txtp => txtp,

      dmltssm => dmltssm,
      dmlinknumL => linknumL,
      dmlanenumL => lanenumL,
      dmdatarateL => datarateL,

      phy_linkup => phy_linkup0,
      dl_linkup => dl_linkup0
    );

  sig_lane : for i in 0 to NLANE-1 generate
    -- inputs
    rxdatakL(i) <= rxdatak(2*(i+1)-1 downto 2*i);
    rxdataL(i) <= rxdata(16*(i+1)-1 downto 16*i);
    rxelecidleL(i) <= rxelecidle(i);
    -- outputs
    txdatak(2*(i+1)-1 downto 2*i) <= txdatakL(i);
    txdata(16*(i+1)-1 downto 16*i) <= txdataL(i);
    txelecidle(i) <= txelecidleL(i);
    rxpolarity(i) <= rxpolarityL(i);
    txcompl(i) <= '0';
  end generate sig_lane;

  powerdown <= "00";
  txdetectrx <= '0';

  dll0 : datalink
    port map (
      st2_trigger => trigger,

      clk => clk,
      rst => not (rstn and phy_linkup0),

      -- DLL-MAC interface
      rxdp => rxdp,
      rxwedp => rxwedp,

      txdp => txdp,
      txemptydp => txemptydp,
      txrdreqdp => txrdreqdp,

      rxwetp => rxwetp,
      rxtptop => rxtptop,
      rxtp => rxtp,
      txtoptp => txtoptp,
      txwetp => txwetp,
      txbetp => txbetp,
      txtp => txtp,
      rxbetp => rxbetp,

      -- TL-DLL interface
      rxtop => rxtop,
      rxwe => rxwe,
      rx => rx,
      txtop => txtop,
      txwe => txwe,
      txbe => txbe,
      tx => tx,
      tpsize => tpsize,

      ca_nph => ca_nph,
      ca_ph => ca_ph,
      ca_ch => ca_ch,
      ca_npd => ca_npd,
      ca_pd => ca_pd,
      ca_cd => ca_cd,
      ca_p_flush => ca_p_flush,
      ca_np_flush => ca_np_flush,
      ca_c_flush => ca_c_flush,

      cl_nph => cl_nph,
      cl_ph => cl_ph,
      cl_ch => cl_ch,
      cl_npd => cl_npd,
      cl_pd => cl_pd,
      cl_cd => cl_cd,

      inf_cl_nph => inf_cl_nph,
      inf_cl_ph => inf_cl_ph,
      inf_cl_ch => inf_cl_ch,
      inf_cl_npd => inf_cl_npd,
      inf_cl_pd => inf_cl_pd,
      inf_cl_cd => inf_cl_cd,

      phy_linkup => phy_linkup0,
      dl_linkup => dl_linkup0
    );

  tl0 : transaction
    port map (
      clk => clk,
      rst => not (rstn and phy_linkup0),
      rx => rx,
      rxtop => rxtop,
      rxwr => rxwe,
      txreq => txreq,
      txack => txack,
      tx => tx,
      tpsize => tpsize,
      txtop => txtop,
      txwr => txwe,
      txbe => txbe,

      ca_nph => ca_nph,
      ca_ph => ca_ph,
      ca_ch => ca_ch,
      ca_npd => ca_npd,
      ca_pd => ca_pd,
      ca_cd => ca_cd,
      ca_p_flush => ca_p_flush,
      ca_np_flush => ca_np_flush,
      ca_c_flush => ca_c_flush,

      cl_nph => cl_nph,
      cl_ph => cl_ph,
      cl_ch => cl_ch,
      cl_npd => cl_npd,
      cl_pd => cl_pd,
      cl_cd => cl_cd,

      inf_cl_nph => inf_cl_nph,
      inf_cl_ph => inf_cl_ph,
      inf_cl_ch => inf_cl_ch,
      inf_cl_npd => inf_cl_npd,
      inf_cl_pd => inf_cl_pd,
      inf_cl_cd => inf_cl_cd,

      mpls => mpls,

      cfg_bar0 => cfg_bar0,
      cfg_bar1 => cfg_bar1,
      cfg_bar2 => cfg_bar2,
      cfg_bar3 => cfg_bar3,
      cfg_bar4 => cfg_bar4,
      cfg_bar5 => cfg_bar5,
      cfg_bar_rom => cfg_bar_rom,

      cfg_read => cfg_read,
      cfg_write => cfg_write,
      cfg_addr => cfg_addr,
      cfg_be => cfg_be,
      cfg_dataout => cfg_dataout,
      cfg_datain => cfg_datain,

      slv_read => slv_read,
      slv_write => slv_write,
      slv_bar => slv_bar,
      slv_addr => slv_addr,
      slv_bytevalid => slv_bytevalid,
      slv_dataout => slv_dataout,
      slv_datain => slv_datain,

      slv_readreq => slv_readreq,
      slv_writereq => slv_writereq,
      slv_accept => slv_accept,
      slv_bytecount => slv_bytecount,

      ms_wrchannel => ms_wrchannel,
      ms_write => ms_write,
      ms_wraddr => ms_wraddr,
      ms_wrdata => ms_wrdata,

      ms_rdchannel => ms_rdchannel,
      ms_read => ms_read,
      ms_rdaddr => ms_rdaddr,
      ms_rddata => ms_rddata,

      dma_fifocnt_off => dma_fifocnt_off,
      dmach_wrreq => dmach_wrreq,
      dmach_wrack => dmach_wrack,
      dmach_rdreq => dmach_rdreq,
      dmach_rdack => dmach_rdack,
      dmach_length => dmach_length,
      dmach_paddr => dmach_paddr,
      dmach_laddr => dmach_laddr,

      dmach_rdcpl => dmach_rdcpl,
      dmach_rdcpllength => dmach_rdcpllength,
      dmach_rdcplbcnt => dmach_rdcplbcnt,
      dmach_rdcpladdr => dmach_rdcpladdr,

      phy_linkup => phy_linkup0,
      dl_linkup => dl_linkup0
    );

  cfgreg0 : cfgreg
    port map (
      clk => clk,
-- rst => not (rstn and phy_linkup0),
      rst => not rstn, -- !!!

      mpls => mpls,
      mrrs => mrrs,
      rcb => rcb,

      cfg_bar0_out => cfg_bar0,
      cfg_bar1_out => cfg_bar1,
      cfg_bar2_out => cfg_bar2,
      cfg_bar3_out => cfg_bar3,
      cfg_bar4_out => cfg_bar4,
      cfg_bar5_out => cfg_bar5,
      cfg_bar_rom_out => cfg_bar_rom,

      rd => cfg_read,
      wr => cfg_write,
      be => cfg_be,
      addr => cfg_addr,
      datain => cfg_dataout,
      dataout => cfg_datain
    );

  dma_ch : for i in 0 to NDMACH-1 generate
    u0 : dmactl
      port map (
        clk => clk,
-- rst => not (rstn and phy_linkup0),
        rst => not rstn, -- !!!

        dma_control => dma_control(i),
        dma_param => dma_param(i),
        dma_status => dma_status(i),
        dma_fifocnt => dma_fifocnt(i),

        dma_paddrlow_in => dma_paddrlow_in(i),
        dma_paddrhigh_in => dma_paddrhigh_in(i),
        dma_laddr_in => dma_laddr_in(i),
        dma_size_in => dma_size_in(i),

        dma_paddrlow_out => dma_paddrlow_out(i),
        dma_paddrhigh_out => dma_paddrhigh_out(i),
        dma_laddr_out => dma_laddr_out(i),
        dma_size_out => dma_size_out(i),

        dma_fifocnt_off => dma_fifocnt_off,

        mpls => mpls,
        mrrs => mrrs,

        wrreq => dmach_wrreq(i),
        wrack => dmach_wrack(i),
        rdreq => dmach_rdreq(i),
        rdack => dmach_rdack(i),
        length => dmach_length(i),
        paddr => dmach_paddr(i),
        laddr => dmach_laddr(i),

        rdcpl => dmach_rdcpl(i),
        rdcpl_length => dmach_rdcpllength,
        rdcpl_bcnt => dmach_rdcplbcnt,
        rdcpl_addr => dmach_rdcpladdr(i)
      );

  end generate dma_ch;

end default;

--
-- Type0 Configuration Register
--
-- !! accepts DW (32-bit) access only.
--
-------------------------------------------------------------
-- addr 31:24 23:16 15:08 07:00
-- byte DW
-------------------------------------------------------------
-- 000 00 device_id vendor_id
-- 004 01 status command
-- 008 02 class_code revision_id
-- 00c 03 BIST hdr_type lat_timer cache_line_size
-- 010 04 bar0
-- 014 05 bar1
-- 018 06 bar2
-- 01c 07 bar3
-- 020 08 bar4
-- 024 09 bar5
-- 028 0a CardBus CIS pointer
-- 02c 0b sub_device_id sub_vendor_id
-- 030 0c bar_rom
-- 034 0d cap_ptr
-- 038 0e
-- 03c 0f max_gnt min_gnt int_pin int_line
-------------------------------------------------------------
--
--
-- PCI Express Capability Register
-------------------------------------------------------------
-- addr 31:24 23:16 15:08 07:00
-- byte DW
-------------------------------------------------------------
-- 080 00 pcie_cap next_ptr cap_id
-- 084 01 device_cap
-- 088 02 device_status device_ctl
-- 08c 03 link_cap
-- 090 04 link_status link_ctl
-- 094 05 not used.
-- 098 06 not used.
-- 09c 07 not used.
-- 0a0 08 not used.
-------------------------------------------------------------
--
-- PCI Express Extended Capability Register
-------------------------------------------------------------
-- addr 31:24 23:16 15:08 07:00
-- byte
-- 100 00000000 (which indicates not available).
-------------------------------------------------------------
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity cfgreg is
  generic (
    CFG_VENDOR_ID_INIT : std_logic_vector(15 downto 0);
    CFG_DEVICE_ID_INIT : std_logic_vector(15 downto 0);
    CFG_COMMAND_INIT : std_logic_vector(15 downto 0);

    CFG_REVISION_ID_INIT : std_logic_vector( 7 downto 0);
    CFG_CLASS_CODE_INIT : std_logic_vector(23 downto 0);
    CFG_HEADER_TYPE_INIT : std_logic_vector( 7 downto 0);
    CFG_BAR0_INIT : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0);
    CFG_SUB_VENDOR_ID_INIT : std_logic_vector(15 downto 0);
    CFG_SUB_DEVICE_ID_INIT : std_logic_vector(15 downto 0);
    CFG_INT_PIN_INIT : std_logic_vector( 7 downto 0);

    PCIE_CAP_INIT : std_logic_vector(15 downto 0);
    PCIE_DEVICE_CAP_INIT : std_logic_vector(31 downto 0);
    PCIE_DEVICE_CTL_INIT : std_logic_vector(15 downto 0);
    PCIE_DEVICE_STATUS_INIT : std_logic_vector(15 downto 0);
    PCIE_LINK_CAP_INIT : std_logic_vector(31 downto 0);
    PCIE_LINK_CTL_INIT : std_logic_vector(15 downto 0);
    PCIE_LINK_STATUS_INIT : std_logic_vector(15 downto 0);

    NLANE : integer;
    MAX_LINK_SPEED : std_logic_vector(3 downto 0);
    MAX_READ_REQ_SIZE : integer;
    MAX_PAYLOAD : integer;
    SIMULATION : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;

    mpls : out std_logic_vector(11 downto 0); -- max payload size
    mrrs : out std_logic_vector(11+1 downto 0); -- max read request size
    rcb : out std_logic; -- read completion boundary

    cfg_bar0_out : out std_logic_vector(31 downto 0);
    cfg_bar1_out : out std_logic_vector(31 downto 0);
    cfg_bar2_out : out std_logic_vector(31 downto 0);
    cfg_bar3_out : out std_logic_vector(31 downto 0);
    cfg_bar4_out : out std_logic_vector(31 downto 0);
    cfg_bar5_out : out std_logic_vector(31 downto 0);
    cfg_bar_rom_out : out std_logic_vector(31 downto 0);

    rd : in std_logic;
    wr : in std_logic;
    addr : in std_logic_vector(11 downto 0);
    be : in std_logic_vector(3 downto 0);
    datain : in std_logic_vector(31 downto 0);
    dataout : out std_logic_vector(31 downto 0)
  );
end cfgreg;

architecture default of cfgreg is

  -- Type0 Configuration Registers
  signal cfg_vendor_id : std_logic_vector(15 downto 0);
  signal cfg_device_id : std_logic_vector(15 downto 0);
  signal cfg_command : std_logic_vector(15 downto 0);
  signal cfg_status : std_logic_vector(15 downto 0);
  signal cfg_revision_id : std_logic_vector( 7 downto 0);
  signal cfg_class_code : std_logic_vector(23 downto 0);
  signal cfg_header_type : std_logic_vector(7 downto 0);
  signal cfg_cache_line_size : std_logic_vector(7 downto 0);
  signal cfg_latency_timer : std_logic_vector(7 downto 0);
  signal cfg_bist : std_logic_vector(7 downto 0);
  signal cfg_bar0 : std_logic_vector(31 downto 0);
  signal cfg_bar1 : std_logic_vector(31 downto 0);
  signal cfg_bar2 : std_logic_vector(31 downto 0);
  signal cfg_bar3 : std_logic_vector(31 downto 0);
  signal cfg_bar4 : std_logic_vector(31 downto 0);
  signal cfg_bar5 : std_logic_vector(31 downto 0);
  signal cfg_cardbus_cis_ptr : std_logic_vector(31 downto 0);
  signal cfg_sub_vendor_id : std_logic_vector(15 downto 0);
  signal cfg_sub_device_id : std_logic_vector(15 downto 0);
  signal cfg_bar_rom : std_logic_vector(31 downto 0);
  signal cfg_cap_ptr : std_logic_vector(7 downto 0);
  signal cfg_int_line : std_logic_vector(7 downto 0);
  signal cfg_int_pin : std_logic_vector(7 downto 0);
  signal cfg_min_gnt : std_logic_vector(7 downto 0);
  signal cfg_max_gnt : std_logic_vector(7 downto 0);

  -- PCI Express Capability Registers
  signal pcie_cap_id : std_logic_vector(7 downto 0);
  signal pcie_cap_ptr : std_logic_vector(7 downto 0);
  signal pcie_cap : std_logic_vector(15 downto 0);
  signal pcie_device_cap : std_logic_vector(31 downto 0);
  signal pcie_device_ctl : std_logic_vector(15 downto 0);
  signal pcie_device_status : std_logic_vector(15 downto 0);
  signal pcie_link_cap : std_logic_vector(31 downto 0);
  signal pcie_link_ctl : std_logic_vector(15 downto 0);
  signal pcie_link_status : std_logic_vector(15 downto 0);

  -- for internal use
  signal addr0 : std_logic_vector(11 downto 0);

begin

  cfg_bar0_out <= cfg_bar0;
  cfg_bar1_out <= cfg_bar1;
  cfg_bar2_out <= cfg_bar2;
  cfg_bar3_out <= cfg_bar3;
  cfg_bar4_out <= cfg_bar4;
  cfg_bar5_out <= cfg_bar5;
  cfg_bar_rom_out <= cfg_bar_rom;

  addr0 <= addr(11 downto 2) & "00";

  -- configuration read
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (rd = '1') then
        case addr0 is

          when x"000" =>
            dataout <= cfg_device_id & cfg_vendor_id;

          when x"004" =>
            dataout <= cfg_status & cfg_command;

          when x"008" =>
            dataout <= cfg_class_code & cfg_revision_id;

          when x"00c" =>
            dataout <= cfg_bist & cfg_header_type & cfg_latency_timer & cfg_cache_line_size;

          when x"010" =>
            dataout <= cfg_bar0;

          when x"014" =>
            dataout <= cfg_bar1;

          when x"018" =>
            dataout <= cfg_bar2;

          when x"01c" =>
            dataout <= cfg_bar3;

          when x"020" =>
            dataout <= cfg_bar4;

          when x"024" =>
            dataout <= cfg_bar5;

          when x"028" =>
            dataout <= cfg_cardbus_cis_ptr;

          when x"02c" =>
            dataout <= cfg_sub_device_id & cfg_sub_vendor_id;

          when x"030" =>
            dataout <= cfg_bar_rom;

          when x"034" =>
            dataout <= conv_std_logic_vector(0, 24) & cfg_cap_ptr;

          when x"038" =>
            dataout <= conv_std_logic_vector(0, 32);

          when x"03c" =>
            dataout <= cfg_max_gnt & cfg_min_gnt & cfg_int_pin & cfg_int_line;

          when x"080" =>
            dataout <= pcie_cap & pcie_cap_ptr & pcie_cap_id;

          when x"084" =>
            dataout <= pcie_device_cap;

          when x"088" =>
            dataout <= pcie_device_status & pcie_device_ctl;

          when x"08c" =>
            dataout <= pcie_link_cap;

          when x"090" =>
            dataout <= pcie_link_status & pcie_link_ctl;

          when x"100" =>
            dataout <= (others => '0');

          when others =>
            dataout <= (others => '1');

        end case;
      end if;
    end if;
  end process;

  -- configuration write
  process (clk, rst)

    variable mask16 : std_logic_vector(15 downto 0);
    variable mask32 : std_logic_vector(31 downto 0);

  begin
    if (rst = '1') then

      cfg_vendor_id <= CFG_VENDOR_ID_INIT;
      cfg_device_id <= CFG_DEVICE_ID_INIT;
      cfg_command <= CFG_COMMAND_INIT;
      cfg_status <= x"0010";
      -- bit 04 cap_list present bit. shoud be '1'.
      -- bit 15:08 error status bits. all set to zero for now.
      -- other bits should always be zero.

      cfg_revision_id <= CFG_REVISION_ID_INIT;
      cfg_class_code <= CFG_CLASS_CODE_INIT;
      cfg_header_type <= CFG_HEADER_TYPE_INIT;
      cfg_cache_line_size <= x"00"; -- not used for PCIe, but should be writable.
      cfg_latency_timer <= x"00"; -- not used for PCIe. must be tied to 0.
      cfg_bist <= x"00"; -- do not support buit-in self-test.
      cfg_bar0 <= CFG_BAR0_INIT;
      cfg_bar1 <= CFG_BAR1_INIT;
      cfg_bar2 <= CFG_BAR2_INIT;
      cfg_bar3 <= CFG_BAR3_INIT;
      cfg_bar4 <= CFG_BAR4_INIT;
      cfg_bar5 <= CFG_BAR5_INIT;
      cfg_cardbus_cis_ptr <= x"00000000"; -- not used.
      cfg_sub_vendor_id <= CFG_SUB_VENDOR_ID_INIT;
      cfg_sub_device_id <= CFG_SUB_DEVICE_ID_INIT;
      cfg_bar_rom <= CFG_BAR_ROM_INIT;
      cfg_cap_ptr <= x"80"; -- points to top address of PCIe capability register.
      cfg_int_line <= x"00"; -- set by RC. 00:IRQ0 01:IRQ1 ... 0f:IRQ15
      cfg_int_pin <= CFG_INT_PIN_INIT;
      cfg_min_gnt <= x"00"; -- not used for PCIe.
      cfg_max_gnt <= x"00"; -- not used for PCIe.

      pcie_cap_id <= x"10"; -- capability ID. 10:PCIe capability register
      pcie_cap_ptr <= x"00"; -- end of linked list.
      pcie_cap <= PCIE_CAP_INIT;
      pcie_device_cap <= PCIE_DEVICE_CAP_INIT(31 downto 3) &
                             conv_std_logic_vector(mylog2(MAX_PAYLOAD) - 7, 3);
                             -- 000:128B 001:256B 010:512B 011:1kB 100:2kB 101:4kB

      pcie_device_ctl <= PCIE_DEVICE_CTL_INIT(15) &
                             conv_std_logic_vector(mylog2(MAX_READ_REQ_SIZE) - 7, 3) &
                             PCIE_DEVICE_CTL_INIT(11 downto 8) &
                             conv_std_logic_vector(0, 3) & -- initial value of max payload size should be set to 128B.
                             PCIE_DEVICE_CTL_INIT(4 downto 0);

      pcie_device_status <= PCIE_DEVICE_STATUS_INIT;

      pcie_link_cap <= PCIE_LINK_CAP_INIT(31 downto 10) &
                             conv_std_logic_vector(NLANE, 6) & -- max link width
                             MAX_LINK_SPEED;

      pcie_link_ctl <= PCIE_LINK_CTL_INIT;
      pcie_link_status <= PCIE_LINK_STATUS_INIT(15 downto 10) &
                             conv_std_logic_vector(NLANE, 6) & -- max link width
                             MAX_LINK_SPEED;

    elsif (rising_edge(clk)) then

      if (wr = '1') then
        case addr0 is

          when x"004" =>
            mask16 := be_to_mask16(be(1 downto 0)) and x"0547"; -- writable bits: 10,8,6,2,1,0
            cfg_command <= (datain(15 downto 0) and mask16) or (cfg_command and not mask16);

          when x"00c" =>
            if (be(0) = '1') then
              cfg_cache_line_size <= datain(7 downto 0);
            end if;

          when x"010" =>
            mask32 := be_to_mask32(be) and (CFG_BAR0_INIT(31 downto 7) & "0000000");
            cfg_bar0 <= (datain(31 downto 0) and mask32) or (cfg_bar0 and not mask32);

          when x"014" =>
            if (CFG_BAR0_INIT(2 downto 1) = "10") then -- 64-bit address
              cfg_bar1 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR1_INIT(31 downto 7) & "0000000");
              cfg_bar1 <= (datain(31 downto 0) and mask32) or (cfg_bar1 and not mask32);
            end if;

          when x"018" =>
            mask32 := be_to_mask32(be) and (CFG_BAR2_INIT(31 downto 7) & "0000000");
            cfg_bar2 <= (datain(31 downto 0) and mask32) or (cfg_bar2 and not mask32);

          when x"01c" =>
            if (CFG_BAR2_INIT(2 downto 1) = "10") then -- 64-bit address
              cfg_bar3 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR3_INIT(31 downto 7) & "0000000");
              cfg_bar3 <= (datain(31 downto 0) and mask32) or (cfg_bar3 and not mask32);
            end if;

          when x"020" =>
            mask32 := be_to_mask32(be) and (CFG_BAR4_INIT(31 downto 7) & "0000000");
            cfg_bar4 <= (datain(31 downto 0) and mask32) or (cfg_bar4 and not mask32);

          when x"024" =>
            if (CFG_BAR4_INIT(2 downto 1) = "10") then -- 64-bit address
              cfg_bar5 <= be_to_mask32(be) and datain(31 downto 0);
            else
              mask32 := be_to_mask32(be) and (CFG_BAR5_INIT(31 downto 7) & "0000000");
              cfg_bar5 <= (datain(31 downto 0) and mask32) or (cfg_bar5 and not mask32);
            end if;

          when x"030" =>
            mask32 := be_to_mask32(be) and (CFG_BAR_ROM_INIT(31 downto 11) & "00000000001");
            cfg_bar_rom <= (datain(31 downto 0) and mask32) or (cfg_bar_rom and not mask32);

          when x"03c" =>
            if (be(0) = '1') then
              cfg_int_line <= datain(7 downto 0);
            end if;

          when x"088" =>
              mask16 := be_to_mask16(be(1 downto 0)) and x"70e0";
              pcie_device_ctl <= (datain(15 downto 0) and mask16) or (pcie_device_ctl and not mask16);

          when x"090" =>
              mask16 := be_to_mask16(be(1 downto 0)) and x"70e0";
              pcie_link_ctl <= (datain(15 downto 0) and mask16) or (pcie_link_ctl and not mask16);

          when others =>
            -- written data is discarded.

        end case;
      end if;
    end if;
  end process;

  -- max payload size 000:128B ... 101:4kB
  process (clk)
  begin
    if (rising_edge(clk)) then
      case pcie_device_ctl(7 downto 5) is
        when "000" =>
          mpls <= conv_std_logic_vector(128, mpls'length);
        when "001" =>
          mpls <= conv_std_logic_vector(256, mpls'length);
        when "010" =>
          mpls <= conv_std_logic_vector(512, mpls'length);
        when "011" =>
          mpls <= conv_std_logic_vector(1024, mpls'length);
        when "100" =>
          mpls <= conv_std_logic_vector(2048, mpls'length);
        when "101" =>
          mpls <= conv_std_logic_vector(4096, mpls'length);
        when others =>
          mpls <= conv_std_logic_vector(128, mpls'length);
      end case;
    end if;
  end process;

  -- max read request size 000:128B ... 101:4kB
  process (clk)
  begin
    if (rising_edge(clk)) then
      case pcie_device_ctl(14 downto 12) is
        when "000" =>
          mrrs <= conv_std_logic_vector(128, mrrs'length);
        when "001" =>
          mrrs <= conv_std_logic_vector(256, mrrs'length);
        when "010" =>
          mrrs <= conv_std_logic_vector(512, mrrs'length);
        when "011" =>
          mrrs <= conv_std_logic_vector(1024, mrrs'length);
        when "100" =>
          mrrs <= conv_std_logic_vector(2048, mrrs'length);
        when "101" =>
          mrrs <= conv_std_logic_vector(4096, mrrs'length);
        when others =>
          mrrs <= conv_std_logic_vector(128, mrrs'length);
      end case;
    end if;
  end process;

  -- read completion boundary 0:64B 1:128B
  rcb <= pcie_link_ctl(3);

end default;

--
-- DMA Controller (for each channel)
--
-- dma_control[0] : we for dma_paddrlow_in.
-- [1] : we for dma_paddrhigh_in.
-- [2] : we for dma_laddr_in.
-- [3] : we for dma_size_in.
-- [4] : we for dma_param.
-- [5] : kickoff a new transfer.
-- [6] : abort currently running transfer.
--
-- dma_param[0] : 0:fifo mode 1:ram mode
-- [1] : not used.
-- [3:2] : read latency. 00:1clk 01:2clk 10:3clk 11:4clk
-- [7:4] : be for DW transfer. not used for burst transfer.
-- [8] : direction. 0 for MemRd burst. 1 for MemWr burst.
-- [10:9] : not used.
-- [13:11] : TC. tied to "000".
-- [15:14] : ATTR.
--
-- dma_status[3] : set to 1 if transfer is done.
-- [2:0] : verbose diagnostics of a transfer being done.
--
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity dmactl is
  generic (
    NLANE : integer;
    NDMACH : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;

    mpls : in std_logic_vector(11 downto 0); -- max payload size
    mrrs : in std_logic_vector(11+1 downto 0); -- max read request size

    -- interface to user-side logic:
    dma_control : in std_logic_vector(6 downto 0);
    dma_param : in std_logic_vector(15 downto 0);
    dma_status : out std_logic_vector(3 downto 0);
    dma_fifocnt : in std_logic_vector(12 downto 0);

    dma_paddrlow_in : in std_logic_vector(31 downto 0);
    dma_paddrhigh_in : in std_logic_vector(31 downto 0);
    dma_laddr_in : in std_logic_vector(31 downto 0);
    dma_size_in : in std_logic_vector(31 downto 0);

    dma_paddrlow_out : out std_logic_vector(31 downto 0);
    dma_paddrhigh_out : out std_logic_vector(31 downto 0);
    dma_laddr_out : out std_logic_vector(31 downto 0);
    dma_size_out : out std_logic_vector(31 downto 0);

    -- interface to master controller:
    dma_fifocnt_off : in std_logic_vector(4 downto 0); -- width >= log2(NLANE*2)
    wrreq : out std_logic;
    wrack : in std_logic;
    rdreq : out std_logic;
    rdack : in std_logic;

    length : out std_logic_vector(PAYLOAD_WIDTH-1 downto 0);
    paddr : out std_logic_vector(63 downto 0);
    laddr : out std_logic_vector(31 downto 0);

    -- interface to slave controller:
    rdcpl : in std_logic;
    rdcpl_length : in std_logic_vector(9 downto 0);
    rdcpl_bcnt : in std_logic_vector(11 downto 0);
    rdcpl_addr : out std_logic_vector(31 downto 0)
  );
end dmactl;

architecture default of dmactl is

  type DMA_STATE is (
    IDLE,

    WAIT_Wr,
    READY_Wr,
    BURST_Wr,
    CPL_Wr,

    WAIT_Rd,
    READY_Rd,
    BURST_Rd,
    CPL_Rd
  );

  signal dmasm : DMA_STATE;
  signal kickoff : std_logic;
  signal abort : std_logic;
  signal done : std_logic;
  signal bcnt : std_logic_vector(mpls'left + 1 downto 0);

  signal paddr_reg : std_logic_vector(63 downto 0);
  signal laddr_reg : std_logic_vector(31 downto 0);
  signal size_reg : std_logic_vector(31 downto 0);
  signal dir_reg : std_logic;

  signal paddr_reg_next : std_logic_vector(63 downto 0);
  signal laddr_reg_next : std_logic_vector(31 downto 0);
  signal size_reg_next : std_logic_vector(31 downto 0);
  signal dma_fifocnt_off0 : std_logic_vector(dma_fifocnt_off'range);

  signal rdcpl_addr0 : std_logic_vector(31 downto 0);
  signal wcnt : std_logic_vector(1 downto 0);

begin

  --
  -- DMA register read
  --
  dma_paddrhigh_out <= paddr_reg(63 downto 32);
  dma_paddrlow_out <= paddr_reg(31 downto 0);
  dma_laddr_out <= laddr_reg;
  dma_size_out <= size_reg;
  dma_status <= done & "000";
  paddr <= paddr_reg;
  laddr <= laddr_reg;

  process(clk, rst)
      variable bcnttmp : std_logic_vector(bcnt'range);
      variable fcnttmp : std_logic_vector(dma_fifocnt'range);
      variable blentmp : std_logic_vector(length'left + 3 downto 0);
  begin
    if (rst = '1') then
      abort <= '0';
      done <= '1';
      wrreq <= '0';
      rdreq <= '0';
      dmasm <= IDLE;
      dma_fifocnt_off0 <= (others => '0');
      wcnt <= conv_std_logic_vector(0, wcnt'length);
    elsif (rising_edge(clk)) then

      kickoff <= '0';

      case dmasm is
        when IDLE =>
          --
          -- DMA register write
          --
          if (dma_control(0) = '1') then
            paddr_reg(31 downto 0) <= dma_paddrlow_in;
            paddr_reg_next(31 downto 0) <= dma_paddrlow_in;
          end if;

          if (dma_control(1) = '1') then
            paddr_reg(63 downto 32) <= dma_paddrhigh_in;
            paddr_reg_next(63 downto 32) <= dma_paddrhigh_in;
          else
            paddr_reg_next <= paddr_reg;
          end if;

          if (dma_control(2) = '1') then
            laddr_reg <= dma_laddr_in;
            laddr_reg_next <= dma_laddr_in;
          else
            laddr_reg_next <= laddr_reg;
          end if;

          if (dma_control(3) = '1') then
            size_reg <= dma_size_in;
            size_reg_next <= dma_size_in;
          else
            size_reg_next <= size_reg;
          end if;

          if (dma_control(4) = '1') then
            dir_reg <= dma_param(8);
          end if;

          if (dma_control(5) = '1') then
            kickoff <= '1';
          end if;

          if (dma_control(6) = '1') then
            abort <= '1';
          end if;

          if (kickoff = '1') then
            done <= '0';
            if (dir_reg = '1') then
              dmasm <= WAIT_Wr;
            else
              dmasm <= WAIT_Rd;
            end if;
          end if;

          dma_fifocnt_off0 <= (others => '0');

        when WAIT_Wr =>

          -- determine length of next burst.
          if (unsigned(size_reg_next) < unsigned(mpls)) then
            bcnttmp := size_reg_next(bcnt'range);
          else
            bcnttmp := '0' & mpls;
          end if;
          bcnt <= bcnttmp;

          -- issue request for burst write when data is ready.
          fcnttmp := unsigned(dma_fifocnt) - unsigned(dma_fifocnt_off0);
          if (unsigned(bcnttmp) <= unsigned(fcnttmp)) then
            wrreq <= '1';
            length <= bcnttmp(bcnttmp'left - 1 downto 2);  -- size in DW unit. note that MSB is discatded,
                                                           -- i.e., length=0 denotes 4096B.
            dma_fifocnt_off0 <= (others => '0');
            dmasm <= READY_Wr;
          end if;

          if (dma_fifocnt_off0 /= conv_std_logic_vector(0, dma_fifocnt_off0'length)) then
            dma_fifocnt_off0 <= unsigned(dma_fifocnt_off0) - NLANE * 2;
          end if;

          size_reg <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;

        when READY_Wr =>
          if (wrack = '1') then
            wrreq <= '0';
            dmasm <= BURST_Wr;
          end if;

        when BURST_Wr =>
          if (wrack = '0') then
            dma_fifocnt_off0 <= dma_fifocnt_off;

            if (unsigned(size_reg) > unsigned(bcnt)) then
              dmasm <= WAIT_Wr;
            else
              done <= '1';
              dmasm <= CPL_Wr;
            end if;
          end if;

          size_reg_next <= unsigned(size_reg) - unsigned(bcnt);
          paddr_reg_next <= unsigned(paddr_reg) + unsigned(bcnt);
          laddr_reg_next <= unsigned(laddr_reg) + unsigned(bcnt);

        when CPL_Wr =>
          size_reg <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          dmasm <= IDLE;

        when WAIT_Rd =>

          -- determine length of next burst.
          if (unsigned(size_reg_next) < unsigned(mrrs)) then
            bcnttmp := size_reg_next(bcnt'range);
          else
            bcnttmp := mrrs(bcnt'range);
          end if;
          bcnt <= bcnttmp;

          -- issue a MemRd TLP when rx fifo has enough room.
          if (unsigned(bcnttmp) <= unsigned(dma_fifocnt)) then
            rdreq <= '1';
            length <= bcnttmp(bcnttmp'left - 1 downto 2);  -- size in DW unit. note that MSB is discatded,
                                                           -- i.e., length=0 denotes 4096B.
            dmasm <= READY_Rd;
          end if;

          size_reg <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          rdcpl_addr0 <= laddr_reg_next;

        when READY_Rd =>
          if (rdack = '1') then
            rdreq <= '0';
            dmasm <= BURST_Rd;
          end if;

        when BURST_Rd =>
          -- received a completion TLP for current request.
          if (rdcpl = '1') then
            blentmp := '0' & rdcpl_length & "00";
            if (rdcpl_length = conv_std_logic_vector(0, rdcpl_length'length)) then
              blentmp := (blentmp'left => '1', others => '0'); -- length=0 denotes 4096B.
            end if;
            rdcpl_addr0 <= unsigned(rdcpl_addr0) + unsigned(blentmp);

            -- received all completion TLPs for current request.
            bcnttmp := '0' & rdcpl_bcnt;
            if (rdcpl_bcnt = conv_std_logic_vector(0, rdcpl_bcnt'length)) then
              bcnttmp := (bcnttmp'left => '1', others => '0'); -- bcnt=0 denotes 4096B.
            end if;
            if (unsigned(bcnttmp) <= unsigned(blentmp)) then
              if (unsigned(size_reg) > unsigned(bcnt)) then
                dmasm <= WAIT_Rd;
              else
                done <= '1';
                dmasm <= CPL_Rd;
              end if;
            end if;
          end if;

          size_reg_next <= unsigned(size_reg) - unsigned(bcnt);
          paddr_reg_next <= unsigned(paddr_reg) + unsigned(bcnt);
          laddr_reg_next <= unsigned(laddr_reg) + unsigned(bcnt);

        when CPL_Rd =>
          size_reg <= size_reg_next;
          paddr_reg <= paddr_reg_next;
          laddr_reg <= laddr_reg_next;
          dmasm <= IDLE;

        when others =>

      end case;

      -- abort DMA after currently running burst is completed.
      if (abort = '1') then
        if (dmasm = IDLE or dmasm = WAIT_Wr or dmasm = WAIT_Rd) then
          abort <= '0';
          done <= '1';
          wrreq <= '0';
          rdreq <= '0';
          dmasm <= IDLE;
        end if;
      end if;

    end if; -- clk
  end process;

end default;
--
-- Transaction Layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
use work.msslvpkg.all;

entity transaction is

  generic (
    NLANE : integer;
    NVC : integer;
    NDMACH : integer;

    MAX_PAYLOAD : integer;

    CA_PH_VC0_INIT : integer;
    CA_NPH_VC0_INIT : integer;
    CA_CH_VC0_INIT : integer;
    CA_PD_VC0_INIT : integer;
    CA_NPD_VC0_INIT : integer;
    CA_CD_VC0_INIT : integer;

    CL_PH_VC0_INIT : integer;
    CL_NPH_VC0_INIT : integer;
    CL_CH_VC0_INIT : integer;
    CL_PD_VC0_INIT : integer;
    CL_NPD_VC0_INIT : integer;
    CL_CD_VC0_INIT : integer;

    CFG_BAR0_INIT : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0);

    SIMULATION : integer
  );
  port (

    triggerout : out std_logic;


    txpd_usedw : out std_logic_vector(8 downto 0);
    txpd_full : out std_logic;
    txpd_ovflw : out std_logic;
    txpd_undflw : out std_logic;
    rxempty_nph : out std_logic;
    rxempty_ph : out std_logic;
    rxempty_ch : out std_logic;
    rxfull_ph : out std_logic;
    rxfull_pd : out std_logic;


    clk : in std_logic;
    rst : in std_logic; -- active high

    -- TL-PHYMAC interface
    txreq : out std_logic;
    txack : in std_logic;

    -- TL-DLL interface
    rxtop : in std_logic;
    rxwr : in std_logic;
    rx : in std_logic_vector(NLANE*16-1 downto 0);

    txtop : out std_logic;
    txwr : out std_logic;
    txbe : out std_logic_vector(NLANE*2-1 downto 0);
    tx : out std_logic_vector(NLANE*16-1 downto 0);
    tpsize : out std_logic_vector(OFFSET_WIDTH-1 downto 0);

    -- notify CA to data-link layer
    ca_nph : out std_logic_vector(7 downto 0);
    ca_ph : out std_logic_vector(7 downto 0);
    ca_ch : out std_logic_vector(7 downto 0);
    ca_npd : out std_logic_vector(11 downto 0);
    ca_pd : out std_logic_vector(11 downto 0);
    ca_cd : out std_logic_vector(11 downto 0);
    ca_p_flush : out std_logic;
    ca_np_flush : out std_logic;
    ca_c_flush : out std_logic;

    -- CL notified by data-link layer
    cl_nph : in std_logic_vector( 7 downto 0);
    cl_ph : in std_logic_vector( 7 downto 0);
    cl_ch : in std_logic_vector( 7 downto 0);
    cl_npd : in std_logic_vector(11 downto 0);
    cl_pd : in std_logic_vector(11 downto 0);
    cl_cd : in std_logic_vector(11 downto 0);

    inf_cl_ph : in std_logic;
    inf_cl_pd : in std_logic;
    inf_cl_nph : in std_logic;
    inf_cl_npd : in std_logic;
    inf_cl_ch : in std_logic;
    inf_cl_cd : in std_logic;

    -- AL-TL interface
    mpls : in std_logic_vector(11 downto 0);

    cfg_bar0 : in std_logic_vector(31 downto 0);
    cfg_bar1 : in std_logic_vector(31 downto 0);
    cfg_bar2 : in std_logic_vector(31 downto 0);
    cfg_bar3 : in std_logic_vector(31 downto 0);
    cfg_bar4 : in std_logic_vector(31 downto 0);
    cfg_bar5 : in std_logic_vector(31 downto 0);
    cfg_bar_rom : in std_logic_vector(31 downto 0);

    cfg_read : out std_logic;
    cfg_write : out std_logic;
    cfg_addr : out std_logic_vector(11 downto 0);
    cfg_be : out std_logic_vector(3 downto 0);
    cfg_dataout : out std_logic_vector(31 downto 0);
    cfg_datain : in std_logic_vector(31 downto 0);

    -- core interface
    slv_read : out std_logic;
    slv_write : out std_logic;
    slv_bar : out std_logic_vector(6 downto 0);
    slv_addr : out std_logic_vector(63 downto 0);
    slv_bytevalid : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain : in std_logic_vector(NLANE*16-1 downto 0);

    slv_readreq : out std_logic;
    slv_writereq : out std_logic;
    slv_accept : in std_logic;
    slv_bytecount : out std_logic_vector(11 downto 0);

    ms_write : out std_logic;
    ms_wrchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_wraddr : out std_logic_vector(31 downto 0);
    ms_wrdata : in std_logic_vector(NLANE*16-1 downto 0);

    ms_read : out std_logic;
    ms_rdchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_rdaddr : out std_logic_vector(31 downto 0);
    ms_rddata : out std_logic_vector(NLANE*16-1 downto 0);

    dma_fifocnt_off : out std_logic_vector(4 downto 0);
    dmach_wrreq : in std_logic_vector(NDMACH-1 downto 0);
    dmach_wrack : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdreq : in std_logic_vector(NDMACH-1 downto 0);
    dmach_rdack : out std_logic_vector(NDMACH-1 downto 0);
    dmach_length : in each10b(NDMACH-1 downto 0);
    dmach_paddr : in each64b(NDMACH-1 downto 0);
    dmach_laddr : in each32b(NDMACH-1 downto 0);

    dmach_rdcpl : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdcpllength : out std_logic_vector(9 downto 0);
    dmach_rdcplbcnt : out std_logic_vector(11 downto 0);
    dmach_rdcpladdr : in each32b(NDMACH-1 downto 0);

    phy_linkup : in std_logic;
    dl_linkup : in std_logic
  );
end transaction;

architecture default of transaction is

  component rxctl
    generic (
      NLANE : integer := NLANE
    );
    port (
      clk : in std_logic;
      rst : in std_logic;

      rxtop : in std_logic;
      rxwr : in std_logic;
      rx : in std_logic_vector(NLANE*16-1 downto 0);

      vc : out std_logic_vector(2 downto 0);
      rxwr_bufsel : out std_logic;
      bufsel : out std_logic_vector(1 downto 0);

      rxwr_header : out std_logic;
      header : out std_logic_vector(127 downto 0);

      rxwr_data : out std_logic;
      data : out std_logic_vector(NLANE*16-1 downto 0)
    );
  end component;

  component txctl
    generic (
      NLANE : integer := NLANE;
      CL_PH_VC0_INIT : integer := CL_PH_VC0_INIT;
      CL_NPH_VC0_INIT : integer := CL_NPH_VC0_INIT;
      CL_CH_VC0_INIT : integer := CL_CH_VC0_INIT;
      CL_PD_VC0_INIT : integer := CL_PD_VC0_INIT;
      CL_NPD_VC0_INIT : integer := CL_NPD_VC0_INIT;
      CL_CD_VC0_INIT : integer := CL_CD_VC0_INIT
    );
    port (
      clk : in std_logic;
      rst : in std_logic;

      -- CL notified by data-link layer
      cl_header : in std_logic_vector( 7 downto 0);
      cl_data : in std_logic_vector(11 downto 0);
      inf_cl_header : in std_logic;
      inf_cl_data : in std_logic;

      -- notify TxFCbuf CA to slvctl
      cat_nph : out std_logic_vector(7 downto 0);
      cat_ph : out std_logic_vector(7 downto 0);
      cat_ch : out std_logic_vector(7 downto 0);
      cat_npd : out std_logic_vector(11 downto 0);
      cat_pd : out std_logic_vector(11 downto 0);
      cat_cd : out std_logic_vector(11 downto 0);

      hbufsel : out std_logic_vector(1 downto 0);
      dbufsel : out std_logic_vector(1 downto 0);

      empty_bufsel : in std_logic;
      empty_header : in std_logic;
      empty_data : in std_logic;

      rd_bufsel : out std_logic;
      rd_header : out std_logic;
      rd_data : out std_logic;

      q_bufsel : in std_logic_vector(1 downto 0);
      q_header : in std_logic_vector(127 downto 0);
      q_data : in std_logic_vector(NLANE*16-1 downto 0);

      txreq : out std_logic;
      txack : in std_logic;
      txtop : out std_logic;
      txwr : out std_logic;
      txbe : out std_logic_vector(NLANE*2-1 downto 0);
      tpsize : out std_logic_vector(OFFSET_WIDTH-1 downto 0);
      tx : out std_logic_vector(NLANE*16-1 downto 0)
    );
  end component;

  component txarb
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high

      c_req : in std_logic;
      c_ack : out std_logic;

      p_req : in std_logic;
      p_ack : out std_logic;

      np_req : in std_logic;
      np_ack : out std_logic;

      srcsel : out std_logic_vector(1 downto 0) -- TPBUFTYPE_P, _NP, or _C
    );
  end component;

  component slvctl
    generic (
      CA_PH_VC0_INIT : integer := CA_PH_VC0_INIT;
      CA_NPH_VC0_INIT : integer := CA_NPH_VC0_INIT;
      CA_CH_VC0_INIT : integer := CA_CH_VC0_INIT;
      CA_PD_VC0_INIT : integer := CA_PD_VC0_INIT;
      CA_NPD_VC0_INIT : integer := CA_NPD_VC0_INIT;
      CA_CD_VC0_INIT : integer := CA_CD_VC0_INIT;

      CFG_BAR0_INIT : std_logic_vector(31 downto 0) := CFG_BAR0_INIT;
      CFG_BAR1_INIT : std_logic_vector(31 downto 0) := CFG_BAR1_INIT;
      CFG_BAR2_INIT : std_logic_vector(31 downto 0) := CFG_BAR2_INIT;
      CFG_BAR3_INIT : std_logic_vector(31 downto 0) := CFG_BAR3_INIT;
      CFG_BAR4_INIT : std_logic_vector(31 downto 0) := CFG_BAR4_INIT;
      CFG_BAR5_INIT : std_logic_vector(31 downto 0) := CFG_BAR5_INIT;
      CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0) := CFG_BAR_ROM_INIT;
      NLANE : integer := NLANE;
      NDMACH : integer := NDMACH
    );
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high

      ca_nph : out std_logic_vector(7 downto 0);
      ca_ph : out std_logic_vector(7 downto 0);
      ca_ch : out std_logic_vector(7 downto 0);
      ca_npd : out std_logic_vector(11 downto 0);
      ca_pd : out std_logic_vector(11 downto 0);
      ca_cd : out std_logic_vector(11 downto 0);
      ca_p_flush : out std_logic;
      ca_np_flush : out std_logic;
      ca_c_flush : out std_logic;

      mpls : in std_logic_vector(11 downto 0); -- max payload size

      cfg_bar0 : in std_logic_vector(31 downto 0);
      cfg_bar1 : in std_logic_vector(31 downto 0);
      cfg_bar2 : in std_logic_vector(31 downto 0);
      cfg_bar3 : in std_logic_vector(31 downto 0);
      cfg_bar4 : in std_logic_vector(31 downto 0);
      cfg_bar5 : in std_logic_vector(31 downto 0);
      cfg_bar_rom : in std_logic_vector(31 downto 0);

      rxhbufsel : out std_logic_vector(1 downto 0);
      rxdbufsel : out std_logic_vector(1 downto 0);
      rxempty_bufsel : in std_logic;

      rxrd_bufsel : out std_logic;
      rxrd_header : out std_logic;
      rxrd_data : out std_logic;

      rxq_bufsel : in std_logic_vector(1 downto 0);
      rxq_header : in std_logic_vector(127 downto 0);
      rxq_data : in std_logic_vector(NLANE*16-1 downto 0);

      txvc : out std_logic_vector(2 downto 0);

      cat_header : in std_logic_vector(7 downto 0);
      cat_data : in std_logic_vector(11 downto 0);

      bdf : out std_logic_vector(15 downto 0);

      txwr_req : out std_logic;
      txwr_ack : in std_logic;
      txwr_bufsel : out std_logic;
      txwr_header : out std_logic;
      txwr_data : out std_logic;

      tx_header : out std_logic_vector(127 downto 0);
      tx_data : out std_logic_vector(NLANE*16-1 downto 0);

      slv_read : out std_logic;
      slv_write : out std_logic;
      slv_bar : out std_logic_vector(6 downto 0);
      slv_addr : out std_logic_vector(63 downto 0);
      slv_bytevalid : out std_logic_vector(NLANE*2-1 downto 0);
      slv_dataout : out std_logic_vector(NLANE*16-1 downto 0);
      slv_datain : in std_logic_vector(NLANE*16-1 downto 0);

      slv_readreq : out std_logic;
      slv_writereq : out std_logic;
      slv_accept : in std_logic;
      slv_bytecount : out std_logic_vector(11 downto 0);

      ms_read : out std_logic;
      ms_rdchannel : out std_logic_vector(NDMACH-1 downto 0);
      ms_rdaddr : out std_logic_vector(31 downto 0);
      ms_rddata : out std_logic_vector(NLANE*16-1 downto 0);

      dmach_rdcpl : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdcpllength : out std_logic_vector(9 downto 0);
      dmach_rdcplbcnt : out std_logic_vector(11 downto 0);
      dmach_rdcpladdr : in each32b(NDMACH-1 downto 0);

      cfg_read : out std_logic;
      cfg_write : out std_logic;
      cfg_addr : out std_logic_vector(11 downto 0);
      cfg_be : out std_logic_vector(3 downto 0);
      cfg_dataout : out std_logic_vector(31 downto 0);
      cfg_datain : in std_logic_vector(31 downto 0)
    );
  end component;

  component msctl
    generic (
      NLANE : integer := NLANE;
      NDMACH : integer := NDMACH
    );
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high

      cat_ph : in std_logic_vector(7 downto 0);
      cat_pd : in std_logic_vector(11 downto 0);
      cat_nph : in std_logic_vector(7 downto 0);
      cat_npd : in std_logic_vector(11 downto 0);
      bdf : in std_logic_vector(15 downto 0);

      -- interface to TxFCbuf and txarb
      p_txwr_req : out std_logic;
      p_txwr_ack : in std_logic;
      p_txwr_bufsel : out std_logic;
      txwr_ph : out std_logic;
      tx_ph : out std_logic_vector(127 downto 0);
      txwr_pd : out std_logic;
      tx_pd : out std_logic_vector(NLANE*16-1 downto 0);

      np_txwr_req : out std_logic;
      np_txwr_ack : in std_logic;
      np_txwr_bufsel : out std_logic;
      txwr_nph : out std_logic;
      tx_nph : out std_logic_vector(127 downto 0);
      txwr_npd : out std_logic;
      tx_npd : out std_logic_vector(NLANE*16-1 downto 0);

      -- interface to master controllers in the user-side logic.
      ms_write : out std_logic;
      ms_wrchannel : out std_logic_vector(NDMACH-1 downto 0);
      ms_wraddr : out std_logic_vector(31 downto 0);
      ms_wrdata : in std_logic_vector(NLANE*16-1 downto 0);

      -- interface to DMA controllers
      dma_fifocnt_off : out std_logic_vector(4 downto 0);
      dmach_wrreq : in std_logic_vector(NDMACH-1 downto 0);
      dmach_wrack : out std_logic_vector(NDMACH-1 downto 0);
      dmach_rdreq : in std_logic_vector(NDMACH-1 downto 0);
      dmach_rdack : out std_logic_vector(NDMACH-1 downto 0);
      dmach_length : in each10b(NDMACH-1 downto 0);
      dmach_paddr : in each64b(NDMACH-1 downto 0);
      dmach_laddr : in each32b(NDMACH-1 downto 0)
    );
  end component;

  component fcbuf
    generic (
      VIRTUAL_CHANNEL : integer;
      DATA_WIDTH : integer;
      DATA_DEPTH : integer
    );
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high reset

      vc : in std_logic_vector(2 downto 0);
      wr : in std_logic;
      rd : in std_logic;
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);

      empty : out std_logic;
      full : out std_logic;
      q : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
  end component;

  component fcbuf2
    generic (
      VIRTUAL_CHANNEL : integer;
      DATA_WIDTH : integer;
      DATA_DEPTH : integer
    );
    port (
      clk : in std_logic;
      rst : in std_logic; -- active high reset

      vc : in std_logic_vector(2 downto 0);
      wr : in std_logic;
      rd : in std_logic;
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);

      empty : out std_logic;
      full : out std_logic;
      usedw : out std_logic_vector(8 downto 0);
      q : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
  end component;

  signal rxvc , txvc : std_logic_vector(2 downto 0);

  --
  -- selection buffer control lines
  --
  signal rxrd_bufsel , txrd_bufsel : std_logic;
  signal rxempty_bufsel , txempty_bufsel : std_logic;
  signal rx_bufselin , tx_bufselin : std_logic_vector(1 downto 0);
  signal tx_hbufselout : std_logic_vector(1 downto 0);
  signal tx_dbufselout : std_logic_vector(1 downto 0);
  signal rx_hbufselout , rx_dbufselout : std_logic_vector(1 downto 0);
  signal rxq_bufsel , txq_bufsel : std_logic_vector(1 downto 0);

  --
  -- header buffer control lines
  --
  signal rxwr_bufsel : std_logic;
  signal rxwr_header , txwr_header : std_logic;
  signal rxrd_header , txrd_header : std_logic;
  signal txempty_header : std_logic;
  signal rx_header : std_logic_vector(127 downto 0);
  signal rxq_header , txq_header : std_logic_vector(127 downto 0);
  signal tx_ch : std_logic_vector(127 downto 0);
  signal tx_cd : std_logic_vector(NLANE*16-1 downto 0);

  -- non-posted
  signal rxwr_nph , txwr_nph : std_logic;
  signal rxrd_nph , txrd_nph : std_logic;
  signal txempty_nph : std_logic;
  signal rxq_nph , txq_nph : std_logic_vector(127 downto 0);
  signal cat_nph : std_logic_vector(7 downto 0);

  -- posted
  signal rxwr_ph , txwr_ph : std_logic;
  signal rxrd_ph , txrd_ph : std_logic;
  signal txempty_ph : std_logic;
  signal rxq_ph , txq_ph : std_logic_vector(127 downto 0);
  signal cat_ph : std_logic_vector(7 downto 0);

  -- completion
  signal rxwr_ch , txwr_ch : std_logic;
  signal rxrd_ch , txrd_ch : std_logic;
  signal txempty_ch : std_logic;
  signal rxq_ch , txq_ch : std_logic_vector(127 downto 0);
  signal cat_ch : std_logic_vector(7 downto 0);

  --
  -- data buffer control lines
  --
  signal rxwr_data , txwr_data : std_logic;
  signal rxrd_data , txrd_data : std_logic;
  signal txempty_data : std_logic;
  signal rx_data : std_logic_vector(NLANE*16-1 downto 0);
  signal rxq_data , txq_data : std_logic_vector(NLANE*16-1 downto 0);

  -- non-posted
  signal rxwr_npd , txwr_npd : std_logic;
  signal rxrd_npd , txrd_npd : std_logic;
  signal txempty_npd : std_logic;
  signal rxq_npd , txq_npd : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_npd : std_logic_vector(11 downto 0);

  -- posted
  signal rxwr_pd , txwr_pd : std_logic;
  signal rxrd_pd , txrd_pd : std_logic;
  signal txempty_pd : std_logic;
  signal rxq_pd , txq_pd : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_pd : std_logic_vector(11 downto 0);

  -- completion
  signal rxwr_cd , txwr_cd : std_logic;
  signal rxrd_cd , txrd_cd : std_logic;
  signal txempty_cd : std_logic;
  signal rxq_cd , txq_cd : std_logic_vector(NLANE*16-1 downto 0);
  signal cat_cd : std_logic_vector(11 downto 0);

  signal cl_header : std_logic_vector( 7 downto 0);
  signal cl_data : std_logic_vector(11 downto 0);
  signal inf_cl_header : std_logic;
  signal inf_cl_data : std_logic;

  signal txwr_bufsel : std_logic;

  signal c_txwr_req : std_logic;
  signal c_txwr_ack : std_logic;
  signal c_txwr_bufsel : std_logic;

  signal p_txwr_req : std_logic;
  signal p_txwr_ack : std_logic;
  signal p_txwr_bufsel : std_logic;
  signal tx_ph : std_logic_vector(127 downto 0);
  signal tx_pd : std_logic_vector(NLANE*16-1 downto 0);

  signal np_txwr_req : std_logic;
  signal np_txwr_ack : std_logic;
  signal np_txwr_bufsel : std_logic;
  signal tx_nph : std_logic_vector(127 downto 0);
  signal tx_npd : std_logic_vector(NLANE*16-1 downto 0);

  signal tx_srcsel : std_logic_vector(1 downto 0);
  signal bdf : std_logic_vector(15 downto 0);


  -- for debugging
  signal txpd_full0 : std_logic;

  attribute keep : boolean;
  attribute keep of txpd_full : signal is true;
  attribute keep of txpd_usedw : signal is true;

-- synthesis translate_off
  attribute keep of txpd_ovflw : signal is true;
  attribute keep of txpd_undflw : signal is true;

  attribute keep : boolean;
  attribute keep of txpd_full : signal is true;
  attribute keep of txrd_bufsel : signal is true;
  attribute keep of txempty_bufsel : signal is true;
  attribute keep of txrd_ph : signal is true;
  attribute keep of txempty_ph : signal is true;
  attribute keep of txrd_ch : signal is true;
  attribute keep of txempty_ch : signal is true;
  attribute keep of txwr_bufsel : signal is true;

  attribute keep of txq_bufsel : signal is true;
  attribute keep of rxfull_ph : signal is true;
  attribute keep of rxfull_pd : signal is true;
-- synthesis translate_on

begin

  -- Rx Flow Control Buffer Input Controller
  rxctl0 : rxctl
    port map (
      clk => clk,
      rst => rst,

      rxtop => rxtop,
      rxwr => rxwr,
      rx => rx,

      vc => rxvc,
      rxwr_bufsel => rxwr_bufsel,
      bufsel => rx_bufselin,

      rxwr_header => rxwr_header,
      rxwr_data => rxwr_data,

      header => rx_header,
      data => rx_data
    );

  -- Rx Flow Control Buffer Input Selector
  process (rx_bufselin, rxwr_header, rxwr_data)
  begin

    rxwr_nph <= '0';
    rxwr_npd <= '0';
    rxwr_ph <= '0';
    rxwr_pd <= '0';
    rxwr_ch <= '0';
    rxwr_cd <= '0';

    case rx_bufselin is

      when TPBUFTYPE_NP =>
        rxwr_nph <= rxwr_header;
        rxwr_npd <= rxwr_data;

      when TPBUFTYPE_P =>
        rxwr_ph <= rxwr_header;
        rxwr_pd <= rxwr_data;

      when TPBUFTYPE_C =>
        rxwr_ch <= rxwr_header;
        rxwr_cd <= rxwr_data;

      when others =>

    end case;
  end process;

  -- Slave Controller
  slvctl0 : slvctl
    port map (
      clk => clk,
      rst => rst,

      ca_nph => ca_nph,
      ca_ph => ca_ph,
      ca_ch => ca_ch,
      ca_npd => ca_npd,
      ca_pd => ca_pd,
      ca_cd => ca_cd,
      ca_p_flush => ca_p_flush,
      ca_np_flush => ca_np_flush,
      ca_c_flush => ca_c_flush,

      mpls => mpls,

      cfg_bar0 => cfg_bar0,
      cfg_bar1 => cfg_bar1,
      cfg_bar2 => cfg_bar2,
      cfg_bar3 => cfg_bar3,
      cfg_bar4 => cfg_bar4,
      cfg_bar5 => cfg_bar5,
      cfg_bar_rom => cfg_bar_rom,

      rxhbufsel => rx_hbufselout,
      rxdbufsel => rx_dbufselout,
      rxempty_bufsel => rxempty_bufsel,

      rxrd_bufsel => rxrd_bufsel,
      rxrd_header => rxrd_header,
      rxrd_data => rxrd_data,

      rxq_bufsel => rxq_bufsel,
      rxq_header => rxq_header,
      rxq_data => rxq_data,

      txvc => txvc,

      cat_header => cat_ch,
      cat_data => cat_cd,

      bdf => bdf,

      txwr_req => c_txwr_req,
      txwr_ack => c_txwr_ack,
      txwr_bufsel => c_txwr_bufsel,
      txwr_header => txwr_ch,
      txwr_data => txwr_cd,

      tx_header => tx_ch,
      tx_data => tx_cd,

      slv_read => slv_read,
      slv_write => slv_write,
      slv_bar => slv_bar,
      slv_addr => slv_addr,
      slv_bytevalid => slv_bytevalid,
      slv_dataout => slv_dataout,
      slv_datain => slv_datain,

      slv_readreq => slv_readreq,
      slv_writereq => slv_writereq,
      slv_accept => slv_accept,
      slv_bytecount => slv_bytecount,

      ms_read => ms_read,
      ms_rdchannel => ms_rdchannel,
      ms_rdaddr => ms_rdaddr,
      ms_rddata => ms_rddata,

      dmach_rdcpl => dmach_rdcpl,
      dmach_rdcpllength => dmach_rdcpllength,
      dmach_rdcplbcnt => dmach_rdcplbcnt,
      dmach_rdcpladdr => dmach_rdcpladdr,

      cfg_read => cfg_read,
      cfg_write => cfg_write,
      cfg_addr => cfg_addr,
      cfg_be => cfg_be,
      cfg_dataout => cfg_dataout,
      cfg_datain => cfg_datain
    );

  -- Rx Header Flow Control Buffer Output Selector
  process (rx_hbufselout, rxrd_header, rxq_nph, rxq_ph, rxq_ch)
  begin

    rxrd_nph <= '0';
    rxrd_ph <= '0';
    rxrd_ch <= '0';

    case rx_hbufselout is
      when TPBUFTYPE_NP =>
        rxq_header <= rxq_nph;
        rxrd_nph <= rxrd_header;

      when TPBUFTYPE_P =>
        rxq_header <= rxq_ph;
        rxrd_ph <= rxrd_header;

      when TPBUFTYPE_C =>
        rxq_header <= rxq_ch;
        rxrd_ch <= rxrd_header;

      when others =>
        rxq_header <= (others => '0');

    end case;
  end process;

  -- Rx Data Flow Control Buffer Output Selector
  process (rx_dbufselout, rxrd_data, rxq_npd, rxq_pd, rxq_cd)
  begin

    rxrd_npd <= '0';
    rxrd_pd <= '0';
    rxrd_cd <= '0';

    case rx_dbufselout is

      when TPBUFTYPE_NP =>
        rxq_data <= rxq_npd;
        rxrd_npd <= rxrd_data;

      when TPBUFTYPE_P =>
        rxq_data <= rxq_pd;
        rxrd_pd <= rxrd_data;

      when TPBUFTYPE_C =>
        rxq_data <= rxq_cd;
        rxrd_cd <= rxrd_data;

      when others =>
        rxq_data <= (others => '0');

    end case;
  end process;

  -- Tx Flow Control Buffer Output Controller
  txctl0 : txctl
    port map (
      clk => clk,
      rst => rst,

      cl_header => cl_header,
      cl_data => cl_data,
      inf_cl_header => inf_cl_header,
      inf_cl_data => inf_cl_data,

      cat_nph => cat_nph,
      cat_ph => cat_ph,
      cat_ch => cat_ch,
      cat_npd => cat_npd,
      cat_pd => cat_pd,
      cat_cd => cat_cd,

      hbufsel => tx_hbufselout,
      dbufsel => tx_dbufselout,

      empty_bufsel => txempty_bufsel,
      empty_header => txempty_header,
      empty_data => txempty_data,

      rd_bufsel => txrd_bufsel,
      rd_header => txrd_header,
      rd_data => txrd_data,

      q_bufsel => txq_bufsel,
      q_header => txq_header,
      q_data => txq_data,

      txreq => txreq,
      txack => txack,
      txtop => txtop,
      txwr => txwr,
      txbe => txbe,
      tpsize => tpsize,
      tx => tx
    );

  -- Tx Header Flow Control Buffer Output Selector
  process (tx_hbufselout, txrd_header,
           txempty_nph, txq_nph,
           txempty_ph , txq_ph,
           txempty_ch , txq_ch)
  begin

    txrd_nph <= '0';
    txrd_ph <= '0';
    txrd_ch <= '0';

    case tx_hbufselout is

      when TPBUFTYPE_NP =>
        txempty_header <= txempty_nph;
        txq_header <= txq_nph;

        txrd_nph <= txrd_header;

      when TPBUFTYPE_P =>
        txempty_header <= txempty_ph;
        txq_header <= txq_ph;

        txrd_ph <= txrd_header;

      when TPBUFTYPE_C =>
        txempty_header <= txempty_ch;
        txq_header <= txq_ch;

        txrd_ch <= txrd_header;

      when others =>
        txempty_header <= '1';
        txq_header <= (others => '0');

    end case;
  end process;


  -- Tx Data Flow Control Buffer Output Selector
  process (tx_dbufselout, txrd_data,
           txempty_npd, txq_npd,
           txempty_pd, txq_ph,
           txempty_cd, txq_ch,
           cl_nph, cl_npd, inf_cl_nph, inf_cl_npd,
           cl_ph , cl_pd , inf_cl_ph , inf_cl_pd,
           cl_ch , cl_cd , inf_cl_ch , inf_cl_cd)

  begin

    txrd_npd <= '0';
    txrd_pd <= '0';
    txrd_cd <= '0';

    case tx_dbufselout is

      when TPBUFTYPE_NP =>
        cl_header <= cl_nph;
        cl_data <= cl_npd;
        inf_cl_header <= inf_cl_nph;
        inf_cl_data <= inf_cl_npd;

        txempty_data <= txempty_npd;
        txq_data <= txq_npd;

        txrd_npd <= txrd_data;

      when TPBUFTYPE_P =>
        cl_header <= cl_ph;
        cl_data <= cl_pd;
        inf_cl_header <= inf_cl_ph;
        inf_cl_data <= inf_cl_pd;

        txempty_data <= txempty_pd;
        txq_data <= txq_pd;

        txrd_pd <= txrd_data;

      when TPBUFTYPE_C =>
        cl_header <= cl_ch;
        cl_data <= cl_cd;
        inf_cl_header <= inf_cl_ch;
        inf_cl_data <= inf_cl_cd;

        txempty_data <= txempty_cd;
        txq_data <= txq_cd;

        txrd_cd <= txrd_data;

      when others =>
        cl_header <= (others => '0');
        cl_data <= (others => '0');
        inf_cl_header <= '0';
        inf_cl_data <= '0';

        txempty_data <= '1';
        txq_data <= (others => '0');

    end case;
  end process;


  msctl0 : msctl
    port map (
      clk => clk,
      rst => rst,

      cat_ph => cat_ph,
      cat_pd => cat_pd,
      cat_nph => cat_nph,
      cat_npd => cat_npd,

      bdf => bdf,

      p_txwr_req => p_txwr_req,
      p_txwr_ack => p_txwr_ack,
      p_txwr_bufsel => p_txwr_bufsel,
      txwr_ph => txwr_ph,
      tx_ph => tx_ph,
      txwr_pd => txwr_pd,
      tx_pd => tx_pd,

      np_txwr_req => np_txwr_req,
      np_txwr_ack => np_txwr_ack,
      np_txwr_bufsel => np_txwr_bufsel,
      txwr_nph => txwr_nph,
      tx_nph => tx_nph,
      txwr_npd => txwr_npd,
      tx_npd => tx_npd,

      ms_wrchannel => ms_wrchannel,
      ms_write => ms_write,
      ms_wraddr => ms_wraddr,
      ms_wrdata => ms_wrdata,

      dma_fifocnt_off => dma_fifocnt_off,
      dmach_wrreq => dmach_wrreq,
      dmach_wrack => dmach_wrack,
      dmach_rdreq => dmach_rdreq,
      dmach_rdack => dmach_rdack,
      dmach_length => dmach_length,
      dmach_paddr => dmach_paddr,
      dmach_laddr => dmach_laddr
    );

  -- exclusively select write requests from slvctl (completion) or
  -- msctl (posted or non-posted).
  txarb0 : txarb
    port map (
      clk => clk,
      rst => rst,

      c_req => c_txwr_req,
      c_ack => c_txwr_ack,

      p_req => p_txwr_req,
      p_ack => p_txwr_ack,

      np_req => np_txwr_req,
      np_ack => np_txwr_ack,

      srcsel => tx_srcsel
    );

  process(tx_srcsel, c_txwr_bufsel, p_txwr_bufsel, np_txwr_bufsel)
  begin
    case (tx_srcsel) is
      when TPBUFTYPE_C =>
        txwr_bufsel <= c_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_C;

      when TPBUFTYPE_P =>
        txwr_bufsel <= p_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_P;

      when TPBUFTYPE_NP =>
        txwr_bufsel <= np_txwr_bufsel;
        tx_bufselin <= TPBUFTYPE_NP;

      when others =>

    end case;
  end process;

  --
  -- Flow Control Buffers
  --

  vc : for i in 0 to NVC-1 generate

    -- Rx Buffer Selection buffer
    rxselbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPH_VC0_INIT + CA_PH_VC0_INIT + CA_CH_VC0_INIT,
        DATA_WIDTH => rx_bufselin'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_bufsel,
        data => rx_bufselin,

        rd => rxrd_bufsel,
        empty => rxempty_bufsel,
        q => rxq_bufsel
      );

    -- Rx Non-Posted Header buffer
    rxnphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_nph,
        data => rx_header,

        rd => rxrd_nph,
        empty => rxempty_nph,
        q => rxq_nph
      );

    -- Rx Posted Header buffer
    rxphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_PH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_ph,
        data => rx_header,

        rd => rxrd_ph,
        empty => rxempty_ph,
        full => rxfull_ph,
        q => rxq_ph
      );

    -- Rx Completion Header buffer
    rxchbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_CH_VC0_INIT,
        DATA_WIDTH => rx_header'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_ch,
        data => rx_header,

        rd => rxrd_ch,
        empty => rxempty_ch,
        q => rxq_ch
      );

    -- Rx Non-Posted Data buffer
    rxnpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_NPD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_npd,
        data => rx_data,

        rd => rxrd_npd,
        empty => open,
        q => rxq_npd
      );

    -- Rx Posted Data buffer
    rxpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_PD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_pd,
        data => rx_data,

        rd => rxrd_pd,
        empty => open,
        q => rxq_pd
      );

    -- Rx Completion Data buffer
    rxcdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CA_CD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => rx_data'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => rxvc,
        wr => rxwr_cd,
        data => rx_data,

        rd => rxrd_cd,
        empty => open,
        full => rxfull_pd,
        q => rxq_cd
      );

    -- Tx Buffer Selection buffer
    txselbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_NPH_VC0_INIT + CL_PH_VC0_INIT + CL_CH_VC0_INIT,
        DATA_WIDTH => tx_bufselin'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_bufsel,
        data => tx_bufselin,

        rd => txrd_bufsel,
        empty => txempty_bufsel,
        q => txq_bufsel
      );

    -- Tx Non-Posted Header buffer
    txnphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_NPH_VC0_INIT,
        DATA_WIDTH => tx_nph'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_nph,
        data => tx_nph,

        rd => txrd_nph,
        empty => txempty_nph,
        q => txq_nph
      );

    -- Tx Posted Header buffer
    txphbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_PH_VC0_INIT,
        DATA_WIDTH => tx_ph'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_ph,
        data => tx_ph,

        rd => txrd_ph,
        empty => txempty_ph,
        q => txq_ph
      );

    -- Tx Completion Header buffer
    txchbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_CH_VC0_INIT,
        DATA_WIDTH => tx_ch'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_ch,
        data => tx_ch,

        rd => txrd_ch,
        empty => txempty_ch,
        q => txq_ch
      );

    -- Tx Non-Posted Data buffer
    txnpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_NPD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_npd'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_npd,
        data => tx_npd,

        rd => txrd_npd,
        empty => txempty_npd,
        q => txq_npd
      );

    -- Tx Posted Data buffer
    txpdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_PD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_pd'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_pd,
        data => tx_pd,

        rd => txrd_pd,
        empty => txempty_pd,
-- full => txpd_full0,
-- usedw => txpd_usedw,
        q => txq_pd
      );

-- txpd_full <= txpd_full0;
-- txpd_ovflw <= '1' when txpd_full0 = '1' and txwr_pd = '1' else '0';
-- txpd_undflw <= '1' when txempty_pd = '1' and txrd_pd = '1' else '0';

    -- Tx Completion Data buffer
    txcdbuf : fcbuf
      generic map (
        VIRTUAL_CHANNEL => i,
        DATA_DEPTH => CL_CD_VC0_INIT*16/(NLANE*2),
        DATA_WIDTH => tx_cd'length
      )
      port map (
        clk => clk,
        rst => rst,

        vc => txvc,
        wr => txwr_cd,
        data => tx_cd,

        rd => txrd_cd,
        empty => txempty_cd,
        q => txq_cd
      );

  end generate vc;

end default;


--
-- TLP receiver
-- parse incoming TLP and output its
-- header/data to Rx{NP,P,C}{H,D}buf,
-- according to its type.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity rxctl is
  generic (
    NLANE : integer
  );
  port (
-- cont : out std_logic;

    clk : in std_logic;
    rst : in std_logic;

    rxtop : in std_logic;
    rxwr : in std_logic; -- not used
    rx : in std_logic_vector(NLANE*16-1 downto 0);

    vc : out std_logic_vector(2 downto 0);
    rxwr_bufsel : out std_logic;
    bufsel : out std_logic_vector(1 downto 0);

    rxwr_header : out std_logic;
    header : out std_logic_vector(127 downto 0);

    rxwr_data : out std_logic;
    data : out std_logic_vector(NLANE*16-1 downto 0)
  );
end rxctl;

architecture default of rxctl is

  type array3x3 is array (7 downto 0) of std_logic_vector(2 downto 0);
  signal tc2vc : array3x3;

-- attribute keep : boolean;
-- attribute keep of cont : signal is true;

begin

  x1 : if (NLANE = 1) generate
    x1block : block

      type RXCTL_STATE is (
        IDLE,
        RECV_B2_B3,
        RECV_HEADRx1,
        RECV_DATA
      );

-- attribute enum_encoding: string;
-- attribute enum_encoding of RXCTL_STATE : type is "00 01 10 11";

      signal rxsm : RXCTL_STATE;
      signal have_data : std_logic;
      signal payload_left : std_logic_vector(11 downto 0); -- in byte unit
      signal header_left : std_logic_vector(3 downto 0);
      signal header_count : std_logic_vector(3 downto 0);
      signal short_header : std_logic;
      signal rx1, rx2 : std_logic_vector(NLANE*16-1 downto 0);
      signal header0 : std_logic_vector(127 downto 0);

    begin

      header <= header0;

      process (clk, rst)
      begin
        if (rst = '1') then
          rxwr_bufsel <= '0';
          rxwr_header <= '0';
          rxwr_data <= '0';
          rxsm <= IDLE;
        elsif (rising_edge(clk)) then

          rxwr_bufsel <= '0';
          rxwr_header <= '0';
          rxwr_data <= '0';
          rx1 <= rx;
          rx2 <= rx1;

          case rxsm is
            when IDLE =>

              -- TLP byte 0 & 1
-- if (rxwr = '1') then
              if (rxtop = '1') then -- !!! replayed TLPs are filtered here.

                -- TLP has data?
                if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA) or
                    (rx(0*8+6 downto 0*8+5) = TPFMT_16BH_DATA)) then
                  have_data <= '1';
                else
                  have_data <= '0';
                end if;

                -- header size
                if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_NODATA) or
                    (rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA)) then
                  short_header <= '1'; -- 12-byte header.
                  if (NLANE = 1) then
                    header_left <= conv_std_logic_vector(12-NLANE*2, header_left'length);
                  end if;
                else
                  short_header <= '0'; -- 16-byte header.
                  header_left <= conv_std_logic_vector(16-NLANE*2, header_left'length);
                end if;

                -- buffer type fmt type bufsel
                if (rx(0*8+6 downto 0*8+0) = "1000000" or -- MWr 10 00000 P
                    rx(0*8+6 downto 0*8+0) = "1100000" or -- MWr 11 00000 P
                    rx(0*8+6 downto 0*8+3) = "0110" or -- Msg 01 10XXX P
                    rx(0*8+6 downto 0*8+3) = "1110") then -- Msg 11 10XXX P
                  bufsel <= TPBUFTYPE_P;
                elsif (rx(0*8+6 downto 0*8+0) = "0001010" or -- Cpl 00 01010 C
                       rx(0*8+6 downto 0*8+0) = "1001010" or -- CplD 10 01010 C
                       rx(0*8+6 downto 0*8+0) = "0001011" or -- CplLk 00 01011 C
                       rx(0*8+6 downto 0*8+0) = "1001011") then -- CplDLk 10 01011 C
                  bufsel <= TPBUFTYPE_C;
                else -- others NP
                  bufsel <= TPBUFTYPE_NP;
                end if;

                -- virtual channel
                for i in 0 to 5 loop
                  if (rx(1*8+6 downto 1*8+4) = conv_std_logic_vector(i, 3)) then
                    vc <= tc2vc(i);
                  end if;
                end loop; -- virtual channel

                -- output header byte 0 & 1.
                header0(1*8+7 downto 0*8+0) <= rx;
                header_count <= conv_std_logic_vector(NLANE*2, header_count'length);
                rxsm <= RECV_B2_B3;

              end if;

            when RECV_B2_B3 =>
              -- payload in byte unit
              payload_left <= rx(0*8+1 downto 0*8+0) & rx(1*8+7 downto 1*8+0) & "00";

              -- output header byte 2 & 3.
              header0(3*8+7 downto 2*8+0) <= rx;
              header_count <= unsigned(header_count) + NLANE*2;
              header_left <= unsigned(header_left) - NLANE*2;
              rxsm <= RECV_HEADRx1;

            when RECV_HEADRx1 =>
              if (header_left = conv_std_logic_vector(NLANE*2, header_left'length)) then
                rxwr_bufsel <= '1';
                rxwr_header <= '1';
                if (have_data = '1') then
                  rxsm <= RECV_DATA;
                else
                  rxsm <= IDLE;
                end if;
              else
                header_left <= unsigned(header_left) - NLANE*2;
              end if;

              -- output header byte : 4&5, 6&7, 8&9, 10&11 (, 12&13, 14&15).
              -- 2*i runs through : 4, 6, 8, 10, 12, 14
              for i in 2 to 7 loop
                if (header_count = conv_std_logic_vector(2*i, header_left'length)) then
                  header0(16*(i+1)-1 downto 16*i) <= rx;
                end if;
              end loop;
              header_count <= unsigned(header_count) + NLANE*2;

            when RECV_DATA =>
              rxwr_data <= '1';
              data <= rx;

              if (payload_left <= conv_std_logic_vector(NLANE*2, payload_left'length)) then
                rxsm <= IDLE;
              else
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;

            when others =>
              rxsm <= IDLE;

          end case;

        end if;
      end process;

    end block x1block;
  end generate x1;

  x4 : if (NLANE >= 4) generate
    x4block : block

      signal have_data : std_logic;
      signal short_header : std_logic;
      signal bufsel1 : std_logic_vector(1 downto 0);
      signal bufsel2 : std_logic_vector(1 downto 0);
      signal top1, top2 : std_logic;
      signal rx1, rx2 : std_logic_vector(NLANE*16-1 downto 0);
      signal payload_left : std_logic_vector(11 downto 0); -- in byte unit

    begin
      bufsel <= bufsel2;
      process (clk, rst)
      begin
        if (rst = '1') then
          rxwr_bufsel <= '0';
          rxwr_header <= '0';
          rxwr_data <= '0';
        elsif (rising_edge(clk)) then
          rxwr_bufsel <= '0';
          rxwr_header <= '0';
          rxwr_data <= '0';
          top1 <= rxtop;
          top2 <= top1;
          rx1 <= rx;
          rx2 <= rx1;
          bufsel2 <= bufsel1;
          if (rxtop = '1') then
            -- TLP has data?
            if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA) or
                (rx(0*8+6 downto 0*8+5) = TPFMT_16BH_DATA)) then
              have_data <= '1';
            else
              have_data <= '0';
            end if;
            -- header size
            if ((rx(0*8+6 downto 0*8+5) = TPFMT_12BH_NODATA) or
                (rx(0*8+6 downto 0*8+5) = TPFMT_12BH_DATA)) then
              short_header <= '1'; -- 12-byte header.
            else
              short_header <= '0'; -- 16-byte header.
            end if;
            -- buffer type fmt type bufsel
            if (rx(0*8+6 downto 0*8+0) = "1000000" or -- MWr 10 00000 P
                rx(0*8+6 downto 0*8+0) = "1100000" or -- MWr 11 00000 P
                rx(0*8+6 downto 0*8+3) = "0110" or -- Msg 01 10XXX P
                rx(0*8+6 downto 0*8+3) = "1110") then -- Msg 11 10XXX P
              bufsel1 <= TPBUFTYPE_P;
            elsif (rx(0*8+6 downto 0*8+0) = "0001010" or -- Cpl 00 01010 C
                   rx(0*8+6 downto 0*8+0) = "1001010" or -- CplD 10 01010 C
                   rx(0*8+6 downto 0*8+0) = "0001011" or -- CplLk 00 01011 C
                   rx(0*8+6 downto 0*8+0) = "1001011") then -- CplDLk 10 01011 C
              bufsel1 <= TPBUFTYPE_C;
            else -- others NP
              bufsel1 <= TPBUFTYPE_NP;
            end if;
            -- virtual channel
            for i in 0 to 5 loop
              if (rx(1*8+6 downto 1*8+4) = conv_std_logic_vector(i, 3)) then
                vc <= tc2vc(i);
              end if;
            end loop; -- virtual channel
          end if;
          -- header and its write enable
          if (top1 = '1') then
            case NLANE is
              when 4 =>
                rxwr_bufsel <= '1';
                rxwr_header <= '1';
                header <= rx & rx1;
              when 8 =>
                rxwr_bufsel <= '1';
                rxwr_header <= '1';
                header <= rx1;
              when others =>
            end case;
          end if;
          -- payload_left
          case NLANE is
            when 4 =>
              if (rxtop = '1') then
                -- load payload size (in byte unit).
                payload_left <= rx(2*8+1 downto 2*8+0) & rx(3*8+7 downto 3*8+0) & "00";
              elsif (top1 = '1') then
                -- do nothing.
              elsif (top2 = '1') then
                -- do nothing.
              elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;
            when 8 =>
              if (rxtop = '1') then
                -- load payload size (in byte unit).
                payload_left <= rx(2*8+1 downto 2*8+0) & rx(3*8+7 downto 3*8+0) & "00";
              elsif (top1 = '1') then
                -- do nothing.
              elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                payload_left <= unsigned(payload_left) - NLANE*2;
              end if;
            when others =>
          end case;
          -- data write enable
          if (have_data = '1') then
            case NLANE is
              when 4 =>
                if (top1 = '1') then
                  -- do nothing.
                elsif (top2 = '1') then
                  rxwr_data <= '1';
                elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                  rxwr_data <= '1';
                end if;
              when 8 =>
                if (top1 = '1') then
                  rxwr_data <= '1';
                elsif (payload_left > conv_std_logic_vector(NLANE*2, payload_left'length)) then
                  rxwr_data <= '1';
                end if;
              when others =>
            end case;
          end if; -- have_data
          -- data
          case NLANE is
            when 4 =>
              if (short_header = '1') then
                data <= rx(3*8+7 downto 0*8+0) & rx1(7*8+7 downto 4*8+0);
              else
                data <= rx;
              end if;
            when 8 =>
              if (short_header = '1') then
                data <= rx(11*8+7 downto 0*8+0) & rx1(15*8+7 downto 12*8+0);
              else
                data <= rx;
              end if;
            when others =>
          end case;
        end if;
      end process;
    end block x4block;
    -- Traffic Class vs Virtual Channel mapping:
    tc2vc(0) <= "000";
    tc2vc(1) <= "000";
    tc2vc(2) <= "000";
    tc2vc(3) <= "000";
    tc2vc(4) <= "000";
    tc2vc(5) <= "000";
    tc2vc(6) <= "000";
    tc2vc(7) <= "000";
  end generate x4;
end default;
--
-- TLP transmitter
-- multiplex outgoing header/data from Tx{NP,P,C}{H,D}buf,
-- into TLPs and pass them to data-link layer.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity txctl is
  generic (
    NLANE : integer;
    CL_PH_VC0_INIT : integer;
    CL_NPH_VC0_INIT : integer;
    CL_CH_VC0_INIT : integer;
    CL_PD_VC0_INIT : integer;
    CL_NPD_VC0_INIT : integer;
    CL_CD_VC0_INIT : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;
    -- CL notified by data-link layer
    cl_header : in std_logic_vector( 7 downto 0);
    cl_data : in std_logic_vector(11 downto 0);
    inf_cl_header : in std_logic;
    inf_cl_data : in std_logic;
    -- notify TxFCbuf CA to slvctl or msctl
    cat_nph : out std_logic_vector(7 downto 0);
    cat_ph : out std_logic_vector(7 downto 0);
    cat_ch : out std_logic_vector(7 downto 0);
    cat_npd : out std_logic_vector(11 downto 0);
    cat_pd : out std_logic_vector(11 downto 0);
    cat_cd : out std_logic_vector(11 downto 0);
    hbufsel : out std_logic_vector(1 downto 0);
    dbufsel : out std_logic_vector(1 downto 0);
    empty_bufsel : in std_logic;
    empty_header : in std_logic;
    empty_data : in std_logic;
    rd_bufsel : out std_logic;
    rd_header : out std_logic;
    rd_data : out std_logic;
    q_bufsel : in std_logic_vector(1 downto 0);
    q_header : in std_logic_vector(127 downto 0);
    q_data : in std_logic_vector(NLANE*16-1 downto 0);
    txreq : out std_logic;
    txack : in std_logic;
    txtop : out std_logic;
    txwr : out std_logic;
    txbe : out std_logic_vector(NLANE*2-1 downto 0);
    tpsize : out std_logic_vector(OFFSET_WIDTH-1 downto 0);
    tx : out std_logic_vector(NLANE*16-1 downto 0)
  );
end txctl;
architecture default of txctl is
  type TXCTL_STATE is (
    IDLE,
    PARSE_HEADER,
    WAIT_CREDITS,
    SEND_HEADERx1,
    SEND_HEADERx4,
    SEND_HEADERx8,
    SEND_DATA,
    WAITx1,
    INVALID_STATE
  );
  signal txsm : TXCTL_STATE;
  -- credit consumed so far
  signal cc_nph : std_logic_vector( 7 downto 0);
  signal cc_ph : std_logic_vector( 7 downto 0);
  signal cc_ch : std_logic_vector( 7 downto 0);
  signal cc_npd : std_logic_vector(11 downto 0);
  signal cc_pd : std_logic_vector(11 downto 0);
  signal cc_cd : std_logic_vector(11 downto 0);
  signal cc_header : std_logic_vector( 7 downto 0);
  signal cc_data : std_logic_vector(11 downto 0);
  signal credit_inc : std_logic_vector(11 downto 0);
  -- TxFCbuf credit allocated
  signal cat_nph0 : std_logic_vector(7 downto 0);
  signal cat_ph0 : std_logic_vector(7 downto 0);
  signal cat_ch0 : std_logic_vector(7 downto 0);
  signal cat_npd0 : std_logic_vector(11 downto 0);
  signal cat_pd0 : std_logic_vector(11 downto 0);
  signal cat_cd0 : std_logic_vector(11 downto 0);
  signal cr_header : std_logic_vector(7 downto 0);
  signal cr_data : std_logic_vector(11 downto 0);
  signal cl_header1 : std_logic_vector(7 downto 0);
  signal cl_data1 : std_logic_vector(11 downto 0);
  signal inf_cl_data1 : std_logic;
  signal inf_cl_header1 : std_logic;
  signal bufsel0 : std_logic_vector(1 downto 0);
  signal header : std_logic_vector(127 downto 0);
  signal have_data : std_logic;
  signal payload_left : std_logic_vector(11 downto 0); -- in byte unit
  signal header_left : std_logic_vector(3 downto 0);
  signal header_count : std_logic_vector(3 downto 0);
  signal short_header : std_logic;
  signal data1 : std_logic_vector(NLANE*16-1 downto 0);
  signal txacked : std_logic;
  signal wcnt : std_logic_vector(1 downto 0);
begin
  cat_ph <= cat_ph0;
  cat_nph <= cat_nph0;
  cat_ch <= cat_ch0;
  cat_pd <= cat_pd0;
  cat_npd <= cat_npd0;
  cat_cd <= cat_cd0;
  hbufsel <= q_bufsel;
  dbufsel <= bufsel0;
  -- chose credit consumed according to bufsel.
  process (bufsel0,
           cc_nph, cc_npd,
           cc_ph , cc_pd,
           cc_ch , cc_cd)
  begin
    case bufsel0 is
      when TPBUFTYPE_NP =>
        cc_header <= cc_nph;
        cc_data <= cc_npd;
      when TPBUFTYPE_P =>
        cc_header <= cc_ph;
        cc_data <= cc_pd;
      when TPBUFTYPE_C =>
        cc_header <= cc_ch;
        cc_data <= cc_cd;
      when others =>
        cc_header <= (others => '0');
        cc_data <= (others => '0');
    end case;
  end process;
  process (clk, rst)
    variable fmttmp : std_logic_vector(1 downto 0);
    variable lengthtmp : std_logic_vector(9 downto 0);
    variable cr_headertmp : std_logic_vector(7 downto 0);
    variable cr_datatmp : std_logic_vector(11 downto 0);
    variable header_lefttmp : std_logic_vector(header_left'range);
    variable short_headertmp : std_logic;
    variable have_datatmp : std_logic;
    variable payload_lefttmp : std_logic_vector(payload_left'range);
  begin
    if (rst = '1') then
      cc_nph <= conv_std_logic_vector(0, cc_nph'length);
      cc_ph <= conv_std_logic_vector(0, cc_ph'length);
      cc_ch <= conv_std_logic_vector(0, cc_ch'length);
      cc_npd <= conv_std_logic_vector(0, cc_npd'length);
      cc_pd <= conv_std_logic_vector(0, cc_pd'length);
      cc_cd <= conv_std_logic_vector(0, cc_cd'length);
      -- init TxFCbuf Credit Allocated.
      cat_ph0 <= conv_std_logic_vector(CL_PH_VC0_INIT, cat_ph0'length);
      cat_nph0 <= conv_std_logic_vector(CL_NPH_VC0_INIT, cat_nph0'length);
      cat_ch0 <= conv_std_logic_vector(CL_CH_VC0_INIT, cat_ch0'length);
      cat_pd0 <= conv_std_logic_vector(CL_PD_VC0_INIT, cat_pd0'length);
      cat_npd0 <= conv_std_logic_vector(CL_NPD_VC0_INIT, cat_npd0'length);
      cat_cd0 <= conv_std_logic_vector(CL_CD_VC0_INIT, cat_cd0'length);
      -- initial values:
      rd_bufsel <= '0';
      rd_header <= '0';
      rd_data <= '0';
      txtop <= '0';
      txwr <= '0';
      txbe <= (others => '1');
      wcnt <= conv_std_logic_vector(0, wcnt'length);
      txsm <= IDLE;
    elsif (rising_edge(clk)) then
      -- default values:
      rd_bufsel <= '0';
      rd_header <= '0';
      rd_data <= '0';
      txtop <= '0';
      txwr <= '0';
      txbe <= (others => '1');
      -- delayed values:
      data1 <= q_data;
      cl_header1 <= cl_header; -- CLs need to be latched just to satisfy timing constraint.
      inf_cl_header1 <= inf_cl_header;
      cl_data1 <= cl_data;
      inf_cl_data1 <= inf_cl_data;
      case txsm is
        when IDLE =>
          -- header field aliases:
          lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
          txreq <= '0';
          txacked <= '0';
          if (empty_bufsel = '0') then -- something resides in the queue.
            bufsel0 <= q_bufsel;
            if (empty_header = '0') then
              header <= q_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rd_bufsel <= '1';
              rd_header <= '1';
              credit_inc <= length_to_credit(lengthtmp);
              txreq <= '1';
              txsm <= PARSE_HEADER;
            end if;
          end if;
        when PARSE_HEADER =>
          -- header field aliases:
          fmttmp := header( 0*8+6 downto 0*8+5);
          lengthtmp(9 downto 0) := header(2*8+1 downto 2*8+0) & header(3*8+7 downto 3*8+0);
          -- header size
          if (fmttmp = TPFMT_12BH_NODATA or fmttmp = TPFMT_12BH_DATA) then
            short_headertmp := '1';
            short_header <= '1';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(12, header_left'length);
              header_left <= conv_std_logic_vector(12, header_left'length);
            end if;
          else
            short_headertmp := '0';
            short_header <= '0';
            if (NLANE = 1) then
              header_lefttmp := conv_std_logic_vector(16, header_left'length);
              header_left <= conv_std_logic_vector(16, header_left'length);
            end if;
          end if;
          -- credit requested for header
          cr_headertmp := unsigned(cc_header) + 1;
          cr_header <= cr_headertmp;
          -- credit requested for data
          if (fmttmp = TPFMT_12BH_DATA or fmttmp = TPFMT_16BH_DATA) then
            have_datatmp := '1';
            have_data <= '1';
            payload_lefttmp := lengthtmp & "00"; -- in byte unit.
            payload_left <= lengthtmp & "00"; -- in byte unit.
            cr_datatmp := cc_data + credit_inc;
            cr_data <= cr_datatmp;
          else
            have_datatmp := '0';
            have_data <= '0';
          end if;
          case fmttmp is
            when TPFMT_12BH_NODATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length);
            when TPFMT_16BH_NODATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length);
            when TPFMT_12BH_DATA =>
              tpsize <= conv_std_logic_vector(12, tpsize'length) + ("0" & lengthtmp & "00");
            when TPFMT_16BH_DATA =>
              tpsize <= conv_std_logic_vector(16, tpsize'length) + ("0" & lengthtmp & "00");
            when others =>
          end case;
          if (txack = '1' or txacked = '1') then
            txacked <= '1';
            txreq <= '0';
          end if;
          txsm <= WAIT_CREDITS;
        when WAIT_CREDITS =>
          if (txack = '1' or txacked = '1') then
            txacked <= '1';
            txreq <= '0';
          end if;
          if (have_enough_credit(cl_header1, cr_header) = '1' or inf_cl_header1 = '1') then -- check credit for header.
            if (have_data = '0' or -- header only TLP.
                (empty_data = '0' and -- data is arrived in Tx buf.
                 (have_enough_credit(cl_data1, cr_data) = '1' or inf_cl_data1 = '1'))) then -- check credit for data.
              if (txack = '1' or txacked = '1') then -- start tx when PHYMAC:Framer is ready.
                -- send the first word.
                case NLANE is
                  when 1 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(1*8+7 downto 0*8+0);
                    header_left <= unsigned(header_left) - NLANE*2;
                    header_count <= conv_std_logic_vector(NLANE*2, header_count'length);
                    txsm <= SEND_HEADERx1;
                  when 4 =>
                    txtop <= '1';
                    txwr <= '1';
                    tx <= header(7*8+7 downto 0*8+0);
                    if (have_data = '1' and short_header = '1') then
                      rd_data <= '1';
                    end if;
                    txsm <= SEND_HEADERx4;
                  when 8 =>
                    if (have_data = '0') then -- TLP w/o data.
                      txtop <= '1';
                      txwr <= '1';
                      tx <= header(15*8+7 downto 0*8+0);
                      if (short_header = '1') then
                        txbe <= x"0fff";
                      end if;
                      txsm <= IDLE;
                    else -- TLP w/data.
                      rd_data <= '1';
                      if (short_header = '0') then
                        txtop <= '1';
                        txwr <= '1';
                        tx <= header(15*8+7 downto 0*8+0);
                        txsm <= SEND_DATA;
                      else
                        txwr <= '0'; -- don't transmit header right now.
                        tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 0*8+0);
                        payload_left <= unsigned(payload_left) - 4; -- pack header & 4-byte data.
                        txsm <= SEND_HEADERx8;
                      end if;
                    end if;
                    if (have_data = '0' or short_header = '0') then
                      --
                      -- update header-credit consumed.
                      --
                      case bufsel0 is
                        when TPBUFTYPE_NP =>
                          cc_nph <= cr_header;
                          cat_nph0 <= unsigned(cat_nph0) + 1;
                        when TPBUFTYPE_P =>
                          cc_ph <= cr_header;
                          cat_ph0 <= unsigned(cat_ph0) + 1;
                        when TPBUFTYPE_C =>
                          cc_ch <= cr_header;
                          cat_ch0 <= unsigned(cat_ch0) + 1;
                        when others =>
                      end case;
                    end if;
                  when others =>
                end case;
              end if; -- acktp
            end if; -- cl_data
          end if; -- cl_header
        when SEND_HEADERx1 =>
          if (NLANE = 1) then
            txwr <= '1';
            -- output header byte : 2&3, 4&5, 6&7, 8&9, 10&11 (, 12&13, 14&15).
            -- 2*i runs through : 2, 4, 6, 8, 10, 12, 14
            for i in 1 to 7 loop
              if (header_count = conv_std_logic_vector(2*i, header_left'length)) then
                tx <= header(16*(i+1)-1 downto 16*i);
              end if;
            end loop;
            header_count <= unsigned(header_count) + NLANE*2;
            if (header_left = conv_std_logic_vector(NLANE*2, header_left'length)) then
              if (have_data = '1') then
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                txsm <= WAITx1;
              end if;
              --
              -- update header-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_nph <= cr_header;
                  cat_nph0 <= unsigned(cat_nph0) + 1;
                when TPBUFTYPE_P =>
                  cc_ph <= cr_header;
                  cat_ph0 <= unsigned(cat_ph0) + 1;
                when TPBUFTYPE_C =>
                  cc_ch <= cr_header;
                  cat_ch0 <= unsigned(cat_ch0) + 1;
                when others =>
              end case;
            else
              header_left <= unsigned(header_left) - NLANE*2;
            end if;
          end if;
        when SEND_HEADERx4 =>
          if (NLANE = 4) then
            txwr <= '1';
            if (have_data = '0') then -- TLP w/o data.
              tx <= header(15*8+7 downto 8*8+0);
              if (short_header = '1') then
                txbe <= x"0f";
              end if;
              txsm <= IDLE;
            else -- TLP w/data.
              if (short_header = '0') then
                -- transmit last half (8bytes) of the header, then go to SEND_DATA.
                -- rd_data is not asserted yet. assert it right now.
                tx <= header(15*8+7 downto 8*8+0);
                rd_data <= '1';
                txsm <= SEND_DATA;
              else
                -- pack 4-byte header & 4-byte data.
                tx <= q_data(3*8+7 downto 0*8+0) & header(11*8+7 downto 8*8+0);
                payload_left <= unsigned(payload_left) - 4;
                -- rd_data is already asserted in the previous state (PARSE_HEADER).
                -- 4-byte data is transmitted so far, and 8-byte can be transmitted
                -- in the next clock. rd_data should be deasserted now, if payload <= 8byte (=1word in the queue).
                if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                  -- all data (4bytes) are transmitted. quit.
                  txsm <= IDLE;
                  --
                  -- update data-credit consumed.
                  --
                  case bufsel0 is
                    when TPBUFTYPE_NP =>
                      cc_npd <= cr_data;
                      cat_npd0 <= cat_npd0 + credit_inc;
                    when TPBUFTYPE_P =>
                      cc_pd <= cr_data;
                      cat_pd0 <= cat_pd0 + credit_inc;
                    when TPBUFTYPE_C =>
                      cc_cd <= cr_data;
                      cat_cd0 <= cat_cd0 + credit_inc;
                    when others =>
                  end case;
                else
                  -- data yet resides in the queue. goto SEND_DATA.
                  if (payload_left <= conv_std_logic_vector(8, payload_left'length)) then
                    -- SEND_DATA state will be finished in a sigle clock.
                    -- rd_data should be deasserted now.
                    rd_data <= '0';
                  else
                    -- SEND_DATA state continues two or more clocks.
                    -- rd_data should be kept asserted.
                    rd_data <= '1';
                  end if;
                  txsm <= SEND_DATA;
                end if;
              end if;
            end if;
            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;
          end if;
        when SEND_HEADERx8 =>
          if (NLANE = 8) then
            txtop <= '1';
            txwr <= '1'; -- transmit header which was set in the previous state (PARSE_HEADER).
            if (payload_left = conv_std_logic_vector(0, payload_left'length)) then
              -- all data (4bytes) are transmitted. quit.
              txsm <= IDLE;
              --
              -- update data-credit consumed.
              --
              case bufsel0 is
                when TPBUFTYPE_NP =>
                  cc_npd <= cr_data;
                  cat_npd0 <= cat_npd0 + credit_inc;
                when TPBUFTYPE_P =>
                  cc_pd <= cr_data;
                  cat_pd0 <= cat_pd0 + credit_inc;
                when TPBUFTYPE_C =>
                  cc_cd <= cr_data;
                  cat_cd0 <= cat_cd0 + credit_inc;
                when others =>
              end case;
            else
              -- data yet resides in the queue. goto SEND_DATA.
              if (payload_left <= conv_std_logic_vector(12, payload_left'length)) then
                -- rd_data should be asserted only for 1clk since payload size is 16B or smaller.
                rd_data <= '0';
              else
                -- SEND_DATA state continues two or more clocks.
                -- rd_data should be kept asserted.
                rd_data <= '1';
              end if;
              txsm <= SEND_DATA;
            end if;
            --
            -- update header-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_nph <= cr_header;
                cat_nph0 <= unsigned(cat_nph0) + 1;
              when TPBUFTYPE_P =>
                cc_ph <= cr_header;
                cat_ph0 <= unsigned(cat_ph0) + 1;
              when TPBUFTYPE_C =>
                cc_ch <= cr_header;
                cat_ch0 <= unsigned(cat_ch0) + 1;
              when others =>
            end case;
          end if;
        when SEND_DATA =>
          txwr <= '1';
          case NLANE is
            when 1 =>
              tx <= q_data; -- always transmit 2-byte data.
              if (payload_left > conv_std_logic_vector(2, payload_left'length)) then
                rd_data <= '1';
              end if;
            when 4 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(7*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;
              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"0f";
              end if;
              if (short_header = '1') then
                if (payload_left > conv_std_logic_vector(12, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                if (payload_left > conv_std_logic_vector(8, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;
            when 8 =>
              if (short_header = '1') then
                tx <= q_data(3*8+7 downto 0*8+0) & data1(15*8+7 downto 4*8+0);
              else
                tx <= q_data;
              end if;
              -- byte enable of packet tail.
              if (payload_left = conv_std_logic_vector(12, payload_left'length)) then
                txbe <= x"0fff";
              elsif (payload_left = conv_std_logic_vector(8, payload_left'length)) then
                txbe <= x"00ff";
              elsif (payload_left = conv_std_logic_vector(4, payload_left'length)) then
                txbe <= x"000f";
              end if;
              if (short_header = '1') then
                -- assert rd_data if payload_left > NLANE*2-4.
                if (payload_left > conv_std_logic_vector(28, payload_left'length)) then
                  rd_data <= '1';
                end if;
              else
                -- assert rd_data if payload_left > NLANE.
                if (payload_left > conv_std_logic_vector(16, payload_left'length)) then
                  rd_data <= '1';
                end if;
              end if;
            when others =>
          end case;
          -- exit to IDLE.
          if (payload_left <= conv_std_logic_vector(NLANE*2, payload_left'length)) then
            if (NLANE = 1) then
              txsm <= WAITx1;
            else
              txsm <= IDLE;
            end if;
            --
            -- update data-credit consumed.
            --
            case bufsel0 is
              when TPBUFTYPE_NP =>
                cc_npd <= cr_data;
                cat_npd0 <= cat_npd0 + credit_inc;
              when TPBUFTYPE_P =>
                cc_pd <= cr_data;
                cat_pd0 <= cat_pd0 + credit_inc;
              when TPBUFTYPE_C =>
                cc_cd <= cr_data;
                cat_cd0 <= cat_cd0 + credit_inc;
              when others =>
            end case;
            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            -- this is necessary only for x8 32-bit addressing TLP w/payload.
            if (NLANE = 8) then
              fmttmp := q_header( 0*8+6 downto 0*8+5);
              lengthtmp(9 downto 0) := q_header(2*8+1 downto 2*8+0) & q_header(3*8+7 downto 3*8+0);
              if (empty_bufsel = '0' and fmttmp = TPFMT_12BH_DATA) then
                txreq <= '0';
                txacked <= '0';
                bufsel0 <= q_bufsel;
                if (empty_header = '0') then
                  header <= q_header;
                  -- bufsel & header is saved to a register.
                  -- fifo output is no longer necessary.
                  rd_bufsel <= '1';
                  rd_header <= '1';
                  credit_inc <= length_to_credit(lengthtmp);
                  txreq <= '1';
                  txsm <= PARSE_HEADER;
                end if;
              end if;
            end if;
          else
            payload_left <= unsigned(payload_left) - NLANE*2;
          end if;
        when WAITx1 =>
          -- wait for 3clk.
          if (wcnt = conv_std_logic_vector(3-1, wcnt'length)) then
            wcnt <= conv_std_logic_vector(0, wcnt'length);
            txsm <= IDLE;
          else
            wcnt <= unsigned(wcnt) + 1;
          end if;
        when others =>
          txsm <= INVALID_STATE;
      end case;
    end if;
  end process;
end default;
--
-- PCIe Slvae (Target) Controller
-- . receive Posted and Non-posted TLPs from Rx FC bufs.
-- . communicate with application layer as master.
-- . transmit Completion TLPs to Tx FC bufs.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;
entity slvctl is
  generic (
    CA_PH_VC0_INIT : integer;
    CA_NPH_VC0_INIT : integer;
    CA_CH_VC0_INIT : integer;
    CA_PD_VC0_INIT : integer;
    CA_NPD_VC0_INIT : integer;
    CA_CD_VC0_INIT : integer;
    CFG_BAR0_INIT : std_logic_vector(31 downto 0);
    CFG_BAR1_INIT : std_logic_vector(31 downto 0);
    CFG_BAR2_INIT : std_logic_vector(31 downto 0);
    CFG_BAR3_INIT : std_logic_vector(31 downto 0);
    CFG_BAR4_INIT : std_logic_vector(31 downto 0);
    CFG_BAR5_INIT : std_logic_vector(31 downto 0);
    CFG_BAR_ROM_INIT : std_logic_vector(31 downto 0);
    NLANE : integer;
    NDMACH : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic; -- active high
    ca_nph : out std_logic_vector(7 downto 0);
    ca_ph : out std_logic_vector(7 downto 0);
    ca_ch : out std_logic_vector(7 downto 0);
    ca_npd : out std_logic_vector(11 downto 0);
    ca_pd : out std_logic_vector(11 downto 0);
    ca_cd : out std_logic_vector(11 downto 0);
    ca_p_flush : out std_logic;
    ca_np_flush : out std_logic;
    ca_c_flush : out std_logic;
    mpls : in std_logic_vector(11 downto 0); -- max payload size
    cfg_bar0 : in std_logic_vector(31 downto 0);
    cfg_bar1 : in std_logic_vector(31 downto 0);
    cfg_bar2 : in std_logic_vector(31 downto 0);
    cfg_bar3 : in std_logic_vector(31 downto 0);
    cfg_bar4 : in std_logic_vector(31 downto 0);
    cfg_bar5 : in std_logic_vector(31 downto 0);
    cfg_bar_rom : in std_logic_vector(31 downto 0);
    rxhbufsel : out std_logic_vector(1 downto 0);
    rxdbufsel : out std_logic_vector(1 downto 0);
    rxempty_bufsel : in std_logic;
    rxrd_bufsel : out std_logic;
    rxrd_header : out std_logic;
    rxrd_data : out std_logic;
    rxq_bufsel : in std_logic_vector(1 downto 0);
    rxq_header : in std_logic_vector(127 downto 0);
    rxq_data : in std_logic_vector(NLANE*16-1 downto 0);
    txvc : out std_logic_vector(2 downto 0);
    cat_header : in std_logic_vector(7 downto 0);
    cat_data : in std_logic_vector(11 downto 0);
    bdf : out std_logic_vector(15 downto 0); -- completer id
    txwr_req : out std_logic;
    txwr_ack : in std_logic;
    txwr_bufsel : out std_logic;
    txwr_header : out std_logic;
    txwr_data : out std_logic;
    tx_header : out std_logic_vector(127 downto 0);
    tx_data : out std_logic_vector(NLANE*16-1 downto 0);
    slv_read : out std_logic;
    slv_write : out std_logic;
    slv_bar : out std_logic_vector(6 downto 0);
    slv_addr : out std_logic_vector(63 downto 0);
    slv_bytevalid : out std_logic_vector(NLANE*2-1 downto 0);
    slv_dataout : out std_logic_vector(NLANE*16-1 downto 0);
    slv_datain : in std_logic_vector(NLANE*16-1 downto 0);
    slv_readreq : out std_logic;
    slv_writereq : out std_logic;
    slv_accept : in std_logic;
    slv_bytecount : out std_logic_vector(11 downto 0);
    ms_read : out std_logic;
    ms_rdchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_rdaddr : out std_logic_vector(31 downto 0);
    ms_rddata : out std_logic_vector(NLANE*16-1 downto 0);
    dmach_rdcpl : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdcpllength : out std_logic_vector(9 downto 0);
    dmach_rdcplbcnt : out std_logic_vector(11 downto 0);
    dmach_rdcpladdr : in each32b(NDMACH-1 downto 0);
    cfg_read : out std_logic;
    cfg_write : out std_logic;
    cfg_addr : out std_logic_vector(11 downto 0);
    cfg_be : out std_logic_vector(3 downto 0);
    cfg_dataout : out std_logic_vector(31 downto 0);
    cfg_datain : in std_logic_vector(31 downto 0)
  );
end slvctl;
architecture default of slvctl is
  type SLV_STATE is (
    IDLE,
    PARSE_HEADER,
    -- CfgRd0
    WAIT_CfgRd0,
    CPL_CfgRd0,
    CPL_CfgRd0_2,
    -- CfgWr0
    WAIT_CfgWr0,
    CPL_CfgWr0,
    CPL_CfgWr0_2,
    CPL_CfgWr0_3,
    -- MsgD
    BURST_MsgD,
    -- MemRd
    WAIT_MemRd,
    WAIT_MemRd_2,
    READY_MemRd,
    BURST_MemRd,
    -- MemWr
    WAIT_MemWr,
    BURST_MemWr,
    -- CplD
    WAIT_CplD,
    BURST_CplD,
    -- Non-Supported Request
    INVALID_STATE
  );
  signal slvsm : SLV_STATE;
  -- credit allocated
  signal ca_nph0 : std_logic_vector(7 downto 0);
  signal ca_ph0 : std_logic_vector(7 downto 0);
  signal ca_ch0 : std_logic_vector(7 downto 0);
  signal ca_npd0 : std_logic_vector(11 downto 0);
  signal ca_pd0 : std_logic_vector(11 downto 0);
  signal ca_cd0 : std_logic_vector(11 downto 0);
  signal credit_inc : std_logic_vector(11 downto 0);
  -- TxFCbuf credit consumed
  signal cct_ch : std_logic_vector(7 downto 0);
  signal cct_cd : std_logic_vector(11 downto 0);
  signal crt_header : std_logic_vector(7 downto 0);
  signal crt_data : std_logic_vector(11 downto 0);
  signal rxbufsel0 : std_logic_vector(1 downto 0);
  signal header : std_logic_vector(127 downto 0);
  signal be_first : std_logic_vector(3 downto 0);
  signal be_last : std_logic_vector(3 downto 0);
  signal byte_ucnt : std_logic_vector(11 downto 0);
  signal byte_dcnt : std_logic_vector(11 downto 0);
  signal barsel : std_logic_vector(6 downto 0);
  signal slv_addr0 : std_logic_vector(63 downto 0);
  signal bdf0 : std_logic_vector(15 downto 0);
  signal loweraddr : std_logic_vector(6 downto 0);
  signal length : std_logic_vector(9 downto 0);
  signal bytecount : std_logic_vector(11 downto 0);
  signal bytecount_next : std_logic_vector(11 downto 0);
  signal dwcount : std_logic_vector(9 downto 0);
  signal dwcount_next : std_logic_vector(9 downto 0);
  signal cpllength : std_logic_vector(9 downto 0);
  signal cplbytelength : std_logic_vector(11 downto 0);
  signal data16 : std_logic_vector(15 downto 0);
  signal ms_rdaddr0 : std_logic_vector(31 downto 0);
  signal ms_rdchannel0 : std_logic_vector(NDMACH-1 downto 0);
begin
  rxhbufsel <= rxq_bufsel;
  rxdbufsel <= rxbufsel0;
  slv_addr <= slv_addr0;
  slv_dataout <= rxq_data;
  slv_bytecount <= byte_dcnt;
  ms_rddata <= rxq_data;
  ms_rdaddr <= ms_rdaddr0;
  ms_rdchannel <= ms_rdchannel0;
  -- credit allocated
  ca_ph <= ca_ph0;
  ca_nph <= ca_nph0;
  ca_ch <= ca_ch0;
  ca_pd <= ca_pd0;
  ca_npd <= ca_npd0;
  ca_cd <= ca_cd0;
  bdf <= bdf0;
  -- Slave Controller State Machine
  process (clk, rst)
    variable typetmp : std_logic_vector(6 downto 0);
    variable be_firsttmp : std_logic_vector(3 downto 0);
    variable be_lasttmp : std_logic_vector(3 downto 0);
    variable lengthtmp : std_logic_vector(9 downto 0);
    variable cpllengthtmp : std_logic_vector(9 downto 0);
    variable cplbytelengthtmp : std_logic_vector(11 downto 0);
    variable bdftmp : std_logic_vector(15 downto 0);
    variable addrtmp32 : std_logic_vector(31 downto 0);
    variable addrtmp64 : std_logic_vector(63 downto 0);
    variable addrtmp : std_logic_vector(6 downto 0);
    variable loweraddrtmp : std_logic_vector(6 downto 0);
    variable bytecounttmp : std_logic_vector(11 downto 0);
    variable crt_datatmp : std_logic_vector(11 downto 0);
    variable byte_dcnttmp : std_logic_vector(11 downto 0);
    variable tagtmp : std_logic_vector(7 downto 0);
    variable bcnttmp : std_logic_vector(11 downto 0);
  begin
    if (rst = '1') then
      -- init Flow Control Credit Allocated.
      ca_ph0 <= conv_std_logic_vector(CA_PH_VC0_INIT, ca_ph0'length);
      ca_nph0 <= conv_std_logic_vector(CA_NPH_VC0_INIT, ca_nph0'length);
      ca_ch0 <= conv_std_logic_vector(CA_CH_VC0_INIT, ca_ch0'length);
      ca_pd0 <= conv_std_logic_vector(CA_PD_VC0_INIT, ca_pd0'length);
      ca_npd0 <= conv_std_logic_vector(CA_NPD_VC0_INIT, ca_npd0'length);
      ca_cd0 <= conv_std_logic_vector(CA_CD_VC0_INIT, ca_cd0'length);
      -- clear TxFCbuf Credit Consumed.
      cct_ch <= conv_std_logic_vector(0, cct_ch'length);
      cct_cd <= conv_std_logic_vector(0, cct_cd'length);
      -- initial values:
      rxrd_bufsel <= '0';
      rxrd_header <= '0';
      rxrd_data <= '0';
      txwr_bufsel <= '0';
      txwr_header <= '0';
      txwr_data <= '0';
      slv_read <= '0';
      slv_write <= '0';
      slv_bytevalid <= (others => '0');
      cfg_read <= '0';
      cfg_write <= '0';
      ca_p_flush <= '0';
      ca_np_flush <= '0';
      ca_c_flush <= '0';
      txwr_req <= '0';
      slv_readreq <= '0';
      slv_writereq <= '0';
      ms_read <= '0';
      slvsm <= IDLE;
    elsif (rising_edge(clk)) then
      -- header field aliases:
      typetmp := header( 0*8+6 downto 0*8+0);
      be_firsttmp(3 downto 0) := header( 7*8+3 downto 7*8+0);
      be_lasttmp(3 downto 0) := header( 7*8+7 downto 7*8+4);
      lengthtmp(9 downto 0) := header(2*8+1 downto 2*8+0) & header(3*8+7 downto 3*8+0);
      bdftmp := header(9*8+7 downto 8*8+0);
      addrtmp32 := header( 8*8+7 downto 8*8+0) & header( 9*8+7 downto 9*8+0) &
                   header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+2) & "00";
      addrtmp64 := header( 8*8+7 downto 8*8+0) & header( 9*8+7 downto 9*8+0) &
                   header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+0) &
                   header(12*8+7 downto 12*8+0) & header(13*8+7 downto 13*8+0) &
                   header(14*8+7 downto 14*8+0) & header(15*8+7 downto 15*8+2) & "00";
      tagtmp := header(10*8+7 downto 10*8+0);
      bcnttmp := header( 6*8+3 downto 6*8+0) & header( 7*8+7 downto 7*8+0);
      loweraddrtmp := header(11*8+6 downto 11*8+0);
      -- default values:
      rxrd_bufsel <= '0';
      rxrd_header <= '0';
      rxrd_data <= '0';
      txwr_bufsel <= '0';
      txwr_header <= '0';
      txwr_data <= '0';
      slv_read <= '0';
      slv_write <= '0';
      cfg_read <= '0';
      cfg_write <= '0';
      ca_p_flush <= '0';
      ca_np_flush <= '0';
      ca_c_flush <= '0';
      ms_read <= '0';
      dmach_rdcpl <= (others => '0');
      case slvsm is
        when IDLE =>
          if (rxempty_bufsel = '0') then -- something arrived in the queue.
            rxbufsel0 <= rxq_bufsel;
            header <= rxq_header;
            -- bufsel & header is saved to a register.
            -- fifo output is no longer necessary.
            rxrd_bufsel <= '1';
            rxrd_header <= '1';
            -- header is parsed in the next state, PARSE_HEADER.
            -- it is difficult to parse header here in this state
            -- due to timing restriction.
            slvsm <= PARSE_HEADER;
          end if;
        when PARSE_HEADER =>
          slv_bar <= barsel;
          slv_addr0 <= mask_addr(barsel, header,
                                 CFG_BAR0_INIT, CFG_BAR1_INIT, CFG_BAR2_INIT,
                                 CFG_BAR3_INIT, CFG_BAR4_INIT, CFG_BAR5_INIT, CFG_BAR_ROM_INIT);
          credit_inc <= length_to_credit(lengthtmp);
          case rxbufsel0 is
            when TPBUFTYPE_NP =>
              ca_nph0 <= unsigned(ca_nph0) + 1;
              crt_header <= unsigned(cct_ch) + 1;
              case typetmp is
                when TPTYPE_MRd32 =>
                  addrtmp := addrtmp32(6 downto 0);
                  -- remaining data size in byte/DW
                  bytecounttmp := memrd_1st_cpl_bytecount(lengthtmp, be_firsttmp, be_lasttmp);
                  bytecount <= bytecounttmp;
                  dwcount <= lengthtmp;
                  -- lower address of the first completion
                  loweraddrtmp := memrd_1st_cpl_loweraddr(addrtmp, be_firsttmp);
                  loweraddr <= loweraddrtmp;
                  -- length of the first completion in byte/DW
                  cplbytelength <= memrd_1st_cpl_bytelength(lengthtmp, mpls, loweraddrtmp, bytecounttmp);
                  cpllength <= memrd_1st_cpl_length(lengthtmp, mpls, loweraddrtmp);
                  txwr_req <= '1';
                  txvc <= "000";
                  slv_readreq <= '1';
                  slvsm <= WAIT_MemRd;
                when TPTYPE_MRd64 =>
                  addrtmp := addrtmp64(6 downto 0);
                when TPTYPE_MRdLk32 =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_MRdLk64 =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_IORd =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_IOWr =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_CfgRd0 =>
                  txwr_req <= '1';
                  cfg_read <= '1';
                  cfg_addr <= "0000" & addrtmp32(7 downto 2) & "00";
                  txvc <= "000";
                  tx_header <= cfgrd0_cpl_header(header);
                  crt_data <= unsigned(cct_cd) + 1;
                  bdf0 <= bdftmp; -- save my BDF (bus id, device id, function id)
                  -- for future use.
                  -- note that the saved value is byte reversed.
                  slvsm <= WAIT_CfgRd0;
                when TPTYPE_CfgWr0 =>
                  txwr_req <= '1';
                  txvc <= "000";
                  tx_header <= cfgwr0_cpl_header(header);
                  bdf0 <= bdftmp;
                  slvsm <= WAIT_CfgWr0;
                when TPTYPE_CfgRd1 =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_CfgWr1 =>
                  slvsm <= INVALID_STATE;
                when others =>
                  slvsm <= INVALID_STATE;
              end case;
            when TPBUFTYPE_P =>
              ca_ph0 <= unsigned(ca_ph0) + 1;
              be_first <= be_firsttmp;
              be_last <= be_lasttmp;
              byte_dcnttmp := lengthtmp & "00";
              byte_dcnt <= byte_dcnttmp;
              byte_ucnt <= conv_std_logic_vector(0, byte_ucnt'length);
              slv_bytevalid <= (others => '1');
              if (NLANE = 1) then
                slv_bytevalid <= be_firsttmp(1 downto 0);
              else
                -- byte-valid bits of the lowest 4 bytes are given by be_firsttmp.
                slv_bytevalid(3 downto 0) <= be_firsttmp(3 downto 0);
              end if;
              -- byte-enable of the last NLANE*2 byte.
              if (byte_dcnttmp <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
                case NLANE is
                  when 4 =>
                    case byte_dcnttmp is
                      when conv_std_logic_vector(8, byte_dcnt'length) =>
                        slv_bytevalid <= be_lasttmp & be_firsttmp;
                      when conv_std_logic_vector(4, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & be_firsttmp;
                      when others =>
                    end case;
                  when 8 =>
                    case byte_dcnttmp is
                      when conv_std_logic_vector(16, byte_dcnt'length) =>
                        slv_bytevalid <= be_lasttmp & "1111" & "1111" & be_firsttmp;
                      when conv_std_logic_vector(12, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & be_lasttmp & "1111" & be_firsttmp;
                      when conv_std_logic_vector(8, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & "0000" & be_lasttmp & be_firsttmp;
                      when conv_std_logic_vector(4, byte_dcnt'length) =>
                        slv_bytevalid <= "0000" & "0000" & "0000" & be_firsttmp;
                      when others =>
                    end case;
                  when others =>
                end case;
              end if;
              case typetmp is
                when TPTYPE_MWr32 | TPTYPE_MWr64 =>
                  slv_writereq <= '1';
                  slvsm <= WAIT_MemWr;
                when others =>
-- msg_write <= '1';
                  if (typetmp(6 downto 3) = "0110") then -- TPTYPE_Msg
                    -- !!! this part should be rewritten. !!!
                    -- !!! all messages are ignored for now. !!!
                    slvsm <= IDLE;
                  elsif (typetmp(6 downto 3) = "1110") then -- TPTYPE_MsgD
                    rxrd_data <= '1';
                    slvsm <= BURST_MsgD;
                  else
                    slvsm <= INVALID_STATE;
                  end if;
              end case;
            when TPBUFTYPE_C =>
              ca_ch0 <= unsigned(ca_ch0) + 1;
              case typetmp is
                when TPTYPE_Cpl =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_CplD =>
                  for i in 0 to NDMACH-1 loop
                    if (tagtmp = conv_std_logic_vector(i, tagtmp'length)) then
                      ms_rdaddr0 <= dmach_rdcpladdr(i);
                      ms_rdchannel0 <= (i => '1', others => '0');
                      dmach_rdcpl(i) <= '1';
                    end if;
                  end loop;
                  dmach_rdcpllength <= lengthtmp;
                  dmach_rdcplbcnt <= bcnttmp;
                  byte_dcnt <= lengthtmp & "00";
                  byte_ucnt <= conv_std_logic_vector(0, byte_ucnt'length);
                  slvsm <= WAIT_CplD;
                when TPTYPE_CplLk =>
                  slvsm <= INVALID_STATE;
                when TPTYPE_CplDLk =>
                  slvsm <= INVALID_STATE;
                when others =>
                  slvsm <= INVALID_STATE;
              end case;
            when others =>
              slvsm <= INVALID_STATE;
          end case; -- end of PARSE_HEADER state
        --
        -- Type0 Configuration Read
        --
        when WAIT_CfgRd0 =>
          -- wait until Tx buffers are ready.
          if (txwr_ack = '1' and -- got write permission to Tx buf.
              have_enough_credit(cat_header, crt_header) = '1' and -- Tx buf have enough room.
              have_enough_credit(cat_data, crt_data) = '1') then
            cct_ch <= crt_header;
            cct_cd <= crt_data;
            txwr_bufsel <= '1';
            txwr_header <= '1';
            slvsm <= CPL_CfgRd0;
          end if;
        when CPL_CfgRd0 =>
          ca_np_flush <= '1';
          txwr_data <= '1';
          if (NLANE = 1) then
            tx_data <= cfg_datain(15 downto 0);
            slvsm <= CPL_CfgRd0_2;
          else
            tx_data(NLANE*16-1 downto 32) <= (others => '0');
            tx_data(31 downto 0) <= cfg_datain;
            txwr_req <= '0';
            slvsm <= IDLE;
          end if;
        -- x1 only
        when CPL_CfgRd0_2 =>
          if (NLANE = 1) then
            txwr_data <= '1';
            tx_data <= cfg_datain(31 downto 16);
            txwr_req <= '0';
            slvsm <= IDLE;
          end if;
        --
        -- Type0 Configuration Write
        --
        when WAIT_CfgWr0 =>
          -- wait until Tx buffers are ready.
          if (txwr_ack = '1' and
              have_enough_credit(cat_header, crt_header) = '1') then
            cct_ch <= crt_header;
            txwr_bufsel <= '1';
            txwr_header <= '1';
            rxrd_data <= '1';
            ca_npd0 <= ca_npd0 + credit_inc;
            slvsm <= CPL_CfgWr0;
          end if;
        when CPL_CfgWr0 =>
          ca_np_flush <= '1';
          cfg_addr <= "0000" & header(11*8+7 downto 11*8+2) & "00";
          cfg_be <= header( 7*8+3 downto 7*8+0);
          if (NLANE = 1) then
            rxrd_data <= '1';
            data16 <= rxq_data;
            slvsm <= CPL_CfgWr0_2;
          else
            cfg_write <= '1';
            cfg_dataout <= rxq_data(31 downto 0);
            slvsm <= CPL_CfgWr0_3;
          end if;
        -- x1 only
        when CPL_CfgWr0_2 =>
          if (NLANE = 1) then
            cfg_write <= '1';
            cfg_dataout <= rxq_data & data16;
            slvsm <= CPL_CfgWr0_3;
          end if;
        -- wait for the written configuration register to be updated.
        when CPL_CfgWr0_3 =>
          txwr_req <= '0';
          slvsm <= IDLE;
        --
        -- Memory Read
        --
        when WAIT_MemRd =>
          byte_dcnt <= cplbytelength;
          bytecount_next <= bytecount - cplbytelength;
          dwcount_next <= dwcount - cpllength;
          tx_header <= memrd_cpl_header(header, bdf0, cpllength, bytecount, loweraddr);
          crt_data <= cct_cd + length_to_credit(cpllength);
          ca_np_flush <= '1';
          slvsm <= WAIT_MemRd_2;
        when WAIT_MemRd_2 =>
          -- wait until Tx buffers are ready.
          if (
              slv_accept = '1' and -- !!!
              txwr_ack = '1' and
              have_enough_credit(cat_header, crt_header) = '1' and
              have_enough_credit(cat_data, crt_data) = '1') then
            txwr_bufsel <= '1';
            txwr_header <= '1';
            cct_ch <= crt_header;
            cct_cd <= crt_data;
            slv_readreq <= '0';
            slv_read <= '1';
            slvsm <= READY_MemRd;
          end if;
        when READY_MemRd =>
          -- keep slv_read asserted if burst continues longer than 1clk.
          if (byte_dcnt > conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            slv_read <= '1';
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;
          end if;
          txwr_req <= '0';
          slvsm <= BURST_MemRd;
        when BURST_MemRd =>
          txwr_data <= '1';
          tx_data <= slv_datain;
          byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          -- deassert slv_addr 2-clk earlier than end of burst.
          if (byte_dcnt > conv_std_logic_vector(NLANE*2*2, byte_dcnt'length)) then
            slv_read <= '1';
          end if;
          if (byte_dcnt > conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then  -- read burst in progress.
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;
          else -- about to finish current completion.
            -- some more data to be sent. prepare for the next completion.
            if (dwcount_next /= conv_std_logic_vector(0, dwcount_next'length)) then
              bytecount <= bytecount_next;
              dwcount <= dwcount_next;
              loweraddr <= (others => '0');
              -- next completion is not the final one.
              -- cpllength is set to max payload size.
              if (mpls(11 downto 2) < dwcount_next) then
                cpllengthtmp := mpls(11 downto 2);
                cpllength <= cpllengthtmp;
                cplbytelengthtmp := mpls;
                cplbytelength <= cplbytelengthtmp;
                dwcount_next <= dwcount_next - mpls(11 downto 2);
                bytecount_next <= bytecount_next - mpls;
              -- next completion is the final one.
              -- cpllength is set to remaining data size.
              else
                cpllengthtmp := dwcount_next;
                cpllength <= cpllengthtmp;
                cplbytelengthtmp := bytecount_next;
                cplbytelength <= cplbytelengthtmp;
              end if;
              slv_readreq <= '1';
              slvsm <= WAIT_MemRd;
            else
-- txwr_req <= '0'; -- !!!
              slvsm <= IDLE;
            end if;
          end if;
        --
        -- Memory Write
        --
        when WAIT_MemWr =>
          if (slv_accept = '1') then
            -- flush CA at the beginning of a burst. flush is
            -- necessary only when RxFCbuf is almost full. but
            -- for simplicity, flush is asserted every time a
            -- TLP is retrieved from RxFCbuf.
-- ca_p_flush <= '1'; -- !!!
            rxrd_data <= '1';
            slv_writereq <= '0';
            slv_write <= '1';
            slvsm <= BURST_MemWr;
          end if;
        when BURST_MemWr =>
          slv_bytevalid <= (others => '1');
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_pd0 <= ca_pd0 + credit_inc;
            ca_p_flush <= '1'; -- !!!
            slvsm <= IDLE;
            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            if (rxempty_bufsel = '0') then -- something arrived in the queue.
              rxbufsel0 <= rxq_bufsel;
              header <= rxq_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rxrd_bufsel <= '1';
              rxrd_header <= '1';
              -- header is parsed in the next state, PARSE_HEADER.
              -- it is difficult to parse header here in this state
              -- due to timing restriction.
              slvsm <= PARSE_HEADER;
            end if;
          else
            rxrd_data <= '1';
            slv_write <= '1';
            slv_addr0 <= unsigned(slv_addr0) + NLANE*2;
            byte_ucnt <= unsigned(byte_ucnt) + NLANE*2;
            byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          end if;
          -- byte-enable of the first DW for x1.
          if (byte_ucnt = conv_std_logic_vector(0, byte_ucnt'length) and NLANE = 1) then
            slv_bytevalid <= be_first(3 downto 2);
          end if;
          -- byte-enable of the last NLANE*2 byte for x1.
          if (byte_dcnt = conv_std_logic_vector(6, byte_dcnt'length) and NLANE = 1) then
            slv_bytevalid <= be_last(1 downto 0);
          end if;
          if (byte_dcnt <= conv_std_logic_vector((NLANE*2)*2, byte_dcnt'length)) then
            case NLANE is
              when 1 =>
                if (byte_ucnt >= conv_std_logic_vector(4, byte_ucnt'length)) then
                  slv_bytevalid <= be_last(3 downto 2);
                end if;
              when 4 =>
                case byte_dcnt is
                  when conv_std_logic_vector(16, byte_dcnt'length) =>
                    slv_bytevalid <= be_last & "1111";
                  when conv_std_logic_vector(12, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & be_last;
                  when others =>
                end case;
              when 8 =>
                case byte_dcnt is
                  when conv_std_logic_vector(32, byte_dcnt'length) =>
                    slv_bytevalid <= be_last & "1111" & "1111" & "1111";
                  when conv_std_logic_vector(28, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & be_last & "1111" & "1111";
                  when conv_std_logic_vector(24, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & "0000" & be_last & "1111";
                  when conv_std_logic_vector(20, byte_dcnt'length) =>
                    slv_bytevalid <= "0000" & "0000" & "0000" & be_last;
                  when others =>
                end case;
              when others =>
            end case;
          end if;
        --
        -- Completion with Data
        --
        when WAIT_CplD =>
-- ca_c_flush <= '1'; -- !!!
          rxrd_data <= '1';
          ms_read <= '1';
          slvsm <= BURST_CplD;
        when BURST_CplD =>
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_cd0 <= ca_cd0 + credit_inc; -- !!!
            ca_c_flush <= '1';
            slvsm <= IDLE;
            -- additional logic for better performance.
            -- skip IDLE state and go directly to PARSE_HEADER if possible.
            if (rxempty_bufsel = '0') then -- something arrived in the queue.
              rxbufsel0 <= rxq_bufsel;
              header <= rxq_header;
              -- bufsel & header is saved to a register.
              -- fifo output is no longer necessary.
              rxrd_bufsel <= '1';
              rxrd_header <= '1';
              -- header is parsed in the next state, PARSE_HEADER.
              -- it is difficult to parse header here in this state
              -- due to timing restriction.
              slvsm <= PARSE_HEADER;
            end if;
          else
            rxrd_data <= '1';
            ms_read <= '1';
            ms_rdaddr0 <= unsigned(ms_rdaddr0) + NLANE*2;
            byte_ucnt <= unsigned(byte_ucnt) + NLANE*2;
            byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          end if;
        when BURST_MsgD =>
          -- !!! this part should be rewritten. !!!
          -- !!! all messages are ignored for now. !!!
          -- just to remove data from rxq_npd.
          if (byte_dcnt <= conv_std_logic_vector(NLANE*2, byte_dcnt'length)) then
            ca_pd0 <= ca_pd0 + credit_inc;
            slvsm <= IDLE;
          else
            rxrd_data <= '1';
            byte_ucnt <= unsigned(byte_ucnt) + NLANE*2;
            byte_dcnt <= unsigned(byte_dcnt) - NLANE*2;
          end if;
        when others =>
          slvsm <= INVALID_STATE;
      end case;
    end if;
  end process;
  --
  -- Address Decoder
  -- decode address by comparing it with values stored in base address registers.
  --
  process(clk)
    variable addrtmp32 : std_logic_vector(31 downto 0);
    variable addrtmp64 : std_logic_vector(63 downto 0);
    variable mask32 : std_logic_vector(31 downto 0);
    variable mask64 : std_logic_vector(63 downto 0);
  begin
    if (rising_edge(clk)) then
      -- alias for address fields in rxq_header
      addrtmp32 := rxq_header( 8*8+7 downto 8*8+0) & rxq_header( 9*8+7 downto 9*8+0) &
                   rxq_header(10*8+7 downto 10*8+0) & rxq_header(11*8+7 downto 11*8+2) & "00";
      addrtmp64 := rxq_header( 8*8+7 downto 8*8+0) & rxq_header( 9*8+7 downto 9*8+0) &
                   rxq_header(10*8+7 downto 10*8+0) & rxq_header(11*8+7 downto 11*8+0) &
                   rxq_header(12*8+7 downto 12*8+0) & rxq_header(13*8+7 downto 13*8+0) &
                   rxq_header(14*8+7 downto 14*8+0) & rxq_header(15*8+7 downto 15*8+2) & "00";
      barsel <= (others => '0');
      -- BAR0 & BAR1, 64-bit address
      if (CFG_BAR0_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR1_INIT(31 downto 0) & CFG_BAR0_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar1 & cfg_bar0) and mask64)) then
          barsel(0) <= '1';
        end if;
      else
        -- BAR0, 32-bit address
        if (CFG_BAR0_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then -- bar0 is in use.
          mask32 := CFG_BAR0_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar0 and mask32)) then
            barsel(0) <= '1';
          end if;
        end if;
        -- BAR1, 32-bit address
        if (CFG_BAR1_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR1_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar1 and mask32)) then
            barsel(1) <= '1';
          end if;
        end if;
      end if;
      -- BAR2 & BAR3, 64-bit address
      if (CFG_BAR2_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR3_INIT(31 downto 0) & CFG_BAR2_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar3 & cfg_bar2) and mask64)) then
          barsel(2) <= '1';
        end if;
      else
        -- BAR2, 32-bit address
        if (CFG_BAR2_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR2_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar2 and mask32)) then
            barsel(2) <= '1';
          end if;
        end if;
        -- BAR3, 32-bit address
        if (CFG_BAR3_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR3_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar3 and mask32)) then
            barsel(3) <= '1';
          end if;
        end if;
      end if;
      -- BAR4 & BAR5, 64-bit address
      if (CFG_BAR4_INIT(2 downto 1) = "10") then
        mask64 := CFG_BAR5_INIT(31 downto 0) & CFG_BAR4_INIT(31 downto 7) & "0000000";
        if ((addrtmp64 and mask64) = ((cfg_bar5 & cfg_bar4) and mask64)) then
          barsel(4) <= '1';
        end if;
      else
        -- BAR4, 32-bit address
        if (CFG_BAR4_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR4_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar4 and mask32)) then
            barsel(4) <= '1';
          end if;
        end if;
        -- BAR5, 32-bit address
        if (CFG_BAR5_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
          mask32 := CFG_BAR5_INIT(31 downto 7) & "0000000";
          if ((addrtmp32 and mask32) = (cfg_bar5 and mask32)) then
            barsel(5) <= '1';
          end if;
        end if;
      end if;
      -- BAR ROM
      if (CFG_BAR_ROM_INIT(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := CFG_BAR_ROM_INIT(31 downto 11) & "00000000001";
        if ((addrtmp32(31 downto 0) and mask32) = (cfg_bar_rom and mask32)) then
          barsel(6) <= '1';
        end if;
      end if;
    end if; -- end of rising_edge(clk)
  end process;
end default;
--
-- functions used in msctl and slvctl
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
package msslvpkg is
  -- completion header generators:
  function cfgrd0_cpl_header(
    rhdr : in std_logic_vector -- header of the requester.
  )
  return std_logic_vector;
  function cfgwr0_cpl_header(
    rhdr : in std_logic_vector -- header of the requester.
  )
  return std_logic_vector;
  function memrd_1st_cpl_loweraddr(
    addr : in std_logic_vector; -- byte address of the request.
    first_be : in std_logic_vector -- byte enable of the 1st DW of the request.
  )
  return std_logic_vector;
  function memrd_1st_cpl_bytecount(
    length : in std_logic_vector; -- requested data size in DW unit.
    first_be : in std_logic_vector; -- byte enable of the 1st DW of the request.
    last_be : in std_logic_vector -- byte enable of the last DW of the request.
  )
  return std_logic_vector;
  function memrd_cpl_header(
    rhdr : in std_logic_vector(127 downto 0); -- header of the requester.
    bdf : in std_logic_vector(15 downto 0); -- completer id (bus id, device id, function id)
    len : in std_logic_vector( 9 downto 0);
    bcnt : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector( 6 downto 0)
  )
  return std_logic_vector;
  function memrd_1st_cpl_bytelength(
    len : in std_logic_vector(9 downto 0);
    mpls : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0);
    bytecount : in std_logic_vector(11 downto 0)
  )
  return std_logic_vector;
  function memrd_1st_cpl_length(
    len : in std_logic_vector(9 downto 0);
    mpls : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector;
  function mask_addr(
    barsel : in std_logic_vector(6 downto 0);
    header : in std_logic_vector(127 downto 0);
    bar0 : in std_logic_vector(31 downto 0);
    bar1 : in std_logic_vector(31 downto 0);
    bar2 : in std_logic_vector(31 downto 0);
    bar3 : in std_logic_vector(31 downto 0);
    bar4 : in std_logic_vector(31 downto 0);
    bar5 : in std_logic_vector(31 downto 0);
    barrom : in std_logic_vector(31 downto 0)
  )
  return std_logic_vector;
  function memwr32_header(
    len : in std_logic_vector(9 downto 0);
    addr : in std_logic_vector(31 downto 0);
    tag : in std_logic_vector(7 downto 0);
    bdf : in std_logic_vector(15 downto 0);
    tc : in std_logic_vector(2 downto 0);
    attr : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;
  function memrd32_header(
    len : in std_logic_vector(9 downto 0);
    addr : in std_logic_vector(31 downto 0);
    tag : in std_logic_vector(7 downto 0);
    bdf : in std_logic_vector(15 downto 0);
    tc : in std_logic_vector(2 downto 0);
    attr : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;
end package;
package body msslvpkg is
  -- CfgRd0 Completion header generator
  function cfgrd0_cpl_header(
    rhdr : in std_logic_vector(127 downto 0) -- header of the requester.
  )
  return std_logic_vector is
    variable chdr : std_logic_vector(127 downto 0); -- header of the completer.
    variable len : std_logic_vector(9 downto 0);
    variable bcnt : std_logic_vector(11 downto 0);
  begin
    len := conv_std_logic_vector(1, len'length);  -- 1DW length
    bcnt := conv_std_logic_vector(4, bcnt'length);  -- 4bytes to read
    chdr :=
      conv_std_logic_vector(0, 32) & -- byte12-15 not used
      x"00" & -- byte11 lower address should be 0 for CfgWr/Rd
      rhdr(6*8+7 downto 6*8+0) & -- byte10 tag
      rhdr(5*8+7 downto 4*8+0) & -- byte8,9 requester id
      bcnt(7 downto 0) & -- byte7 byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) & -- byte6 cpl stat, bcm, byte count higher 4bits.
      rhdr(9*8+7 downto 8*8+0) & -- byte4,5 completer id
      len(7 downto 0) & len(9 downto 8) & -- byte3 length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" & -- byte2 td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" & -- byte1 tc
      '0' & TPTYPE_CplD; -- byte0 fmt,type
    return chdr;
  end cfgrd0_cpl_header;
  -- CfgWr0 Completion header generator
  function cfgwr0_cpl_header(
    rhdr : in std_logic_vector(127 downto 0) -- header of the requester.
  )
  return std_logic_vector is
    variable chdr : std_logic_vector(127 downto 0); -- header of the completer.
    variable len : std_logic_vector(9 downto 0);
    variable bcnt : std_logic_vector(11 downto 0);
  begin
    len := conv_std_logic_vector(0, len'length);  -- 0DW length
    bcnt := conv_std_logic_vector(4, bcnt'length);  -- 4bytes to write
    chdr :=
      conv_std_logic_vector(0, 32) & -- byte12-15 not used
      x"00" & -- byte11 lower address should be 0 for CfgWr/Rd
      rhdr(6*8+7 downto 6*8+0) & -- byte10 tag
      rhdr(5*8+7 downto 4*8+0) & -- byte8,9 requester id
      bcnt(7 downto 0) & -- byte7 byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) & -- byte6 cpl stat, bcm, byte count higher 4bits.
      rhdr(9*8+7 downto 8*8+0) & -- byte4,5 completer id
      len(7 downto 0) & len(9 downto 8) & -- byte3 length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" & -- byte2 td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" & -- byte1 tc
      '0' & TPTYPE_Cpl; -- byte0 fmt,type
    return chdr;
  end cfgwr0_cpl_header;
  function memrd_1st_cpl_loweraddr(
    addr : in std_logic_vector(6 downto 0); -- lower 7 bits of byte address of the request.
    first_be : in std_logic_vector -- byte enable of the 1st DW of the request.
  )
  return std_logic_vector is
    variable loweraddr : std_logic_vector(6 downto 0);
  begin
    -- higher 5 bits
    loweraddr(6 downto 2) := addr(6 downto 2);
    -- lower 2 bits
    if (first_be = "0000") then
      loweraddr(1 downto 0) := "00";
    elsif (first_be(0) = '1') then
      loweraddr(1 downto 0) := "00";
    elsif (first_be(1) = '1') then
      loweraddr(1 downto 0) := "01";
    elsif (first_be(2) = '1') then
      loweraddr(1 downto 0) := "10";
    else
      loweraddr(1 downto 0) := "11";
    end if;
    return loweraddr;
  end memrd_1st_cpl_loweraddr;
  function memrd_1st_cpl_bytecount(
    length : in std_logic_vector; -- requested data size in DW unit.
    first_be : in std_logic_vector; -- byte enable of the 1st DW of the request.
    last_be : in std_logic_vector -- byte enable of the last DW of the request.
  )
  return std_logic_vector is
    variable bytecount : std_logic_vector(11 downto 0);
  begin
    if (last_be = "0000") then -- 1DW or zero length request
      if (first_be(3) = '1' and first_be(0) = '1') then
        bytecount := conv_std_logic_vector(4, bytecount'length);
      elsif (first_be(3 downto 2) = "01" and first_be(0) = '1') then
        bytecount := conv_std_logic_vector(3, bytecount'length);
      elsif (first_be(3) = '1' and first_be(1 downto 0) = "10") then
        bytecount := conv_std_logic_vector(3, bytecount'length);
      elsif (first_be = "0011") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      elsif (first_be = "0110") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      elsif (first_be = "1100") then
        bytecount := conv_std_logic_vector(2, bytecount'length);
      else
        bytecount := conv_std_logic_vector(1, bytecount'length);
      end if;
    else -- multi-DW length request
      if (first_be(0) = '1') then
        if (last_be(3) = '1') then
          bytecount := length & "00";
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(1, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        end if;
      elsif (first_be(1) = '1') then
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(1, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        end if;
      elsif (first_be(2) = '1') then
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(2, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(5, bytecount'length);
        end if;
      else
        if (last_be(3) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(3, bytecount'length);
        elsif (last_be(2) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(4, bytecount'length);
        elsif (last_be(1) = '1') then
          bytecount := (length & "00") - conv_std_logic_vector(5, bytecount'length);
        else
          bytecount := (length & "00") - conv_std_logic_vector(6, bytecount'length);
        end if;
      end if;
    end if;
    return bytecount;
  end memrd_1st_cpl_bytecount;
  -- MemRd Completion header generator
  function memrd_cpl_header(
    rhdr : in std_logic_vector(127 downto 0); -- header of the requester.
    bdf : in std_logic_vector(15 downto 0); -- completer id (bus id, device id, function id)
    len : in std_logic_vector(9 downto 0);
    bcnt : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector is
    variable chdr : std_logic_vector(127 downto 0); -- header of the completer.
  begin
    chdr :=
      conv_std_logic_vector(0, 32) & -- byte12-15 not used
      '0' & loweraddr(6 downto 0) & -- byte11 lower address
      rhdr(6*8+7 downto 6*8+0) & -- byte10 tag
      rhdr(5*8+7 downto 4*8+0) & -- byte8,9 requester id
      bcnt(7 downto 0) & -- byte7 byte count lower 8bits.
      CPLSTAT_SC & '0' & bcnt(11 downto 8) & -- byte6 cpl stat, bcm, byte count higher 4bits.
      bdf & -- byte4,5 completer id
      len(7 downto 0) & len(9 downto 8) & -- byte3 length
      "00" & rhdr(2*8+5 downto 2*8+4) & "00" & -- byte2 td,ep,attr
      '0' & rhdr(1*8+6 downto 1*8+4) & "0000" & -- byte1 tc
      '0' & TPTYPE_CplD; -- byte0 fmt,type
    return chdr;
  end memrd_cpl_header;
  function memrd_1st_cpl_bytelength(
    len : in std_logic_vector(9 downto 0);
    mpls : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0);
    bytecount : in std_logic_vector(11 downto 0)
  )
  return std_logic_vector is
    variable blen : std_logic_vector(11 downto 0);
  begin
    if (len < mpls(11 downto 2)) then
      blen := bytecount;
    else
      blen := mpls - ("00000" & loweraddr);
    end if;
    return blen;
  end memrd_1st_cpl_bytelength;
  function memrd_1st_cpl_length(
    len : in std_logic_vector(9 downto 0);
    mpls : in std_logic_vector(11 downto 0);
    loweraddr : in std_logic_vector(6 downto 0)
  )
  return std_logic_vector is
    variable dwlen : std_logic_vector(9 downto 0);
  begin
    if (len < mpls(11 downto 2)) then
      dwlen := len;
    else
      dwlen := mpls(11 downto 2) - ("00000" & loweraddr(6 downto 2));
    end if;
    return dwlen;
  end memrd_1st_cpl_length;
  function mask_addr(
    barsel : in std_logic_vector(6 downto 0);
    header : in std_logic_vector(127 downto 0);
    bar0 : in std_logic_vector(31 downto 0);
    bar1 : in std_logic_vector(31 downto 0);
    bar2 : in std_logic_vector(31 downto 0);
    bar3 : in std_logic_vector(31 downto 0);
    bar4 : in std_logic_vector(31 downto 0);
    bar5 : in std_logic_vector(31 downto 0);
    barrom : in std_logic_vector(31 downto 0)
  )
  return std_logic_vector is
    variable maddr : std_logic_vector(63 downto 0);
    variable addrtmp32 : std_logic_vector(31 downto 0);
    variable addrtmp64 : std_logic_vector(63 downto 0);
    variable mask32 : std_logic_vector(31 downto 0);
    variable mask64 : std_logic_vector(63 downto 0);
  begin
    -- alias for address fields in rxq_header
    addrtmp32 := header( 8*8+7 downto 8*8+0) & header( 9*8+7 downto 9*8+0) &
                 header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+2) & "00";
    addrtmp64 := header( 8*8+7 downto 8*8+0) & header( 9*8+7 downto 9*8+0) &
                 header(10*8+7 downto 10*8+0) & header(11*8+7 downto 11*8+0) &
                 header(12*8+7 downto 12*8+0) & header(13*8+7 downto 13*8+0) &
                 header(14*8+7 downto 14*8+0) & header(15*8+7 downto 15*8+2) & "00";
    -- BAR0 & BAR1, 64-bit address
    if (bar0(2 downto 1) = "10") then
      mask64 := bar1(31 downto 0) & bar0(31 downto 7) & "0000000";
      if (barsel(0) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR0, 32-bit address
      if (bar0(31 downto 0) /= conv_std_logic_vector(0, 32)) then -- bar0 is in use.
        mask32 := bar0(31 downto 7) & "0000000";
        if (barsel(0) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR1, 32-bit address
      if (bar1(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar1(31 downto 7) & "0000000";
        if (barsel(1) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;
    -- BAR2 & BAR3, 64-bit address
    if (bar2(2 downto 1) = "10") then
      mask64 := bar3(31 downto 0) & bar2(31 downto 7) & "0000000";
      if (barsel(2) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR2, 32-bit address
      if (bar2(31 downto 0) /= conv_std_logic_vector(0, 32)) then -- bar0 is in use.
        mask32 := bar2(31 downto 7) & "0000000";
        if (barsel(2) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR3, 32-bit address
      if (bar3(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar3(31 downto 7) & "0000000";
        if (barsel(3) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;
    -- BAR4 & BAR5, 64-bit address
    if (bar4(2 downto 1) = "10") then
      mask64 := bar5(31 downto 0) & bar4(31 downto 7) & "0000000";
      if (barsel(4) = '1') then
        maddr := (addrtmp64 and not mask64);
      end if;
    else
      -- BAR4, 32-bit address
      if (bar4(31 downto 0) /= conv_std_logic_vector(0, 32)) then -- bar0 is in use.
        mask32 := bar4(31 downto 7) & "0000000";
        if (barsel(4) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
      -- BAR5, 32-bit address
      if (bar5(31 downto 0) /= conv_std_logic_vector(0, 32)) then
        mask32 := bar5(31 downto 7) & "0000000";
        if (barsel(5) = '1') then
          maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
        end if;
      end if;
    end if;
    -- BAR ROM, 32-bit address
    if (barrom(31 downto 0) /= conv_std_logic_vector(0, 32)) then
      mask32 := barrom(31 downto 7) & "0000000";
      if (barsel(6) = '1') then
        maddr := conv_std_logic_vector(0, 32) & (addrtmp32 and not mask32);
      end if;
    end if;
    return maddr;
  end mask_addr;
  -- MemWr32 header generator
  function memwr32_header(
    len : in std_logic_vector(9 downto 0);
    addr : in std_logic_vector(31 downto 0);
    tag : in std_logic_vector(7 downto 0);
    bdf : in std_logic_vector(15 downto 0); -- completer id (bus id, device id, function id)
    tc : in std_logic_vector(2 downto 0);
    attr : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector is
    variable hdr : std_logic_vector(127 downto 0); -- header of the completer.
    variable rvsdatmp32 : std_logic_vector(31 downto 0);
  begin
    -- reverse byte order.
    rvsdatmp32 := addr(0*8+7 downto 0*8+0) & addr(1*8+7 downto 1*8+0) &
                  addr(2*8+7 downto 2*8+0) & addr(3*8+7 downto 3*8+0);
    hdr :=
      conv_std_logic_vector(0, 32) & -- byte12-15 not used
      rvsdatmp32(31 downto 0) & -- byte8-11 address
      x"ff" & -- byte7 last BE, first BE
      tag & -- byte6 tag
      bdf & -- byte4,5 completer id
      len(7 downto 0) & -- byte3 length
      "00" & attr & "00" & len(9 downto 8) & -- byte2 no td, no ep, attr
      '0' & tc & "0000" & -- byte1 tc
      '0' & TPTYPE_MWr32; -- byte0 fmt,type
    return hdr;
  end memwr32_header;
  -- MemRd32 header generator
  function memrd32_header(
    len : in std_logic_vector(9 downto 0);
    addr : in std_logic_vector(31 downto 0);
    tag : in std_logic_vector(7 downto 0);
    bdf : in std_logic_vector(15 downto 0); -- completer id (bus id, device id, function id)
    tc : in std_logic_vector(2 downto 0);
    attr : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector is
    variable hdr : std_logic_vector(127 downto 0); -- header of the completer.
    variable rvsdatmp32 : std_logic_vector(31 downto 0);
  begin
    -- reverse byte order.
    rvsdatmp32 := addr(0*8+7 downto 0*8+0) & addr(1*8+7 downto 1*8+0) &
                  addr(2*8+7 downto 2*8+0) & addr(3*8+7 downto 3*8+0);
    hdr :=
      conv_std_logic_vector(0, 32) & -- byte12-15 not used
      rvsdatmp32(31 downto 0) & -- byte8-11 address
      x"ff" & -- byte7 last BE, first BE
      tag & -- byte6 tag
      bdf & -- byte4,5 completer id
      len(7 downto 0) & -- byte3 length
      "00" & attr & "00" & len(9 downto 8) & -- byte2 no td, no ep, attr
      '0' & tc & "0000" & -- byte1 tc
      '0' & TPTYPE_MRd32; -- byte0 fmt,type
    return hdr;
  end memrd32_header;
end msslvpkg;
--
-- Flow Control Buffer
--
-- . Single clock FIFO, show ahead.
-- . Read/write operation is valid only when 'vc' is
-- matched with 'VIRTUAL_CHANNEL'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
entity fcbuf is
  generic (
    VIRTUAL_CHANNEL : integer;
    DATA_WIDTH : integer;
    DATA_DEPTH : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic; -- active high reset
    vc : in std_logic_vector(2 downto 0);
    wr : in std_logic;
    rd : in std_logic;
    data : in std_logic_vector(DATA_WIDTH-1 downto 0);
    empty : out std_logic;
    full : out std_logic;
    usedw : out std_logic_vector(8 downto 0);
    q : out std_logic_vector(DATA_WIDTH-1 downto 0)
  );
end fcbuf;
architecture default of fcbuf is
  component scfifo
    generic (
      LPM_WIDTH : natural := DATA_WIDTH;
      LPM_NUMWORDS : natural := DATA_DEPTH;
      LPM_SHOWAHEAD : string := "ON";
      ADD_RAM_OUTPUT_REGISTER : string := "ON";
      OVERFLOW_CHECKING : string := "OFF";
      UNDERFLOW_CHECKING : string := "OFF";
      USE_EAB : string := "ON";
      INTENDED_DEVICE_FAMILY : string := "Stratix II GX";
      LPM_TYPE : string := "scfifo"
    );
    port (
      rdreq : in std_logic;
      empty : out std_logic;
      full : out std_logic;
      clock : in std_logic;
      sclr : in std_logic;
      q : out std_logic_vector(LPM_WIDTH-1 downto 0);
      wrreq : in std_logic;
      data : in std_logic_vector(LPM_WIDTH-1 downto 0)
    );
  end component;
  signal we : std_logic;
  signal full0 : std_logic;
  signal fullout : std_logic;
begin
    buf0 : scfifo
      generic map (
        LPM_NUMWORDS => DATA_DEPTH,
        LPM_WIDTH => data'length
      )
      port map (
        clock => clk,
        sclr => rst,
        wrreq => we,
        rdreq => rd,
        data => data,
        empty => empty,
        full => full0,
        q => q
      );
    we <= '1' when wr = '1' and vc = conv_std_logic_vector(VIRTUAL_CHANNEL, vc'length)
              else '0';
   full <= full0;
end default;
--
-- Tx FC Buffer Input Arbiter
-- . receives write requests from slvctl:c_req, msctl:np_req, and msctl:p_req.
-- . returns acknowledge to one of them with first come first served policy.
-- . c_req has the highest priority and p_req has the 2nd.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;
entity txarb is
  port (
    clk : in std_logic;
    rst : in std_logic; -- active high
    c_req : in std_logic;
    c_ack : out std_logic;
    p_req : in std_logic;
    p_ack : out std_logic;
    np_req : in std_logic;
    np_ack : out std_logic;
    srcsel : out std_logic_vector(1 downto 0) -- TPBUFTYPE_P, _NP, or _C
  );
end txarb;
architecture default of txarb is
  signal c_ack0 : std_logic;
  signal p_ack0 : std_logic;
  signal np_ack0 : std_logic;
begin
  c_ack <= c_ack0;
  p_ack <= p_ack0;
  np_ack <= np_ack0;
  process(c_req, np_req, p_req)
  begin
    if (c_req = '1') then
      if (c_ack0 = '1') then -- already locked.
        c_ack0 <= '1';
      elsif (np_ack0 = '0' and p_ack0 = '0') then -- no one locked.
        c_ack0 <= '1';
      end if;
    else
      c_ack0 <= '0';
    end if;
    if (p_req = '1') then
      if (p_ack0 = '1') then -- already locked.
        p_ack0 <= '1';
      elsif (c_ack0 = '0' and np_ack0 = '0') then -- no one locked.
        if (c_req = '0') then -- no pending request from slvctl:c_req.
          p_ack0 <= '1';
        else
          p_ack0 <= '0';
        end if;
      end if;
    else
      p_ack0 <= '0';
    end if;
    if (np_req = '1') then
      if (np_ack0 = '1') then -- already locked.
        np_ack0 <= '1';
      elsif (c_ack0 = '0' and p_ack0 = '0') then -- no one locked.
        if (c_req = '0' and p_req = '0') then -- no pending request from slvctl:c_req nor msctl:p_req.
          np_ack0 <= '1';
        else
          np_ack0 <= '0';
        end if;
      end if;
    else
      np_ack0 <= '0';
    end if;
  end process;
  process(p_ack0, np_ack0)
  begin
    srcsel <= TPBUFTYPE_C;
    if (p_ack0 = '1') then
      srcsel <= TPBUFTYPE_P;
    elsif (np_ack0 = '1') then
      srcsel <= TPBUFTYPE_NP;
    end if;
  end process;
end default;
--
-- PCIe Master (Initiator) Controller
-- . transmit Posted and Non-posted TLPs to Tx FC bufs.
-- . communicate with application layer (DMA manager) as slave.
-- . receive Completion TLPs from Rx FC bufs.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.msslvpkg.all;
entity msctl is
  generic (
    NLANE : integer;
    NDMACH : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic; -- active high
    cat_ph : in std_logic_vector(7 downto 0);
    cat_pd : in std_logic_vector(11 downto 0);
    cat_nph : in std_logic_vector(7 downto 0);
    cat_npd : in std_logic_vector(11 downto 0);
    bdf : in std_logic_vector(15 downto 0); -- completer id
    -- interface to TxFCbuf and txarb
    p_txwr_req : out std_logic;
    p_txwr_ack : in std_logic;
    p_txwr_bufsel : out std_logic;
    txwr_ph : out std_logic;
    tx_ph : out std_logic_vector(127 downto 0);
    txwr_pd : out std_logic;
    tx_pd : out std_logic_vector(NLANE*16-1 downto 0);
    np_txwr_req : out std_logic;
    np_txwr_ack : in std_logic;
    np_txwr_bufsel : out std_logic;
    txwr_nph : out std_logic;
    tx_nph : out std_logic_vector(127 downto 0);
    txwr_npd : out std_logic;
    tx_npd : out std_logic_vector(NLANE*16-1 downto 0);
    -- interface to master controllers in the user-side logic.
    ms_write : out std_logic;
    ms_wrchannel : out std_logic_vector(NDMACH-1 downto 0);
    ms_wraddr : out std_logic_vector(31 downto 0);
    ms_wrdata : in std_logic_vector(NLANE*16-1 downto 0);
    -- interface to DMA controllers
    dma_fifocnt_off : out std_logic_vector(4 downto 0); -- width >= log2(NLANE*2)
    dmach_wrreq : in std_logic_vector(NDMACH-1 downto 0);
    dmach_wrack : out std_logic_vector(NDMACH-1 downto 0);
    dmach_rdreq : in std_logic_vector(NDMACH-1 downto 0);
    dmach_rdack : out std_logic_vector(NDMACH-1 downto 0);
    dmach_length : in each10b(NDMACH-1 downto 0);
    dmach_paddr : in each64b(NDMACH-1 downto 0);
    dmach_laddr : in each32b(NDMACH-1 downto 0)
  );
end msctl;
architecture default of msctl is
  type WR_STATE is (
    IDLE,
    WAIT_MemWr,
    READY_MemWr,
    BURST_MemWr
  );
  type RD_STATE is (
    IDLE,
    WAIT_MemRd
  );
  signal wrsm : WR_STATE;
  signal rdsm : RD_STATE;
  -- TxFCbuf credit consumed
  signal cct_nph : std_logic_vector(7 downto 0);
  signal cct_ph : std_logic_vector(7 downto 0);
  signal cct_npd : std_logic_vector(11 downto 0);
  signal cct_pd : std_logic_vector(11 downto 0);
  signal wr : std_logic;
  signal wrladdr : std_logic_vector(31 downto 0);
  signal wrpaddr : std_logic_vector(63 downto 0);
  signal wrchannel : std_logic_vector(NDMACH-1 downto 0);
  signal wrack : std_logic_vector(NDMACH-1 downto 0);
  signal wrcnt : std_logic_vector(PAYLOAD_WIDTH+2-1 downto 0); -- size in byte.
  signal rd : std_logic;
  signal rdpaddr : std_logic_vector(63 downto 0);
  signal rdack : std_logic_vector(NDMACH-1 downto 0);
  signal rdcnt : std_logic_vector(PAYLOAD_WIDTH+2-1 downto 0);
  signal wrtag, rdtag : std_logic_vector(7 downto 0);
  signal crt_ph : std_logic_vector(7 downto 0);
  signal crt_pd : std_logic_vector(11 downto 0);
begin
  ms_wrchannel <= wrchannel;
  dmach_wrack <= wrack;
  dmach_rdack <= rdack;
  ms_write <= wr;
  ms_wraddr <= wrladdr;
  tx_pd <= ms_wrdata;
  tx_npd <= (others => '0'); -- not used (reserved for non-posted data).
  --
  -- DMA Write State Machine
  -- transmit MemWr TLP on request from a DMA channel.
  -- Tx Posted Buffer is inter locked by the channel.
  -- No other channel can start a new transfer until the one completes.
  --
  process(clk, rst)
    variable wrreqtmp : std_logic;
    variable wrcnttmp : std_logic_vector(wrcnt'range); -- size in byte.
    variable cat_pdtmp : std_logic_vector(11 downto 0);
  begin
    if (rst = '1') then
      cct_ph <= conv_std_logic_vector(0, cct_ph'length);
      cct_pd <= conv_std_logic_vector(0, cct_pd'length);
      p_txwr_req <= '0';
      wrchannel <= (others => '0');
      wrack <= (others => '0');
      wrsm <= IDLE;
    elsif (rising_edge(clk)) then
      wr <= '0';
      p_txwr_bufsel <= '0';
      txwr_ph <= '0';
      txwr_pd <= '0';
      dma_fifocnt_off <= (others => '0');
      case wrsm is
        when IDLE =>
          -- accept write request from the youngest channel.
          wrchannel <= (others => '0');
          wrack <= (others => '0');
          wrreqtmp := '0';
          for i in 0 to NDMACH-1 loop
            if (dmach_wrreq(i) = '1' and wrreqtmp = '0') then
              wrreqtmp := '1';
              wrack(i) <= '1'; -- acknowledge.
              wrchannel(i) <= '1'; -- alloc DMA channel.
              wrladdr <= dmach_laddr(i);
              wrpaddr <= dmach_paddr(i);
              wrcnttmp := dmach_length(i) & "00"; -- size in byte.
              wrcnt <= wrcnttmp;
              crt_ph <= unsigned(cct_ph) + 1;
              crt_pd <= cct_pd + length_to_credit(wrcnttmp(wrcnt'left downto 2));
              wrtag <= conv_std_logic_vector(i, wrtag'length);
              p_txwr_req <= '1'; -- write request to TxFCbuf.
              wrsm <= WAIT_MemWr;
            end if;
          end loop;
        when WAIT_MemWr =>
          if (p_txwr_ack = '1' and
              have_enough_credit(cat_ph, crt_ph) = '1' and
              have_enough_credit(cat_pd, crt_pd) = '1') then
            tx_ph <= memwr32_header(wrcnt(wrcnt'left downto 2),
                                    wrpaddr(31 downto 0),
                                    wrtag, bdf, "000", "00");
            cct_ph <= crt_ph;
            cct_pd <= crt_pd;
            wr <= '1';
            p_txwr_bufsel <= '1';
            txwr_ph <= '1';
            wrsm <= READY_MemWr;
          end if;
        when READY_MemWr =>
          -- keep wr asserted if burst continues longer than 1clk.
          if (wrcnt > conv_std_logic_vector(NLANE*2, wrcnt'length)) then
            wr <= '1';
            wrladdr <= unsigned(wrladdr) + NLANE*2;
          end if;
          txwr_pd <= '1';
          p_txwr_req <= '0'; -- !!!
          wrsm <= BURST_MemWr;
        when BURST_MemWr =>
          -- deassert write acknowledge 4 or 3-clk earlier than end of burst
          -- (for better perormance).
-- if (NLANE >= 4) then
-- if (wrcnt = conv_std_logic_vector(NLANE*2*4, wrcnt'length)) then
-- wrack <= (others => '0');
-- dma_fifocnt_off <= conv_std_logic_vector(NLANE*2*2, dma_fifocnt_off'length);
-- end if;
-- if (wrcnt = conv_std_logic_vector(NLANE*2*3, wrcnt'length)) then
-- wrack <= (others => '0');
-- dma_fifocnt_off <= conv_std_logic_vector(NLANE*2*1, dma_fifocnt_off'length);
-- end if;
-- end if;
          if (wrcnt <= conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wrack <= (others => '0');
            dma_fifocnt_off <= conv_std_logic_vector(0, dma_fifocnt_off'length);
          end if;
          -- deassert wr 2-clk earlier than end of burst.
          -- release write channel at the same clk.
          if (wrcnt > conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wr <= '1';
            wrladdr <= unsigned(wrladdr) + NLANE*2;
          end if;
          if (wrcnt <= conv_std_logic_vector(NLANE*2*2, wrcnt'length)) then
            wrchannel <= (others => '0');
          end if;
          if (wrcnt > conv_std_logic_vector(NLANE*2, wrcnt'length)) then
            -- proceed burst.
            -- receive wdata from user-side logic and push it into TxFCbuf.
            wrcnt <= unsigned(wrcnt) - NLANE*2;
            txwr_pd <= '1';
          else
            -- burst going to end.
            wrsm <= IDLE;
            if (NLANE >= 4) then -- !!!
              -- skip IDLE state and go directly to WAIT_MemWr if possible.
              wrchannel <= (others => '0');
              wrack <= (others => '0');
              wrreqtmp := '0';
              for i in 0 to NDMACH-1 loop
                if (dmach_wrreq(i) = '1' and wrreqtmp = '0') then
                  wrreqtmp := '1';
                  wrack(i) <= '1'; -- acknowledge.
                  wrchannel(i) <= '1'; -- alloc DMA channel.
                  wrladdr <= dmach_laddr(i);
                  wrpaddr <= dmach_paddr(i);
                  wrcnttmp := dmach_length(i) & "00"; -- size in byte.
                  wrcnt <= wrcnttmp;
                  crt_ph <= unsigned(cct_ph) + 1;
                  crt_pd <= cct_pd + length_to_credit(wrcnttmp(wrcnt'left downto 2));
                  wrtag <= conv_std_logic_vector(i, wrtag'length);
                  p_txwr_req <= '1'; -- write request to TxFCbuf.
                  wrsm <= WAIT_MemWr;
                end if;
              end loop;
              if (wrreqtmp = '1') then
                wrsm <= WAIT_MemWr;
              end if;
            end if;
          end if;
        when others =>
      end case;
    end if;
  end process;
  --
  -- DMA Read State Machine
  -- transmit MemRd TLP on request from a DMA channel.
  --
  process(clk, rst)
    variable rdreqtmp : std_logic;
    variable crt_nphtmp : std_logic_vector(7 downto 0);
  begin
    if (rst = '1') then
      cct_nph <= conv_std_logic_vector(0, cct_nph'length);
      cct_npd <= conv_std_logic_vector(0, cct_npd'length);
      np_txwr_req <= '0';
      rdack <= (others => '0');
      rdsm <= IDLE;
    elsif (rising_edge(clk)) then
      np_txwr_bufsel <= '0';
      txwr_nph <= '0';
      txwr_npd <= '0';
      case rdsm is
        when IDLE =>
          np_txwr_req <= '0'; -- !!!
          -- accept read request from the youngest channel.
          rdack <= (others => '0');
          rdreqtmp := '0';
          for i in 0 to NDMACH-1 loop
            if (dmach_rdreq(i) = '1' and rdreqtmp = '0') then
              rdreqtmp := '1';
              rdack(i) <= '1';
              rdpaddr <= dmach_paddr(i);
              rdcnt <= dmach_length(i) & "00";
              rdtag <= conv_std_logic_vector(i, rdtag'length);
              np_txwr_req <= '1';
              rdsm <= WAIT_MemRd;
            end if;
          end loop;
        when WAIT_MemRd =>
          crt_nphtmp := unsigned(cct_nph) + 1;
          if (np_txwr_ack = '1' and have_enough_credit(cat_nph, crt_nphtmp) = '1') then
            tx_nph <= memrd32_header(rdcnt(rdcnt'left downto 2),
                                      rdpaddr(31 downto 0),
                                      rdtag, bdf, "000", "00");
            cct_nph <= crt_nphtmp;
            np_txwr_bufsel <= '1';
            txwr_nph <= '1';
            rdack <= (others => '0');
            rdsm <= IDLE;
          end if;
        when others =>
      end case;
    end if;
  end process;
end default;
--
-- Data-Link Layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
entity datalink is
  generic (
    -- Flow-control buffer size
    MAX_PAYLOAD : integer;
    CA_PH_VC0_INIT : integer;
    CA_NPH_VC0_INIT : integer;
    CA_CH_VC0_INIT : integer;
    CA_PD_VC0_INIT : integer;
    CA_NPD_VC0_INIT : integer;
    CA_CD_VC0_INIT : integer;
    TXDPBUF_DEPTH : integer := 6;
    NLANE : integer;
    SIMULATION : integer
  );
  port (
    st2_trigger : out std_logic;
    txdpbuffull : out std_logic;
    clk : in std_logic;
    rst : in std_logic; -- active high
    -- DLL-MAC interface
    rxwedp : in std_logic;
    rxdp : in std_logic_vector(47 downto 0);
    txemptydp : out std_logic;
    txrdreqdp : in std_logic;
    txdp : out std_logic_vector(47 downto 0);
    rxwetp : in std_logic;
    rxbetp : in std_logic_vector(NLANE*2-1 downto 0);
    rxtptop : in std_logic;
    rxtp : in std_logic_vector(NLANE*16-1 downto 0);
    txtoptp : out std_logic;
    txwetp : out std_logic;
    txbetp : out std_logic_vector(NLANE*2-1 downto 0);
    txtp : out std_logic_vector(NLANE*16-1 downto 0);
    -- TL-DLL interface
    rxtop : out std_logic;
    rxwe : out std_logic;
    rx : out std_logic_vector(NLANE*16-1 downto 0);
    txtop : in std_logic;
    txwe : in std_logic;
    txbe : in std_logic_vector(NLANE*2-1 downto 0);
    tx : in std_logic_vector(NLANE*16-1 downto 0);
    tpsize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
    -- CA notified by transaction layer
    ca_nph : in std_logic_vector(7 downto 0);
    ca_ph : in std_logic_vector(7 downto 0);
    ca_ch : in std_logic_vector(7 downto 0);
    ca_npd : in std_logic_vector(11 downto 0);
    ca_pd : in std_logic_vector(11 downto 0);
    ca_cd : in std_logic_vector(11 downto 0);
    ca_p_flush : in std_logic;
    ca_np_flush : in std_logic;
    ca_c_flush : in std_logic;
    -- notify CL to transaction layer
    cl_nph : out std_logic_vector( 7 downto 0);
    cl_ph : out std_logic_vector( 7 downto 0);
    cl_ch : out std_logic_vector( 7 downto 0);
    cl_npd : out std_logic_vector(11 downto 0);
    cl_pd : out std_logic_vector(11 downto 0);
    cl_cd : out std_logic_vector(11 downto 0);
    inf_cl_ph : out std_logic;
    inf_cl_pd : out std_logic;
    inf_cl_nph : out std_logic;
    inf_cl_npd : out std_logic;
    inf_cl_ch : out std_logic;
    inf_cl_cd : out std_logic;
    -- core status
    phy_linkup : in std_logic;
    dl_linkup : out std_logic
  );
end datalink;
architecture default of datalink is
  component acknakmgr
    generic (
      NLANE : integer := NLANE;
      MAX_PAYLOAD : integer := MAX_PAYLOAD
    );
    port (
      clk : in std_logic;
      rst : in std_logic;
      we : in std_logic;
      seqno : in std_logic_vector(11 downto 0);
      lcrcerr : in std_logic;
      next_seqno : out std_logic_vector(11 downto 0);
      req : out std_logic;
      antype : out std_logic_vector(7 downto 0);
      data : out std_logic_vector(23 downto 0);
      ack : in std_logic
    );
  end component;
  component updatefcmgr
    generic (
      NLANE : integer := NLANE;
      MAX_PAYLOAD : integer := MAX_PAYLOAD
    );
    port (
      clk : in std_logic;
      rst : in std_logic;
      -- CA notified by transaction layer
      ca_ph : in std_logic_vector( 7 downto 0);
      ca_nph : in std_logic_vector( 7 downto 0);
      ca_ch : in std_logic_vector( 7 downto 0);
      ca_pd : in std_logic_vector(11 downto 0);
      ca_npd : in std_logic_vector(11 downto 0);
      ca_cd : in std_logic_vector(11 downto 0);
      -- notify CL to transaction layer
      cl_ph : out std_logic_vector( 7 downto 0);
      cl_nph : out std_logic_vector( 7 downto 0);
      cl_ch : out std_logic_vector( 7 downto 0);
      cl_pd : out std_logic_vector(11 downto 0);
      cl_npd : out std_logic_vector(11 downto 0);
      cl_cd : out std_logic_vector(11 downto 0);
      inf_cl_ph : out std_logic;
      inf_cl_pd : out std_logic;
      inf_cl_nph : out std_logic;
      inf_cl_npd : out std_logic;
      inf_cl_ch : out std_logic;
      inf_cl_cd : out std_logic;
      -- Rx InitFC/UpdateFC DLLP
      we : in std_logic;
      fctype_in : in std_logic_vector(7 downto 0);
      hsize_in : in std_logic_vector(7 downto 0);
      dsize_in : in std_logic_vector(11 downto 0);
      -- Tx UpdateFC DLLP
      flush_p : in std_logic;
      flush_np : in std_logic;
      flush_c : in std_logic;
      req : out std_logic;
      fctype_out : out std_logic_vector(7 downto 0);
      hsize_out : out std_logic_vector(7 downto 0);
      dsize_out : out std_logic_vector(11 downto 0);
      ack : in std_logic
    );
  end component;
  component txdpctl
    port (
      clk : in std_logic;
      rst : in std_logic;
      dptype : in std_logic_vector(7 downto 0);
      data : in std_logic_vector(23 downto 0);
      kickoff : in std_logic;
      dp : out std_logic_vector(47 downto 0);
      done : out std_logic
    );
  end component;
  component rxdpctl
    port (
      clk : in std_logic;
      rst : in std_logic;
      we : in std_logic;
      dp : in std_logic_vector(47 downto 0);
      dptype : out std_logic_vector(7 downto 0);
      dpbody : out std_logic_vector(23 downto 0);
      crcerr : out std_logic;
      done : out std_logic
    );
  end component;
  component txtpctl
    generic (
      NLANE : integer := NLANE
    );
    port (
      topin : in std_logic;
      wein : in std_logic;
      bein : in std_logic_vector(NLANE*2-1 downto 0);
      tpbody : in std_logic_vector(NLANE*16-1 downto 0);
      tpbodysize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
      topout : out std_logic;
      weout : out std_logic;
      beout : out std_logic_vector(NLANE*2-1 downto 0);
      tp : out std_logic_vector(NLANE*16-1 downto 0);
      clk : in std_logic;
      rst : in std_logic
      );
  end component;
  component rxtpctl
    generic (
      NLANE : integer := NLANE
    );
    port (
      clk : in std_logic;
      rst : in std_logic;
      wein : in std_logic;
      be : in std_logic_vector(NLANE*2-1 downto 0);
      tptop : in std_logic;
      tp : in std_logic_vector(NLANE*16-1 downto 0);
      next_seqno : in std_logic_vector(11 downto 0);
      topout : out std_logic;
      weout : out std_logic;
      tpbody : out std_logic_vector(NLANE*16-1 downto 0);
      seqno : out std_logic_vector(11 downto 0);
      crcerr : out std_logic;
      done : out std_logic
    );
  end component;
  component scfifo
    generic (
      ADD_RAM_OUTPUT_REGISTER : string := "ON";
      INTENDED_DEVICE_FAMILY : string := "Stratix II GX";
      LPM_NUMWORDS : natural;
      LPM_SHOWAHEAD : string := "ON";
      LPM_TYPE : string := "scfifo";
      LPM_WIDTH : natural;
      OVERFLOW_CHECKING : string := "OFF";
      UNDERFLOW_CHECKING : string := "OFF";
      USE_EAB : string := "ON"
    );
    port (
      rdreq : in std_logic;
      empty : out std_logic;
      clock : in std_logic;
      sclr : in std_logic;
      q : out std_logic_vector(LPM_WIDTH-1 downto 0);
      wrreq : in std_logic;
      data : in std_logic_vector(LPM_WIDTH-1 downto 0);
      full : out std_logic
    );
  end component;
  -- DLCMSM States
  type DLCMSM_STATE is (
    DL_INACTIVE,
    DL_INIT1,
    DL_INIT2,
    DL_ACTIVE,
    INVALID_STATE
  );
-- attribute enum_encoding: string;
-- attribute enum_encoding of DLCMSM_STATE : type is "000 001 010 011 100 101 110 111";
  -- internal use
  signal dlcmsm : DLCMSM_STATE;
  signal prev_dlcmsm : DLCMSM_STATE;
  signal txdp_type : std_logic_vector(7 downto 0);
  signal txdp_data : std_logic_vector(23 downto 0);
  signal txdp_kickoff : std_logic;
  signal txdp_done_pulse : std_logic;
  signal txdp_done_p : std_logic;
  signal txdp_done_np : std_logic;
  signal txdp_done_c : std_logic;
  signal txdp_dp : std_logic_vector(47 downto 0);
  signal txdp_clear : std_logic;
  signal rxdp_type : std_logic_vector(7 downto 0);
  signal rxdp_data : std_logic_vector(23 downto 0);
  signal rxdp_dcrcerr : std_logic;
  signal rxdp_done_pulse : std_logic;
  signal rxdp_done_p : std_logic;
  signal rxdp_done_np : std_logic;
  signal rxdp_done_c : std_logic;
  signal rxtp_seqno : std_logic_vector(11 downto 0);
  signal rxtp_lcrcerr : std_logic;
  signal rxtp_done_pulse : std_logic;
  signal an_req : std_logic;
  signal an_ack : std_logic;
  signal an_type : std_logic_vector(7 downto 0);
  signal an_data : std_logic_vector(23 downto 0);
  signal next_seqno : std_logic_vector(11 downto 0);
  signal fc_we : std_logic;
  signal fc_type_in : std_logic_vector(7 downto 0);
  signal fc_hsize_in : std_logic_vector(7 downto 0);
  signal fc_dsize_in : std_logic_vector(11 downto 0);
  signal fc_req : std_logic;
  signal fc_ack : std_logic;
  signal fc_type_out : std_logic_vector(7 downto 0);
  signal fc_hsize_out : std_logic_vector(7 downto 0);
  signal fc_dsize_out : std_logic_vector(11 downto 0);
  signal dl_linkup0 : std_logic;
  signal txdpcleared : std_logic;
-- synthesis translate_off
  attribute keep : boolean;
  attribute keep of txdp_done_pulse : signal is true;
  attribute keep of txdpbuffull : signal is true;
  attribute keep of rxtp_seqno : signal is true;
  attribute keep of rxtp_lcrcerr : signal is true;
  attribute keep of rxtp_done_pulse : signal is true;
-- synthesis translate_on
begin
  dl_linkup <= dl_linkup0;
  process (clk, rst)
  begin
    if (rst = '1') then
      dl_linkup0 <= '0';
      txdp_kickoff <= '0';
      txdp_done_p <= '0';
      txdp_done_np <= '0';
      txdp_done_c <= '0';
      rxdp_done_p <= '0';
      rxdp_done_np <= '0';
      rxdp_done_c <= '0';
      if (SIMULATION = 0) then
        dlcmsm <= DL_INACTIVE;
      else
        dlcmsm <= DL_ACTIVE; -- skip training sequence for simulation.
      end if;
    elsif (rising_edge(clk)) then
      dl_linkup0 <= '0';
      prev_dlcmsm <= dlcmsm;
      case dlcmsm is
        when DL_INACTIVE =>
          txdp_kickoff <= '0';
          txdp_done_p <= '0';
          txdp_done_np <= '0';
          txdp_done_c <= '0';
          rxdp_done_p <= '0';
          rxdp_done_np <= '0';
          rxdp_done_c <= '0';
          txdpcleared <= '0';
          if (phy_linkup = '1') then
            dlcmsm <= DL_INIT1;
          end if;
        when DL_INIT1 =>
          -- repeatedly send InitFC1-P,Np,Cpl w/Credit Allocated.
          if (dlcmsm /= prev_dlcmsm) then
            txdp_type <= DPTYPE_INITFC1_P_VC0;
            txdp_data <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                             "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
            txdp_kickoff <= '1';
          elsif (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC1_C_VC0) then
              txdp_type <= DPTYPE_INITFC1_P_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_P_VC0) then
              txdp_type <= DPTYPE_INITFC1_NP_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_NPH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_NPD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            else
              txdp_type <= DPTYPE_INITFC1_C_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_CH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_CD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            end if;
          else
            txdp_kickoff <= '0';
          end if;
          -- sent at least one InitFC1-P,Np,Cpl.
          if (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC1_P_VC0) then
              txdp_done_p <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_NP_VC0) then
              txdp_done_np <= '1';
            elsif (txdp_type = DPTYPE_INITFC1_C_VC0) then
              txdp_done_c <= '1';
            end if;
          end if;
          -- receive InitFC1-P,Np,Cpl and set Credit Limit.
          fc_we <= '0';
          if (rxdp_done_pulse = '1') then
            if ((rxdp_type = DPTYPE_INITFC1_P_VC0) or
               (rxdp_type = DPTYPE_INITFC1_NP_VC0) or
               (rxdp_type = DPTYPE_INITFC1_C_VC0)) then
              fc_we <= '1';
              fc_type_in <= rxdp_type;
              fc_hsize_in <= rxdp_data(21 downto 14);
              fc_dsize_in <= rxdp_data(11 downto 0);
              if (rxdp_type = DPTYPE_INITFC1_P_VC0) then
                rxdp_done_p <= '1';
              elsif (rxdp_type = DPTYPE_INITFC1_NP_VC0) then
                rxdp_done_np <= '1';
              elsif (rxdp_type = DPTYPE_INITFC1_C_VC0) then
                rxdp_done_c <= '1';
              end if;
            end if;
          end if;
          -- sent and received at least one InitFC1-P,Np,Cpl.
          if (txdp_done_p = '1' and txdp_done_np = '1' and txdp_done_c = '1' and
              rxdp_done_p = '1' and rxdp_done_np = '1' and rxdp_done_c = '1') then
            txdp_kickoff <= '0';
            txdp_done_p <= '0';
            txdp_done_np <= '0';
            txdp_done_c <= '0';
            rxdp_done_p <= '0';
            rxdp_done_np <= '0';
            rxdp_done_c <= '0';
            fc_we <= '0';
            dlcmsm <= DL_INIT2;
          end if;
          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;
        when DL_INIT2 =>
-- dl_linkup0 <= '1';
          -- repeatedly send InitFC2-P,Np,Cpl w/Credit Allocated.
          if (dlcmsm /= prev_dlcmsm) then
            txdp_type <= DPTYPE_INITFC2_P_VC0;
            txdp_data <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                             "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
            txdp_kickoff <= '1';
          elsif (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC2_C_VC0) then
              txdp_type <= DPTYPE_INITFC2_P_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_PH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_PD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_P_VC0) then
              txdp_type <= DPTYPE_INITFC2_NP_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_NPH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_NPD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            else
              txdp_type <= DPTYPE_INITFC2_C_VC0;
              txdp_data <= "00" & conv_std_logic_vector(CA_CH_VC0_INIT, 8) &
                              "00" & conv_std_logic_vector(CA_CD_VC0_INIT, 12);
              txdp_kickoff <= '1';
            end if;
          else
            txdp_kickoff <= '0';
          end if;
          -- sent at least one InitFC2-P,Np,Cpl.
          if (txdp_done_pulse = '1') then
            if (txdp_type = DPTYPE_INITFC2_P_VC0) then
              txdp_done_p <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_NP_VC0) then
              txdp_done_np <= '1';
            elsif (txdp_type = DPTYPE_INITFC2_C_VC0) then
              txdp_done_c <= '1';
            end if;
          end if;
          -- receive InitFC2-P,Np,Cpl and set Credit Limit.
          if (rxdp_done_pulse = '1') then
            if (rxdp_type = DPTYPE_INITFC2_P_VC0) then
              rxdp_done_p <= '1';
            elsif (rxdp_type = DPTYPE_INITFC2_NP_VC0) then
              rxdp_done_np <= '1';
            elsif (rxdp_type = DPTYPE_INITFC2_C_VC0) then
              rxdp_done_c <= '1';
            end if;
          end if;
          -- sent and received at least one InitFC2-P,Np,Cpl.
          if (txdp_done_p = '1' and txdp_done_np = '1' and txdp_done_c = '1' and
              rxdp_done_p = '1' and rxdp_done_np = '1' and rxdp_done_c = '1') then
            txdp_kickoff <= '0';
            txdp_done_p <= '0';
            txdp_done_np <= '0';
            txdp_done_c <= '0';
            rxdp_done_p <= '0';
            rxdp_done_np <= '0';
            rxdp_done_c <= '0';
            dlcmsm <= DL_ACTIVE;
          end if;
          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;
        when DL_ACTIVE =>
          st2_trigger <= '0';
          if (rxdp_done_pulse = '1') then
            -- Ack/Nak
            if (rxdp_type = DPTYPE_NAK) then
              st2_trigger <= '1';
            end if;
          end if;
          dl_linkup0 <= '1';
          txdp_clear <= '0';
          an_ack <= '0';
          fc_ack <= '0';
          txdp_kickoff <= '0';
          -- transmit Ack/Nak DLLP.
          if (an_req = '1' and an_ack = '0') then
            txdp_type <= an_type;
            txdp_data <= an_data;
            txdp_kickoff <= '1';
            an_ack <= '1';
            -- the 1st TLP from Root Complex arrived.
            -- clear unnecessary InitFC2s in TxDPbuf.
            if (an_data = conv_std_logic_vector(0, an_data'length) and txdpcleared = '0') then
              txdp_clear <= '1';
              txdpcleared <= '1';
            end if;
          -- transmit UpdateFC DLLP.
          elsif (fc_req = '1' and fc_ack = '0') then
            txdp_type <= fc_type_out;
            txdp_data <= "00" & fc_hsize_out & "00" & fc_dsize_out;
            txdp_kickoff <= '1';
            fc_ack <= '1';
          end if;
          -- receive UpdateFC-P,Np,Cpl and set Credit Limit.
          fc_we <= '0';
          if (rxdp_done_pulse = '1') then
            if ((rxdp_type = DPTYPE_UPDATEFC_P_VC0) or
               (rxdp_type = DPTYPE_UPDATEFC_NP_VC0) or
               (rxdp_type = DPTYPE_UPDATEFC_C_VC0)) then
              fc_we <= '1';
              fc_type_in <= rxdp_type;
              fc_hsize_in <= rxdp_data(21 downto 14);
              fc_dsize_in <= rxdp_data(11 downto 0);
            end if;
          end if;
          if (phy_linkup = '0') then
            dlcmsm <= DL_INACTIVE;
          end if;
        when others =>
          dlcmsm <= DL_INACTIVE;
      end case;
    end if;
  end process;
  acknakmgr0 : acknakmgr
    port map (
      clk => clk,
      rst => not dl_linkup0,
      we => rxtp_done_pulse,
      seqno => rxtp_seqno,
      lcrcerr => rxtp_lcrcerr,
      next_seqno => next_seqno,
      req => an_req,
      antype => an_type,
      data => an_data,
      ack => an_ack
    );
  updatefcmgr0 : updatefcmgr
    port map (
      clk => clk,
      rst => rst,
      ca_nph => ca_nph,
      ca_ph => ca_ph,
      ca_ch => ca_ch,
      ca_npd => ca_npd,
      ca_pd => ca_pd,
      ca_cd => ca_cd,
      cl_nph => cl_nph,
      cl_ph => cl_ph,
      cl_ch => cl_ch,
      cl_npd => cl_npd,
      cl_pd => cl_pd,
      cl_cd => cl_cd,
      inf_cl_ph => inf_cl_ph,
      inf_cl_pd => inf_cl_pd,
      inf_cl_nph => inf_cl_nph,
      inf_cl_npd => inf_cl_npd,
      inf_cl_ch => inf_cl_ch,
      inf_cl_cd => inf_cl_cd,
      we => fc_we,
      fctype_in => fc_type_in,
      hsize_in => fc_hsize_in,
      dsize_in => fc_dsize_in,
      flush_p => ca_p_flush,
      flush_np => ca_np_flush,
      flush_c => ca_c_flush,
      req => fc_req,
      fctype_out => fc_type_out,
      hsize_out => fc_hsize_out,
      dsize_out => fc_dsize_out,
      ack => fc_ack
    );
  txdpctl0 : txdpctl
    port map (
      clk => clk,
      rst => rst,
      dptype => txdp_type,
      data => txdp_data,
      kickoff => txdp_kickoff,
      dp => txdp_dp,
      done => txdp_done_pulse
    );
  rxdpctl0 : rxdpctl
    port map (
      clk => clk,
      rst => rst,
      we => rxwedp,
      dp => rxdp,
      dptype => rxdp_type,
      dpbody => rxdp_data,
      crcerr => rxdp_dcrcerr,
      done => rxdp_done_pulse
    );
  txdpbuf0 : scfifo
    generic map (
      LPM_NUMWORDS => 2**TXDPBUF_DEPTH,
      LPM_WIDTH => 48
    )
    port map (
      clock => clk,
      sclr => rst or txdp_clear,
      wrreq => txdp_done_pulse,
      data => txdp_dp,
      rdreq => txrdreqdp,
      empty => txemptydp,
-- full => open,
      full => txdpbuffull,
      q => txdp
    );
  rxtpctl0 : rxtpctl
    port map (
      clk => clk,
      rst => not dl_linkup0,
      wein => rxwetp,
      be => rxbetp,
      tptop => rxtptop,
      tp => rxtp,
      next_seqno => next_seqno,
      topout => rxtop,
      weout => rxwe,
      tpbody => rx,
      seqno => rxtp_seqno,
      crcerr => rxtp_lcrcerr,
      done => rxtp_done_pulse
    );
  txtpctl0 : txtpctl
    port map (
      clk => clk,
      rst => rst,
      topin => txtop,
      wein => txwe,
      bein => txbe,
      tpbody => tx,
      tpbodysize => tpsize,
      topout => txtoptp,
      weout => txwetp,
      beout => txbetp,
      tp => txtp
    );
end default;
--
-- Ack/Nak Manager
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity acknakmgr is
  generic (
    NLANE : integer;
    MAX_PAYLOAD : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;
    we : in std_logic;
    seqno : in std_logic_vector(11 downto 0);
    lcrcerr : in std_logic;
    next_seqno : out std_logic_vector(11 downto 0);
    req : out std_logic;
    antype : out std_logic_vector(7 downto 0);
    data : out std_logic_vector(23 downto 0);
    ack : in std_logic
  );
end acknakmgr;
architecture default of acknakmgr is
  signal latency : std_logic_vector(15 downto 0);
  signal timer : std_logic_vector(15 downto 0);
  signal next_rcv_seq : std_logic_vector(11 downto 0);
  signal ack_scheduled : std_logic;
  signal nak_scheduled : std_logic;
  signal last_good_seq : std_logic_vector(11 downto 0);
begin
  data <= conv_std_logic_vector(0, 12) & last_good_seq;
  next_seqno <= next_rcv_seq;
  process (clk, rst)
    variable reqtmp : std_logic;
  begin
    if (rst = '1') then
      last_good_seq <= (others => '1'); -- no good-TLP received yet.
      next_rcv_seq <= (others => '0');
      ack_scheduled <= '0';
      nak_scheduled <= '0';
      req <= '0';
      timer <= (others => '0');
    elsif (rising_edge(clk)) then
      reqtmp := '0';
      if (we = '1') then
        if (lcrcerr = '0' and seqno = next_rcv_seq) then -- a new TLP coming in.
          -- do not schedule any Ack/Nak here.
          -- just update seq# and type.
          last_good_seq <= seqno;
          antype <= DPTYPE_ACK;
          next_rcv_seq <= unsigned(next_rcv_seq) + 1;
          ack_scheduled <= '1';
          nak_scheduled <= '0';
          reqtmp := '1'; -- transmit Ack immediately. !!!
        elsif (lcrcerr = '0' and valid_seqno(seqno, next_rcv_seq) = '1') then -- a replayed TLP coming in.
          antype <= DPTYPE_ACK;
          ack_scheduled <= '1';
          nak_scheduled <= '0';
          reqtmp := '1'; -- transmit Ack immediately.
        elsif (nak_scheduled = '0') then -- an invalid TLP coming in.
          -- schedule a Nak.
          -- only one Nak is scheduled for a sequence of bad-TLPs.
          nak_scheduled <= '1';
          antype <= DPTYPE_NAK;
          reqtmp := '1';
        end if;
      end if;
      -- flush an Ack due to latency-timer expiration.
      if (timer > latency) then
        reqtmp := '1';
      elsif (ack_scheduled = '1') then
        timer <= unsigned(timer) + 1;
      end if;
      if (reqtmp = '1') then
        req <= '1';
      elsif (ack = '1') then -- scheduled request is acknowledged by TxDPctl and transmitted.
        req <= '0';
        ack_scheduled <= '0';
        nak_scheduled <= '0';
        timer <= (others => '0');
      end if;
      --
      -- Ack/Nak latency timer load value (in 125MHz UI, i.e., 8 ns unit):
      --
      case NLANE is
        when 1 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(237/2 - (20+128)/1/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(416/2 - (20+256)/1/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when 4 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(73/2 - (20+128)/4/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(118/2 - (20+256)/4/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when 8 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(67/2 - (20+128)/8/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(107/2 - (20+256)/8/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when others =>
          latency <= conv_std_logic_vector(100/2, latency'length);
      end case;
    end if;
  end process;
end default;
--
-- UpdateFC Manager
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity updatefcmgr is
  generic (
    NLANE : integer;
    MAX_PAYLOAD : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;
    -- CA notified by transaction layer
    ca_ph : in std_logic_vector( 7 downto 0);
    ca_nph : in std_logic_vector( 7 downto 0);
    ca_ch : in std_logic_vector( 7 downto 0);
    ca_pd : in std_logic_vector(11 downto 0);
    ca_npd : in std_logic_vector(11 downto 0);
    ca_cd : in std_logic_vector(11 downto 0);
    -- notify CL to transaction layer
    cl_ph : out std_logic_vector( 7 downto 0);
    cl_nph : out std_logic_vector( 7 downto 0);
    cl_ch : out std_logic_vector( 7 downto 0);
    cl_pd : out std_logic_vector(11 downto 0);
    cl_npd : out std_logic_vector(11 downto 0);
    cl_cd : out std_logic_vector(11 downto 0);
    inf_cl_ph : out std_logic;
    inf_cl_pd : out std_logic;
    inf_cl_nph : out std_logic;
    inf_cl_npd : out std_logic;
    inf_cl_ch : out std_logic;
    inf_cl_cd : out std_logic;
    -- Rx InitFC/UpdateFC DLLP
    we : in std_logic;
    fctype_in : in std_logic_vector(7 downto 0);
    hsize_in : in std_logic_vector(7 downto 0);
    dsize_in : in std_logic_vector(11 downto 0);
    -- Tx UpdateFC DLLP
    flush_p : in std_logic;
    flush_np : in std_logic;
    flush_c : in std_logic;
    req : out std_logic;
    fctype_out : out std_logic_vector(7 downto 0);
    hsize_out : out std_logic_vector(7 downto 0);
    dsize_out : out std_logic_vector(11 downto 0);
    ack : in std_logic
  );
end updatefcmgr;
architecture default of updatefcmgr is
  signal latency : std_logic_vector(15 downto 0);
  signal timer : std_logic_vector(15 downto 0);
  signal expired : std_logic;
 constant TIME30us : std_logic_vector(timer'left downto 0) := x"0ea6";
-- constant TIME30us : std_logic_vector(timer'left downto 0) := x"0da6";  -- !! 30us - 4kB/16 clk margin.
-- constant TIME120us : std_logic_vector(timer'left downto 0) := x"3a98";
  type FCOUT_STATE is (
    IDLE,
    REQ_NP,
    REQ_P,
    REQ_C
  );
-- attribute enum_encoding: string;
-- attribute enum_encoding of FCOUT_STATE : type is "00 01 10 11";
  signal fcoutsm : FCOUT_STATE;
  signal flush_p0 : std_logic;
  signal flush_np0 : std_logic;
  signal flush_c0 : std_logic;
begin
  --
  -- Reception of Credit Limit:
  -- 1) take in 'hsize_in' and 'dsize_in' with 'we'.
  -- 2) update CL of type specified by 'fctype_in'.
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      inf_cl_ph <= '0';
      inf_cl_pd <= '0';
      inf_cl_nph <= '0';
      inf_cl_npd <= '0';
      inf_cl_ch <= '0';
      inf_cl_cd <= '0';
    elsif (rising_edge(clk)) then
      if (we = '1') then
        case fctype_in is
          when DPTYPE_INITFC1_P_VC0 =>
            cl_ph <= hsize_in;
            cl_pd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_ph <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_pd <= '1';
            end if;
          when DPTYPE_INITFC1_NP_VC0 =>
            cl_nph <= hsize_in;
            cl_npd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_nph <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_npd <= '1';
            end if;
          when DPTYPE_INITFC1_C_VC0 =>
            cl_ch <= hsize_in;
            cl_cd <= dsize_in;
            if (hsize_in = conv_std_logic_vector(0, hsize_in'length)) then
              inf_cl_ch <= '1';
            end if;
            if (dsize_in = conv_std_logic_vector(0, dsize_in'length)) then
              inf_cl_cd <= '1';
            end if;
          when DPTYPE_UPDATEFC_P_VC0 =>
            cl_ph <= hsize_in;
            cl_pd <= dsize_in;
          when DPTYPE_UPDATEFC_NP_VC0 =>
            cl_nph <= hsize_in;
            cl_npd <= dsize_in;
          when DPTYPE_UPDATEFC_C_VC0 =>
            cl_ch <= hsize_in;
            cl_cd <= dsize_in;
          when others =>
        end case;
      end if;
      --
      -- UpdateFC latency timer load value (in 125MHz UI, i.e., 8 ns unit):
      --
      case NLANE is
        when 1 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(237/2 - (20+128)/1/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(416/2 - (20+256)/1/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when 4 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(73/2 - (20+128)/4/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(118/2 - (20+256)/4/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when 8 =>
          case MAX_PAYLOAD is
            when 128 =>
              latency <= conv_std_logic_vector(67/2 - (20+128)/8/2, latency'length);
            when 256 =>
              latency <= conv_std_logic_vector(107/2 - (20+256)/8/2, latency'length);
            when others =>
              latency <= conv_std_logic_vector(100/2, latency'length);
          end case;
        when others =>
          latency <= conv_std_logic_vector(100/2, latency'length);
      end case;
    end if;
  end process;
  --
  -- Transmission of Credit Allocated:
  -- 1) assert 'req' pulse with 'fctype_out' and 'h/dsize_out',
  -- when UpdateFC should be transmitted,
  -- i.e., 'timer' expired or 'flush' asserted.
  -- 2) deassert 'req' and reset UpdateFC timer
  -- when 'ack' detect a pulse.
  --
  process (clk, rst)
  begin
    if (rst = '1') then
      timer <= (others => '0');
      expired <= '0';
      flush_p0 <= '0';
      flush_np0 <= '0';
      flush_c0 <= '0';
      req <= '0';
      fcoutsm <= IDLE;
    elsif (rising_edge(clk)) then
      -- save a flush request.
      if (flush_p = '1') then
        flush_p0 <= '1';
      end if;
      if (flush_np = '1') then
        flush_np0 <= '1';
      end if;
      if (flush_c = '1') then
        flush_c0 <= '1';
      end if;
      case fcoutsm is
        when IDLE =>
          if (flush_p0 = '1') then
            flush_p0 <= '0'; -- update due to flush request.
            req <= '1'; -- send UpdateFC NP only.
            fctype_out <= DPTYPE_UPDATEFC_P_VC0;
            hsize_out <= ca_ph;
            dsize_out <= ca_pd;
            fcoutsm <= REQ_P;
          elsif (flush_np0 = '1') then
            flush_np0 <= '0';
            req <= '1';
            fctype_out <= DPTYPE_UPDATEFC_NP_VC0;
            hsize_out <= ca_nph;
            dsize_out <= ca_npd;
            fcoutsm <= REQ_NP;
          elsif (flush_c0 = '1') then
            flush_c0 <= '0';
            req <= '1';
            fctype_out <= DPTYPE_UPDATEFC_C_VC0;
            hsize_out <= ca_ch;
            dsize_out <= ca_cd;
            fcoutsm <= REQ_C;
-- elsif (timer > latency) then
          elsif (timer > TIME30us) then -- !!
            expired <= '1'; -- update due to timer expiration.
            req <= '1'; -- send all three UpdateFCs (P/NP/C).
            fctype_out <= DPTYPE_UPDATEFC_NP_VC0;
            hsize_out <= ca_nph;
            dsize_out <= ca_npd;
            fcoutsm <= REQ_NP;
          else
            timer <= unsigned(timer) + 1;
          end if;
        when REQ_NP =>
          if (ack = '1') then
            if (expired = '1') then
              req <= '1';
              fctype_out <= DPTYPE_UPDATEFC_P_VC0;
              hsize_out <= ca_ph;
              dsize_out <= ca_pd;
              fcoutsm <= REQ_P;
            else
              req <= '0';
              fcoutsm <= IDLE;
            end if;
          end if;
        when REQ_P =>
          if (ack = '1') then
            if (expired = '1') then
              req <= '1';
              fctype_out <= DPTYPE_UPDATEFC_C_VC0;
              hsize_out <= ca_ch;
              dsize_out <= ca_cd;
              fcoutsm <= REQ_C;
            else
              req <= '0';
              fcoutsm <= IDLE;
            end if;
          end if;
        when REQ_C =>
          if (ack = '1') then
            req <= '0';
            timer <= (others => '0');
            expired <= '0';
            fcoutsm <= IDLE;
          end if;
        when others =>
          fcoutsm <= IDLE;
      end case;
    end if; -- rising(clk)
  end process;
end default;
--
-- DLLP transmitter
-- 1) set 'dptype' and 'data' when 'kickoff'.
-- 2) calc CRC of 'dptype' & 'data'.
-- 3) output DLLP to 'dp' when 'done'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity txdpctl is
  port (
    clk : in std_logic;
    rst : in std_logic;
    dptype : in std_logic_vector(7 downto 0);
    data : in std_logic_vector(23 downto 0);
    kickoff : in std_logic;
    dp : out std_logic_vector(47 downto 0);
    done : out std_logic
  );
end txdpctl;
architecture default of txdpctl is
  component calc_crc
    generic (
      DIVISOR : integer := 16#100B#;
      DATA_WIDTH : integer := 32;
      CRC_WIDTH : integer := 16
    );
    port (
      clr : in std_logic;
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  constant TXDP_WAIT : integer := 0;
  constant CRCDELAY : integer := 2;
  signal dpbody : std_logic_vector(31 downto 0);
  signal crcout : std_logic_vector(15 downto 0);
  type arrayNWAITxDPW is array (TXDP_WAIT+CRCDELAY downto 0) of std_logic_vector(dpbody'left downto 0);
  type arrayNWAITxCRCW is array (TXDP_WAIT+CRCDELAY downto 0) of std_logic_vector(crcout'left downto 0);
  signal pkickoff : std_logic_vector(TXDP_WAIT+CRCDELAY downto 0);
  signal pdp : arrayNWAITxDPW;
  signal pcrcout : arrayNWAITxCRCW;
begin
  -- alias for input data.
  dpbody <= data(7 downto 0) & data(15 downto 8) & data(23 downto 16) & dptype;
  pkickoff(0) <= kickoff;
  process (clk)
  begin
    if (rising_edge(clk)) then
      pkickoff(pkickoff'left downto 1) <= pkickoff(pkickoff'left-1 downto 0);
    end if;
  end process;
  done <= pkickoff(TXDP_WAIT+CRCDELAY);
  pdp(0)(31 downto 0) <= dpbody;
  process (clk)
  begin
    if (rising_edge(clk)) then
      pdp(pdp'left downto 1) <= pdp(pdp'left-1 downto 0);
    end if;
  end process;
  pcrcout(CRCDELAY) <= crcout;
  pcrcoutdelay0 : if (TXDP_WAIT > 0) generate
    process (clk)
    begin
      if (rising_edge(clk)) then
        pcrcout(pcrcout'left downto CRCDELAY+1) <= pcrcout(pcrcout'left-1 downto CRCDELAY);
      end if;
    end process;
  end generate pcrcoutdelay0;
  dp(39 downto 32) <= pcrcout(TXDP_WAIT+CRCDELAY)(15 downto 8);
  dp(47 downto 40) <= pcrcout(TXDP_WAIT+CRCDELAY)( 7 downto 0);
  dp(31 downto 0) <= pdp(TXDP_WAIT+CRCDELAY);
  txdpcrc0 : calc_crc
    port map (
      clr => '1',
      we => kickoff,
      be => (others => '1'),
      data => dpbody,
      crc => crcout,
      rst => rst,
      clk => clk
    );
end default;
--
-- DLLP receiver
-- 1) DLLP comes in. check CRC of it.
-- 2) output its 'dptype', 'dpbody' and 'crcerr' with 'done' pulse.
--
-- input to output delay : 3clk
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity rxdpctl is
  port (
    clk : in std_logic;
    rst : in std_logic;
    we : in std_logic;
    dp : in std_logic_vector(47 downto 0);
    dptype : out std_logic_vector(7 downto 0);
    dpbody : out std_logic_vector(23 downto 0);
    crcerr : out std_logic;
    done : out std_logic
  );
end rxdpctl;
architecture default of rxdpctl is
  component calc_crc
    generic (
      DIVISOR : integer := 16#100B#;
      DATA_WIDTH : integer := 32;
      CRC_WIDTH : integer := 16
    );
    port (
      clr : in std_logic;
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  signal we1 : std_logic;
  signal we2 : std_logic;
  signal dp1 : std_logic_vector(47 downto 0);
  signal dp2 : std_logic_vector(47 downto 0);
  signal crc_recvd : std_logic_vector(15 downto 0);
  signal crc_calcd : std_logic_vector(15 downto 0);
begin
  process (clk, rst)
  begin
    if (rst = '1') then
      done <= '0';
    elsif (rising_edge(clk)) then
      dp1 <= dp;
      dp2 <= dp1;
      -- dptype & dpbody
      dptype <= dp2( 7 downto 0);
      dpbody <= dp2(15 downto 8) & dp2(23 downto 16) & dp2(31 downto 24);
      -- crcerr
      crc_recvd <= dp1(39 downto 32) & dp1(47 downto 40);
      if (we2 = '1') then
        if (crc_recvd = crc_calcd) then
          crcerr <= '0';
        else
          crcerr <= '1';
        end if;
      end if;
      -- done pulse
      we1 <= we;
      we2 <= we1;
      done <= we2;
    end if;
  end process;
  rxdpcrc0 : calc_crc
    port map (
      clr => '1',
      we => we,
      be => (others => '1'),
      data => dp(31 downto 0),
      crc => crc_calcd,
      rst => rst,
      clk => clk
    );
end default;
--
-- TLP receiver
-- TLP comes in from 'tp' port with 'wein' asserted.
-- TLP contents are decoded and sent out to 'tpdata' port
-- with 'tptype', 'weout', 'crcerr'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;
entity rxtpctl is
  generic (
    NLANE : integer
  );
  port (
    clk : in std_logic;
    rst : in std_logic;
    wein : in std_logic;
    be : in std_logic_vector(NLANE*2-1 downto 0);
    tptop : in std_logic;
    tp : in std_logic_vector(NLANE*16-1 downto 0);
    next_seqno : in std_logic_vector(11 downto 0);
    topout : out std_logic;
    weout : out std_logic; -- not used for x4 or wider.
    tpbody : out std_logic_vector(NLANE*16-1 downto 0);
    seqno : out std_logic_vector(11 downto 0);
    crcerr : out std_logic;
    done : out std_logic
  );
end rxtpctl;
architecture default of rxtpctl is
  component calc_crc
    generic (
      DIVISOR : integer := 16#04C11DB7#;
      DATA_WIDTH : integer := NLANE*16;
      CRC_WIDTH : integer := 32
    );
    port (
      clr : in std_logic;
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  component calc_lcrc_walma
    generic (
      DIVISOR : integer := 16#04C11DB7#;
      DATA_WIDTH : integer := NLANE*16;
      NPIPE : integer := NLANE/2;
      CRC_WIDTH : integer := 32
    );
    port (
      top : in std_logic;
      packetsize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
begin
  x1 : if (NLANE = 1) generate
    x1block : block
      constant CRCDELAY : integer := NSTAGE+1; -- num of clks from the falling edge of wecrc to crcout.
      type arrayNDELAYxSEQNOW is array (CRCDELAY+1 downto 0) of std_logic_vector(11 downto 0);
      signal seqno1 : std_logic_vector(11 downto 0);
      signal tpend1 : std_logic;
      signal wecrc : std_logic;
      signal crcout : std_logic_vector(31 downto 0);
      signal we1, we2, we3 : std_logic;
      signal tptop1, tptop2, tptop3 : std_logic;
      signal next_rcv_seq : std_logic_vector(11 downto 0); -- !! should be removed later.
      signal tp1, tp2 : std_logic_vector(NLANE*16-1 downto 0);
    begin
      wecrc <= wein;
      topout <= tptop3;
      weout <= wein and we1 and we2 and we3;
      tpbody <= tp2;
      process (clk, rst)
        variable seqnotmp : std_logic_vector(seqno'range);
      begin
        if (rst = '1') then
          next_rcv_seq <= (others => '0'); -- !! should be modified later.
          crcerr <= '0';
        elsif (rising_edge(clk)) then
          -- delayed inputs.
          we1 <= wein;
          we2 <= we1;
          we3 <= we2;
          tptop2 <= tptop1;
          tptop3 <= tptop2;
          tp1 <= tp;
          tp2 <= tp1;
          -- detect end of packet.
          if (wein = '0' and we1 = '1') then
            tpend1 <= '1';
          else
            tpend1 <= '0';
          end if;
          -- sequence number of incoming TLP.
          if (tptop = '1') then
            seqno1 <= tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
          end if;
          -- check if sequence number of incoming TLP is the expected one.
          -- tptop should be passed to transaction layer if and only if
          -- incoming TLP is a new (i.e. non-replayed) one.
          -- !! this version skips LCRC check. should be modified later. !!
          tptop1 <= '0';
          if (tptop = '1') then
            seqnotmp := tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
            if (seqnotmp = next_rcv_seq) then -- a new TLP is coming in.
              next_rcv_seq <= unsigned(next_rcv_seq) + 1;
              tptop1 <= '1';
            end if;
          end if;
          -- check CRC. it should be all-one.
          if (tpend1 = '1') then
            if (crcout = x"ffffffff") then
              crcerr <= '0';
            else
              crcerr <= '1';
            end if;
          end if;
        end if;
      end process;
      seqno <= seqno1;
      done <= tpend1;
-- synthesis translate_off
      rxlcrc0 : calc_crc
        port map (
          clr => '0',
          we => wecrc,
          be => (others => '1'),
          data => tp,
          crc => crcout,
          rst => rst,
          clk => clk
        );
-- synthesis translate_on
      crcout <= (others => '1'); -- just for debugging
    end block x1block;
  end generate x1;
  x4 : if (NLANE >= 4) generate
    x4block : block
      constant CRCDELAY : integer := NSTAGE+1; -- num of clks from the falling edge of wecrc to crcout.
      type arrayNDELAYxSEQNOW is array (CRCDELAY+1 downto 0) of std_logic_vector(11 downto 0);
      signal pseqno : arrayNDELAYxSEQNOW;
      signal ptpend : std_logic_vector(CRCDELAY+1 downto 0);
      signal we1 : std_logic;
      signal be1 : std_logic_vector(NLANE*2-1 downto 0);
      signal tp1 : std_logic_vector(NLANE*16-1 downto 0);
      signal crctop : std_logic;
      signal wecrc : std_logic;
      signal becrc : std_logic_vector(NLANE*2-1 downto 0);
      signal crcin : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout : std_logic_vector(31 downto 0);
      signal beall1 : std_logic_vector(NLANE*2-1 downto 0);
      signal tptop1 : std_logic;
      signal next_rcv_seq : std_logic_vector(11 downto 0); -- !! should be removed later.
    begin
      beall1 <= (others => '1');
      topout <= tptop1;
      weout <= '0'; -- not used.
      process (clk, rst)
        variable seqnotmp : std_logic_vector(seqno'range);
      begin
        if (rst = '1') then
          next_rcv_seq <= (others => '0'); -- !! should be modified later.
          crcerr <= '0';
        elsif (rising_edge(clk)) then
          -- delayed inputs.
          we1 <= wein;
          be1 <= be;
          tp1 <= tp;
          -- detect end of packet.
          if (wein = '1' and be /= beall1) then
            ptpend(0) <= '1';
          else
            ptpend(0) <= '0';
          end if;
          ptpend(ptpend'left downto 1) <= ptpend(ptpend'left-1 downto 0);
          -- inputs to crc calculator.
          crctop <= tptop;
          wecrc <= wein;
          becrc <= be;
          -- sequence number of incoming TLP.
          if (tptop = '1') then
            pseqno(0) <= tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
          end if;
          pseqno(pseqno'left downto 1) <= pseqno(pseqno'left-1 downto 0);
          -- check if sequence number of incoming TLP is the expected one.
          -- tptop should be passed to transaction layer if and only if
          -- incoming TLP is a new (i.e. non-replayed) one.
          -- !! this version skips LCRC check. should be modified later. !!
          tptop1 <= '0';
          if (tptop = '1') then
            seqnotmp := tp(0*8+3 downto 0*8+0) & tp(1*8+7 downto 1*8+0);
            if (seqnotmp = next_rcv_seq) then -- a new TLP is coming in.
              next_rcv_seq <= unsigned(next_rcv_seq) + 1;
              tptop1 <= '1';
            end if;
          end if;
          -- check CRC. it should be all-one.
          if (ptpend(CRCDELAY) = '1') then
            if (crcout = x"ffffffff") then
              crcerr <= '0';
            else
              crcerr <= '1';
            end if;
          end if;
        end if;
      end process;
      seqno <= pseqno(CRCDELAY+1);
      done <= ptpend(CRCDELAY+1);
-- tpbody <= tp(1*8+7 downto 0*8+0) & tp1(7*8+7 downto 2*8+0);
      tpbody <= tp(1*8+7 downto 0*8+0) & tp1(NLANE*16-1 downto 2*8+0);
      -- pass incoming TLP into crc calculator.
      process (tp1, wein, be, we1, be1)
      begin
        if (NLANE = 4) then
          if (wein = '1' and be = x"03") then -- CRC at packet tail is negated.
            crcin(7*8+7 downto 6*8+0) <= not tp1(7*8+7 downto 6*8+0);
            crcin(5*8+7 downto 0*8+0) <= tp1(5*8+7 downto 0*8+0);
          elsif (we1 = '1' and be1 = x"03") then
            crcin(1*8+7 downto 0*8+0) <= not tp1(1*8+7 downto 0*8+0);
          elsif (we1 = '1' and be1 = x"3f") then
            crcin(5*8+7 downto 2*8+0) <= not tp1(5*8+7 downto 2*8+0);
            crcin(1*8+7 downto 0*8+0) <= tp1(1*8+7 downto 0*8+0);
          else
            crcin <= tp1;
          end if;
        end if;
        if (NLANE = 8) then
          -- pass incoming TLP to CRC calculator.
          crcin <= tp1;
          -- the last 4-byte of TLP (LCRC) is negated.
          if (wein = '1' and be = x"0003") then
            crcin(15*8+7 downto 14*8+0) <= not tp1(15*8+7 downto 14*8+0);
          elsif (we1 = '1' and be1 = x"0003") then
            crcin( 1*8+7 downto 0*8+0) <= not tp1( 1*8+7 downto 0*8+0);
          elsif (we1 = '1' and be1 = x"003f") then
            crcin( 5*8+7 downto 2*8+0) <= not tp1( 5*8+7 downto 2*8+0);
          elsif (we1 = '1' and be1 = x"03ff") then
            crcin( 9*8+7 downto 6*8+0) <= not tp1( 9*8+7 downto 6*8+0);
          elsif (we1 = '1' and be1 = x"3fff") then
            crcin(13*8+7 downto 10*8+0) <= not tp1(13*8+7 downto 10*8+0);
          end if;
        end if;
      end process;
-- synthesis translate_off
      rxlcrc0 : calc_lcrc_walma
        port map (
          top => crctop,
          packetsize => packetsize,
          we => wecrc,
          be => becrc,
          data => crcin,
          crc => crcout,
          rst => rst,
          clk => clk
        );
-- synthesis translate_on
      crcout <= (others => '1'); -- just for debugging
    end block x4block;
  end generate x4;
end default;
--
-- TLP transmitter
-- bare TLP-body comes in from 'tpbody' port with 'wein' asserted.
-- the body is wrapped with sequencial number and LCRC,
-- and then sent out to 'tp' port.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;
entity txtpctl is
  generic (
    NLANE : integer
  );
  port (
    topin : in std_logic;
    wein : in std_logic;
    bein : in std_logic_vector(NLANE*2-1 downto 0);
    tpbody : in std_logic_vector(NLANE*16-1 downto 0);
    tpbodysize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
    topout : out std_logic;
    weout : out std_logic;
    beout : out std_logic_vector(NLANE*2-1 downto 0);
    tp : out std_logic_vector(NLANE*16-1 downto 0);
    clk : in std_logic;
    rst : in std_logic
  );
end txtpctl;
architecture default of txtpctl is
  component calc_crc
    generic (
      DIVISOR : integer := 16#04C11DB7#;
      DATA_WIDTH : integer := NLANE*16;
      CRC_WIDTH : integer := 32
    );
    port (
      clr : in std_logic;
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  component calc_lcrc_walma
    generic (
      DIVISOR : integer := 16#04C11DB7#;
      DATA_WIDTH : integer := NLANE*16;
      NPIPE : integer := NLANE/2;
      CRC_WIDTH : integer := 32
    );
    port (
      top : in std_logic;
      packetsize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
      we : in std_logic;
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
      );
  end component;
  constant CRCDELAY : integer := NSTAGE+1; -- num of clks from the falling edge of wecrc to crcout.
begin
  x1 : if (NLANE = 1) generate
    x1block : block
      type TXTP_STATE is (
        IDLE,
        SEND_TLP,
        WAIT_LCRC,
        SEND_LCRC0,
        SEND_LCRC1,
        INVALID_STATE
      );
      signal txtpsm : TXTP_STATE;
      signal seqno : std_logic_vector(11 downto 0);
      signal seqno16 : std_logic_vector(15 downto 0);
      signal we : std_logic_vector(4 downto 0);
      signal wecrc : std_logic;
      signal tpbody1 : std_logic_vector(NLANE*16-1 downto 0);
      signal tpbody2 : std_logic_vector(NLANE*16-1 downto 0);
      signal crcin : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout : std_logic_vector(31 downto 0);
    begin
      process (clk, rst)
      begin
        if (rst = '1') then
          seqno <= (others => '0');
          txtpsm <= IDLE;
        elsif (rising_edge(clk)) then
          case txtpsm is
            when IDLE =>
              if (wein = '1') then
                txtpsm <= SEND_TLP;
              end if;
            when SEND_TLP =>
              if (wein = '0') then
                txtpsm <= WAIT_LCRC;
              end if;
            when WAIT_LCRC =>
              seqno <= unsigned(seqno) + 1;
              txtpsm <= SEND_LCRC0;
            when SEND_LCRC0 =>
              txtpsm <= SEND_LCRC1;
            when SEND_LCRC1 =>
              txtpsm <= IDLE;
            when others =>
              txtpsm <= INVALID_STATE;
          end case;
        end if;
      end process;
      beout <= (others => '1');
      process (clk)
      begin
        if (rising_edge(clk)) then
          topout <= topin;
        end if;
      end process;
      -- calc_crc has 2-clk delay from 'data' input to 'crc' output.
      weout <= we(1) or we(2) or we(3) or we(4);
      wecrc <= we(0) or we(1);
      we(0) <= wein;
      process (clk)
      begin
        if (rising_edge(clk)) then
          we(4 downto 1) <= we(3 downto 0);
          tpbody1 <= tpbody;
          tpbody2 <= tpbody1;
        end if;
      end process;
      -- extend seqno to 16-bit width and then swap the higher and lower bytes.
      seqno16 <= seqno(7 downto 0) &
                 conv_std_logic_vector(0, 16 - seqno'length) & seqno(11 downto 8);
      crcin <= tpbody1 when we(1) = '1' else seqno16;
      process (we, txtpsm, tpbody2, crcout, seqno, seqno16)
      begin
        if (we(2) = '1') then
          tp <= tpbody2;
        elsif (txtpsm = SEND_LCRC0) then
          tp <= crcout(23 downto 16) & crcout(31 downto 24);
        elsif (txtpsm = SEND_LCRC1) then
          tp <= crcout(7 downto 0) & crcout(15 downto 8);
        else
          tp <= seqno16;
        end if;
      end process;
      txlcrc0 : calc_crc
        port map (
          clr => '0',
          we => wecrc,
          be => (others => '1'),
          data => crcin,
          crc => crcout,
          rst => rst,
          clk => clk
        );
    end block x1block;
  end generate x1;
  x4 : if (NLANE >= 4) generate
    x4block : block
      signal seqno : std_logic_vector(11 downto 0);
      signal top : std_logic;
      signal wecrc : std_logic;
      signal becrc : std_logic_vector(NLANE*2-1 downto 0);
      signal crcin : std_logic_vector(NLANE*16-1 downto 0);
      signal crcout : std_logic_vector(31 downto 0);
      signal crcout1 : std_logic_vector(31 downto 0);
      type arrayNDELAYxDATAW is array (CRCDELAY+2 downto 0) of std_logic_vector(NLANE*16-1 downto 0);
      type arrayNDELAYxHALFNLANE is array (CRCDELAY+2 downto 0) of std_logic_vector(NLANE/2-1 downto 0);
      signal ptp : arrayNDELAYxDATAW;
      signal ptop : std_logic_vector(CRCDELAY downto 0);
      signal ptail : std_logic_vector(CRCDELAY downto 0);
      signal pwe : std_logic_vector(CRCDELAY+2 downto 0);
      signal be0 : std_logic_vector(NLANE*2-1 downto 0);
      signal be1 : std_logic_vector(NLANE*2-1 downto 0);
      signal tpbody1 : std_logic_vector(NLANE*16-1 downto 0);
      signal endat : arrayNDELAYxHALFNLANE;
      signal packetsize : std_logic_vector(OFFSET_WIDTH-1 downto 0);
    begin
      -- set initial values of pipeline registers.
      pwe(0) <= wein;
      -- shift tbody by 2 bytes in order to insert a sequence number.
      ptp(0)(NLANE*16-1 downto 2*8+0) <= tpbody(NLANE*16-1-2*8 downto 0*8+0);
      ptp(0)(1*8+7 downto 0*8+0) <= seqno(7 downto 0) & "0000" & seqno(11 downto 8) when ptop(0) = '1'
                                         else tpbody1(NLANE*16-1 downto NLANE*16-2*8);
      -- check byte-enable at packet tail and
      -- set endat(0) to 1, 2, 4, 8, if packet ends at lane 3L, 7L, 3H, 7H,
      -- for x8, for example.
      process(pwe, be0)
        variable endattmp : std_logic;
      begin
        endat(0) <= (others => '0');
        if (pwe(0) = '1') then
          endattmp := '0';
          for i in NLANE/2-1 downto 0 loop
            if (be0(i*4) = '1' and endattmp = '0') then
              endat(0)(i) <= '1';
              endattmp := '1';
            end if;
          end loop;
        end if;
      end process;
      ptop(0) <= topin; -- top of TLP input.
      topout <= ptop(CRCDELAY); -- top of TLP output.
      -- tail of TLP input.
      process (pwe, topin)
      begin
        ptail(0) <= '0';
        if (pwe(1) = '1' and (pwe(0) = '0' or topin = '1')) then
          ptail(0) <= '1';
        end if;
      end process;
      packetsize <= tpbodysize + conv_std_logic_vector(2, packetsize'length);  -- add 2 bytes for sequence number.
      crcin <= ptp(0);
      be0 <= bein;
      process(clk, rst)
      begin
        if (rst = '1') then
          seqno <= (others => '0');
        elsif (rising_edge(clk)) then
          -- delayed inputs:
          ptop(ptop'left downto 1) <= ptop(ptop'left-1 downto 0);
          ptail(ptail'left downto 1) <= ptail(ptail'left-1 downto 0);
          pwe(pwe'left downto 1) <= pwe(pwe'left-1 downto 0);
          ptp(ptp'left downto 1) <= ptp(ptp'left-1 downto 0);
          endat(endat'left downto 1) <= endat(endat'left-1 downto 0);
          -- count up sequence number at the rising edge of wein.
          if (ptop(0) = '1') then
            seqno <= unsigned(seqno) + 1;
          end if;
          crcout1 <= crcout;
          be1 <= be0;
          tpbody1 <= tpbody;
        end if;
      end process;
      -- write enable for CRC calculation.
      process(endat, pwe)
      begin
        if (NLANE = 4) then
          wecrc <= '0';
          if (pwe(0) = '1') then
            -- wecrc is the same as wein, in most cases.
            wecrc <= '1';
          end if;
          if (endat(1)(1) = '1' and ptail(0) = '1') then
            -- assert wecrc 1clk longer if TLP end at lane 3H.
            wecrc <= '1';
          end if;
        end if;
        if (NLANE = 8) then
          wecrc <= '0';
          if (pwe(0) = '1') then
            -- wecrc is the same as wein, in most cases.
            wecrc <= '1';
          end if;
          if (endat(1)(3) = '1' and ptail(0) = '1') then
            -- assert wecrc 1clk longer if TLP end at lane 7H.
            wecrc <= '1';
          end if;
        end if;
      end process;
      -- byte enable for CRC calculation.
      process(pwe, be0, be1)
      begin
        if (NLANE = 4) then
          becrc <= (others => '1');
          if (ptail(0) = '1' and be1(4) = '1') then
            becrc <= x"03";
          elsif (pwe(0) = '1' and be0(4) = '0') then
            becrc <= x"3f";
          end if;
        end if;
        if (NLANE = 8) then
          becrc <= (others => '1');
          if (ptail(0) = '1' and be1(12) = '1') then
            becrc <= x"0003";
          elsif (pwe(0) = '1') then
            if (be0(4) = '0') then
              becrc <= x"003f";
            elsif (be0(8) = '0') then
              becrc <= x"03ff";
            elsif (be0(12) = '0') then
              becrc <= x"3fff";
            end if;
          end if;
        end if;
      end process;
      -- write enable for final output.
      process(clk)
      begin
        if (rising_edge(clk)) then
          if (NLANE = 4) then
            if (pwe(CRCDELAY-1) = '1') then
              weout <= '1';
            elsif (pwe(CRCDELAY) = '1') then
              weout <= '1';
            else
              weout <= '0';
            end if;
          end if;
          if (NLANE = 8) then
            -- weout is the same as wein delayed by CRCDELAY-clk, in most cases.
            weout <= pwe(CRCDELAY-1);
            -- assert weout 1clk longer if TLP end at lane 3H or 7H.
            if (pwe(CRCDELAY) = '1') then
              if (endat(CRCDELAY)(3) = '1' or endat(CRCDELAY)(2) = '1') then
                weout <= '1';
              end if;
            else
            end if;
          end if;
        end if;
      end process;
      -- byte enable for final output.
      process(clk)
      begin
        if (rising_edge(clk)) then
          if (NLANE = 4) then
            beout <= (others => '1');
            if (ptail(CRCDELAY-1) = '1') then
              if (endat(CRCDELAY)(1) = '1') then
                beout <= x"3f";
              else
                beout <= x"03";
              end if;
            end if;
          end if;
          if (NLANE = 8) then
            beout <= (others => '1');
            if (endat(CRCDELAY)(3) = '1' and ptail(CRCDELAY-1) = '1') then
              beout <= x"003f";
            end if;
            if (endat(CRCDELAY)(2) = '1' and ptail(CRCDELAY-1) = '1') then
              beout <= x"0003";
            end if;
            if (endat(CRCDELAY-1)(1) = '1' and ptail(CRCDELAY-2) = '1') then
              beout <= x"3fff";
            end if;
            if (endat(CRCDELAY-1)(0) = '1' and ptail(CRCDELAY-2) = '1') then
              beout <= x"03ff";
            end if;
          end if;
        end if;
      end process;
      -- final data output.
      process(endat, pwe, ptp, crcout, crcout1)
      begin
        tp <= ptp(CRCDELAY); -- default output other than packet tail.
        if (NLANE = 4) then
          if (endat(CRCDELAY+1)(1) = '1' and ptail(CRCDELAY) = '1') then
            tp(5*8+7 downto 2*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
          if (endat(CRCDELAY+1)(0) = '1' and ptail(CRCDELAY) = '1') then
            tp(1*8+7 downto 0*8+0) <= crcout1(0*8+7 downto 0*8+0) & crcout1(1*8+7 downto 1*8+0);
          end if;
          if (endat(CRCDELAY)(0) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(7*8+7 downto 6*8+0) <= crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
        end if;
        if (NLANE = 8) then
          if (endat(CRCDELAY+1)(3) = '1' and ptail(CRCDELAY) = '1') then
            tp(5*8+7 downto 2*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
          if (endat(CRCDELAY+1)(2) = '1' and ptail(CRCDELAY) = '1') then
            tp(1*8+7 downto 0*8+0) <= crcout1(0*8+7 downto 0*8+0) & crcout1(1*8+7 downto 1*8+0);
          end if;
          if (endat(CRCDELAY)(2) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(15*8+7 downto 14*8+0) <= crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
          if (endat(CRCDELAY)(1) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(13*8+7 downto 10*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
          if (endat(CRCDELAY)(0) = '1' and ptail(CRCDELAY-1) = '1') then
            tp(9*8+7 downto 6*8+0) <= crcout(0*8+7 downto 0*8+0) & crcout(1*8+7 downto 1*8+0) &
                                      crcout(2*8+7 downto 2*8+0) & crcout(3*8+7 downto 3*8+0);
          end if;
        end if;
      end process;
      txlcrc0 : calc_lcrc_walma
        port map (
          top => ptop(0),
          packetsize => packetsize,
          we => wecrc,
          be => becrc,
          data => crcin,
          crc => crcout,
          rst => rst,
          clk => clk
        );
    end block x4block;
  end generate x4;
end default;
--
-- calculate CRC_WIDTH-bit CRC of input data stream of DATA_WIDTH-bit width.
-- input data to output crc delay : 2clk
-- output remains unchanged while we is deasserted.
--
-- this version calculates CRC using the straight forward, simplest algorithm.
-- it is handy to calculate 16-bit CRC for DLLP,
-- but is not fast enough to calculate 32-bit LCRC for lanes wider than x2.
-- for x4 (i.e. 64-bit input per clock) and wider lanes, use calc_crc_walma.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity calc_crc is
  generic (
    DIVISOR : integer;
    DATA_WIDTH : integer;
    CRC_WIDTH : integer
  );
  port (
    clr : in std_logic;
    we : in std_logic;
    be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
    data : in std_logic_vector(DATA_WIDTH-1 downto 0);
    crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
    rst : in std_logic; -- active high reset
    clk : in std_logic
  );
end calc_crc;
architecture default of calc_crc is
  signal we1 : std_logic;
  signal crc_reg : std_logic_vector(CRC_WIDTH-1 downto 0);
  signal ncrc1 : std_logic_vector(CRC_WIDTH-1 downto 0);
  signal crc2 : std_logic_vector(CRC_WIDTH-1 downto 0);
begin
  process (clk)
    variable crc_var : std_logic_vector(CRC_WIDTH-1 downto 0);
  begin
    if (rising_edge(clk)) then
      we1 <= we;
      if (we = '1') then
        if (we1 = '0' or clr = '1') then -- init crc_reg at the rising edge of we.
          crc_var := (others => '1');
        else
          crc_var := crc_reg;
        end if;
        -- shift CRC 1-bit to the left.
        eachbit : for i in 0 to DATA_WIDTH-1 loop
          -- CRC is XORed with the divisor if its MSB is 1.
          if (be(i/8) = '1') then
            if ((data(i) xor crc_var(CRC_WIDTH-1)) = '1') then
              crc_var := (crc_var(CRC_WIDTH-2 downto 0) & '0') xor
                         conv_std_logic_vector(DIVISOR, CRC_WIDTH);
            else
              crc_var := (crc_var(CRC_WIDTH-2 downto 0) & '0');
            end if;
          end if;
        end loop;
        crc_reg <= crc_var;
      end if; -- we
    end if; -- clk
  end process;
  -- negate.
  ncrc1 <= not crc_reg;
  -- reverse bit order of each byte.
  process (clk)
  begin
    if (rising_edge(clk)) then
      for j in 0 to CRC_WIDTH / 8 - 1 loop
        for i in 0 to 7 loop
          crc2(j * 8 + i) <= ncrc1(j * 8 + 7 - i);
        end loop;
      end loop;
    end if;
  end process;
  crc <= crc2;
end default;
--
-- CRC calculater, pipelined & parallelized with
-- algorithm proposed by Walma (2007).
--
-- calculate 32-bit CRC of input data stream of 64-bit width.
-- + input data to output crc delay : ??? clk
-- + support DIVISOR : 16#04C11DB7# so far.
-- + top must be asserted 1clk at the top of data stream.
-- + be is handled correctly only at the tail of data stream.
-- be='0' in the middle of data-stream causes an incorrect result.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;
entity calc_lcrc_walma is
  generic (
    DIVISOR : integer;
    DATA_WIDTH : integer;
    NPIPE : integer;
    CRC_WIDTH : integer
  );
  port (
    top : in std_logic;
    packetsize : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
    we : in std_logic;
    be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
    data : in std_logic_vector(DATA_WIDTH-1 downto 0);
    crc : out std_logic_vector(CRC_WIDTH-1 downto 0);
    rst : in std_logic; -- active high reset
    clk : in std_logic
  );
end calc_lcrc_walma;
architecture default of calc_lcrc_walma is
  component crcpipe
    generic (
      DIVISOR : integer;
      DATA_WIDTH : integer;
      CRC_WIDTH : integer
    );
    port (
      sop : in std_logic;
      data : in std_logic_vector(DATA_WIDTH-1 downto 0);
      be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      offset : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
      eop : in std_logic;
      offadj : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
      lfsr : out std_logic_vector(CRC_WIDTH-1 downto 0);
      done : out std_logic;
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  type arrayNPIPExDATAW is array (NPIPE-1 downto 0) of std_logic_vector(DATA_WIDTH/NPIPE-1 downto 0);
  type arrayNPIPExBEW is array (NPIPE-1 downto 0) of std_logic_vector(DATA_WIDTH/NPIPE/8-1 downto 0);
  -- input to each pipe.
  signal sop : std_logic;
  signal eop : std_logic;
  signal pdata : arrayNPIPExDATAW;
  signal pbe : arrayNPIPExBEW;
  signal poffset : arrayNPIPExOFFW;
  signal offadj : std_logic_vector(OFFSET_WIDTH-1 downto 0);
  -- output from each pipe.
  signal plfsr : arrayNPIPExCRCW;
  signal pdone : std_logic_vector(NPIPE-1 downto 0);
  -- LFSR reduced from all pipes.
  signal lfsr : std_logic_vector(CRC_WIDTH-1 downto 0);
  -- for internal use.
  signal we1 : std_logic;
begin
  process(clk, rst)
    variable lfsrtmp : std_logic_vector(CRC_WIDTH-1 downto 0);
    variable nbytetmp : integer;
  begin
    if (rst = '1') then
      for i in 0 to NPIPE-1 loop
        poffset(i) <= (others => '0'); -- just to remove warning message :-)
      end loop;
    elsif (rising_edge(clk)) then
      sop <= top;
      we1 <= we;
      --
      -- at start of a new data stream...
      --
      for i in 0 to NPIPE-1 loop
        -- negate the first CRC_WIDTH-bit of the input.
        if (top = '1' and DATA_WIDTH/NPIPE*i < CRC_WIDTH) then
          pdata(i) <= not data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
        else
          pdata(i) <= data((i+1)*DATA_WIDTH/NPIPE-1 downto i*DATA_WIDTH/NPIPE+0);
        end if;
        pbe(i) <= be((i+1)*DATA_WIDTH/NPIPE/8-1 downto i*DATA_WIDTH/NPIPE/8);
        -- set offset, the number of bytes from the end of max-possible TLP.
        if (top = '1') then
          if (USE_G = 1) then
            poffset(i) <= conv_std_logic_vector(MAX_TLP_SIZE-DATA_WIDTH/NPIPE/8*(i+1), OFFSET_WIDTH); -- !!
            -- eg) for x4, the 1st data to be processed by the 1st pipe (crcpipe0) has (8k-8/NPIPE)-byte offset.
          else
            if (unsigned(packetsize) > DATA_WIDTH/NPIPE/8*(i+1)) then -- care for packet tail.
              poffset(i) <= packetsize - conv_std_logic_vector(DATA_WIDTH/NPIPE/8*(i+1), OFFSET_WIDTH);
            else
              poffset(i) <= (others => '0');
            end if;
          end if;
        else
          if (unsigned(poffset(i)) > DATA_WIDTH/8) then -- care for packet tail.
            poffset(i) <= unsigned(poffset(i)) - DATA_WIDTH/8;
          else
            poffset(i) <= (others => '0');
          end if;
        end if;
      end loop;
      --
      -- while pipelines processing a data stream...
      --
      if (top = '1') then
        offadj <= conv_std_logic_vector(MAX_TLP_SIZE-DATA_WIDTH/8, OFFSET_WIDTH);
      elsif (we = '1') then
        -- count number of bytes taken in.
        nbytetmp := 0;
        for i in 0 to DATA_WIDTH/8-1 loop
          if (be(i) = '1') then
            nbytetmp := nbytetmp + 1;
          end if;
        end loop;
        -- as data-stream grows bigger, offadj (i.e., difference between
        -- data-stream size and MAX_TLP_SIZE) becomes shorter. the
        -- final value of offadj can be obtained when 'we' goes down,
        -- and it can be read out by pipelines when 'eop' is asserted.
        offadj <= unsigned(offadj) - nbytetmp;
      end if;
      --
      -- at end of a data stream...
      --
      lfsrtmp := (others => '0');
      for i in 0 to NPIPE-1 loop
        -- reduce LFSR output from each pipe.
        if (pdone(0) = '1') then
          lfsrtmp := lfsrtmp xor plfsr(i);
        end if;
      end loop;
      lfsr <= lfsrtmp;
    end if;
  end process;
  -- assert 'eop' at
  -- . the falling edge of 'we'.
  -- . 1-clk before 'top' of the next packet.
  eop <= '1' when (we = '0' and we1 = '1') or (top = '1' and we1 = '1') else '0';
    u0 : crcpipe
      generic map (
        DIVISOR => DIVISOR,
        DATA_WIDTH => DATA_WIDTH/NPIPE,
        CRC_WIDTH => CRC_WIDTH
      )
      port map (
        sop => sop,
        eop => eop,
        -- taken in every clock cycle.
        data => pdata(0),
        be => pbe(0),
        offset => poffset(0),
        -- taken in at the end of a packet.
        offadj => offadj,
        -- output becomes valid several clocks after eop pulse.
        -- the delay is fixed and thus pdone pulse is not really necessary.
        -- just for debugging and easy maintenance.
        done => pdone(0),
        lfsr => plfsr(0),
        rst => rst,
        clk => clk
      );
  pipe : for i in 1 to NPIPE-1 generate
    u0 : crcpipe
      generic map (
        DIVISOR => DIVISOR,
        DATA_WIDTH => DATA_WIDTH/NPIPE,
        CRC_WIDTH => CRC_WIDTH
      )
      port map (
        sop => sop,
        eop => eop,
        -- taken in every clock cycle.
        data => pdata(i),
        be => pbe(i),
        offset => poffset(i),
        -- taken in at the end of a packet.
        offadj => offadj,
        -- output becomes valid several clocks after eop pulse.
        -- the delay is fixed and thus pdone pulse is not really necessary.
        -- just for debugging and easy maintenance.
        done => pdone(i),
        lfsr => plfsr(i),
        rst => rst,
        clk => clk
      );
  end generate pipe;
  -- convert LFSR to CRC.
  process (lfsr)
    variable nlfsrtmp : std_logic_vector(CRC_WIDTH-1 downto 0);
  begin
    -- negate.
    nlfsrtmp := not lfsr;
    -- reverse bit order of each byte.
    for j in 0 to CRC_WIDTH / 8 - 1 loop
      for i in 0 to 7 loop
        crc(j * 8 + i) <= nlfsrtmp(j * 8 + 7 - i);
      end loop;
    end loop;
  end process;
end default;
--
-- A pipeline to calculate a CRC fragment.
--
-- + sop pulse indicates start of a new packet.
-- it clears an internal LFSR register.
--
-- + data & offset are taken in every clock cycle, and pushed
-- into the pipeline. the pipeline consists of several stages:
--
-- - prestage: calculate LFSR for data taken in.
--
-- - H stages: multiply H-matrices to LFSR for offset-times.
--
-- - accumulation stage: accumulate LFSRs.
--
-- - G stages: multiply G-matrices to the accumulated LFSR
-- for offadj-times. this part is not pipelined. each of
-- G stages is activated only once per packet.
--
-- + offadj is taken in at end of a packet with eop pulse.
-- it indicates the number of over-processed 'zero's.
-- contribution to LFSR values from these zeros are canceled
-- by multiplying G-matrix with LFSR offadj times.
--
-- + output becomes valid several clocks after eop pulse, when the
-- above mentioned cancelation has done. the delay is fixed, and
-- thus done pulse is not really necessary. it exists just for
-- debugging and easy maintenance.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
library altera_mf;
use altera_mf.all;
use work.gpciepkg.all;
use work.pcrcpkg.all;
entity crcpipe is
  generic (
    DIVISOR : integer := 16#04C11DB7#;
    DATA_WIDTH : integer;
    CRC_WIDTH : integer
  );
  port (
    sop : in std_logic;
    data : in std_logic_vector(DATA_WIDTH-1 downto 0);
    be : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
    offset : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
    eop : in std_logic;
    offadj : in std_logic_vector(OFFSET_WIDTH-1 downto 0);
    lfsr : out std_logic_vector(CRC_WIDTH-1 downto 0);
    done : out std_logic;
    rst : in std_logic; -- active high reset
    clk : in std_logic
  );
end crcpipe;
architecture default of crcpipe is
  component lpm_rom
    generic (
      LPM_WIDTH : natural;
      LPM_WIDTHAD : natural;
      LPM_NUMWORDS : natural;
      LPM_FILE : string
    );
    port (
      address : in std_logic_vector(LPM_WIDTHAD-1 downto 0);
      q : out std_logic_vector(LPM_WIDTH-1 downto 0)
    );
  end component;
  -- pipeline registers.
  signal psop : std_logic_vector(FINAL_STAGE downto 0);
  signal peop : std_logic_vector(FINAL_STAGE downto 0);
  signal plfsr : arrayNSTGxCRCW;
  signal paccum : arrayNSTGxCRCW;
  signal poffset : arrayNSTGxOFFW;
  signal poffadj : arrayNSTGxOFFW;
begin
  process(clk, rst)
    variable plfsrtmp : std_logic_vector(CRC_WIDTH-1 downto 0);
    variable paccumtmp : std_logic_vector(CRC_WIDTH-1 downto 0);
  begin
    if (rst = '1') then
    elsif (rising_edge(clk)) then
      --
      -- input to pipeline registers
      --
      psop(0) <= sop;
      psop(FINAL_STAGE downto 1) <= psop(FINAL_STAGE-1 downto 0);
      peop(0) <= eop;
      peop(FINAL_STAGE downto 1) <= peop(FINAL_STAGE-1 downto 0);
      poffset(0) <= offset;
      poffset(FINAL_STAGE downto 1) <= poffset(FINAL_STAGE-1 downto 0);
      if (eop = '1') then
        poffadj(0) <= offadj;
      end if;
      poffadj(FINAL_STAGE downto 1) <= poffadj(FINAL_STAGE-1 downto 0);
      --
      -- prestage:
      -- calculate LFSR for a single-byte input.
      --
-- plfsr(0) <= calc_lfsr_serial(data, DIVISOR, CRC_WIDTH);
      plfsr(0) <= calc_lfsr_wbe_serial(data, be, DIVISOR, CRC_WIDTH);
      --
      -- H stages:
      -- multiply H-matrices to LFSR for offset-times.
      --
      if (USE_H11 = 0) then
        case MULT_PER_STAGE is
          when 1 =>
            for i in 0 to ACCUM_STAGE-2 loop
              if (poffset(i)(i) = '1') then
                plfsr(i+1) <= mult_m_v(H_MATRIX(i), plfsr(i));
              else
                plfsr(i+1) <= plfsr(i);
              end if;
            end loop;
          when others =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp := plfsr(i);
              for k in 0 to MULT_PER_STAGE-1 loop
                if (poffset(i)(i*MULT_PER_STAGE+k) = '1') then
                  plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE+k), plfsrtmp);
                end if;
              end loop;
              plfsr(i+1) <= plfsrtmp;
            end loop;
        end case;
      else -- USE_H11
        case MULT_PER_STAGE is
          when 1 =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp := plfsr(i);
              case poffset(i)(i*2+1 downto i*2+0) is
                when "00" =>
                  plfsrtmp := plfsrtmp;
                when "01" =>
                  plfsrtmp := mult_m_v(H_MATRIX(i*2+0), plfsrtmp);
                when "10" =>
                  plfsrtmp := mult_m_v(H_MATRIX(i*2+1), plfsrtmp);
                when "11" =>
                  plfsrtmp := mult_m_v(H11_MATRIX(i*2+0), plfsrtmp);
                when others =>
              end case;
              plfsr(i+1) <= plfsrtmp;
            end loop;
          when others =>
            for i in 0 to ACCUM_STAGE-2 loop
              plfsrtmp := plfsr(i);
              for k in 0 to MULT_PER_STAGE-1 loop
                case poffset(i)(i*MULT_PER_STAGE*2+k*2+1 downto i*MULT_PER_STAGE*2+k*2+0) is
                  when "00" =>
                    plfsrtmp := plfsrtmp;
                  when "01" =>
                    plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE*2+k*2+0), plfsrtmp);
                  when "10" =>
                    plfsrtmp := mult_m_v(H_MATRIX(i*MULT_PER_STAGE*2+k*2+1), plfsrtmp);
                  when "11" =>
                    plfsrtmp := mult_m_v(H11_MATRIX(i*MULT_PER_STAGE*2+k*2+0), plfsrtmp);
                  when others =>
                end case;
              end loop;
              plfsr(i+1) <= plfsrtmp;
            end loop;
        end case;
      end if;
      --
      -- accumulation stage
      --
      if (psop(ACCUM_STAGE-1) = '1') then
        -- paccum is set to plfsr at start of a packet.
        if (poffset(ACCUM_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(ACCUM_STAGE) <= mult_m_v(H_MATRIX(OFFSET_WIDTH-1), plfsr(ACCUM_STAGE-1));
        else
          paccum(ACCUM_STAGE) <= plfsr(ACCUM_STAGE-1);
        end if;
      else
        -- plfsr is accumulated to paccum.
        if (poffset(ACCUM_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(ACCUM_STAGE) <= paccum(ACCUM_STAGE) xor mult_m_v(H_MATRIX(OFFSET_WIDTH-1), plfsr(ACCUM_STAGE-1));
        else
          paccum(ACCUM_STAGE) <= paccum(ACCUM_STAGE) xor plfsr(ACCUM_STAGE-1);
        end if;
      end if;
      --
      -- G stages:
      -- multiply G-matrices to the accumulated LFSR for offadj-times.
      --
      if (USE_G = 1) then
        if (USE_H11 = 0) then
          case MULT_PER_STAGE is
            when 1 =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                if (poffadj(ACCUM_STAGE+i)(i) = '1') then
                  paccum(ACCUM_STAGE+i+1) <= mult_m_v(G_MATRIX(i), paccum(ACCUM_STAGE+i));
                else
                  paccum(ACCUM_STAGE+i+1) <= paccum(ACCUM_STAGE+i);
                end if;
              end loop;
            when others =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                for k in 0 to MULT_PER_STAGE-1 loop
                  if (poffadj(ACCUM_STAGE+i)(i*MULT_PER_STAGE+k) = '1') then
                    paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE+k), paccumtmp);
                  end if;
                end loop;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;
          end case;
        else
          case MULT_PER_STAGE is
            when 1 =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                case poffadj(ACCUM_STAGE+i)(i*2+1 downto i*2+0) is
                  when "00" =>
                    paccumtmp := paccumtmp;
                  when "01" =>
                    paccumtmp := mult_m_v(G_MATRIX(i*2+0), paccumtmp);
                  when "10" =>
                    paccumtmp := mult_m_v(G_MATRIX(i*2+1), paccumtmp);
                  when "11" =>
                    paccumtmp := mult_m_v(G11_MATRIX(i*2+0), paccumtmp);
                  when others =>
                end case;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;
            when others =>
              for i in 0 to FINAL_STAGE-ACCUM_STAGE-2 loop
                paccumtmp := paccum(ACCUM_STAGE+i);
                for k in 0 to MULT_PER_STAGE-1 loop
                  case poffadj(ACCUM_STAGE+i)(i*MULT_PER_STAGE*2+k*2+1 downto i*MULT_PER_STAGE*2+k*2+0) is
                    when "00" =>
                      paccumtmp := paccumtmp;
                    when "01" =>
                      paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE*2+k*2+0), paccumtmp);
                    when "10" =>
                      paccumtmp := mult_m_v(G_MATRIX(i*MULT_PER_STAGE*2+k*2+1), paccumtmp);
                    when "11" =>
                      paccumtmp := mult_m_v(G11_MATRIX(i*MULT_PER_STAGE*2+k*2+0), paccumtmp);
                    when others =>
                  end case;
                end loop;
                paccum(ACCUM_STAGE+i+1) <= paccumtmp;
              end loop;
          end case;
        end if; -- USE_H11
        if (poffadj(FINAL_STAGE-1)(OFFSET_WIDTH-1) = '1') then
          paccum(FINAL_STAGE) <= mult_m_v(G_MATRIX(OFFSET_WIDTH-1), paccum(FINAL_STAGE-1));
        else
          paccum(FINAL_STAGE) <= paccum(FINAL_STAGE-1);
        end if;
      end if; -- USE_G
    end if; -- clk
  end process;
  done <= peop(FINAL_STAGE);
  lfsr <= paccum(FINAL_STAGE);
end default;
--
-- Media Access Control sub layer
-- in Physical layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity phymac is
  generic (
    NLANE : integer;
    N_FTS : integer;
    DATARATE : std_logic_vector(7 downto 0);
    TSCTRL : std_logic_vector(7 downto 0);
    DEFRAMER_TLPBUF_DEPTH : integer;
    DEFRAMER_DLLPBUF_DEPTH : integer;
    SIMULATION : integer
  );
  port (
    -- for debugging purpose.
    st2_txd : out each16b(NLANE-1 downto 0);
    st2_txk : out each2b(NLANE-1 downto 0);
    st2_rxd : out each16b(NLANE-1 downto 0);
    st2_rxk : out each2b(NLANE-1 downto 0);
    clk : in std_logic;
    rst : in std_logic; -- active high
    -- PIPE interface
    txelecidleL : out each1b(NLANE-1 downto 0);
    txdatakL : out each2b(NLANE-1 downto 0);
    txdataL : out each16b(NLANE-1 downto 0);
    rxelecidleL : in each1b(NLANE-1 downto 0);
    rxdatakL : in each2b(NLANE-1 downto 0);
    rxdataL : in each16b(NLANE-1 downto 0);
    rxpolarityL : out each1b(NLANE-1 downto 0);
    -- DLL-MAC interface
    rxwedp : out std_logic;
    rxdp : out std_logic_vector(47 downto 0);
    rxwetp : out std_logic;
    rxbetp : out std_logic_vector(NLANE*2-1 downto 0);
    rxtptop : out std_logic;
    rxtp : out std_logic_vector(NLANE*16-1 downto 0);
    txdp : in std_logic_vector(47 downto 0);
    txemptydp : in std_logic;
    txrdreqdp : out std_logic;
    txtp : in std_logic_vector(NLANE*16-1 downto 0);
    txtoptp : in std_logic;
    txwetp : in std_logic;
    txbetp : in std_logic_vector(NLANE*2-1 downto 0);
    -- TL-PHYMAC interface
    txreq : in std_logic;
    txack : out std_logic;
    dmltssm : out LTSSM_STATE;
    dmlinknumL : out each8b(NLANE-1 downto 0);
    dmlanenumL : out each8b(NLANE-1 downto 0);
    dmdatarateL : out each8b(NLANE-1 downto 0);
    phy_linkup : out std_logic;
    dl_linkup : in std_logic
  );
end phymac;
architecture default of phymac is
  component txts
    port (
      clk : in std_logic;
      rst : in std_logic;
      npacket : in std_logic_vector(11 downto 0);
      link : in std_logic_vector(7 downto 0);
      linkpad : in std_logic;
      lane : in std_logic_vector(7 downto 0);
      lanepad : in std_logic;
      nfts : in std_logic_vector(7 downto 0);
      datarate : in std_logic_vector(7 downto 0);
      ctrl : in std_logic_vector(7 downto 0);
      id : in std_logic_vector(7 downto 0);
      kickoff : in std_logic;
      txdatak : out std_logic_vector(1 downto 0);
      txdata : out std_logic_vector(15 downto 0);
      doskp : in std_logic;
      sentskp : out std_logic;
      done : out std_logic
    );
  end component;
  component rxts
    port (
      clk : in std_logic;
      rst : in std_logic;
      npacket : out std_logic_vector(11 downto 0);
      link : out std_logic_vector(7 downto 0);
      linkpad : out std_logic;
      lane : out std_logic_vector(7 downto 0);
      lanepad : out std_logic;
      nfts : out std_logic_vector(7 downto 0);
      datarate : out std_logic_vector(7 downto 0);
      ctrl : out std_logic_vector(7 downto 0);
      id : out std_logic_vector(7 downto 0);
      clear : in std_logic;
      rxdatak : in std_logic_vector(1 downto 0);
      rxdata : in std_logic_vector(15 downto 0)
    );
  end component;
  component txli
    port (
      clk : in std_logic;
      rst : in std_logic;
      npacket : in std_logic_vector(11 downto 0);
      kickoff : in std_logic;
      txdatak : out std_logic_vector(1 downto 0);
      txdata : out std_logic_vector(15 downto 0);
      done : out std_logic
    );
  end component;
  component rxli
    port (
      clk : in std_logic;
      rst : in std_logic;
      npacket : out std_logic_vector(11 downto 0);
      clear : in std_logic;
      rxdatak : in std_logic_vector(1 downto 0);
      rxdata : in std_logic_vector(15 downto 0)
    );
  end component;
  component scramble
    generic (
      DATA_WIDTH : integer;
      SIMULATION : integer := SIMULATION
    );
    port (
      src : in std_logic_vector(DATA_WIDTH-1 downto 0);
      iskcode : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
      doscramble : in std_logic;
      dstk : out std_logic_vector(DATA_WIDTH/8-1 downto 0);
      dst : out std_logic_vector(DATA_WIDTH-1 downto 0);
      rst : in std_logic; -- active high reset
      clk : in std_logic
    );
  end component;
  component deskew
    generic (
      NLANE : integer := NLANE
    );
    port (
      enable : in std_logic;
      srcd : in each16b(NLANE-1 downto 0);
      srck : in each2b(NLANE-1 downto 0);
      dstd : out each16b(NLANE-1 downto 0);
      dstk : out each2b(NLANE-1 downto 0);
      rst : in std_logic; -- active high
      clk : in std_logic
      );
  end component;
  component deframer
    generic (
      DEFRAMER_TLPBUF_DEPTH : integer := DEFRAMER_TLPBUF_DEPTH;
      DEFRAMER_DLLPBUF_DEPTH : integer := DEFRAMER_DLLPBUF_DEPTH;
      NLANE : integer := NLANE
    );
    port (
-- synthesis translate_off
      -- for debugging
      dpusedwmax : out std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
      usedwmax : out std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
-- synthesis translate_on
      -- PLP input
      data : in each16b(NLANE-1 downto 0);
      datak : in each2b(NLANE-1 downto 0);
      -- DLLP output
      wedp : out std_logic;
      dp : out std_logic_vector(47 downto 0);
      -- TLP output
      wetp : out std_logic;
      tptop : out std_logic;
      tp : out std_logic_vector(NLANE*16-1 downto 0);
      betp : out std_logic_vector(NLANE*2-1 downto 0);
      rst : in std_logic; -- active high
      clk : in std_logic
    );
  end component;
  component framer
    generic (
      NLANE : integer := NLANE
    );
    port (
      -- DLLP input
      emptydp : in std_logic;
      rdreqdp : out std_logic;
      dp : in std_logic_vector(47 downto 0);
      -- TLP input
      reqtp : in std_logic;
      acktp : out std_logic;
      toptp : in std_logic;
      wetp : in std_logic;
      betp : in std_logic_vector(NLANE*2-1 downto 0);
      tp : in std_logic_vector(NLANE*16-1 downto 0);
      -- PLP output
      data : out each16b(NLANE-1 downto 0);
      datak : out each2b(NLANE-1 downto 0);
      rst : in std_logic; -- active high
      clk : in std_logic
    );
  end component;
  component count_skp
    port (
      dosend : out std_logic;
      sent : in std_logic;
      rst : in std_logic; -- active high
      clk : in std_logic
    );
  end component;
  -- time unit conversion from milli second to UI @ 125MHz (i.e. 8 ns unit)
  constant TIME1H : std_logic_vector := x"01e848"; -- for actual hardware
  constant TIME2H : std_logic_vector := x"03d090"; -- for actual hardware
  constant TIME12H : std_logic_vector := x"16e360";
  constant TIME24H : std_logic_vector := x"2dc6c0";
  constant TIME48H : std_logic_vector := x"5b8d80";
  constant TIME1S : std_logic_vector := x"000040"; -- for simulation
  constant TIME2S : std_logic_vector := x"000080";
  constant TIME12S : std_logic_vector := x"0000c0";
  constant TIME24S : std_logic_vector := x"000180";
  constant TIME48S : std_logic_vector := x"000300";
  constant TIME16UI : std_logic_vector := x"000010";
  signal time1 : std_logic_vector(23 downto 0);
  signal time2 : std_logic_vector(23 downto 0);
  signal time12 : std_logic_vector(23 downto 0);
  signal time24 : std_logic_vector(23 downto 0);
  signal time48 : std_logic_vector(23 downto 0);
  constant NTS1H : std_logic_vector := x"001";
  constant NTS2H : std_logic_vector := x"002";
  constant NTS8H : std_logic_vector := x"008";
  constant NTS16H : std_logic_vector := x"010";
  constant NTS1KH : std_logic_vector := x"400";
  constant NTS1S : std_logic_vector := x"001";
  constant NTS2S : std_logic_vector := x"002";
  constant NTS8S : std_logic_vector := x"002";
  constant NTS16S : std_logic_vector := x"002";
  constant NTS1KS : std_logic_vector := x"004";
  signal nts1 : std_logic_vector(11 downto 0);
  signal nts2 : std_logic_vector(11 downto 0);
  signal nts8 : std_logic_vector(11 downto 0);
  signal nts16 : std_logic_vector(11 downto 0);
  signal nts1k : std_logic_vector(11 downto 0);
  -- DLL-MAC interface
  signal ltssm : LTSSM_STATE;
  signal rxpolarityL0 : each1b(NLANE-1 downto 0);
  signal datarateL0 : each8b(NLANE-1 downto 0);
  signal nftsL0 : each8b(NLANE-1 downto 0);
  signal linknumL0 : each8b(NLANE-1 downto 0);
  signal lanenumL0 : each8b(NLANE-1 downto 0);
  -- interface to TxTS
  signal txtskickoffL : each1b(NLANE-1 downto 0);
  signal txtsnumL : each12b(NLANE-1 downto 0);
  signal txtslinkL : each8b(NLANE-1 downto 0);
  signal txtslinkpadL : each1b(NLANE-1 downto 0);
  signal txtslaneL : each8b(NLANE-1 downto 0);
  signal txtslanepadL : each1b(NLANE-1 downto 0);
  signal txtsnftsL : each8b(NLANE-1 downto 0);
  signal txtsdatarateL : each8b(NLANE-1 downto 0);
  signal txtsctrlL : each8b(NLANE-1 downto 0);
  signal txtsidL : each8b(NLANE-1 downto 0);
  signal txtsdatakL : each2b(NLANE-1 downto 0);
  signal txtsdataL : each16b(NLANE-1 downto 0);
  signal txtsdoneL : each1b(NLANE-1 downto 0);
  signal txtsdone_pulseL : each1b(NLANE-1 downto 0);
  -- interface to TxLI
  signal txlikickoffL : each1b(NLANE-1 downto 0);
  signal txlinumL : each12b(NLANE-1 downto 0);
  signal txlidatakL : each2b(NLANE-1 downto 0);
  signal txlidataL : each16b(NLANE-1 downto 0);
  signal txlidoneL : each1b(NLANE-1 downto 0);
  signal txlidone_pulseL : each1b(NLANE-1 downto 0);
  -- interface to RxTS
  signal rxtsclearL : each1b(NLANE-1 downto 0);
  signal rxtsnumL : each12b(NLANE-1 downto 0);
  signal rxtslinkL : each8b(NLANE-1 downto 0);
  signal rxtslinkpadL : each1b(NLANE-1 downto 0);
  signal rxtslaneL : each8b(NLANE-1 downto 0);
  signal rxtslanepadL : each1b(NLANE-1 downto 0);
  signal rxtsnftsL : each8b(NLANE-1 downto 0);
  signal rxtsdatarateL : each8b(NLANE-1 downto 0);
  signal rxtsctrlL : each8b(NLANE-1 downto 0);
  signal rxtsidL : each8b(NLANE-1 downto 0);
  signal rxtsdoneL : each1b(NLANE-1 downto 0);
  -- interface to RxLI
  signal rxliclearL : each1b(NLANE-1 downto 0);
  signal rxlinumL : each12b(NLANE-1 downto 0);
  signal rxlidoneL : each1b(NLANE-1 downto 0);
  -- internal use
  signal prev_ltssm : LTSSM_STATE;
  signal timer : std_logic_vector(23 downto 0);
  signal isos : std_logic;
  signal isli : std_logic;
  signal sentskpL : each1b(NLANE-1 downto 0);
  signal doskp : std_logic;
  signal txkL : each2b(NLANE-1 downto 0);
  signal txdL : each16b(NLANE-1 downto 0);
  signal txtoscrkL : each2b(NLANE-1 downto 0);
  signal txtoscrdL : each16b(NLANE-1 downto 0);
  signal rxkL : each2b(NLANE-1 downto 0); -- descrambled, but not deskewed yet.
  signal rxdL : each16b(NLANE-1 downto 0);
  signal rxdeskwkL : each2b(NLANE-1 downto 0); -- descrambled and deskewed.
  signal rxdeskwdL : each16b(NLANE-1 downto 0);
  signal doscrambleL : each1b(NLANE-1 downto 0);
begin
  -- for debugging purpose
  dmltssm <= ltssm;
  st2 : for i in 0 to NLANE-1 generate
    st2_txd(i) <= txtoscrdL(i);
    st2_txk(i) <= txtoscrkL(i);
    st2_rxd(i) <= rxdeskwdL(i);
    st2_rxk(i) <= rxdeskwkL(i);
  end generate st2;
  dmlinknumL <= linknumL0;
  dmlanenumL <= lanenumL0;
  dmdatarateL <= datarateL0;
  rxpolarityL <= rxpolarityL0;
  -- Link Training & Sequence State Machine
  process (clk, rst)
    variable txtmp : std_logic;
    variable rxtmp : std_logic;
  begin
    if (rst = '1') then
      phy_linkup <= '0';
      if (SIMULATION = 0) then
        ltssm <= DETECT_QUIET;
      else
        ltssm <= L0; -- skip training sequence for simulation.
      end if;
    elsif (rising_edge(clk)) then
      prev_ltssm <= ltssm;
      case ltssm is
        --
        -- DETECT Sub State
        --
        when DETECT_QUIET =>
          -- reset all registers.
          isos <= '1';
          isli <= '0';
          for i in 0 to NLANE-1 loop
            txtskickoffL(i) <= '0';
            txtsdoneL(i) <= '0';
            txlikickoffL(i) <= '0';
            txlidoneL(i) <= '0';
            rxtsdoneL(i) <= '0';
            rxtsclearL(i) <= '1';
            rxlidoneL(i) <= '0';
            rxliclearL(i) <= '1';
            rxpolarityL0(i) <= '0';
            lanenumL0(i) <= (others => '1'); -- set an invalid lane#.
          end loop;
          -- down the link.
          phy_linkup <= '0';
          for i in 0 to NLANE-1 loop
            txelecidleL(i) <= '1';
          end loop;
          -- Rx of all lanes exit electrical idle, or timed out.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxelecidleL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;
          if (rxtmp = '0' or timer = time12) then
            ltssm <= DETECT_ACTIVE;
          end if;
        when DETECT_ACTIVE =>
          -- !!! behavior of this state is not spec compliant !!!
          -- Rx detection is not performed for simplicity.
          -- just wait for 1ms and transition to Polling.Active.
          for i in 0 to NLANE-1 loop
            txelecidleL(i) <= '0';
          end loop;
          if (timer = time1) then
            ltssm <= POLLING_ACTIVE;
          end if;
        --
        -- POLLING Sub State
        --
        when POLLING_ACTIVE =>
          -- send TS1(PAD, PAD) x 1024.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts1k;
              txtslinkpadL(i) <= '1';
              txtslinkL(i) <= K_PAD;
              txtslanepadL(i) <= '1';
              txtslaneL(i) <= K_PAD;
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS1ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then
              txtsnumL(i) <= nts1k; -- send more TS just in case.
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          -- receive TS1/TS2/TS1_D/TS2_D(PAD, PAD) x 8.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS1(PAD, PAD) x 1024
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxtsdoneL(i) <= '1'; -- received TS1/2/1D/2D(PAD, PAD) x 8
              end if;
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          -- all lanes sent 1024 TSs and received 8 TSs.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= POLLING_CONFIGWAIT;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          if (timer = time24) then
            ltssm <= DETECT_QUIET;
          end if;
        when POLLING_CONFIGWAIT =>
          -- send TS2(PAD, PAD) x 16.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '1';
              txtslinkL(i) <= K_PAD;
              txtslanepadL(i) <= '1';
              txtslaneL(i) <= K_PAD;
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          -- receive TS1/2/1D/2D(PAD, PAD) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          -- invert polarity if necessary.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and
                (rxtsidL(i) = TS1ID_N or rxtsidL(i) = TS2ID_N)) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxpolarityL0(i) <= '1';
              end if;
            end if;
          end loop;
          -- save datarate & nfts reveived.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxtsdoneL(i) <= '1'; -- received TS2(PAD, PAD) x 8
                datarateL0(i) <= rxtsdatarateL(i);
                nftsL0(i) <= rxtsnftsL(i);
              end if;
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(PAD, PAD) x 8
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          -- all lanes sent TS2(PAD, PAD) x 16 and
          -- all lanes received TS2(PAD, PAD) x 8.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= POLLING_CONFIGACC;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;
        when POLLING_CONFIGACC =>
          -- send TS2(PAD, PAD) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '1';
              txtslinkL(i) <= K_PAD;
              txtslanepadL(i) <= '1';
              txtslaneL(i) <= K_PAD;
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(PAD, PAD) x 16
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          -- all lanes sent 16 TSs.
          if (txtmp = '1' and rxtmp = '1') then
            rxtmp := '1';
            for i in 0 to NLANE-1 loop
              if (datarateL0(i) /= DATARATE) then
                rxtmp := '0';
              end if;
            end loop;
            if (rxtmp = '1') then -- all peer lanes are Gen1.
              ltssm <= CONFIG_LINKSTART;
            else -- one or more peer lanes try to communicate at 5.0Gb/s or faster.
              ltssm <= POLLING_SPEED;
            end if;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;
        when POLLING_SPEED =>
          -- stay electrical idle for at least 20ns but no longer than 2ms.
          if (timer = TIME16UI) then
            ltssm <= POLLING_ACTIVE;
            for i in 0 to NLANE-1 loop
              txelecidleL(i) <= '0';
            end loop;
          else
            for i in 0 to NLANE-1 loop
              txelecidleL(i) <= '1';
            end loop;
          end if;
        when CONFIG_LINKSTART =>
          -- send TS1(PAD, PAD) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '1';
              txtslinkL(i) <= K_PAD;
              txtslanepadL(i) <= '1';
              txtslaneL(i) <= K_PAD;
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS1ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS1(PAD, PAD) x 16
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          -- receive TS1(link#, PAD) x 2
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '1') then -- received TS1(link#, PAD) x 2
                rxtsdoneL(i) <= '1';
                linknumL0(i) <= rxtslinkL(i); -- link# is determined.
              end if;
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          -- all lanes sent TS1(PAD, PAD) x 16 and
          -- all lanes received TS1(link#, PAD) x 2.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_LINKACC;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          -- received TS1(PAD, PAD) x 2, causing transition to Detect.Quiet.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= DETECT_QUIET;
              end if;
            end if;
          end loop;
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when CONFIG_LINKACC =>
          -- send TS1(link#, PAD) x 1
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16; -- one TS should be enough. but some chipsets need more.
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- share link# of lane0 among all lanes.
                                                 -- NLANE lanes are configured into one link.
              txtslanepadL(i) <= '1';
              txtslaneL(i) <= K_PAD;
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS1ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS1(link#, PAD) x 1
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          if (txtmp = '1') then
            ltssm <= CONFIG_LANEWAIT;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;
        when CONFIG_LANEWAIT =>
          -- receive TS2(link#, lane#) x 2 or TS1(link#, lane# diff from previous one) x 2
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2) then
              -- received TS2(link#, lane#) x 2
              if (rxtsidL(i) = TS2ID and rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then
                rxtsdoneL(i) <= '1';
                lanenumL0(i) <= rxtslaneL(i); -- lane# is determined.
              end if;
              -- received TS1(link#, lane# diff from previous one) x 2
              if (rxtsidL(i) = TS1ID and rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then
                if (rxtslaneL(i) /= lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then -- w/correct link# & unseen lane#.
                  rxtsdoneL(i) <= '1';
                  lanenumL0(i) <= rxtslaneL(i); -- lane# is modified.
                end if;
              end if;
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          -- all lanes received TS2(link#, lane#) x 2.
          if (rxtmp = '1') then
            ltssm <= CONFIG_LANEACC;
            for i in 0 to NLANE-1 loop
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          -- received TS1(PAD, PAD) x 2, causing transition to Detect.Quiet.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= DETECT_QUIET;
              end if;
            end if;
          end loop;
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when CONFIG_LANEACC =>
          -- send TS1(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16; -- one TS should be enough. but some chipsets need more.
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS1ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS1(link#, lane#) x 1
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          -- all lanes sent TS1(link#, lane#) x 16 and
          if (txtmp = '1') then
            ltssm <= CONFIG_COMPLETE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;
        when CONFIG_COMPLETE =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(link#, lane#) x 16
            end if;
          end loop;
          -- receive TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_COMPLETE2;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              rxtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when CONFIG_COMPLETE2 =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(link#, lane#).
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          if (txtmp = '1') then
            ltssm <= CONFIG_IDLE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when CONFIG_IDLE =>
          isos <= '0';
          isli <= '1';
          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then -- send more logical idle just in case.
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;
          -- reveive logical idle x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxliclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1'; -- sent logical idle x 16
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxliclearL(i) = '0' and rxlinumL(i) >= nts8) then
              rxlidoneL(i) <= '1'; -- received logical idle x 8
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxlidoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_IDLE2;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
              rxlidoneL(i) <= '0';
              rxliclearL(i) <= '1';
            end loop;
          end if;
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when CONFIG_IDLE2 =>
          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then -- send more logical idle just in case.
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1'; -- sent logical idle x 16
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          if (txtmp = '1') then
            ltssm <= L0;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
            end loop;
          end if;
        when L0 =>
          phy_linkup <= '1';
          isli <= '0';
          -- watch receiver for TS.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          -- any of the lanes received TS, causing transition to Recovery.
          -- LTSSM transitions to Detect.Quiet instead of Recovery state.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxtsnumL(i) >= nts1) then
              rxtmp := '1';
            end if;
          end loop;
          if (rxtmp = '1') then
            ltssm <= RECOVERY_LOCK;
          end if;
          -- any Rx of the lanes gets into electrical idle, causing transition to Detect.Quiet.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxelecidleL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;
          if (rxtmp = '1') then
            ltssm <= DETECT_QUIET;
          end if;
        when RECOVERY_LOCK =>
          isos <= '1';
          isli <= '0';
          -- send TS1(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS1ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS1(link#, lane#).
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          -- receive TS1 or TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and (rxtsidL(i) = TS2ID or rxtsidL(i) = TS1ID)) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_CONFIG;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          -- timed out.
          -- transition to Config.Linkstart if any lane received TS1 or TS2.
          -- transision to Detect.Quiet otherwise.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;
          if (timer = time24) then
            if (rxtmp = '1') then
              ltssm <= CONFIG_LINKSTART;
            else
              ltssm <= DETECT_QUIET;
            end if;
          end if;
        when RECOVERY_CONFIG =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(link#, lane#).
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          -- receive TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_CONFIG2;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i) <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          -- timed out.
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;
        when RECOVERY_CONFIG2 =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i) <= nts16;
              txtslinkpadL(i) <= '0';
              txtslinkL(i) <= linknumL0(0); -- set link# to that of lane0.
              txtslanepadL(i) <= '0';
              txtslaneL(i) <= lanenumL0(i); -- set lane# to that received from the upstream.
              txtsnftsL(i) <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i) <= TSCTRL;
              txtsidL(i) <= TS2ID;
              txtskickoffL(i) <= '1';
            elsif (txtsdone_pulseL(i) = '1') then -- send more TS just in case.
              txtsnumL(i) <= nts1k;
              txtskickoffL(i) <= '1';
            else
              txtskickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1'; -- sent TS2(link#, lane#).
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          if (txtmp = '1') then
            ltssm <= RECOVERY_IDLE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;
          -- timed out.
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;
        when RECOVERY_IDLE =>
          isos <= '0';
          isli <= '1';
          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then -- send more logical idle just in case.
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;
          -- reveive logical idle x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxliclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1'; -- sent logical idle x 16
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxliclearL(i) = '0' and rxlinumL(i) >= nts8) then
              rxlidoneL(i) <= '1'; -- received logical idle x 8
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxlidoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_IDLE2;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
              rxlidoneL(i) <= '0';
              rxliclearL(i) <= '1';
            end loop;
          end if;
          -- received TS1(PAD, PAD) x 2, causing transition to Config.Linkstart.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= CONFIG_LINKSTART;
              end if;
            end if;
          end loop;
          -- timed out.
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when RECOVERY_IDLE2 =>
          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then -- send more logical idle just in case.
              txlinumL(i) <= nts16;
              txlikickoffL(i) <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;
          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1'; -- sent logical idle x 16
            end if;
          end loop;
          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          if (txtmp = '1') then
            ltssm <= L0;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
            end loop;
          end if;
          -- timed out.
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
        when RECOVERY_SPEED =>
          ltssm <= DETECT_QUIET;
        when others =>
          ltssm <= INVALID_STATE;
      end case;
    end if;
  end process;
  -- timeout counter
  -- cleared to 0 when LTSSM transition occured.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (ltssm /= prev_ltssm) then
        timer <= (others => '0');
      else
        timer <= unsigned(timer) + 1;
      end if;
    end if;
  end process;
  -- TS1/TS2 transmitter
  txtslane : for i in 0 to NLANE-1 generate
    u0 : txts
      port map (
        clk => clk,
        rst => rst,
        npacket => txtsnumL(i),
        link => txtslinkL(i),
        linkpad => txtslinkpadL(i),
        lane => txtslaneL(i),
        lanepad => txtslanepadL(i),
        nfts => txtsnftsL(i),
        datarate => txtsdatarateL(i),
        ctrl => txtsctrlL(i),
        id => txtsidL(i),
        kickoff => txtskickoffL(i),
        txdata => txtsdataL(i),
        txdatak => txtsdatakL(i),
        doskp => doskp,
        sentskp => sentskpL(i),
        done => txtsdone_pulseL(i)
      );
  end generate txtslane;
  count_skp0 :count_skp
    port map (
      dosend => doskp,
      sent => sentskpL(0),
      rst => rst,
      clk => clk
    );
  -- TS1/TS2 receiver
  rxtslane : for i in 0 to NLANE-1 generate
    u0 : rxts
      port map (
        clk => clk,
        rst => rst,
        npacket => rxtsnumL(i),
        link => rxtslinkL(i),
        linkpad => rxtslinkpadL(i),
        lane => rxtslaneL(i),
        lanepad => rxtslanepadL(i),
        nfts => rxtsnftsL(i),
        datarate => rxtsdatarateL(i),
        ctrl => rxtsctrlL(i),
        id => rxtsidL(i),
        clear => rxtsclearL(i),
        rxdatak => rxdatakL(i),
        rxdata => rxdataL(i)
      );
  end generate rxtslane;
  -- Logical idle transmitter
  txlilane : for i in 0 to NLANE-1 generate
    u0 : txli
      port map (
        clk => clk,
        rst => rst,
        npacket => txlinumL(i),
        kickoff => txlikickoffL(i),
        txdata => txlidataL(i),
        txdatak => txlidatakL(i),
        done => txlidone_pulseL(i)
      );
  end generate txlilane;
  -- Logical idle reveiver
  rxlilane : for i in 0 to NLANE-1 generate
    u0 : rxli
      port map (
        clk => clk,
        rst => rst,
        npacket => rxlinumL(i),
        clear => rxliclearL(i),
        rxdatak => rxkL(i),
        rxdata => rxdL(i)
      );
  end generate rxlilane;
  --
  -- Tx data selectors
  --
  -- feed one of txdL, txlidataL and txtsdataL to the scrambler.
  txmux0 : for i in 0 to NLANE-1 generate
    process (clk)
    begin
      if (rising_edge(clk)) then
        if (isli = '1') then -- transmit logical idle.
          doscrambleL(i) <= '1';
          txtoscrkL(i) <= txlidatakL(i);
          txtoscrdL(i) <= txlidataL(i);
        elsif (isos = '1') then -- transmit TS ordered set.
          doscrambleL(i) <= '0';
          txtoscrkL(i) <= txtsdatakL(i);
          txtoscrdL(i) <= txtsdataL(i);
        else -- transmit TLP or DLLP.
          doscrambleL(i) <= '1';
          txtoscrkL(i) <= txkL(i);
          txtoscrdL(i) <= txdL(i);
        end if;
      end if;
    end process;
  end generate txmux0;
  -- Tx scrambler
  txscr0 : for i in 0 to NLANE-1 generate
    u0 : scramble
      generic map (
        DATA_WIDTH => 16
      )
      port map (
        clk => clk,
        rst => rst,
        src => txtoscrdL(i),
        iskcode => txtoscrkL(i),
        doscramble => doscrambleL(i),
        dstk => txdatakL(i),
        dst => txdataL(i)
      );
  end generate txscr0;
  -- Rx descrambler
  rxdescr0 : for i in 0 to NLANE-1 generate
    u0 : scramble
      generic map (
        DATA_WIDTH => 16
      )
      port map (
        clk => clk,
        rst => rst,
        src => rxdataL(i),
        doscramble => '1',
        iskcode => rxdatakL(i),
        dstk => rxkL(i),
        dst => rxdL(i)
      );
  end generate rxdescr0;
  deskew0 : deskew
    port map (
      enable => isos,
      srcd => rxdL,
      srck => rxkL,
      dstd => rxdeskwdL,
      dstk => rxdeskwkL,
      rst => rst,
      clk => clk
    );
  deframer0 : deframer
    port map (
      data => rxdeskwdL,
      datak => rxdeskwkL,
      wedp => rxwedp,
      dp => rxdp,
      wetp => rxwetp,
      betp => rxbetp,
      tptop => rxtptop,
      tp => rxtp,
      rst => rst,
      clk => clk
    );
  framer0 : framer
    port map (
      emptydp => txemptydp,
      rdreqdp => txrdreqdp,
      dp => txdp,
      reqtp => txreq,
      acktp => txack,
      toptp => txtoptp,
      wetp => txwetp,
      betp => txbetp,
      tp => txtp,
      data => txdL,
      datak => txkL,
      rst => rst,
      clk => clk
    );
  -- constants to which different values are assigned
  -- for simulation and for actual hardware.
  gennosim : if (SIMULATION = 0) generate
    time1 <= TIME1H;
    time2 <= TIME2H;
    time12 <= TIME12H;
    time24 <= TIME24H;
    time48 <= TIME48H;
    nts1 <= NTS1H;
    nts2 <= NTS2H;
    nts8 <= NTS8H;
    nts16 <= NTS16H;
    nts1k <= NTS1KH;
  end generate gennosim;
  gensim : if (SIMULATION = 1) generate
    time1 <= TIME1S;
    time2 <= TIME2S;
    time12 <= TIME12S;
    time24 <= TIME24S;
    time48 <= TIME48S;
    nts1 <= NTS1S;
    nts2 <= NTS2S;
    nts8 <= NTS8S;
    nts16 <= NTS16S;
    nts1k <= NTS1KS;
  end generate gensim;
end default;
--
-- TS1/TS2 transmitter
-- 1) set 'npacket', 'link', 'lane', 'id' when 'kickoff'.
-- 2) transmit 'npacket' TSs.
-- 3) output 1-clk pulse when 'done'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity txts is
  port (
    clk : in std_logic;
    rst : in std_logic;
    npacket : in std_logic_vector(11 downto 0);
    link : in std_logic_vector(7 downto 0);
    linkpad : in std_logic;
    lane : in std_logic_vector(7 downto 0);
    lanepad : in std_logic;
    nfts : in std_logic_vector(7 downto 0);
    datarate : in std_logic_vector(7 downto 0);
    ctrl : in std_logic_vector(7 downto 0);
    id : in std_logic_vector(7 downto 0);
    kickoff : in std_logic;
    txdatak : out std_logic_vector(1 downto 0);
    txdata : out std_logic_vector(15 downto 0);
    doskp : in std_logic;
    sentskp : out std_logic;
    done : out std_logic
  );
end txts;
architecture default of txts is
  type TS_STATE is (
    IDLE,
    SEND_COM_LINK,
    SEND_LANE_NFTS,
    SEND_DATARATE_CTRL,
    SEND_TSIDs,
    INVALID_STATE
  );
-- attribute enum_encoding: string;
-- attribute enum_encoding of TS_STATE : type is "000 001 010 011 100 101 110 111";
  signal txtssm : TS_STATE;
  signal npacket_left : std_logic_vector(11 downto 0);
  signal link0 : std_logic_vector(7 downto 0);
  signal linkpad0 : std_logic;
  signal lane0 : std_logic_vector(7 downto 0);
  signal lanepad0 : std_logic;
  signal nfts0 : std_logic_vector(7 downto 0);
  signal datarate0 : std_logic_vector(7 downto 0);
  signal ctrl0 : std_logic_vector(7 downto 0);
  signal id0 : std_logic_vector(7 downto 0);
  signal nid_left : std_logic_vector(3 downto 0);
  signal sentskp0 : std_logic;
begin
  sentskp <= sentskp0;
  process (clk, rst)
  begin
    if (rst = '1') then
      txtssm <= IDLE;
    elsif (rising_edge(clk)) then
      -- set # of TS packets to be sent,
      -- as well as LINK#, LANE#, N_FTS, and TSID.
      -- and then start TS transmission.
      if (kickoff = '1') then
        txtssm <= SEND_COM_LINK;
        npacket_left <= npacket;
        nid_left <= "1010";
        linkpad0 <= linkpad;
        link0 <= link;
        lanepad0 <= lanepad;
        lane0 <= lane;
        nfts0 <= nfts;
        datarate0 <= datarate;
        ctrl0 <= ctrl;
        id0 <= id;
        done <= '0';
        sentskp0 <= '0';
      else
        case txtssm is
          when IDLE =>
            done <= '0';
            if (sentskp0 = '1') then -- SKP os transmission is in progress.
              sentskp0 <= '0';
              txdatak <= "11";
              txdata <= K_SKP & K_SKP;
            elsif (doskp = '1') then -- issue SKP os.
              sentskp0 <= '1';
              txdatak <= "11";
              txdata <= K_SKP & K_COM;
            elsif (npacket_left /= conv_std_logic_vector(0, npacket_left'length)) then
              nid_left <= "1010";
              txdatak <= linkpad0 & '1';
              txdata <= link0 & K_COM;
              txtssm <= SEND_LANE_NFTS;
            else
              txdatak <= (others => '0');
              txdata <= (others => '0');
            end if;
          when SEND_COM_LINK =>
            txdatak <= linkpad0 & '1';
            txdata <= link0 & K_COM;
            txtssm <= SEND_LANE_NFTS;
          when SEND_LANE_NFTS =>
            txdatak <= '0' & lanepad0;
            txdata <= nfts0 & lane0;
            txtssm <= SEND_DATARATE_CTRL;
          when SEND_DATARATE_CTRL =>
            txdatak <= "00";
            txdata <= ctrl0 & datarate0;
            txtssm <= SEND_TSIDs;
          when SEND_TSIDs =>
            nid_left <= unsigned(nid_left) - 2;
            if (nid_left = conv_std_logic_vector(2, nid_left'length)) then
              npacket_left <= unsigned(npacket_left) - 1;
              txtssm <= IDLE;
              if (npacket_left = conv_std_logic_vector(1, npacket_left'length)) then
                done <= '1';
              end if;
            else
              txdatak <= "00";
              txdata <= id0 & id0;
            end if;
          when others =>
            txtssm <= IDLE;
        end case;
      end if; -- if (kickoff = '1')
    end if; -- if (rising_edge(clk))
  end process;
end default;
--
-- TS1/TS2 receiver
-- 1) reset packet counter when 'clear'ed.
-- 2) output 'npacket', 'link', 'lane', 'id' of the latest
-- TS1/TS2 received if any. 'npacket' is the num of
-- the same packets sequentially obtained.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity rxts is
  port (
    clk : in std_logic;
    rst : in std_logic;
    npacket : out std_logic_vector(11 downto 0);
    link : out std_logic_vector(7 downto 0);
    linkpad : out std_logic;
    lane : out std_logic_vector(7 downto 0);
    lanepad : out std_logic;
    nfts : out std_logic_vector(7 downto 0);
    datarate : out std_logic_vector(7 downto 0);
    ctrl : out std_logic_vector(7 downto 0);
    id : out std_logic_vector(7 downto 0);
    clear : in std_logic;
    rxdatak : in std_logic_vector(1 downto 0);
    rxdata : in std_logic_vector(15 downto 0)
  );
end rxts;
architecture default of rxts is
  type TS_STATE is (
    IDLE,
    RECV_ALIGN,
    RECV_COM_LINK,
    RECV_LANE_NFTS,
    RECV_DATARATE_CTRL,
    RECV_TSID,
    INVALID_STATE
  );
-- attribute enum_encoding: string;
-- attribute enum_encoding of TS_STATE : type is "000 001 010 011 100 101 110 111";
  signal rxtssm : TS_STATE;
  signal npacket0 : std_logic_vector(11 downto 0);
  signal nid_left : std_logic_vector(3 downto 0);
  signal rxdatak1 : std_logic_vector(1 downto 0);
  signal rxdata1 : std_logic_vector(15 downto 0);
  signal adatak : std_logic_vector(1 downto 0);
  signal adata : std_logic_vector(15 downto 0);
  signal aligned : std_logic;
  signal samets : std_logic;
  signal commit : std_logic;
  -- parameters of currently parsing TS.
  -- note that it is uncertain if the receiving data
  -- is a TS or something else (eg. SKP ordered set).
  signal link0 : std_logic_vector(7 downto 0);
  signal linkpad0 : std_logic;
  signal lane0 : std_logic_vector(7 downto 0);
  signal lanepad0 : std_logic;
  signal nfts0 : std_logic_vector(7 downto 0);
  signal datarate0 : std_logic_vector(7 downto 0);
  signal ctrl0 : std_logic_vector(7 downto 0);
  signal id0 : std_logic_vector(7 downto 0);
  -- parameters of recently received TS.
  signal link1 : std_logic_vector(7 downto 0);
  signal linkpad1 : std_logic;
  signal lane1 : std_logic_vector(7 downto 0);
  signal lanepad1 : std_logic;
  signal nfts1 : std_logic_vector(7 downto 0);
  signal datarate1 : std_logic_vector(7 downto 0);
  signal ctrl1 : std_logic_vector(7 downto 0);
  signal id1 : std_logic_vector(7 downto 0);
-- attribute preserve : boolean;
-- attribute preserve of rxdatak1 : signal is true;
-- attribute preserve of rxdata1 : signal is true;
begin
  npacket <= npacket0; -- # of same TSs received in sequence.
  -- parameters of recently received TS.
  link <= link1;
  linkpad <= linkpad1;
  lane <= lane1;
  lanepad <= lanepad1;
  nfts <= nfts1;
  datarate <= datarate1;
  ctrl <= ctrl1;
  id <= id1;
  process (clk, rst)
  begin
    if (rst = '1') then
      rxtssm <= IDLE;
    elsif (rising_edge(clk)) then
      if (clear = '1') then
        link0 <= (others => '0');
        linkpad0 <= '0';
        lane0 <= (others => '0');
        lanepad0 <= '0';
        nfts0 <= (others => '0');
        datarate0 <= (others => '0');
        ctrl0 <= (others => '0');
        id0 <= (others => '0');
        commit <= '0';
        rxtssm <= IDLE;
      else
        case rxtssm is
          when IDLE =>
            commit <= '0';
            nid_left <= "1010";
            -- detect a COM. start parsing TS ordered sets.
            if (rxdatak(0) = '1' and rxdata(7 downto 0) = K_COM) then -- TS is on 16-bit alignment.
              aligned <= '1';
              rxtssm <= RECV_ALIGN;
            elsif (rxdatak(1) = '1' and rxdata(15 downto 8) = K_COM) then -- TS is not on 16-bit alignment.
              aligned <= '0';
              rxtssm <= RECV_ALIGN;
            end if;
          when RECV_ALIGN =>
            rxtssm <= RECV_COM_LINK; -- wait for 1clk to adata becomes valid.
          when RECV_COM_LINK =>
            commit <= '0';
            if (adatak(0) = '1') then
              if (adatak(1) = '1') then
                if (adata(15 downto 8) = K_PAD) then
                  linkpad0 <= '1';
                  link0 <= adata(15 downto 8);
                  rxtssm <= RECV_LANE_NFTS;
                else
                  rxtssm <= IDLE; -- ordered set other than TS coming in.
                end if;
              else
                linkpad0 <= '0';
                link0 <= adata(15 downto 8);
                rxtssm <= RECV_LANE_NFTS;
              end if;
            else
              rxtssm <= IDLE;
            end if;
          when RECV_LANE_NFTS =>
            if (adatak(0) = '1' and adata(7 downto 0) = K_PAD) then
              lanepad0 <= '1';
            else
              lanepad0 <= '0';
            end if;
            lane0 <= adata(7 downto 0);
            nfts0 <= adata(15 downto 8);
            rxtssm <= RECV_DATARATE_CTRL;
          when RECV_DATARATE_CTRL =>
            if (adatak = "00") then
              datarate0 <= adata(7 downto 0);
              ctrl0 <= adata(15 downto 8);
              rxtssm <= RECV_TSID;
            else
              rxtssm <= IDLE;
            end if;
          when RECV_TSID =>
            if (adatak = "00" and
                (adata = TS1ID & TS1ID or
                 adata = TS2ID & TS2ID or
                 adata = TS1ID_N & TS1ID_N or
                 adata = TS2ID_N & TS2ID_N)) then
              id0 <= adata(7 downto 0);
              if (nid_left = conv_std_logic_vector(2, nid_left'length)) then
                commit <= '1';
                nid_left <= "1010";
                rxtssm <= RECV_COM_LINK;
              else
                nid_left <= unsigned(nid_left) - 2;
              end if;
            else
              rxtssm <= IDLE;
            end if;
          when others =>
            rxtssm <= IDLE;
        end case;
      end if;
    end if;
  end process;
  -- when receiving the final byte of a TS (i.e. when commit = 1),
  -- increment npacket0 if currently receiving TS and
  -- the recently received one is the same. set npacket0 to 1
  -- otherwise.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (clear = '1') then
        npacket0 <= (others => '0');
        id1 <= (others => '0'); -- so that next coming TS never match the previous one.
      elsif (commit = '1') then
        link1 <= link0;
        linkpad1 <= linkpad0;
        lane1 <= lane0;
        lanepad1 <= lanepad0;
        nfts1 <= nfts0;
        datarate1 <= datarate0;
        ctrl1 <= ctrl0;
        id1 <= id0;
        if (samets = '1') then
          npacket0 <= unsigned(npacket0) + 1;
        else
          npacket0 <= conv_std_logic_vector(1, npacket0'length);
        end if;
      end if;
    end if;
  end process;
  -- check if current TS and previously received one have
  -- the same parameters.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (link0 = link1 and linkpad0 = linkpad1 and
          lane0 = lane1 and lanepad0 = lanepad1 and
          nfts0 = nfts1 and datarate0 = datarate1 and
          ctrl0 = ctrl1 and id0 = id1) then
        samets <= '1';
      else
        samets <= '0';
      end if;
    end if;
  end process;
  -- shift TS so that it is 16-bit aligned.
  -- adata has 2-clk delay to rxdata.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (aligned = '1') then
        adatak <= rxdatak1;
        adata <= rxdata1;
      else
        adatak(1) <= rxdatak(0);
        adatak(0) <= rxdatak1(1);
        adata(15 downto 8) <= rxdata(7 downto 0);
        adata(7 downto 0) <= rxdata1(15 downto 8);
      end if;
      rxdatak1 <= rxdatak;
      rxdata1 <= rxdata;
    end if;
  end process;
end default;
--
-- Logical idle transmitter
-- 1) set 'npacket' when 'kickoff'.
-- 2) transmit 'npacket' logical idles.
-- 3) output 1-clk pulse when 'done'.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity txli is
  port (
    clk : in std_logic;
    rst : in std_logic;
    npacket : in std_logic_vector(11 downto 0);
    kickoff : in std_logic;
    txdatak : out std_logic_vector(1 downto 0);
    txdata : out std_logic_vector(15 downto 0);
    done : out std_logic
  );
end txli;
architecture default of txli is
  type LI_STATE is (
    IDLE,
    SEND_LI,
    INVALID_STATE
  );
  signal txlism : LI_STATE;
  signal npacket_left : std_logic_vector(11 downto 0);
begin
  process (clk, rst)
  begin
    if (rst = '1') then
      txlism <= IDLE;
    elsif (rising_edge(clk)) then
      if (kickoff = '1') then
        txlism <= SEND_LI;
        npacket_left <= npacket;
        done <= '0';
      else
        case txlism is
          when IDLE =>
            done <= '0';
            txdatak <= "00";
            txdata <= LOGICAL_IDLE & LOGICAL_IDLE;
          when SEND_LI =>
            txdatak <= "00";
            txdata <= LOGICAL_IDLE & LOGICAL_IDLE;
            if (npacket_left = conv_std_logic_vector(2, npacket_left'length)) then
              done <= '1';
              txlism <= IDLE;
            else
              npacket_left <= unsigned(npacket_left) - 2;
            end if;
          when others =>
            txlism <= IDLE;
        end case;
      end if;
    end if;
  end process;
end default;
--
-- Logical idle receiver
-- 1) reset packet counter when 'clear'ed.
-- 2) output 'npacket', the num of LOGICAL_IDLE
-- sequentially obtained.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity rxli is
  port (
    clk : in std_logic;
    rst : in std_logic;
    npacket : out std_logic_vector(11 downto 0);
    clear : in std_logic;
    rxdatak : in std_logic_vector(1 downto 0);
    rxdata : in std_logic_vector(15 downto 0)
  );
end rxli;
architecture default of rxli is
  type LI_STATE is (
    IDLE,
    RECV_LI,
    INVALID_STATE
  );
  signal rxlism : LI_STATE;
  signal npacket0 : std_logic_vector(11 downto 0);
begin
  npacket <= npacket0;
  process (clk, rst)
  begin
    if (rst = '1') then
      rxlism <= IDLE;
    elsif (rising_edge(clk)) then
      if (clear = '1') then
        npacket0 <= (others => '0');
        rxlism <= IDLE;
      else
        case rxlism is
          when IDLE =>
            -- detect a pair of logical idle.
            if (rxdatak = "00" and rxdata = LOGICAL_IDLE & LOGICAL_IDLE) then
              npacket0 <= conv_std_logic_vector(2, npacket0'length);
              rxlism <= RECV_LI;
            end if;
          when RECV_LI =>
            if (rxdatak = "00" and rxdata = LOGICAL_IDLE & LOGICAL_IDLE) then
              npacket0 <= unsigned(npacket0) + 2;
            else
              rxlism <= IDLE;
            end if;
          when others =>
            rxlism <= IDLE;
        end case;
      end if;
    end if;
  end process;
end default;
--
-- scrambler/descrambler
-- src to dst pipeline delay : 2clk
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity scramble is
  generic (
    DATA_WIDTH : integer;
    SIMULATION : integer
  );
  port (
    src : in std_logic_vector(DATA_WIDTH-1 downto 0);
    iskcode : in std_logic_vector(DATA_WIDTH/8-1 downto 0);
    doscramble : in std_logic;
    dstk : out std_logic_vector(DATA_WIDTH/8-1 downto 0);
    dst : out std_logic_vector(DATA_WIDTH-1 downto 0);
    rst : in std_logic; -- active high reset
    clk : in std_logic
  );
end scramble;
architecture default of scramble is
  signal lfsr : std_logic_vector(15 downto 0); -- linear feedback shift register
  signal src1 : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal src2 : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal scrambled2 : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal iskcode1 : std_logic_vector(DATA_WIDTH/8-1 downto 0);
  signal iskcode2 : std_logic_vector(DATA_WIDTH/8-1 downto 0);
  signal doscramble1 : std_logic;
  signal doscramble2 : std_logic;
  constant MASK : std_logic_vector := "0000000000111001"; -- X16 + X5 + X4 + X3 + 1
begin
  process (clk, rst)
    variable lfsrtmp : std_logic_vector(15 downto 0);
  begin
    if (rst = '1') then
      lfsr <= x"ffff";
    elsif (rising_edge(clk)) then
      -- delayed values.
      iskcode1 <= iskcode;
      iskcode2 <= iskcode1;
      src1 <= src;
      src2 <= src1;
      doscramble1 <= doscramble;
      doscramble2 <= doscramble1;
      -- update LFSR.
      lfsrtmp := lfsr;
      eachbyte : for j in 0 to DATA_WIDTH/8-1 loop
        -- shift LFSR 1-bit to the left if src is not a SKP.
        if (not (iskcode1(j) = '1' and src1(8 * j + 7 downto 8 * j + 0) = K_SKP)) then
          eachbit : for i in 0 to 7 loop
            -- scramble src with LFSR.
            scrambled2(8 * j + i) <= src1(8 * j + i) xor lfsrtmp(15);
            if (lfsrtmp(15) = '1') then
              lfsrtmp := (lfsrtmp(14 downto 0) & '0') xor MASK;
            else
              lfsrtmp := (lfsrtmp(14 downto 0) & '0');
            end if;
          end loop eachbit;
        end if;
        -- reset LFSR when detect a COM.
        if (iskcode1(j) = '1' and src1(8 * j + 7 downto 8 * j + 0) = K_COM) then
          lfsrtmp := x"ffff";
        end if;
      end loop eachbyte;
      lfsr <= lfsrtmp;
    end if;
  end process;
  process (clk)
  begin
    if (rising_edge(clk)) then
      -- do scramble a char if it is not a K-code.
      for j in 0 to DATA_WIDTH/8-1 loop
        dstk(j) <= iskcode2(j);
        if (SIMULATION = 0) then
          if (iskcode2(j) = '1' or doscramble2 = '0') then
            dst(7 + 8*j downto 0 + 8*j) <= src2(7 + 8*j downto 0 + 8*j);
          else
            dst(7 + 8*j downto 0 + 8*j) <= scrambled2(7 + 8*j downto 0 + 8*j);
          end if;
        else
          -- no scrambing (only for debugging purpose).
          dst(7 + 8*j downto 0 + 8*j) <= src2(7 + 8*j downto 0 + 8*j);
        end if;
      end loop;
    end if;
  end process;
end default;
--
-- Deskewer
-- remove lane-to-lane skew by inserting 'delay's.
-- 'delay' of each lane is determined using COM.
-- delay values demain unchanged while 'enable' is 0.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity deskew is
  generic (
    NLANE : integer
  );
  port (
    enable : in std_logic;
    srcd : in each16b(NLANE-1 downto 0);
    srck : in each2b(NLANE-1 downto 0);
    dstd : out each16b(NLANE-1 downto 0);
    dstk : out each2b(NLANE-1 downto 0);
    rst : in std_logic; -- active high
    clk : in std_logic
  );
end deskew;
architecture default of deskew is
  signal srcd0 : each16b(NLANE-1 downto 0);
  signal srcd1 : each16b(NLANE-1 downto 0);
  signal srcd2 : each16b(NLANE-1 downto 0);
  signal srcd3 : each16b(NLANE-1 downto 0);
  signal srcd4 : each16b(NLANE-1 downto 0);
  signal srck0 : each2b(NLANE-1 downto 0);
  signal srck1 : each2b(NLANE-1 downto 0);
  signal srck2 : each2b(NLANE-1 downto 0);
  signal srck3 : each2b(NLANE-1 downto 0);
  signal srck4 : each2b(NLANE-1 downto 0);
  signal foundlow0 : each1b(NLANE-1 downto 0); -- found COM at lower 8-byte on a lane.
  signal foundlow1 : each1b(NLANE-1 downto 0);
  signal foundhigh0 : each1b(NLANE-1 downto 0); -- found COM at higer 8-byte on a lane.
  signal foundhigh1 : each1b(NLANE-1 downto 0);
  signal foundall : std_logic;
  signal foundany0 : std_logic;
  signal foundany1 : std_logic;
  signal swap : each1b(NLANE-1 downto 0); -- swap higer & lower 8-bytes.
  signal swapall : std_logic;
  signal swapany : std_logic;
  signal delay : each2b(NLANE-1 downto 0); -- delay can be as large as 3.
begin
  -- generate delayed signals.
  src_lane : for i in 0 to NLANE-1 generate
    srcd0(i) <= srcd(i);
    srck0(i) <= srck(i);
    process(clk)
    begin
      if (rising_edge(clk)) then
        srcd1(i) <= srcd0(i);
        srck1(i) <= srck0(i);
        srcd2(i) <= srcd1(i);
        srck2(i) <= srck1(i);
        srcd3(i) <= srcd2(i);
        srck3(i) <= srck2(i);
        srcd4(i) <= srcd3(i);
        srck4(i) <= srck3(i);
        foundlow1(i) <= foundlow0(i);
        foundhigh1(i) <= foundhigh0(i);
        foundany1 <= foundany0;
      end if;
    end process;
  end generate src_lane;
  -- check if each lane have seen a COM or not.
  com_lane : for i in 0 to NLANE-1 generate
    process(clk, rst)
    begin
      if (rst = '1') then
        foundlow0(i) <= '0';
        foundhigh0(i) <= '0';
      elsif (rising_edge(clk)) then
        if (enable = '1') then
          if (srck0(i)(0) = '1' and srcd0(i)( 7 downto 0) = K_COM) then
            foundlow0(i) <= '1';
          end if;
          if (srck0(i)(1) = '1' and srcd0(i)(15 downto 8) = K_COM) then
            foundhigh0(i) <= '1';
          end if;
          if (foundall = '1') then -- clear when delay adjustment done.
            foundlow0(i) <= '0';
            foundhigh0(i) <= '0';
          end if;
        end if;
      end if;
    end process;
  end generate com_lane;
  -- found COMs on all lanes.
  process(foundlow0, foundhigh0)
    variable foundalltmp : std_logic;
  begin
    foundalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '0' and foundhigh0(i) = '0') then
        foundalltmp := '0';
      end if;
    end loop;
    foundall <= foundalltmp;
  end process;
  -- found firstly arrived COM(s).
  process(foundlow0, foundhigh0)
    variable foundanytmp : std_logic;
  begin
    foundanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '1') then
        foundanytmp := '1';
      end if;
      if (foundhigh0(i) = '1') then
        foundanytmp := '1';
      end if;
    end loop;
    foundany0 <= foundanytmp;
  end process;
  delay_lane : for i in 0 to NLANE-1 generate
    process(clk)
    begin
      if (rising_edge(clk)) then
        if (foundall = '0') then -- delay adjustment is in progress.
          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i) <= '0';
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i) <= '1'; -- add a bit more (0.5clk = 1/250MHz) delay.
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;
        else -- delay adjustment is about to complete.
          -- insert no delay for lanes which haven't see a COM so far.
          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i) <= '0';
            end if;
          end if;
          -- insert 0.5clk delay
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i) <= '1';
            end if;
          end if;
        end if;
      end if;
    end process;
  end generate delay_lane;
  -- all lanes are swapped.
  process(swap)
    variable swapalltmp : std_logic;
  begin
    swapalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '0') then
        swapalltmp := '0';
      end if;
    end loop;
    swapall <= swapalltmp;
  end process;
  -- at least one lane is swapped.
  process(swap)
    variable swapanytmp : std_logic;
  begin
    swapanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '1') then
        swapanytmp := '1';
      end if;
    end loop;
    swapany <= swapanytmp;
  end process;
  mux_lane : for i in 0 to NLANE-1 generate
    process(srcd0, srcd1, srcd2, srcd3, srcd4,
            srck0, srck1, srck2, srck3, srck4,
            delay, swap, swapany, swapall)
    begin
      -- add delay, and swap if necessary.
      if (swap(i) = '0' or swapall = '1' or swapany = '0') then -- no swap is necessary.
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i) <= srcd0(i);
          dstk(i) <= srck0(i);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i) <= srcd1(i);
          dstk(i) <= srck1(i);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i) <= srcd2(i);
          dstk(i) <= srck2(i);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i) <= srcd3(i);
          dstk(i) <= srck3(i);
        end if;
      else
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd0(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd1(i)(15 downto 8);
          dstk(i)(1) <= srck0(i)(0);
          dstk(i)(0) <= srck1(i)(1);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd1(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd2(i)(15 downto 8);
          dstk(i)(1) <= srck1(i)(0);
          dstk(i)(0) <= srck2(i)(1);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd2(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd3(i)(15 downto 8);
          dstk(i)(1) <= srck2(i)(0);
          dstk(i)(0) <= srck3(i)(1);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd3(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd4(i)(15 downto 8);
          dstk(i)(1) <= srck3(i)(0);
          dstk(i)(0) <= srck4(i)(1);
        end if;
      end if;
    end process;
  end generate mux_lane;
end default;
--
-- Deframer
-- decode a Physical-Layer Packet (PLP) and output its contents to
-- Data-Link-Layer Packet (DLLP) or Transaction-Layer Packet (TLP) port
-- depending on the framing symbol SDP/STP.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity deframer is
  generic (
    DEFRAMER_DLLPBUF_DEPTH : integer;
    DEFRAMER_TLPBUF_DEPTH : integer;
    NLANE : integer
  );
  port (
-- synthesis translate_off
    -- for debugging
    dpusedwmax : out std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
    usedwmax : out std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
-- synthesis translate_on
    -- PLP input
    data : in each16b(NLANE-1 downto 0);
    datak : in each2b(NLANE-1 downto 0);
    -- DLLP output
    wedp : out std_logic;
    dp : out std_logic_vector(47 downto 0);
    -- TLP output
    wetp : out std_logic;
    betp : out std_logic_vector(NLANE*2-1 downto 0);
    tptop : out std_logic;
    tp : out std_logic_vector(NLANE*16-1 downto 0);
    rst : in std_logic; -- active high
    clk : in std_logic
  );
end deframer;
architecture default of deframer is
begin
  x1 : if (NLANE = 1) generate
    x1block : block
      type DP_STATE is (
        IDLE,
        RECV_D0_D1,
        RECV_D2_C0,
        RECV_C1,
        RECV_TYPE_D0,
        RECV_D1_D2,
        RECV_C0_C1,
        INVALID_STATE
      );
      type TP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_NONALIGNED,
        INVALID_STATE
      );
      signal wedp0 : std_logic;
      signal dp0 : std_logic_vector(47 downto 0);
      signal wetp0 : std_logic;
      signal wetp1 : std_logic;
      signal tptop0 : std_logic;
      signal tptop1 : std_logic;
      signal tp0 : std_logic_vector(NLANE*16-1 downto 0);
      signal tp1 : std_logic_vector(NLANE*16-1 downto 0);
      signal dpsm : DP_STATE;
      signal tpsm : TP_STATE;
    begin
      --
      -- deframe 64-bit PLP containing 48-bit DLLP:
      --
      -- DLLP format:
      -- 1B 1B 3B-data 2B-crc 1B
      -- 0 1 2 3 4 5 6 7
      -- SDP type data0 data1 data2 crc0 crc1 END
      --
      -- 16-bit aligned input:
      -- type data1 crc0 END
      -- SDP data0 data2 crc1
      --
      -- 16-bit non-aligned input:
      -- SDP data0 data2 crc1
      -- type data1 crc0 END
      --
      wedp <= wedp0;
      dp <= dp0;
      process (clk, rst)
      begin
        if (rst = '1') then
          wedp0 <= '0';
          dpsm <= IDLE;
        elsif (rising_edge(clk)) then
          wedp0 <= '0';
          case dpsm is
            when IDLE =>
              -- detect a SDP on data0 lower byte.
              if (datak(0)(0) = '1' and data(0)(7 downto 0) = K_SDP) then
                dp0( 7 downto 0) <= data(0)(15 downto 8); -- type
                dpsm <= RECV_D0_D1;
                -- detect a SDP on data0 higher byte.
              elsif (datak(0)(1) = '1' and data(0)(15 downto 8) = K_SDP) then
                dpsm <= RECV_TYPE_D0;
              end if;
              -- write aligned DP
            when RECV_D0_D1 =>
              dp0(15 downto 8) <= data(0)( 7 downto 0); -- data0
              dp0(23 downto 16) <= data(0)(15 downto 8); -- data1
              dpsm <= RECV_D2_C0;
            when RECV_D2_C0 =>
              dp0(31 downto 24) <= data(0)( 7 downto 0); -- data2
              dp0(39 downto 32) <= data(0)(15 downto 8); -- crc high
              dpsm <= RECV_C1;
            when RECV_C1 =>
              dp0(47 downto 40) <= data(0)( 7 downto 0); -- crc low
              wedp0 <= '1';
              dpsm <= IDLE;
              -- write non-aligned DP
            when RECV_TYPE_D0 =>
              dp0( 7 downto 0) <= data(0)( 7 downto 0); -- type
              dp0(15 downto 8) <= data(0)(15 downto 8); -- data0
              dpsm <= RECV_D1_D2;
            when RECV_D1_D2 =>
              dp0(23 downto 16) <= data(0)( 7 downto 0); -- data1
              dp0(31 downto 24) <= data(0)(15 downto 8); -- data2
              dpsm <= RECV_C0_C1;
            when RECV_C0_C1 =>
              wedp0 <= '1';
              dp0(39 downto 32) <= data(0)( 7 downto 0); -- crc high
              dp0(47 downto 40) <= data(0)(15 downto 8); -- crc low
              dpsm <= IDLE;
            when others =>
              wedp0 <= '0';
              dpsm <= IDLE;
          end case;
        end if;
      end process;
      --
      -- deframe PLP containing TLP:
      --
      -- TLP format:
      -- 1B 2B 12/16B 12-4116B 4B 4B 1B
      -- STP Seq# Header Data (ECRC) LCRC EDP/EDB
      --
      -- 16-bit aligned input:
      -- seq0 h0 h2 h4 h6 h8 h10 . . . . . . END
      -- STP seq1 h1 h3 h5 h7 h9 h11 . . . . lcrc3
      --
      -- 16-bit non-aligned input:
      -- STP seq1 h1 h3 h5 h7 h9 h11 . . . lcrc3
      -- seq0 h0 h2 h4 h6 h8 h10 . . . . . . END
      --
      tp <= tp1;
      wetp <= '1' when wetp0 = '1' or wetp1 = '1' else '0';
      tptop <= tptop1;
      process (clk, rst)
      begin
        if (rst = '1') then
          wetp0 <= '0';
          wetp1 <= '0';
          tpsm <= IDLE;
        elsif (rising_edge(clk)) then
          wetp0 <= '0';
          wetp1 <= '0';
          betp <= (others => '1');
          tptop0 <= '0';
          tptop1 <= tptop0;
          case tpsm is
            when IDLE =>
              -- detect a STP on data0 lower byte.
              if (datak(0)(0) = '1' and data(0)(7 downto 0) = K_STP) then
                tp0( 7 downto 0) <= data(0)(15 downto 8); -- seq0
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
                -- detect a STP on data0 higher byte.
              elsif (datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP) then
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              end if;
              -- write aligned TP
            when RECV_ALIGNED =>
              tp1(15 downto 8) <= data(0)( 7 downto 0);
              tp1( 7 downto 0) <= tp0( 7 downto 0);
              tp0( 7 downto 0) <= data(0)(15 downto 8);
              wetp1 <= wetp0;
              if (datak(0)(1) = '1' and
                  (data(0)(15 downto 8) = K_END or data(0)(15 downto 8) = K_EDB)) then
                tpsm <= IDLE;
              else
                wetp0 <= '1';
              end if;
            when RECV_NONALIGNED =>
              tp1(15 downto 8) <= data(0)(15 downto 8);
              tp1( 7 downto 0) <= data(0)( 7 downto 0);
              if (datak(0)(0) = '1' and
                  (data(0)( 7 downto 0) = K_END or data(0)( 7 downto 0) = K_EDB)) then
                -- detect a STP on data0 higher byte.
                -- i.e., END is immediately followed by STP of next TLP.
                if (datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP) then
                  tptop0 <= '1';
                else
                  tpsm <= IDLE;
                end if;
              else
                wetp0 <= '1';
                wetp1 <= '1';
              end if;
            when others =>
              tpsm <= IDLE;
          end case;
        end if;
      end process;
    end block x1block;
  end generate x1;
  x4 : if (NLANE = 4) generate
    x4block : block
      component scfifo
        generic (
          LPM_NUMWORDS : natural := 2**DEFRAMER_TLPBUF_DEPTH;
          LPM_WIDTHU : natural := DEFRAMER_TLPBUF_DEPTH;
          LPM_WIDTH : natural := NLANE*18;
          LPM_SHOWAHEAD : string := "ON";
          ADD_RAM_OUTPUT_REGISTER : string := "ON";
          OVERFLOW_CHECKING : string := "ON";
          UNDERFLOW_CHECKING : string := "OFF";
          USE_EAB : string := "ON";
          INTENDED_DEVICE_FAMILY : string := "Stratix II GX";
          LPM_TYPE : string := "scfifo"
        );
        port (
          usedw : out std_logic_vector(LPM_WIDTHU-1 downto 0);
          rdreq : in std_logic;
          full : out std_logic;
          empty : out std_logic;
          clock : in std_logic;
          sclr : in std_logic;
          q : out std_logic_vector(LPM_WIDTH-1 downto 0);
          wrreq : in std_logic;
          data : in std_logic_vector(LPM_WIDTH-1 downto 0)
        );
      end component;
      type DP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_NONALIGNED,
        INVALID_STATE
      );
      type TP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_END_ALIGNED,
        RECV_NONALIGNED,
        RECV_END_NONALIGNED,
        INVALID_STATE
      );
      type BUF_STATE is (
        IDLE,
        RECV
      );
      signal dpsm : DP_STATE;
      signal tpsm : TP_STATE;
      signal bufsm : BUF_STATE;
      signal wedp0 : std_logic;
      signal wedp1 : std_logic;
      signal dp0 : std_logic_vector(47 downto 0);
      signal dp1 : std_logic_vector(47 downto 0);
      signal wetp0 : std_logic;
      signal wetp1 : std_logic;
      signal tptop0 : std_logic;
      signal tptop1 : std_logic;
      signal tp0 : std_logic_vector(NLANE*16-1 downto 0);
      signal tp1 : std_logic_vector(NLANE*16-1 downto 0);
      signal end_at_low : std_logic;
      signal bufin : std_logic_vector(NLANE*18-1 downto 0);
      signal bufout : std_logic_vector(NLANE*18-1 downto 0);
      signal wrreq : std_logic;
      signal rdreq : std_logic;
      signal rdack : std_logic;
      signal full : std_logic;
      signal empty : std_logic;
      signal d : each16b(NLANE-1 downto 0);
      signal k : each2b(NLANE-1 downto 0);
      signal sdp_at_0l : std_logic;
      signal sdp_at_0h : std_logic;
      signal end_at_3l : std_logic;
      signal end_at_3h : std_logic;
      signal stp_at_0l_in : std_logic;
      signal stp_at_0h_in : std_logic;
      signal end_at_3l_in : std_logic;
      signal end_at_3h_in : std_logic;
      signal stp_at_0l_out : std_logic;
      signal stp_at_0h_out : std_logic;
      signal end_at_3l_out : std_logic;
      signal end_at_3h_out : std_logic;
      signal usedw0 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
      signal usedwmax0 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
    begin
-- synthesis translate_off
      -- for debugging
      fullout <= full;
      usedwmax <= usedwmax0;
-- synthesis translate_on
      -- DLLP 16-bit aligned input:
      --
      -- lane3 H END
      -- lane2 H crc1
      -- lane1 H crc0
      -- lane0 H data2
      -- lane3 L data1
      -- lane2 L data0
      -- lane1 L type
      -- lane0 L SDP
      --
      -- DLLP 16-bit non-aligned input:
      -- lane3 H data1
      -- lane2 H data0
      -- lane1 H type
      -- lane0 H SDP
      -- lane3 L END
      -- lane2 L crc1
      -- lane1 L crc0
      -- lane0 L data2
      wedp <= wedp1;
      dp <= dp1;
      sdp_at_0l <= '1' when datak(0)(0) = '1' and data(0)( 7 downto 0) = K_SDP else '0';
      sdp_at_0h <= '1' when datak(0)(1) = '1' and data(0)(15 downto 8) = K_SDP else '0';
      end_at_3l <= '1' when datak(3)(0) = '1' and data(3)( 7 downto 0) = K_END else '0';
      end_at_3h <= '1' when datak(3)(1) = '1' and data(3)(15 downto 8) = K_END else '0';
      process (clk, rst)
      begin
        if (rst = '1') then
          wedp0 <= '0';
          wedp1 <= '0';
          dpsm <= IDLE;
        elsif (rising_edge(clk)) then
          wedp0 <= '0';
          wedp1 <= '0';
          case dpsm is
            when IDLE =>
              dp1 <= dp0;
              wedp1 <= wedp0;
              -- detect a SDP on data0 lower byte.
              if (sdp_at_0l = '1') then
                dp0(47 downto 40) <= data(2)(15 downto 8); -- crc low
                dp0(39 downto 32) <= data(1)(15 downto 8); -- crc high
                dp0(31 downto 24) <= data(0)(15 downto 8); -- data2
                dp0(23 downto 16) <= data(3)( 7 downto 0); -- data1
                dp0(15 downto 8) <= data(2)( 7 downto 0); -- data0
                dp0( 7 downto 0) <= data(1)( 7 downto 0); -- type
                wedp0 <= '1';
                -- detect a SDP on data0 higher byte.
              elsif (sdp_at_0h = '1') then
                dp0(23 downto 16) <= data(3)(15 downto 8); -- data1
                dp0(15 downto 8) <= data(2)(15 downto 8); -- data0
                dp0( 7 downto 0) <= data(1)(15 downto 8); -- type
                dpsm <= RECV_NONALIGNED;
              end if;
            when RECV_NONALIGNED =>
              wedp1 <= '1';
              dp1(47 downto 40) <= data(2)( 7 downto 0); -- crc low
              dp1(39 downto 32) <= data(1)( 7 downto 0); -- crc high
              dp1(31 downto 24) <= data(0)( 7 downto 0); -- data2
              dp1(23 downto 0) <= dp0(23 downto 0);
              if (sdp_at_0h = '1') then
                dp0(23 downto 16) <= data(3)(15 downto 8); -- data1
                dp0(15 downto 8) <= data(2)(15 downto 8); -- data0
                dp0( 7 downto 0) <= data(1)(15 downto 8); -- type
              else
                dpsm <= IDLE;
              end if;
            when others =>
              dpsm <= IDLE;
          end case;
        end if;
      end process;
      --
      -- TLP
      --
      tp <= tp1;
      tptop <= tptop1;
      wetp <= '1' when wetp0 = '1' or wetp1 = '1' else '0';
      process (clk)
      begin
        if (rising_edge(clk)) then
          for i in 0 to NLANE-1 loop
            bufin(i*18+7 downto i*18+0) <= data(i)(7 downto 0);
            bufin(i*18+8) <= datak(i)(0);
            bufin(i*18+16 downto i*18+9) <= data(i)(15 downto 8);
            bufin(i*18+17) <= datak(i)(1);
          end loop;
        end if;
      end process;
      lane: for i in 0 to NLANE-1 generate
        d(i)(7 downto 0) <= bufout(i*18+7 downto i*18+0);
        k(i)(0) <= bufout(i*18+8);
        d(i)(15 downto 8) <= bufout(i*18+16 downto i*18+9);
        k(i)(1) <= bufout(i*18+17);
      end generate lane;
      stp_at_0l_in <= '1' when datak(0)(0) = '1' and data(0)( 7 downto 0) = K_STP else '0';
      stp_at_0h_in <= '1' when datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP else '0';
      end_at_3l_in <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
      end_at_3h_in <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
      stp_at_0l_out <= '1' when k(0)(0) = '1' and d(0)( 7 downto 0) = K_STP else '0';
      stp_at_0h_out <= '1' when k(0)(1) = '1' and d(0)(15 downto 8) = K_STP else '0';
      end_at_3l_out <= '1' when k(3)(0) = '1' and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
      end_at_3h_out <= '1' when k(3)(1) = '1' and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
      rxbuf0 : scfifo
        port map (
          usedw => usedw0,
          clock => clk,
          sclr => rst,
          wrreq => wrreq,
          data => bufin,
          rdreq => rdreq,
          full => full,
          empty => empty,
          q => bufout
        );
      process (clk, rst)
      begin
        if (rst = '1') then
          usedwmax0 <= (others => '0');
        elsif (rising_edge(clk)) then
          if (unsigned(usedwmax0) < unsigned(usedw0)) then
            usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
          end if;
        end if;
      end process;
      process (clk, rst)
      begin
        if (rst = '1') then
          wrreq <= '0';
          bufsm <= IDLE;
        elsif (rising_edge(clk)) then
          wrreq <= '0';
          case bufsm is
            when IDLE =>
              if (stp_at_0l_in = '1' or stp_at_0h_in = '1') then
                wrreq <= '1';
                bufsm <= RECV;
              end if;
            when RECV =>
              wrreq <= '1';
              if (end_at_3h_in = '1' or (end_at_3l_in = '1' and stp_at_0h_in = '0')) then
                bufsm <= IDLE;
              end if;
            when others =>
          end case;
        end if;
      end process;
      rdreq <= '1' when empty = '0' and rdack = '1' else '0';
      rdack <= '0' when end_at_3l_out = '1' and stp_at_0h_out = '1' and tpsm = RECV_ALIGNED else '1';
      process (clk, rst)
      begin
        if (rst = '1') then
          wetp0 <= '0';
          wetp1 <= '0';
          tpsm <= IDLE;
        elsif (rising_edge(clk)) then
          wetp0 <= '0';
          wetp1 <= '0';
          betp <= (others => '1');
          tptop0 <= '0';
          tptop1 <= tptop0;
          case tpsm is
            when IDLE =>
              if (empty = '0') then
                -- detect a STP on lane0 lower byte.
                if (k(0)(0) = '1' and d(0)(7 downto 0) = K_STP) then
                  tp0(55 downto 48) <= d(3)(15 downto 8);
                  tp0(47 downto 40) <= d(2)(15 downto 8);
                  tp0(39 downto 32) <= d(1)(15 downto 8);
                  tp0(31 downto 24) <= d(0)(15 downto 8);
                  tp0(23 downto 16) <= d(3)( 7 downto 0);
                  tp0(15 downto 8) <= d(2)( 7 downto 0);
                  tp0( 7 downto 0) <= d(1)( 7 downto 0);
                  tptop0 <= '1';
                  tpsm <= RECV_ALIGNED;
                  -- detect a STP on lane0 higher byte.
                elsif (k(0)(1) = '1' and d(0)(15 downto 8) = K_STP) then
                  tp0(23 downto 16) <= d(3)(15 downto 8);
                  tp0(15 downto 8) <= d(2)(15 downto 8);
                  tp0( 7 downto 0) <= d(1)(15 downto 8);
                  tptop0 <= '1';
                  tpsm <= RECV_NONALIGNED;
                end if;
              end if;
            when RECV_ALIGNED =>
              tp1(63 downto 56) <= d(0)(7 downto 0);
              tp1(55 downto 0) <= tp0(55 downto 0);
              tp0(55 downto 48) <= d(3)(15 downto 8);
              tp0(47 downto 40) <= d(2)(15 downto 8);
              tp0(39 downto 32) <= d(1)(15 downto 8);
              tp0(31 downto 24) <= d(0)(15 downto 8);
              tp0(23 downto 16) <= d(3)( 7 downto 0);
              tp0(15 downto 8) <= d(2)( 7 downto 0);
              tp0( 7 downto 0) <= d(1)( 7 downto 0);
              wetp1 <= wetp0;
              wetp0 <= '1';
              if (end_at_3h_out = '1') then
                end_at_low <= '0';
                tpsm <= RECV_END_ALIGNED;
              elsif (end_at_3l_out = '1') then
                end_at_low <= '1';
                tpsm <= RECV_END_ALIGNED;
              end if;
            when RECV_END_ALIGNED =>
              wetp1 <= wetp0;
              wetp0 <= '1';
              tp1(47 downto 0) <= tp0(47 downto 0);
              if (end_at_low = '0') then
                betp <= "00111111";
                tp1(63 downto 48) <= (others => '0');
              else
                betp <= "00000011";
                tp1(63 downto 16) <= (others => '0');
              end if;
              if (stp_at_0l_out = '1') then
                tp0(55 downto 48) <= d(3)(15 downto 8);
                tp0(47 downto 40) <= d(2)(15 downto 8);
                tp0(39 downto 32) <= d(1)(15 downto 8);
                tp0(31 downto 24) <= d(0)(15 downto 8);
                tp0(23 downto 16) <= d(3)( 7 downto 0);
                tp0(15 downto 8) <= d(2)( 7 downto 0);
                tp0( 7 downto 0) <= d(1)( 7 downto 0);
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
              elsif (stp_at_0h_out = '1') then
                tp0(23 downto 16) <= d(3)(15 downto 8);
                tp0(15 downto 8) <= d(2)(15 downto 8);
                tp0( 7 downto 0) <= d(1)(15 downto 8);
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              else
                tpsm <= IDLE;
              end if;
            when RECV_NONALIGNED =>
              tp1(63 downto 56) <= d(0)(15 downto 8);
              tp1(55 downto 48) <= d(3)( 7 downto 0);
              tp1(47 downto 40) <= d(2)( 7 downto 0);
              tp1(39 downto 32) <= d(1)( 7 downto 0);
              tp1(31 downto 24) <= d(0)( 7 downto 0);
              tp1(23 downto 0) <= tp0(23 downto 0);
              tp0(23 downto 16) <= d(3)(15 downto 8);
              tp0(15 downto 8) <= d(2)(15 downto 8);
              tp0( 7 downto 0) <= d(1)(15 downto 8);
              wetp1 <= wetp0;
              wetp0 <= '1';
              if (end_at_3h_out = '1') then
                end_at_low <= '0';
                tpsm <= RECV_END_NONALIGNED;
              end if;
              if (end_at_3l_out = '1') then
                end_at_low <= '1';
                if (stp_at_0h_out = '1') then
                  betp <= "00111111";
                  tp1(63 downto 48) <= (others => '0');
                  tptop0 <= '1';
                else
                  betp <= "00111111";
                  tp1(63 downto 48) <= (others => '0');
-- tp0(23 downto 0) <= (others => '0');
                  tpsm <= IDLE;
                end if;
              end if;
            when RECV_END_NONALIGNED =>
              wetp1 <= wetp0;
              wetp0 <= '1';
              tp1(23 downto 0) <= tp0(23 downto 0);
              if (end_at_low = '0') then
                betp <= "00000011";
                tp1(63 downto 16) <= (others => '0');
              else
                betp <= "00111111";
                tp1(63 downto 48) <= (others => '0');
              end if;
              if (stp_at_0l_out = '1') then
                tp0(55 downto 48) <= d(3)(15 downto 8);
                tp0(47 downto 40) <= d(2)(15 downto 8);
                tp0(39 downto 32) <= d(1)(15 downto 8);
                tp0(31 downto 24) <= d(0)(15 downto 8);
                tp0(23 downto 16) <= d(3)( 7 downto 0);
                tp0(15 downto 8) <= d(2)( 7 downto 0);
                tp0( 7 downto 0) <= d(1)( 7 downto 0);
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
              elsif (stp_at_0h_out = '1') then
                tp0(23 downto 16) <= d(3)(15 downto 8);
                tp0(15 downto 8) <= d(2)(15 downto 8);
                tp0( 7 downto 0) <= d(1)(15 downto 8);
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              else
                tpsm <= IDLE;
              end if;
            when others =>
              tpsm <= IDLE;
          end case;
        end if;
      end process;
    end block x4block;
  end generate x4;
  x8 : if (NLANE = 8) generate
    x8block : block
    begin
      --
      -- deframe DLLP
      --
      sdpblock : block
        component scfifo
          generic (
            LPM_NUMWORDS : natural := 2**DEFRAMER_DLLPBUF_DEPTH;
            LPM_WIDTHU : natural := DEFRAMER_DLLPBUF_DEPTH;
            LPM_WIDTH : natural := NLANE*18 + 2;
            LPM_SHOWAHEAD : string := "ON";
            ADD_RAM_OUTPUT_REGISTER : string := "ON";
            OVERFLOW_CHECKING : string := "ON";
            UNDERFLOW_CHECKING : string := "OFF";
            USE_EAB : string := "ON";
            INTENDED_DEVICE_FAMILY : string := "Stratix II GX";
            LPM_TYPE : string := "scfifo"
          );
          port (
            usedw : out std_logic_vector(LPM_WIDTHU-1 downto 0);
            rdreq : in std_logic;
            full : out std_logic;
            empty : out std_logic;
            clock : in std_logic;
            sclr : in std_logic;
            q : out std_logic_vector(LPM_WIDTH-1 downto 0);
            wrreq : in std_logic;
            data : in std_logic_vector(LPM_WIDTH-1 downto 0)
          );
        end component;
        signal wrcnt : std_logic_vector(1 downto 0);
        signal rdcnt, rdcnt0 : std_logic_vector(1 downto 0);
        signal sdpat0 : std_logic_vector(NLANE/2-1 downto 0);
        signal wedp1 : std_logic;
        signal dp0, dp1 : std_logic_vector(47 downto 0);
        signal bufin : std_logic_vector(NLANE*18+1 downto 0);
        signal bufout : std_logic_vector(NLANE*18+1 downto 0);
        signal wrreq : std_logic;
        signal rdreq : std_logic;
        signal rdack : std_logic;
        signal full : std_logic;
        signal empty : std_logic;
        signal usedw0 : std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
        signal usedwmax0 : std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
        signal d : each16b(NLANE-1 downto 0);
        signal k : each2b(NLANE-1 downto 0);
        signal sdp_at_0l_in, sdp_at_4l_in, sdp_at_0h_in, sdp_at_4h_in : std_logic;
        signal end_at_3l_in, end_at_7l_in, end_at_3h_in, end_at_7h_in : std_logic;
        signal sdp_at_any_lane_in : std_logic;
        signal end_at_any_lane_in : std_logic;
        signal sdp_at_0l_out, sdp_at_4l_out, sdp_at_0h_out, sdp_at_4h_out : std_logic;
        signal end_at_3l_out, end_at_7l_out, end_at_3h_out, end_at_7h_out : std_logic;
        signal sdp_at_any_lane_out : std_logic;
        signal end_at_any_lane_out : std_logic;
-- attribute keep : boolean;
-- attribute keep of usedwmax0 : signal is true;
      begin
-- synthesis translate_off
        -- for debugging
        dpfullout <= full;
        dpusedwmax <= usedwmax0;
-- synthesis translate_on
        dp <= dp1;
        wedp <= wedp1;
        process (clk)
        begin
          if (rising_edge(clk)) then
            for i in 0 to NLANE-1 loop
              bufin(i*18+7 downto i*18+0) <= data(i)(7 downto 0);
              bufin(i*18+8) <= datak(i)(0);
              bufin(i*18+16 downto i*18+9) <= data(i)(15 downto 8);
              bufin(i*18+17) <= datak(i)(1);
            end loop;
          end if;
        end process;
        bufin(NLANE*18+1 downto NLANE*18) <= wrcnt; -- direct connection with no register.
        process(empty, bufout)
        begin
          for i in 0 to NLANE-1 loop
            if (empty = '0') then
              d(i)(7 downto 0) <= bufout(i*18+7 downto i*18+0);
              k(i)(0) <= bufout(i*18+8);
              d(i)(15 downto 8) <= bufout(i*18+16 downto i*18+9);
              k(i)(1) <= bufout(i*18+17);
            else
              d(i) <= (others => '0');
              k(i) <= (others => '0');
            end if;
          end loop;
        end process;
        rdcnt <= bufout(NLANE*18+1 downto NLANE*18);
        sdp_at_0l_in <= '1' when datak(0)(0) = '1' and data(0)( 7 downto 0) = K_SDP else '0';
        sdp_at_4l_in <= '1' when datak(4)(0) = '1' and data(4)( 7 downto 0) = K_SDP else '0';
        sdp_at_0h_in <= '1' when datak(0)(1) = '1' and data(0)(15 downto 8) = K_SDP else '0';
        sdp_at_4h_in <= '1' when datak(4)(1) = '1' and data(4)(15 downto 8) = K_SDP else '0';
        end_at_3l_in <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_in <= '1' when datak(7)(0) = '1' and (data(7)( 7 downto 0) = K_END or data(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_in <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_in <= '1' when datak(7)(1) = '1' and (data(7)(15 downto 8) = K_END or data(7)(15 downto 8) = K_EDB) else '0';
        sdp_at_any_lane_in <= '1' when (sdp_at_0l_in = '1' or sdp_at_4l_in = '1' or sdp_at_0h_in = '1' or sdp_at_4h_in = '1') else '0';
        end_at_any_lane_in <= '1' when (end_at_3l_in = '1' or end_at_7l_in = '1' or end_at_3h_in = '1' or end_at_7h_in = '1') else '0';
        sdp_at_0l_out <= '1' when k(0)(0) = '1' and d(0)( 7 downto 0) = K_SDP else '0';
        sdp_at_4l_out <= '1' when k(4)(0) = '1' and d(4)( 7 downto 0) = K_SDP else '0';
        sdp_at_0h_out <= '1' when k(0)(1) = '1' and d(0)(15 downto 8) = K_SDP else '0';
        sdp_at_4h_out <= '1' when k(4)(1) = '1' and d(4)(15 downto 8) = K_SDP else '0';
        end_at_3l_out <= '1' when k(3)(0) = '1' and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_out <= '1' when k(7)(0) = '1' and (d(7)( 7 downto 0) = K_END or d(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_out <= '1' when k(3)(1) = '1' and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_out <= '1' when k(7)(1) = '1' and (d(7)(15 downto 8) = K_END or d(7)(15 downto 8) = K_EDB) else '0';
        sdp_at_any_lane_out <= '1' when (sdp_at_0l_out = '1' or sdp_at_4l_out = '1' or sdp_at_0h_out = '1' or sdp_at_4h_out = '1') else '0';
        end_at_any_lane_out <= '1' when (end_at_3l_out = '1' or end_at_7l_out = '1' or end_at_3h_out = '1' or end_at_7h_out = '1') else '0';
        rxbuf0 : scfifo
          port map (
            usedw => usedw0,
            clock => clk,
            sclr => rst,
            wrreq => wrreq,
            data => bufin,
            rdreq => rdreq,
            full => full,
            empty => empty,
            q => bufout
          );
        process (clk, rst)
        begin
          if (rst = '1') then
            usedwmax0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (unsigned(usedwmax0) < unsigned(usedw0)) then
              usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
            end if;
          end if;
        end process;
        process (clk, rst)
        begin
          if (rst = '1') then
            wrreq <= '0';
          elsif (rising_edge(clk)) then
            wrreq <= '0';
            -- buffer when see an SDP or an END.
            if (sdp_at_any_lane_in = '1' or end_at_3l_in = '1') then
              wrreq <= '1';
            end if;
            -- stop rdreq 2clk when see an END, an SDP-END pair, and SDP in the same clk.
            if (end_at_3l_in = '1' and sdp_at_4l_in = '1' and sdp_at_4h_in = '1') then
              wrcnt <= "10";
            -- stop rdreq 1clk when see an END followed by SDP in the same clk.
            elsif (end_at_3l_in = '1' and (sdp_at_4l_in = '1' or sdp_at_0h_in = '1' or sdp_at_4h_in = '1')) then
              wrcnt <= "01";
            elsif (end_at_7l_in = '1' and (sdp_at_0h_in = '1' or sdp_at_4h_in = '1')) then
              wrcnt <= "01";
            elsif (end_at_3h_in = '1' and sdp_at_4h_in = '1') then
              wrcnt <= "01";
            else
              wrcnt <= "00";
            end if;
          end if;
        end process;
        process(clk, rst)
        begin
          if (rst = '1') then
            rdcnt0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (rdcnt0 = conv_std_logic_vector(0, rdcnt'length)) then
              if (empty = '0' and rdcnt /= conv_std_logic_vector(0, rdcnt'length)) then
                rdcnt0 <= rdcnt;
              end if;
            else
              rdcnt0 <= unsigned(rdcnt0) - 1;
            end if;
          end if;
        end process;
        rdreq <= '1' when empty = '0' and rdack = '1' else '0';
        rdack <= '0' when rdcnt /= conv_std_logic_vector(0, rdcnt'length) and
                          rdcnt0 /= conv_std_logic_vector(1, rdcnt0'length) else '1';
        process(clk)
          variable sdpattmp : std_logic_vector(3 downto 0);
        begin
          if (rising_edge(clk)) then
            --
            -- prestage
            --
            sdpattmp := (others => '0');
            if (empty = '0') then
              if (rdcnt = conv_std_logic_vector(2, rdcnt'length)) then  -- END, SDP-END, SDP in a single clk.
                if (rdcnt0 = conv_std_logic_vector(2, rdcnt0'length)) then
                  sdpattmp := "0010";
                elsif (rdcnt0 = conv_std_logic_vector(1, rdcnt0'length)) then
                  sdpattmp := "1000";
                end if;
              elsif (sdp_at_0l_out = '1' and sdp_at_0h_out = '1') then -- two SDPs at 0L & 0H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0001";
                else
                  sdpattmp := "0100";
                end if;
              elsif (sdp_at_0l_out = '1' and sdp_at_4h_out = '1') then -- two SDPs at 0L & 4H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0001";
                else
                  sdpattmp := "1000";
                end if;
              elsif (sdp_at_4l_out = '1' and sdp_at_4h_out = '1') then -- two SDPs at 4L & 4H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0010";
                else
                  sdpattmp := "1000";
                end if;
              else -- signle SDP.
                if (sdp_at_0l_out = '1') then
                  sdpattmp := "0001";
                elsif (sdp_at_4l_out = '1') then
                  sdpattmp := "0010";
                elsif (sdp_at_0h_out = '1') then
                  sdpattmp := "0100";
                elsif (sdp_at_4h_out = '1') then
                  sdpattmp := "1000";
                end if;
              end if;
              -- last-half of the previous DLLP.
              if (end_at_3l_out = '1' and rdcnt0 = conv_std_logic_vector(0, rdcnt'length)) then
                sdpattmp := "0000";
              end if;
            end if;
            sdpat0 <= sdpattmp;
            case sdpattmp is
              when "0001" =>
                dp0( 47 downto 40) <= d(6)( 7 downto 0);
                dp0( 39 downto 32) <= d(5)( 7 downto 0);
                dp0( 31 downto 24) <= d(4)( 7 downto 0);
                dp0( 23 downto 16) <= d(3)( 7 downto 0);
                dp0( 15 downto 8) <= d(2)( 7 downto 0);
                dp0( 7 downto 0) <= d(1)( 7 downto 0);
              when "0010" =>
                dp0( 47 downto 40) <= d(2)(15 downto 8);
                dp0( 39 downto 32) <= d(1)(15 downto 8);
                dp0( 31 downto 24) <= d(0)(15 downto 8);
                dp0( 23 downto 16) <= d(7)( 7 downto 0);
                dp0( 15 downto 8) <= d(6)( 7 downto 0);
                dp0( 7 downto 0) <= d(5)( 7 downto 0);
              when "0100" =>
                dp0( 47 downto 40) <= d(6)(15 downto 8);
                dp0( 39 downto 32) <= d(5)(15 downto 8);
                dp0( 31 downto 24) <= d(4)(15 downto 8);
                dp0( 23 downto 16) <= d(3)(15 downto 8);
                dp0( 15 downto 8) <= d(2)(15 downto 8);
                dp0( 7 downto 0) <= d(1)(15 downto 8);
              when "1000" =>
                dp0( 23 downto 16) <= d(7)(15 downto 8);
                dp0( 15 downto 8) <= d(6)(15 downto 8);
                dp0( 7 downto 0) <= d(5)(15 downto 8);
              when others =>
            end case;
            --
            -- stage 0
            --
            if (sdpat0 = conv_std_logic_vector(0, sdpat0'length)) then
              wedp1 <= '0';
            else
              wedp1 <= '1';
            end if;
            dp1 <= dp0;
            if (sdpat0 = "1000") then
              dp1( 47 downto 40) <= d(2)(7 downto 0);
              dp1( 39 downto 32) <= d(1)(7 downto 0);
              dp1( 31 downto 24) <= d(0)(7 downto 0);
            end if;
          end if; -- clk
        end process;
      end block sdpblock;
      --
      -- deframe TLP
      --
      stpblock : block
        component scfifo
          generic (
            LPM_NUMWORDS : natural := 2**DEFRAMER_TLPBUF_DEPTH;
            LPM_WIDTHU : natural := DEFRAMER_TLPBUF_DEPTH;
            LPM_WIDTH : natural := NLANE*18;
            LPM_SHOWAHEAD : string := "ON";
            ADD_RAM_OUTPUT_REGISTER : string := "ON";
            OVERFLOW_CHECKING : string := "ON";
            UNDERFLOW_CHECKING : string := "OFF";
            USE_EAB : string := "ON";
            INTENDED_DEVICE_FAMILY : string := "Stratix II GX";
            LPM_TYPE : string := "scfifo"
            );
          port (
            usedw : out std_logic_vector(LPM_WIDTHU-1 downto 0);
            rdreq : in std_logic;
            full : out std_logic;
            empty : out std_logic;
            clock : in std_logic;
            sclr : in std_logic;
            q : out std_logic_vector(LPM_WIDTH-1 downto 0);
            wrreq : in std_logic;
            data : in std_logic_vector(LPM_WIDTH-1 downto 0)
            );
        end component;
        type BUF_STATE is (
          IDLE,
          RECV
        );
        signal bufsm : BUF_STATE;
        signal rdcnt : std_logic;
        signal stpat0, stpat1 : std_logic_vector(NLANE/2-1 downto 0);
        signal endat0, endat1 : std_logic_vector(NLANE/2-1 downto 0);
        signal wetp0, wetp1, wetp2 : std_logic;
        signal tptop0, tptop1, tptop2 : std_logic;
        signal tpend0, tpend1 : std_logic;
        signal betp2 : std_logic_vector(NLANE*2-1 downto 0);
        signal tp0, tp1, tp2 : std_logic_vector(NLANE*16-1 downto 0);
        signal bufin : std_logic_vector(NLANE*18-1 downto 0);
        signal bufout : std_logic_vector(NLANE*18-1 downto 0);
        signal wrreq : std_logic;
        signal rdreq : std_logic;
        signal rdack : std_logic;
        signal full : std_logic;
        signal empty : std_logic;
        signal usedw0 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
        signal usedwmax0 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
        signal d : each16b(NLANE-1 downto 0);
        signal k : each2b(NLANE-1 downto 0);
        signal stp_at_0l_in, stp_at_4l_in, stp_at_0h_in, stp_at_4h_in : std_logic;
        signal end_at_3l_in, end_at_7l_in, end_at_3h_in, end_at_7h_in : std_logic;
        signal stp_at_any_lane_in : std_logic;
        signal end_at_any_lane_in : std_logic;
        signal stp_at_0l_out, stp_at_4l_out, stp_at_0h_out, stp_at_4h_out : std_logic;
        signal end_at_3l_out, end_at_7l_out, end_at_3h_out, end_at_7h_out : std_logic;
        signal stp_at_any_lane_out : std_logic;
        signal end_at_any_lane_out : std_logic;
-- synthesis translate_off
        attribute keep : boolean;
        attribute keep of usedwmax0 : signal is true;
-- synthesis translate_on
      begin
-- synthesis translate_off
        -- for debugging
        fullout <= full;
        usedwmax <= usedwmax0;
-- synthesis translate_on
        tp <= tp2;
        tptop <= tptop2;
        wetp <= wetp2;
        betp <= betp2;
        process (clk)
        begin
          if (rising_edge(clk)) then
            for i in 0 to NLANE-1 loop
              bufin(i*18+7 downto i*18+0) <= data(i)(7 downto 0);
              bufin(i*18+8) <= datak(i)(0);
              bufin(i*18+16 downto i*18+9) <= data(i)(15 downto 8);
              bufin(i*18+17) <= datak(i)(1);
            end loop;
          end if;
        end process;
        lane: for i in 0 to NLANE-1 generate
          d(i)(7 downto 0) <= bufout(i*18+7 downto i*18+0);
          k(i)(0) <= bufout(i*18+8);
          d(i)(15 downto 8) <= bufout(i*18+16 downto i*18+9);
          k(i)(1) <= bufout(i*18+17);
        end generate lane;
        stp_at_0l_in <= '1' when datak(0)(0) = '1' and data(0)( 7 downto 0) = K_STP else '0';
        stp_at_4l_in <= '1' when datak(4)(0) = '1' and data(4)( 7 downto 0) = K_STP else '0';
        stp_at_0h_in <= '1' when datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP else '0';
        stp_at_4h_in <= '1' when datak(4)(1) = '1' and data(4)(15 downto 8) = K_STP else '0';
        end_at_3l_in <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_in <= '1' when datak(7)(0) = '1' and (data(7)( 7 downto 0) = K_END or data(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_in <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_in <= '1' when datak(7)(1) = '1' and (data(7)(15 downto 8) = K_END or data(7)(15 downto 8) = K_EDB) else '0';
        stp_at_any_lane_in <= '1' when (stp_at_0l_in = '1' or stp_at_4l_in = '1' or stp_at_0h_in = '1' or stp_at_4h_in = '1') else '0';
        end_at_any_lane_in <= '1' when (end_at_3l_in = '1' or end_at_7l_in = '1' or end_at_3h_in = '1' or end_at_7h_in = '1') else '0';
        stp_at_0l_out <= '1' when k(0)(0) = '1' and d(0)( 7 downto 0) = K_STP else '0';
        stp_at_4l_out <= '1' when k(4)(0) = '1' and d(4)( 7 downto 0) = K_STP else '0';
        stp_at_0h_out <= '1' when k(0)(1) = '1' and d(0)(15 downto 8) = K_STP else '0';
        stp_at_4h_out <= '1' when k(4)(1) = '1' and d(4)(15 downto 8) = K_STP else '0';
        end_at_3l_out <= '1' when k(3)(0) = '1' and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_out <= '1' when k(7)(0) = '1' and (d(7)( 7 downto 0) = K_END or d(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_out <= '1' when k(3)(1) = '1' and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_out <= '1' when k(7)(1) = '1' and (d(7)(15 downto 8) = K_END or d(7)(15 downto 8) = K_EDB) else '0';
        stp_at_any_lane_out <= '1' when (stp_at_0l_out = '1' or stp_at_4l_out = '1' or stp_at_0h_out = '1' or stp_at_4h_out = '1') else '0';
        end_at_any_lane_out <= '1' when (end_at_3l_out = '1' or end_at_7l_out = '1' or end_at_3h_out = '1' or end_at_7h_out = '1') else '0';
        rxbuf0 : scfifo
          port map (
            usedw => usedw0,
            clock => clk,
            sclr => rst,
            wrreq => wrreq,
            data => bufin,
            rdreq => rdreq,
            full => full,
            empty => empty,
            q => bufout
            );
        process (clk, rst)
        begin
          if (rst = '1') then
            usedwmax0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (unsigned(usedwmax0) < unsigned(usedw0)) then
              usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
            end if;
          end if;
        end process;
        process (clk, rst)
        begin
          if (rst = '1') then
            wrreq <= '0';
            bufsm <= IDLE;
          elsif (rising_edge(clk)) then
            wrreq <= '0';
            case bufsm is
              when IDLE =>
                -- start buffering when see an STP.
                if (stp_at_any_lane_in = '1') then
                  wrreq <= '1';
                  bufsm <= RECV;
                end if;
              when RECV =>
                wrreq <= '1';
                -- stop buffering when see an END with no STP at any higher lane in the same clk.
                if (end_at_any_lane_in = '1' and stp_at_any_lane_in = '0') then
                  bufsm <= IDLE;
                end if;
              when others =>
            end case;
          end if;
        end process;
        -- stop rdreq 1-clk when see an END and an STP in the same clk.
        rdreq <= '1' when empty = '0' and rdack = '1' else '0';
        rdack <= '0' when end_at_any_lane_out = '1' and stp_at_any_lane_out = '1' and rdcnt = '0' else '1';
        process(clk, rst)
        begin
          if (rst = '1') then
            rdcnt <= '0';
          elsif (rising_edge(clk)) then
            if (rdcnt = '0') then
              if (end_at_any_lane_out = '1' and stp_at_any_lane_out = '1') then
                rdcnt <= '1';
              end if;
            else
              rdcnt <= '0';
            end if;
          end if;
        end process;
        process (clk)
          variable stpattmp : std_logic_vector(NLANE/2-1 downto 0);
          variable endattmp : std_logic_vector(NLANE/2-1 downto 0);
          variable setmp : std_logic_vector(NLANE-1 downto 0);
          variable betptmp : std_logic_vector(NLANE*2-1 downto 0);
        begin
          if (rising_edge(clk)) then
            --
            -- prestage
            --
            wetp0 <= '0';
            tptop0 <= '0';
            tpend0 <= '0';
            if (empty = '0') then
              wetp0 <= '1';
              stpattmp := stpat0;
              if (end_at_any_lane_out = '0' or rdcnt = '1') then
                -- assert tptop0 when see an STP with no END,
                -- or see combination of an STP and an END duplicated (i.e., rdcnt=1)
                -- by rxbuf0.
                if (stp_at_any_lane_out = '1') then
                  tptop0 <= '1';
                end if;
                -- modify alignment of TLP top if detect an STP.
                -- use current alignment otherwise.
                if (stp_at_0l_out = '1') then
                  stpattmp := "0001";
                elsif (stp_at_4l_out = '1') then
                  stpattmp := "0010";
                elsif (stp_at_0h_out = '1') then
                  stpattmp := "0100";
                elsif (stp_at_4h_out = '1') then
                  stpattmp := "1000";
                end if;
              end if;
              stpat0 <= stpattmp;
              endattmp := endat0;
              if (end_at_any_lane_out = '1' and rdcnt = '0') then
                -- assert tpend0 when see an END which is not duplicated (i.e., rdcnt=0)
                -- by rxbuf0.
                tpend0 <= '1';
                -- modify alignment of TLP end.
                if (end_at_3l_out = '1') then
                  endattmp := "0001";
                elsif (end_at_7l_out = '1') then
                  endattmp := "0010";
                elsif (end_at_3h_out = '1') then
                  endattmp := "0100";
                elsif (end_at_7h_out = '1') then
                  endattmp := "1000";
                end if;
              end if;
              endat0 <= endattmp;
            end if;
            case stpattmp is
              when "0001" =>
                tp0(119 downto 112) <= d(7)(15 downto 8);
                tp0(111 downto 104) <= d(6)(15 downto 8);
                tp0(103 downto 96) <= d(5)(15 downto 8);
                tp0( 95 downto 88) <= d(4)(15 downto 8);
                tp0( 87 downto 80) <= d(3)(15 downto 8);
                tp0( 79 downto 72) <= d(2)(15 downto 8);
                tp0( 71 downto 64) <= d(1)(15 downto 8);
                tp0( 63 downto 56) <= d(0)(15 downto 8);
                tp0( 55 downto 48) <= d(7)( 7 downto 0);
                tp0( 47 downto 40) <= d(6)( 7 downto 0);
                tp0( 39 downto 32) <= d(5)( 7 downto 0);
                tp0( 31 downto 24) <= d(4)( 7 downto 0);
                tp0( 23 downto 16) <= d(3)( 7 downto 0);
                tp0( 15 downto 8) <= d(2)( 7 downto 0);
                tp0( 7 downto 0) <= d(1)( 7 downto 0);
              when "0010" =>
                tp0( 87 downto 80) <= d(7)(15 downto 8);
                tp0( 79 downto 72) <= d(6)(15 downto 8);
                tp0( 71 downto 64) <= d(5)(15 downto 8);
                tp0( 63 downto 56) <= d(4)(15 downto 8);
                tp0( 55 downto 48) <= d(3)(15 downto 8);
                tp0( 47 downto 40) <= d(2)(15 downto 8);
                tp0( 39 downto 32) <= d(1)(15 downto 8);
                tp0( 31 downto 24) <= d(0)(15 downto 8);
                tp0( 23 downto 16) <= d(7)( 7 downto 0);
                tp0( 15 downto 8) <= d(6)( 7 downto 0);
                tp0( 7 downto 0) <= d(5)( 7 downto 0);
              when "0100" =>
                tp0( 55 downto 48) <= d(7)(15 downto 8);
                tp0( 47 downto 40) <= d(6)(15 downto 8);
                tp0( 39 downto 32) <= d(5)(15 downto 8);
                tp0( 31 downto 24) <= d(4)(15 downto 8);
                tp0( 23 downto 16) <= d(3)(15 downto 8);
                tp0( 15 downto 8) <= d(2)(15 downto 8);
                tp0( 7 downto 0) <= d(1)(15 downto 8);
              when "1000" =>
                tp0( 23 downto 16) <= d(7)(15 downto 8);
                tp0( 15 downto 8) <= d(6)(15 downto 8);
                tp0( 7 downto 0) <= d(5)(15 downto 8);
              when others =>
            end case;
            --
            -- stage 0
            --
            wetp1 <= wetp0;
            tptop1 <= tptop0;
            tpend1 <= tpend0;
            stpat1 <= stpat0;
            endat1 <= endat0;
            case stpat0 is
              when "0001" =>
                tp1(127 downto 120) <= d(0)(7 downto 0);
                tp1(119 downto 0) <= tp0(119 downto 0);
              when "0010" =>
                tp1(127 downto 120) <= d(4)(7 downto 0);
                tp1(119 downto 112) <= d(3)(7 downto 0);
                tp1(111 downto 104) <= d(2)(7 downto 0);
                tp1(103 downto 96) <= d(1)(7 downto 0);
                tp1( 95 downto 88) <= d(0)(7 downto 0);
                tp1( 87 downto 0) <= tp0( 87 downto 0);
              when "0100" =>
                tp1(127 downto 120) <= d(0)(15 downto 8);
                tp1(119 downto 112) <= d(7)(7 downto 0);
                tp1(111 downto 104) <= d(6)(7 downto 0);
                tp1(103 downto 96) <= d(5)(7 downto 0);
                tp1( 95 downto 88) <= d(4)(7 downto 0);
                tp1( 87 downto 80) <= d(3)(7 downto 0);
                tp1( 79 downto 72) <= d(2)(7 downto 0);
                tp1( 71 downto 64) <= d(1)(7 downto 0);
                tp1( 63 downto 56) <= d(0)(7 downto 0);
                tp1( 55 downto 0) <= tp0(55 downto 0);
              when "1000" =>
                tp1(127 downto 120) <= d(4)(15 downto 8);
                tp1(119 downto 112) <= d(3)(15 downto 8);
                tp1(111 downto 104) <= d(2)(15 downto 8);
                tp1(103 downto 96) <= d(1)(15 downto 8);
                tp1( 95 downto 88) <= d(0)(15 downto 8);
                tp1( 87 downto 80) <= d(7)(7 downto 0);
                tp1( 79 downto 72) <= d(6)(7 downto 0);
                tp1( 71 downto 64) <= d(5)(7 downto 0);
                tp1( 63 downto 56) <= d(4)(7 downto 0);
                tp1( 55 downto 48) <= d(3)(7 downto 0);
                tp1( 47 downto 40) <= d(2)(7 downto 0);
                tp1( 39 downto 32) <= d(1)(7 downto 0);
                tp1( 31 downto 24) <= d(0)(7 downto 0);
                tp1( 23 downto 0) <= tp0(23 downto 0);
              when others =>
            end case;
            --
            -- stage 1
            --
            wetp2 <= wetp1;
            tptop2 <= tptop1;
            tp2 <= tp1;
            betptmp := (others => '1');
            if (tpend0 = '1') then
              setmp := stpat0 & endat0;
              case setmp is
                when "0010" & "0001" =>
                  betptmp := x"3fff";
                when "0100" & "0001" =>
                  betptmp := x"03ff";
                when "0100" & "0010" =>
                  betptmp := x"3fff";
                when "1000" & "0001" =>
                  betptmp := x"003f";
                when "1000" & "0010" =>
                  betptmp := x"03ff";
                when "1000" & "0100" =>
                  betptmp := x"3fff";
                when others =>
              end case;
            end if;
            if (tpend1 = '1') then
              setmp := stpat1 & endat1;
              case setmp is
                when "0001" & "0001" =>
                  betptmp := x"0003";
                when "0001" & "0010" =>
                  betptmp := x"003f";
                when "0001" & "0100" =>
                  betptmp := x"03ff";
                when "0001" & "1000" =>
                  betptmp := x"3fff";
                when "0010" & "0010" =>
                  betptmp := x"0003";
                when "0010" & "0100" =>
                  betptmp := x"003f";
                when "0010" & "1000" =>
                  betptmp := x"03ff";
                when "0100" & "0100" =>
                  betptmp := x"0003";
                when "0100" & "1000" =>
                  betptmp := x"003f";
                when "1000" & "1000" =>
                  betptmp := x"0003";
                when others =>
                  wetp2 <= '0';
              end case;
            end if;
            betp2 <= betptmp;
            -- mask tp2
            for i in 0 to NLANE*2-1 loop
              if (betptmp(i) = '0') then
                tp2(8*i+7 downto 8*i+0) <= x"00";
              end if;
            end loop;
          end if; -- clk
        end process;
      end block stpblock;
    end block x8block;
  end generate x8;
end default;
--
-- Framer
-- encode a Data-Link-Layer-Packet (DLLP) or a Transaction-Layer-Packet (TLP)
-- into a Physical-Layer Packet (PLP).
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity framer is
  generic (
    NLANE : integer
  );
  port (
    -- DLLP input
    emptydp : in std_logic;
    rdreqdp : out std_logic;
    dp : in std_logic_vector(47 downto 0);
    -- TLP input
    reqtp : in std_logic;
    acktp : out std_logic;
    toptp : in std_logic;
    wetp : in std_logic;
    betp : in std_logic_vector(NLANE*2-1 downto 0);
    tp : in std_logic_vector(NLANE*16-1 downto 0);
    -- PLP output
    data : out each16b(NLANE-1 downto 0);
    datak : out each2b(NLANE-1 downto 0);
    rst : in std_logic; -- active high
    clk : in std_logic
  );
end framer;
architecture default of framer is
  component count_skp
    port (
      dosend : out std_logic;
      sent : in std_logic;
      rst : in std_logic; -- active high
      clk : in std_logic
    );
  end component;
  signal doskp : std_logic;
  signal sentskp : std_logic;
begin
  x1 : if (NLANE = 1) generate
    x1block : block
      type FRAMER_STATE is (
        IDLE,
        SEND_SKP,
        SEND_SDP_TYPE,
        SEND_D0_D1,
        SEND_D2_C0,
        SEND_C1_END,
        WAIT_TLP,
        SEND_TLP,
        INVALID_STATE
      );
      signal framersm : FRAMER_STATE;
      signal dp1 : std_logic_vector(47 downto 0);
      signal tp1 : std_logic_vector(NLANE*16-1 downto 0);
      signal ntpcoming : std_logic_vector(7 downto 0); -- # of TLPs coming in.
      signal acktp0 : std_logic;
    begin
      acktp <= acktp0;
      -- accept TLP request, whenever there's no request for SKP-OS nor DLLP.
      process (reqtp, doskp, emptydp, framersm)
      begin
        acktp0 <= '0';
        if (reqtp = '1') then
          if (doskp = '0' and emptydp = '1' and
              (framersm = IDLE or framersm = SEND_TLP or framersm = WAIT_TLP)) then
            acktp0 <= '1';
          end if;
        end if;
      end process;
      process (clk, rst)
        variable ntpcomingtmp : std_logic_vector(ntpcoming'range);
      begin
        if (rst = '1') then
          rdreqdp <= '0';
          ntpcoming <= (others => '0');
          framersm <= IDLE;
        elsif (rising_edge(clk)) then
          rdreqdp <= '0';
          tp1 <= tp;
          ntpcomingtmp := ntpcoming;
          -- transmit logical idle, if nothing to do.
          datak(0) <= (others => '0');
          data(0) <= (others => '0');
          -- count up the num of TLP coming in, whenever acktp0 is asserted in the previous clk.
          if (acktp0 = '1') then
            ntpcomingtmp := unsigned(ntpcomingtmp) + 1;
          end if;
          case framersm is
            when IDLE =>
              --
              -- transmission priority:
              -- 1) TLP (which cannot be delayed)
              -- 2) SKP-OS
              -- 3) DLLP
              -- 4) logical idle
              --
              -- TLP is coming in. should not start SKP-OS nor DLLP.
              if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                framersm <= WAIT_TLP;
              elsif (doskp = '1') then -- issue SKP os.
                sentskp <= '1';
                datak(0) <= "11";
                data(0) <= K_SKP & K_COM;
                framersm <= SEND_SKP;
              elsif (emptydp = '0') then -- transmit a DLLP in the queue.
                rdreqdp <= '1';
                framersm <= SEND_SDP_TYPE;
              end if;
            when SEND_SKP =>
              sentskp <= '0';
              datak(0) <= "11";
              data(0) <= K_SKP & K_SKP;
              framersm <= IDLE;
            when SEND_SDP_TYPE =>
              dp1 <= dp; -- latch dp.
              datak(0) <= "01";
              data(0) <= dp(7 downto 0) & K_SDP; -- output type & SDP
              framersm <= SEND_D0_D1;
            when SEND_D0_D1 =>
              datak(0) <= "00";
              data(0) <= dp1(23 downto 8); -- output data1 & data0
              framersm <= SEND_D2_C0;
            when SEND_D2_C0 =>
              datak(0) <= "00";
              data(0) <= dp1(39 downto 24); -- output crc high & data2
              framersm <= SEND_C1_END;
            when SEND_C1_END =>
              datak(0) <= "10";
              data(0) <= K_END & dp1(47 downto 40); -- output END & crc low
              framersm <= IDLE;
            when WAIT_TLP =>
              if (wetp = '1') then
                datak(0) <= "01";
                data(0) <= tp(7 downto 0) & K_STP;
                framersm <= SEND_TLP;
              end if;
            when SEND_TLP =>
              if (wetp = '0') then
                datak(0) <= "10";
                data(0) <= K_END & tp1(15 downto 8);
                framersm <= IDLE;
                -- pass through IDLE state if next TLP is going to arrive.
                if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                  ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                  framersm <= WAIT_TLP;
                end if;
              else
                datak(0) <= "00";
                data(0) <= tp(7 downto 0) & tp1(15 downto 8);
              end if;
            when others =>
              framersm <= IDLE;
          end case;
        end if;
        ntpcoming <= ntpcomingtmp;
      end process;
    end block x1block;
  end generate x1;
  x4 : if (NLANE >= 4) generate
    x4block : block
      type FRAMER_STATE is (
        IDLE,
        SEND_SKP,
        SEND_DP,
        WAIT_TLP,
        SEND_TLP,
        INVALID_STATE
      );
      signal framersm : FRAMER_STATE;
      signal rdreqdp0 : std_logic;
      signal dp1 : std_logic_vector(47 downto 0);
      signal toptp1 : std_logic;
      signal wetp1 : std_logic;
      signal betp1 : std_logic_vector(NLANE*2-1 downto 0);
      signal tp1 : std_logic_vector(NLANE*16-1 downto 0);
      signal tp2 : std_logic_vector(NLANE*16-1 downto 0);
      signal ntpcoming : std_logic_vector(7 downto 0); -- # of TLPs coming in.
      signal acktp0 : std_logic;
    begin
      rdreqdp <= rdreqdp0 and not emptydp;
      acktp <= acktp0;
      -- accept TLP request, whenever there's no request for SKP-OS nor DLLP.
      process (reqtp, doskp, emptydp, framersm)
      begin
        acktp0 <= '0';
        if (reqtp = '1') then
          if (doskp = '0' and emptydp = '1' and
              (framersm = IDLE or framersm = SEND_TLP or framersm = WAIT_TLP)) then -- !!
            acktp0 <= '1';
          end if;
        end if;
      end process;
      process (clk, rst)
        variable ntpcomingtmp : std_logic_vector(ntpcoming'range);
      begin
        if (rst = '1') then
          rdreqdp0 <= '0';
          ntpcoming <= (others => '0');
          framersm <= IDLE;
        elsif (rising_edge(clk)) then
          rdreqdp0 <= '0';
          toptp1 <= toptp;
          wetp1 <= wetp;
          betp1 <= betp;
          tp1 <= tp;
          tp2 <= tp1;
          ntpcomingtmp := ntpcoming;
          -- transmit logical idle, if nothing to do.
          for i in 0 to NLANE-1 loop
            datak(i) <= (others => '0');
            data(i) <= (others => '0');
          end loop;
          -- count up the num of TLP coming in, whenever acktp0 is asserted in the previous clk.
          if (acktp0 = '1') then
            ntpcomingtmp := unsigned(ntpcomingtmp) + 1;
          end if;
          case framersm is
            when IDLE =>
              --
              -- transmission priority:
              -- 1) TLP (which cannot be delayed)
              -- 2) SKP-OS
              -- 3) DLLP
              -- 4) logical idle
              --
              -- TLP is coming in. should not start SKP-OS nor DLLP.
              if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                framersm <= WAIT_TLP;
              elsif (doskp = '1') then -- issue SKP os.
                sentskp <= '1';
                for i in 0 to NLANE-1 loop
                  datak(i) <= "11";
                  data(i) <= K_SKP & K_COM;
                end loop;
                framersm <= SEND_SKP;
              elsif (emptydp = '0') then -- transmit a DLLP in the queue.
                rdreqdp0 <= '1';
                framersm <= SEND_DP;
              end if;
            when SEND_SKP =>
              sentskp <= '0';
              for i in 0 to NLANE-1 loop
                datak(i) <= "11";
                data(i) <= K_SKP & K_SKP;
              end loop;
              framersm <= IDLE;
            when SEND_DP =>
              if (emptydp = '0') then
                if (NLANE = 4) then
                  datak(3)(1) <= '1';
                  datak(0)(0) <= '1';
                  data(3)(15 downto 8) <= K_END;
                  data(2)(15 downto 8) <= dp(47 downto 40); -- crc0
                  data(1)(15 downto 8) <= dp(39 downto 32); -- crc0
                  data(0)(15 downto 8) <= dp(31 downto 24); -- data2
                  data(3)( 7 downto 0) <= dp(23 downto 16); -- data1
                  data(2)( 7 downto 0) <= dp(15 downto 8); -- data0
                  data(1)( 7 downto 0) <= dp( 7 downto 0); -- type
                  data(0)( 7 downto 0) <= K_SDP;
                end if;
                if (NLANE = 8) then
                  datak(7)(0) <= '1';
                  datak(0)(0) <= '1';
                  data(7)( 7 downto 0) <= K_END;
                  data(6)( 7 downto 0) <= dp(47 downto 40); -- crc0
                  data(5)( 7 downto 0) <= dp(39 downto 32); -- crc0
                  data(4)( 7 downto 0) <= dp(31 downto 24); -- data2
                  data(3)( 7 downto 0) <= dp(23 downto 16); -- data1
                  data(2)( 7 downto 0) <= dp(15 downto 8); -- data0
                  data(1)( 7 downto 0) <= dp( 7 downto 0); -- type
                  data(0)( 7 downto 0) <= K_SDP;
                end if;
                rdreqdp0 <= '1';
              else -- no more DLLP remains in the queue.
                framersm <= IDLE;
              end if;
            when WAIT_TLP =>
              if (toptp1 = '1') then -- the 1st word.
                datak(0)(0) <= '1';
                if (NLANE = 4) then
                  data(3)(15 downto 8) <= tp1(6*8+7 downto 6*8+0);
                  data(2)(15 downto 8) <= tp1(5*8+7 downto 5*8+0);
                  data(1)(15 downto 8) <= tp1(4*8+7 downto 4*8+0);
                  data(0)(15 downto 8) <= tp1(3*8+7 downto 3*8+0);
                  data(3)( 7 downto 0) <= tp1(2*8+7 downto 2*8+0);
                  data(2)( 7 downto 0) <= tp1(1*8+7 downto 1*8+0);
                  data(1)( 7 downto 0) <= tp1(0*8+7 downto 0*8+0);
                  data(0)( 7 downto 0) <= K_STP;
                end if;
                if (NLANE = 8) then
                  data(7)(15 downto 8) <= tp1(14*8+7 downto 14*8+0);
                  data(6)(15 downto 8) <= tp1(13*8+7 downto 13*8+0);
                  data(5)(15 downto 8) <= tp1(12*8+7 downto 12*8+0);
                  data(4)(15 downto 8) <= tp1(11*8+7 downto 11*8+0);
                  data(3)(15 downto 8) <= tp1(10*8+7 downto 10*8+0);
                  data(2)(15 downto 8) <= tp1( 9*8+7 downto 9*8+0);
                  data(1)(15 downto 8) <= tp1( 8*8+7 downto 8*8+0);
                  data(0)(15 downto 8) <= tp1( 7*8+7 downto 7*8+0);
                  data(7)( 7 downto 0) <= tp1( 6*8+7 downto 6*8+0);
                  data(6)( 7 downto 0) <= tp1( 5*8+7 downto 5*8+0);
                  data(5)( 7 downto 0) <= tp1( 4*8+7 downto 4*8+0);
                  data(4)( 7 downto 0) <= tp1( 3*8+7 downto 3*8+0);
                  data(3)( 7 downto 0) <= tp1( 2*8+7 downto 2*8+0);
                  data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                  data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                  data(0)( 7 downto 0) <= K_STP;
                end if;
                framersm <= SEND_TLP;
              end if;
            when SEND_TLP =>
              -- the last word.
              if (wetp = '0' or
                  (wetp1 = '1' and toptp = '1')) then
                if (NLANE = 4) then
                  if (betp1(5) = '1') then
                    datak(3)(1) <= '1'; -- put K_END on higher lane3.
                    data(3)(15 downto 8) <= K_END;
                    data(2)(15 downto 8) <= tp1(5*8+7 downto 5*8+0);
                    data(1)(15 downto 8) <= tp1(4*8+7 downto 4*8+0);
                    data(0)(15 downto 8) <= tp1(3*8+7 downto 3*8+0);
                    data(3)( 7 downto 0) <= tp1(2*8+7 downto 2*8+0);
                    data(2)( 7 downto 0) <= tp1(1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1(0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(7*8+7 downto 7*8+0);
                  else
                    datak(3)(0) <= '1'; -- put K_END on lower lane3.
                    data(3)(15 downto 8) <= (others => '0');
                    data(2)(15 downto 8) <= (others => '0');
                    data(1)(15 downto 8) <= (others => '0');
                    data(0)(15 downto 8) <= (others => '0');
                    data(3)( 7 downto 0) <= K_END;
                    data(2)( 7 downto 0) <= tp1(1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1(0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(7*8+7 downto 7*8+0);
                  end if;
                end if;
                if (NLANE = 8) then
                  if (betp1(13) = '1') then
                    datak(7)(1) <= '1'; -- put K_END on higher lane7.
                    data(7)(15 downto 8) <= K_END;
                    data(6)(15 downto 8) <= tp1(13*8+7 downto 13*8+0);
                    data(5)(15 downto 8) <= tp1(12*8+7 downto 12*8+0);
                    data(4)(15 downto 8) <= tp1(11*8+7 downto 11*8+0);
                    data(3)(15 downto 8) <= tp1(10*8+7 downto 10*8+0);
                    data(2)(15 downto 8) <= tp1( 9*8+7 downto 9*8+0);
                    data(1)(15 downto 8) <= tp1( 8*8+7 downto 8*8+0);
                    data(0)(15 downto 8) <= tp1( 7*8+7 downto 7*8+0);
                    data(7)( 7 downto 0) <= tp1( 6*8+7 downto 6*8+0);
                    data(6)( 7 downto 0) <= tp1( 5*8+7 downto 5*8+0);
                    data(5)( 7 downto 0) <= tp1( 4*8+7 downto 4*8+0);
                    data(4)( 7 downto 0) <= tp1( 3*8+7 downto 3*8+0);
                    data(3)( 7 downto 0) <= tp1( 2*8+7 downto 2*8+0);
                    data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(15*8+7 downto 15*8+0);
                  elsif (betp1(9) = '1') then
                    datak(7)(1) <= '1';
                    datak(6)(1) <= '1';
                    datak(5)(1) <= '1';
                    datak(4)(1) <= '1';
                    datak(3)(1) <= '1'; -- put K_END on higher lane3.
                    data(7)(15 downto 8) <= K_PAD;
                    data(6)(15 downto 8) <= K_PAD;
                    data(5)(15 downto 8) <= K_PAD;
                    data(4)(15 downto 8) <= K_PAD;
                    data(3)(15 downto 8) <= K_END;
                    data(2)(15 downto 8) <= tp1( 9*8+7 downto 9*8+0);
                    data(1)(15 downto 8) <= tp1( 8*8+7 downto 8*8+0);
                    data(0)(15 downto 8) <= tp1( 7*8+7 downto 7*8+0);
                    data(7)( 7 downto 0) <= tp1( 6*8+7 downto 6*8+0);
                    data(6)( 7 downto 0) <= tp1( 5*8+7 downto 5*8+0);
                    data(5)( 7 downto 0) <= tp1( 4*8+7 downto 4*8+0);
                    data(4)( 7 downto 0) <= tp1( 3*8+7 downto 3*8+0);
                    data(3)( 7 downto 0) <= tp1( 2*8+7 downto 2*8+0);
                    data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(15*8+7 downto 15*8+0);
                  elsif (betp1(5) = '1') then
                    datak(7)(0) <= '1'; -- put K_END on lower lane7.
                    data(7)( 7 downto 0) <= K_END;
                    data(6)( 7 downto 0) <= tp1( 5*8+7 downto 5*8+0);
                    data(5)( 7 downto 0) <= tp1( 4*8+7 downto 4*8+0);
                    data(4)( 7 downto 0) <= tp1( 3*8+7 downto 3*8+0);
                    data(3)( 7 downto 0) <= tp1( 2*8+7 downto 2*8+0);
                    data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(15*8+7 downto 15*8+0);
                  else
                    datak(7)(0) <= '1';
                    datak(6)(0) <= '1';
                    datak(5)(0) <= '1';
                    datak(4)(0) <= '1';
                    datak(3)(0) <= '1'; -- put K_END on lower lane3.
                    data(7)( 7 downto 0) <= K_PAD;
                    data(6)( 7 downto 0) <= K_PAD;
                    data(5)( 7 downto 0) <= K_PAD;
                    data(4)( 7 downto 0) <= K_PAD;
                    data(3)( 7 downto 0) <= K_END;
                    data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                    data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                    data(0)( 7 downto 0) <= tp2(15*8+7 downto 15*8+0);
                  end if;
                end if;
                framersm <= IDLE;
                -- pass through IDLE state if next TLP is going to arrive.
                if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                  ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                  framersm <= WAIT_TLP;
                end if;
              else
                if (NLANE = 4) then
                  data(3)(15 downto 8) <= tp1(6*8+7 downto 6*8+0);
                  data(2)(15 downto 8) <= tp1(5*8+7 downto 5*8+0);
                  data(1)(15 downto 8) <= tp1(4*8+7 downto 4*8+0);
                  data(0)(15 downto 8) <= tp1(3*8+7 downto 3*8+0);
                  data(3)( 7 downto 0) <= tp1(2*8+7 downto 2*8+0);
                  data(2)( 7 downto 0) <= tp1(1*8+7 downto 1*8+0);
                  data(1)( 7 downto 0) <= tp1(0*8+7 downto 0*8+0);
                  data(0)( 7 downto 0) <= tp2(7*8+7 downto 7*8+0);
                end if;
                if (NLANE = 8) then
                  data(7)(15 downto 8) <= tp1(14*8+7 downto 14*8+0);
                  data(6)(15 downto 8) <= tp1(13*8+7 downto 13*8+0);
                  data(5)(15 downto 8) <= tp1(12*8+7 downto 12*8+0);
                  data(4)(15 downto 8) <= tp1(11*8+7 downto 11*8+0);
                  data(3)(15 downto 8) <= tp1(10*8+7 downto 10*8+0);
                  data(2)(15 downto 8) <= tp1( 9*8+7 downto 9*8+0);
                  data(1)(15 downto 8) <= tp1( 8*8+7 downto 8*8+0);
                  data(0)(15 downto 8) <= tp1( 7*8+7 downto 7*8+0);
                  data(7)( 7 downto 0) <= tp1( 6*8+7 downto 6*8+0);
                  data(6)( 7 downto 0) <= tp1( 5*8+7 downto 5*8+0);
                  data(5)( 7 downto 0) <= tp1( 4*8+7 downto 4*8+0);
                  data(4)( 7 downto 0) <= tp1( 3*8+7 downto 3*8+0);
                  data(3)( 7 downto 0) <= tp1( 2*8+7 downto 2*8+0);
                  data(2)( 7 downto 0) <= tp1( 1*8+7 downto 1*8+0);
                  data(1)( 7 downto 0) <= tp1( 0*8+7 downto 0*8+0);
                  data(0)( 7 downto 0) <= tp2(15*8+7 downto 15*8+0);
                end if;
              end if;
            when others =>
              framersm <= IDLE;
          end case;
        end if;
        ntpcoming <= ntpcomingtmp;
      end process;
    end block x4block;
  end generate x4;
  count_skp0 : count_skp
    port map (
      dosend => doskp,
      sent => sentskp,
      rst => rst,
      clk => clk
    );
end default;
--
-- count the num of SKP ordered sets to be sent.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
entity count_skp is
  port (
    dosend : out std_logic;
    sent : in std_logic;
    rst : in std_logic; -- active high
    clk : in std_logic
  );
end count_skp;
architecture default of count_skp is
  signal nskpleft : std_logic_vector(7 downto 0); -- # of SKP-os to be sent.
  signal timer : std_logic_vector(9 downto 0);
begin
  process (clk, rst)
    variable nskplefttmp : std_logic_vector(7 downto 0);
  begin
    if (rst = '1') then
      nskpleft <= (others => '0');
      timer <= (others => '0');
    elsif (rising_edge(clk)) then
      -- SKP os should be transmitted in every 700clk (1400 symbols, 0x2bc clk).
      nskplefttmp := nskpleft;
      if (timer(9 downto 0) = "1010111100") then
        timer <= (others => '0');
        nskplefttmp := unsigned(nskplefttmp) + 1;
      else
        timer <= unsigned(timer) + 1;
      end if;
      if (sent = '1') then
        if (nskplefttmp /= conv_std_logic_vector(0, nskplefttmp'length)) then
          nskplefttmp := unsigned(nskplefttmp) - 1;
        end if;
      end if;
      nskpleft <= nskplefttmp;
      if (nskplefttmp = conv_std_logic_vector(0, nskplefttmp'length)) then
        dosend <= '0';
      else
        dosend <= '1';
      end if;
    end if;
  end process;
end default;
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
package gpciepkg is
  --
  -- PCI Express-specific constants
  --
  -- flow control credit
  constant MAX_CREDIT_H : integer := 128; -- max credit for header
  constant MAX_CREDIT_D : integer := 2048; -- max credit for data (in 16byte unit)
  constant PAYLOAD_WIDTH : integer := 10; -- payload size width in DW unit (10 bit denotes max possible payload is 4kB).
  -- payload size
  constant MAX_TLP_SIZE : integer := 8192; -- max-possible TLP size in byte unit. rounded up to powers of 2.
  constant OFFSET_WIDTH : integer := 13; -- ceil(log2(MAX_TLP_SIZE))
  -- 10-bit character symbols (K-code)
  constant K_COM : std_logic_vector := "10111100"; -- comma 0xbc
  constant K_STP : std_logic_vector := "11111011"; -- start TLP 0xfb
  constant K_SDP : std_logic_vector := "01011100"; -- start DLLP 0x5c
  constant K_END : std_logic_vector := "11111101"; -- end 0xfd
  constant K_EDB : std_logic_vector := "11111110"; -- end bad 0xfe
  constant K_PAD : std_logic_vector := "11110111"; -- pad 0xf7
  constant K_SKP : std_logic_vector := "00011100"; -- skip 0x1c
  constant K_FTS : std_logic_vector := "00111100"; -- fast training sequence 0x3c
  constant K_IDL : std_logic_vector := "01111100"; -- idle 0x7c
  constant LOGICAL_IDLE : std_logic_vector := "00000000"; -- send 0x00 if no data to send.
  constant TS1ID : std_logic_vector := "01001010"; -- D10.2 0x4a
  constant TS1ID_N : std_logic_vector := "10110101"; -- D21.5 0xb5, TS1ID inverted.
  constant TS2ID : std_logic_vector := "01000101"; -- D5.2 0x45
  constant TS2ID_N : std_logic_vector := "10111010"; -- D26.5 0xba, TS2ID inverted.
  -- DLLP type
  constant DPTYPE_ACK : std_logic_vector := "00000000"; -- acknowledged 0x00
  constant DPTYPE_NAK : std_logic_vector := "00010000"; -- not acknowledged 0x10
  constant DPTYPE_PM_ENTER_L1 : std_logic_vector := "00100000"; -- PM enter L1 0x20
  constant DPTYPE_PM_ENTER_L23 : std_logic_vector := "00100001"; -- PM enter L2 or L3 0x21
  constant DPTYPE_PM_ASR1 : std_logic_vector := "00100011"; -- PM active state request1 0x23
  constant DPTYPE_PM_REQACK : std_logic_vector := "00100100"; -- PM request acknowledge 0x24
  constant DPTYPE_VENDOR : std_logic_vector := "00110000"; -- vendor specific 0x30
  constant DPTYPE_INITFC1_P_VC0 : std_logic_vector := "01000000"; -- init FC1, posted 0x40
  constant DPTYPE_INITFC1_NP_VC0 : std_logic_vector := "01010000"; -- init FC1, non-posted 0x50
  constant DPTYPE_INITFC1_C_VC0 : std_logic_vector := "01100000"; -- init FC1, completion 0x60
  constant DPTYPE_INITFC2_P_VC0 : std_logic_vector := "11000000"; -- init FC2, posted 0xc0
  constant DPTYPE_INITFC2_NP_VC0 : std_logic_vector := "11010000"; -- init FC2, non-posted 0xd0
  constant DPTYPE_INITFC2_C_VC0 : std_logic_vector := "11100000"; -- init FC2, completion 0xe0
  constant DPTYPE_UPDATEFC_P_VC0 : std_logic_vector := "10000000"; -- update FC, posted 0x80
  constant DPTYPE_UPDATEFC_NP_VC0 : std_logic_vector := "10010000"; -- update FC, non-posted 0x90
  constant DPTYPE_UPDATEFC_C_VC0 : std_logic_vector := "10100000"; -- update FC, completion 0xa0
  -- TLP format
  constant TPFMT_12BH_NODATA : std_logic_vector := "00"; -- 12-byte header w/o data.
  constant TPFMT_16BH_NODATA : std_logic_vector := "01"; -- 16-byte header w/o data.
  constant TPFMT_12BH_DATA : std_logic_vector := "10"; -- 12-byte header w/data.
  constant TPFMT_16BH_DATA : std_logic_vector := "11"; -- 16-byte header w/data.
  -- Flow Control Buffer type
  constant TPBUFTYPE_P : std_logic_vector := "00"; -- posted.
  constant TPBUFTYPE_NP : std_logic_vector := "01"; -- non posted.
  constant TPBUFTYPE_C : std_logic_vector := "10"; -- completion.
  -- TLP type
  constant TPTYPE_MRd32 : std_logic_vector := "0000000"; -- 0x00
  constant TPTYPE_MRd64 : std_logic_vector := "0100000"; -- 0x20
  constant TPTYPE_MRdLk32 : std_logic_vector := "0000001";
  constant TPTYPE_MRdLk64 : std_logic_vector := "0100001";
  constant TPTYPE_MWr32 : std_logic_vector := "1000000"; -- 0x40
  constant TPTYPE_MWr64 : std_logic_vector := "1100000"; -- 0x60
  constant TPTYPE_IORd : std_logic_vector := "0000010";
  constant TPTYPE_IOWr : std_logic_vector := "1000010";
  constant TPTYPE_CfgRd0 : std_logic_vector := "0000100"; -- 0x04
  constant TPTYPE_CfgWr0 : std_logic_vector := "1000100"; -- 0x44
  constant TPTYPE_CfgRd1 : std_logic_vector := "0000101";
  constant TPTYPE_CfgWr1 : std_logic_vector := "1000101";
  constant TPTYPE_Msg : std_logic_vector := "0110XXX";
  constant TPTYPE_MsgD : std_logic_vector := "1110XXX"; -- 0x7X
  constant TPTYPE_Cpl : std_logic_vector := "0001010"; -- 0x0a
  constant TPTYPE_CplD : std_logic_vector := "1001010"; -- 0x4a
  constant TPTYPE_CplLk : std_logic_vector := "0001011";
  constant TPTYPE_CplDLk : std_logic_vector := "1001011";
  -- completion status
  constant CPLSTAT_SC : std_logic_vector := "000"; -- successful completion
  constant CPLSTAT_UR : std_logic_vector := "001"; -- unsupported request
  constant CPLSTAT_CRS : std_logic_vector := "010"; -- configuration request retry status
  constant CPLSTAT_CA : std_logic_vector := "100"; -- completer abort
  --
  -- type definitions
  --
  -- LTSSM States
  type LTSSM_STATE is (
    DETECT_QUIET, DETECT_ACTIVE,
    POLLING_ACTIVE, POLLING_COMPLIANCE, POLLING_CONFIGWAIT, POLLING_CONFIGACC, POLLING_SPEED,
    CONFIG_LINKSTART, CONFIG_LINKACC, CONFIG_LANEWAIT, CONFIG_LANEACC, CONFIG_COMPLETE, CONFIG_COMPLETE2, CONFIG_IDLE, CONFIG_IDLE2,
    L0,
    RECOVERY_LOCK, RECOVERY_CONFIG, RECOVERY_CONFIG2, RECOVERY_IDLE, RECOVERY_IDLE2, RECOVERY_SPEED,
    INVALID_STATE
  );
-- attribute enum_encoding: string;
-- attribute enum_encoding of LTSSM_STATE : type is "00000 00001 00010 00011 00100 00101 00110 00111 01000 01001 01010 01011 01100 01101 01110 01111 10000 10001 10010 10011 10100 10101 10110 10111 11000 11001 11010 11011 11100 11101 11110 11111 ";
  -- define arrays of std_logic_vector.
  type each1b is array (natural range <>) of std_logic;
  type each2b is array (natural range <>) of std_logic_vector(1 downto 0);
  type each4b is array (natural range <>) of std_logic_vector(3 downto 0);
  type each7b is array (natural range <>) of std_logic_vector(6 downto 0);
  type each8b is array (natural range <>) of std_logic_vector(7 downto 0);
  type each10b is array (natural range <>) of std_logic_vector(9 downto 0);
  type each12b is array (natural range <>) of std_logic_vector(11 downto 0);
  type each13b is array (natural range <>) of std_logic_vector(12 downto 0);
  type each16b is array (natural range <>) of std_logic_vector(15 downto 0);
  type each32b is array (natural range <>) of std_logic_vector(31 downto 0);
  type each64b is array (natural range <>) of std_logic_vector(63 downto 0);
  --
  -- utility functions
  --
  -- returns log2(ceil(n)), where n is a natural number.
  function mylog2(
    n : integer
  ) return integer;
  -- convert 2-bit byte enable bits to 16-bit mask.
  function be_to_mask16(
    be : in std_logic_vector(1 downto 0)
  )
  return std_logic_vector;
  -- convert 4-bit byte enable bits to 32-bit mask.
  function be_to_mask32(
    be : in std_logic_vector(3 downto 0)
  )
  return std_logic_vector;
  function length_to_credit(
    len : in std_logic_vector(9 downto 0)
  )
  return std_logic_vector;
  function have_enough_credit(
    limit : in std_logic_vector;
    request : in std_logic_vector
  )
  return std_logic;
  function valid_seqno(
    received : in std_logic_vector;
    expected : in std_logic_vector
  )
  return std_logic;
end package;
package body gpciepkg is
  --
  -- a function to return ceil(log2(n)).
  -- n must be a natural number.
  --
  function mylog2(n : in integer) return integer is
    variable i : integer := 0;
  begin
    while (2**i < n) loop
      i := i + 1;
    end loop;
    return i;
  end mylog2;
  --
  -- convert 2-bit byte enable bits to 16-bit mask.
  --
  function be_to_mask16(
    be : in std_logic_vector(1 downto 0) -- byte enable
  )
  return std_logic_vector is
    variable mask : std_logic_vector(15 downto 0);
  begin
    case be is
      when "11" =>
        mask := x"ffff";
      when "10" =>
        mask := x"ff00";
      when "01" =>
        mask := x"00ff";
      when others =>
          mask := (others => '0');
    end case;
    return mask;
  end be_to_mask16;
  --
  -- convert 4-bit byte enable bits to 32-bit mask.
  --
  function be_to_mask32(
    be : in std_logic_vector(3 downto 0) -- byte enable
  )
  return std_logic_vector is
    variable mask : std_logic_vector(31 downto 0);
  begin
    case be is
      when "1111" =>
        mask := x"ffffffff";
      when "1110" =>
        mask := x"ffffff00";
      when "1101" =>
        mask := x"ffff00ff";
      when "1100" =>
        mask := x"ffff0000";
      when "1011" =>
        mask := x"ff00ffff";
      when "1010" =>
        mask := x"ff00ff00";
      when "1001" =>
        mask := x"ff0000ff";
      when "1000" =>
        mask := x"ff000000";
      when "0111" =>
        mask := x"00ffffff";
      when "0110" =>
        mask := x"00ffff00";
      when "0101" =>
        mask := x"00ff00ff";
      when "0100" =>
        mask := x"00ff0000";
      when "0011" =>
        mask := x"0000ffff";
      when "0010" =>
        mask := x"0000ff00";
      when "0001" =>
        mask := x"000000ff";
      when others =>
        mask := (others => '0');
    end case;
    return mask;
  end be_to_mask32;
  --
  -- convert length in DW unit into credit in 16-byte unit.
  -- the result is rounded up to 16-byte boundary.
  -- i.e., credit = ceil(length/4) = (length+3)>>2
  --
  function length_to_credit(
    len : in std_logic_vector(9 downto 0)
  )
  return std_logic_vector is
    variable lenplus3 : std_logic_vector(9 downto 0);
    variable credit : std_logic_vector(11 downto 0);
  begin
    lenplus3 := unsigned(len) + 3;
    credit := conv_std_logic_vector(0, credit'length - lenplus3'length + 2) &
              lenplus3(lenplus3'left downto 2);
    return credit;
  end length_to_credit;
  function have_enough_credit(
    limit : in std_logic_vector;
    request : in std_logic_vector
  )
  return std_logic is
    variable diff : std_logic_vector(limit'range);
    variable enough : std_logic;
  begin
    diff := unsigned(limit) - unsigned(request);
    enough := not diff(diff'left);
    return enough;
  end have_enough_credit;
  function valid_seqno(
    received : in std_logic_vector;
    expected : in std_logic_vector
  )
  return std_logic is
    variable diff : std_logic_vector(expected'range);
    variable larger_than_expected : std_logic;
  begin
    diff := unsigned(expected) - unsigned(received);
    larger_than_expected := diff(diff'left);
    -- received TLP is valid if its seqno is equal to or smaller than
    -- expected seqno.
    return not larger_than_expected;
  end valid_seqno;
end gpciepkg;
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;
package pcrcpkg is
  -- parameters determined by PCIe spec.
  constant CRC_WIDTH : integer := 32;
  constant NELEM : integer := CRC_WIDTH;
  constant NMATRICE : integer := OFFSET_WIDTH;
  -- parameters which may be adjusted to achieve
  -- better trade off between performance & area.
  constant NPIPEMAX : integer := 4; -- set to 4 or larger for x8, 2 or larger for x4.
                                           -- both ArriaGX and StratixII GX can handle 32bit/pipe.
  constant USE_G : integer := 0; -- 0 or 1
  constant USE_H11 : integer := 1; -- 0 or 1
  constant MULT_PER_STAGE : integer := 1; -- fmax on StratixII 1:300MHz 2:180MHz 3:130MHz
  constant ACCUM_STAGE : integer := 7;
  constant FINAL_STAGE : integer := ACCUM_STAGE*(USE_G+1);
  constant NSTAGE : integer := FINAL_STAGE+2; -- prestage, stage0, 1, ... ,FINAL_STAGE
  --
  -- USE_G USE_H11 MULT_PER_STAGE ACCUM_STAGE FINAL_STAGE NSTAGE DELAY(of calc_crc)
  -- 1 0 1 13 26 28 29
  -- 1 0 2 7 14 16 17
  -- 1 0 3 5 10 12 13
  -- 1 0 4 4 8 10 11
  -- 1 1 1 7 14 16 17
  -- 1 1 2 4 8 10 11
  -- 1 1 3 3 6 8 9
  -- 0 0 1 13 13 15 16
  -- 0 0 2 7 7 9 10
  -- 0 0 3 5 5 7 8
  -- 0 0 4 4 4 6 7
  -- 0 1 1 7 7 9 10 <- ArriaGX default setting.
  -- 0 1 2 4 4 6 7 <- StratixII default setting.
  -- 0 1 3 3 3 5 6 <- probably OK with StratixII GX.
  type arrayNPIPExOFFW is array (NPIPEMAX-1 downto 0) of std_logic_vector(OFFSET_WIDTH-1 downto 0);
  type arrayNPIPExCRCW is array (NPIPEMAX-1 downto 0) of std_logic_vector(CRC_WIDTH-1 downto 0);
  type arrayNSTGxCRCW is array (NSTAGE-1 downto 0) of std_logic_vector(CRC_WIDTH-1 downto 0);
  type arrayNSTGxOFFW is array (NSTAGE-1 downto 0) of std_logic_vector(OFFSET_WIDTH-1 downto 0);
  type matrix is array (0 to NELEM-1, NELEM-1 downto 0) of std_logic;
  type matrice is array (0 to NMATRICE-1) of matrix;
  --
  -- powers of H-matrix are listed below.
  -- H-matrix is defined as a 32x32 matrix that converts CRC32(data) to CRC32(data & '0').
  -- you can obtain CRC32 of data trailed by n zeros (CRC32(data & conv_std_logic_vector(0, n)))
  -- by multiplying H to CRC32(data) for n times.
  --
  -- H^2^(i+3) is stored to H_MATRIX(i).
  -- for example, H_MATRIX(0) := H^8,
  -- H_MATRIX(1) := H^16,
  -- H_MATRIX(2) := H^32,
  -- ...
  -- H_MATRIX(12) := H^32k,
  constant H_MATRIX : matrice :=
    (
      -- H8
      (
       x"41000000",
       x"c3000000",
       x"c7000000",
       x"8e000000",
       x"5d000000",
       x"fb000000",
       x"f6000000",
       x"ad000000",
       x"1b000001",
       x"36000002",
       x"2d000004",
       x"1b000008",
       x"77000010",
       x"ee000020",
       x"dc000040",
       x"b8000080",
       x"31000100",
       x"62000200",
       x"c4000400",
       x"88000800",
       x"10001000",
       x"20002000",
       x"01004000",
       x"43008000",
       x"86010000",
       x"0c020000",
       x"59040000",
       x"b2080000",
       x"64100000",
       x"c8200000",
       x"90400000",
       x"20800000"
      ),
      -- H16
      (
       x"16410000",
       x"3ac30000",
       x"63c70000",
       x"c78e0000",
       x"995d0000",
       x"24fb0000",
       x"49f60000",
       x"85ad0000",
       x"1d1b0000",
       x"3a360000",
       x"622d0000",
       x"d21b0000",
       x"b2770000",
       x"64ee0000",
       x"c9dc0000",
       x"93b80000",
       x"31310001",
       x"62620002",
       x"c4c40004",
       x"89880008",
       x"13100010",
       x"26200020",
       x"5a010040",
       x"a2430080",
       x"44860100",
       x"890c0200",
       x"04590400",
       x"08b20800",
       x"11641000",
       x"22c82000",
       x"45904000",
       x"8b208000"
      ),
      -- H32
      (
       x"f7011641",
       x"19033ac3",
       x"c50763c7",
       x"8a0ec78e",
       x"e31c995d",
       x"313824fb",
       x"627049f6",
       x"33e185ad",
       x"90c21d1b",
       x"21843a36",
       x"b409622d",
       x"9f13d21b",
       x"c926b277",
       x"924d64ee",
       x"249ac9dc",
       x"493593b8",
       x"656a3131",
       x"cad46262",
       x"95a8c4c4",
       x"2b518988",
       x"56a31310",
       x"ad462620",
       x"ad8d5a01",
       x"ac1ba243",
       x"58374486",
       x"b06e890c",
       x"97dc0459",
       x"2fb808b2",
       x"5f701164",
       x"bee022c8",
       x"7dc04590",
       x"fb808b20"
      ),
      -- H64
      (
       x"b4e5b025",
       x"dd2ed06e",
       x"0eb810f9",
       x"1d7021f3",
       x"8e05f3c2",
       x"a8ee57a0",
       x"51dcaf40",
       x"175ceea5",
       x"9a5c6d6f",
       x"34b8dadf",
       x"dd94059b",
       x"0fcdbb12",
       x"ab7ec600",
       x"56fd8c01",
       x"adfb1803",
       x"5bf63006",
       x"0309d029",
       x"0613a052",
       x"0c2740a5",
       x"184e814b",
       x"309d0296",
       x"613a052c",
       x"7691ba7c",
       x"59c6c4dc",
       x"b38d89b9",
       x"671b1372",
       x"7ad396c0",
       x"f5a72d81",
       x"eb4e5b02",
       x"d69cb604",
       x"ad396c09",
       x"5a72d812"
      ),
      -- H128
      (
       x"e8f6c5af",
       x"391b4ef0",
       x"9ac0584f",
       x"3580b09e",
       x"83f7a492",
       x"ef198c8a",
       x"de331914",
       x"5490f786",
       x"41d72aa2",
       x"83ae5545",
       x"efaa6f25",
       x"37a21be4",
       x"87b2f267",
       x"0f65e4ce",
       x"1ecbc99d",
       x"3d97933a",
       x"93d9e3da",
       x"27b3c7b4",
       x"4f678f69",
       x"9ecf1ed2",
       x"3d9e3da4",
       x"7b3c7b49",
       x"1e8e333c",
       x"d5eaa3d7",
       x"abd547ae",
       x"57aa8f5c",
       x"47a3db16",
       x"8f47b62d",
       x"1e8f6c5a",
       x"3d1ed8b5",
       x"7a3db16b",
       x"f47b62d7"
      ),
      -- H256
      (
       x"9108245d",
       x"b3186ce7",
       x"f738fd93",
       x"ee71fb26",
       x"4debd211",
       x"0adf807f",
       x"15bf00fe",
       x"ba7625a1",
       x"e5e46f1e",
       x"cbc8de3d",
       x"06999827",
       x"9c3b1412",
       x"a97e0c78",
       x"52fc18f0",
       x"a5f831e1",
       x"4bf063c2",
       x"06e8e3d8",
       x"0dd1c7b1",
       x"1ba38f63",
       x"37471ec7",
       x"6e8e3d8f",
       x"dd1c7b1f",
       x"2b30d262",
       x"c7698099",
       x"8ed30133",
       x"1da60266",
       x"aa442091",
       x"54884122",
       x"a9108245",
       x"5221048b",
       x"a4420917",
       x"4884122e"
      ),
      -- H512
      (
       x"dc70ff4f",
       x"649101d0",
       x"1552fcee",
       x"2aa5f9dc",
       x"893b0cf7",
       x"ce06e6a0",
       x"9c0dcd40",
       x"e46b65ce",
       x"14a634d3",
       x"294c69a7",
       x"8ee82c00",
       x"c1a0a74f",
       x"5f31b1d0",
       x"be6363a0",
       x"7cc6c740",
       x"f98d8e81",
       x"2f6be24c",
       x"5ed7c499",
       x"bdaf8932",
       x"7b5f1264",
       x"f6be24c8",
       x"ed7c4991",
       x"06886c6c",
       x"d1602796",
       x"a2c04f2c",
       x"45809e59",
       x"5771c3fd",
       x"aee387fa",
       x"5dc70ff4",
       x"bb8e1fe9",
       x"771c3fd3",
       x"ee387fa7"
      ),
      -- H1k
      (
       x"2d9b71d5",
       x"76ad927f",
       x"c0c0552a",
       x"8180aa55",
       x"2e9a257e",
       x"70af3b29",
       x"e15e7653",
       x"ef279d73",
       x"f3d44b33",
       x"e7a89666",
       x"e2ca5d19",
       x"e80fcbe7",
       x"fd84e61b",
       x"fb09cc36",
       x"f613986d",
       x"ec2730db",
       x"f5d51063",
       x"ebaa20c7",
       x"d754418f",
       x"aea8831f",
       x"5d51063f",
       x"baa20c7f",
       x"58df692b",
       x"9c25a382",
       x"384b4704",
       x"70968e09",
       x"ccb66dc7",
       x"996cdb8e",
       x"32d9b71d",
       x"65b36e3a",
       x"cb66dc75",
       x"96cdb8ea"
      ),
      -- H2k
      (
       x"b7739aa3",
       x"d994afe5",
       x"045ac569",
       x"08b58ad2",
       x"a6188f07",
       x"fb4284ad",
       x"f685095a",
       x"5a798816",
       x"03808a8e",
       x"0701151d",
       x"b971b098",
       x"c590fb92",
       x"3c526d86",
       x"78a4db0d",
       x"f149b61a",
       x"e2936c34",
       x"725542ca",
       x"e4aa8595",
       x"c9550b2b",
       x"92aa1657",
       x"25542caf",
       x"4aa8595f",
       x"2223281d",
       x"f335ca99",
       x"e66b9532",
       x"ccd72a64",
       x"2eddce6a",
       x"5dbb9cd5",
       x"bb7739aa",
       x"76ee7354",
       x"eddce6a8",
       x"dbb9cd51"
      ),
      -- H4k
      (
       x"1aedbae1",
       x"2f36cf22",
       x"448024a4",
       x"89004948",
       x"08ed2871",
       x"0b37ea03",
       x"166fd407",
       x"363212ee",
       x"76899f3c",
       x"ed133e79",
       x"c0cbc613",
       x"9b7a36c7",
       x"2c19d76f",
       x"5833aedf",
       x"b0675dbf",
       x"60cebb7e",
       x"db70cc1d",
       x"b6e1983a",
       x"6dc33075",
       x"db8660ea",
       x"b70cc1d4",
       x"6e1983a8",
       x"c6debdb0",
       x"9750c181",
       x"2ea18302",
       x"5d430605",
       x"a06bb6eb",
       x"40d76dd7",
       x"81aedbae",
       x"035db75c",
       x"06bb6eb8",
       x"0d76dd70"
      ),
      -- H8k
      (
       x"9d928f1c",
       x"a6b79125",
       x"d0fdad57",
       x"a1fb5aae",
       x"de643a40",
       x"215afb9d",
       x"42b5f73a",
       x"18f96168",
       x"ac604dcc",
       x"58c09b98",
       x"2c13b82d",
       x"c5b5ff46",
       x"16f97190",
       x"2df2e321",
       x"5be5c643",
       x"b7cb8c87",
       x"f2059613",
       x"e40b2c26",
       x"c816584c",
       x"902cb098",
       x"20596130",
       x"40b2c260",
       x"1cf70bdc",
       x"a47c98a4",
       x"48f93148",
       x"91f26290",
       x"be764a3c",
       x"7cec9478",
       x"f9d928f1",
       x"f3b251e3",
       x"e764a3c7",
       x"cec9478e"
      ),
      -- H16k
      (
       x"d8f285c0",
       x"69178e41",
       x"0add9942",
       x"15bb3284",
       x"f384e0c9",
       x"3ffb4453",
       x"7ff688a6",
       x"271f948d",
       x"96cdacda",
       x"2d9b59b5",
       x"83c436aa",
       x"df7ae894",
       x"660754e8",
       x"cc0ea9d1",
       x"981d53a3",
       x"303aa747",
       x"b887cb4f",
       x"710f969e",
       x"e21f2d3d",
       x"c43e5a7b",
       x"887cb4f7",
       x"10f969ee",
       x"f900561d",
       x"2af229fa",
       x"55e453f5",
       x"abc8a7eb",
       x"8f63ca17",
       x"1ec7942e",
       x"3d8f285c",
       x"7b1e50b8",
       x"f63ca170",
       x"ec7942e0"
      ),
      -- H32k
      (
       x"7eaf4a9e",
       x"83f1dfa2",
       x"794cf5da",
       x"f299ebb4",
       x"9b9c9df7",
       x"49967170",
       x"932ce2e0",
       x"58f68f5f",
       x"cf425421",
       x"9e84a843",
       x"43a61a19",
       x"f9e37ead",
       x"8d69b7c5",
       x"1ad36f8b",
       x"35a6df17",
       x"6b4dbe2e",
       x"a83436c2",
       x"50686d85",
       x"a0d0db0a",
       x"41a1b615",
       x"83436c2a",
       x"0686d855",
       x"73a2fa34",
       x"99eabef6",
       x"33d57ded",
       x"67aafbda",
       x"b1fabd2a",
       x"63f57a54",
       x"c7eaf4a9",
       x"8fd5e953",
       x"1fabd2a7",
       x"3f57a54f"
      )
    );
  --
  -- powers of G-matrix are listed below.
  -- G-matrix is an inverse of H-matrix.
  -- you can obtain CRC32(data) from CRC32 of data trailed by n zeros
  -- (CRC32(data & conv_std_logic_vector(0, n))) by multiplying G to
  -- CRC32(data) for n times.
  --
  -- G^2^(i+3) is stored to G_MATRIX(i).
  -- for example, G_MATRIX(0) := G^8,
  -- G_MATRIX(1) := G^16,
  -- G_MATRIX(2) := G^32,
  -- ...
  -- G_MATRIX(12) := G^32k,
  constant G_MATRIX : matrice :=
    (
      -- G8
      (
       x"000001aa",
       x"000002ff",
       x"00000455",
       x"000008aa",
       x"000010fe",
       x"00002057",
       x"000040ae",
       x"000080f7",
       x"00010044",
       x"00020089",
       x"000400b9",
       x"000800d8",
       x"0010001a",
       x"00200035",
       x"0040006b",
       x"008000d7",
       x"01000005",
       x"0200000b",
       x"04000016",
       x"0800002c",
       x"10000059",
       x"200000b2",
       x"400000cf",
       x"80000035",
       x"0000006b",
       x"000000d6",
       x"00000006",
       x"0000000d",
       x"0000001a",
       x"00000035",
       x"0000006a",
       x"000000d5"
      ),
      -- G16
      (
       x"0001aab1",
       x"0002ffd3",
       x"00045516",
       x"0008aa2d",
       x"0010feea",
       x"00205765",
       x"0040aecb",
       x"0080f727",
       x"010044fe",
       x"020089fc",
       x"0400b948",
       x"0800d821",
       x"10001af2",
       x"200035e4",
       x"40006bc9",
       x"8000d792",
       x"00000594",
       x"00000b29",
       x"00001652",
       x"00002ca5",
       x"0000594a",
       x"0000b294",
       x"0000cf99",
       x"00003583",
       x"00006b06",
       x"0000d60d",
       x"000006aa",
       x"00000d55",
       x"00001aab",
       x"00003556",
       x"00006aac",
       x"0000d558"
      ),
      -- G32
      (
       x"aab14226",
       x"ffd3c66b",
       x"5516cef0",
       x"aa2d9de1",
       x"feea79e5",
       x"5765b1ec",
       x"aecb63d9",
       x"f7278595",
       x"44fe490c",
       x"89fc9218",
       x"b9486617",
       x"d8218e09",
       x"1af25e34",
       x"35e4bc69",
       x"6bc978d2",
       x"d792f1a5",
       x"0594a16d",
       x"0b2942da",
       x"165285b4",
       x"2ca50b68",
       x"594a16d1",
       x"b2942da3",
       x"cf991961",
       x"358370e5",
       x"6b06e1cb",
       x"d60dc397",
       x"06aac508",
       x"0d558a11",
       x"1aab1422",
       x"35562844",
       x"6aac5089",
       x"d558a113"
      ),
      -- G64
      (
       x"ff0a7996",
       x"011e8aba",
       x"fd376ce3",
       x"fa6ed9c6",
       x"0bd7ca1b",
       x"e8a5eda1",
       x"d14bdb43",
       x"5d9dcf10",
       x"4431e7b7",
       x"8863cf6e",
       x"efcde74a",
       x"2091b703",
       x"be291791",
       x"7c522f22",
       x"f8a45e45",
       x"f148bc8b",
       x"1d9b0081",
       x"3b360102",
       x"766c0205",
       x"ecd8040b",
       x"d9b00816",
       x"b360102d",
       x"99ca59cc",
       x"cc9eca0e",
       x"993d941c",
       x"327b2838",
       x"9bfc29e6",
       x"37f853cc",
       x"6ff0a799",
       x"dfe14f32",
       x"bfc29e65",
       x"7f853ccb"
      ),
      -- G128
      (
       x"54e65c12",
       x"fd2ae436",
       x"aeb3947e",
       x"5d6728fd",
       x"ee280de8",
       x"88b647c2",
       x"116c8f85",
       x"763f4318",
       x"b898da23",
       x"7131b446",
       x"b685349f",
       x"39ec352c",
       x"273e364a",
       x"4e7c6c95",
       x"9cf8d92a",
       x"39f1b255",
       x"270538b9",
       x"4e0a7172",
       x"9c14e2e4",
       x"3829c5c9",
       x"70538b93",
       x"e0a71726",
       x"95a8725f",
       x"7fb6b8ac",
       x"ff6d7158",
       x"fedae2b1",
       x"a9539970",
       x"52a732e0",
       x"a54e65c1",
       x"4a9ccb82",
       x"95399704",
       x"2a732e09"
      ),
      -- G256
      (
       x"662e22fb",
       x"aa72670c",
       x"32caece2",
       x"6595d9c4",
       x"ad059173",
       x"3c25001d",
       x"784a003a",
       x"96ba228f",
       x"4b5a67e5",
       x"96b4cfcb",
       x"4b47bd6d",
       x"f0a15821",
       x"876c92b9",
       x"0ed92572",
       x"1db24ae4",
       x"3b6495c9",
       x"10e70968",
       x"21ce12d1",
       x"439c25a2",
       x"87384b44",
       x"0e709688",
       x"1ce12d10",
       x"5fec78da",
       x"d9f6d34e",
       x"b3eda69c",
       x"67db4d38",
       x"a998b88b",
       x"53317117",
       x"a662e22f",
       x"4cc5c45f",
       x"998b88be",
       x"3317117d"
      ),
      -- G512
      (
       x"acf22d24",
       x"f516776d",
       x"46dec3fe",
       x"8dbd87fd",
       x"b78922de",
       x"c3e06898",
       x"87c0d130",
       x"a3738f45",
       x"ea1533ae",
       x"d42a675c",
       x"04a6e39d",
       x"a5bfea1f",
       x"e78df91b",
       x"cf1bf236",
       x"9e37e46c",
       x"3c6fc8d9",
       x"d42dbc97",
       x"a85b792f",
       x"50b6f25e",
       x"a16de4bd",
       x"42dbc97b",
       x"85b792f7",
       x"a79d08cb",
       x"e3c83cb2",
       x"c7907964",
       x"8f20f2c8",
       x"b2b3c8b4",
       x"65679169",
       x"cacf22d2",
       x"959e45a4",
       x"2b3c8b49",
       x"56791692"
      ),
      -- G1k
      (
       x"3e787f4d",
       x"428881d6",
       x"bb697ce1",
       x"76d2f9c3",
       x"d3dd8ccb",
       x"99c366da",
       x"3386cdb5",
       x"5975e426",
       x"8c93b700",
       x"19276e00",
       x"0c36a34d",
       x"261539d7",
       x"72520ce2",
       x"e4a419c5",
       x"c948338a",
       x"92906714",
       x"1b58b164",
       x"36b162c8",
       x"6d62c590",
       x"dac58b20",
       x"b58b1640",
       x"6b162c80",
       x"e854264d",
       x"eed033d6",
       x"dda067ac",
       x"bb40cf58",
       x"48f9e1fd",
       x"91f3c3fa",
       x"23e787f4",
       x"47cf0fe9",
       x"8f9e1fd3",
       x"1f3c3fa6"
      ),
      -- G2k
      (
       x"ca02a002",
       x"5e07e006",
       x"760d600f",
       x"ec1ac01e",
       x"1237203f",
       x"ee6ce07c",
       x"dcd9c0f8",
       x"73b121f3",
       x"2d60e3e5",
       x"5ac1c7cb",
       x"7f812f94",
       x"3500ff2a",
       x"a0035e56",
       x"4006bcad",
       x"800d795a",
       x"001af2b4",
       x"ca37456a",
       x"946e8ad5",
       x"28dd15aa",
       x"51ba2b55",
       x"a37456aa",
       x"46e8ad55",
       x"47d3faa9",
       x"45a55550",
       x"8b4aaaa0",
       x"16955541",
       x"e7280a80",
       x"ce501500",
       x"9ca02a00",
       x"39405400",
       x"7280a800",
       x"e5015001"
      ),
      -- G4k
      (
       x"c5319c8a",
       x"4f52a59e",
       x"5b94d7b6",
       x"b729af6d",
       x"ab62c250",
       x"93f4182a",
       x"27e83055",
       x"8ae1fc20",
       x"d0f264cb",
       x"a1e4c997",
       x"86f80fa5",
       x"c8c183c1",
       x"54b29b09",
       x"a9653612",
       x"52ca6c25",
       x"a594d84b",
       x"8e182c1d",
       x"1c30583b",
       x"3860b076",
       x"70c160ec",
       x"e182c1d8",
       x"c30583b0",
       x"433a9bea",
       x"4344ab5f",
       x"868956be",
       x"0d12ad7c",
       x"df14c672",
       x"be298ce4",
       x"7c5319c8",
       x"f8a63391",
       x"f14c6722",
       x"e298ce45"
      ),
      -- G8k
      (
       x"2b536e2b",
       x"7df5b27d",
       x"d0b80ad1",
       x"a17015a2",
       x"69b3456f",
       x"f835e4f5",
       x"f06bc9ea",
       x"cb84fdfe",
       x"bc5a95d7",
       x"78b52baf",
       x"da393974",
       x"9f211cc3",
       x"151157ac",
       x"2a22af58",
       x"54455eb1",
       x"a88abd63",
       x"7a4614ec",
       x"f48c29d8",
       x"e91853b0",
       x"d230a760",
       x"a4614ec1",
       x"48c29d83",
       x"bad6552c",
       x"5effc472",
       x"bdff88e4",
       x"7bff11c9",
       x"dcad4db8",
       x"b95a9b71",
       x"72b536e2",
       x"e56a6dc5",
       x"cad4db8a",
       x"95a9b715"
      ),
      -- G16k
      (
       x"46649bb0",
       x"caadacd0",
       x"d33fc211",
       x"a67f8423",
       x"0a9b93f6",
       x"5353bc5d",
       x"a6a778bb",
       x"0b2a6ac7",
       x"50304e3e",
       x"a0609c7d",
       x"06a5a34a",
       x"4b2fdd25",
       x"d03b21fa",
       x"a07643f5",
       x"40ec87ea",
       x"81d90fd5",
       x"45d6841b",
       x"8bad0837",
       x"175a106f",
       x"2eb420de",
       x"5d6841bd",
       x"bad0837a",
       x"33c59d45",
       x"21efa13b",
       x"43df4277",
       x"87be84ef",
       x"4919926e",
       x"923324dd",
       x"246649bb",
       x"48cc9376",
       x"919926ec",
       x"23324dd8"
      ),
      -- G32k
      (
       x"790e5a18",
       x"8b12ee28",
       x"6f2b8649",
       x"de570c92",
       x"c5a0433c",
       x"f24edc61",
       x"e49db8c2",
       x"b0352b9c",
       x"19640d21",
       x"32c81a43",
       x"1c9e6e9f",
       x"40328726",
       x"f96b5455",
       x"f2d6a8ab",
       x"e5ad5157",
       x"cb5aa2ae",
       x"efbb1f44",
       x"df763e89",
       x"beec7d12",
       x"7dd8fa24",
       x"fbb1f448",
       x"f763e891",
       x"97c98b3b",
       x"569d4c6e",
       x"ad3a98dc",
       x"5a7531b8",
       x"cde43968",
       x"9bc872d0",
       x"3790e5a1",
       x"6f21cb43",
       x"de439686",
       x"bc872d0c"
      )
    );
  -- H^(2^(i+3)+2^(i+4)) is stored to H11_MATRIX(i).
  -- for example, H11_MATRIX(0) := H^24,
  -- H11_MATRIX(1) := H^48,
  -- H11_MATRIX(2) := H^96,
  -- ...
  -- H11_MATRIX(11) := H^48k,
  constant H11_MATRIX : matrice :=
    (
      -- H24
      (
       x"01164100",
       x"033ac300",
       x"0763c700",
       x"0ec78e00",
       x"1c995d00",
       x"3824fb00",
       x"7049f600",
       x"e185ad00",
       x"c21d1b00",
       x"843a3600",
       x"09622d00",
       x"13d21b00",
       x"26b27700",
       x"4d64ee00",
       x"9ac9dc00",
       x"3593b800",
       x"6a313100",
       x"d4626200",
       x"a8c4c400",
       x"51898800",
       x"a3131000",
       x"46262000",
       x"8d5a0100",
       x"1ba24300",
       x"37448601",
       x"6e890c02",
       x"dc045904",
       x"b808b208",
       x"70116410",
       x"e022c820",
       x"c0459040",
       x"808b2080"
      ),
      -- H48
      (
       x"b025f701",
       x"d06e1903",
       x"10f9c507",
       x"21f38a0e",
       x"f3c2e31c",
       x"57a03138",
       x"af406270",
       x"eea533e1",
       x"6d6f90c2",
       x"dadf2184",
       x"059bb409",
       x"bb129f13",
       x"c600c926",
       x"8c01924d",
       x"1803249a",
       x"30064935",
       x"d029656a",
       x"a052cad4",
       x"40a595a8",
       x"814b2b51",
       x"029656a3",
       x"052cad46",
       x"ba7cad8d",
       x"c4dcac1b",
       x"89b95837",
       x"1372b06e",
       x"96c097dc",
       x"2d812fb8",
       x"5b025f70",
       x"b604bee0",
       x"6c097dc0",
       x"d812fb80"
      ),
      -- H96
      (
       x"c0be831e",
       x"41c38523",
       x"43398959",
       x"867312b2",
       x"cc58a67a",
       x"580fcfeb",
       x"b01f9fd7",
       x"a081bcb0",
       x"81bdfa7e",
       x"037bf4fd",
       x"c6496ae4",
       x"4c2c56d7",
       x"58e62eb0",
       x"b1cc5d61",
       x"6398bac2",
       x"c7317585",
       x"4edc6814",
       x"9db8d028",
       x"3b71a050",
       x"76e340a0",
       x"edc68140",
       x"db8d0280",
       x"77a4861e",
       x"2ff78f22",
       x"5fef1e44",
       x"bfde3c89",
       x"bf02fa0c",
       x"7e05f418",
       x"fc0be831",
       x"f817d063",
       x"f02fa0c7",
       x"e05f418f"
      ),
      -- H192
      (
       x"d4c01ec6",
       x"7d40234a",
       x"2e405853",
       x"5c80b0a7",
       x"6dc17f88",
       x"0f42e1d6",
       x"1e85c3ad",
       x"e9cb999c",
       x"07572dff",
       x"0eae5bfe",
       x"c99ca93b",
       x"47f94cb1",
       x"5b3287a4",
       x"b6650f48",
       x"6cca1e91",
       x"d9943d23",
       x"67e86481",
       x"cfd0c902",
       x"9fa19204",
       x"3f432409",
       x"7e864813",
       x"fd0c9027",
       x"2ed93e88",
       x"897263d7",
       x"12e4c7af",
       x"25c98f5e",
       x"9f53007b",
       x"3ea600f6",
       x"7d4c01ec",
       x"fa9803d8",
       x"f53007b1",
       x"ea600f63"
      ),
      -- H384
      (
       x"25534ce2",
       x"6ff5d526",
       x"fab8e6ae",
       x"f571cd5d",
       x"cfb0d658",
       x"ba32e053",
       x"7465c0a6",
       x"cd98cdaf",
       x"be62d7bc",
       x"7cc5af78",
       x"dcd81212",
       x"9ce368c7",
       x"1c959d6c",
       x"392b3ad9",
       x"725675b2",
       x"e4aceb64",
       x"ec0a9a2a",
       x"d8153454",
       x"b02a68a8",
       x"6054d151",
       x"c0a9a2a3",
       x"81534547",
       x"27f5c66c",
       x"6ab8c03a",
       x"d5718074",
       x"aae300e8",
       x"70954d33",
       x"e12a9a67",
       x"c25534ce",
       x"84aa699c",
       x"0954d338",
       x"12a9a671"
      ),
      -- H768
      (
       x"ac5a5971",
       x"f4eeeb93",
       x"45878e57",
       x"8b0f1cae",
       x"ba44602c",
       x"d8d29929",
       x"b1a53252",
       x"cf103dd5",
       x"327a22db",
       x"64f445b6",
       x"65b2d21d",
       x"673ffd4b",
       x"6225a3e6",
       x"c44b47cd",
       x"88968f9a",
       x"112d1f35",
       x"8e00671b",
       x"1c00ce36",
       x"38019c6c",
       x"700338d9",
       x"e00671b2",
       x"c00ce364",
       x"2c439fb9",
       x"f4dd6602",
       x"e9bacc05",
       x"d375980a",
       x"0ab16965",
       x"1562d2cb",
       x"2ac5a597",
       x"558b4b2e",
       x"ab16965c",
       x"562d2cb8"
      ),
      -- H1_5k
      (
       x"7caca516",
       x"85f5ef3b",
       x"77477b61",
       x"ee8ef6c2",
       x"a1b14892",
       x"3fce3432",
       x"7f9c6865",
       x"839475dc",
       x"7b844eaf",
       x"f7089d5e",
       x"92bd9fab",
       x"59d79a41",
       x"cf039194",
       x"9e072329",
       x"3c0e4653",
       x"781c8ca6",
       x"8c95bc5b",
       x"192b78b7",
       x"3256f16e",
       x"64ade2dc",
       x"c95bc5b9",
       x"92b78b72",
       x"59c3b3f3",
       x"cf2bc2f0",
       x"9e5785e0",
       x"3caf0bc1",
       x"05f2b294",
       x"0be56528",
       x"17caca51",
       x"2f9594a2",
       x"5f2b2945",
       x"be56528b"
      ),
      -- H3k
      (
       x"3e255879",
       x"426fe88a",
       x"bafa896d",
       x"75f512db",
       x"d5cf7dce",
       x"95bba3e5",
       x"2b7747cb",
       x"68cbd7ef",
       x"efb2f7a6",
       x"df65ef4c",
       x"80ee86e1",
       x"3ff855ba",
       x"41d5f30d",
       x"83abe61b",
       x"0757cc36",
       x"0eaf986d",
       x"237a68a2",
       x"46f4d144",
       x"8de9a289",
       x"1bd34512",
       x"37a68a25",
       x"6f4d144a",
       x"e0bf70ed",
       x"ff5bb9a3",
       x"feb77346",
       x"fd6ee68c",
       x"c4f89561",
       x"89f12ac3",
       x"13e25587",
       x"27c4ab0f",
       x"4f89561e",
       x"9f12ac3c"
      ),
      -- H6k
      (
       x"1c36c004",
       x"245b400c",
       x"5480401c",
       x"a9008038",
       x"4e37c074",
       x"805940ed",
       x"00b281da",
       x"1d53c3b1",
       x"26914766",
       x"4d228ecd",
       x"8673dd9f",
       x"10d17b3a",
       x"3d943670",
       x"7b286ce0",
       x"f650d9c0",
       x"eca1b381",
       x"c575a706",
       x"8aeb4e0d",
       x"15d69c1b",
       x"2bad3836",
       x"575a706d",
       x"aeb4e0db",
       x"415f01b2",
       x"9e88c360",
       x"3d1186c1",
       x"7a230d82",
       x"e870db00",
       x"d0e1b600",
       x"a1c36c00",
       x"4386d800",
       x"870db001",
       x"0e1b6002"
      ),
      -- H12k
      (
       x"b1f069fd",
       x"d210ba06",
       x"15d11df1",
       x"2ba23be3",
       x"e6b41e3b",
       x"7c98558b",
       x"f930ab17",
       x"43913fd3",
       x"36d2165a",
       x"6da42cb4",
       x"6ab83094",
       x"648008d5",
       x"78f07856",
       x"f1e0f0ac",
       x"e3c1e158",
       x"c783c2b1",
       x"3ef7ec9e",
       x"7defd93c",
       x"fbdfb278",
       x"f7bf64f0",
       x"ef7ec9e1",
       x"defd93c3",
       x"0c0b4e7b",
       x"a9e6f50b",
       x"53cdea16",
       x"a79bd42d",
       x"fec7c1a7",
       x"fd8f834f",
       x"fb1f069f",
       x"f63e0d3f",
       x"ec7c1a7f",
       x"d8f834fe"
      ),
      -- H24k
      (
       x"fd07abfa",
       x"0708fc0f",
       x"f31653e5",
       x"e62ca7ca",
       x"315ee46e",
       x"9fba6326",
       x"3f74c64d",
       x"83ee2761",
       x"fadbe538",
       x"f5b7ca71",
       x"16683f18",
       x"d1d7d5ca",
       x"5ea8006e",
       x"bd5000dd",
       x"7aa001bb",
       x"f5400377",
       x"1787ad15",
       x"2f0f5a2a",
       x"5e1eb455",
       x"bc3d68aa",
       x"787ad154",
       x"f0f5a2a9",
       x"1ceceea8",
       x"c4de76aa",
       x"89bced55",
       x"1379daaa",
       x"dbf41eaf",
       x"b7e83d5f",
       x"6fd07abf",
       x"dfa0f57f",
       x"bf41eafe",
       x"7e83d5fd"
      ),
      -- H48k
      (
       x"9775f984",
       x"b99e0a8d",
       x"e449ec9e",
       x"c893d93d",
       x"06524bfe",
       x"9bd16e79",
       x"37a2dcf2",
       x"f8304061",
       x"67157946",
       x"ce2af28c",
       x"0b201c9d",
       x"8135c0be",
       x"951e78f9",
       x"2a3cf1f3",
       x"5479e3e6",
       x"a8f3c7cd",
       x"c692761e",
       x"8d24ec3d",
       x"1a49d87a",
       x"3493b0f4",
       x"692761e8",
       x"d24ec3d0",
       x"33e87e24",
       x"f0a505cc",
       x"e14a0b98",
       x"c2941731",
       x"125dd7e6",
       x"24bbafcc",
       x"49775f98",
       x"92eebf30",
       x"25dd7e61",
       x"4bbafcc2"
      ),
      -- dummy
      (
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000"
      )
    );
  -- G^(2^(i+3)+2^(i+4)) is stored to G11_MATRIX(i).
  -- for example, G11_MATRIX(0) := G^24,
  -- G11_MATRIX(1) := G^48,
  -- G11_MATRIX(2) := G^96,
  -- ...
  -- G11_MATRIX(11) := G^48k,
  constant G11_MATRIX : matrice :=
    (
      -- G24
      (
       x"01aab142",
       x"02ffd3c6",
       x"045516ce",
       x"08aa2d9d",
       x"10feea79",
       x"205765b1",
       x"40aecb63",
       x"80f72785",
       x"0044fe49",
       x"0089fc92",
       x"00b94866",
       x"00d8218e",
       x"001af25e",
       x"0035e4bc",
       x"006bc978",
       x"00d792f1",
       x"000594a1",
       x"000b2942",
       x"00165285",
       x"002ca50b",
       x"00594a16",
       x"00b2942d",
       x"00cf9919",
       x"00358370",
       x"006b06e1",
       x"00d60dc3",
       x"0006aac5",
       x"000d558a",
       x"001aab14",
       x"00355628",
       x"006aac50",
       x"00d558a1"
      ),
      -- G48
      (
       x"4226ff0a",
       x"c66b011e",
       x"cef0fd37",
       x"9de1fa6e",
       x"79e50bd7",
       x"b1ece8a5",
       x"63d9d14b",
       x"85955d9d",
       x"490c4431",
       x"92188863",
       x"6617efcd",
       x"8e092091",
       x"5e34be29",
       x"bc697c52",
       x"78d2f8a4",
       x"f1a5f148",
       x"a16d1d9b",
       x"42da3b36",
       x"85b4766c",
       x"0b68ecd8",
       x"16d1d9b0",
       x"2da3b360",
       x"196199ca",
       x"70e5cc9e",
       x"e1cb993d",
       x"c397327b",
       x"c5089bfc",
       x"8a1137f8",
       x"14226ff0",
       x"2844dfe1",
       x"5089bfc2",
       x"a1137f85"
      ),
      -- G96
      (
       x"7a79406a",
       x"8e8bc0be",
       x"676ec117",
       x"cedd822f",
       x"e7c24434",
       x"b5fdc803",
       x"6bfb9007",
       x"ad8e6064",
       x"216580a2",
       x"42cb0145",
       x"ffef42e0",
       x"85a7c5ab",
       x"7136cb3c",
       x"e26d9678",
       x"c4db2cf0",
       x"89b659e1",
       x"6915f3a8",
       x"d22be750",
       x"a457cea0",
       x"48af9d41",
       x"915f3a82",
       x"22be7504",
       x"3f05aa63",
       x"047214ad",
       x"08e4295a",
       x"11c852b5",
       x"59e9e501",
       x"b3d3ca03",
       x"67a79406",
       x"cf4f280d",
       x"9e9e501a",
       x"3d3ca035"
      ),
      -- G192
      (
       x"63b8c9ce",
       x"a4c95a52",
       x"2a2a7d6b",
       x"5454fad6",
       x"cb113c62",
       x"f59ab10a",
       x"eb356214",
       x"b5d20de7",
       x"081cd200",
       x"1039a400",
       x"43cb81cf",
       x"e42fca50",
       x"abe75d6f",
       x"57cebade",
       x"af9d75bc",
       x"5f3aeb78",
       x"ddcd1f3e",
       x"bb9a3e7c",
       x"77347cf8",
       x"ee68f9f0",
       x"dcd1f3e1",
       x"b9a3e7c3",
       x"10ff0649",
       x"4246c55d",
       x"848d8aba",
       x"091b1574",
       x"718ee327",
       x"e31dc64e",
       x"c63b8c9c",
       x"8c771939",
       x"18ee3273",
       x"31dc64e7"
      ),
      -- G384
      (
       x"4897a883",
       x"d9b8f984",
       x"fbe65b8a",
       x"f7ccb715",
       x"a70ec6a9",
       x"068a25d1",
       x"0d144ba2",
       x"52bf3fc7",
       x"ede9d70c",
       x"dbd3ae19",
       x"ff30f4b1",
       x"b6f641e1",
       x"257b2b40",
       x"4af65681",
       x"95ecad03",
       x"2bd95a06",
       x"1f251c8e",
       x"3e4a391d",
       x"7c94723a",
       x"f928e474",
       x"f251c8e8",
       x"e4a391d0",
       x"81d08b23",
       x"4b36bec4",
       x"966d7d88",
       x"2cdafb10",
       x"11225ea2",
       x"2244bd44",
       x"44897a88",
       x"8912f510",
       x"1225ea20",
       x"244bd441"
      ),
      -- G768
      (
       x"c0b45547",
       x"41dcffc9",
       x"430daad4",
       x"861b55a9",
       x"cc82fe15",
       x"59b1a96c",
       x"b36352d9",
       x"a672f0f5",
       x"8c51b4ac",
       x"18a36958",
       x"f1f287f6",
       x"23515aab",
       x"8616e011",
       x"0c2dc023",
       x"185b8046",
       x"30b7008c",
       x"a1da545e",
       x"43b4a8bd",
       x"8769517a",
       x"0ed2a2f4",
       x"1da545e9",
       x"3b4a8bd2",
       x"b62142e2",
       x"acf6d082",
       x"59eda104",
       x"b3db4209",
       x"a702d155",
       x"4e05a2aa",
       x"9c0b4554",
       x"38168aa8",
       x"702d1551",
       x"e05a2aa3"
      ),
      -- G1_5k
      (
       x"0fef71bf",
       x"103192c1",
       x"2f8c543d",
       x"5f18a87b",
       x"b1de2149",
       x"6c53332c",
       x"d8a66659",
       x"bea3bd0d",
       x"72a80ba5",
       x"e550174a",
       x"c54f5f2a",
       x"8571cfea",
       x"050cee6a",
       x"0a19dcd5",
       x"1433b9aa",
       x"28677354",
       x"5f219716",
       x"be432e2d",
       x"7c865c5a",
       x"f90cb8b4",
       x"f2197169",
       x"e432e2d3",
       x"c78ab418",
       x"80fa198f",
       x"01f4331e",
       x"03e8663c",
       x"083fbdc6",
       x"107f7b8d",
       x"20fef71b",
       x"41fdee37",
       x"83fbdc6f",
       x"07f7b8df"
      ),
      -- G3k
      (
       x"b81cb118",
       x"c825d328",
       x"28571749",
       x"50ae2e92",
       x"1940ec3d",
       x"8a9d6963",
       x"153ad2c6",
       x"92691494",
       x"9cce9831",
       x"399d3062",
       x"cb26d1dc",
       x"2e5112a0",
       x"e4be9458",
       x"c97d28b1",
       x"92fa5163",
       x"25f4a2c7",
       x"f3f5f496",
       x"e7ebe92d",
       x"cfd7d25a",
       x"9fafa4b5",
       x"3f5f496a",
       x"7ebe92d4",
       x"456194b1",
       x"32df987b",
       x"65bf30f7",
       x"cb7e61ee",
       x"2ee072c4",
       x"5dc0e588",
       x"bb81cb11",
       x"77039623",
       x"ee072c46",
       x"dc0e588c"
      ),
      -- G6k
      (
       x"ff7c9b03",
       x"0185ad04",
       x"fc77c10a",
       x"f8ef8214",
       x"0ea39f2b",
       x"e23ba554",
       x"c4774aa9",
       x"77920e51",
       x"105887a0",
       x"20b10f41",
       x"be1e8581",
       x"83419001",
       x"f9ffbb00",
       x"f3ff7600",
       x"e7feec00",
       x"cffdd801",
       x"60872b00",
       x"c10e5600",
       x"821cac01",
       x"04395802",
       x"0872b005",
       x"10e5600a",
       x"deb65b17",
       x"42102d2d",
       x"84205a5b",
       x"0840b4b7",
       x"effdf26c",
       x"dffbe4d8",
       x"bff7c9b0",
       x"7fef9360",
       x"ffdf26c0",
       x"ffbe4d81"
      ),
      -- G12k
      (
       x"4748fb34",
       x"c9d90d5c",
       x"d4fae18d",
       x"a9f5c31b",
       x"14a37d03",
       x"6e0e0133",
       x"dc1c0266",
       x"ff70fff8",
       x"b9a904c4",
       x"73520989",
       x"a1ece826",
       x"04912b78",
       x"4e6aadc5",
       x"9cd55b8b",
       x"39aab716",
       x"73556e2c",
       x"a1e2276c",
       x"43c44ed9",
       x"87889db2",
       x"0f113b64",
       x"1e2276c8",
       x"3c44ed91",
       x"3fc12017",
       x"38cabb1b",
       x"71957636",
       x"e32aec6c",
       x"811d23ec",
       x"023a47d9",
       x"04748fb3",
       x"08e91f66",
       x"11d23ecd",
       x"23a47d9a"
      ),
      -- G24k
      (
       x"cadd08f4",
       x"5f67191d",
       x"74133acf",
       x"e826759f",
       x"1a91e3ca",
       x"fffecf61",
       x"fffd9ec2",
       x"35263571",
       x"a0916217",
       x"4122c42f",
       x"489880aa",
       x"5bec09a0",
       x"7d051bb5",
       x"fa0a376b",
       x"f4146ed6",
       x"e828ddac",
       x"1a8cb3ad",
       x"3519675a",
       x"6a32ceb4",
       x"d4659d68",
       x"a8cb3ad0",
       x"519675a1",
       x"69f1e3b7",
       x"193ecf9a",
       x"327d9f35",
       x"64fb3e6b",
       x"032b7423",
       x"0656e847",
       x"0cadd08f",
       x"195ba11e",
       x"32b7423d",
       x"656e847a"
      ),
      -- G48k
      (
       x"49e05290",
       x"da20f7b1",
       x"fda1bdf3",
       x"fb437be6",
       x"bf66a55c",
       x"372d1829",
       x"6e5a3053",
       x"95543236",
       x"634836fd",
       x"c6906dfb",
       x"c4c08966",
       x"c061405c",
       x"c922d228",
       x"9245a451",
       x"248b48a2",
       x"49169145",
       x"dbcd701b",
       x"b79ae037",
       x"6f35c06e",
       x"de6b80dc",
       x"bcd701b9",
       x"79ae0372",
       x"babc5475",
       x"3c98fa7b",
       x"7931f4f6",
       x"f263e9ed",
       x"ad27814a",
       x"5a4f0294",
       x"b49e0529",
       x"693c0a52",
       x"d27814a4",
       x"a4f02948"
      ),
      -- dummy
      (
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000",
        x"00000000"
      )
    );
  function calc_lfsr_serial(
    data : in std_logic_vector;
    divisor : in integer;
    crcwidth : in integer
  )
  return std_logic_vector;
  function calc_lfsr_wbe_serial(
    data : in std_logic_vector;
    be : in std_logic_vector;
    divisor : in integer;
    crcwidth : in integer
  )
  return std_logic_vector;
  function mult_m_v(
    msrc : in matrix;
    vsrc : in std_logic_vector(NELEM-1 downto 0)
  )
  return std_logic_vector;
end package;
package body pcrcpkg is
  --
  -- feed data'length input to LFSR and process it with naive algorithm.
  --
  function calc_lfsr_serial(
    data : in std_logic_vector;
    divisor : in integer;
    crcwidth : in integer
  )
  return std_logic_vector is
    variable lfsr : std_logic_vector(crcwidth-1 downto 0);
  begin
      -- initial LFSR value should be all-zero, not all-one.
      lfsr := (others => '0');
      -- proces input bit by bit.
      eachbit : for i in 0 to data'length - 1 loop
        if ((data(i) xor lfsr(lfsr'left)) = '1') then
          -- LFSR is XORed with the divisor
          lfsr := (lfsr(lfsr'left-1 downto 0) & '0') xor
                     conv_std_logic_vector(divisor, lfsr'length);
        else
          -- shift LFSR 1-bit to the left.
          lfsr := (lfsr(lfsr'left-1 downto 0) & '0');
        end if;
      end loop eachbit;
      return lfsr;
  end calc_lfsr_serial;
  --
  -- feed data'length input and its byte enable to LFSR and process it with naive algorithm.
  --
  function calc_lfsr_wbe_serial(
    data : in std_logic_vector;
    be : in std_logic_vector;
    divisor : in integer;
    crcwidth : in integer
  )
  return std_logic_vector is
    variable lfsr : std_logic_vector(crcwidth-1 downto 0);
  begin
      -- initial LFSR value should be all-zero, not all-one.
      lfsr := (others => '0');
      -- proces input bit by bit.
      eachbit : for i in 0 to data'length - 1 loop
        if (be(i/8) = '1') then
          if ((data(i) xor lfsr(lfsr'left)) = '1') then
            -- LFSR is XORed with the divisor
            lfsr := (lfsr(lfsr'left-1 downto 0) & '0') xor
                    conv_std_logic_vector(divisor, lfsr'length);
          else
            -- shift LFSR 1-bit to the left.
            lfsr := (lfsr(lfsr'left-1 downto 0) & '0');
          end if;
        end if;
      end loop eachbit;
      return lfsr;
  end calc_lfsr_wbe_serial;
  function mult_m_v(
    msrc : in matrix;
    vsrc : in std_logic_vector(NELEM-1 downto 0)
  )
  return std_logic_vector is
    variable vdst, vtmpA, vtmpB : std_logic_vector(NELEM-1 downto 0);
  begin
    for r in 0 to NELEM-1 loop
      vtmpA(r) := '0';
      for c in 0 to NELEM/2-1 loop
        if (msrc(r, c) = '1' and vsrc(c) = '1') then
          vtmpA(r) := not vtmpA(r);
        end if;
      end loop;
    end loop;
    for r in 0 to NELEM-1 loop
      vtmpB(r) := '0';
      for c in NELEM/2 to NELEM-1 loop
        if (msrc(r, c) = '1' and vsrc(c) = '1') then
          vtmpB(r) := not vtmpB(r);
        end if;
      end loop;
    end loop;
    for r in 0 to NELEM-1 loop
      vdst(r) := vtmpA(r) xor vtmpB(r);
    end loop;
    return vdst;
  end mult_m_v;
end pcrcpkg;
