--
-- Media Access Control sub layer
-- in Physical layer
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity phymac is
  generic (
    NLANE                  : integer;
    N_FTS                  : integer;
    DATARATE               : std_logic_vector(7 downto 0);
    TSCTRL                 : std_logic_vector(7 downto 0);
    DEFRAMER_TLPBUF_DEPTH  : integer;
    DEFRAMER_DLLPBUF_DEPTH : integer;
    SIMULATION             : integer
  );
  port (
    -- for debugging purpose.
    st2_txd           : out each16b(NLANE-1 downto 0);
    st2_txk           : out each2b(NLANE-1 downto 0);
    st2_rxd           : out each16b(NLANE-1 downto 0);
    st2_rxk           : out each2b(NLANE-1 downto 0);

    clk               : in  std_logic;
    rst               : in  std_logic;   -- active high

    -- PIPE interface
    txelecidleL       : out each1b(NLANE-1 downto 0);
    txdatakL          : out each2b(NLANE-1 downto 0);
    txdataL           : out each16b(NLANE-1 downto 0);
    rxelecidleL       : in  each1b(NLANE-1 downto 0);
    rxdatakL          : in  each2b(NLANE-1 downto 0);
    rxdataL           : in  each16b(NLANE-1 downto 0);
    rxpolarityL       : out each1b(NLANE-1 downto 0);

    -- DLL-MAC interface
    rxwedp            : out std_logic;
    rxdp              : out std_logic_vector(47 downto 0);
    rxwetp            : out std_logic;
    rxbetp            : out std_logic_vector(NLANE*2-1 downto 0);
    rxtptop           : out std_logic;
    rxtp              : out std_logic_vector(NLANE*16-1 downto 0);

    txdp              : in  std_logic_vector(47 downto 0);
    txemptydp         : in  std_logic;
    txrdreqdp         : out std_logic;
    txtp              : in  std_logic_vector(NLANE*16-1 downto 0);
    txtoptp           : in  std_logic;
    txwetp            : in  std_logic;
    txbetp            : in  std_logic_vector(NLANE*2-1 downto 0);

    -- TL-PHYMAC interface
    txreq             : in  std_logic;
    txack             : out std_logic;
    
    dmltssm           : out LTSSM_STATE;
    dmlinknumL        : out each8b(NLANE-1 downto 0);
    dmlanenumL        : out each8b(NLANE-1 downto 0);
    dmdatarateL       : out each8b(NLANE-1 downto 0);
    phy_linkup        : out std_logic;
    dl_linkup         : in  std_logic
  );
end phymac;

architecture default of phymac is

  component txts
    port (
      clk        : in  std_logic;
      rst        : in  std_logic;
      npacket    : in  std_logic_vector(11 downto 0);
      link       : in  std_logic_vector(7 downto 0);
      linkpad    : in  std_logic;
      lane       : in  std_logic_vector(7 downto 0);
      lanepad    : in  std_logic;
      nfts       : in  std_logic_vector(7 downto 0);
      datarate   : in  std_logic_vector(7 downto 0);
      ctrl       : in  std_logic_vector(7 downto 0);
      id         : in  std_logic_vector(7 downto 0);
      kickoff    : in  std_logic;
      txdatak    : out std_logic_vector(1 downto 0);
      txdata     : out std_logic_vector(15 downto 0);
      doskp      : in  std_logic;
      sentskp    : out std_logic;
      done       : out std_logic
    );
  end component;

  component rxts
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      npacket   : out std_logic_vector(11 downto 0);
      link      : out std_logic_vector(7 downto 0);
      linkpad   : out std_logic;
      lane      : out std_logic_vector(7 downto 0);
      lanepad   : out std_logic;
      nfts      : out std_logic_vector(7 downto 0);
      datarate  : out std_logic_vector(7 downto 0);
      ctrl      : out std_logic_vector(7 downto 0);
      id        : out std_logic_vector(7 downto 0);
      clear     : in  std_logic;
      rxdatak   : in  std_logic_vector(1 downto 0);
      rxdata    : in  std_logic_vector(15 downto 0)
    );
  end component;

  component txli
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      npacket   : in  std_logic_vector(11 downto 0);
      kickoff   : in  std_logic;
      txdatak   : out std_logic_vector(1 downto 0);
      txdata    : out std_logic_vector(15 downto 0);
      done      : out std_logic
    );
  end component;

  component rxli
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      npacket   : out std_logic_vector(11 downto 0);
      clear     : in  std_logic;
      rxdatak   : in  std_logic_vector(1 downto 0);
      rxdata    : in  std_logic_vector(15 downto 0)
    );
  end component;

  component scramble
    generic (
      DATA_WIDTH  : integer;
      SIMULATION  : integer := SIMULATION
    );
    port (
      src         : in  std_logic_vector(DATA_WIDTH-1 downto 0);
      iskcode     : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
      doscramble  : in  std_logic;
      dstk        : out std_logic_vector(DATA_WIDTH/8-1 downto 0);
      dst         : out std_logic_vector(DATA_WIDTH-1 downto 0);
      rst         : in  std_logic;     -- active high reset
      clk         : in  std_logic
    );
  end component;

  component deskew
    generic (
      NLANE       : integer := NLANE
    );
    port (
      enable      : in  std_logic;
      srcd        : in  each16b(NLANE-1 downto 0);
      srck        : in  each2b(NLANE-1 downto 0);
      dstd        : out each16b(NLANE-1 downto 0);
      dstk        : out each2b(NLANE-1 downto 0);
      rst         : in  std_logic;        -- active high
      clk         : in  std_logic
      );
  end component;

  component deframer
    generic (
      DEFRAMER_TLPBUF_DEPTH  : integer := DEFRAMER_TLPBUF_DEPTH;
      DEFRAMER_DLLPBUF_DEPTH : integer := DEFRAMER_DLLPBUF_DEPTH;
      NLANE              : integer := NLANE
    );
    port (
-- synthesis translate_off
      -- for debugging
      dpusedwmax  : out std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
      usedwmax    : out std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
-- synthesis translate_on

      -- PLP input
      data        : in  each16b(NLANE-1 downto 0);
      datak       : in  each2b(NLANE-1 downto 0);
      -- DLLP output
      wedp        : out std_logic;
      dp          : out std_logic_vector(47 downto 0);
      -- TLP output
      wetp        : out std_logic;
      tptop       : out std_logic;
      tp          : out std_logic_vector(NLANE*16-1 downto 0);
      betp        : out std_logic_vector(NLANE*2-1 downto 0);

      rst         : in  std_logic;        -- active high
      clk         : in  std_logic
    );
  end component;

  component framer
    generic (
      NLANE       : integer := NLANE
    );
    port (
      -- DLLP input
      emptydp     : in  std_logic;
      rdreqdp     : out std_logic;
      dp          : in  std_logic_vector(47 downto 0);

      -- TLP input
      reqtp       : in  std_logic;
      acktp       : out std_logic;
      toptp       : in  std_logic;
      wetp        : in  std_logic;
      betp        : in  std_logic_vector(NLANE*2-1 downto 0);
      tp          : in  std_logic_vector(NLANE*16-1 downto 0);

      -- PLP output
      data        : out each16b(NLANE-1 downto 0);
      datak       : out each2b(NLANE-1 downto 0);

      rst         : in  std_logic;        -- active high
      clk         : in  std_logic
    );
  end component;

  component count_skp
    port (
      dosend      : out std_logic;
      sent        : in  std_logic;
      rst         : in  std_logic;        -- active high
      clk         : in  std_logic
    );
  end component;

  -- time unit conversion from milli second to UI @ 125MHz (i.e. 8 ns unit)
  constant TIME1H               : std_logic_vector := x"01e848";  -- for actual hardware
  constant TIME2H               : std_logic_vector := x"03d090";  -- for actual hardware
  constant TIME12H              : std_logic_vector := x"16e360";
  constant TIME24H              : std_logic_vector := x"2dc6c0";
  constant TIME48H              : std_logic_vector := x"5b8d80";
  constant TIME1S               : std_logic_vector := x"000040";  -- for simulation
  constant TIME2S               : std_logic_vector := x"000080";
  constant TIME12S              : std_logic_vector := x"0000c0";
  constant TIME24S              : std_logic_vector := x"000180";
  constant TIME48S              : std_logic_vector := x"000300";
  constant TIME16UI             : std_logic_vector := x"000010";
  signal time1                  : std_logic_vector(23 downto 0);
  signal time2                  : std_logic_vector(23 downto 0);
  signal time12                 : std_logic_vector(23 downto 0);
  signal time24                 : std_logic_vector(23 downto 0);
  signal time48                 : std_logic_vector(23 downto 0);

  constant NTS1H                : std_logic_vector := x"001";
  constant NTS2H                : std_logic_vector := x"002";
  constant NTS8H                : std_logic_vector := x"008";
  constant NTS16H               : std_logic_vector := x"010";
  constant NTS1KH               : std_logic_vector := x"400";
  constant NTS1S                : std_logic_vector := x"001";
  constant NTS2S                : std_logic_vector := x"002";
  constant NTS8S                : std_logic_vector := x"002";
  constant NTS16S               : std_logic_vector := x"002";
  constant NTS1KS               : std_logic_vector := x"004";
  signal   nts1                 : std_logic_vector(11 downto 0);
  signal   nts2                 : std_logic_vector(11 downto 0);
  signal   nts8                 : std_logic_vector(11 downto 0);
  signal   nts16                : std_logic_vector(11 downto 0);
  signal   nts1k                : std_logic_vector(11 downto 0);

  -- DLL-MAC interface
  signal ltssm                  : LTSSM_STATE;
  signal rxpolarityL0           : each1b(NLANE-1 downto 0);
  signal datarateL0             : each8b(NLANE-1 downto 0);
  signal nftsL0                 : each8b(NLANE-1 downto 0);
  signal linknumL0              : each8b(NLANE-1 downto 0);
  signal lanenumL0              : each8b(NLANE-1 downto 0);

  -- interface to TxTS
  signal txtskickoffL           : each1b(NLANE-1 downto 0);
  signal txtsnumL               : each12b(NLANE-1 downto 0);
  signal txtslinkL              : each8b(NLANE-1 downto 0);
  signal txtslinkpadL           : each1b(NLANE-1 downto 0);
  signal txtslaneL              : each8b(NLANE-1 downto 0);
  signal txtslanepadL           : each1b(NLANE-1 downto 0);
  signal txtsnftsL              : each8b(NLANE-1 downto 0);
  signal txtsdatarateL          : each8b(NLANE-1 downto 0);
  signal txtsctrlL              : each8b(NLANE-1 downto 0);
  signal txtsidL                : each8b(NLANE-1 downto 0);
  signal txtsdatakL             : each2b(NLANE-1 downto 0);
  signal txtsdataL              : each16b(NLANE-1 downto 0);
  signal txtsdoneL              : each1b(NLANE-1 downto 0);
  signal txtsdone_pulseL        : each1b(NLANE-1 downto 0);

  -- interface to TxLI
  signal txlikickoffL           : each1b(NLANE-1 downto 0);
  signal txlinumL               : each12b(NLANE-1 downto 0);
  signal txlidatakL             : each2b(NLANE-1 downto 0);
  signal txlidataL              : each16b(NLANE-1 downto 0);
  signal txlidoneL              : each1b(NLANE-1 downto 0);
  signal txlidone_pulseL        : each1b(NLANE-1 downto 0);

  -- interface to RxTS
  signal rxtsclearL             : each1b(NLANE-1 downto 0);
  signal rxtsnumL               : each12b(NLANE-1 downto 0);
  signal rxtslinkL              : each8b(NLANE-1 downto 0);
  signal rxtslinkpadL           : each1b(NLANE-1 downto 0);
  signal rxtslaneL              : each8b(NLANE-1 downto 0);
  signal rxtslanepadL           : each1b(NLANE-1 downto 0);
  signal rxtsnftsL              : each8b(NLANE-1 downto 0);
  signal rxtsdatarateL          : each8b(NLANE-1 downto 0);
  signal rxtsctrlL              : each8b(NLANE-1 downto 0);
  signal rxtsidL                : each8b(NLANE-1 downto 0);
  signal rxtsdoneL              : each1b(NLANE-1 downto 0);

  -- interface to RxLI
  signal rxliclearL             : each1b(NLANE-1 downto 0);
  signal rxlinumL               : each12b(NLANE-1 downto 0);
  signal rxlidoneL              : each1b(NLANE-1 downto 0);
  
  -- internal use
  signal prev_ltssm             : LTSSM_STATE;
  signal timer                  : std_logic_vector(23 downto 0);
  signal isos                   : std_logic;
  signal isli                   : std_logic;
  signal sentskpL               : each1b(NLANE-1 downto 0);
  signal doskp                  : std_logic;

  signal txkL                   : each2b(NLANE-1 downto 0);
  signal txdL                   : each16b(NLANE-1 downto 0);
  signal txtoscrkL              : each2b(NLANE-1 downto 0);
  signal txtoscrdL              : each16b(NLANE-1 downto 0);
  
  signal rxkL                   : each2b(NLANE-1 downto 0);  -- descrambled, but not deskewed yet.
  signal rxdL                   : each16b(NLANE-1 downto 0);
  signal rxdeskwkL              : each2b(NLANE-1 downto 0);  -- descrambled and deskewed.
  signal rxdeskwdL              : each16b(NLANE-1 downto 0);

  signal doscrambleL            : each1b(NLANE-1 downto 0);

begin

  -- for debugging purpose
  dmltssm <= ltssm;
  st2 : for i in 0 to NLANE-1 generate
    st2_txd(i) <= txtoscrdL(i);
    st2_txk(i) <= txtoscrkL(i);
    st2_rxd(i) <= rxdeskwdL(i);
    st2_rxk(i) <= rxdeskwkL(i);
  end generate st2;
  
  dmlinknumL <= linknumL0;
  dmlanenumL <= lanenumL0;
  dmdatarateL <= datarateL0;
  rxpolarityL <= rxpolarityL0;
 
  -- Link Training & Sequence State Machine
  process (clk, rst)
    variable txtmp : std_logic;
    variable rxtmp : std_logic;
  begin
    if (rst = '1') then

      phy_linkup  <= '0';

      if (SIMULATION = 0) then
        ltssm <= DETECT_QUIET;
      else
        ltssm <= L0; -- skip training sequence for simulation.
      end if;

    elsif (rising_edge(clk)) then

      prev_ltssm <= ltssm;

      case ltssm is
        --
        -- DETECT Sub State
        --
        when DETECT_QUIET =>

          -- reset all registers.
          isos <= '1';
          isli <= '0';
          for i in 0 to NLANE-1 loop
            txtskickoffL(i) <= '0';
            txtsdoneL(i)    <= '0';
            txlikickoffL(i) <= '0';
            txlidoneL(i)    <= '0';
            rxtsdoneL(i)    <= '0';
            rxtsclearL(i)   <= '1';
            rxlidoneL(i)    <= '0';
            rxliclearL(i)   <= '1';
            rxpolarityL0(i) <= '0';
            lanenumL0(i) <= (others => '1');  -- set an invalid lane#.
          end loop;

          -- down the link.
          phy_linkup  <= '0';
          for i in 0 to NLANE-1 loop
            txelecidleL(i) <= '1';
          end loop;

          -- Rx of all lanes exit electrical idle, or timed out.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxelecidleL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;

          if (rxtmp = '0' or timer = time12) then
            ltssm <= DETECT_ACTIVE;
          end if;

        when DETECT_ACTIVE =>
          -- !!! behavior of this state is not spec compliant !!!
          -- Rx detection is not performed for simplicity.
          -- just wait for 1ms and transition to Polling.Active.
          for i in 0 to NLANE-1 loop
            txelecidleL(i) <= '0';
          end loop;

          if (timer = time1) then
            ltssm <= POLLING_ACTIVE;
          end if;

        --
        -- POLLING Sub State
        --
        when POLLING_ACTIVE =>

          -- send TS1(PAD, PAD) x 1024.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts1k;
              txtslinkpadL(i)  <= '1';
              txtslinkL(i)     <= K_PAD;
              txtslanepadL(i)  <= '1';
              txtslaneL(i)     <= K_PAD;
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS1ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then
              txtsnumL(i)      <= nts1k;  -- send more TS just in case.
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          -- receive TS1/TS2/TS1_D/TS2_D(PAD, PAD) x 8.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i)   <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';  -- sent TS1(PAD, PAD) x 1024
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxtsdoneL(i) <= '1';  -- received TS1/2/1D/2D(PAD, PAD) x 8
              end if;
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          -- all lanes sent 1024 TSs and received 8 TSs.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm  <= POLLING_CONFIGWAIT;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i)    <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i)    <= '0';
              rxtsclearL(i)   <= '1';
            end loop;
          end if;

          if (timer = time24) then
            ltssm <= DETECT_QUIET;
          end if;
          
        when POLLING_CONFIGWAIT =>

          -- send TS2(PAD, PAD) x 16.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '1';
              txtslinkL(i)     <= K_PAD;
              txtslanepadL(i)  <= '1';
              txtslaneL(i)     <= K_PAD;
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          -- receive TS1/2/1D/2D(PAD, PAD) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          -- invert polarity if necessary.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and
                (rxtsidL(i) = TS1ID_N or rxtsidL(i) = TS2ID_N)) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxpolarityL0(i) <= '1';
              end if;
            end if;
          end loop;

          -- save datarate & nfts reveived.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                rxtsdoneL(i)  <= '1';  -- received TS2(PAD, PAD) x 8
                datarateL0(i) <= rxtsdatarateL(i);
                nftsL0(i)     <= rxtsnftsL(i);
              end if;
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';  -- sent TS2(PAD, PAD) x 8
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          -- all lanes sent TS2(PAD, PAD) x 16 and
          -- all lanes received TS2(PAD, PAD) x 8.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm  <= POLLING_CONFIGACC;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i)    <= '0';
              txtskickoffL(i) <= '0';
              rxtsdoneL(i)    <= '0';
              rxtsclearL(i)   <= '1';
            end loop;
          end if;

          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;

        when POLLING_CONFIGACC =>

          -- send TS2(PAD, PAD) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '1';
              txtslinkL(i)     <= K_PAD;
              txtslanepadL(i)  <= '1';
              txtslaneL(i)     <= K_PAD;
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';  -- sent TS2(PAD, PAD) x 16
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;
          
          -- all lanes sent 16 TSs.
          if (txtmp = '1' and rxtmp = '1') then
            rxtmp := '1';
            for i in 0 to NLANE-1 loop
              if (datarateL0(i) /= DATARATE) then
                rxtmp := '0';
              end if;
            end loop;
            if (rxtmp = '1') then       -- all peer lanes are Gen1.
              ltssm  <= CONFIG_LINKSTART;
            else                        -- one or more peer lanes try to communicate at 5.0Gb/s or faster.
              ltssm  <= POLLING_SPEED;
            end if;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
            end loop;
          end if;

          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;

        when POLLING_SPEED =>
          -- stay electrical idle for at least 20ns but no longer than 2ms.
          if (timer = TIME16UI) then
            ltssm <= POLLING_ACTIVE;
            for i in 0 to NLANE-1 loop
              txelecidleL(i) <= '0';
            end loop;
          else
            for i in 0 to NLANE-1 loop
              txelecidleL(i) <= '1';
            end loop;
          end if;

        when CONFIG_LINKSTART =>
          -- send TS1(PAD, PAD) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '1';
              txtslinkL(i)     <= K_PAD;
              txtslanepadL(i)  <= '1';
              txtslaneL(i)     <= K_PAD;
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS1ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';  -- sent TS1(PAD, PAD) x 16
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          -- receive TS1(link#, PAD) x 2
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;
          
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '1') then -- received TS1(link#, PAD) x 2
                rxtsdoneL(i) <= '1';
                linknumL0(i) <= rxtslinkL(i);     -- link# is determined.
              end if;
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          -- all lanes sent TS1(PAD, PAD) x 16 and
          -- all lanes received TS1(link#, PAD) x 2.
          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_LINKACC;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
              rxtsdoneL(i)  <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;
          
          -- received TS1(PAD, PAD) x 2, causing transition to Detect.Quiet.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= DETECT_QUIET;
              end if;
            end if;
          end loop;
          
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;
          
        when CONFIG_LINKACC =>

          -- send TS1(link#, PAD) x 1
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;         -- one TS should be enough. but some chipsets need more.
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- share link# of lane0 among all lanes.
                                                 -- NLANE lanes are configured into one link.
              txtslanepadL(i)  <= '1';
              txtslaneL(i)     <= K_PAD;
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS1ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS1(link#, PAD) x 1
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          if (txtmp = '1') then
            ltssm <= CONFIG_LANEWAIT;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
            end loop;
          end if;

        when CONFIG_LANEWAIT =>

          -- receive TS2(link#, lane#) x 2 or TS1(link#, lane# diff from previous one) x 2
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2) then

              -- received TS2(link#, lane#) x 2
              if (rxtsidL(i) = TS2ID and rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then
                rxtsdoneL(i) <= '1';
                lanenumL0(i) <= rxtslaneL(i);     -- lane# is determined.
              end if;
              -- received TS1(link#, lane# diff from previous one) x 2
              if (rxtsidL(i) = TS1ID and rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then
                if (rxtslaneL(i) /= lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then  -- w/correct link# & unseen lane#.
                  rxtsdoneL(i) <= '1';
                  lanenumL0(i) <= rxtslaneL(i);     -- lane# is modified.
                end if;
              end if;

            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          -- all lanes received TS2(link#, lane#) x 2.
          if (rxtmp = '1') then
            ltssm <= CONFIG_LANEACC;
            for i in 0 to NLANE-1 loop
              rxtsdoneL(i)  <= '0';
              rxtsclearL(i) <= '1';
            end loop;
          end if;

          -- received TS1(PAD, PAD) x 2, causing transition to Detect.Quiet.
          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= DETECT_QUIET;
              end if;
            end if;
          end loop;

          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when CONFIG_LANEACC =>

          -- send TS1(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;         -- one TS should be enough. but some chipsets need more.
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS1ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS1(link#, lane#) x 1
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          -- all lanes sent TS1(link#, lane#) x 16 and
          if (txtmp = '1') then
            ltssm <= CONFIG_COMPLETE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
            end loop;
          end if;

        when CONFIG_COMPLETE =>
          
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS2(link#, lane#) x 16
            end if;
          end loop;

          -- receive TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then  -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then  -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_COMPLETE2;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i)    <= '0';
              rxtsdoneL(i)    <= '0';
              txtskickoffL(i) <= '0';
              rxtsclearL(i)   <= '1';
            end loop;
          end if;

          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when CONFIG_COMPLETE2 =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS2(link#, lane#).
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          if (txtmp = '1') then
            ltssm <= CONFIG_IDLE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i)    <= '0';
              txtskickoffL(i) <= '0';
            end loop;
          end if;

          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when CONFIG_IDLE =>
          isos <= '0';
          isli <= '1';

          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i)  <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then  -- send more logical idle just in case.
              txlinumL(i)      <= nts16;
              txlikickoffL(i)  <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;

          -- reveive logical idle x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxliclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1';            -- sent logical idle x 16
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxliclearL(i) = '0' and rxlinumL(i) >= nts8) then
              rxlidoneL(i) <= '1';            -- received logical idle x 8
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxlidoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= CONFIG_IDLE2;
            for i in 0 to NLANE-1 loop
              txlidoneL(i)    <= '0';
              txlikickoffL(i) <= '0';
              rxlidoneL(i)    <= '0';
              rxliclearL(i)   <= '1';
            end loop;
          end if;

          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when CONFIG_IDLE2 =>

          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i)  <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then  -- send more logical idle just in case.
              txlinumL(i)      <= nts16;
              txlikickoffL(i)  <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1';            -- sent logical idle x 16
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          if (txtmp = '1') then
            ltssm <= L0;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
            end loop;
          end if;

        when L0 =>
          phy_linkup  <= '1';
          isli <= '0';

          -- watch receiver for TS.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          -- any of the lanes received TS, causing transition to Recovery.
          -- LTSSM transitions to Detect.Quiet instead of Recovery state.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxtsnumL(i) >= nts1) then
              rxtmp := '1';
            end if;
          end loop;
          if (rxtmp = '1') then
            ltssm <= RECOVERY_LOCK;
          end if;

          -- any Rx of the lanes gets into electrical idle, causing transition to Detect.Quiet.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxelecidleL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;
          if (rxtmp = '1') then
            ltssm <= DETECT_QUIET;
          end if;

        when RECOVERY_LOCK =>
          isos <= '1';
          isli <= '0';

          -- send TS1(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS1ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS1(link#, lane#).
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          -- receive TS1 or TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and (rxtsidL(i) = TS2ID or rxtsidL(i) = TS1ID)) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then  -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then  -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_CONFIG;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
              rxtsdoneL(i)    <= '0';
              rxtsclearL(i)   <= '1';
            end loop;
          end if;

          -- timed out.
          -- transition to Config.Linkstart if any lane received TS1 or TS2.
          -- transision to Detect.Quiet otherwise.
          rxtmp := '0';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '1') then
              rxtmp := '1';
            end if;
          end loop;

          if (timer = time24) then
            if (rxtmp = '1') then
              ltssm <= CONFIG_LINKSTART;
            else
              ltssm <= DETECT_QUIET;
            end if;
          end if;

        when RECOVERY_CONFIG =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS2(link#, lane#).
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          -- receive TS2(link#, lane#) x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts8 and rxtsidL(i) = TS2ID) then
              if (rxtslinkpadL(i) = '0' and rxtslanepadL(i) = '0') then  -- received TS2(link#, lane#) x 8,
                if (rxtslaneL(i) = lanenumL0(i) and rxtslinkL(i) = linknumL0(i)) then  -- w/correct link# & lane#,
                  rxtsdoneL(i) <= '1';
                end if;
              end if;
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxtsdoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_CONFIG2;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
              rxtsdoneL(i)    <= '0';
              rxtsclearL(i)   <= '1';
            end loop;
          end if;

          -- timed out.
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;

        when RECOVERY_CONFIG2 =>
          -- send TS2(link#, lane#) x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txtsnumL(i)      <= nts16;
              txtslinkpadL(i)  <= '0';
              txtslinkL(i)     <= linknumL0(0);  -- set link# to that of lane0.
              txtslanepadL(i)  <= '0';
              txtslaneL(i)     <= lanenumL0(i);  -- set lane# to that received from the upstream.
              txtsnftsL(i)     <= conv_std_logic_vector(N_FTS, 8);
              txtsdatarateL(i) <= DATARATE;
              txtsctrlL(i)     <= TSCTRL;
              txtsidL(i)       <= TS2ID;
              txtskickoffL(i)  <= '1';
            elsif (txtsdone_pulseL(i) = '1') then  -- send more TS just in case.
              txtsnumL(i)      <= nts1k;
              txtskickoffL(i)  <= '1';
            else 
              txtskickoffL(i)  <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txtsdone_pulseL(i) = '1') then
              txtsdoneL(i) <= '1';   -- sent TS2(link#, lane#).
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txtsdoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          if (txtmp = '1') then
            ltssm <= RECOVERY_IDLE;
            for i in 0 to NLANE-1 loop
              txtsdoneL(i) <= '0';
              txtskickoffL(i)  <= '0';
            end loop;
          end if;

          -- timed out.
          if (timer = time48) then
            ltssm <= DETECT_QUIET;
          end if;

        when RECOVERY_IDLE =>
          isos <= '0';
          isli <= '1';

          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i)  <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then  -- send more logical idle just in case.
              txlinumL(i)      <= nts16;
              txlikickoffL(i)  <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;

          -- reveive logical idle x 8
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxliclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1';            -- sent logical idle x 16
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxliclearL(i) = '0' and rxlinumL(i) >= nts8) then
              rxlidoneL(i) <= '1';            -- received logical idle x 8
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          rxtmp := '1';
          for i in 0 to NLANE-1 loop
            if (rxlidoneL(i) = '0') then
              rxtmp := '0';
            end if;
          end loop;

          if (txtmp = '1' and rxtmp = '1') then
            ltssm <= RECOVERY_IDLE2;
            for i in 0 to NLANE-1 loop
              txlidoneL(i)    <= '0';
              txlikickoffL(i) <= '0';
              rxlidoneL(i)    <= '0';
              rxliclearL(i)   <= '1';
            end loop;
          end if;

          -- received TS1(PAD, PAD) x 2, causing transition to Config.Linkstart.
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              rxtsclearL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (rxtsclearL(i) = '0' and rxtsnumL(i) >= nts2 and rxtsidL(i) = TS1ID) then
              if (rxtslinkpadL(i) = '1' and rxtslanepadL(i) = '1') then
                ltssm <= CONFIG_LINKSTART;
              end if;
            end if;
          end loop;

          -- timed out.
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when RECOVERY_IDLE2 =>
          -- send logical idle x 16
          for i in 0 to NLANE-1 loop
            if (timer = conv_std_logic_vector(0, timer'length)) then
              txlinumL(i)  <= nts16;
              txlikickoffL(i) <= '1';
            elsif (txlidone_pulseL(i) = '1') then  -- send more logical idle just in case.
              txlinumL(i)      <= nts16;
              txlikickoffL(i)  <= '1';
            else
              txlikickoffL(i) <= '0';
            end if;
          end loop;

          for i in 0 to NLANE-1 loop
            if (txlidone_pulseL(i) = '1') then
              txlidoneL(i) <= '1';            -- sent logical idle x 16
            end if;
          end loop;

          txtmp := '1';
          for i in 0 to NLANE-1 loop
            if (txlidoneL(i) = '0') then
              txtmp := '0';
            end if;
          end loop;

          if (txtmp = '1') then
            ltssm <= L0;
            for i in 0 to NLANE-1 loop
              txlidoneL(i) <= '0';
              txlikickoffL(i) <= '0';
            end loop;
          end if;

          -- timed out.
          if (timer = time2) then
            ltssm <= DETECT_QUIET;
          end if;

        when RECOVERY_SPEED =>
          ltssm <= DETECT_QUIET;

        when others =>
          ltssm <= INVALID_STATE;

      end case;
    end if;
  end process;

  -- timeout counter
  -- cleared to 0 when LTSSM transition occured.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (ltssm  /= prev_ltssm) then
        timer <= (others => '0');
      else
        timer <= unsigned(timer) + 1;
      end if;
    end if;
  end process;
    
  -- TS1/TS2 transmitter
  txtslane : for i in 0 to NLANE-1 generate
    u0 : txts
      port map (
        clk        => clk,
        rst        => rst,
        npacket    => txtsnumL(i),
        link       => txtslinkL(i),
        linkpad    => txtslinkpadL(i),
        lane       => txtslaneL(i),
        lanepad    => txtslanepadL(i),
        nfts       => txtsnftsL(i),
        datarate   => txtsdatarateL(i),
        ctrl       => txtsctrlL(i),
        id         => txtsidL(i),
        kickoff    => txtskickoffL(i),
        txdata     => txtsdataL(i),
        txdatak    => txtsdatakL(i),
        doskp      => doskp,
        sentskp    => sentskpL(i),
        done       => txtsdone_pulseL(i)
      );
  end generate txtslane;

  count_skp0 :count_skp
    port map (
      dosend => doskp,
      sent   => sentskpL(0),
      rst    => rst,
      clk    => clk
    );
  
  -- TS1/TS2 receiver
  rxtslane : for i in 0 to NLANE-1 generate
    u0 : rxts
      port map (
        clk       => clk,
        rst       => rst,
        npacket   => rxtsnumL(i),
        link      => rxtslinkL(i),
        linkpad   => rxtslinkpadL(i),
        lane      => rxtslaneL(i),
        lanepad   => rxtslanepadL(i),
        nfts      => rxtsnftsL(i),
        datarate  => rxtsdatarateL(i),
        ctrl      => rxtsctrlL(i),
        id        => rxtsidL(i),
        clear     => rxtsclearL(i),
        rxdatak   => rxdatakL(i),
        rxdata    => rxdataL(i)
      );
  end generate rxtslane;
  
  -- Logical idle transmitter
  txlilane : for i in 0 to NLANE-1 generate
    u0 : txli
      port map (
        clk       => clk,
        rst       => rst,
        npacket   => txlinumL(i),
        kickoff   => txlikickoffL(i),
        txdata    => txlidataL(i),
        txdatak   => txlidatakL(i),
        done      => txlidone_pulseL(i)
      );
  end generate txlilane;

  -- Logical idle reveiver
  rxlilane : for i in 0 to NLANE-1 generate
    u0 : rxli
      port map (
        clk       => clk,
        rst       => rst,
        npacket   => rxlinumL(i),
        clear     => rxliclearL(i),
        rxdatak   => rxkL(i),
        rxdata    => rxdL(i)
      );
  end generate rxlilane;

  --
  -- Tx data selectors
  --

  -- feed one of txdL, txlidataL and txtsdataL to the scrambler.
  txmux0 : for i in 0 to NLANE-1 generate
    process (clk)
    begin
      if (rising_edge(clk)) then
        if (isli = '1') then              -- transmit logical idle.
          doscrambleL(i) <= '1';
          txtoscrkL(i)   <= txlidatakL(i);
          txtoscrdL(i)   <= txlidataL(i);
        elsif (isos = '1') then           -- transmit TS ordered set.
          doscrambleL(i) <= '0';
          txtoscrkL(i)   <= txtsdatakL(i);
          txtoscrdL(i)   <= txtsdataL(i);
        else                              -- transmit TLP or DLLP.
          doscrambleL(i) <= '1';
          txtoscrkL(i)   <= txkL(i);
          txtoscrdL(i)   <= txdL(i);
        end if;
      end if;
    end process;
  end generate txmux0;

  -- Tx scrambler
  txscr0 : for i in 0 to NLANE-1 generate
    u0 : scramble
      generic map (
        DATA_WIDTH => 16
      )
      port map (
        clk        => clk,
        rst        => rst,
        src        => txtoscrdL(i),
        iskcode    => txtoscrkL(i),
        doscramble => doscrambleL(i),
        dstk       => txdatakL(i),
        dst        => txdataL(i)
      );
  end generate txscr0;

  -- Rx descrambler
  rxdescr0 : for i in 0 to NLANE-1 generate
    u0 : scramble
      generic map (
        DATA_WIDTH => 16
      )
      port map (
        clk        => clk,
        rst        => rst,
        src        => rxdataL(i),
        doscramble => '1',
        iskcode    => rxdatakL(i),
        dstk       => rxkL(i),
        dst        => rxdL(i)
      );
  end generate rxdescr0;

  deskew0 : deskew
    port map (
      enable  => isos,
      srcd    => rxdL,
      srck    => rxkL,
      dstd    => rxdeskwdL,
      dstk    => rxdeskwkL,
      rst     => rst,
      clk     => clk
    );

  deframer0 : deframer
    port map (
      data    => rxdeskwdL,
      datak   => rxdeskwkL,

      wedp    => rxwedp,
      dp      => rxdp,

      wetp    => rxwetp,
      betp    => rxbetp,
      tptop   => rxtptop,
      tp      => rxtp,

      rst     => rst,
      clk     => clk
    );

  framer0 : framer
    port map (
      emptydp => txemptydp,
      rdreqdp => txrdreqdp,
      dp      => txdp,

      reqtp   => txreq,
      acktp   => txack,

      toptp   => txtoptp,
      wetp    => txwetp,
      betp    => txbetp,
      tp      => txtp,

      data    => txdL,
      datak   => txkL,

      rst     => rst,
      clk     => clk
    );

  -- constants to which different values are assigned
  -- for simulation and for actual hardware.
  gennosim : if (SIMULATION = 0) generate
    time1   <= TIME1H;
    time2   <= TIME2H;
    time12  <= TIME12H;
    time24  <= TIME24H;
    time48  <= TIME48H;
    nts1    <= NTS1H;
    nts2    <= NTS2H;
    nts8    <= NTS8H;
    nts16   <= NTS16H;
    nts1k   <= NTS1KH;
  end generate gennosim;

  gensim : if (SIMULATION = 1) generate
    time1   <= TIME1S;
    time2   <= TIME2S;
    time12  <= TIME12S;
    time24  <= TIME24S;
    time48  <= TIME48S;
    nts1    <= NTS1S;
    nts2    <= NTS2S;
    nts8    <= NTS8S;
    nts16   <= NTS16S;
    nts1k   <= NTS1KS;
  end generate gensim;
  
end default;

--
-- TS1/TS2 transmitter
--   1) set 'npacket', 'link', 'lane', 'id' when 'kickoff'.
--   2) transmit 'npacket' TSs.
--   3) output 1-clk pulse when 'done'.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity txts is
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;
    npacket    : in  std_logic_vector(11 downto 0);
    link       : in  std_logic_vector(7 downto 0);
    linkpad    : in  std_logic;
    lane       : in  std_logic_vector(7 downto 0);
    lanepad    : in  std_logic;
    nfts       : in  std_logic_vector(7 downto 0);
    datarate   : in  std_logic_vector(7 downto 0);
    ctrl       : in  std_logic_vector(7 downto 0);
    id         : in  std_logic_vector(7 downto 0);
    kickoff    : in  std_logic;
    txdatak    : out std_logic_vector(1 downto 0);
    txdata     : out std_logic_vector(15 downto 0);
    doskp      : in  std_logic;
    sentskp    : out std_logic;
    done       : out std_logic
  );
end txts;

architecture default of txts is

  type TS_STATE is (
    IDLE,
    SEND_COM_LINK,
    SEND_LANE_NFTS,
    SEND_DATARATE_CTRL,
    SEND_TSIDs,
    INVALID_STATE
  );

--  attribute enum_encoding: string;
--  attribute enum_encoding of TS_STATE : type is "000 001 010 011 100 101 110 111";

  signal txtssm       : TS_STATE;
  signal npacket_left : std_logic_vector(11 downto 0);
  signal link0        : std_logic_vector(7 downto 0);
  signal linkpad0     : std_logic;
  signal lane0        : std_logic_vector(7 downto 0);
  signal lanepad0     : std_logic;
  signal nfts0        : std_logic_vector(7 downto 0);
  signal datarate0    : std_logic_vector(7 downto 0);
  signal ctrl0        : std_logic_vector(7 downto 0);
  signal id0          : std_logic_vector(7 downto 0);
  signal nid_left     : std_logic_vector(3 downto 0);
  signal sentskp0     : std_logic;
  
begin

  sentskp <= sentskp0;
  
  process (clk, rst)
  begin
    if (rst = '1') then
      txtssm <= IDLE;
    elsif (rising_edge(clk)) then

      -- set # of TS packets to be sent,
      -- as well as LINK#, LANE#, N_FTS, and TSID.
      -- and then start TS transmission.
      if (kickoff = '1') then
        txtssm <= SEND_COM_LINK;
        npacket_left <= npacket;
        nid_left <= "1010";
        linkpad0 <= linkpad;
        link0 <= link;
        lanepad0 <= lanepad;
        lane0 <= lane;
        nfts0 <= nfts;
        datarate0 <= datarate;
        ctrl0 <= ctrl;
        id0   <= id;
        done <= '0';
        sentskp0 <= '0';
      else
        case txtssm is
          when IDLE =>
            done <= '0';
            if (sentskp0 = '1') then       -- SKP os transmission is in progress.
              sentskp0 <= '0';
              txdatak <= "11";
              txdata  <= K_SKP & K_SKP;
            elsif (doskp = '1') then      -- issue SKP os.
              sentskp0 <= '1';
              txdatak <= "11";
              txdata  <= K_SKP & K_COM;
            elsif (npacket_left /= conv_std_logic_vector(0, npacket_left'length)) then
              nid_left <= "1010";
              txdatak <= linkpad0 & '1';
              txdata <= link0 & K_COM;
              txtssm <= SEND_LANE_NFTS;
            else
              txdatak <= (others => '0');
              txdata  <= (others => '0');
            end if;
          when SEND_COM_LINK =>
            txdatak <= linkpad0 & '1';
            txdata <= link0 & K_COM;
            txtssm <= SEND_LANE_NFTS;
          when SEND_LANE_NFTS =>
            txdatak <= '0' & lanepad0;
            txdata <= nfts0 & lane0;
            txtssm <= SEND_DATARATE_CTRL;
          when SEND_DATARATE_CTRL =>
            txdatak <= "00";
            txdata <= ctrl0 & datarate0;
            txtssm <= SEND_TSIDs;
          when SEND_TSIDs =>
            nid_left <= unsigned(nid_left) - 2;
            if (nid_left = conv_std_logic_vector(2, nid_left'length)) then
              npacket_left <= unsigned(npacket_left) - 1;
              txtssm <= IDLE;
              if (npacket_left = conv_std_logic_vector(1, npacket_left'length)) then
                done <= '1';
              end if;
            else
              txdatak <= "00";
              txdata <= id0 & id0;
            end if;
          when others =>
            txtssm <= IDLE;
        end case;
      end if;                           -- if (kickoff = '1')
    end if;                             -- if (rising_edge(clk))
  end process;

end default;

--
-- TS1/TS2 receiver
--   1) reset packet counter when 'clear'ed.
--   2) output 'npacket', 'link', 'lane', 'id' of the latest
--      TS1/TS2 received if any. 'npacket' is the num of
--      the same packets sequentially obtained.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity rxts is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;
    npacket   : out std_logic_vector(11 downto 0);
    link      : out std_logic_vector(7 downto 0);
    linkpad   : out std_logic;
    lane      : out std_logic_vector(7 downto 0);
    lanepad   : out std_logic;
    nfts      : out std_logic_vector(7 downto 0);
    datarate  : out std_logic_vector(7 downto 0);
    ctrl      : out std_logic_vector(7 downto 0);
    id        : out std_logic_vector(7 downto 0);
    clear     : in  std_logic;
    rxdatak   : in  std_logic_vector(1 downto 0);
    rxdata    : in  std_logic_vector(15 downto 0)
  );
end rxts;

architecture default of rxts is

  type TS_STATE is (
    IDLE,
    RECV_ALIGN,
    RECV_COM_LINK,
    RECV_LANE_NFTS,
    RECV_DATARATE_CTRL,
    RECV_TSID,
    INVALID_STATE
  );

--  attribute enum_encoding: string;
--  attribute enum_encoding of TS_STATE : type is "000 001 010 011 100 101 110 111";

  signal rxtssm       : TS_STATE;
  signal npacket0     : std_logic_vector(11 downto 0);
  signal nid_left     : std_logic_vector(3 downto 0);
  signal rxdatak1     : std_logic_vector(1 downto 0);
  signal rxdata1      : std_logic_vector(15 downto 0);
  signal adatak       : std_logic_vector(1 downto 0);
  signal adata        : std_logic_vector(15 downto 0);
  signal aligned      : std_logic;
  signal samets       : std_logic;
  signal commit       : std_logic;
  
  -- parameters of currently parsing TS.
  -- note that it is uncertain if the receiving data
  -- is a TS or something else (eg. SKP ordered set).
  signal link0        : std_logic_vector(7 downto 0);
  signal linkpad0     : std_logic;
  signal lane0        : std_logic_vector(7 downto 0);
  signal lanepad0     : std_logic;
  signal nfts0        : std_logic_vector(7 downto 0);
  signal datarate0    : std_logic_vector(7 downto 0);
  signal ctrl0        : std_logic_vector(7 downto 0);
  signal id0          : std_logic_vector(7 downto 0);

  -- parameters of recently received TS.
  signal link1        : std_logic_vector(7 downto 0);
  signal linkpad1     : std_logic;
  signal lane1        : std_logic_vector(7 downto 0);
  signal lanepad1     : std_logic;
  signal nfts1        : std_logic_vector(7 downto 0);
  signal datarate1    : std_logic_vector(7 downto 0);
  signal ctrl1        : std_logic_vector(7 downto 0);
  signal id1          : std_logic_vector(7 downto 0);

--  attribute preserve : boolean;
--  attribute preserve of rxdatak1 : signal is true;
--  attribute preserve of rxdata1  : signal is true;

begin

  npacket   <= npacket0;  -- # of same TSs received in sequence.

  -- parameters of recently received TS.
  link      <= link1;
  linkpad   <= linkpad1;
  lane      <= lane1;
  lanepad   <= lanepad1;
  nfts      <= nfts1;
  datarate  <= datarate1;
  ctrl      <= ctrl1;
  id        <= id1;
  
  process (clk, rst)
  begin
    if (rst = '1') then
      rxtssm <= IDLE;
    elsif (rising_edge(clk)) then
      if (clear = '1') then
        link0     <= (others => '0');
        linkpad0  <= '0';
        lane0     <= (others => '0');
        lanepad0  <= '0';
        nfts0     <= (others => '0');
        datarate0 <= (others => '0');
        ctrl0     <= (others => '0');
        id0       <= (others => '0');
        commit    <= '0';
        rxtssm <= IDLE;
      else        
        case rxtssm is
          when IDLE =>

            commit    <= '0';
            nid_left <= "1010";

            -- detect a COM. start parsing TS ordered sets.
            if (rxdatak(0) = '1' and rxdata(7 downto 0) = K_COM) then -- TS is on 16-bit alignment.
              aligned <= '1';
              rxtssm <= RECV_ALIGN;
            elsif (rxdatak(1) = '1' and rxdata(15 downto 8) = K_COM) then -- TS is not on 16-bit alignment.
              aligned <= '0';
              rxtssm <= RECV_ALIGN;
            end if;

          when RECV_ALIGN =>
            rxtssm <= RECV_COM_LINK;    -- wait for 1clk to adata becomes valid.
            
          when RECV_COM_LINK =>
            
            commit <= '0';

            if (adatak(0) = '1') then
              if (adatak(1) = '1') then
                if (adata(15 downto 8) = K_PAD) then
                  linkpad0 <= '1';
                  link0 <= adata(15 downto 8);
                  rxtssm <= RECV_LANE_NFTS;
                else
                  rxtssm <= IDLE;       -- ordered set other than TS coming in.
                end if;
              else
                linkpad0 <= '0';
                link0 <= adata(15 downto 8);
                rxtssm <= RECV_LANE_NFTS;
              end if;
            else
              rxtssm <= IDLE;
            end if;

          when RECV_LANE_NFTS =>
            if (adatak(0) = '1' and adata(7 downto 0) = K_PAD) then
              lanepad0 <= '1';
            else
              lanepad0 <= '0';
            end if;
            lane0 <= adata(7 downto 0);
            nfts0 <= adata(15 downto 8);
            rxtssm <= RECV_DATARATE_CTRL;

          when RECV_DATARATE_CTRL =>
            if (adatak = "00") then
              datarate0 <= adata(7 downto 0);
              ctrl0 <= adata(15 downto 8);
              rxtssm <= RECV_TSID;
            else
              rxtssm <= IDLE;
            end if;

          when RECV_TSID =>
            if (adatak = "00" and
                (adata = TS1ID & TS1ID or
                 adata = TS2ID & TS2ID or
                 adata = TS1ID_N & TS1ID_N or
                 adata = TS2ID_N & TS2ID_N)) then
                 
              id0 <= adata(7 downto 0);
              if (nid_left = conv_std_logic_vector(2, nid_left'length)) then
                commit <= '1';
                nid_left <= "1010";
                rxtssm <= RECV_COM_LINK;
              else
                nid_left <= unsigned(nid_left) - 2;
              end if;
            else
              rxtssm <= IDLE;
            end if;

          when others =>
            rxtssm <= IDLE;
        end case;
      end if;
    end if;
  end process;

  -- when receiving the final byte of a TS (i.e. when commit = 1),
  -- increment npacket0 if currently receiving TS and
  -- the recently received one is the same. set npacket0 to 1
  -- otherwise.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (clear = '1') then

        npacket0  <= (others => '0');
        id1       <= (others => '0');   -- so that next coming TS never match the previous one.

      elsif (commit = '1') then

        link1     <= link0;
        linkpad1  <= linkpad0;
        lane1     <= lane0;
        lanepad1  <= lanepad0;
        nfts1     <= nfts0;
        datarate1 <= datarate0;
        ctrl1     <= ctrl0;
        id1       <= id0;

        if (samets = '1') then
          npacket0 <= unsigned(npacket0) + 1;
        else
          npacket0 <= conv_std_logic_vector(1, npacket0'length);
        end if;

      end if;
    end if;
  end process;

  -- check if current TS and previously received one have
  -- the same parameters.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (link0 = link1 and linkpad0 = linkpad1 and
          lane0 = lane1 and lanepad0 = lanepad1 and
          nfts0 = nfts1 and datarate0 = datarate1 and
          ctrl0 = ctrl1 and id0 = id1) then
        samets <= '1';
      else
        samets <= '0';
      end if;
    end if;
  end process;


  -- shift TS so that it is 16-bit aligned.
  -- adata has 2-clk delay to rxdata.
  process (clk)
  begin
    if (rising_edge(clk)) then
      if (aligned = '1') then
        adatak <= rxdatak1;
        adata  <= rxdata1;
      else
        adatak(1) <= rxdatak(0);
        adatak(0) <= rxdatak1(1);
        adata(15 downto 8) <= rxdata(7 downto 0);
        adata(7 downto 0)  <= rxdata1(15 downto 8);
      end if;
      rxdatak1 <= rxdatak;
      rxdata1 <= rxdata;
    end if;
  end process;

end default;

--
-- Logical idle transmitter
--   1) set 'npacket' when 'kickoff'.
--   2) transmit 'npacket' logical idles.
--   3) output 1-clk pulse when 'done'.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity txli is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;
    npacket   : in  std_logic_vector(11 downto 0);
    kickoff   : in  std_logic;
    txdatak   : out std_logic_vector(1 downto 0);
    txdata    : out std_logic_vector(15 downto 0);
    done      : out std_logic
  );
end txli;

architecture default of txli is

  type LI_STATE is (
    IDLE,
    SEND_LI,
    INVALID_STATE
  );

  signal txlism       : LI_STATE;
  signal npacket_left : std_logic_vector(11 downto 0);

begin

  process (clk, rst)
  begin
    if (rst = '1') then
      txlism <= IDLE;
    elsif (rising_edge(clk)) then
      if (kickoff = '1') then
        txlism <= SEND_LI;
        npacket_left <= npacket;
        done <= '0';
      else
        case txlism is
          when IDLE =>
            done <= '0';
            txdatak <= "00";
            txdata <= LOGICAL_IDLE & LOGICAL_IDLE;
          when SEND_LI =>
            txdatak <= "00";
            txdata <= LOGICAL_IDLE & LOGICAL_IDLE;
            if (npacket_left = conv_std_logic_vector(2, npacket_left'length)) then
              done <= '1';
              txlism <= IDLE;
            else
              npacket_left <= unsigned(npacket_left) - 2;
            end if;
          when others =>
            txlism <= IDLE;
        end case;
      end if;
    end if;
  end process;
end default;


--
-- Logical idle receiver
--   1) reset packet counter when 'clear'ed.
--   2) output 'npacket', the num of LOGICAL_IDLE
--      sequentially obtained.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity rxli is
  port (
    clk       : in  std_logic;
    rst       : in  std_logic;
    npacket   : out std_logic_vector(11 downto 0);
    clear     : in  std_logic;
    rxdatak   : in  std_logic_vector(1 downto 0);
    rxdata    : in  std_logic_vector(15 downto 0)
  );
end rxli;

architecture default of rxli is

  type LI_STATE is (
    IDLE,
    RECV_LI,
    INVALID_STATE
  );

  signal rxlism       : LI_STATE;
  signal npacket0     : std_logic_vector(11 downto 0);
  
begin

  npacket   <= npacket0;
  
  process (clk, rst)
  begin
    if (rst = '1') then
      rxlism <= IDLE;
    elsif (rising_edge(clk)) then
      if (clear = '1') then
        npacket0 <= (others => '0');
        rxlism <= IDLE;
      else
        case rxlism is
          when IDLE =>

            -- detect a pair of logical idle.
            if (rxdatak = "00" and rxdata = LOGICAL_IDLE & LOGICAL_IDLE) then
              npacket0 <= conv_std_logic_vector(2, npacket0'length);
              rxlism <= RECV_LI;
            end if;

          when RECV_LI =>
            if (rxdatak = "00" and rxdata = LOGICAL_IDLE & LOGICAL_IDLE) then
              npacket0 <= unsigned(npacket0) + 2;
            else
              rxlism <= IDLE;
            end if;

          when others =>
            rxlism <= IDLE;
        end case;
      end if;
    end if;
  end process;
end default;

--
-- scrambler/descrambler
-- src to dst pipeline delay : 2clk
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity scramble is
  generic (
    DATA_WIDTH  : integer;
    SIMULATION  : integer
  );
  port (
    src         : in  std_logic_vector(DATA_WIDTH-1 downto 0);
    iskcode     : in  std_logic_vector(DATA_WIDTH/8-1 downto 0);
    doscramble  : in  std_logic;
    dstk        : out std_logic_vector(DATA_WIDTH/8-1 downto 0);
    dst         : out std_logic_vector(DATA_WIDTH-1 downto 0);
    rst         : in  std_logic;     -- active high reset
    clk         : in  std_logic
  );
end scramble;

#if 0 // w/2-clk delay

architecture default of scramble is

  signal   lfsr          : std_logic_vector(15 downto 0); -- linear feedback shift register
  signal   src1          : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal   scrambled1    : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal   iskcode1      : std_logic_vector(DATA_WIDTH/8-1 downto 0);
  signal   doscramble1   : std_logic;
  
  constant MASK          : std_logic_vector := "0000000000111001"; -- X16 + X5 + X4 + X3 + 1

begin

  process (clk, rst)
    variable lfsr_var  : std_logic_vector(15 downto 0);
  begin
    if (rst = '1') then
      lfsr <= x"ffff";
    elsif (rising_edge(clk)) then

      -- update LFSR.
      lfsr_var := lfsr;
      eachbyte : for j in 0 to DATA_WIDTH/8-1 loop
        -- shift LFSR 1-bit to the left if src is not a SKP.
        if (not (iskcode(j) = '1' and src(8 * j + 7 downto 8 * j + 0) = K_SKP)) then
          eachbit : for i in 0 to 7 loop
            -- scramble src with LFSR.
            scrambled1(8 * j + i) <= src(8 * j + i) xor lfsr_var(15);
            if (lfsr_var(15) = '1') then
              lfsr_var := (lfsr_var(14 downto 0) & '0') xor MASK;
            else
              lfsr_var := (lfsr_var(14 downto 0) & '0');
            end if;
          end loop eachbit;
        end if;
        -- reset LFSR when detect a COM.
        if (iskcode(j) = '1' and src(8 * j + 7 downto 8 * j + 0) = K_COM) then
          lfsr_var := x"ffff";
        end if;    
      end loop eachbyte;
      lfsr <= lfsr_var;

    end if;
  end process;

  process (clk)
  begin
    if (rising_edge(clk)) then
      src1 <= src;
      doscramble1 <= doscramble;
    end if;
  end process;

  process (clk)
  begin
    if (rising_edge(clk)) then
      iskcode1 <= iskcode;

      -- do scramble a char if it is not a K-code.
      for j in 0 to DATA_WIDTH/8-1 loop
        dstk(j) <= iskcode1(j);
        if (SIMULATION = 0) then
          if (iskcode1(j) = '1' or doscramble = '0') then
            dst(7 + 8*j downto 0 + 8*j) <= src1(7 + 8*j downto 0 + 8*j);
          else
            dst(7 + 8*j downto 0 + 8*j) <= scrambled1(7 + 8*j downto 0 + 8*j);
          end if;
        else
          -- no scrambing (only for debugging purpose).
          dst(7 + 8*j downto 0 + 8*j) <= src1(7 + 8*j downto 0 + 8*j);
        end if;
      end loop;
    end if;
  end process;
  
end default;

#else // w/3-clk delay

architecture default of scramble is

  signal   lfsr          : std_logic_vector(15 downto 0); -- linear feedback shift register
  signal   src1          : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal   src2          : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal   scrambled2    : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal   iskcode1      : std_logic_vector(DATA_WIDTH/8-1 downto 0);
  signal   iskcode2      : std_logic_vector(DATA_WIDTH/8-1 downto 0);
  signal   doscramble1   : std_logic;
  signal   doscramble2   : std_logic;
  
  constant MASK          : std_logic_vector := "0000000000111001"; -- X16 + X5 + X4 + X3 + 1

begin

  process (clk, rst)
    variable lfsrtmp  : std_logic_vector(15 downto 0);
  begin
    if (rst = '1') then
      lfsr <= x"ffff";
    elsif (rising_edge(clk)) then
      -- delayed values.
      iskcode1    <= iskcode;
      iskcode2    <= iskcode1;
      src1        <= src;
      src2        <= src1;
      doscramble1 <= doscramble;
      doscramble2 <= doscramble1;

      -- update LFSR.
      lfsrtmp := lfsr;
      eachbyte : for j in 0 to DATA_WIDTH/8-1 loop
        -- shift LFSR 1-bit to the left if src is not a SKP.
        if (not (iskcode1(j) = '1' and src1(8 * j + 7 downto 8 * j + 0) = K_SKP)) then
          eachbit : for i in 0 to 7 loop
            -- scramble src with LFSR.
            scrambled2(8 * j + i) <= src1(8 * j + i) xor lfsrtmp(15);
            if (lfsrtmp(15) = '1') then
              lfsrtmp := (lfsrtmp(14 downto 0) & '0') xor MASK;
            else
              lfsrtmp := (lfsrtmp(14 downto 0) & '0');
            end if;
          end loop eachbit;
        end if;
        -- reset LFSR when detect a COM.
        if (iskcode1(j) = '1' and src1(8 * j + 7 downto 8 * j + 0) = K_COM) then
          lfsrtmp := x"ffff";
        end if;    
      end loop eachbyte;
      lfsr <= lfsrtmp;

    end if;
  end process;

  process (clk)
  begin
    if (rising_edge(clk)) then
      -- do scramble a char if it is not a K-code.
      for j in 0 to DATA_WIDTH/8-1 loop
        dstk(j) <= iskcode2(j);
        if (SIMULATION = 0) then
          if (iskcode2(j) = '1' or doscramble2 = '0') then
            dst(7 + 8*j downto 0 + 8*j) <= src2(7 + 8*j downto 0 + 8*j);
          else
            dst(7 + 8*j downto 0 + 8*j) <= scrambled2(7 + 8*j downto 0 + 8*j);
          end if;
        else
          -- no scrambing (only for debugging purpose).
          dst(7 + 8*j downto 0 + 8*j) <= src2(7 + 8*j downto 0 + 8*j);
        end if;
      end loop;
    end if;
  end process;
  
end default;

#endif

-- 
-- Deskewer
-- remove lane-to-lane skew by inserting 'delay's.
-- 'delay' of each lane is determined using COM.
-- delay values demain unchanged while 'enable' is 0.
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity deskew is
  generic (
    NLANE       : integer
  );
  port (
    enable      : in  std_logic;
    srcd        : in  each16b(NLANE-1 downto 0);
    srck        : in  each2b(NLANE-1 downto 0);
    dstd        : out each16b(NLANE-1 downto 0);
    dstk        : out each2b(NLANE-1 downto 0);
    rst         : in  std_logic;        -- active high
    clk         : in  std_logic
  );
end deskew;

architecture default of deskew is

  signal srcd0      : each16b(NLANE-1 downto 0);
  signal srcd1      : each16b(NLANE-1 downto 0);
  signal srcd2      : each16b(NLANE-1 downto 0);
  signal srcd3      : each16b(NLANE-1 downto 0);
  signal srcd4      : each16b(NLANE-1 downto 0);
  signal srck0      : each2b(NLANE-1 downto 0);
  signal srck1      : each2b(NLANE-1 downto 0);
  signal srck2      : each2b(NLANE-1 downto 0);
  signal srck3      : each2b(NLANE-1 downto 0);
  signal srck4      : each2b(NLANE-1 downto 0);
  signal foundlow0  : each1b(NLANE-1 downto 0);       -- found COM at lower 8-byte on a lane.
  signal foundlow1  : each1b(NLANE-1 downto 0);
  signal foundhigh0 : each1b(NLANE-1 downto 0);       -- found COM at higer 8-byte on a lane.
  signal foundhigh1 : each1b(NLANE-1 downto 0);
  signal foundall   : std_logic;
  signal foundany0  : std_logic;
  signal foundany1  : std_logic;
  signal swap       : each1b(NLANE-1 downto 0);       -- swap higer & lower 8-bytes.
  signal swapall    : std_logic;
  signal swapany    : std_logic;
  signal delay      : each2b(NLANE-1 downto 0);       -- delay can be as large as 3.

begin
  
  -- generate delayed signals.
  src_lane : for i in 0 to NLANE-1 generate
    srcd0(i) <= srcd(i);
    srck0(i) <= srck(i);
    process(clk)
    begin
      if (rising_edge(clk)) then
        srcd1(i) <= srcd0(i);
        srck1(i) <= srck0(i);

        srcd2(i) <= srcd1(i);
        srck2(i) <= srck1(i);

        srcd3(i) <= srcd2(i);
        srck3(i) <= srck2(i);

        srcd4(i) <= srcd3(i);
        srck4(i) <= srck3(i);

        foundlow1(i)  <= foundlow0(i);
        foundhigh1(i) <= foundhigh0(i);
        foundany1 <= foundany0;
      end if;
    end process;  
  end generate src_lane;

  -- check if each lane have seen a COM or not.
  com_lane : for i in 0 to NLANE-1 generate
    process(clk, rst)
    begin
      if (rst = '1') then
        foundlow0(i)  <= '0';
        foundhigh0(i) <= '0';
      elsif (rising_edge(clk)) then
        if (enable = '1') then
          if (srck0(i)(0) = '1' and srcd0(i)( 7 downto 0) = K_COM) then
            foundlow0(i) <= '1';
          end if;
          if (srck0(i)(1) = '1' and srcd0(i)(15 downto 8) = K_COM) then
            foundhigh0(i) <= '1';
          end if;
          if (foundall = '1') then     -- clear when delay adjustment done.
            foundlow0(i)  <= '0';
            foundhigh0(i) <= '0';
          end if;
        end if;
      end if;
    end process;  
  end generate com_lane;

  -- found COMs on all lanes.
  process(foundlow0, foundhigh0)
    variable foundalltmp : std_logic;
  begin
    foundalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '0' and foundhigh0(i) = '0') then
        foundalltmp := '0';
      end if;
    end loop;
    foundall <= foundalltmp;
  end process;

  -- found firstly arrived COM(s).
  process(foundlow0, foundhigh0)
    variable foundanytmp : std_logic;
  begin
    foundanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '1') then
        foundanytmp := '1';
      end if;
      if (foundhigh0(i) = '1') then
        foundanytmp := '1';
      end if;
    end loop;
    foundany0 <= foundanytmp;
  end process;
  
  delay_lane : for i in 0 to NLANE-1 generate
    process(clk)
    begin
      if (rising_edge(clk)) then

        if (foundall = '0') then        -- delay adjustment is in progress.

          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i)  <= '0';
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i)  <= '1'; -- add a bit more (0.5clk = 1/250MHz) delay.
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;

        else        -- delay adjustment is about to complete.

          -- insert no delay for lanes which haven't see a COM so far.
          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i)  <= '0';
            end if;
          end if;
          -- insert 0.5clk delay
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i)  <= '1';
            end if;
          end if;

        end if;
      end if;
    end process;  
  end generate delay_lane;

  -- all lanes are swapped.
  process(swap)
    variable swapalltmp : std_logic;
  begin
    swapalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '0') then
        swapalltmp := '0';
      end if;
    end loop;
    swapall <= swapalltmp;
  end process;

  -- at least one lane is swapped.
  process(swap)
    variable swapanytmp : std_logic;
  begin
    swapanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '1') then
        swapanytmp := '1';
      end if;
    end loop;
    swapany <= swapanytmp;
  end process;

    
  mux_lane : for i in 0 to NLANE-1 generate
    process(srcd0, srcd1, srcd2, srcd3, srcd4,
            srck0, srck1, srck2, srck3, srck4,
            delay, swap, swapany, swapall)
    begin

      -- add delay, and swap if necessary.
      if (swap(i) = '0' or swapall = '1' or swapany = '0') then  -- no swap is necessary.
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i) <= srcd0(i);
          dstk(i) <= srck0(i);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i) <= srcd1(i);
          dstk(i) <= srck1(i);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i) <= srcd2(i);
          dstk(i) <= srck2(i);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i) <= srcd3(i);
          dstk(i) <= srck3(i);
        end if;
      else
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd0(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd1(i)(15 downto 8);
          dstk(i)(1) <= srck0(i)(0);
          dstk(i)(0) <= srck1(i)(1);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd1(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd2(i)(15 downto 8);
          dstk(i)(1) <= srck1(i)(0);
          dstk(i)(0) <= srck2(i)(1);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd2(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd3(i)(15 downto 8);
          dstk(i)(1) <= srck2(i)(0);
          dstk(i)(0) <= srck3(i)(1);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd3(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd4(i)(15 downto 8);
          dstk(i)(1) <= srck3(i)(0);
          dstk(i)(0) <= srck4(i)(1);
        end if;
      end if;

    end process;  
  end generate mux_lane;

end default;


#if 0 // deskewer that can handle very large skew.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity deskew is
  generic (
    NLANE       : integer
  );
  port (
    enable      : in  std_logic;
    srcd        : in  each16b(NLANE-1 downto 0);
    srck        : in  each2b(NLANE-1 downto 0);
    dstd        : out each16b(NLANE-1 downto 0);
    dstk        : out each2b(NLANE-1 downto 0);
    rst         : in  std_logic;        -- active high
    clk         : in  std_logic
  );
end deskew;

architecture default of deskew is

  signal srcd0      : each16b(NLANE-1 downto 0);
  signal srcd1      : each16b(NLANE-1 downto 0);
  signal srcd2      : each16b(NLANE-1 downto 0);
  signal srcd3      : each16b(NLANE-1 downto 0);
  signal srcd4      : each16b(NLANE-1 downto 0);
  signal srcd5      : each16b(NLANE-1 downto 0);
  signal srcd6      : each16b(NLANE-1 downto 0);
  signal srcd7      : each16b(NLANE-1 downto 0);
  signal srcd8      : each16b(NLANE-1 downto 0);

  signal srck0      : each2b(NLANE-1 downto 0);
  signal srck1      : each2b(NLANE-1 downto 0);
  signal srck2      : each2b(NLANE-1 downto 0);
  signal srck3      : each2b(NLANE-1 downto 0);
  signal srck4      : each2b(NLANE-1 downto 0);
  signal srck5      : each2b(NLANE-1 downto 0);
  signal srck6      : each2b(NLANE-1 downto 0);
  signal srck7      : each2b(NLANE-1 downto 0);
  signal srck8      : each2b(NLANE-1 downto 0);

  signal foundlow0  : each1b(NLANE-1 downto 0);       -- found COM at lower 8-byte on a lane.
  signal foundlow1  : each1b(NLANE-1 downto 0);
  signal foundhigh0 : each1b(NLANE-1 downto 0);       -- found COM at higer 8-byte on a lane.
  signal foundhigh1 : each1b(NLANE-1 downto 0);
  signal foundall   : std_logic;
  signal foundany0  : std_logic;
  signal foundany1  : std_logic;
  signal swap       : each1b(NLANE-1 downto 0);       -- swap higer & lower 8-bytes.
  signal swapall    : std_logic;
  signal swapany    : std_logic;
  signal delay      : each4b(NLANE-1 downto 0);       -- assign 4bits so that delay can be as large as 16-1.

begin
  
  -- generate delayed signals.
  src_lane : for i in 0 to NLANE-1 generate
    srcd0(i) <= srcd(i);
    srck0(i) <= srck(i);
    process(clk)
    begin
      if (rising_edge(clk)) then
        srcd1(i) <= srcd0(i);
        srck1(i) <= srck0(i);

        srcd2(i) <= srcd1(i);
        srck2(i) <= srck1(i);

        srcd3(i) <= srcd2(i);
        srck3(i) <= srck2(i);

        srcd4(i) <= srcd3(i);
        srck4(i) <= srck3(i);

        srcd5(i) <= srcd4(i);
        srck5(i) <= srck4(i);

        srcd6(i) <= srcd5(i);
        srck6(i) <= srck5(i);

        srcd7(i) <= srcd6(i);
        srck7(i) <= srck6(i);

        srcd8(i) <= srcd7(i);
        srck8(i) <= srck7(i);

        foundlow1(i)  <= foundlow0(i);
        foundhigh1(i) <= foundhigh0(i);
        foundany1 <= foundany0;
      end if;
    end process;  
  end generate src_lane;

  -- check if each lane have seen a COM or not.
  com_lane : for i in 0 to NLANE-1 generate
    process(clk, rst)
    begin
      if (rst = '1') then
        foundlow0(i)  <= '0';
        foundhigh0(i) <= '0';
      elsif (rising_edge(clk)) then
        if (enable = '1') then
          if (srck0(i)(0) = '1' and srcd0(i)( 7 downto 0) = K_COM) then
            foundlow0(i) <= '1';
          end if;
          if (srck0(i)(1) = '1' and srcd0(i)(15 downto 8) = K_COM) then
            foundhigh0(i) <= '1';
          end if;
          if (foundall = '1') then     -- clear when delay adjustment done.
            foundlow0(i)  <= '0';
            foundhigh0(i) <= '0';
          end if;
        end if;
      end if;
    end process;  
  end generate com_lane;

  -- found COMs on all lanes.
  process(foundlow0, foundhigh0)
    variable foundalltmp : std_logic;
  begin
    foundalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '0' and foundhigh0(i) = '0') then
        foundalltmp := '0';
      end if;
    end loop;
    foundall <= foundalltmp;
  end process;

  -- found firstly arrived COM(s).
  process(foundlow0, foundhigh0)
    variable foundanytmp : std_logic;
  begin
    foundanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (foundlow0(i) = '1') then
        foundanytmp := '1';
      end if;
      if (foundhigh0(i) = '1') then
        foundanytmp := '1';
      end if;
    end loop;
    foundany0 <= foundanytmp;
  end process;
  
  delay_lane : for i in 0 to NLANE-1 generate
    process(clk)
    begin
      if (rising_edge(clk)) then

        if (foundall = '0') then        -- delay adjustment is in progress.

          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i)  <= '0';
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              -- insert 1-clk delay when see a COM.
              delay(i) <= conv_std_logic_vector(1, delay(0)'length);
              swap(i)  <= '1'; -- add a bit more (0.5clk = 1/250MHz) delay.
            else
              -- increment delay until all other lanes see COMs.
              delay(i) <= unsigned(delay(i)) + 1;
            end if;
          end if;

        else        -- delay adjustment is about to complete.

          -- insert no delay for lanes which haven't see a COM so far.
          if (foundhigh0(i) = '1') then
            if (foundhigh1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i)  <= '0';
            end if;
          end if;
          -- insert 0.5clk delay
          if (foundlow0(i) = '1') then
            if (foundlow1(i) = '0') then
              delay(i) <= conv_std_logic_vector(0, delay(0)'length);
              swap(i)  <= '1';
            end if;
          end if;

        end if;
      end if;
    end process;  
  end generate delay_lane;

  -- all lanes are swapped.
  process(swap)
    variable swapalltmp : std_logic;
  begin
    swapalltmp := '1';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '0') then
        swapalltmp := '0';
      end if;
    end loop;
    swapall <= swapalltmp;
  end process;

  -- at least one lane is swapped.
  process(swap)
    variable swapanytmp : std_logic;
  begin
    swapanytmp := '0';
    for i in 0 to NLANE-1 loop
      if (swap(i) = '1') then
        swapanytmp := '1';
      end if;
    end loop;
    swapany <= swapanytmp;
  end process;

    
  mux_lane : for i in 0 to NLANE-1 generate
    process(srcd0, srcd1, srcd2, srcd3, srcd4, srcd5, srcd6, srcd7, srcd8,
            srck0, srck1, srck2, srck3, srck4, srck5, srck6, srck7, srck8,
            delay, swap, swapany, swapall)
    begin

      -- add delay, and swap if necessary.
      if (swap(i) = '0' or swapall = '1' or swapany = '0') then  -- no swap is necessary.
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i) <= srcd0(i);
          dstk(i) <= srck0(i);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i) <= srcd1(i);
          dstk(i) <= srck1(i);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i) <= srcd2(i);
          dstk(i) <= srck2(i);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i) <= srcd3(i);
          dstk(i) <= srck3(i);
        elsif (delay(i) = conv_std_logic_vector(4, delay(0)'length)) then
          dstd(i) <= srcd4(i);
          dstk(i) <= srck4(i);
        elsif (delay(i) = conv_std_logic_vector(5, delay(0)'length)) then
          dstd(i) <= srcd5(i);
          dstk(i) <= srck5(i);
        elsif (delay(i) = conv_std_logic_vector(6, delay(0)'length)) then
          dstd(i) <= srcd6(i);
          dstk(i) <= srck6(i);
        elsif (delay(i) = conv_std_logic_vector(7, delay(0)'length)) then
          dstd(i) <= srcd7(i);
          dstk(i) <= srck7(i);
        end if;
      else
        if (delay(i) = conv_std_logic_vector(0, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd0(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd1(i)(15 downto 8);
          dstk(i)(1) <= srck0(i)(0);
          dstk(i)(0) <= srck1(i)(1);
        elsif (delay(i) = conv_std_logic_vector(1, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd1(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd2(i)(15 downto 8);
          dstk(i)(1) <= srck1(i)(0);
          dstk(i)(0) <= srck2(i)(1);
        elsif (delay(i) = conv_std_logic_vector(2, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd2(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd3(i)(15 downto 8);
          dstk(i)(1) <= srck2(i)(0);
          dstk(i)(0) <= srck3(i)(1);
        elsif (delay(i) = conv_std_logic_vector(3, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd3(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd4(i)(15 downto 8);
          dstk(i)(1) <= srck3(i)(0);
          dstk(i)(0) <= srck4(i)(1);
        elsif (delay(i) = conv_std_logic_vector(4, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd4(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd5(i)(15 downto 8);
          dstk(i)(1) <= srck4(i)(0);
          dstk(i)(0) <= srck5(i)(1);
        elsif (delay(i) = conv_std_logic_vector(5, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd5(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd6(i)(15 downto 8);
          dstk(i)(1) <= srck5(i)(0);
          dstk(i)(0) <= srck6(i)(1);
        elsif (delay(i) = conv_std_logic_vector(6, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd6(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd7(i)(15 downto 8);
          dstk(i)(1) <= srck6(i)(0);
          dstk(i)(0) <= srck7(i)(1);
        elsif (delay(i) = conv_std_logic_vector(7, delay(0)'length)) then
          dstd(i)(15 downto 8) <= srcd7(i)( 7 downto 0);
          dstd(i)( 7 downto 0) <= srcd8(i)(15 downto 8);
          dstk(i)(1) <= srck7(i)(0);
          dstk(i)(0) <= srck8(i)(1);
        end if;
      end if;

    end process;  
  end generate mux_lane;

end default;

#endif

-- 
-- Deframer
-- decode a Physical-Layer Packet (PLP) and output its contents to
-- Data-Link-Layer Packet (DLLP) or Transaction-Layer Packet (TLP) port
-- depending on the framing symbol SDP/STP.
-- 
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity deframer is
  generic (
    DEFRAMER_DLLPBUF_DEPTH : integer;
    DEFRAMER_TLPBUF_DEPTH  : integer;
    NLANE                  : integer
  );
  port (
-- synthesis translate_off
    -- for debugging
    dpusedwmax  : out std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
    usedwmax    : out std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
-- synthesis translate_on

    -- PLP input
    data        : in  each16b(NLANE-1 downto 0);
    datak       : in  each2b(NLANE-1 downto 0);
    -- DLLP output
    wedp        : out std_logic;
    dp          : out std_logic_vector(47 downto 0);
    -- TLP output
    wetp        : out std_logic;
    betp        : out std_logic_vector(NLANE*2-1 downto 0);
    tptop       : out std_logic;
    tp          : out std_logic_vector(NLANE*16-1 downto 0);

    rst         : in  std_logic;        -- active high
    clk         : in  std_logic
  );
end deframer;

architecture default of deframer is

begin

  x1 : if (NLANE = 1) generate

    x1block : block

      type DP_STATE is (
        IDLE,
        RECV_D0_D1,
        RECV_D2_C0,
        RECV_C1,
        RECV_TYPE_D0,
        RECV_D1_D2,
        RECV_C0_C1,
        INVALID_STATE
      );

      type TP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_NONALIGNED,
        INVALID_STATE
      );

      signal wedp0        : std_logic;
      signal dp0          : std_logic_vector(47 downto 0);
      signal wetp0        : std_logic;
      signal wetp1        : std_logic;
      signal tptop0       : std_logic;
      signal tptop1       : std_logic;
      signal tp0          : std_logic_vector(NLANE*16-1 downto 0);
      signal tp1          : std_logic_vector(NLANE*16-1 downto 0);
      
      signal dpsm         : DP_STATE;
      signal tpsm         : TP_STATE;

    begin
      --
      -- deframe 64-bit PLP containing 48-bit DLLP:
      --
      -- DLLP format:
      --   1B  1B   3B-data           2B-crc    1B
      --   0   1    2     3     4     5    6    7
      --   SDP type data0 data1 data2 crc0 crc1 END
      --
      --   16-bit aligned input:
      --     type data1 crc0  END
      --     SDP  data0 data2 crc1
      --
      --   16-bit non-aligned input:
      --     SDP  data0 data2 crc1
      --          type  data1 crc0 END
      --
      wedp <= wedp0;
      dp   <= dp0;

      process (clk, rst)
      begin
        if (rst = '1') then
          wedp0  <= '0';
          dpsm <= IDLE;
        elsif (rising_edge(clk)) then

          wedp0  <= '0';

          case dpsm is
            when IDLE =>
              -- detect a SDP on data0 lower byte.
              if (datak(0)(0) = '1' and data(0)(7 downto 0) = K_SDP) then
                dp0( 7 downto 0) <= data(0)(15 downto 8);  -- type
                dpsm <= RECV_D0_D1;
                -- detect a SDP on data0 higher byte.
              elsif (datak(0)(1) = '1' and data(0)(15 downto 8) = K_SDP) then
                dpsm <= RECV_TYPE_D0;
              end if;

              -- write aligned DP
            when RECV_D0_D1 =>
              dp0(15 downto  8) <= data(0)( 7 downto 0);  -- data0
              dp0(23 downto 16) <= data(0)(15 downto 8);  -- data1
              dpsm <= RECV_D2_C0;

            when RECV_D2_C0 =>
              dp0(31 downto 24) <= data(0)( 7 downto 0);  -- data2
              dp0(39 downto 32) <= data(0)(15 downto 8);  -- crc high
              dpsm <= RECV_C1;

            when RECV_C1 =>
              dp0(47 downto 40) <= data(0)( 7 downto 0);  -- crc low
              wedp0 <= '1';
              dpsm <= IDLE;

              -- write non-aligned DP
            when RECV_TYPE_D0 =>
              dp0( 7 downto  0) <= data(0)( 7 downto 0);  -- type
              dp0(15 downto  8) <= data(0)(15 downto 8);  -- data0
              dpsm <= RECV_D1_D2;

            when RECV_D1_D2 =>
              dp0(23 downto 16) <= data(0)( 7 downto 0);  -- data1
              dp0(31 downto 24) <= data(0)(15 downto 8);  -- data2
              dpsm <= RECV_C0_C1;

            when RECV_C0_C1 =>
              wedp0 <= '1';
              dp0(39 downto 32) <= data(0)( 7 downto 0);  -- crc high
              dp0(47 downto 40) <= data(0)(15 downto 8);  -- crc low
              dpsm <= IDLE;

            when others =>
              wedp0  <= '0';
              dpsm <= IDLE;

          end case;
        end if;
      end process;  

      --
      -- deframe PLP containing TLP:
      --
      -- TLP format:
      --   1B  2B   12/16B   12-4116B     4B    4B   1B
      --   STP Seq# Header Data        (ECRC) LCRC EDP/EDB
      --
      --   16-bit aligned input:
      --     seq0 h0   h2  h4  h6  h8  h10 . . . . . . END
      --     STP  seq1 h1  h3  h5  h7  h9  h11 . . . . lcrc3
      --
      --   16-bit non-aligned input:
      --     STP  seq1 h1  h3  h5  h7  h9  h11 . . . lcrc3
      --          seq0 h0  h2  h4  h6  h8  h10 . . . . . . END
      --
      tp <= tp1;
      wetp <= '1' when wetp0 = '1' or wetp1 = '1' else '0';
      tptop <= tptop1;

      process (clk, rst)
      begin
        if (rst = '1') then
          wetp0  <= '0';
          wetp1  <= '0';
          tpsm <= IDLE;
        elsif (rising_edge(clk)) then

          wetp0  <= '0';
          wetp1  <= '0';
          betp   <= (others => '1');
          tptop0 <= '0';
          tptop1 <= tptop0;

          case tpsm is
            when IDLE =>
              -- detect a STP on data0 lower byte.
              if (datak(0)(0) = '1' and data(0)(7 downto 0) = K_STP) then
                tp0( 7 downto 0) <= data(0)(15 downto 8);  -- seq0
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
                -- detect a STP on data0 higher byte.
              elsif (datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP) then
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              end if;

              -- write aligned TP
            when RECV_ALIGNED =>
              tp1(15 downto  8) <= data(0)( 7 downto 0);
              tp1( 7 downto  0) <= tp0( 7 downto 0);
              tp0( 7 downto  0) <= data(0)(15 downto 8);
              wetp1 <= wetp0;
              if (datak(0)(1) = '1' and
                  (data(0)(15 downto 8) = K_END or data(0)(15 downto 8) = K_EDB)) then
                tpsm <= IDLE;
              else
                wetp0 <= '1';
              end if;

            when RECV_NONALIGNED =>
              tp1(15 downto  8) <= data(0)(15 downto 8);
              tp1( 7 downto  0) <= data(0)( 7 downto 0);
              if (datak(0)(0) = '1' and
                  (data(0)( 7 downto 0) = K_END or data(0)( 7 downto 0) = K_EDB)) then

                -- detect a STP on data0 higher byte.
                -- i.e., END is immediately followed by STP of next TLP.
                if (datak(0)(1) = '1' and data(0)(15 downto 8) = K_STP) then
                  tptop0 <= '1';
                else
                  tpsm <= IDLE;
                end if;

              else
                wetp0 <= '1';
                wetp1 <= '1';
              end if;

            when others =>
              tpsm <= IDLE;

          end case;
        end if;
      end process;

    end block x1block;

  end generate x1;


  x4 : if (NLANE = 4) generate

    x4block : block

      component scfifo
        generic (
          LPM_NUMWORDS            : natural := 2**DEFRAMER_TLPBUF_DEPTH;
          LPM_WIDTHU              : natural := DEFRAMER_TLPBUF_DEPTH;
          LPM_WIDTH               : natural := NLANE*18;
          LPM_SHOWAHEAD           : string  := "ON";
          ADD_RAM_OUTPUT_REGISTER : string  := "ON";
          OVERFLOW_CHECKING       : string  := "ON";
          UNDERFLOW_CHECKING      : string  := "OFF";
          USE_EAB                 : string  := "ON";
          INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
          LPM_TYPE                : string  := "scfifo"
        );
        port (
          usedw    : out std_logic_vector(LPM_WIDTHU-1 downto 0);

          rdreq    : in  std_logic;
          full     : out std_logic;
          empty    : out std_logic;
          clock    : in  std_logic;
          sclr     : in  std_logic;
          q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
          wrreq    : in  std_logic;
          data     : in  std_logic_vector(LPM_WIDTH-1 downto 0)
        );
      end component;

      type DP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_NONALIGNED,
        INVALID_STATE
      );

      type TP_STATE is (
        IDLE,
        RECV_ALIGNED,
        RECV_END_ALIGNED,
        RECV_NONALIGNED,
        RECV_END_NONALIGNED,
        INVALID_STATE
      );

      type BUF_STATE is (
        IDLE,
        RECV
      );
      
      signal dpsm         : DP_STATE;
      signal tpsm         : TP_STATE;
      signal bufsm        : BUF_STATE;

      signal wedp0        : std_logic;
      signal wedp1        : std_logic;
      signal dp0          : std_logic_vector(47 downto 0);
      signal dp1          : std_logic_vector(47 downto 0);
      signal wetp0        : std_logic;
      signal wetp1        : std_logic;
      signal tptop0       : std_logic;
      signal tptop1       : std_logic;
      signal tp0          : std_logic_vector(NLANE*16-1 downto 0);
      signal tp1          : std_logic_vector(NLANE*16-1 downto 0);

      signal end_at_low : std_logic;
      signal bufin      : std_logic_vector(NLANE*18-1 downto 0); 
      signal bufout     : std_logic_vector(NLANE*18-1 downto 0); 
      signal wrreq      : std_logic;
      signal rdreq      : std_logic;
      signal rdack      : std_logic;
      signal full       : std_logic;
      signal empty      : std_logic;
      signal d          : each16b(NLANE-1 downto 0);
      signal k          : each2b(NLANE-1 downto 0);

      signal sdp_at_0l     : std_logic;
      signal sdp_at_0h     : std_logic;
      signal end_at_3l     : std_logic;
      signal end_at_3h     : std_logic;
      signal stp_at_0l_in  : std_logic;
      signal stp_at_0h_in  : std_logic;
      signal end_at_3l_in  : std_logic;
      signal end_at_3h_in  : std_logic;
      signal stp_at_0l_out : std_logic;
      signal stp_at_0h_out : std_logic;
      signal end_at_3l_out : std_logic;
      signal end_at_3h_out : std_logic;

      signal usedw0    : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
      signal usedwmax0 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);

    begin

-- synthesis translate_off
      -- for debugging
      fullout  <= full;
      usedwmax <= usedwmax0;
-- synthesis translate_on

      --   DLLP 16-bit aligned input:
      --
      --   lane3 H  END
      --   lane2 H  crc1
      --   lane1 H  crc0
      --   lane0 H  data2
      --   lane3 L  data1
      --   lane2 L  data0
      --   lane1 L  type
      --   lane0 L  SDP
      --
      --   DLLP 16-bit non-aligned input:
      --   lane3 H  data1
      --   lane2 H  data0
      --   lane1 H  type
      --   lane0 H  SDP
      --   lane3 L        END
      --   lane2 L        crc1
      --   lane1 L        crc0
      --   lane0 L        data2
      wedp <= wedp1;
      dp   <= dp1;

      sdp_at_0l  <= '1' when datak(0)(0) = '1' and data(0)( 7 downto 0) = K_SDP else '0';
      sdp_at_0h  <= '1' when datak(0)(1) = '1' and data(0)(15 downto 8) = K_SDP else '0';
      end_at_3l  <= '1' when datak(3)(0) = '1' and data(3)( 7 downto 0) = K_END else '0';
      end_at_3h  <= '1' when datak(3)(1) = '1' and data(3)(15 downto 8) = K_END else '0';

      process (clk, rst)
      begin
        if (rst = '1') then
          wedp0 <= '0';
          wedp1 <= '0';
          dpsm <= IDLE;
        elsif (rising_edge(clk)) then

          wedp0 <= '0';
          wedp1 <= '0';

          case dpsm is
            when IDLE =>
              dp1   <= dp0;
              wedp1 <= wedp0;

              -- detect a SDP on data0 lower byte.
              if (sdp_at_0l = '1') then
                dp0(47 downto 40) <= data(2)(15 downto 8);  -- crc low
                dp0(39 downto 32) <= data(1)(15 downto 8);  -- crc high
                dp0(31 downto 24) <= data(0)(15 downto 8);  -- data2
                dp0(23 downto 16) <= data(3)( 7 downto 0);  -- data1
                dp0(15 downto  8) <= data(2)( 7 downto 0);  -- data0
                dp0( 7 downto  0) <= data(1)( 7 downto 0);  -- type
                wedp0 <= '1';
                -- detect a SDP on data0 higher byte.
              elsif (sdp_at_0h = '1') then
                dp0(23 downto 16) <= data(3)(15 downto 8);  -- data1
                dp0(15 downto  8) <= data(2)(15 downto 8);  -- data0
                dp0( 7 downto  0) <= data(1)(15 downto 8);  -- type
                dpsm <= RECV_NONALIGNED;
              end if;

            when RECV_NONALIGNED =>
              wedp1 <= '1';
              dp1(47 downto 40) <= data(2)( 7 downto 0);  -- crc low
              dp1(39 downto 32) <= data(1)( 7 downto 0);  -- crc high
              dp1(31 downto 24) <= data(0)( 7 downto 0);  -- data2
              dp1(23 downto 0)  <= dp0(23 downto 0);

              if (sdp_at_0h = '1') then
                dp0(23 downto 16) <= data(3)(15 downto 8);  -- data1
                dp0(15 downto  8) <= data(2)(15 downto 8);  -- data0
                dp0( 7 downto  0) <= data(1)(15 downto 8);  -- type
              else
                dpsm <= IDLE;
              end if;
              
            when others =>
              dpsm <= IDLE;

          end case;
        end if;
      end process;  

      --
      -- TLP
      --
      tp <= tp1;
      tptop <= tptop1;
      wetp <= '1' when wetp0 = '1' or wetp1 = '1' else '0';

      process (clk)
      begin
        if (rising_edge(clk)) then
          for i in 0 to NLANE-1 loop
            bufin(i*18+7 downto i*18+0)    <= data(i)(7 downto 0);
            bufin(i*18+8)                  <= datak(i)(0);
            bufin(i*18+16 downto i*18+9)   <= data(i)(15 downto 8);
            bufin(i*18+17)                 <= datak(i)(1);
          end loop;
        end if;
      end process;

      lane: for i in 0 to NLANE-1 generate
        d(i)(7 downto 0)               <= bufout(i*18+7 downto i*18+0);
        k(i)(0)                        <= bufout(i*18+8);
        d(i)(15 downto 8)              <= bufout(i*18+16 downto i*18+9);
        k(i)(1)                        <= bufout(i*18+17);
      end generate lane;

      stp_at_0l_in  <= '1' when datak(0)(0) = '1' and  data(0)( 7 downto 0) = K_STP else '0';
      stp_at_0h_in  <= '1' when datak(0)(1) = '1' and  data(0)(15 downto 8) = K_STP else '0';
      end_at_3l_in  <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
      end_at_3h_in  <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
      stp_at_0l_out <= '1' when k(0)(0) = '1'     and  d(0)( 7 downto 0) = K_STP else '0';
      stp_at_0h_out <= '1' when k(0)(1) = '1'     and  d(0)(15 downto 8) = K_STP else '0';
      end_at_3l_out <= '1' when k(3)(0) = '1'     and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
      end_at_3h_out <= '1' when k(3)(1) = '1'     and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
                   
      rxbuf0 : scfifo
        port map (
          usedw    => usedw0,

          clock    => clk,
          sclr     => rst,

          wrreq    => wrreq,
          data     => bufin,

          rdreq    => rdreq,
          full     => full,
          empty    => empty,
          q        => bufout
        );

      process (clk, rst)
      begin
        if (rst = '1') then
          usedwmax0 <= (others => '0');
        elsif (rising_edge(clk)) then
          if (unsigned(usedwmax0) < unsigned(usedw0)) then
            usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
          end if;
        end if;
      end process;

      process (clk, rst)
      begin
        if (rst = '1') then
          wrreq <= '0';
          bufsm <= IDLE;
        elsif (rising_edge(clk)) then
          wrreq <= '0';
          case bufsm is
            when IDLE =>
              if (stp_at_0l_in = '1' or stp_at_0h_in = '1') then
                wrreq <= '1';
                bufsm <= RECV;
              end if;
            when RECV =>
              wrreq <= '1';
              if (end_at_3h_in = '1' or (end_at_3l_in = '1' and stp_at_0h_in = '0')) then
                bufsm <= IDLE;
              end if;
            when others =>
          end case;
        end if;
      end process;

      rdreq <= '1' when empty = '0' and rdack = '1' else '0';
      rdack <= '0' when end_at_3l_out = '1' and stp_at_0h_out = '1' and tpsm = RECV_ALIGNED else '1';

      process (clk, rst)
      begin
        if (rst = '1') then
          wetp0  <= '0';
          wetp1  <= '0';
          tpsm <= IDLE;
        elsif (rising_edge(clk)) then

          wetp0  <= '0';
          wetp1  <= '0';
          betp   <= (others => '1');
          tptop0 <= '0';
          tptop1 <= tptop0;

          case tpsm is
            when IDLE =>
              if (empty = '0') then
                -- detect a STP on lane0 lower byte.
                if (k(0)(0) = '1' and d(0)(7 downto 0) = K_STP) then
                  tp0(55 downto 48) <= d(3)(15 downto 8);
                  tp0(47 downto 40) <= d(2)(15 downto 8);
                  tp0(39 downto 32) <= d(1)(15 downto 8);
                  tp0(31 downto 24) <= d(0)(15 downto 8);
                  tp0(23 downto 16) <= d(3)( 7 downto 0);
                  tp0(15 downto  8) <= d(2)( 7 downto 0);
                  tp0( 7 downto  0) <= d(1)( 7 downto 0);
                  tptop0 <= '1';
                  tpsm <= RECV_ALIGNED;
                  -- detect a STP on lane0 higher byte.
                elsif (k(0)(1) = '1' and d(0)(15 downto 8) = K_STP) then
                  tp0(23 downto 16) <= d(3)(15 downto 8);
                  tp0(15 downto  8) <= d(2)(15 downto 8);
                  tp0( 7 downto  0) <= d(1)(15 downto 8);
                  tptop0 <= '1';
                  tpsm <= RECV_NONALIGNED;
                end if;
              end if;

            when RECV_ALIGNED =>
              tp1(63 downto 56) <= d(0)(7 downto 0);
              tp1(55 downto  0) <= tp0(55 downto  0);

              tp0(55 downto 48) <= d(3)(15 downto 8);
              tp0(47 downto 40) <= d(2)(15 downto 8);
              tp0(39 downto 32) <= d(1)(15 downto 8);
              tp0(31 downto 24) <= d(0)(15 downto 8);
              tp0(23 downto 16) <= d(3)( 7 downto 0);
              tp0(15 downto  8) <= d(2)( 7 downto 0);
              tp0( 7 downto  0) <= d(1)( 7 downto 0);

              wetp1 <= wetp0;
              wetp0 <= '1';

              if (end_at_3h_out = '1') then
                end_at_low <= '0';
                tpsm <= RECV_END_ALIGNED;
              elsif (end_at_3l_out = '1') then
                end_at_low <= '1';
                tpsm <= RECV_END_ALIGNED;
              end if;

            when RECV_END_ALIGNED =>
              wetp1 <= wetp0;
              wetp0 <= '1';
              tp1(47 downto  0) <= tp0(47 downto  0);
              if (end_at_low = '0') then
                betp <= "00111111";
                tp1(63 downto 48) <= (others => '0');
              else
                betp <= "00000011";
                tp1(63 downto 16) <= (others => '0');
              end if;

              if (stp_at_0l_out = '1') then
                tp0(55 downto 48) <= d(3)(15 downto 8);
                tp0(47 downto 40) <= d(2)(15 downto 8);
                tp0(39 downto 32) <= d(1)(15 downto 8);
                tp0(31 downto 24) <= d(0)(15 downto 8);
                tp0(23 downto 16) <= d(3)( 7 downto 0);
                tp0(15 downto  8) <= d(2)( 7 downto 0);
                tp0( 7 downto  0) <= d(1)( 7 downto 0);
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
              elsif (stp_at_0h_out = '1') then
                tp0(23 downto 16) <= d(3)(15 downto 8);
                tp0(15 downto  8) <= d(2)(15 downto 8);
                tp0( 7 downto  0) <= d(1)(15 downto 8);
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              else
                tpsm <= IDLE;
              end if;

            when RECV_NONALIGNED =>
              tp1(63 downto 56) <= d(0)(15 downto 8);
              tp1(55 downto 48) <= d(3)( 7 downto 0);
              tp1(47 downto 40) <= d(2)( 7 downto 0);
              tp1(39 downto 32) <= d(1)( 7 downto 0);
              tp1(31 downto 24) <= d(0)( 7 downto 0);
              tp1(23 downto  0) <= tp0(23 downto  0);

              tp0(23 downto 16) <= d(3)(15 downto 8);
              tp0(15 downto  8) <= d(2)(15 downto 8);
              tp0( 7 downto  0) <= d(1)(15 downto 8);

              wetp1 <= wetp0;
              wetp0 <= '1';

              if (end_at_3h_out = '1') then
                end_at_low <= '0';
                tpsm <= RECV_END_NONALIGNED;
              end if;

              if (end_at_3l_out = '1') then
                end_at_low <= '1';
                if (stp_at_0h_out = '1') then
                  betp <= "00111111";
                  tp1(63 downto 48) <= (others => '0');
                  tptop0 <= '1';
                else
                  betp <= "00111111";
                  tp1(63 downto 48) <= (others => '0');
--                  tp0(23 downto  0) <= (others => '0');
                  tpsm <= IDLE;
                end if;
              end if;

            when RECV_END_NONALIGNED =>
              wetp1 <= wetp0;
              wetp0 <= '1';
              tp1(23 downto  0) <= tp0(23 downto  0);
              if (end_at_low = '0') then
                betp <= "00000011";
                tp1(63 downto 16) <= (others => '0');
              else
                betp <= "00111111";
                tp1(63 downto 48) <= (others => '0');
              end if;

              if (stp_at_0l_out = '1') then
                tp0(55 downto 48) <= d(3)(15 downto 8);
                tp0(47 downto 40) <= d(2)(15 downto 8);
                tp0(39 downto 32) <= d(1)(15 downto 8);
                tp0(31 downto 24) <= d(0)(15 downto 8);
                tp0(23 downto 16) <= d(3)( 7 downto 0);
                tp0(15 downto  8) <= d(2)( 7 downto 0);
                tp0( 7 downto  0) <= d(1)( 7 downto 0);
                tptop0 <= '1';
                tpsm <= RECV_ALIGNED;
              elsif (stp_at_0h_out = '1') then
                tp0(23 downto 16) <= d(3)(15 downto 8);
                tp0(15 downto  8) <= d(2)(15 downto 8);
                tp0( 7 downto  0) <= d(1)(15 downto 8);
                tptop0 <= '1';
                tpsm <= RECV_NONALIGNED;
              else
                tpsm <= IDLE;
              end if;
              
            when others =>
              tpsm <= IDLE;

          end case;
        end if;
      end process;  

    end block x4block;

  end generate x4;

  x8 : if (NLANE = 8) generate

    x8block : block

    begin

      --
      -- deframe DLLP
      --
      sdpblock : block

        component scfifo
          generic (
            LPM_NUMWORDS            : natural := 2**DEFRAMER_DLLPBUF_DEPTH;
            LPM_WIDTHU              : natural := DEFRAMER_DLLPBUF_DEPTH;
            LPM_WIDTH               : natural := NLANE*18 + 2;
            LPM_SHOWAHEAD           : string  := "ON";
            ADD_RAM_OUTPUT_REGISTER : string  := "ON";
            OVERFLOW_CHECKING       : string  := "ON";
            UNDERFLOW_CHECKING      : string  := "OFF";
            USE_EAB                 : string  := "ON";
            INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
            LPM_TYPE                : string  := "scfifo"
          );
          port (
            usedw    : out std_logic_vector(LPM_WIDTHU-1 downto 0);

            rdreq    : in  std_logic;
            full     : out std_logic;
            empty    : out std_logic;
            clock    : in  std_logic;
            sclr     : in  std_logic;
            q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
            wrreq    : in  std_logic;
            data     : in  std_logic_vector(LPM_WIDTH-1 downto 0)
          );
        end component;

        signal wrcnt                  : std_logic_vector(1 downto 0);
        signal rdcnt, rdcnt0          : std_logic_vector(1 downto 0);
        signal sdpat0                 : std_logic_vector(NLANE/2-1 downto 0);
        signal wedp1                  : std_logic;
        signal dp0, dp1               : std_logic_vector(47 downto 0);

        signal bufin                  : std_logic_vector(NLANE*18+1 downto 0); 
        signal bufout                 : std_logic_vector(NLANE*18+1 downto 0); 
        signal wrreq                  : std_logic;
        signal rdreq                  : std_logic;
        signal rdack                  : std_logic;
        signal full                   : std_logic;
        signal empty                  : std_logic;
        signal usedw0                 : std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);
        signal usedwmax0              : std_logic_vector(DEFRAMER_DLLPBUF_DEPTH-1 downto 0);

        signal d                      : each16b(NLANE-1 downto 0);
        signal k                      : each2b(NLANE-1 downto 0);

        signal sdp_at_0l_in, sdp_at_4l_in, sdp_at_0h_in, sdp_at_4h_in  : std_logic;
        signal end_at_3l_in, end_at_7l_in, end_at_3h_in, end_at_7h_in  : std_logic;
        signal sdp_at_any_lane_in : std_logic;
        signal end_at_any_lane_in : std_logic;

        signal sdp_at_0l_out, sdp_at_4l_out, sdp_at_0h_out, sdp_at_4h_out : std_logic;
        signal end_at_3l_out, end_at_7l_out, end_at_3h_out, end_at_7h_out : std_logic;
        signal sdp_at_any_lane_out : std_logic;
        signal end_at_any_lane_out : std_logic;

--        attribute keep : boolean;
--        attribute keep of usedwmax0 : signal is true;

      begin
-- synthesis translate_off
        -- for debugging
        dpfullout  <= full;
        dpusedwmax <= usedwmax0;
-- synthesis translate_on

        dp    <= dp1;
        wedp  <= wedp1;

        process (clk)
        begin
          if (rising_edge(clk)) then
            for i in 0 to NLANE-1 loop
              bufin(i*18+7 downto i*18+0)    <= data(i)(7 downto 0);
              bufin(i*18+8)                  <= datak(i)(0);
              bufin(i*18+16 downto i*18+9)   <= data(i)(15 downto 8);
              bufin(i*18+17)                 <= datak(i)(1);
            end loop;
          end if;
        end process;
        bufin(NLANE*18+1 downto NLANE*18) <= wrcnt;  -- direct connection with no register.

        process(empty, bufout)
        begin
          for i in 0 to NLANE-1 loop
            if (empty = '0') then
              d(i)(7 downto 0)               <= bufout(i*18+7 downto i*18+0);
              k(i)(0)                        <= bufout(i*18+8);
              d(i)(15 downto 8)              <= bufout(i*18+16 downto i*18+9);
              k(i)(1)                        <= bufout(i*18+17);
            else
              d(i) <= (others => '0');
              k(i) <= (others => '0');
            end if;
          end loop;
        end process;
        rdcnt <= bufout(NLANE*18+1 downto NLANE*18);

        sdp_at_0l_in  <= '1' when datak(0)(0) = '1' and  data(0)( 7 downto 0) = K_SDP else '0';
        sdp_at_4l_in  <= '1' when datak(4)(0) = '1' and  data(4)( 7 downto 0) = K_SDP else '0';
        sdp_at_0h_in  <= '1' when datak(0)(1) = '1' and  data(0)(15 downto 8) = K_SDP else '0';
        sdp_at_4h_in  <= '1' when datak(4)(1) = '1' and  data(4)(15 downto 8) = K_SDP else '0';

        end_at_3l_in  <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_in  <= '1' when datak(7)(0) = '1' and (data(7)( 7 downto 0) = K_END or data(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_in  <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_in  <= '1' when datak(7)(1) = '1' and (data(7)(15 downto 8) = K_END or data(7)(15 downto 8) = K_EDB) else '0';

        sdp_at_any_lane_in <= '1' when (sdp_at_0l_in = '1' or sdp_at_4l_in = '1' or sdp_at_0h_in = '1' or sdp_at_4h_in = '1') else '0';
        end_at_any_lane_in <= '1' when (end_at_3l_in = '1' or end_at_7l_in = '1' or end_at_3h_in = '1' or end_at_7h_in = '1') else '0';

        sdp_at_0l_out <= '1' when k(0)(0) = '1'     and  d(0)( 7 downto 0) = K_SDP else '0';
        sdp_at_4l_out <= '1' when k(4)(0) = '1'     and  d(4)( 7 downto 0) = K_SDP else '0';
        sdp_at_0h_out <= '1' when k(0)(1) = '1'     and  d(0)(15 downto 8) = K_SDP else '0';
        sdp_at_4h_out <= '1' when k(4)(1) = '1'     and  d(4)(15 downto 8) = K_SDP else '0';

        end_at_3l_out <= '1' when k(3)(0) = '1'     and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_out <= '1' when k(7)(0) = '1'     and (d(7)( 7 downto 0) = K_END or d(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_out <= '1' when k(3)(1) = '1'     and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_out <= '1' when k(7)(1) = '1'     and (d(7)(15 downto 8) = K_END or d(7)(15 downto 8) = K_EDB) else '0';

        sdp_at_any_lane_out <= '1' when (sdp_at_0l_out = '1' or sdp_at_4l_out = '1' or sdp_at_0h_out = '1' or sdp_at_4h_out = '1') else '0';
        end_at_any_lane_out <= '1' when (end_at_3l_out = '1' or end_at_7l_out = '1' or end_at_3h_out = '1' or end_at_7h_out = '1') else '0';

        rxbuf0 : scfifo
          port map (
            usedw    => usedw0,

            clock    => clk,
            sclr     => rst,

            wrreq    => wrreq,
            data     => bufin,

            rdreq    => rdreq,
            full     => full,
            empty    => empty,
            q        => bufout
          );

        process (clk, rst)
        begin
          if (rst = '1') then
            usedwmax0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (unsigned(usedwmax0) < unsigned(usedw0)) then
              usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
            end if;
          end if;
        end process;

        process (clk, rst)
        begin
          if (rst = '1') then
            wrreq <= '0';
          elsif (rising_edge(clk)) then
            wrreq <= '0';
            -- buffer when see an SDP or an END.
            if (sdp_at_any_lane_in = '1' or end_at_3l_in = '1') then
              wrreq <= '1';
            end if;
            -- stop rdreq 2clk when see an END, an SDP-END pair, and SDP in the same clk.
            if (end_at_3l_in = '1' and sdp_at_4l_in = '1' and sdp_at_4h_in = '1') then
              wrcnt <= "10";
            -- stop rdreq 1clk when see an END followed by SDP in the same clk.
            elsif (end_at_3l_in = '1' and (sdp_at_4l_in = '1' or sdp_at_0h_in = '1' or sdp_at_4h_in = '1')) then
              wrcnt <= "01";
            elsif (end_at_7l_in = '1' and (sdp_at_0h_in = '1' or sdp_at_4h_in = '1')) then
              wrcnt <= "01";
            elsif (end_at_3h_in = '1' and sdp_at_4h_in = '1') then
              wrcnt <= "01";
            else
              wrcnt <= "00";
            end if;
          end if;
        end process;

        process(clk, rst)
        begin
          if (rst = '1') then
            rdcnt0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (rdcnt0 = conv_std_logic_vector(0, rdcnt'length)) then
              if (empty = '0' and rdcnt /= conv_std_logic_vector(0, rdcnt'length)) then
                rdcnt0 <= rdcnt;
              end if;
            else
              rdcnt0 <= unsigned(rdcnt0) - 1;
            end if;
          end if;
        end process;

        rdreq <= '1' when empty = '0' and rdack = '1' else '0';
        rdack <= '0' when rdcnt  /= conv_std_logic_vector(0, rdcnt'length) and
                          rdcnt0 /= conv_std_logic_vector(1, rdcnt0'length) else '1';

        process(clk)
          variable sdpattmp   : std_logic_vector(3 downto 0);
        begin
          if (rising_edge(clk)) then
            --
            -- prestage
            --
            sdpattmp := (others => '0');
            if (empty = '0') then
              if (rdcnt = conv_std_logic_vector(2, rdcnt'length)) then  -- END, SDP-END, SDP in a single clk.
                if (rdcnt0 = conv_std_logic_vector(2, rdcnt0'length)) then
                  sdpattmp := "0010";
                elsif (rdcnt0 = conv_std_logic_vector(1, rdcnt0'length)) then
                  sdpattmp := "1000";
                end if;
              elsif (sdp_at_0l_out = '1' and sdp_at_0h_out = '1') then  -- two SDPs at 0L & 0H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0001";
                else
                  sdpattmp := "0100";
                end if;
              elsif (sdp_at_0l_out = '1' and sdp_at_4h_out = '1') then  -- two SDPs at 0L & 4H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0001";
                else
                  sdpattmp := "1000";
                end if;
              elsif (sdp_at_4l_out = '1' and sdp_at_4h_out = '1') then  -- two SDPs at 4L & 4H.
                if (rdcnt0 = conv_std_logic_vector(0, rdcnt0'length)) then
                  sdpattmp := "0010";
                else
                  sdpattmp := "1000";
                end if;
              else  -- signle SDP.
                if (sdp_at_0l_out = '1') then
                  sdpattmp := "0001";
                elsif (sdp_at_4l_out = '1') then
                  sdpattmp := "0010";
                elsif (sdp_at_0h_out = '1') then
                  sdpattmp := "0100";
                elsif (sdp_at_4h_out = '1') then
                  sdpattmp := "1000";
                end if;
              end if;

              -- last-half of the previous DLLP.
              if (end_at_3l_out = '1' and rdcnt0 = conv_std_logic_vector(0, rdcnt'length)) then
                sdpattmp := "0000";
              end if;

            end if;
            sdpat0 <= sdpattmp;

            case sdpattmp is
              when "0001" =>
                dp0( 47 downto  40) <= d(6)( 7 downto 0);
                dp0( 39 downto  32) <= d(5)( 7 downto 0);
                dp0( 31 downto  24) <= d(4)( 7 downto 0);
                dp0( 23 downto  16) <= d(3)( 7 downto 0);
                dp0( 15 downto   8) <= d(2)( 7 downto 0);
                dp0(  7 downto   0) <= d(1)( 7 downto 0);

              when "0010" =>
                dp0( 47 downto  40) <= d(2)(15 downto 8);
                dp0( 39 downto  32) <= d(1)(15 downto 8);
                dp0( 31 downto  24) <= d(0)(15 downto 8);
                dp0( 23 downto  16) <= d(7)( 7 downto 0);
                dp0( 15 downto   8) <= d(6)( 7 downto 0);
                dp0(  7 downto   0) <= d(5)( 7 downto 0);

              when "0100" =>
                dp0( 47 downto  40) <= d(6)(15 downto 8);
                dp0( 39 downto  32) <= d(5)(15 downto 8);
                dp0( 31 downto  24) <= d(4)(15 downto 8);
                dp0( 23 downto  16) <= d(3)(15 downto 8);
                dp0( 15 downto   8) <= d(2)(15 downto 8);
                dp0(  7 downto   0) <= d(1)(15 downto 8);

              when "1000" =>
                dp0( 23 downto  16) <= d(7)(15 downto 8);
                dp0( 15 downto   8) <= d(6)(15 downto 8);
                dp0(  7 downto   0) <= d(5)(15 downto 8);

              when others =>
            end case;

            --
            -- stage 0
            --
            if (sdpat0 = conv_std_logic_vector(0, sdpat0'length)) then
              wedp1 <= '0';
            else
              wedp1 <= '1';
            end if;

            dp1 <= dp0;
            if (sdpat0 = "1000") then
              dp1( 47 downto  40) <= d(2)(7 downto 0);
              dp1( 39 downto  32) <= d(1)(7 downto 0);
              dp1( 31 downto  24) <= d(0)(7 downto 0);
            end if;

          end if; -- clk
        end process;

      end block sdpblock;

      --
      -- deframe TLP
      --
      stpblock : block

        component scfifo
          generic (
            LPM_NUMWORDS            : natural := 2**DEFRAMER_TLPBUF_DEPTH;
            LPM_WIDTHU              : natural := DEFRAMER_TLPBUF_DEPTH;
            LPM_WIDTH               : natural := NLANE*18;
            LPM_SHOWAHEAD           : string  := "ON";
            ADD_RAM_OUTPUT_REGISTER : string  := "ON";
            OVERFLOW_CHECKING       : string  := "ON";
            UNDERFLOW_CHECKING      : string  := "OFF";
            USE_EAB                 : string  := "ON";
            INTENDED_DEVICE_FAMILY  : string  := "Stratix II GX";
            LPM_TYPE                : string  := "scfifo"
            );
          port (
            usedw    : out std_logic_vector(LPM_WIDTHU-1 downto 0);

            rdreq    : in  std_logic;
            full     : out std_logic;
            empty    : out std_logic;
            clock    : in  std_logic;
            sclr     : in  std_logic;
            q        : out std_logic_vector(LPM_WIDTH-1 downto 0);
            wrreq    : in  std_logic;
            data     : in  std_logic_vector(LPM_WIDTH-1 downto 0)
            );
        end component;

        type BUF_STATE is (
          IDLE,
          RECV
        );

        signal bufsm                  : BUF_STATE;
        signal rdcnt                  : std_logic;
        signal stpat0, stpat1         : std_logic_vector(NLANE/2-1 downto 0);
        signal endat0, endat1         : std_logic_vector(NLANE/2-1 downto 0);
        signal wetp0, wetp1, wetp2    : std_logic;
        signal tptop0, tptop1, tptop2 : std_logic;
        signal tpend0, tpend1         : std_logic;
        signal betp2                  : std_logic_vector(NLANE*2-1 downto 0); 
        signal tp0, tp1, tp2          : std_logic_vector(NLANE*16-1 downto 0); 

        signal bufin                  : std_logic_vector(NLANE*18-1 downto 0); 
        signal bufout                 : std_logic_vector(NLANE*18-1 downto 0); 
        signal wrreq                  : std_logic;
        signal rdreq                  : std_logic;
        signal rdack                  : std_logic;
        signal full                   : std_logic;
        signal empty                  : std_logic;
        signal usedw0                 : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);
        signal usedwmax0              : std_logic_vector(DEFRAMER_TLPBUF_DEPTH-1 downto 0);

        signal d                      : each16b(NLANE-1 downto 0);
        signal k                      : each2b(NLANE-1 downto 0);

        signal stp_at_0l_in, stp_at_4l_in, stp_at_0h_in, stp_at_4h_in  : std_logic;
        signal end_at_3l_in, end_at_7l_in, end_at_3h_in, end_at_7h_in  : std_logic;
        signal stp_at_any_lane_in : std_logic;
        signal end_at_any_lane_in : std_logic;

        signal stp_at_0l_out, stp_at_4l_out, stp_at_0h_out, stp_at_4h_out : std_logic;
        signal end_at_3l_out, end_at_7l_out, end_at_3h_out, end_at_7h_out : std_logic;
        signal stp_at_any_lane_out : std_logic;
        signal end_at_any_lane_out : std_logic;

-- synthesis translate_off
        attribute keep : boolean;
        attribute keep of usedwmax0 : signal is true;
-- synthesis translate_on

      begin

-- synthesis translate_off
        -- for debugging
        fullout  <= full;
        usedwmax <= usedwmax0;
-- synthesis translate_on

        tp    <= tp2;
        tptop <= tptop2;
        wetp  <= wetp2;
        betp  <= betp2;

        process (clk)
        begin
          if (rising_edge(clk)) then
            for i in 0 to NLANE-1 loop
              bufin(i*18+7 downto i*18+0)    <= data(i)(7 downto 0);
              bufin(i*18+8)                  <= datak(i)(0);
              bufin(i*18+16 downto i*18+9)   <= data(i)(15 downto 8);
              bufin(i*18+17)                 <= datak(i)(1);
            end loop;
          end if;
        end process;

        lane: for i in 0 to NLANE-1 generate
          d(i)(7 downto 0)               <= bufout(i*18+7 downto i*18+0);
          k(i)(0)                        <= bufout(i*18+8);
          d(i)(15 downto 8)              <= bufout(i*18+16 downto i*18+9);
          k(i)(1)                        <= bufout(i*18+17);
        end generate lane;

        stp_at_0l_in  <= '1' when datak(0)(0) = '1' and  data(0)( 7 downto 0) = K_STP else '0';
        stp_at_4l_in  <= '1' when datak(4)(0) = '1' and  data(4)( 7 downto 0) = K_STP else '0';
        stp_at_0h_in  <= '1' when datak(0)(1) = '1' and  data(0)(15 downto 8) = K_STP else '0';
        stp_at_4h_in  <= '1' when datak(4)(1) = '1' and  data(4)(15 downto 8) = K_STP else '0';

        end_at_3l_in  <= '1' when datak(3)(0) = '1' and (data(3)( 7 downto 0) = K_END or data(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_in  <= '1' when datak(7)(0) = '1' and (data(7)( 7 downto 0) = K_END or data(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_in  <= '1' when datak(3)(1) = '1' and (data(3)(15 downto 8) = K_END or data(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_in  <= '1' when datak(7)(1) = '1' and (data(7)(15 downto 8) = K_END or data(7)(15 downto 8) = K_EDB) else '0';

        stp_at_any_lane_in <= '1' when (stp_at_0l_in = '1' or stp_at_4l_in = '1' or stp_at_0h_in = '1' or stp_at_4h_in = '1') else '0';
        end_at_any_lane_in <= '1' when (end_at_3l_in = '1' or end_at_7l_in = '1' or end_at_3h_in = '1' or end_at_7h_in = '1') else '0';

        stp_at_0l_out <= '1' when k(0)(0) = '1'     and  d(0)( 7 downto 0) = K_STP else '0';
        stp_at_4l_out <= '1' when k(4)(0) = '1'     and  d(4)( 7 downto 0) = K_STP else '0';
        stp_at_0h_out <= '1' when k(0)(1) = '1'     and  d(0)(15 downto 8) = K_STP else '0';
        stp_at_4h_out <= '1' when k(4)(1) = '1'     and  d(4)(15 downto 8) = K_STP else '0';

        end_at_3l_out <= '1' when k(3)(0) = '1'     and (d(3)( 7 downto 0) = K_END or d(3)( 7 downto 0) = K_EDB) else '0';
        end_at_7l_out <= '1' when k(7)(0) = '1'     and (d(7)( 7 downto 0) = K_END or d(7)( 7 downto 0) = K_EDB) else '0';
        end_at_3h_out <= '1' when k(3)(1) = '1'     and (d(3)(15 downto 8) = K_END or d(3)(15 downto 8) = K_EDB) else '0';
        end_at_7h_out <= '1' when k(7)(1) = '1'     and (d(7)(15 downto 8) = K_END or d(7)(15 downto 8) = K_EDB) else '0';

        stp_at_any_lane_out <= '1' when (stp_at_0l_out = '1' or stp_at_4l_out = '1' or stp_at_0h_out = '1' or stp_at_4h_out = '1') else '0';
        end_at_any_lane_out <= '1' when (end_at_3l_out = '1' or end_at_7l_out = '1' or end_at_3h_out = '1' or end_at_7h_out = '1') else '0';

        rxbuf0 : scfifo
          port map (
            usedw    => usedw0,

            clock    => clk,
            sclr     => rst,

            wrreq    => wrreq,
            data     => bufin,

            rdreq    => rdreq,
            full     => full,
            empty    => empty,
            q        => bufout
            );

        process (clk, rst)
        begin
          if (rst = '1') then
            usedwmax0 <= (others => '0');
          elsif (rising_edge(clk)) then
            if (unsigned(usedwmax0) < unsigned(usedw0)) then
              usedwmax0 <= conv_std_logic_vector(0, usedwmax0'length - usedw0'length) & usedw0;
            end if;
          end if;
        end process;

        process (clk, rst)
        begin
          if (rst = '1') then
            wrreq <= '0';
            bufsm <= IDLE;
          elsif (rising_edge(clk)) then
            wrreq <= '0';
            case bufsm is
              when IDLE =>
                -- start buffering when see an STP.
                if (stp_at_any_lane_in = '1') then
                  wrreq <= '1';
                  bufsm <= RECV;
                end if;
              when RECV =>
                wrreq <= '1';
                -- stop buffering when see an END with no STP at any higher lane in the same clk.
                if (end_at_any_lane_in = '1' and stp_at_any_lane_in = '0') then
                  bufsm <= IDLE;
                end if;
              when others =>
            end case;
          end if;
        end process;

        -- stop rdreq 1-clk when see an END and an STP in the same clk.
        rdreq <= '1' when empty = '0' and rdack = '1' else '0';
        rdack <= '0' when end_at_any_lane_out = '1' and stp_at_any_lane_out = '1' and rdcnt = '0' else '1';

        process(clk, rst)
        begin
          if (rst = '1') then
            rdcnt <= '0';
          elsif (rising_edge(clk)) then
            if (rdcnt = '0') then
              if (end_at_any_lane_out = '1' and stp_at_any_lane_out = '1') then
                rdcnt <= '1';
              end if;
            else
              rdcnt <= '0';
            end if;
          end if;
        end process;


        process (clk)
          variable stpattmp   : std_logic_vector(NLANE/2-1 downto 0);
          variable endattmp   : std_logic_vector(NLANE/2-1 downto 0);
          variable setmp      : std_logic_vector(NLANE-1 downto 0);
          variable betptmp    : std_logic_vector(NLANE*2-1 downto 0);
        begin
          if (rising_edge(clk)) then
            --
            -- prestage
            --
            wetp0  <= '0';
            tptop0 <= '0';
            tpend0 <= '0';

            if (empty = '0') then
              wetp0 <= '1';

              stpattmp := stpat0;
              if (end_at_any_lane_out = '0' or rdcnt = '1') then
                -- assert tptop0 when see an STP with no END,
                -- or see combination of an STP and an END duplicated (i.e., rdcnt=1)
                -- by rxbuf0.
                if (stp_at_any_lane_out = '1') then
                  tptop0 <= '1';
                end if;

                -- modify alignment of TLP top if detect an STP.
                -- use current alignment otherwise.
                if (stp_at_0l_out = '1') then
                  stpattmp := "0001";
                elsif (stp_at_4l_out = '1') then
                  stpattmp := "0010";
                elsif (stp_at_0h_out = '1') then
                  stpattmp := "0100";
                elsif (stp_at_4h_out = '1') then
                  stpattmp := "1000";
                end if;
              end if;
              stpat0 <= stpattmp;

              endattmp := endat0;
              if (end_at_any_lane_out = '1' and rdcnt = '0') then
                -- assert tpend0 when see an END which is not duplicated (i.e., rdcnt=0)
                -- by rxbuf0.
                tpend0 <= '1';

                -- modify alignment of TLP end.
                if (end_at_3l_out = '1') then
                  endattmp := "0001";
                elsif (end_at_7l_out = '1') then
                  endattmp := "0010";
                elsif (end_at_3h_out = '1') then
                  endattmp := "0100";
                elsif (end_at_7h_out = '1') then
                  endattmp := "1000";
                end if;
              end if;
              endat0 <= endattmp;

            end if;

            case stpattmp is
              when "0001" =>
                tp0(119 downto 112) <= d(7)(15 downto 8);
                tp0(111 downto 104) <= d(6)(15 downto 8);
                tp0(103 downto  96) <= d(5)(15 downto 8);
                tp0( 95 downto  88) <= d(4)(15 downto 8);
                tp0( 87 downto  80) <= d(3)(15 downto 8);
                tp0( 79 downto  72) <= d(2)(15 downto 8);
                tp0( 71 downto  64) <= d(1)(15 downto 8);
                tp0( 63 downto  56) <= d(0)(15 downto 8);
                tp0( 55 downto  48) <= d(7)( 7 downto 0);
                tp0( 47 downto  40) <= d(6)( 7 downto 0);
                tp0( 39 downto  32) <= d(5)( 7 downto 0);
                tp0( 31 downto  24) <= d(4)( 7 downto 0);
                tp0( 23 downto  16) <= d(3)( 7 downto 0);
                tp0( 15 downto   8) <= d(2)( 7 downto 0);
                tp0(  7 downto   0) <= d(1)( 7 downto 0);

              when "0010" =>
                tp0( 87 downto  80) <= d(7)(15 downto 8);
                tp0( 79 downto  72) <= d(6)(15 downto 8);
                tp0( 71 downto  64) <= d(5)(15 downto 8);
                tp0( 63 downto  56) <= d(4)(15 downto 8);
                tp0( 55 downto  48) <= d(3)(15 downto 8);
                tp0( 47 downto  40) <= d(2)(15 downto 8);
                tp0( 39 downto  32) <= d(1)(15 downto 8);
                tp0( 31 downto  24) <= d(0)(15 downto 8);
                tp0( 23 downto  16) <= d(7)( 7 downto 0);
                tp0( 15 downto   8) <= d(6)( 7 downto 0);
                tp0(  7 downto   0) <= d(5)( 7 downto 0);

              when "0100" =>
                tp0( 55 downto  48) <= d(7)(15 downto 8);
                tp0( 47 downto  40) <= d(6)(15 downto 8);
                tp0( 39 downto  32) <= d(5)(15 downto 8);
                tp0( 31 downto  24) <= d(4)(15 downto 8);
                tp0( 23 downto  16) <= d(3)(15 downto 8);
                tp0( 15 downto   8) <= d(2)(15 downto 8);
                tp0(  7 downto   0) <= d(1)(15 downto 8);

              when "1000" =>
                tp0( 23 downto  16) <= d(7)(15 downto 8);
                tp0( 15 downto   8) <= d(6)(15 downto 8);
                tp0(  7 downto   0) <= d(5)(15 downto 8);

              when others =>
            end case;

            --
            -- stage 0
            --
            wetp1  <= wetp0;
            tptop1 <= tptop0;
            tpend1 <= tpend0;
            stpat1 <= stpat0;
            endat1 <= endat0;

            case stpat0 is
              when "0001" =>
                tp1(127 downto 120) <= d(0)(7 downto 0);
                tp1(119 downto   0) <= tp0(119 downto 0);

              when "0010" =>
                tp1(127 downto 120) <= d(4)(7 downto 0);
                tp1(119 downto 112) <= d(3)(7 downto 0);
                tp1(111 downto 104) <= d(2)(7 downto 0);
                tp1(103 downto  96) <= d(1)(7 downto 0);
                tp1( 95 downto  88) <= d(0)(7 downto 0);
                tp1( 87 downto   0) <= tp0( 87 downto 0);

              when "0100" =>
                tp1(127 downto 120) <= d(0)(15 downto 8);
                tp1(119 downto 112) <= d(7)(7 downto 0);
                tp1(111 downto 104) <= d(6)(7 downto 0);
                tp1(103 downto  96) <= d(5)(7 downto 0);
                tp1( 95 downto  88) <= d(4)(7 downto 0);
                tp1( 87 downto  80) <= d(3)(7 downto 0);
                tp1( 79 downto  72) <= d(2)(7 downto 0);
                tp1( 71 downto  64) <= d(1)(7 downto 0);
                tp1( 63 downto  56) <= d(0)(7 downto 0);
                tp1( 55 downto   0) <= tp0(55 downto 0);

              when "1000" =>
                tp1(127 downto 120) <= d(4)(15 downto 8);
                tp1(119 downto 112) <= d(3)(15 downto 8);
                tp1(111 downto 104) <= d(2)(15 downto 8);
                tp1(103 downto  96) <= d(1)(15 downto 8);
                tp1( 95 downto  88) <= d(0)(15 downto 8);
                tp1( 87 downto  80) <= d(7)(7 downto 0);
                tp1( 79 downto  72) <= d(6)(7 downto 0);
                tp1( 71 downto  64) <= d(5)(7 downto 0);
                tp1( 63 downto  56) <= d(4)(7 downto 0);
                tp1( 55 downto  48) <= d(3)(7 downto 0);
                tp1( 47 downto  40) <= d(2)(7 downto 0);
                tp1( 39 downto  32) <= d(1)(7 downto 0);
                tp1( 31 downto  24) <= d(0)(7 downto 0);
                tp1( 23 downto   0) <= tp0(23 downto 0);

              when others =>
            end case;

            --
            -- stage 1
            --
            wetp2   <= wetp1;
            tptop2  <= tptop1;
            tp2     <= tp1;

            betptmp := (others => '1');
            if (tpend0 = '1') then
              setmp := stpat0 & endat0;
              case setmp is
                when "0010" & "0001" =>
                  betptmp := x"3fff";
                when "0100" & "0001" =>
                  betptmp := x"03ff";
                when "0100" & "0010" =>
                  betptmp := x"3fff";

                when "1000" & "0001" =>
                  betptmp := x"003f";
                when "1000" & "0010" =>
                  betptmp := x"03ff";
                when "1000" & "0100" =>
                  betptmp := x"3fff";

                when others =>
              end case;
            end if;

            if (tpend1 = '1') then
              setmp := stpat1 & endat1;
              case setmp is
                when "0001" & "0001" =>
                  betptmp := x"0003";
                when "0001" & "0010" =>
                  betptmp := x"003f";
                when "0001" & "0100" =>
                  betptmp := x"03ff";
                when "0001" & "1000" =>
                  betptmp := x"3fff";

                when "0010" & "0010" =>
                  betptmp := x"0003";
                when "0010" & "0100" =>
                  betptmp := x"003f";
                when "0010" & "1000" =>
                  betptmp := x"03ff";

                when "0100" & "0100" =>
                  betptmp := x"0003";
                when "0100" & "1000" =>
                  betptmp := x"003f";

                when "1000" & "1000" =>
                  betptmp := x"0003";

                when others =>
                  wetp2 <= '0';
              end case;
            end if;
            betp2 <= betptmp;

            -- mask tp2
            for i in 0 to NLANE*2-1 loop
              if (betptmp(i) = '0') then
                tp2(8*i+7 downto 8*i+0) <= x"00";
              end if;
            end loop;

          end if; -- clk
        end process;

      end block stpblock;
    end block x8block;
  end generate x8;
  
end default;


-- 
-- Framer
-- encode a Data-Link-Layer-Packet (DLLP) or a Transaction-Layer-Packet (TLP)
-- into a Physical-Layer Packet (PLP).
-- 
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity framer is
  generic (
    NLANE       : integer
  );
  port (
    -- DLLP input
    emptydp     : in  std_logic;
    rdreqdp     : out std_logic;
    dp          : in  std_logic_vector(47 downto 0);

    -- TLP input
    reqtp       : in  std_logic;
    acktp       : out std_logic;
    toptp       : in  std_logic;
    wetp        : in  std_logic;
    betp        : in  std_logic_vector(NLANE*2-1 downto 0);
    tp          : in  std_logic_vector(NLANE*16-1 downto 0);

    -- PLP output
    data        : out each16b(NLANE-1 downto 0);
    datak       : out each2b(NLANE-1 downto 0);

    rst         : in  std_logic;        -- active high
    clk         : in  std_logic
  );
end framer;

architecture default of framer is

  component count_skp
    port (
      dosend      : out std_logic;
      sent        : in  std_logic;
      rst         : in  std_logic;        -- active high
      clk         : in  std_logic
    );
  end component;

  signal doskp         : std_logic;
  signal sentskp       : std_logic;

begin

  x1 : if (NLANE = 1) generate

    x1block : block

      type FRAMER_STATE is (
        IDLE,
        SEND_SKP,
        SEND_SDP_TYPE,
        SEND_D0_D1,
        SEND_D2_C0,
        SEND_C1_END,
        WAIT_TLP,
        SEND_TLP,
        INVALID_STATE
      );

      signal framersm      : FRAMER_STATE;
      signal dp1           : std_logic_vector(47 downto 0);
      signal tp1           : std_logic_vector(NLANE*16-1 downto 0);
      signal ntpcoming     : std_logic_vector(7 downto 0);  -- # of TLPs coming in.
      signal acktp0        : std_logic;

    begin

      acktp <= acktp0;

      -- accept TLP request, whenever there's no request for SKP-OS nor DLLP.
      process (reqtp, doskp, emptydp, framersm)
      begin
        acktp0 <= '0';
        if (reqtp = '1') then
          if (doskp = '0' and emptydp = '1' and
              (framersm = IDLE or framersm = SEND_TLP or framersm = WAIT_TLP)) then
            acktp0 <= '1';
          end if;
        end if;
      end process;

      process (clk, rst)
        variable ntpcomingtmp : std_logic_vector(ntpcoming'range);
      begin

        if (rst = '1') then

          rdreqdp   <= '0';
          ntpcoming <= (others => '0');
          framersm  <= IDLE;

        elsif (rising_edge(clk)) then

          rdreqdp  <= '0';
          tp1      <= tp;

          ntpcomingtmp := ntpcoming;

          -- transmit logical idle, if nothing to do.
          datak(0) <= (others => '0');
          data(0)  <= (others => '0');

          -- count up the num of TLP coming in, whenever acktp0 is asserted in the previous clk.
          if (acktp0 = '1') then
            ntpcomingtmp := unsigned(ntpcomingtmp) + 1;
          end if;

          case framersm is

            when IDLE =>            
              --
              -- transmission priority:
              --   1) TLP (which cannot be delayed)
              --   2) SKP-OS
              --   3) DLLP
              --   4) logical idle
              --

              -- TLP is coming in. should not start SKP-OS nor DLLP.
              if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                framersm <= WAIT_TLP;
              elsif (doskp = '1') then    -- issue SKP os.
                sentskp <= '1';
                datak(0) <= "11";
                data(0)  <= K_SKP & K_COM;
                framersm <= SEND_SKP;
              elsif (emptydp = '0') then  -- transmit a DLLP in the queue.
                rdreqdp <= '1';
                framersm <= SEND_SDP_TYPE;
              end if;

            when SEND_SKP =>
              sentskp <= '0';
              datak(0) <= "11";
              data(0)  <= K_SKP & K_SKP;
              framersm <= IDLE;

            when SEND_SDP_TYPE =>
              dp1      <= dp;             -- latch dp.
              datak(0) <= "01";
              data(0)  <= dp(7 downto 0) & K_SDP;  -- output type & SDP
              framersm <= SEND_D0_D1;

            when SEND_D0_D1 =>
              datak(0) <= "00";
              data(0)  <= dp1(23 downto 8);  -- output data1 & data0
              framersm <= SEND_D2_C0;

            when SEND_D2_C0 =>
              datak(0) <= "00";
              data(0)  <= dp1(39 downto 24);  -- output crc high & data2
              framersm <= SEND_C1_END;

            when SEND_C1_END =>
              datak(0) <= "10";
              data(0)  <= K_END & dp1(47 downto 40);  -- output END & crc low
              framersm <= IDLE;

            when WAIT_TLP =>
              if (wetp = '1') then
                datak(0) <= "01";
                data(0) <= tp(7 downto 0) & K_STP;
                framersm <= SEND_TLP;
              end if;

            when SEND_TLP =>
              if (wetp = '0') then
                datak(0) <= "10";
                data(0) <= K_END & tp1(15 downto 8);
                framersm <= IDLE;

                -- pass through IDLE state if next TLP is going to arrive.
                if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                  ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                  framersm <= WAIT_TLP;
                end if;

              else
                datak(0) <= "00";
                data(0) <= tp(7 downto 0) & tp1(15 downto 8);
              end if;

            when others =>
              framersm <= IDLE;

          end case;

        end if;

        ntpcoming <= ntpcomingtmp;

      end process;  

    end block x1block;

  end generate x1;


  x4 : if (NLANE >= 4) generate

    x4block : block

      type FRAMER_STATE is (
        IDLE,
        SEND_SKP,
        SEND_DP,
        WAIT_TLP,
        SEND_TLP,
        INVALID_STATE
      );

      signal framersm      : FRAMER_STATE;
      signal rdreqdp0      : std_logic;
      signal dp1           : std_logic_vector(47 downto 0);
      signal toptp1        : std_logic;
      signal wetp1         : std_logic;
      signal betp1         : std_logic_vector(NLANE*2-1 downto 0);
      signal tp1           : std_logic_vector(NLANE*16-1 downto 0);
      signal tp2           : std_logic_vector(NLANE*16-1 downto 0);
      signal ntpcoming     : std_logic_vector(7 downto 0);  -- # of TLPs coming in.
      signal acktp0        : std_logic;

    begin

      rdreqdp <= rdreqdp0 and not emptydp;
      acktp <= acktp0;

      -- accept TLP request, whenever there's no request for SKP-OS nor DLLP.
      process (reqtp, doskp, emptydp, framersm)
      begin
        acktp0 <= '0';
        if (reqtp = '1') then
          if (doskp = '0' and emptydp = '1' and
              (framersm = IDLE or framersm = SEND_TLP or framersm = WAIT_TLP)) then  -- !!
            acktp0 <= '1';
          end if;
        end if;
      end process;

      process (clk, rst)
        variable ntpcomingtmp : std_logic_vector(ntpcoming'range);
      begin

        if (rst = '1') then

          rdreqdp0  <= '0';
          ntpcoming <= (others => '0');
          framersm  <= IDLE;

        elsif (rising_edge(clk)) then

          rdreqdp0 <= '0';
          toptp1   <= toptp;
          wetp1    <= wetp;
          betp1    <= betp;
          tp1      <= tp;
          tp2      <= tp1;

          ntpcomingtmp := ntpcoming;

          -- transmit logical idle, if nothing to do.
          for i in 0 to NLANE-1 loop
            datak(i) <= (others => '0');
            data(i)  <= (others => '0');
          end loop;

          -- count up the num of TLP coming in, whenever acktp0 is asserted in the previous clk.
          if (acktp0 = '1') then
            ntpcomingtmp := unsigned(ntpcomingtmp) + 1;
          end if;

          case framersm is

            when IDLE =>            
              --
              -- transmission priority:
              --   1) TLP (which cannot be delayed)
              --   2) SKP-OS
              --   3) DLLP
              --   4) logical idle
              --

              -- TLP is coming in. should not start SKP-OS nor DLLP.
              if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                framersm <= WAIT_TLP;
              elsif (doskp = '1') then    -- issue SKP os.
                sentskp <= '1';
                for i in 0 to NLANE-1 loop
                  datak(i) <= "11";
                  data(i)  <= K_SKP & K_COM;
                end loop;
                framersm <= SEND_SKP;
              elsif (emptydp = '0') then  -- transmit a DLLP in the queue.
                rdreqdp0 <= '1';
                framersm <= SEND_DP;
              end if;

            when SEND_SKP =>
              sentskp <= '0';
              for i in 0 to NLANE-1 loop
                datak(i) <= "11";
                data(i)  <= K_SKP & K_SKP;
              end loop;
              framersm <= IDLE;

            when SEND_DP =>
              if (emptydp = '0') then
                if (NLANE = 4) then
                  datak(3)(1) <= '1';
                  datak(0)(0) <= '1';
                  data(3)(15 downto  8) <= K_END;
                  data(2)(15 downto  8) <= dp(47 downto 40);  -- crc0
                  data(1)(15 downto  8) <= dp(39 downto 32);  -- crc0
                  data(0)(15 downto  8) <= dp(31 downto 24);  -- data2
                  data(3)( 7 downto  0) <= dp(23 downto 16);  -- data1
                  data(2)( 7 downto  0) <= dp(15 downto  8);  -- data0
                  data(1)( 7 downto  0) <= dp( 7 downto  0);  -- type
                  data(0)( 7 downto  0) <= K_SDP;
                end if;

                if (NLANE = 8) then
                  datak(7)(0) <= '1';
                  datak(0)(0) <= '1';
                  data(7)( 7 downto  0) <= K_END;
                  data(6)( 7 downto  0) <= dp(47 downto 40);  -- crc0
                  data(5)( 7 downto  0) <= dp(39 downto 32);  -- crc0
                  data(4)( 7 downto  0) <= dp(31 downto 24);  -- data2
                  data(3)( 7 downto  0) <= dp(23 downto 16);  -- data1
                  data(2)( 7 downto  0) <= dp(15 downto  8);  -- data0
                  data(1)( 7 downto  0) <= dp( 7 downto  0);  -- type
                  data(0)( 7 downto  0) <= K_SDP;
                end if;

                rdreqdp0 <= '1';
              else  -- no more DLLP remains in the queue.
                framersm <= IDLE;
              end if;

            when WAIT_TLP =>
              if (toptp1 = '1') then  -- the 1st word.
                datak(0)(0) <= '1';

                if (NLANE = 4) then
                  data(3)(15 downto  8) <= tp1(6*8+7 downto 6*8+0);
                  data(2)(15 downto  8) <= tp1(5*8+7 downto 5*8+0);
                  data(1)(15 downto  8) <= tp1(4*8+7 downto 4*8+0);
                  data(0)(15 downto  8) <= tp1(3*8+7 downto 3*8+0);
                  data(3)( 7 downto  0) <= tp1(2*8+7 downto 2*8+0);
                  data(2)( 7 downto  0) <= tp1(1*8+7 downto 1*8+0);
                  data(1)( 7 downto  0) <= tp1(0*8+7 downto 0*8+0);
                  data(0)( 7 downto  0) <= K_STP;
                end if;

                if (NLANE = 8) then
                  data(7)(15 downto  8) <= tp1(14*8+7 downto 14*8+0);
                  data(6)(15 downto  8) <= tp1(13*8+7 downto 13*8+0);
                  data(5)(15 downto  8) <= tp1(12*8+7 downto 12*8+0);
                  data(4)(15 downto  8) <= tp1(11*8+7 downto 11*8+0);
                  data(3)(15 downto  8) <= tp1(10*8+7 downto 10*8+0);
                  data(2)(15 downto  8) <= tp1( 9*8+7 downto  9*8+0);
                  data(1)(15 downto  8) <= tp1( 8*8+7 downto  8*8+0);
                  data(0)(15 downto  8) <= tp1( 7*8+7 downto  7*8+0);
                  data(7)( 7 downto  0) <= tp1( 6*8+7 downto  6*8+0);
                  data(6)( 7 downto  0) <= tp1( 5*8+7 downto  5*8+0);
                  data(5)( 7 downto  0) <= tp1( 4*8+7 downto  4*8+0);
                  data(4)( 7 downto  0) <= tp1( 3*8+7 downto  3*8+0);
                  data(3)( 7 downto  0) <= tp1( 2*8+7 downto  2*8+0);
                  data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                  data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                  data(0)( 7 downto  0) <= K_STP;
                end if;

                framersm <= SEND_TLP;
              end if;

            when SEND_TLP =>
              -- the last word.
              if (wetp = '0' or
                  (wetp1 = '1' and toptp = '1')) then

                if (NLANE = 4) then
                  if (betp1(5) = '1') then
                    datak(3)(1) <= '1';  -- put K_END on higher lane3.
                    data(3)(15 downto  8) <= K_END;
                    data(2)(15 downto  8) <= tp1(5*8+7 downto 5*8+0);
                    data(1)(15 downto  8) <= tp1(4*8+7 downto 4*8+0);
                    data(0)(15 downto  8) <= tp1(3*8+7 downto 3*8+0);
                    data(3)( 7 downto  0) <= tp1(2*8+7 downto 2*8+0);
                    data(2)( 7 downto  0) <= tp1(1*8+7 downto 1*8+0);
                    data(1)( 7 downto  0) <= tp1(0*8+7 downto 0*8+0);
                    data(0)( 7 downto  0) <= tp2(7*8+7 downto 7*8+0);
                  else
                    datak(3)(0) <= '1';  -- put K_END on lower lane3.
                    data(3)(15 downto  8) <= (others => '0');
                    data(2)(15 downto  8) <= (others => '0');
                    data(1)(15 downto  8) <= (others => '0');
                    data(0)(15 downto  8) <= (others => '0');
                    data(3)( 7 downto  0) <= K_END;
                    data(2)( 7 downto  0) <= tp1(1*8+7 downto 1*8+0);
                    data(1)( 7 downto  0) <= tp1(0*8+7 downto 0*8+0);
                    data(0)( 7 downto  0) <= tp2(7*8+7 downto 7*8+0);
                  end if;
                end if;

                if (NLANE = 8) then
                  if (betp1(13) = '1') then
                    datak(7)(1) <= '1';  -- put K_END on higher lane7.
                    data(7)(15 downto  8) <= K_END;
                    data(6)(15 downto  8) <= tp1(13*8+7 downto 13*8+0);
                    data(5)(15 downto  8) <= tp1(12*8+7 downto 12*8+0);
                    data(4)(15 downto  8) <= tp1(11*8+7 downto 11*8+0);
                    data(3)(15 downto  8) <= tp1(10*8+7 downto 10*8+0);
                    data(2)(15 downto  8) <= tp1( 9*8+7 downto  9*8+0);
                    data(1)(15 downto  8) <= tp1( 8*8+7 downto  8*8+0);
                    data(0)(15 downto  8) <= tp1( 7*8+7 downto  7*8+0);
                    data(7)( 7 downto  0) <= tp1( 6*8+7 downto  6*8+0);
                    data(6)( 7 downto  0) <= tp1( 5*8+7 downto  5*8+0);
                    data(5)( 7 downto  0) <= tp1( 4*8+7 downto  4*8+0);
                    data(4)( 7 downto  0) <= tp1( 3*8+7 downto  3*8+0);
                    data(3)( 7 downto  0) <= tp1( 2*8+7 downto  2*8+0);
                    data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                    data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                    data(0)( 7 downto  0) <= tp2(15*8+7 downto 15*8+0);
                  elsif (betp1(9) = '1') then
                    datak(7)(1) <= '1';
                    datak(6)(1) <= '1';
                    datak(5)(1) <= '1';
                    datak(4)(1) <= '1';
                    datak(3)(1) <= '1';  -- put K_END on higher lane3.
                    data(7)(15 downto  8) <= K_PAD;
                    data(6)(15 downto  8) <= K_PAD;
                    data(5)(15 downto  8) <= K_PAD;
                    data(4)(15 downto  8) <= K_PAD;
                    data(3)(15 downto  8) <= K_END;
                    data(2)(15 downto  8) <= tp1( 9*8+7 downto  9*8+0);
                    data(1)(15 downto  8) <= tp1( 8*8+7 downto  8*8+0);
                    data(0)(15 downto  8) <= tp1( 7*8+7 downto  7*8+0);
                    data(7)( 7 downto  0) <= tp1( 6*8+7 downto  6*8+0);
                    data(6)( 7 downto  0) <= tp1( 5*8+7 downto  5*8+0);
                    data(5)( 7 downto  0) <= tp1( 4*8+7 downto  4*8+0);
                    data(4)( 7 downto  0) <= tp1( 3*8+7 downto  3*8+0);
                    data(3)( 7 downto  0) <= tp1( 2*8+7 downto  2*8+0);
                    data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                    data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                    data(0)( 7 downto  0) <= tp2(15*8+7 downto 15*8+0);
                  elsif (betp1(5) = '1') then
                    datak(7)(0) <= '1';  -- put K_END on lower lane7.
                    data(7)( 7 downto  0) <= K_END;
                    data(6)( 7 downto  0) <= tp1( 5*8+7 downto  5*8+0);
                    data(5)( 7 downto  0) <= tp1( 4*8+7 downto  4*8+0);
                    data(4)( 7 downto  0) <= tp1( 3*8+7 downto  3*8+0);
                    data(3)( 7 downto  0) <= tp1( 2*8+7 downto  2*8+0);
                    data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                    data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                    data(0)( 7 downto  0) <= tp2(15*8+7 downto 15*8+0);
                  else
                    datak(7)(0) <= '1';
                    datak(6)(0) <= '1';
                    datak(5)(0) <= '1';
                    datak(4)(0) <= '1';
                    datak(3)(0) <= '1';  -- put K_END on lower lane3.
                    data(7)( 7 downto  0) <= K_PAD;
                    data(6)( 7 downto  0) <= K_PAD;
                    data(5)( 7 downto  0) <= K_PAD;
                    data(4)( 7 downto  0) <= K_PAD;
                    data(3)( 7 downto  0) <= K_END;
                    data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                    data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                    data(0)( 7 downto  0) <= tp2(15*8+7 downto 15*8+0);
                  end if;
                end if;

                framersm <= IDLE;
#if 1
                -- pass through IDLE state if next TLP is going to arrive.
                if (ntpcoming /= conv_std_logic_vector(0, ntpcoming'length)) then
                  ntpcomingtmp := unsigned(ntpcomingtmp) - 1;
                  framersm <= WAIT_TLP;
                end if;
#endif
              else

                if (NLANE = 4) then
                  data(3)(15 downto  8) <= tp1(6*8+7 downto 6*8+0);
                  data(2)(15 downto  8) <= tp1(5*8+7 downto 5*8+0);
                  data(1)(15 downto  8) <= tp1(4*8+7 downto 4*8+0);
                  data(0)(15 downto  8) <= tp1(3*8+7 downto 3*8+0);
                  data(3)( 7 downto  0) <= tp1(2*8+7 downto 2*8+0);
                  data(2)( 7 downto  0) <= tp1(1*8+7 downto 1*8+0);
                  data(1)( 7 downto  0) <= tp1(0*8+7 downto 0*8+0);
                  data(0)( 7 downto  0) <= tp2(7*8+7 downto 7*8+0);
                end if;

                if (NLANE = 8) then
                  data(7)(15 downto  8) <= tp1(14*8+7 downto 14*8+0);
                  data(6)(15 downto  8) <= tp1(13*8+7 downto 13*8+0);
                  data(5)(15 downto  8) <= tp1(12*8+7 downto 12*8+0);
                  data(4)(15 downto  8) <= tp1(11*8+7 downto 11*8+0);
                  data(3)(15 downto  8) <= tp1(10*8+7 downto 10*8+0);
                  data(2)(15 downto  8) <= tp1( 9*8+7 downto  9*8+0);
                  data(1)(15 downto  8) <= tp1( 8*8+7 downto  8*8+0);
                  data(0)(15 downto  8) <= tp1( 7*8+7 downto  7*8+0);
                  data(7)( 7 downto  0) <= tp1( 6*8+7 downto  6*8+0);
                  data(6)( 7 downto  0) <= tp1( 5*8+7 downto  5*8+0);
                  data(5)( 7 downto  0) <= tp1( 4*8+7 downto  4*8+0);
                  data(4)( 7 downto  0) <= tp1( 3*8+7 downto  3*8+0);
                  data(3)( 7 downto  0) <= tp1( 2*8+7 downto  2*8+0);
                  data(2)( 7 downto  0) <= tp1( 1*8+7 downto  1*8+0);
                  data(1)( 7 downto  0) <= tp1( 0*8+7 downto  0*8+0);
                  data(0)( 7 downto  0) <= tp2(15*8+7 downto 15*8+0);
                end if;

              end if;

            when others =>
              framersm <= IDLE;

          end case;

        end if;

        ntpcoming <= ntpcomingtmp;

      end process;  

    end block x4block;

  end generate x4;

  count_skp0 : count_skp
    port map (
      dosend => doskp,
      sent   => sentskp,
      rst    => rst,
      clk    => clk
    );

end default;


--
-- count the num of SKP ordered sets to be sent.
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.gpciepkg.all;

entity count_skp is
  port (
    dosend      : out std_logic;
    sent        : in  std_logic;
    rst         : in  std_logic;        -- active high
    clk         : in  std_logic
  );
end count_skp;

architecture default of count_skp is

  signal nskpleft    : std_logic_vector(7 downto 0);  -- # of SKP-os to be sent.
  signal timer       : std_logic_vector(9 downto 0);

begin

  process (clk, rst)
    variable nskplefttmp : std_logic_vector(7 downto 0);
  begin
    if (rst = '1') then
      nskpleft <= (others => '0');
      timer <= (others => '0');
    elsif (rising_edge(clk)) then
      -- SKP os should be transmitted in every 700clk (1400 symbols, 0x2bc clk).
      nskplefttmp := nskpleft;
      if (timer(9 downto 0) = "1010111100") then
        timer <= (others => '0');
        nskplefttmp := unsigned(nskplefttmp) + 1;
      else
        timer <= unsigned(timer) + 1;
      end if;

      if (sent = '1') then
        if (nskplefttmp /= conv_std_logic_vector(0, nskplefttmp'length)) then
          nskplefttmp := unsigned(nskplefttmp) - 1;
        end if;
      end if;
      nskpleft <= nskplefttmp;

      if (nskplefttmp = conv_std_logic_vector(0, nskplefttmp'length)) then
        dosend <= '0';
      else
        dosend <= '1';
      end if;
    end if;
  end process;

end default;
